/*
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Sun May 27 19:37:19 EDT 2007
 *
 * Will use the estimator library to connect to the estimator at source:port
 * and request avail-bandwidth measurements to estimator on dest:port.
 */

#include <stdio.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <unistd.h>

#include "mgrp/bhtp.h"
#include "mgrp/udpex.h"

/* libcommon */
#include "alog.h"
#include "net.h"
#include "tick.h"

#include "libestimator.h"

struct global_state {
	int mgrp;
	int iterate;
	int iterate_interval_sec;
	int useprobestart;
	struct sockaddr_in addr_src;
	struct sockaddr_in addr_dst;
} gb = {
	.mgrp	= 0,
	.iterate_interval_sec = 5,
};

void usage(void);
void parse_cmdline(int argc, char **argv);

void
avail_bw(void)
{
	int rc = 0;
	struct est_channel ch;
	struct est_results results;
	int final = 0;
	int replies = 0;
	void *session = NULL;

	rc = est_open_channel(&ch, &gb.addr_dst);
	if (rc)
		goto error;

	LOGFN("opened channel");

	rc = est_create_pathload_probe_session(&ch, &gb.addr_src,
				gb.iterate, gb.iterate_interval_sec,
                                gb.mgrp, 0, 0,
				!gb.useprobestart, &session);
	if (rc) {
		LOGTR("failed to create pathload probe session");
		goto error;
	}

	if (gb.useprobestart) {
		/* session was not started when it was created */
		rc = est_start_pathload_probe_session(&ch, session);
		if (rc) {
			LOGTR("failed to start pathload probe session");
			goto error;
		}
	}

	LOGFN("probing started between endpoints");

	do {
		rc = est_receive_probe_result(&ch, &results, &final);
		if (rc) {
			LOGTR("PROBE reply failed");
			goto error;
		}
		
		LOG(">>>>>>>>>>>>> GOT PROBE REPLY %d", ++replies);
		EST_LOGRESULTS(&results);

	} while (!final);

	est_close_channel(&ch);
	return;

error:
	LOGTR("libestimator failed: %s", est_strerror(rc));
	est_close_channel(&ch);
	return;
}

int
main (int argc, char **argv)
{
	parse_cmdline(argc, argv);
	avail_bw();
	return 0;
}

void
usage(void)
{
	printf("USAGE\n");
	printf("\tavbw -s <src> -d <dst> [options]\n");
	printf("\n");
	printf("OPTIONS\n");
	printf("\t-x         use MGRP for transport\n");
	printf("\t-i         iterate\n");
	printf("\n");
	exit(1);
}

void
check_params(void)
{
	if (net_addr_empty(&gb.addr_src))
		usage();

	if (net_addr_empty(&gb.addr_dst))
		usage();
}

void
parse_cmdline(int argc, char **argv)
{
	char short_options[] = "hs:d:xiP";

	char c = '?';
	int found = 0;
	char error[200];
	char buf[100];
	struct sockaddr_in *addr = NULL;

	/* parse command line options */

	while (1) {

		c = getopt(argc, argv, short_options);

		if (c == EOF)
		    break;

		switch (c) {

		case 's':
			addr = &gb.addr_src;
			found = net_resolve_host_port(optarg, addr,
					error, sizeof(error));
			ERR_ON(!found, "failed to resolve src host %s", optarg);
			LOG("src %s (%s:%d)",
				optarg,
				inet_ntop(addr->sin_family, &addr->sin_addr,
					buf, sizeof(buf)),
				ntohs(addr->sin_port));
			break;

		case 'd':
			addr = &gb.addr_dst;
			found = net_resolve_host_port(optarg, addr,
					error, sizeof(error));
			ERR_ON(!found, "failed to resolve dst host %s", optarg);
			LOG("dst %s (%s:%d)",
				optarg,
				inet_ntop(addr->sin_family, &addr->sin_addr,
					buf, sizeof(buf)),
				ntohs(addr->sin_port));
			break;

		case 'P':
			gb.useprobestart = 1;
			break;

		case 'x':
			gb.mgrp = 1;
			break;

		case 'i':
			gb.iterate = 1;
			break;

		case 'h':
			usage();
			exit(0);
			break;
		    
		case '?':
			break;
		    
		default:
			break;
		}
	}

	check_params();
}

