/*
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Sun May 27 19:37:19 EDT 2007
 *
 * Will use the estimator library to connect to the estimator at source:port
 * and request avail-bandwidth measurements to estimator on dest:port.
 */

#include <stdio.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <unistd.h>

#include "mgrp/bhtp.h"
#include "mgrp/udpex.h"

/* libcommon */
#include "alog.h"
#include "net.h"
#include "tick.h"
#include "cmp.h"
#include "parse.h"

#include "libestimator.h"

#define MAX_SESSIONS	10

struct global_state {
        struct timeval start_time;

	int mgrp;
        int piggyback;
        int fast;
	int iterate;
	int iterate_sec;
	int iterate_interval_sec;
	struct sockaddr_in addr_src[MAX_SESSIONS];
	struct sockaddr_in addr_dst[MAX_SESSIONS];

	int sessions;
	struct est_probe_session *ses[MAX_SESSIONS];
	char names[MAX_SESSIONS][200];

	int fdmax;
	fd_set fdreads;
	
	int replies;
} gb = {
	.mgrp	= 0,
        .iterate_interval_sec = 5,
};

void usage(void);
void parse_cmdline(int argc, char **argv);

static inline void
include_rfd(int fd)
{
	gb.fdmax = cmp_max(gb.fdmax, fd + 1);
	FD_SET(fd, &gb.fdreads);
}

static inline void
exclude_rfd(int fd)
{
	FD_CLR(fd, &gb.fdreads);
	/* find the new max */
	while (gb.fdmax > 0) {
		if (FD_ISSET(gb.fdmax - 1, &gb.fdreads))
			break;
		gb.fdmax--;
	}
}



void
collect_replies(void)
{
	int i = 0;
	int rc = 0;
	int timeout = 0;
	int interrupt = 0;
	int fdready = 0;
	int total_fdready = 0;
	fd_set rfds = gb.fdreads;

	rc = select(gb.fdmax, &rfds, NULL, NULL, NULL);

	timeout = (rc == 0);
	interrupt = (rc == -1 && errno == EINTR);
	fdready = (rc > 0 ? rc : 0);

	total_fdready = fdready;

	for (i = 0; i < gb.sessions && fdready > 0; i++) {
		int final = 0;
		struct est_results results;
		int sock = est_probe_session_socket(gb.ses[i]);
		
		if (!FD_ISSET(sock, &rfds))
			continue;
		
		rc = est_next_probe_session_estimate(gb.ses[i], &results, &final);
		/* remove the channel on error */
		ERR_ON(rc, "next estimate error: %d", rc);

		LOG(">>>>>>>>>>>>> [%s] GOT PROBE REPLY %d",
			gb.names[i], ++gb.replies);
		EST_LOGRESULTS(&results);

		fdready--;
	}

	if (fdready > 0) {
		ERR("failed to process all %d sockets: %d left",
			total_fdready, fdready);
	}
}

static inline int
finished()
{
        struct timeval now;
        long long usec_diff = 0;

        if (!gb.iterate)
                return 1;

        // no limit set; iterate forever
        if (gb.iterate_sec == 0)
                return 0;

        gettimeofday(&now, NULL);
        usec_diff = atimer_timeval_diff_usec(&now, &gb.start_time);
        if (usec_diff >= (gb.iterate_sec * ATIMER_USEC_PER_SEC))
                return 1;
        
        return 0;
}

void
avail_bw2(void)
{
	int i = 0;
	int rc = 0;

	gb.fdmax = 0;
	FD_ZERO(&gb.fdreads);

	/* create probe sessions; probing does not start yet */

	for (i = 0; i < gb.sessions; i++) {
		gb.ses[i] = est_create_probe_session(
				&gb.addr_src[i],
				&gb.addr_dst[i],
				gb.iterate, gb.iterate_interval_sec,
                                gb.mgrp, gb.piggyback, gb.fast);
		ERR_ON(gb.ses[i] == NULL, "session %d failed", i);
		include_rfd(est_probe_session_socket(gb.ses[i]));
	}

	/* start probing */

	for (i = 0; i < gb.sessions; i++) {
		rc = est_start_probe_session(gb.ses[i]);
		ERR_ON(rc, "failed to start session %d", i);
	}

	/* wait and collect replies */

        gettimeofday(&gb.start_time, NULL);
        do {
		collect_replies();
	} while(!finished());

	/* destroy probe sessions */

	for (i = 0; i < gb.sessions; i++) {
		exclude_rfd(est_probe_session_socket(gb.ses[i]));
		est_destroy_probe_session(gb.ses[i]);
	}


}

int
main (int argc, char **argv)
{
	parse_cmdline(argc, argv);
	avail_bw2();
	return 0;
}

void
usage(void)
{
	printf("USAGE\n");
	printf("\tavbw -P <src1,dst1> [-P <src2,dst2>] [options]\n");
	printf("\n");
	printf("Starts pathload sessions between src-dst pairs.\n");
	printf("The 'estimator' process must run on the specified\n");
	printf("ports on every host for this to work.\n");
	printf("\n");
	printf("OPTIONS\n");
	printf("\t-x         use MGRP for transport\n");
	printf("\t-i         iterate forever\n");
	printf("\t-t <sec>   iterate for <sec> seconds\n");
	printf("\t-p         piggyback\n");
	printf("\n");
	exit(1);
}

void
check_params(void)
{
	if (gb.sessions <= 0)
		usage();
}

void
parse_host_pair(const char *optstr)
{
	char tmp[100];
	int j = gb.sessions++;

	strncpy(gb.names[j], optarg, sizeof(gb.names[j]));

	parse_two_hosts_with_ports(optarg,
			&gb.addr_src[j],
			&gb.addr_dst[j],
			EST_DEFAULT_CONTROL_TCP_PORT);

	net_srcdst_str(tmp, sizeof(tmp),
			&gb.addr_src[j],
			&gb.addr_dst[j]);

	LOG("[%d] %s: %s", j, gb.names[j], tmp);
}

void
parse_cmdline(int argc, char **argv)
{
        int found = 0;
	char short_options[] = "hs:d:xpift:P:";

	char c = '?';

	/* parse command line options */

	while (1) {

		c = getopt(argc, argv, short_options);

		if (c == EOF)
		    break;

		switch (c) {

		case 'P':
			if (gb.sessions >= MAX_SESSIONS)
				ERR("reached max sessions %d", gb.sessions);
			parse_host_pair(optarg);
			break;

		case 'x':
			gb.mgrp = 1;
			break;

		case 'p':
			gb.piggyback = 1;
			break;

		case 'i':
			gb.iterate = 1;
			break;

		case 'f':
			gb.fast = 1;
			break;

		case 't':
                        found = sscanf(optarg, "%u", &gb.iterate_sec);
                        if (found != 1)
                                ERR("error: cannot parse seconds for -t option: %s", optarg);
			gb.iterate = 1;
			break;

		case 'h':
			usage();
			exit(0);
			break;
		    
		case '?':
			break;
		    
		default:
			break;
		}
	}

	check_params();
}

