/*
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Tue Jan 29 11:51:32 MST 2008 (based on traceinfo.c)
 *
 * Retrieves packet counts from multiple network devices and sends
 * statistics to the specified estimator through INFO/dev entries.
 *
 * The device counters are retrieved from /proc/net/dev.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#include "net.h"
#include "atimer.h"
#include "astring.h"
#include "libestimator.h"

#define MAX_DEV_NUM             10

struct global_state {
        char hostid[EST_INFO_IDLEN];

        int estimator_count;
        struct est_estimator estimator;

        int device_count;
        struct net_device devices[MAX_DEV_NUM];

        int interval_sec;

        int debug;
} gb = {
        .interval_sec          = 2,
        .debug                  = 0,
};

void
fill_info_dev(struct est_info_dev *devinfo, struct net_device *dev)
{
        est_info_dev_init(devinfo);

        est_fill_info_srcdst(&devinfo->hdr, dev->src, dev->dst);
        est_fill_info_timestamp(&devinfo->hdr, gb.hostid, &dev->counts.timestamp);

        strncpy(devinfo->stats.name, dev->name, sizeof(devinfo->stats.name));
        strncpy(devinfo->stats.tag, dev->tag, sizeof(devinfo->stats.tag));

        devinfo->stats.rxbps = dev->avg_rate.rxbps;
        devinfo->stats.rxpps = dev->avg_rate.rxpps;
        devinfo->stats.txbps = dev->avg_rate.txbps;
        devinfo->stats.txpps = dev->avg_rate.txpps;
}

struct net_device *
find_matching_dev(const char *buf)
{
        int i = 0;

        for (i = 0; i < gb.device_count; i++) {
                struct net_device *dev = &gb.devices[i];

                if (strstr(buf, dev->name_pattern) != NULL)
                        return dev;
        }
        return NULL;
}

void
update_dev_stats()
{
        char buf[300];
        FILE *procdev = NULL;

        procdev = fopen("/proc/net/dev", "r");
        SYSERR_ON(procdev == NULL);

        while (fgets(buf, sizeof(buf), procdev) != NULL) {
                struct net_device *dev = find_matching_dev(buf);

                if (dev == NULL)
                        continue;
                
                net_device_update_rate_from_buf(dev, buf);

                if (gb.debug) {
                        LOGNR("%s: ", dev->name);
                        LOG_NETCOUNTERS(&dev->counts);
                }
        }
        fclose(procdev);
}

void
send_dev_stats()
{
        int i = 0;
        int info_count = 0;
        struct est_info info_array[MAX_DEV_NUM];

        if (gb.debug)
                LOG("---");

        for (i = 0; i < gb.device_count; i++) {
                struct net_device *dev = &gb.devices[i];
                struct est_info *info = &info_array[info_count];
                char infostr[200];

                BUG_ON(info_count >= MAX_DEV_NUM);

                fill_info_dev(&info->dev, dev);

                est_info_oneline_str(infostr, sizeof(infostr), info);
                LOG("%s", infostr);

                info_count++;
        }
        
        if (gb.estimator_count > 0) {
                int rc = est_push_info_entries(&gb.estimator.ch,
                                               info_array, info_count);
                BUG_ON(rc);
        }

        if (gb.debug)
                LOG("---");
}

void usage();
void parse_cmdline(int argc, char **argv);

int
main(int argc, char **argv)
{
        int rc = 0;

        est_get_host_id(gb.hostid, sizeof(gb.hostid));
	parse_cmdline(argc, argv);

        if (gb.estimator_count > 0) {
                rc = est_open_channel(&gb.estimator.ch, &gb.estimator.addr);
                if (rc) {
                        LOG("error: failed to open channel to estimator on %s (%s)",
                            gb.estimator.unparsed, gb.estimator.addrstr);
                        goto done;
                }
        }

        // initialize device stats
        update_dev_stats();

        // main loop
        while (1) {
                sleep(gb.interval_sec);
                update_dev_stats();
                send_dev_stats();
        }

        if (gb.estimator_count > 0)
                est_close_channel(&gb.estimator.ch);
done:
        return 0;
}

void
check_params(void)
{
        if (gb.device_count < 1) {
                printf("error: no devices specified with the -n option\n");
                printf("error: use option '-h' for help\n\n");
                exit(2);
        }
}

void
parse_cmdline(int argc, char **argv)
{
	char short_options[] = "hP:dn:i:";
	char c = '?';
        int found = 0;
        int rc = 0;

	/* parse command line options */

	while (1) {

		c = getopt(argc, argv, short_options);

		if (c == EOF)
		    break;

		switch (c) {

		case 'P':
                        gb.estimator_count++;
                        if (gb.estimator_count > 1)
                                ERR("error: too many -P options");
                        est_parse_estimator(optarg, &gb.estimator);
			break;

		case 'n':
                        if (gb.device_count >= MAX_DEV_NUM)
                                ERR("error: too many -n options");
                        rc = net_device_init(&gb.devices[gb.device_count], optarg);
                        ERR_ON(rc, "error: failed to parse device: %s", optarg);
                        LOG_ERR("device[%d]: name: %s src: %s dst: %s tag: %s",
                                gb.device_count,
                                gb.devices[gb.device_count].name,
                                gb.devices[gb.device_count].src,
                                gb.devices[gb.device_count].dst,
                                gb.devices[gb.device_count].tag);
                        gb.device_count++;
			break;

                case 'i':
                        found = sscanf(optarg, "%u", &gb.interval_sec);
                        ERR_ON(!found, "failed to parse reporting interval: %s", optarg);
                        break;

		case 'd':
                        gb.debug = 1;
			break;
		    
		case 'h':
			usage();
			exit(0);
			break;
		    
		case '?':
			break;
		    
		default:
			break;
		}
	}

	check_params();
}

void
usage(void)
{
	printf("\n");
	printf("    devinfo -- retrieves network device stats and sends INFO/dev entries\n");
	printf("\n");

	printf("DESCRIPTION\n");
        printf("    Retrieves packet counts from multiple network devices and sends\n");
        printf("    statistics to the specified estimator through INFO/dev entries.\n");
	printf("\n");
        printf("    The device counters are retrieved from /proc/net/dev.\n");
	printf("\n");

	printf("USAGE\n");
	printf("\n");
	printf("    devinfo -n <devname>[:src:dst:tag] [-P <estimator[:port]>] [options]\n");
	printf("\n");

	printf("RESULTS\n");
	printf("\n");
        printf("    The results have the following format:\n");
        printf("    <nodeID> <timestamp> <devname> <tag> RX <bps> <pps> TX <bps> <pps>\n");
	printf("\n");

	printf("BUGS\n");
	printf("\n");
	printf("    It seems that an odd interval produces incorrect results.\n");
	printf("    Every other sample is zero. Use an even interval (default: 2 sec).\n");
	printf("\n");
	printf("    The way we treat a counter rollover is to to just ignore the entry\n");
	printf("    for that interval and keep the rate of the previous interval.\n");
        printf("    To cause a rollover use iperf on gigabit ethernet.\n");
	printf("    It takes about 200 seconds for the counter to roll over.\n");
	printf("\n");

	printf("OPTIONS\n");
	printf("        -n <devname>[:src:dst:tag]      eth0, eth1, ..\n");
	printf("        -i <interval-sec>               reporing interval in sec\n");
	printf("        -d                              print debug info\n");
	printf("\n");
	exit(1);
}

