/*
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Tue Feb  5 11:44:44 MST 2008
 *
 * Retrieves packet counts and bytes from MGRP and sends interval
 * statistics to the specified estimator through INFO/mgrp entries.
 *
 * The counters are retrieved from /proc/net/mgrp_counters.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#include "mgrp_counters.h"
#include "net.h"
#include "aptr.h"
#include "atimer.h"
#include "astring.h"
#include "libestimator.h"

#define MAX_FLOW_NUM            10

// for every flow we send three INFO entries
// (one est_info_mgrp and two est_info_stats)
//
// see send_mgrp_stats()
#define MAX_INFO_NUM            (3*MAX_FLOW_NUM)

struct mgrpflow {
        char dst_name[100];
        struct sockaddr_in dst_addr;

        char pattern[100];

        struct mgrp_counters counts_prev;
        struct timeval counts_prev_timestamp;

        struct mgrp_counters counts;
        struct timeval counts_timestamp;
};

// 0 success, -1 error
static inline int
mgrpflow_parse_init(struct mgrpflow *flow, const char *flow_dst_name)
{
        int found = 0;
        char error[200];
        char addr_str[100];

        memset(flow, 0, sizeof(struct mgrpflow));
        strncpy(flow->dst_name, flow_dst_name, sizeof(flow->dst_name));

        found = net_resolve_host(flow->dst_name, &flow->dst_addr,
                                 error, sizeof(error));
        if (!found) {
                LOG("error: failed to resolve the destination '%s': %s",
                    flow->dst_name, error);
                return -1;
        }

        // pattern used to match relevant line from /proc/net/mgrp_counters
        // the destination in /proc appears as an IP address
        net_sinaddr_str(addr_str, sizeof(addr_str), &flow->dst_addr);
        snprintf(flow->pattern, sizeof(flow->pattern), "%s ", addr_str);

        return 0;
}

struct global_state {
        char hostid[EST_INFO_IDLEN];

        int estimator_count;
        struct est_estimator estimator;

        int flow_count;
        struct mgrpflow flows[MAX_FLOW_NUM];

        int interval_sec;

        int debug;
} gb = {
        .interval_sec           = 1,
        .debug                  = 0,
};

// INFO/mgrp
//

#define PPS(end, start, usec, field) \
        net_calculate_rate_per_second((end)->field, \
                                      (start)->field, \
                                      (usec))

#define BPS(end, start, usec, field) \
        net_calculate_rate_per_second(8 * ((end)->field), \
                                      8 * ((start)->field), \
                                      (usec))

void
fill_info_mgrp(struct est_info_mgrp *mgrp, const struct mgrpflow *flow)
{
        const struct mgrp_counters *start = &flow->counts_prev;
        const struct timeval *start_timestamp = &flow->counts_prev_timestamp;

        const struct mgrp_counters *end = &flow->counts;
        const struct timeval *end_timestamp = &flow->counts_timestamp;

        long long interval_usec = atimer_timeval_diff_usec(end_timestamp,
                                                           start_timestamp);

        est_info_mgrp_init(mgrp);

        // Set src, dst in the header key.
        //
        // This way infoplot.c will add them to the preamble of the datafile
        // and create one datafile for each src,dst combination.
        //
        // The source is the current host (where MGRP is running)
        // The dest is part of the /proc/net/mgrp_counters entry
        est_fill_info_srcdst(&mgrp->hdr, gb.hostid, flow->dst_name);
        est_fill_info_timestamp(&mgrp->hdr, gb.hostid, end_timestamp);

        // STATS

        strncpy(mgrp->stats.dst, flow->dst_name, sizeof(mgrp->stats.dst));

        // mgrp (totals)
        mgrp->stats.mgrp_pps =
                PPS(end, start, interval_usec, mgrp_pkts);
        mgrp->stats.mgrp_bps_header =
                BPS(end, start, interval_usec, mgrp_bytes_header);
        mgrp->stats.mgrp_bps_payload =
                BPS(end, start, interval_usec, mgrp_bytes_payload);

        // tcp
        mgrp->stats.tcp_pps = PPS(end, start, interval_usec, tcp_pkts);
        mgrp->stats.tcp_bps = BPS(end, start, interval_usec, tcp_bytes);

        // udp
        mgrp->stats.udp_pps = PPS(end, start, interval_usec, udp_pkts);
        mgrp->stats.udp_bps = BPS(end, start, interval_usec, udp_bytes);

        // probe
        mgrp->stats.probe_pps = PPS(end, start, interval_usec, probe_pkts);
        mgrp->stats.probe_bps_actual =
                BPS(end, start, interval_usec, probe_bytes_actual);
        mgrp->stats.probe_bps_orig =
                BPS(end, start, interval_usec, probe_bytes_orig);

        // pad
        mgrp->stats.probe_pps_reused =
                PPS(end, start, interval_usec, probe_pkts_reused);
        mgrp->stats.probe_bps_nonpad =
                BPS(end, start, interval_usec, probe_bytes_nonpad);
        mgrp->stats.probe_bps_pad_unused =
                BPS(end, start, interval_usec, probe_bytes_pad_unused);
        mgrp->stats.probe_bps_pad_reused =
                BPS(end, start, interval_usec, probe_bytes_pad_reused);

        // recv
        mgrp->stats.recv.mgrp_pps =
                PPS(end, start, interval_usec, recv.mgrp_pkts);
        mgrp->stats.recv.mgrp_lost_pps =
                PPS(end, start, interval_usec, recv.mgrp_pkts_lost);
        mgrp->stats.recv.mgrp_missing_pps =
                PPS(end, start, interval_usec, recv.mgrp_pkts_missing);
        mgrp->stats.recv.mgrp_outoforder_pps =
                PPS(end, start, interval_usec, recv.mgrp_pkts_outoforder);
}

enum mgrpinfo_stats_type {
        MGRPINFO_SEND = 0,
        MGRPINFO_RECV,
};

// INFO/stats-mgrp
//
// In addition to the INFO/mgrp stats we also use the newer generic
// est_info_stats to send the the actual diffs of the counters,
// instead of computing rates (PPS/BPS)

static inline void
complete_info_stats_mgrp_send(struct est_info_stats *stats,
                              const struct mgrp_counters *end,
                              const struct mgrp_counters *start,
                              long long usec_diff)
{
        // Fill everything under the tag.
        // Headers, src, dst and timestamps already filled

        snprintf(stats->tag, sizeof(stats->tag), "send");

        // Follow order in fill_info_mgrp()

        est_info_stats_setint(stats, 0, usec_diff);

        // mgrp (totals)
        est_info_stats_setint(stats, 1,
                              APTR_FIELD_DIFF(end, start, mgrp_pkts));
        est_info_stats_setint(stats, 2,
                              APTR_FIELD_DIFF(end, start, mgrp_bytes_header));
        est_info_stats_setint(stats, 3,
                              APTR_FIELD_DIFF(end, start, mgrp_bytes_payload));

        // tcp
        est_info_stats_setint(stats, 4,
                              APTR_FIELD_DIFF(end, start, tcp_pkts));
        est_info_stats_setint(stats, 5,
                              APTR_FIELD_DIFF(end, start, tcp_bytes));

        // udp
        est_info_stats_setint(stats, 6,
                              APTR_FIELD_DIFF(end, start, udp_pkts));
        est_info_stats_setint(stats, 7,
                              APTR_FIELD_DIFF(end, start, udp_bytes));

        // probe
        est_info_stats_setint(stats, 8,
                              APTR_FIELD_DIFF(end, start, probe_pkts));
        est_info_stats_setint(stats, 9,
                              APTR_FIELD_DIFF(end, start, probe_bytes_actual));
        est_info_stats_setint(stats, 10,
                              APTR_FIELD_DIFF(end, start, probe_bytes_orig));

        // pad
        est_info_stats_setint(stats, 11,
                              APTR_FIELD_DIFF(end, start, probe_pkts_reused));
        est_info_stats_setint(stats, 12,
                              APTR_FIELD_DIFF(end, start, probe_bytes_nonpad));
        est_info_stats_setint(stats, 13,
                              APTR_FIELD_DIFF(end, start, probe_bytes_pad_unused));
        est_info_stats_setint(stats, 14,
                              APTR_FIELD_DIFF(end, start, probe_bytes_pad_reused));
}

static inline void
complete_info_stats_mgrp_recv(struct est_info_stats *stats,
                              const struct mgrp_counters *end,
                              const struct mgrp_counters *start,
                              long long usec_diff)
{
        // Fill everything under the tag.
        // Headers, src, dst and timestamps already filled

        snprintf(stats->tag, sizeof(stats->tag), "recv");

        // Follow order in fill_info_mgrp()

        est_info_stats_setint(stats, 0, usec_diff);

        // recv
        est_info_stats_setint(stats, 1,
                              APTR_FIELD_DIFF(end, start, recv.mgrp_pkts));
        est_info_stats_setint(stats, 2,
                              APTR_FIELD_DIFF(end, start, recv.mgrp_pkts_lost));
        est_info_stats_setint(stats, 3,
                              APTR_FIELD_DIFF(end, start, recv.mgrp_pkts_missing));
        est_info_stats_setint(stats, 4,
                              APTR_FIELD_DIFF(end, start, recv.mgrp_pkts_outoforder));
}

static inline void
fill_info_stats_mgrp(struct est_info_stats *stats,
                     const struct mgrpflow *flow,
                     enum mgrpinfo_stats_type type)
{
        const struct mgrp_counters *start = &flow->counts_prev;
        const struct timeval *start_timestamp = &flow->counts_prev_timestamp;

        const struct mgrp_counters *end = &flow->counts;
        const struct timeval *end_timestamp = &flow->counts_timestamp;

        long long interval_usec = atimer_timeval_diff_usec(end_timestamp,
                                                           start_timestamp);
        const char *src_host = gb.hostid;
        const char *dst_host = flow->dst_name;
        const struct timeval *timestamp = end_timestamp;

        // Fill the header: key and timestamp.
        //
        // Infoplot will create a separate data file whose name has the format:
        //
        //      stats-nuttcp.<src>.<dst>.dat
        //
        // where type="stats" and subtype="mgrp".
        //
        est_info_stats_init(stats, "mgrp");
        est_fill_info_srcdst(&stats->hdr, src_host, dst_host);
        est_fill_info_timestamp(&stats->hdr, src_host, timestamp);

        if (type == MGRPINFO_SEND)
                complete_info_stats_mgrp_send(stats, end, start, interval_usec);
        else if (type == MGRPINFO_RECV)
                complete_info_stats_mgrp_recv(stats, end, start, interval_usec);
        else
                ERRFN("unknown mgrp_stats type %d", type);
}

struct mgrpflow *
find_matching_flow(const char *buf)
{
        int i = 0;

        for (i = 0; i < gb.flow_count; i++) {
                struct mgrpflow *flow = &gb.flows[i];

                if (strstr(buf, flow->pattern) != NULL)
                        return flow;
        }
        return NULL;
}

int
counter_rollover(struct est_info_mgrp *mgrp)
{
        return (mgrp->stats.mgrp_pps < 0 ||
                mgrp->stats.mgrp_bps_header < 0 ||
                mgrp->stats.mgrp_bps_payload < 0 ||
                mgrp->stats.tcp_pps < 0 ||
                mgrp->stats.tcp_bps < 0 ||
                mgrp->stats.udp_pps < 0 ||
                mgrp->stats.udp_bps < 0 ||
                mgrp->stats.probe_pps < 0 ||
                mgrp->stats.probe_bps_actual < 0 ||
                mgrp->stats.probe_bps_orig < 0 ||
                mgrp->stats.probe_pps_reused < 0 ||
                mgrp->stats.probe_bps_nonpad < 0 ||
                mgrp->stats.probe_bps_pad_unused < 0 ||
                mgrp->stats.probe_bps_pad_reused < 0 ||
                mgrp->stats.recv.mgrp_pps < 0 ||
                mgrp->stats.recv.mgrp_missing_pps < 0 ||
                mgrp->stats.recv.mgrp_outoforder_pps < 0);

        // XXX mgrp_lost_pps can be negative, in case we over-estimated
        // XXX the lost packets in the previous epoch
        // XXX see comments in mgrp/proto.c:mgrp_counters_update_receive_side()
}

void
advance_flow_stats()
{
        int i = 0;

        for (i = 0; i < gb.flow_count; i++) {
                struct mgrpflow *flow = &gb.flows[i];

                flow->counts_prev = flow->counts;
                flow->counts_prev_timestamp = flow->counts_timestamp;

                gettimeofday(&flow->counts_timestamp, NULL);
        }
}

void
update_mgrp_stats()
{
        char buf[300];
        FILE *procmgrp = NULL;

        procmgrp = fopen("/proc/net/mgrp_counters", "r");
        SYSERR_ON(procmgrp == NULL);

        while (fgets(buf, sizeof(buf), procmgrp) != NULL) {
                int rc = 0;
                char *counters = NULL;
                struct mgrpflow *flow = find_matching_flow(buf);

                if (flow == NULL)
                        continue;
                
                // skip first token (IP address) and the space
                counters = strchr(buf, ' ');
                if (counters == NULL) {
                        LOG_ERR("error: cannot find where counters start");
                        LOG_ERR("error: %s", buf);
                        continue;
                }
                counters += 1;

                rc = mgrp_counters_parse_oneliner(&flow->counts, counters);
                if (rc) {
                        LOG_ERR("error: cannot parse all counters");
                        LOG_ERR("error: %s", buf);
                        continue;
                }

                // Even if the flow has the same destination, the epoch may
                // have changed, which means that MGRP has reset the counters.
                // In this case the previous counters have to be reset in order
                // to get correct results.
                //
                // Note that we may understimate bandwidth because we miss
                // the difference between our last read of the counters and
                // the actual value of the counters when the epoch was changed.
                // This is only a problem if the space between the two epochs
                // is smaller than our update interval. Not really a problem.
                //
                // All this comes from the fact that MGRP drops a flow when
                // no socket is using it. This is good for normal operation
                // (to clean up state) but interferes with our experiments.
                //
                // We may consider making flows permanent.
                if (flow->counts.epoch != flow->counts_prev.epoch) {
                        LOG_ERR("warning: epoch changed for %s: %ld -> %ld",
                               flow->pattern,
                               flow->counts_prev.epoch,
                               flow->counts.epoch);
                        memset(&flow->counts_prev, 0, sizeof(flow->counts_prev));
                }
                                             
                if (gb.debug) {
                        char tmp[300];
                        LOG("---");
                        mgrp_counters_oneliner(tmp, sizeof(tmp), &flow->counts_prev);
                        LOG("update-prev  : %s%s", flow->pattern, tmp);
                        mgrp_counters_oneliner(tmp, sizeof(tmp), &flow->counts);
                        LOG("update-parsed: %s%s", flow->pattern, tmp);
                        LOGNR("update-orig  : %s", buf);
                }
        }
        fclose(procmgrp);
}

void
send_mgrp_stats()
{
        int i = 0;
        int info_count = 0;
        struct est_info info_array[MAX_INFO_NUM];

        if (gb.debug)
                LOG("---");

        for (i = 0; i < gb.flow_count; i++) {
                struct mgrpflow *flow = &gb.flows[i];
                struct est_info *info_rate = &info_array[info_count];
                struct est_info *info_send = &info_array[info_count+1];
                struct est_info *info_recv = &info_array[info_count+2];
                char infostr[300];

                BUG_ON(info_count >= MAX_INFO_NUM);

                fill_info_mgrp(&info_rate->mgrp, flow);
                est_info_oneline_str(infostr, sizeof(infostr), info_rate);
                LOG("%s", infostr);

                fill_info_stats_mgrp(&info_send->stats, flow, MGRPINFO_SEND);
                est_info_oneline_str(infostr, sizeof(infostr), info_send);
                LOG("%s", infostr);

                fill_info_stats_mgrp(&info_recv->stats, flow, MGRPINFO_RECV);
                est_info_oneline_str(infostr, sizeof(infostr), info_recv);
                LOG("%s", infostr);

                if (counter_rollover(&info_rate->mgrp)) {
                        LOG_ERR("warning: counter rollover; skipping");
                        continue;
                }

                // advance only if no rollover
                info_count += 3;
        }
        
        if (gb.estimator_count > 0 && info_count > 0) {
                int rc = est_push_info_entries(&gb.estimator.ch,
                                               info_array, info_count);
                BUG_ON(rc);
        }

        if (gb.debug)
                LOG("---");
}

void usage();
void parse_cmdline(int argc, char **argv);

int
main(int argc, char **argv)
{
        int rc = 0;

        est_get_host_id(gb.hostid, sizeof(gb.hostid));
	parse_cmdline(argc, argv);

        if (gb.estimator_count > 0) {
                rc = est_open_channel(&gb.estimator.ch, &gb.estimator.addr);
                if (rc) {
                        LOG("error: failed to open channel to estimator on %s (%s)",
                            gb.estimator.unparsed, gb.estimator.addrstr);
                        goto done;
                }
        }

        // initialize stats
        advance_flow_stats();
        update_mgrp_stats();

        // main loop
        while (1) {
                sleep(gb.interval_sec);
                advance_flow_stats();
                update_mgrp_stats();
                send_mgrp_stats();
        }

        if (gb.estimator_count > 0)
                est_close_channel(&gb.estimator.ch);
done:
        return 0;
}

void
check_params(void)
{
        if (gb.flow_count < 1) {
                printf("error: no flows specified with the -f option\n");
                printf("error: use option '-h' for help\n\n");
                exit(2);
        }
}

void
parse_cmdline(int argc, char **argv)
{
	char short_options[] = "hP:df:i:";
	char c = '?';
        char tmp[100];
        int found = 0;
        int rc = 0;

	/* parse command line options */

	while (1) {

		c = getopt(argc, argv, short_options);

		if (c == EOF)
		    break;

		switch (c) {

		case 'P':
                        gb.estimator_count++;
                        if (gb.estimator_count > 1)
                                ERR("error: too many -P options");
                        est_parse_estimator(optarg, &gb.estimator);
			break;

		case 'f':
                        if (gb.flow_count >= MAX_FLOW_NUM)
                                ERR("error: too many -f options");
                        rc = mgrpflow_parse_init(&gb.flows[gb.flow_count],
                                                 optarg);
                        ERR_ON(rc, "error: failed to parse flow: %s", optarg);
                        LOG_ERR("flow[%d]: dst_name: %s dst_addr: %s",
                                gb.flow_count,
                                gb.flows[gb.flow_count].dst_name,
                                net_sinaddr_str(
                                        tmp, sizeof(tmp),
                                        &gb.flows[gb.flow_count].dst_addr));
                        gb.flow_count++;
			break;

                case 'i':
                        found = sscanf(optarg, "%u", &gb.interval_sec);
                        ERR_ON(!found, "failed to parse reporting interval: %s", optarg);
                        break;

		case 'd':
                        gb.debug = 1;
			break;
		    
		case 'h':
			usage();
			exit(0);
			break;
		    
		case '?':
			break;
		    
		default:
			break;
		}
	}

	check_params();
}

void
usage(void)
{
	printf("\n");
	printf("    mgrpinfo -- retrieves MGRP stats and sends INFO/mgrp entries\n");
	printf("\n");

	printf("DESCRIPTION\n");
        printf("    Retrieves packet counts from multiple MGRP flows and sends\n");
        printf("    statistics to the specified estimator through INFO/mgrp entries.\n");
	printf("\n");
        printf("    The MGRP counters are retrieved from /proc/net/mgrp_counters.\n");
	printf("\n");

	printf("USAGE\n");
	printf("\n");
	printf("    mgrpinfo -f <flow_dst> [ -f <flow2_dst> ] [-P <estimator[:port]>] [options]\n");
	printf("\n");

	printf("RESULTS\n");
	printf("\n");
        printf("    The results have the following format (all in one line):\n");
        printf("    (P = packets, B = bytes)\n");
	printf("\n");
        printf("    <nodeID> <timestamp> <flow-dst>\n");
        printf("        mgrp <P> <headerB> <payloadB>\n");
        printf("        tcp <P> <B>\n");
        printf("        udp <P> <B>\n");
        printf("        probe <P> <actualB> <origB>\n");
        printf("        pad <reusedP> <nonpadB> <padunusedB> <padreusedB>\n");
        printf("        recv <P> <lostP> <missingP> <outoforderP>\n");
	printf("\n");

	printf("BUGS\n");
	printf("\n");
	printf("    The way we treat a counter rollover is to to just ignore the entry\n");
	printf("    for that interval.\n");
	printf("\n");

	printf("OPTIONS\n");
	printf("        -f <flow-dst>           m4, m5, <hostname>, ..\n");
	printf("        -i <interval-sec>       reporing interval in sec\n");
	printf("        -d                      print debug info\n");
	printf("\n");
	exit(1);
}

