/*
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Wed Mar 26 14:01:37 MST 2008
 *
 */

#include "pcapstats.h"

struct global_state {
        char hostid[EST_INFO_IDLEN];


        int verbose;

        int use_estimator;
        struct est_estimator estimator;

        char send_trace_host[EST_INFO_IDLEN];

        char send_host[EST_INFO_IDLEN];
        char send_pcap_name[300];

        char recv_host[EST_INFO_IDLEN];
        char recv_pcap_name[300];

        char filter[300];

        int pktmap_initialized;
        char pktmap_name[500];
        FILE *pktmapfp;
        int pktmap_saw_first_timestamp;
} gb = {
        .verbose = 0,
};

uint64_t
pcap_pkt_hash(void *data, void *data_key)
{
        return _pcap_pkt_hash(data, data_key);
}

int
pcap_pkt_cmp(void *lhs_data, void *rhs_data)
{
        return _pcap_pkt_cmp(lhs_data, rhs_data);
}

static inline void
pcapstats_fill_stats_common(struct est_info_stats *stats,
                            const char *tag,
                            const struct timeval *timestamp)
{
        const char *hostid = gb.send_trace_host;
        const char *src_host = gb.send_host;
        const char *dst_host = gb.recv_host;

        // Fill the header: key and timestamp.
        //
        // Infoplot will create a separate data file whose name has the format:
        //
        //      stats-pcap.<src>.<dst>.dat
        //
        // where type="pcap" and subtype=@tag
        //
        est_info_stats_init(stats, "pcap");
        est_fill_info_srcdst(&stats->hdr, src_host, dst_host);
        est_fill_info_timestamp(&stats->hdr, hostid, timestamp);

        snprintf(stats->tag, sizeof(stats->tag), tag);

}

static inline void
pcapstats_fill_stats_lost(struct est_info_stats *stats, struct pcap_pkt *pkt)
{
        char str[100];

        pcap_pkt_oneliner(pkt, str, sizeof(str));

        pcapstats_fill_stats_common(stats, "lost", &pkt->send_pcap.ts);

        // send stats
        est_info_stats_setint(stats, 0, pkt->send_seqno);
        est_info_stats_setint(stats, 1, pkt->send_pcap.len);

        est_info_stats_setbigstr(stats, 2, str);
}

// out-of-order
static inline void
pcapstats_fill_stats_oord(struct est_info_stats *stats, struct pcap_pkt *pkt)
{
        char str[100];

        pcap_pkt_oneliner(pkt, str, sizeof(str));

        pcapstats_fill_stats_common(stats, "oord", &pkt->send_pcap.ts);

        // send stats
        est_info_stats_setint(stats, 0, pkt->send_seqno);
        est_info_stats_setint(stats, 1, pkt->send_pcap.len);

        // recv stats
        est_info_stats_settimeval(stats, 2, &pkt->recv_pcap.ts);
        est_info_stats_setint(stats, 3, pkt->recv_seqno_diff);

        est_info_stats_setbigstr(stats, 4, str);
}


void
load_match_pcap_stats(struct pcap_stats *stats)
{
        int pkts = 0;
        int loaded = 0;
        int matched = 0;

        if (gb.verbose)
                LOG("Loading send packets..");

        while ((pkts = load_next_send_pkts(stats, PCAPSTATS_LOAD_STEP)) > 0) {
                loaded += pkts;
                if (gb.verbose)
                        LOG("    loaded  %10d total %10d", pkts, loaded);
        }

        if (gb.verbose)
                LOG("Matching recv packets..");

        matched = match_next_recv_pkts(stats);

        if (gb.verbose)
                LOG("    matched %10d total %10d", matched, matched);
}

struct pcapstats_counters_pktsbytes {
        long long sent;
        long long sent_bytes;
        long long lost;
        long long lost_bytes;
        long long oord;
        long long oord_bytes;
};

struct pcapstats_counters {
        // when epoch started
        // usec is always zero
        struct timeval ts;

        // all = udp + tcp - tcp_mgrp_pbk + other 
        // (piggybacked packets are counted twice in udp, tcp)
        struct pcapstats_counters_pktsbytes all;

        // udp_mgrp = udp - udp_nonmgrp
        // tcp_mgrp = tcp - tcp_nonmgrp
        // mgrp = udp_mgrp + tcp_mgrp
        struct pcapstats_counters_pktsbytes mgrp;

        // udp = udp_nopbk + udp_mgrp_pbk
        // udp_nopbk = udp_nonmgrp + udp_mgrp_nopbk
        struct pcapstats_counters_pktsbytes udp;
        struct pcapstats_counters_pktsbytes udp_nonmgrp;
        struct pcapstats_counters_pktsbytes udp_nopbk;
        struct pcapstats_counters_pktsbytes udp_mgrp_pbk;

        // tcp = tcp_nopbk + tcp_mgrp_pbk
        // tcp_nopbk = tcp_nonmgrp + tcp_mgrp_nopbk
        struct pcapstats_counters_pktsbytes tcp;
        struct pcapstats_counters_pktsbytes tcp_nonmgrp;
        struct pcapstats_counters_pktsbytes tcp_nopbk;
        struct pcapstats_counters_pktsbytes tcp_mgrp_pbk;

        struct pcapstats_counters_pktsbytes other;

        // HOWTO: add a new counter set
        //      - add set in struct pcapstats_counters
        //      - add flag in struct pcap_pkt_flags
        //      - add in pcap_pkt_detect_transports()
        //      - add in pcapstats_counters_diff()
        //      - add in pcapstats_counters_update()
        //      - add in pcapstats_add_info_stats_counters()
};

static inline void
pcapstats_counters_pktsbytes_diff(struct pcapstats_counters_pktsbytes *diff,
                                  struct pcapstats_counters_pktsbytes *end,
                                  struct pcapstats_counters_pktsbytes *begin)
{
        APTR_FIELD_DIFF_STORE(diff, end, begin, sent);
        APTR_FIELD_DIFF_STORE(diff, end, begin, sent_bytes);
        APTR_FIELD_DIFF_STORE(diff, end, begin, lost);
        APTR_FIELD_DIFF_STORE(diff, end, begin, lost_bytes);
        APTR_FIELD_DIFF_STORE(diff, end, begin, oord);
        APTR_FIELD_DIFF_STORE(diff, end, begin, oord_bytes);
}

static inline void
pcapstats_counters_diff(struct pcapstats_counters *diff,
                        struct pcapstats_counters *end,
                        struct pcapstats_counters *begin)
{
        pcapstats_counters_pktsbytes_diff(&diff->all, &end->all, &begin->all);

        // mgrp
        pcapstats_counters_pktsbytes_diff(&diff->mgrp, &end->mgrp, &begin->mgrp);

        // udp
        pcapstats_counters_pktsbytes_diff(&diff->udp, &end->udp, &begin->udp);
        pcapstats_counters_pktsbytes_diff(&diff->udp_nonmgrp,
                                          &end->udp_nonmgrp, &begin->udp_nonmgrp);
        pcapstats_counters_pktsbytes_diff(&diff->udp_nopbk,
                                          &end->udp_nopbk, &begin->udp_nopbk);
        pcapstats_counters_pktsbytes_diff(&diff->udp_mgrp_pbk,
                                          &end->udp_mgrp_pbk, &begin->udp_mgrp_pbk);

        // tcp
        pcapstats_counters_pktsbytes_diff(&diff->tcp, &end->tcp, &begin->tcp);
        pcapstats_counters_pktsbytes_diff(&diff->tcp_nonmgrp,
                                          &end->tcp_nonmgrp, &begin->tcp_nonmgrp);
        pcapstats_counters_pktsbytes_diff(&diff->tcp_nopbk,
                                          &end->tcp_nopbk, &begin->tcp_nopbk);
        pcapstats_counters_pktsbytes_diff(&diff->tcp_mgrp_pbk,
                                          &end->tcp_mgrp_pbk, &begin->tcp_mgrp_pbk);

        pcapstats_counters_pktsbytes_diff(&diff->other, &end->other, &begin->other);
}

static inline void
pcapstats_counters_pktsbytes_update(struct pcapstats_counters_pktsbytes *counters,
                                    int len,
                                    int lost, int oord)
{
        BUG_ON(lost && oord);

        counters->sent++;
        counters->sent_bytes += len;

        if (lost) {
                counters->lost++;
                counters->lost_bytes += len;
        }
        if (oord) {
                counters->oord++;
                counters->oord_bytes += len;
        }

        BUG_ON(counters->sent < counters->lost);
        BUG_ON(counters->sent < counters->oord);
}

static inline void
pcapstats_counters_update(struct pcapstats_counters *counters,
                          const struct pcap_pkt *pkt)
{
        struct pcap_pkt_flags flags;

        int lost = pcap_pkt_lost(pkt);
        int oord = pcap_pkt_outoforder(pkt);

        pcap_pkt_detect_transports(&flags, pkt);

        pcapstats_counters_pktsbytes_update(&counters->all,
                                            pkt->transport_len,
                                            lost, oord);

        if (flags.is_mgrp)
                pcapstats_counters_pktsbytes_update(&counters->mgrp,
                                                    pkt->transport_len,
                                                    lost, oord);
        if (flags.is_udp)
                pcapstats_counters_pktsbytes_update(&counters->udp,
                                                    flags.udp_len_orig,
                                                    lost, oord);
        if (flags.is_udp_nonmgrp)
                pcapstats_counters_pktsbytes_update(&counters->udp_nonmgrp,
                                                    flags.udp_len_orig,
                                                    lost, oord);
        if (flags.is_udp_nopbk)
                pcapstats_counters_pktsbytes_update(&counters->udp_nopbk,
                                                    flags.udp_len_orig,
                                                    lost, oord);
        if (flags.is_udp_mgrp_pbk)
                pcapstats_counters_pktsbytes_update(&counters->udp_mgrp_pbk,
                                                    flags.udp_len_orig,
                                                    lost, oord);
        if (flags.is_tcp)
                pcapstats_counters_pktsbytes_update(&counters->tcp,
                                                    flags.tcp_len,
                                                    lost, oord);
        if (flags.is_tcp_nonmgrp)
                pcapstats_counters_pktsbytes_update(&counters->tcp_nonmgrp,
                                                    flags.tcp_len,
                                                    lost, oord);
        if (flags.is_tcp_nopbk)
                pcapstats_counters_pktsbytes_update(&counters->tcp_nopbk,
                                                    flags.tcp_len,
                                                    lost, oord);
        if (flags.is_tcp_mgrp_pbk)
                pcapstats_counters_pktsbytes_update(&counters->tcp_mgrp_pbk,
                                                    flags.tcp_len,
                                                    lost, oord);

        if (!flags.is_udp && !flags.is_tcp)
                pcapstats_counters_pktsbytes_update(&counters->other,
                                                    pkt->transport_len,
                                                    lost, oord);
}

static inline void
pcapstats_fill_stats_counters_pktsbytes(struct est_info_stats *stats,
                                        const char *tag,
                                        struct timeval *ts,
                                        struct pcapstats_counters_pktsbytes *counters)
{
        pcapstats_fill_stats_common(stats, tag, ts);

        // counters
        est_info_stats_setint(stats, 0, counters->sent);
        est_info_stats_setint(stats, 1, counters->sent_bytes);
        est_info_stats_setint(stats, 2, counters->lost);
        est_info_stats_setint(stats, 3, counters->lost_bytes);
        est_info_stats_setint(stats, 4, counters->oord);
        est_info_stats_setint(stats, 5, counters->oord_bytes);
}

// sends multiple INFO/stats-pcap entries
void
pcapstats_add_info_stats_counters(struct est_info_group *igroup,
                                  const char *tag_prefix,
                                  struct pcapstats_counters *counters)
{
        struct est_info info;
        char *tag_suffix = NULL;
        char tag[EST_INFO_STATS_TAGLEN];

        tag_suffix = "all";
        snprintf(tag, sizeof(tag), "%s-%s", tag_prefix, tag_suffix);
        pcapstats_fill_stats_counters_pktsbytes(&info.stats,
                                                tag, &counters->ts,
                                                &counters->all);
        est_info_group_add(igroup, &info, EST_INFO_PRINT);

        tag_suffix = "mgrp";
        snprintf(tag, sizeof(tag), "%s-%s", tag_prefix, tag_suffix);
        pcapstats_fill_stats_counters_pktsbytes(&info.stats,
                                                tag, &counters->ts,
                                                &counters->mgrp);
        est_info_group_add(igroup, &info, EST_INFO_PRINT);

        tag_suffix = "udp";
        snprintf(tag, sizeof(tag), "%s-%s", tag_prefix, tag_suffix);
        pcapstats_fill_stats_counters_pktsbytes(&info.stats,
                                                tag, &counters->ts,
                                                &counters->udp);
        est_info_group_add(igroup, &info, EST_INFO_PRINT);

        tag_suffix = "udp-nonmgrp";
        snprintf(tag, sizeof(tag), "%s-%s", tag_prefix, tag_suffix);
        pcapstats_fill_stats_counters_pktsbytes(&info.stats,
                                                tag, &counters->ts,
                                                &counters->udp_nonmgrp);
        est_info_group_add(igroup, &info, EST_INFO_PRINT);

        tag_suffix = "udp-nopbk";
        snprintf(tag, sizeof(tag), "%s-%s", tag_prefix, tag_suffix);
        pcapstats_fill_stats_counters_pktsbytes(&info.stats,
                                                tag, &counters->ts,
                                                &counters->udp_nopbk);
        est_info_group_add(igroup, &info, EST_INFO_PRINT);

        tag_suffix = "udp-mgrppbk";
        snprintf(tag, sizeof(tag), "%s-%s", tag_prefix, tag_suffix);
        pcapstats_fill_stats_counters_pktsbytes(&info.stats,
                                                tag, &counters->ts,
                                                &counters->udp_mgrp_pbk);
        est_info_group_add(igroup, &info, EST_INFO_PRINT);

        tag_suffix = "tcp";
        snprintf(tag, sizeof(tag), "%s-%s", tag_prefix, tag_suffix);
        pcapstats_fill_stats_counters_pktsbytes(&info.stats,
                                                tag, &counters->ts,
                                                &counters->tcp);
        est_info_group_add(igroup, &info, EST_INFO_PRINT);

        tag_suffix = "tcp-nonmgrp";
        snprintf(tag, sizeof(tag), "%s-%s", tag_prefix, tag_suffix);
        pcapstats_fill_stats_counters_pktsbytes(&info.stats,
                                                tag, &counters->ts,
                                                &counters->tcp_nonmgrp);
        est_info_group_add(igroup, &info, EST_INFO_PRINT);

        tag_suffix = "tcp-nopbk";
        snprintf(tag, sizeof(tag), "%s-%s", tag_prefix, tag_suffix);
        pcapstats_fill_stats_counters_pktsbytes(&info.stats,
                                                tag, &counters->ts,
                                                &counters->tcp_nopbk);
        est_info_group_add(igroup, &info, EST_INFO_PRINT);

        tag_suffix = "tcp-mgrppbk";
        snprintf(tag, sizeof(tag), "%s-%s", tag_prefix, tag_suffix);
        pcapstats_fill_stats_counters_pktsbytes(&info.stats,
                                                tag, &counters->ts,
                                                &counters->tcp_mgrp_pbk);
        est_info_group_add(igroup, &info, EST_INFO_PRINT);

        tag_suffix = "other";
        snprintf(tag, sizeof(tag), "%s-%s", tag_prefix, tag_suffix);
        pcapstats_fill_stats_counters_pktsbytes(&info.stats,
                                                tag, &counters->ts,
                                                &counters->other);
        est_info_group_add(igroup, &info, EST_INFO_PRINT);
}

void
pcapstats_pktmap_init()
{
        if (strlen(gb.pktmap_name) > 0) {
                gb.pktmapfp = fopen(gb.pktmap_name, "w");
                if (gb.pktmapfp == NULL)
                        ERR("error: failed to open pktmap file %s", gb.pktmap_name);
        }
        gb.pktmap_initialized = 1;
}

void
pcapstats_pktmap_done()
{
        if (gb.pktmapfp != NULL) {
                fclose(gb.pktmapfp);
                gb.pktmapfp = NULL;
        }
        gb.pktmap_initialized = 0;
}

#define IS_MGRP_NO_PBK(flags) \
    ((flags.is_mgrp && !flags.is_udp_mgrp_pbk && !flags.is_tcp_mgrp_pbk) ? 1 : 0)

#define IS_MGRP_PBK(flags) \
    ((flags.is_mgrp && (flags.is_udp_mgrp_pbk || flags.is_tcp_mgrp_pbk)) ? 1 : 0)

#define IS_MGRP_UDP_NO_PBK(flags) \
    ((flags.is_mgrp && flags.is_udp_nopbk) ? 1 : 0)

#define IS_MGRP_TCP_NO_PBK(flags)                       \
    ((flags.is_mgrp && flags.is_tcp_nopbk) ? 1 : 0)

#define IS_UDP_NO_MGRP(flags) \
    ((!flags.is_mgrp && flags.is_udp) ? 1 : 0)

#define IS_TCP_NO_MGRP(flags) \
    ((!flags.is_mgrp && flags.is_tcp) ? 1 : 0)
    
#define IS_OTHER_NO_MGRP(flags) \
    ((!flags.is_mgrp && !flags.is_udp && !flags.is_tcp) ? 1 : 0)

void
pcapstats_pktmap_print_header(unsigned long base_sec)
{
        fprintf(gb.pktmapfp, "#&node_id %s\n", gb.send_trace_host);
        fprintf(gb.pktmapfp, "#&src_id %s\n", gb.send_host);
        fprintf(gb.pktmapfp, "#&dst_id %s\n", gb.recv_host);
        fprintf(gb.pktmapfp, "#&src_pcap %s\n", gb.send_pcap_name);
        fprintf(gb.pktmapfp, "#&dst_pcap %s\n", gb.recv_pcap_name);
        fprintf(gb.pktmapfp, "#&base %lu\n", base_sec);
        fprintf(gb.pktmapfp, "#&timestamp_id %s\n", gb.send_trace_host);
        fprintf(gb.pktmapfp,
                "#h, ts,           send_seqno,  len, STATUS, l, o, owd,      FLAGS  M, UDP  T,NM,NP,PB, TCP, T,NM,NP,PB, OTHER, T\n"
                );
}

void
pcapstats_pktmap_update(struct pcap_pkt *pkt)
{
        struct pcap_pkt_flags flags;
        int lost = pcap_pkt_lost(pkt);
        int oord = pcap_pkt_outoforder(pkt);
        unsigned long long owd = -1L;
        unsigned long long owd_sec = 0;
        unsigned long long owd_usec = 0;
        unsigned long long snd_time = 0, rcv_time = 0;

        BUG_ON(!gb.pktmap_initialized);

        if (gb.pktmapfp == NULL)
                return;

        pcap_pkt_detect_transports(&flags, pkt);

        snd_time = pkt->send_pcap.ts.tv_sec * 1000000 + pkt->send_pcap.ts.tv_usec;
        rcv_time = pkt->recv_pcap.ts.tv_sec * 1000000 + pkt->recv_pcap.ts.tv_usec;
        owd = (rcv_time - snd_time);
        owd_sec  = owd / 1000000;
        owd_usec = owd % 1000000;
        if (rcv_time == 0) {
            owd_sec = 0LL;
            owd_usec = 0LL;
        }

        if (!gb.pktmap_saw_first_timestamp) {
                pcapstats_pktmap_print_header(pkt->send_pcap.ts.tv_sec);
                gb.pktmap_saw_first_timestamp = 1;
        }

        /* hostname, timestamp, seqno, len, lost, oord, owd, ... */
        fprintf(gb.pktmapfp,
                "%s, %lu.%.6ld, %5d, %4u, "
                "status, %d, %d, %llu.%.6llu, "
                "flags, %d, udp, %d, %d, %d, %d, tcp, %d, %d, %d, %d, other, %d\n",

                gb.send_trace_host,
                pkt->send_pcap.ts.tv_sec, pkt->send_pcap.ts.tv_usec,
                pkt->send_seqno,
                pkt->send_pcap.len,
                
                lost, oord, owd_sec, owd_usec,

                flags.is_mgrp,

                flags.is_udp,
                flags.is_udp_nonmgrp,
                flags.is_udp_nopbk,
                flags.is_udp_mgrp_pbk,

                flags.is_tcp,
                flags.is_tcp_nonmgrp,
                flags.is_tcp_nopbk,
                flags.is_tcp_mgrp_pbk,

                (!flags.is_udp && !flags.is_tcp)

                );
}

void
walk_compute_stats(struct pcap_stats *stats)
{
        int rc = 0;
        anode *node = NULL;

        struct est_info info;
        struct est_info_group igroup;

        struct pcapstats_counters totals;
        int epoch_count = 0;
        long epoch_sec = 0;
        struct pcapstats_counters epoch_start_totals;

        pcapstats_pktmap_init();
        est_info_group_init(&igroup, (gb.use_estimator ? &gb.estimator : NULL));

        rc = alist_iterator_set(&stats->pktlist,
                                ALIST_FORWARD, ALIST_BEFORE_FIRST, NULL);
        BUG_ON(rc);

        while ((node = alist_iterator_next(&stats->pktlist)) != NULL) {
                struct pcap_pkt *pkt = node->data;
                
                // first epoch
                if (epoch_count == 0) {
                        epoch_count = 1;
                        epoch_sec = pkt->send_pcap.ts.tv_sec;

                        memset(&totals, 0, sizeof(totals));
                        totals.ts.tv_sec = epoch_sec;

                        memset(&epoch_start_totals, 0, sizeof(epoch_start_totals));
                        epoch_start_totals.ts.tv_sec = epoch_sec;

                // epoch transition
                } else if (epoch_sec < pkt->send_pcap.ts.tv_sec) {
                        struct pcapstats_counters epoch_counters = {
                                .ts = epoch_start_totals.ts,
                        };
                        pcapstats_counters_diff(&epoch_counters,
                                                &totals,
                                                &epoch_start_totals);
                        pcapstats_add_info_stats_counters(&igroup, "epoch", &epoch_counters);

                        // epochs should be continuous
                        while (epoch_sec < pkt->send_pcap.ts.tv_sec) {
                                epoch_count++;
                                epoch_sec++;
                                if (epoch_sec == pkt->send_pcap.ts.tv_sec)
                                        break;
                                epoch_start_totals = totals;
                                epoch_start_totals.ts.tv_sec = epoch_sec;
                                epoch_counters.ts = epoch_start_totals.ts;
                                // always returns zero counters
                                pcapstats_counters_diff(&epoch_counters,
                                                        &totals,
                                                        &epoch_start_totals);
                                pcapstats_add_info_stats_counters(&igroup, "epoch", &epoch_counters);
                        }

                        epoch_start_totals = totals;
                        epoch_start_totals.ts.tv_sec = epoch_sec;
                }

                // update counters (detects losses, out-of-order, tcp, udp, ..)
                pcapstats_counters_update(&totals, pkt);

                if (pcap_pkt_outoforder(pkt)) {
                        // generate one info event for every out-of-order packet
                        pcapstats_fill_stats_oord(&info.stats, pkt);
                        est_info_group_add(&igroup, &info, EST_INFO_PRINT);

                } else if (pcap_pkt_lost(pkt)) {
                        // generate one info event for every lost packet
                        pcapstats_fill_stats_lost(&info.stats, pkt);
                        est_info_group_add(&igroup, &info, EST_INFO_PRINT);
                }

                if (gb.verbose >= PCAPSTATS_VERBOSE_PKTS)
                        pcap_pkt_log(pkt);

                // update the per-packet log (if turned on)
                pcapstats_pktmap_update(pkt);
        }

        pcapstats_add_info_stats_counters(&igroup, "total", &totals);

        // we are done; make sure all remaining INFO events are sent to the estimator
        est_info_group_flush(&igroup);
        pcapstats_pktmap_done();
}

void
compute_pcap_stats(pcap_t *send, pcap_t *recv)
{
        int rc = 0;
        struct pcap_stats stats;

        memset(&stats, 0, sizeof(struct pcap_stats));

        stats.send = send;
        stats.recv = recv;

        rc = alist_init(&stats.pktlist, ALIST_NO_MUTEX, &free);
        BUG_ON(rc);

        // first level = 2^16, second level 2^2 elements
        rc = ahash_init(&stats.pkthash, 16, 2, pcap_pkt_cmp, pcap_pkt_hash);
        BUG_ON(rc);

        load_match_pcap_stats(&stats);
        walk_compute_stats(&stats);

        ahash_done(&stats.pkthash);
        alist_done(&stats.pktlist);
}

void usage();
void parse_cmdline(int argc, char **argv);

int
main(int argc, char **argv)
{

        pcap_t *send_pcap = NULL;
        pcap_t *recv_pcap = NULL;
        char err[NET_PCAP_ERRBUF_SIZE];

        est_get_host_id(gb.hostid, sizeof(gb.hostid));
	parse_cmdline(argc, argv);

        // estimator connection
        if (gb.use_estimator) {
                int rc = est_open_channel(&gb.estimator.ch, &gb.estimator.addr);
                if (rc) {
                        ERR("error: failed to open channel to estimator on %s (%s)",
                            gb.estimator.unparsed, gb.estimator.addrstr);
                }
        }

        // pcap files
        send_pcap = net_pcap_open_offline_filter(gb.send_pcap_name,
                                                 gb.filter,
                                                 err, sizeof(err));
        if (send_pcap == NULL)
                ERR("error: send trace: %s", err);

        recv_pcap = net_pcap_open_offline_filter(gb.recv_pcap_name,
                                                 gb.filter,
                                                 err, sizeof(err));
        if (recv_pcap == NULL)
                ERR("error: recv trace: %s", err);

        // main pcap processing
        compute_pcap_stats(send_pcap, recv_pcap);

        // cleanup
        pcap_close(send_pcap);
        pcap_close(recv_pcap);
        if (gb.use_estimator)
                est_close_channel(&gb.estimator.ch);

        return 0;
}

static inline void
check_params_error(char *what_not_specified, char option)
{
        printf("error: %s not specified with the -%c option\n",
               what_not_specified, option);
        printf("error: use option '-h' for help\n\n");
        exit(2);
}

void
check_params(void)
{
        if (strlen(gb.send_host) == 0)
                check_params_error("send host", 'S');

        if (strlen(gb.send_host) == 0)
                check_params_error("send trace host", 'T');

        if (strlen(gb.send_pcap_name) == 0)
                check_params_error("send pcap file", 's');
        
        if (strlen(gb.recv_pcap_name) == 0)
                check_params_error("recv pcap file", 'r');
}

void
parse_cmdline(int argc, char **argv)
{
	char short_options[] = "hr:s:f:E:H:v:S:R:T:m:";
        int found = 0;
	char c = '?';

	/* parse command line options */

	while (1) {

		c = getopt(argc, argv, short_options);

		if (c == EOF)
		    break;

		switch (c) {

                case 'v':
                        found = sscanf(optarg, "%u", &gb.verbose);
                        BUG_ON(found != 1);
                        break;

		case 'E':
                        if (gb.use_estimator)
                                ERR("error: too many -%c options", c);
                        gb.use_estimator = 1;
                        est_parse_estimator(optarg, &gb.estimator);
			break;

                case 'H':
                        strncpy(gb.hostid, optarg, sizeof(gb.hostid));
                        LOG_ERR("Overriding hostid: <%s> [actual arg <%s>]",
                            gb.hostid, optarg);
                        break;

                case 'S':
                        snprintf(gb.send_host, sizeof(gb.send_host), optarg);
                        break;

                case 'T':
                        snprintf(gb.send_trace_host, sizeof(gb.send_trace_host), optarg);
                        break;

                case 'R':
                        snprintf(gb.recv_host, sizeof(gb.recv_host), optarg);
                        break;

                case 's':
                        snprintf(gb.send_pcap_name, sizeof(gb.send_pcap_name),
                                 "%s", optarg);
                        break;

                case 'r':
                        snprintf(gb.recv_pcap_name, sizeof(gb.recv_pcap_name),
                                 "%s", optarg);
                        break;

                case 'm':
                        snprintf(gb.pktmap_name, sizeof(gb.pktmap_name), optarg);
                        break;

                case 'f':
                        snprintf(gb.filter, sizeof(gb.filter), "%s", optarg);
                        break;

		case 'h':
			usage();
			exit(0);
			break;
		    
		case '?':
                        ERRFN("invalid options");
			break;
		    
		default:
			break;
		}
	}

	check_params();
}

void
usage(void)
{
	printf("\n\nOBSOLETE --- UPDATE PLEASE! -- Mon Jun 16 11:53:23 EDT 2008\n");

	printf("\n");
	printf("    pcapstats -- extracts statistics from two tcpdump files\n");
	printf("\n");

	printf("USAGE\n");
	printf("\n");
	printf("    pcapstats -s <send-pcap> -r <recv-pcap> [options]\n");
	printf("\n");

	printf("OPTIONS\n");
	printf("        -s <send-pcap>  pcap file close to sender\n");
	printf("        -r <recv-pcap>  pcap file close to receiver\n");
	printf("        -f <filter-str> tcpdump-type filter string\n");
	printf("\n");
	printf("        -v <level>      verbosity level\n");
	printf("        -v4             one line per packet\n");
	printf("\n");

	exit(1);
}
