#ifndef __PCAPSTATS_H__
#define __PCAPSTATS_H__

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <pcap.h>
#include <time.h>

#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ether.h>

#include <math.h>

// libcommon
#include "alog.h"
#include "alist.h"
#include "atimer.h"
#include "cmp.h"
#include "ahash.h"
#include "net.h"

// libestimator
#include "libestimator.h"

// mgrp
#include "mgrphdr.h"

#define PCAPSTATS_MATCH_WITHIN_SEC      5
#define PCAPSTATS_LOAD_STEP             50
#define PCAPSTATS_VERBOSE_PKTS          4

struct pcap_key {
        int protocol;
        int ip_id;
        int transport_id;

        // not part of the hash
        // used only to match the received packet
        // the receive timestamp should be within a few seconds
        // (to deal with tcpreplay that sends the same packets in a loop)
        struct timeval ts;
};

struct pcap_pkt {
        struct pcap_key key;

        // sender

        int send_seqno;
        struct pcap_pkthdr send_pcap;

        // receiver

        short received;

        // =1 packet received back-to-back
        // >0 packet received in-order but missing pkts in between
        //    (in this case the packes missing are: recv_seqno_diff-1)
        // <0 packet received out-of-order
        int recv_seqno_diff;
        struct pcap_pkthdr recv_pcap;

        struct iphdr ip;
        union {
                struct tcphdr tcp;
                struct udphdr udp;
                struct {
                        struct bhtphdr bh;
                        int fh_valid;
                        struct mgrp_fraghdr fh;
                } mgrp;
        } th;
        short transport_len;
};

struct pcap_stats {

        alist pktlist;
        ahash pkthash;

        pcap_t *send;
        pcap_t *recv;

        // counter that assigns seqno to send packets
        // not associated with anything inside the packet
        int send_seqno;

        // keeps track of maximum seqno received
        // so that we can check for out-of-order packets
        // (monotonically increasing)
        int max_seqno_recvd;

        struct pcap_pkthdr last_recv_pcap;
        const u_char *last_recv_packet;
};

struct pcap_pkt_flags {
        // lengths include headers
        int udp_len;
        int udp_len_orig;
        int tcp_len;

        int is_mgrp;

        int is_udp;
        int is_udp_nonmgrp;
        int is_udp_nopbk;
        int is_udp_mgrp_pbk;

        int is_tcp;
        int is_tcp_nonmgrp;
        int is_tcp_nopbk;
        int is_tcp_mgrp_pbk;

        int pbk_len;
};

static inline int
pcap_pkt_outoforder(const struct pcap_pkt *pkt)
{
        return (pkt->received && pkt->recv_seqno_diff < 0);
}

static inline int
pcap_pkt_lost(const struct pcap_pkt *pkt)
{
        return (!pkt->received);
}

// very similar to spitburst.c:process_packet()
static inline int
pcap_pkt_parse(struct pcap_pkt *pkt,
               const struct pcap_pkthdr *pcap,
               const u_char *packet)
{
        int sport = 0;
        int dport = 0;
        char *error = "unknown error";
        char *transport = "unknown transport";

        const struct ethhdr *eh = (struct ethhdr *)packet;
        int ehlen = ETH_HLEN;

        const struct iphdr *iph = (struct iphdr *)(packet + ehlen);
        int iphlen = iph->ihl * 4;

        // transport portion is under ethernet and IP header
        int transport_len = pcap->len - (ehlen + iphlen);
        int transport_caplen = pcap->caplen - (ehlen + iphlen);
        const u_char *transport_header = packet + (ehlen + iphlen);

        // we need at least eth and IP headers
        BUG_ON(transport_caplen < 0);

        memset(pkt, 0, sizeof(*pkt));

        if (eh->h_proto != htons(ETHERTYPE_IP)) {
                error = "not IP";
                goto skip_error;
        }

        pkt->ip = *iph;
        pkt->key.ts = pcap->ts;
        pkt->key.ip_id = iph->id;
        pkt->key.protocol = iph->protocol;
        pkt->transport_len = transport_len;

        if (iph->protocol == IPPROTO_TCP) {
                const struct tcphdr *hdr = (struct tcphdr *)transport_header;
                BUG_ON(transport_caplen < sizeof(struct tcphdr));

                transport = "tcp";
                sport = ntohs(hdr->source);
                dport = ntohs(hdr->dest);

                pkt->th.tcp = *hdr;
                pkt->key.transport_id = hdr->seq;

        } else if (iph->protocol == IPPROTO_UDP) {
                const struct udphdr *hdr = (struct udphdr *)transport_header;
                BUG_ON(transport_caplen < sizeof(struct udphdr));

                transport = "udp";
                sport = ntohs(hdr->source);
                dport = ntohs(hdr->dest);

                pkt->th.udp = *hdr;
                pkt->key.transport_id = hdr->check;

        } else if (iph->protocol == IPPROTO_MGRP) {

                const struct bhtphdr *bh = (struct bhtphdr *)transport_header;
                const struct mgrp_fraghdr *fh = NULL;
                BUG_ON(transport_caplen < sizeof(struct bhtphdr));

                transport = "mgrp";
                sport = 0;
                dport = 0;

                pkt->th.mgrp.bh = *bh;
                pkt->key.transport_id = bh->seqno;

                // if one of the protocol chunks is fragmented,
                // try to include the fragmentation header too
                //
                // not always possible, since fragmentation header
                // is in the middle part of the packet and may not
                // have been captured by tcpdump

                fh = mgrp_get_fraghdr(bh, transport_header, transport_caplen);
                if (fh != NULL) {
                        pkt->th.mgrp.fh = *fh;
                        pkt->th.mgrp.fh_valid = 1;
                }
                        
        } else if (iph->protocol == IPPROTO_ICMP) {
                transport = "imcp";
                goto skip;

        } else {
                error = "not TCP, UDP or MGRP";
                goto skip_error;
        }

        return 1;

skip_error:
        LOG_ERR("skipping: %s: %ld.%.6ld %d",
                error,
                pcap->ts.tv_sec,
                pcap->ts.tv_usec,
                pcap->len);
skip:
        return 0;
}

static inline void
pcap_pkt_oneliner(struct pcap_pkt *pkt, char *str, int len)
{
        // we have IP and transport headers but nothing else

        if (pkt->ip.protocol == IPPROTO_TCP)
                snprintf(str, len, "tcp:%d", pkt->transport_len);

        else if (pkt->ip.protocol == IPPROTO_UDP)
                snprintf(str, len, "udp:%d:%d", pkt->transport_len, pkt->transport_len);

        else if (pkt->ip.protocol == IPPROTO_MGRP)
                // we have the MGRP header but not any of the TUN headers
                mgrp_header_oneliner(&pkt->th.mgrp.bh, pkt->transport_len,
                                     &pkt->th.mgrp.fh,
                                     str, len);

        else
                snprintf(str, len, "UNKNOWN:PROTO");

}

static inline void
pcap_pkt_log(const struct pcap_pkt *pkt)
{
        long long usec_diff = 0;
        
        if (pkt->received)
                usec_diff = atimer_timeval_diff_usec(
                                &pkt->recv_pcap.ts,
                                &pkt->send_pcap.ts);

        LOG("%8d %ld.%.6ld %ld.%.6ld %5d %8lld %d %3d",
            pkt->send_seqno,
            pkt->send_pcap.ts.tv_sec,
            pkt->send_pcap.ts.tv_usec,
            pkt->recv_pcap.ts.tv_sec,
            pkt->recv_pcap.ts.tv_usec,
            pkt->send_pcap.len,
            usec_diff,
            pkt->received,
            pkt->recv_seqno_diff);
}

static inline uint64_t
_pcap_pkt_hash(void *data, void *data_key)
{
        uint64_t hkey = 0;
        struct pcap_key *pkey = NULL;

        if (data != NULL && data_key == NULL)
                pkey = &((struct pcap_pkt *)data)->key;
        else if (data == NULL && data_key != NULL)
                pkey = (struct pcap_key *)data_key;
        else {
                BUG_ON(data != NULL && data_key != NULL);
                BUG_ON(data == NULL && data_key == NULL);
                BUG();
        }

        // A heuristic unique packet ID
        hkey = pkey->ip_id | pkey->protocol;
        hkey <<= 32;
        hkey |= pkey->transport_id;
        hkey = ahash_mix64(hkey);

        return hkey;
}

static inline int
_pcap_pkt_cmp(void *lhs_data, void *rhs_data)
{
        int diff_sec = 0;
        struct pcap_pkt *lhs = lhs_data;
        struct pcap_pkt *rhs = rhs_data;
        
        if (lhs->key.ip_id != rhs->key.ip_id)
                return cmp_ascend(lhs->key.ip_id, rhs->key.ip_id);

        if (lhs->key.protocol != rhs->key.protocol)
                return cmp_ascend(lhs->key.protocol, rhs->key.protocol);

        if (lhs->key.transport_id != rhs->key.transport_id)
                return cmp_ascend(lhs->key.transport_id, rhs->key.transport_id);

        // it is a match if it is within a few seconds (to deal with tcpreplay)
        diff_sec = abs(rhs->key.ts.tv_sec - lhs->key.ts.tv_sec);
        if (diff_sec > PCAPSTATS_MATCH_WITHIN_SEC)
                return -1;

        return 0;
}

static inline void
pcap_pkt_flags_log(struct pcap_pkt_flags *flags)
{
        LOGNR(">>> flags ulen %d ulen_orig %d tlen %d\t",
            flags->udp_len,
            flags->udp_len_orig,
            flags->tcp_len);
        LOGNR(">>> flags u %d unm %d umnopbk %d umpbk %d\t",
            flags->is_udp,
            flags->is_udp_nonmgrp,
            flags->is_udp_nopbk,
            flags->is_udp_mgrp_pbk);
        LOG(">>> flags t %d tnm %d tmnopbk %d tmpbk %d",
            flags->is_tcp,
            flags->is_tcp_nonmgrp,
            flags->is_tcp_nopbk,
            flags->is_tcp_mgrp_pbk);
}

static inline void
pcap_pkt_detect_transports(struct pcap_pkt_flags *flags,
                           const struct pcap_pkt *pkt)
{
        memset(flags, 0, sizeof(*flags));

        if (pkt->ip.protocol == IPPROTO_TCP) {
                flags->is_tcp = 1;
                flags->is_tcp_nopbk = 1;
                flags->is_tcp_nonmgrp = 1;
                flags->tcp_len = pkt->transport_len;

        } else if (pkt->ip.protocol == IPPROTO_UDP) {
                flags->is_udp = 1;
                flags->is_udp_nopbk = 1;
                flags->is_udp_nonmgrp = 1;
                flags->udp_len = pkt->transport_len;
                flags->udp_len_orig = pkt->transport_len;

        } else if (pkt->ip.protocol == IPPROTO_MGRP) {
                char errstr[200];
                int rc = mgrp_header_detect_transports(
                                        &pkt->th.mgrp.bh,
                                        pkt->transport_len,
                                        &pkt->th.mgrp.fh,
                                        &flags->is_udp,
                                        &flags->udp_len,
                                        &flags->udp_len_orig,
                                        &flags->is_tcp,
                                        &flags->tcp_len,
                                        errstr, sizeof(errstr));
                if (rc) {
                        // XXX HACK
                        // XXX Ignore errors with the MGRP proto field
                        // XXX These may be the nasty TCP packets that
                        // XXX look like MGRP packets (bug well known)
                        if (rc != MGRP_HEADER_ERROR_MGRPHDR)
                                ERRFN("%s rc(%d)", errstr, rc);
                }

                flags->is_mgrp = 1;

                if (flags->is_tcp && flags->is_udp) {
                        flags->is_udp_mgrp_pbk = 1;
                        flags->is_tcp_mgrp_pbk = 1;
                        // TCP is the only transport that piggybacks
                        flags->pbk_len = flags->tcp_len;

                } else if (flags->is_tcp) {
                        flags->is_tcp_nopbk = 1;

                } else if (flags->is_udp) {
                        flags->is_udp_nopbk = 1;
                }
        }
}

// stops at the first pkt not found in the pkthash
// packet that did not match is stored in stats->last_recv_xxxx
static inline int
match_next_recv_pkts(struct pcap_stats *stats)
{
        int pkts = 0;
        int first_time = 1;
        int retrying_recv_match = 0;

        struct pcap_pkthdr pcap;
        const u_char *packet = NULL;

        do {
                struct pcap_pkt recv_pkt;
                struct pcap_pkt *pkt = NULL;

                // retry packet from last invocation
                if (first_time) {
                        first_time = 0;
                        if (stats->last_recv_packet == NULL) {
                                continue;
                        }
                        pcap = stats->last_recv_pcap;
                        packet = stats->last_recv_packet;
                        retrying_recv_match = 1;
                }

                pcap_pkt_parse(&recv_pkt, &pcap, packet);

                pkt = ahash_lookup(&stats->pkthash, &recv_pkt.key);
                if (pkt == NULL) {
                        // check if match failed twice, bug!
                        // This means that recv trace has a packet
                        // not present in the send trace.
                        BUG_ON(retrying_recv_match);

                        // store so that we can continue next time
                        stats->last_recv_pcap = pcap;
                        stats->last_recv_packet = packet;
                        break;
                }

                pkts++;
                retrying_recv_match = 0;

                pkt->received++;
                // in case of duplicates, record only the first receipt
                if (pkt->received == 1) {
                        // timestamp
                        pkt->recv_pcap = pcap;
                        // order
                        pkt->recv_seqno_diff = pkt->send_seqno -
                                               stats->max_seqno_recvd;
                        // max_seqno_recvd is monotonically increasing
                        stats->max_seqno_recvd = cmp_max(pkt->send_seqno,
                                                         stats->max_seqno_recvd);
                }

        } while ((packet = pcap_next(stats->recv, &pcap)) != NULL);

        return pkts;
}

static inline int
load_next_send_pkts(struct pcap_stats *stats, int seconds)
{
        int pkts = 0;
        struct timeval first_ts;

        struct pcap_pkthdr pcap;
        const u_char *packet = NULL;

        while ((packet = pcap_next(stats->send, &pcap)) != NULL) {
                struct pcap_pkt *pkt = malloc(sizeof(struct pcap_pkt));
                long long diff_usec = 0;

                BUG_ON(pkt == NULL);

                pkts++;
                pcap_pkt_parse(pkt, &pcap, packet);
                pkt->send_pcap = pcap;
                pkt->send_seqno = ++stats->send_seqno;

                alist_insert_last_r(&stats->pktlist, pkt, ALIST_OWNER);
                ahash_insert(&stats->pkthash, pkt, &pkt->key);

                if (pkts == 1)
                        first_ts = pcap.ts;

                diff_usec = atimer_timeval_diff_usec(&pcap.ts, &first_ts);
                if (diff_usec > (seconds * ATIMER_USEC_PER_SEC))
                        break;
        }
        return pkts;
}
#endif
