/*
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Fri Nov  2 08:36:31 EDT 2007
 *
 * Will use the estimator library to connect to the estimator at source:port
 * and pull INFO entries;
 */

#include <stdio.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <unistd.h>

#include "mgrp/bhtp.h"
#include "mgrp/udpex.h"

/* libcommon */
#include "alog.h"
#include "net.h"
#include "tick.h"
#include "cmp.h"
#include "parse.h"

#include "libestimator.h"

#define MAX_SESSIONS	2

struct global_state {

	int sessions;
	char names[MAX_SESSIONS][200];
	struct sockaddr_in addr_src[MAX_SESSIONS];
	struct est_channel channels[MAX_SESSIONS];

	int fdmax;
	fd_set fdreads;
	
	int interval_sec;

} gb = {
        .interval_sec = 1,
};

void usage(void);
void parse_cmdline(int argc, char **argv);

static inline void
include_rfd(int fd)
{
	gb.fdmax = cmp_max(gb.fdmax, fd + 1);
	FD_SET(fd, &gb.fdreads);
}

static inline void
exclude_rfd(int fd)
{
	FD_CLR(fd, &gb.fdreads);
	/* find the new max */
	while (gb.fdmax > 0) {
		if (FD_ISSET(gb.fdmax - 1, &gb.fdreads))
			break;
		gb.fdmax--;
	}
}

void
create_sessions()
{
	int i = 0;

	gb.fdmax = 0;
	FD_ZERO(&gb.fdreads);

	/* create trafgen sessions on src nodes */
	for (i = 0; i < gb.sessions; i++) {
                int rc = est_open_channel(&gb.channels[i], &gb.addr_src[i]);
                if (rc)
                        ERR("failed to open channel for session %d", i);
                include_rfd(est_channel_socket(&gb.channels[i]));
	}
}

void
destroy_sessions()
{
	int i = 0;

	/* create trafgen sessions on src nodes */
	for (i = 0; i < gb.sessions; i++) {
                exclude_rfd(est_channel_socket(&gb.channels[i]));
                est_close_channel(&gb.channels[i]);
	}
}


#define MAX_INFO_TO_PULL        10

void
info_pull()
{
	int i = 0;

	for (i = 0; i < gb.sessions; i++) {
                int rc = 0;
                int j = 0;
                int pulled = 0;
                struct est_info info[MAX_INFO_TO_PULL];

                rc = est_pull_info_entries(&gb.channels[i],
                                           MAX_INFO_TO_PULL,
                                           info,
                                           &pulled);
                if (rc)
                        ERR("failed to pull info entries for %d", i);
                
                if (pulled == 0)
                        LOG("no info entries for %d", i);
                for (j = 0; j < pulled; j++) {
                        est_info_log(&info[j]);
                }
	}
}

int
main (int argc, char **argv)
{
	parse_cmdline(argc, argv);
        create_sessions();
        while (1) {
                info_pull();
                if (gb.interval_sec == 0)
                        break;
                sleep(gb.interval_sec);
        }
        destroy_sessions();
	return 0;
}

void
check_params(void)
{
	if (gb.sessions <= 0) {
		printf("error: no sessions specified with the -P option\n");
		printf("error: see 'trafgen -h' for more details\n\n");
		exit(2);
	}
}

void
parse_host_port(const char *optstr)
{
        int found = 0;
        char error[100];
	char tmp[100];
	int j = gb.sessions++;

	strncpy(gb.names[j], optarg, sizeof(gb.names[j]));

        found = net_resolve_host_port(optarg, &gb.addr_src[j],
                                      error, sizeof(error));
        if (!found) {
                ERR("failed to parse host:port '%s': %s", optarg, error);
        }
        if (gb.addr_src[j].sin_port == 0)
                gb.addr_src[j].sin_port = htons(EST_DEFAULT_CONTROL_TCP_PORT);

	net_sockaddr_str(tmp, sizeof(tmp), &gb.addr_src[j]);
	LOG("[%d] %s: %s", j, gb.names[j], tmp);
}

void
parse_cmdline(int argc, char **argv)
{
	char short_options[] = "hP:i:";

	int found = 0;
	char c = '?';

	/* parse command line options */

	while (1) {

		c = getopt(argc, argv, short_options);

		if (c == EOF)
		    break;

		switch (c) {

		case 'P':
			if (gb.sessions >= MAX_SESSIONS)
				ERR("reached max sessions %d", gb.sessions);
			parse_host_port(optarg);
			break;

		case 'i':
                        found = sscanf(optarg, "%u", &gb.interval_sec);
                        if (found != 1)
                                ERR("invalid interval: '%s'", optarg);
			break;

		case 'h':
			usage();
			exit(0);
			break;
		    
		case '?':
			break;
		    
		default:
			break;
		}
	}

	check_params();
}

void
usage(void)
{
	printf("\n");
	printf("        pinfo -- drains all INFO events from the estimator queue at <src:port>\n");
	printf("\n");

	printf("USAGE\n");
	printf("\n");
	printf("        pinfo -P <src[:port]> [options]\n");
	printf("\n");

	printf("OPTIONS\n");
	printf("        -i <sec>        interval in seconds\n");
	printf("\n");
	exit(1);
}


