/*
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Thu Nov  8 17:49:01 EST 2007
 *
 * Generates plots out of INFO entries:
 *      - connects to estimators
 *      - pulls INFO entries
 *      - dumps INFO entries in data file
 *      - invokes gnuplot on data file
 *      - replaces the plot png file
 *      - png file can be served by web server
 */

#include <stdio.h>
#include <pthread.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <unistd.h>

#include "mgrp/bhtp.h"
#include "mgrp/udpex.h"

/* libcommon */
#include "alog.h"
#include "net.h"
#include "tick.h"
#include "cmp.h"
#include "parse.h"

#include "libestimator.h"

#define MAX_SESSIONS	        10
#define MAX_INFO_TO_PULL        200

struct session_data {
        /* input */
        int valid;
        int id;
        char name[200];
        char estimator[50];
        struct sockaddr_in src;
        int interval_sec;

        /* operational */
        pthread_t self;
        struct est_channel ch;

        int info_max;
        int info_count;
        struct est_info *info;
};

struct global_state {

	int sessions;
        struct session_data sdata[MAX_SESSIONS];

	int interval_sec;

} gb = {
        .interval_sec = 1,
};

void usage(void);
void parse_cmdline(int argc, char **argv);

/* This thread keeps retrieving INFO events from one estimator.
 * The events are places in appropriate queues.
 */

void
process_info_entries(struct session_data *data)
{
        int j = 0;
        
        if (data->info_count == 0) {
                LOG("[%d] no info entries from estimator %s",
                    data->id, data->estimator);
                return;
        }

        for (j = 0; j < data->info_count; j++) {
                est_info_log(&data->info[j]);
        }
        data->info_count = 0;
}


void *
session_thread(void *arg)
{
        int rc = 0;
        struct session_data *data = (struct session_data *)arg;

        data->self = pthread_self();

        rc = est_open_channel(&data->ch, &data->src);
        if (rc) {
                LOG("[%d] error: "
                    "failed to open channel to estimator %s",
                    data->id, data->estimator);
                goto done;
        }

        data->info_max = MAX_INFO_TO_PULL;
        data->info_count = 0;
        data->info = malloc(sizeof(struct est_info) * data->info_max);
        BUG_ON(data->info == NULL);

        while (1) {
                rc = est_pull_info_entries(&data->ch,
                                           data->info_max,
                                           data->info,
                                           &data->info_count);
                if (rc) {
                        LOG("[%d] error: "
                            "failed to pull info entries from estimator %s",
                            data->id, data->estimator);
                        goto done;
                }

                process_info_entries(data);

                if (data->interval_sec == 0)
                        break;
                sleep(data->interval_sec);
        }

done:
        free(data->info);
        est_close_channel(&data->ch);
        
        return NULL;
}



int
main (int argc, char **argv)
{
        int i = 0;
        int rc = 0;

        /* initialize all session data */
        /* parse_cmdline() will fill some of them */
        for (i = 0; i < MAX_SESSIONS; i++) {
                struct session_data *data = &gb.sdata[i];

                memset(data, 0, sizeof(struct session_data));
                data->valid = 1;
                data->id = i;
        }

	parse_cmdline(argc, argv);

        /* start the threads */
        for (i = 0; i < gb.sessions; i++) {
                pthread_t thread;
                struct session_data *data = &gb.sdata[i];

                data->interval_sec = gb.interval_sec;

                rc = pthread_create(&thread, NULL, session_thread, data);
                SYSERR_ON(rc);
        }

        while(1) {
            sleep(10);
        }
	return 0;
}

void
check_params(void)
{
	if (gb.sessions <= 0) {
		printf("error: no sessions specified with the -P option\n");
		printf("error: use option '-h' for help\n\n");
		exit(2);
	}
}

void
parse_host_port(const char *optstr)
{
        int found = 0;
        char error[100];
	int j = gb.sessions++;
        struct session_data *data = &gb.sdata[j];

        BUG_ON(j != data->id);

	strncpy(data->name, optarg, sizeof(data->name));

        found = net_resolve_host_port(optarg, &data->src,
                                      error, sizeof(error));
        if (!found) {
                ERR("failed to parse host:port '%s': %s", optarg, error);
        }
        if (data->src.sin_port == 0)
                data->src.sin_port = htons(EST_DEFAULT_CONTROL_TCP_PORT);

	net_sockaddr_str(data->estimator, sizeof(data->estimator), &data->src);
	LOG("[%d] %s: %s", data->id, data->name, data->estimator);
}

void
parse_cmdline(int argc, char **argv)
{
	char short_options[] = "hP:i:";

	int found = 0;
	char c = '?';

	/* parse command line options */

	while (1) {

		c = getopt(argc, argv, short_options);

		if (c == EOF)
		    break;

		switch (c) {

		case 'P':
			if (gb.sessions >= MAX_SESSIONS)
				ERR("reached max sessions %d", gb.sessions);
			parse_host_port(optarg);
			break;

		case 'i':
                        found = sscanf(optarg, "%u", &gb.interval_sec);
                        if (found != 1)
                                ERR("invalid interval: '%s'", optarg);
			break;

		case 'h':
			usage();
			exit(0);
			break;
		    
		case '?':
			break;
		    
		default:
			break;
		}
	}

	check_params();
}

void
usage(void)
{
	printf("\n");
	printf("        pinfo -- drains all INFO events from the estimator queue at <src:port>\n");
	printf("\n");

	printf("USAGE\n");
	printf("\n");
	printf("        pinfo -P <src[:port]> [options]\n");
	printf("\n");

	printf("OPTIONS\n");
	printf("        -i <sec>        interval in seconds\n");
	printf("\n");
	exit(1);
}


