/*
 * pktrain.c
 *
 * Created:     Mon Feb  6 16:12:14 EST 2006
 *
 * Authors:     Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <unistd.h>

#include "mgrp/bhtp.h"
#include "mgrp/udpex.h"

#include "net.h"
#include "err.h"
#include "tick.h"

#include "libestimator.h"

#include "pktrain.h"

/* HACK: copied from linux/time.h */
/* 	 is this in a user-space include file? */

#define NSEC_PER_SEC	1000000000L
#define USEC_PER_SEC    1000000L
#define MSEC_PER_SEC    1000L

#define NSEC_PER_MSEC	1000000L
#define USEC_PER_MSEC   1000L

#define NSEC_PER_USEC	1000L

static inline long long timeval_to_ns(const struct timeval *tv)
{                                    
	return ((long long) tv->tv_sec * NSEC_PER_SEC) +       
		tv->tv_usec * NSEC_PER_USEC;                            
}                                                 

static inline long long timeval_to_usec(const struct timeval *tv)
{                                    
	return ((long long) tv->tv_sec * USEC_PER_SEC) + tv->tv_usec;
}                                                 

struct pktrain_state {
	int is_receiver;
	int receiver_port;
	int sender_port;
        char receiver_host[100];
	struct sockaddr_in receiver_addr;
        int debug_gap;
        int examine_gap;
        unsigned int duration;
	unsigned int gap;
	unsigned int train;
	int print_packet;
	int print_results;
	int len;                        // includes everyhting below IP
	int len_step;
	unsigned int rtt_estimate;
	long snd_pkts;
	long rcv_pkts;
	int ping;
        int transport_overhead;
	int overhead;
	struct timeval started;
	int total_trains;
	int use_kernel;
	int use_mgrp;
        int use_so_timestamp;
        unsigned int train_gap_msec;

        char hostid[EST_INFO_IDLEN];
        struct est_estimator estimator;
};

static struct pktrain_state pk = {
	.is_receiver	= -1,
	.receiver_port	= 47918,
	.sender_port	= 47919,
	.gap		= 10 * USEC_PER_MSEC,
	.train		= 10,
	.print_packet	= 0,
	.len		= 1250,
	.rtt_estimate	= 20,
	.total_trains	= 1,
	.len_step	= 0,
        .train_gap_msec = 1 * MSEC_PER_SEC,
        .use_so_timestamp = 1,
};

#define PKTRAIN_MAX_TRAINLEN    1000

struct pktrain_gap_stats {
        // in microsec
        int gap_desired;
        int gap_actual;
        int gap_effective;
        int gap_slept;

        // number of loops
        int gap_busywait_loops;
};

static inline void
pktrain_gap_stats_log(struct pktrain_gap_stats *stats, int count)
{
        int i = 0;

        LOG("");
        for (i = 0; i < count; i++) {
                LOG(" %3d | desired %6d | actual %6d | effective %6d | slept %6d | busywait %8d",
                    i,
                    stats[i].gap_desired,
                    stats[i].gap_actual,
                    stats[i].gap_effective,
                    stats[i].gap_slept,
                    stats[i].gap_busywait_loops);
        }

}


void usage(void);
void parse_cmdline(int argc, char **argv);

void receiver(void);
void sender(void);

void
pktrain_fill_param_stats(struct est_info_stats *stats)
{
        struct timeval now;
        const char *src_host = pk.hostid;
        const char *dst_host = pk.receiver_host;

        // only for sender
        BUG_ON(pk.is_receiver);

        gettimeofday(&now, NULL);

        // Fill the header: key and timestamp.
        //
        // Infoplot will create a separate data file whose name has the format:
        //
        //      stats-pktrain.<src>.<dst>.dat
        //
        // where type="stats" and subtype="pktrain".
        //
        est_info_stats_init(stats, "pktrain");
        est_fill_info_srcdst(&stats->hdr, src_host, dst_host);
        est_fill_info_timestamp(&stats->hdr, pk.hostid, &now);

        // Actual stats
        snprintf(stats->tag, sizeof(stats->tag), "params");

        // packets per train
        est_info_stats_setint(stats, 0, pk.train);

        // packet size
        est_info_stats_setint(stats, 1, pk.len);

        // gap between packet sends (usec)
        est_info_stats_setint(stats, 2, pk.gap);

        // number of trains to send
        est_info_stats_setint(stats, 3, pk.total_trains);

        // pause between trains (usec)
        est_info_stats_setint(stats, 4, pk.train_gap_msec * USEC_PER_MSEC);
}

void
pktrain_send_param_stats()
{
        int rc = 0;
        char infostr[300];
        struct est_info info;

        if (!pk.estimator.valid)
                return;

        pktrain_fill_param_stats(&info.stats);
        est_info_oneline_str(infostr, sizeof(infostr), &info);
        LOG("%s", infostr);

        rc = est_push_info_entries(&pk.estimator.ch, &info, 1);
        BUG_ON(rc);
}

int
main (int argc, char **argv)
{
	parse_cmdline(argc, argv);

	gettimeofday(&pk.started, NULL);

	pk.overhead	= 14 +	// ethernet
			  20; 	// IP

        pk.transport_overhead = 0;

        if (pk.use_mgrp) {
                printf("** Using MGRP\n");
                printf("** Using UDPex (UDP over MGRP)\n");
                pk.transport_overhead += sizeof(struct bhtphdr) +
                                         sizeof(struct udpexhdr);
        } else {
                printf("** Using classic UDP\n");
                pk.transport_overhead += 8;
        }

        printf("** overhead (ethernet + IP) %d bytes\n", pk.overhead);
        printf("** transport overhead (UDP or UDP/MGRP) %d bytes\n", pk.transport_overhead);

        est_get_host_id(pk.hostid, sizeof(pk.hostid));
        if (pk.estimator.valid) {
                int rc = est_open_channel(&pk.estimator.ch, &pk.estimator.addr);
                if (rc) {
                        ERR("error: failed to open channel to estimator on %s (%s)",
                            pk.estimator.unparsed, pk.estimator.addrstr);
			exit(1);
                }
        }

	if (pk.is_receiver)
		receiver();
	else {
                BUG_ON((pk.len + pk.overhead) > 1500);
                pktrain_send_param_stats();
		sender();
		if (pk.len_step) {
			while (1) {
				pk.len += pk.len_step;
				if ((pk.len + pk.overhead) > 1500)
					break;
				sender();
			}
		}
        }

        if (pk.estimator.valid) {
                est_close_channel(&pk.estimator.ch);
        }

	return 0;
}

static inline void
send_pong(int sock, void *buf, int bufsize, struct sockaddr_in *to)
{
	int rc = 0;
	int sent = 0;
	struct pktrain_payload *pd = (struct pktrain_payload *)buf;

	pk.snd_pkts++;

	memcpy(&pd->tstamp_ping, &pd->tstamp, sizeof(pd->tstamp_ping));
	pd->count_ping = pd->count;

	pd->tag[0] = 'E';
	pd->tag[1] = 'C';
	pd->tag[2] = 'O';
	pd->tag[3] = 0;
	pd->type = PKTRAIN_PONG;
	pd->count = pk.snd_pkts;

	rc = gettimeofday(&pd->tstamp, NULL);
	BUG_ON(rc);

	sent = sendto(sock, buf, bufsize, 0, (struct sockaddr *)to, sizeof(*to));
	SYSERR_ON(sent == -1);
	BUG_ON(sent != bufsize);

	if (pk.print_packet) {
		printf("SEND size %d: ", sent);
		pktrain_print_pd(pd);
		printf("\n");
	}
}

static inline void
receive_pong(int sock, int expect_count)
{
	int rc = 0;
	fd_set rfds;
	struct timeval tv;
	struct timeval now;
	long diff = 0;
	long epoch_diff = 0;

	int size = 0;
	char buf[2000];
	struct pktrain_payload *pd = (struct pktrain_payload *)buf;

	FD_ZERO(&rfds);
	FD_SET(sock, &rfds);

	tv.tv_sec = 0;
	tv.tv_usec = (3 * pk.rtt_estimate) * 1000;

	rc = select(sock+1, &rfds, NULL, NULL, &tv);
	SYSERR_ON(rc == -1);

	if (rc == 0) {
		printf("PONG error: lost\n");
		return;
	}

	size = recv(sock, buf, sizeof(buf), 0);
	SYSERR_ON(size == -1);

	if (expect_count != pd->count_ping) {
		printf("PONG error: out of order, expected %d got %d\n",
			expect_count, pd->count_ping);
		return;
	}

	diff = tick_now_udiff_past(&pd->tstamp_ping, &now);
	epoch_diff = tick_tv_udiff(&now, &pk.started);

	printf("PONG %.6f %d %d %ld\n",
		(0.000001 * epoch_diff),
		pk.len, pk.len + pk.overhead, diff);

	pk.rcv_pkts++;

	if (pk.print_packet) {
		printf("RECV size %d: ", size);
		pktrain_print_pd(pd);
		printf("\n");
	}
}

static inline void
keep_serving(int sock)
{
	static struct bhtp_timestamps prev_stamps;
	static int prev_count = -1;
	static struct timeval prev_pdstamp;
	static struct timeval prev_sys_tstamp;

	int size = 0;

	struct bhtp_timestamps *stamps = NULL;
        struct timeval *sys_tstamp = NULL;
	struct sockaddr_in from;

	char buf[2000];
	struct iovec iov = {
		.iov_base = buf,
		.iov_len = sizeof(buf),
	};

	struct cmsghdr *cmsg = NULL;
	char ctrl[CMSG_SPACE(sizeof(struct bhtp_timestamps))];

	struct msghdr msg = {
		.msg_name = (void *)&from,
		.msg_namelen = sizeof(from),
		.msg_iov = &iov,
		.msg_iovlen = 1,
		.msg_control = (void *)ctrl,
		.msg_controllen = sizeof(ctrl),
		.msg_flags = 0,
	};

	struct pktrain_payload *pd = (struct pktrain_payload *)buf;

	memset(buf, 0, sizeof(buf));
	memset(ctrl, 0, sizeof(ctrl));
	memset(&from, 0, sizeof(from));

	size = recvmsg(sock, &msg, 0);
	SYSERR_ON(size == -1);
	BUG_ON(msg.msg_namelen != sizeof(from));

	pk.rcv_pkts++;

	cmsg = CMSG_FIRSTHDR(&msg);

	if (cmsg &&
	    cmsg->cmsg_level == SOL_MGRP &&
	    cmsg->cmsg_type == UDPEX_TSTAMP) {
		stamps = (struct bhtp_timestamps *)CMSG_DATA(cmsg);

        } else if (cmsg &&
                   cmsg->cmsg_level == SOL_SOCKET &&
                   cmsg->cmsg_type == SO_TIMESTAMP) {
                sys_tstamp = (struct timeval *)CMSG_DATA(cmsg);

        } else if (cmsg) {
                printf(">>> CMSG len %d level %d type %d timeval %d\n",
                       cmsg->cmsg_len,
                       cmsg->cmsg_level,
                       cmsg->cmsg_type,
                       sizeof(struct timeval));
        }

	if (pk.rcv_pkts % 1000 == 0)
		printf("RECV pkts %ld\n", pk.rcv_pkts);

	if (pk.print_packet) {
		printf("RECV size %d: ", size);
		pktrain_print_pd(pd);
		printf("\n");

		if (stamps != NULL && pd->count == prev_count+1) {
			printf("diff count %d ", pd->count);
			bhtp_print_stamp_diff_usec(stamps, &prev_stamps);
			printf("dapp %llu",
				timeval_to_usec(&pd->tstamp) -
				timeval_to_usec(&prev_pdstamp));
			printf("\n");
			prev_stamps = *stamps;
		}
		if (sys_tstamp != NULL && pd->count == prev_count+1) {
			printf("diff count %d ", pd->count);
			printf("dsys %llu ",
				timeval_to_usec(sys_tstamp) -
				timeval_to_usec(&prev_sys_tstamp));
			printf("dapp %llu",
				timeval_to_usec(&pd->tstamp) -
				timeval_to_usec(&prev_pdstamp));
			printf("\n");
			prev_sys_tstamp = *sys_tstamp;
		}
		prev_pdstamp = pd->tstamp;
		prev_count = pd->count;
	}

	if (pd->type & PKTRAIN_PING)
		send_pong(sock, buf, size, &from);
}


void
receiver(void)
{
	int rc = 0;
	int sock = 0;
	int on = 0;

	struct sockaddr_in listen_addr = {
		.sin_family		= AF_INET,
		.sin_addr.s_addr	= htonl(INADDR_ANY),
		.sin_port		= htons(pk.receiver_port),
	};

        sock = socket(AF_INET, SOCK_DGRAM, pk.use_mgrp ? IPPROTO_MGRP : 0);
	SYSERR_ON(sock < 0);

	rc = bind(sock, (struct sockaddr *)&listen_addr, sizeof(listen_addr));
	SYSERR_ON(rc);

	printf("** Listening on UDP port %d\n", pk.receiver_port);

        if (pk.use_kernel) {
                printf("** Using MGRP timestamps\n");
		on = 1;
		rc = setsockopt(sock, SOL_MGRP, UDPEX_TSTAMP, &on, sizeof(on));
		SYSERR_ON(rc);
	} else if (pk.use_so_timestamp) {
                printf("** Using SO_TIMESTAMP timestamps\n");
		on = 1;
		rc = setsockopt(sock, SOL_SOCKET, SO_TIMESTAMP, &on, sizeof(on));
		SYSERR_ON(rc);
        } else {
                printf("** Using application timestamps\n");
        }

	while (1)
		keep_serving(sock);

	rc = close(sock);
	SYSERR_ON(rc);
}

static inline int
inside_pktrain(int train_pkt)
{
	return (train_pkt < pk.train && train_pkt > 1);
}

static inline int
first_in_pktrain(int train_pkt)
{
	return (train_pkt == 1);
}

static inline int
last_in_pktrain(int train_pkt)
{
	return (train_pkt == pk.train);
}


static inline void
fill_payload(struct pktrain_payload *pd, int order)
{
	int rc = 0;

	memset(pd, 0, sizeof(*pd));
	pd->tag[0] = 'T';
	pd->tag[1] = 'A';
	pd->tag[2] = 'G';
	pd->tag[3] = 0;
	pd->type = pk.ping ? PKTRAIN_PING : PKTRAIN_ONEWAY;
        pd->order = order;
	pd->count = pk.snd_pkts;
        pd->gap_usec = pk.gap;
        pd->pkts_in_train = pk.train;
	rc = gettimeofday(&pd->tstamp, NULL);
	BUG_ON(rc);
}

static inline int
pktrain_finished(int sent_trains)
{
        if (pk.total_trains && sent_trains >= pk.total_trains)
                return 1;
        if (pk.duration) {
                long long elapsed_usec = tick_now_udiff_past(&pk.started, NULL);
                if (elapsed_usec >= pk.duration * TICK_USEC_PER_SEC)
                        return 1;
        }
        return 0;
}

static inline int
keep_sending(int sock)
{
	static int train = 0;
	static int total_trains = 0;
        static struct timeval last_send_ts;
        static struct timeval after_gap;
        static struct pktrain_gap_stats stats[PKTRAIN_MAX_TRAINLEN];

	char buf[2000];
	struct pktrain_payload *pd = (struct pktrain_payload *)buf;

	int sent = 0;
	int bufsize = 0;

	train++;
	pk.snd_pkts++;

	fill_payload(pd, train);

	bufsize = pk.len - pk.transport_overhead;
	BUG_ON(bufsize < sizeof(struct pktrain_payload));

        gettimeofday(&last_send_ts, NULL);
        if (!pk.debug_gap) {
                sent = send(sock, buf, bufsize, 0);
                SYSERR_ON(sent != bufsize);
        }

	if (pk.print_packet) {
		printf("SEND size %d: ", sent);
		pktrain_print_pd(pd);
		printf("\n");
	}

	if (pk.ping)
		receive_pong(sock, pd->count);

	if (train < pk.train) {
		int actual = 0;
                
                // @train is number of probe in train, starts at 1
                if (train == 1) {
                        // first probe never has a gap
                        memset(&stats[0], 0, sizeof(struct pktrain_gap_stats));
                }

		/* generate packet gap inside packet train */
                actual = tick_usleep_w_feedback(pk.gap,
                                                &last_send_ts,
                                                &after_gap,
                                                &stats[train].gap_effective,
                                                &stats[train].gap_slept,
                                                &stats[train].gap_busywait_loops);

                stats[train].gap_desired = pk.gap;
                stats[train].gap_actual = actual;

	} else {
		total_trains++;

                if (pk.debug_gap || pk.examine_gap)
                        pktrain_gap_stats_log(stats, train);

                if (pktrain_finished(total_trains))
                        return 0;

		/* gap between trains */
		tick_msleep(pk.train_gap_msec);

		train = 0;
	}
	return 1;
}

static inline int
tell_results(struct msghdr *msg)
{
	struct udpex_probe *probe = NULL;
	struct cmsghdr *cmsg = NULL;

	cmsg = CMSG_FIRSTHDR(msg);

	if (cmsg == NULL ||
	    cmsg->cmsg_level != SOL_MGRP ||
	    cmsg->cmsg_type != UDPEX_PROBE ||
	    cmsg->cmsg_len < CMSG_LEN(sizeof(struct udpex_probe))) {

/* 		printf("ERROR: control SOL_MGRP, UDPEX_PROBE not found"); */
		return 0;
	}
		
	probe = (struct udpex_probe *)CMSG_DATA(cmsg);
/* 	printf("tell_results: flag %d\n", probe->tell_results); */
	return probe->tell_results;
	
}

static inline void
fill_msg(struct msghdr *msg, int train_pkt, int gap, int pdlen, int padlen)
{
	struct udpex_probe *probe = NULL;
	struct cmsghdr *cmsg = NULL;

	cmsg = CMSG_FIRSTHDR(msg);
	cmsg->cmsg_level = SOL_MGRP;
	cmsg->cmsg_type = UDPEX_PROBE;
	cmsg->cmsg_len = CMSG_LEN(sizeof(struct udpex_probe));

	probe = (struct udpex_probe *)CMSG_DATA(cmsg);
	memset(probe, 0, sizeof(*probe));

	if (inside_pktrain(train_pkt)) {
		DEB("pk: internal %d\n", train_pkt);
		/* internal packet in train */
		probe->barrier = 1;
		probe->gap_usec = gap;
		probe->pad = padlen;

	} else if (first_in_pktrain(train_pkt)) {
		DEB("pk: first %d\n", train_pkt);
		/* first packet in train */
		probe->barrier = 1;
		probe->gap_usec = 0;
		probe->pad = padlen;

	} else {
		/* last packet in train */
		DEB("pk: last %d\n", train_pkt);
		probe->barrier = 0;
		probe->gap_usec = gap;
		probe->pad = padlen;
		probe->can_wait_for_usec = 10000;
		probe->tell_results = 1;
	}
}

static inline int
keep_sending_kernel_probes(int sock)
{
	static int pkt = 0;
	static int total_trains = 0;

	char buf[2000];
	struct iovec iov = {
		.iov_base = buf,
		.iov_len = 0,
	};

	char ctrl[CMSG_SPACE(sizeof(struct udpex_probe))];

	struct msghdr msg = {
		.msg_name = NULL,
		.msg_namelen = 0,
		.msg_iov = &iov,
		.msg_iovlen = 1,
		.msg_control = (void *)ctrl,
		.msg_controllen = sizeof(ctrl),
		.msg_flags = 0,
	};

	struct pktrain_payload *pd = (struct pktrain_payload *)buf;

	int sent = 0;
	int padlen = 0;
	int pdlen = 0;

	pkt++;
	pk.snd_pkts++;

	memset(buf, 0, sizeof(buf));
	memset(ctrl, 0, sizeof(ctrl));

	pdlen = sizeof(struct pktrain_payload);
	padlen = (pk.len - pk.transport_overhead) - pdlen;
	BUG_ON(padlen < 0);

	iov.iov_len = pdlen;
	fill_msg(&msg, pkt, pk.gap, pdlen, padlen);
	fill_payload(pd, pkt);

	sent = sendmsg(sock, &msg, 0);
	SYSERR_ON(sent == -1);

	if (pk.print_packet) {
		printf("SEND size %d pad %d: ", sent, padlen);
		pktrain_print_pd(pd);
		printf("\n");
	}

	BUG_ON(sent != pdlen);

	if (pk.ping)
		receive_pong(sock, pd->count);

	/* packet gap inside packet train is taken care by kernel */

	if (last_in_pktrain(pkt)) {
		total_trains++;

                if (pk.print_results) {
                        struct udpex_probe_results res = {.sent = 0};
                        char tmpstr[200];
                        int rc = ioctl(sock, UDPEX_IOC_PROBE_RESULTS, &res);
                        if (rc)
                                printf("ERROR: IOC_PROBE_RESULTS %d\n", rc);
                        else
                                printf("train %d results: %s\n",
                                        total_trains,
                                        udpex_probe_results_str(tmpstr,
                                                sizeof(tmpstr), &res));
                }

                if (pktrain_finished(total_trains))
                        return 0;

		/* gap between trains */
		tick_msleep(pk.train_gap_msec);

		pkt = 0;
	}
	return 1;
}

void
sender(void)
{
	int rc = 0;
	int sock = 0;

	struct sockaddr_in from_addr = {
		.sin_family		= AF_INET,
		.sin_addr.s_addr	= htonl(INADDR_ANY),
		.sin_port		= htons(pk.sender_port),
	};

	struct sockaddr_in to_addr = {
		.sin_family		= AF_INET,
		.sin_addr.s_addr	= pk.receiver_addr.sin_addr.s_addr,
		.sin_port		= htons(pk.receiver_port),
	};

        sock = socket(AF_INET, SOCK_DGRAM, pk.use_mgrp ? IPPROTO_MGRP : 0);
	SYSERR_ON(sock < 0);

	if (pk.sender_port != 0) {
		rc = bind(sock, (struct sockaddr *)&from_addr, sizeof(from_addr));
		SYSERR_ON(rc);
	}

	rc = connect(sock, (struct sockaddr *)&to_addr, sizeof(to_addr));
	SYSERR_ON(rc);

	if (pk.use_kernel) {
                printf("** Using MGRP barriers\n");
		while (keep_sending_kernel_probes(sock));
	} else {
		while (keep_sending(sock));
        }

	rc = close(sock);
	SYSERR_ON(rc);
}

void
parse_cmdline(int argc, char **argv)
{
	/*
	 * Valid Options:
	 *
	 * -h			Help screen
	 */

	char short_options[] = "RD:vdmhrs:g:p:t:T:l:qen:i:xkE:";

	int found = 0;
	char c = '?';
	char error[200];

	/* parse command line options */

	while (1) {

		c = getopt(argc, argv, short_options);

		if (c == EOF)
		    break;

		switch (c) {

		case 'x':
			pk.use_mgrp = 1;
			break;

		case 'k':
			pk.use_kernel = 1;
			break;

		case 'q':
                        pk.print_results = 0;
			pk.print_packet = 0;
			break;

		case 'v':
                        pk.print_results = 1;
			pk.print_packet = 1;
                        pk.examine_gap = 1;
			break;

		case 'R':
                        pk.print_results = 1;
			break;

		case 'r':
			pk.is_receiver = 1;
			break;

		case 'e':
			pk.ping = 1;
			break;

		case 'd':
			pk.debug_gap = 1;
			break;

		case 'm':
			pk.examine_gap = 1;
			break;

		case 's':
			pk.is_receiver = 0;
                        strncpy(pk.receiver_host, optarg,
                                sizeof(pk.receiver_host));
			found = net_resolve_host(pk.receiver_host,
				&pk.receiver_addr,
				error, sizeof(error));

			BUG_ON(!found);
			break;

		case 'g':
			found = sscanf(optarg, "%u", &pk.gap);
			BUG_ON(found != 1);
			break;

		case 'D':
			found = sscanf(optarg, "%u", &pk.duration);
			BUG_ON(found != 1);
                        pk.total_trains = 0;
			break;

		case 'n':
			found = sscanf(optarg, "%d", &pk.total_trains);
			BUG_ON(found != 1);
			break;

		case 'l':
			found = sscanf(optarg, "%u", &pk.len);
			BUG_ON(found != 1);
			break;

		case 'i':
			found = sscanf(optarg, "%u", &pk.len_step);
			BUG_ON(found != 1);
			break;

		case 't':
			found = sscanf(optarg, "%u", &pk.train);
			BUG_ON(found != 1);
			break;

		case 'T':
			found = sscanf(optarg, "%u", &pk.train_gap_msec);
			BUG_ON(found != 1);
			break;

		case 'p':
			found = sscanf(optarg, "%u:%u",
                                       &pk.receiver_port, &pk.sender_port);
			BUG_ON(found != 2);
			break;
		    
                case 'E':
                        est_parse_estimator(optarg, &pk.estimator);
                        break;

		case 'h':
			usage();
			exit(0);
			break;
		    
		case '?':
			BUG();
			break;
		    
		default:
			BUG();
			break;
		}
	}

        if (pk.is_receiver == -1) {
                LOG("error: receiver or sender?");
                LOG("Run 'pktrain -h' for help");
                exit(1);
        }
}

void
usage(void)
{
	printf("\n");
	printf("pktrain -- generates UDP packet trains\n");
	printf("\n");

	printf("USAGE\n");
	printf("\n");
	printf("    [receiver]  pktrain -r [options]\n");
	printf("    [ sender ]  pktrain -s <receiver> [options]\n");
	printf("\n");

	printf("SENDER OPTIONS\n");
	printf("    -g <usec>   Time gap between packets in the packet train (in usec)\n");
	printf("    -t <pkts>   Number of packets in each packet train\n");
	printf("    -T <msec>   Time gap between packet trains (in msec)\n");
	printf("    -l <bytes>  Size of probe payload (below IP header)\n");
	printf("\n");
	printf("    -n <trains> Number of packet trains to send\n");
	printf("    -D <sec>    Stop sending packet trains after <sec> seconds\n");
	printf("\n");
	printf("    -i <step>   (bytes) If non-zero, keeps sending trains\n");
	printf("                and every time increases the probe payload\n");
	printf("                by <step> bytes (up to max packet length)\n");
	printf("\n");
	printf("    -d          Debug gap generation (do not send any packets)\n");
	printf("    -m          Examine gap generation (prints summary after each train)\n");
	printf("    -R          Show MGRP probe results (prints summary after each train)\n");
	printf("\n");

	printf("COMMON OPTIONS\n");
	printf("    -x          Turns on MGRP (UDPex socket instead of UDP)\n");
	printf("    -k          Turns on kernel probes (barriers; needs -x)\n");
	printf("\n");
	printf("    -v          Be verbose! (also turns on -m)\n");
	printf("    -q          Be quiet! shhh!\n");
	printf("    -h          Help!\n");
	printf("\n");

	exit(1);
}

