/*
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Fri Feb 29 14:35:30 MST 2008
 *
 * Sends to an estimator INFO/post entries.
 *
 * Every INFO/post entries triggers a post-processing
 * step in infoplot (to create plots across data files).
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#include "net.h"
#include "atimer.h"
#include "astring.h"
#include "libestimator.h"

struct global_state {
        char hostid[EST_INFO_IDLEN];

        int estimator_count;
        struct est_estimator estimator;

        char src[EST_INFO_IDLEN];
        char dst[EST_INFO_IDLEN];
        char post_cmd[EST_INFO_POST_CMDLEN];
} gb = {
};

void usage();
void parse_cmdline(int argc, char **argv);

void
send_post_cmd()
{
        struct est_info info;
        struct est_info_post *post = &info.post;
        char infostr[300];

        est_info_post_init(post);
        est_fill_info_srcdst(&post->hdr, gb.src, gb.dst);
        est_fill_info_timestamp_now(&post->hdr, gb.hostid);
        strncpy(post->post_cmd, gb.post_cmd, sizeof(post->post_cmd));

        est_info_oneline_str(infostr, sizeof(infostr), &info);
        LOG("%s", infostr);

        if (gb.estimator_count > 0) {
                int rc = est_push_info_entries(&gb.estimator.ch, &info, 1);
                BUG_ON(rc);
        }
}

int
main(int argc, char **argv)
{
        int rc = 0;

        est_get_host_id(gb.hostid, sizeof(gb.hostid));
	parse_cmdline(argc, argv);

        if (gb.estimator_count > 0) {
                rc = est_open_channel(&gb.estimator.ch, &gb.estimator.addr);
                if (rc) {
                        LOG("error: failed to open channel to estimator on %s (%s)",
                            gb.estimator.unparsed, gb.estimator.addrstr);
                        goto done;
                }
        }

        send_post_cmd();

        if (gb.estimator_count > 0)
                est_close_channel(&gb.estimator.ch);
done:
        return 0;
}

void
check_params(void)
{
        if (strlen(gb.post_cmd) == 0) {
                printf("error: no command specified with the -c option\n");
                printf("error: use option '-h' for help\n\n");
                exit(2);
        }
}

void
parse_cmdline(int argc, char **argv)
{
	char short_options[] = "hP:c:s:d:";
	char c = '?';

	/* parse command line options */

	while (1) {

		c = getopt(argc, argv, short_options);

		if (c == EOF)
		    break;

		switch (c) {

		case 'P':
                        gb.estimator_count++;
                        if (gb.estimator_count > 1)
                                ERR("error: too many -P options");
                        est_parse_estimator(optarg, &gb.estimator);
			break;

                case 'c':
                        snprintf(gb.post_cmd, sizeof(gb.post_cmd), "%s", optarg);
                        break;

                case 's':
                        snprintf(gb.src, sizeof(gb.src), "%s", optarg);
                        break;

                case 'd':
                        snprintf(gb.dst, sizeof(gb.dst), "%s", optarg);
                        break;

		case 'h':
			usage();
			exit(0);
			break;
		    
		case '?':
			break;
		    
		default:
			break;
		}
	}

	check_params();
}

void
usage(void)
{
	printf("\n");
	printf("    postinfo -- sends INFO/post entries to an estimator\n");
	printf("\n");

	printf("USAGE\n");
	printf("\n");
	printf("    postinfo [-P <estimator[:port]>] -c <post-command>\n");
	printf("\n");

	printf("OPTIONS\n");
	printf("        -s <src>        fill the INFO source field\n");
	printf("        -d <dst>        fill the INFO destination field\n");
	printf("\n");

	exit(1);
}

