/*
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Wed Mar 26 14:01:37 MST 2008
 *
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <pcap.h>
#include <time.h>

#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ether.h>

#include <math.h>

// libcommon
#include "alog.h"
#include "atimer.h"
#include "quicksort.h"
#include "net.h"

// mgrp
#include "mgrp.h"
#include "bhtp.h"

#define MAX_PKT_BURST   200
#define MAX_VALUES      MAX_PKT_BURST

struct maxmin {
        int finished;
        long long values[MAX_VALUES];
        long long count;
        long long max;
        long long min;
        double avg;
        long long sum;
};

struct packet_burst {
        long long count;
        long long bytes;
        struct pcap_pkthdr pcap[MAX_PKT_BURST];

        struct timeval start;
        struct timeval end;
        long long duration_usec;

        // stats
        struct maxmin pkt_bytes;
        struct maxmin pkt_usec;
};

struct global_state {
        int debug;
        char fname[300];
        char filter[300];

        int max_burst_pkts;
        int max_burst_usec;
        unsigned int max_burst_pkt_interval_usec;

        struct packet_burst burst;

        int filter_mgrp_probes;
} gb = {
        .max_burst_pkts = MAX_PKT_BURST,
        .max_burst_usec = 100000,
        .max_burst_pkt_interval_usec = 3000,
        .burst = { 0 },
};

void usage();
void parse_cmdline(int argc, char **argv);
void log_cmdline();

static inline void
maxmin_add_update(struct maxmin *mm, int val)
{
        BUG_ON(mm->count >= MAX_VALUES);
        mm->values[mm->count] = val;
        mm->count++;

        if (mm->count == 1) {
                mm->max = val;
                mm->min = val;
        } else {
                mm->max = cmp_max(mm->max, val);
                mm->min = cmp_min(mm->min, val);
        }

        mm->sum += val;
        mm->avg = (1.0 * mm->sum) / mm->count;
}

void
maxmin_finish(struct maxmin *mm)
{
        mm->finished = 1;
        // so that we can calculate percentiles
        quicksort(mm->values, 0, mm->count-1);
}

static inline long long
maxmin_percentile(struct maxmin *mm, double percentile)
{
        double n = 0;
        int k = 0;
        double d = 0;

        if (mm->count == 0)
                return 0;
        
        // values must be sorted and have at lease one
        BUG_ON(!mm->finished);
        BUG_ON(mm->count <= 0);

        // use excel's method taken from wikipedia (see Alternative methods)
        // http://en.wikipedia.org/wiki/Percentile

        n = (percentile / 100.0) * (mm->count - 1) + 1;
        // so that n = k + d (k is integer)
        k = n;
        d = n - k;

        BUG_ON(k > mm->count);

        if (k == 0)
                return mm->values[0];
        else if (k == mm->count)
                return mm->values[mm->count-1];
        else
                return mm->values[k-1] + d * (mm->values[k] - mm->values[k-1]);
}

void
log_prologue()
{
        time_t now = time(NULL);

        LOG("#");
        // ctime adds newline
        LOGNR("# %s", ctime(&now));
        LOG("#");

        LOGNR("# ");
        log_cmdline();

        LOG("#");
        LOG("# start           end        duration-usec "
            "BURST  pks bytes "
            "BYTES 90th  min  max "
            "USEC  90th   min   max");
}

void
packet_burst_log(struct packet_burst *burst)
{
        LOG("%ld.%.6ld %ld.%.6ld %6lld "
            "burst %3lld %6lld "
            "bytes %4lld %4lld %4lld "
            "usec %5lld %5lld %5lld ",

            burst->start.tv_sec,
            burst->start.tv_usec,
            burst->end.tv_sec,
            burst->end.tv_usec,
            burst->duration_usec,

            burst->count,
            burst->bytes,

            maxmin_percentile(&burst->pkt_bytes, 90),
            burst->pkt_bytes.min,
            burst->pkt_bytes.max,

            maxmin_percentile(&burst->pkt_usec, 90),
            burst->pkt_usec.min,
            burst->pkt_usec.max);
}

void
packet_burst_finish(struct packet_burst *burst)
{
        // so that we can calculate percentiles
        maxmin_finish(&burst->pkt_bytes);
        maxmin_finish(&burst->pkt_usec);
}

// 1 if added
// 0 if it starts a new burst; not added
//
// A new burst starts after the count has reached the max limit or
// or the inter-packet interval is above threshold.
int
packet_burst_add_pkt(struct packet_burst *burst,
                     const struct pcap_pkthdr *pcap)
{
        // usec from last packet captured: interval between packets
        long long diff_usec = atimer_timeval_diff_usec(&pcap->ts, &burst->end);
        long long burst_usec = atimer_timeval_diff_usec(&pcap->ts, &burst->start);

        if (burst->count >= gb.max_burst_pkts)
                return 0;

        if (burst->count > 0) {
                if (diff_usec > gb.max_burst_pkt_interval_usec)
                        return 0;
                if (burst_usec > gb.max_burst_usec)
                        return 0;
        }

        if (burst->count == 0)
                burst->start = pcap->ts;

        burst->end = pcap->ts;
        burst->duration_usec = atimer_timeval_diff_usec(
                                        &burst->end,
                                        &burst->start);

        BUG_ON(burst->count >= MAX_PKT_BURST);
        burst->pcap[burst->count] = *pcap;
        burst->bytes += pcap->len;
        maxmin_add_update(&burst->pkt_bytes, pcap->len);
        if (burst->count > 0)
                maxmin_add_update(&burst->pkt_usec, diff_usec);

        burst->count++;
        return 1;
}

// very similar to process_packet:pcap_pkt_parse()
int
process_packet(const struct pcap_pkthdr *pcap,
               const u_char *packet,
               const struct timeval *prev_ts)
{
        int sport = 0;
        int dport = 0;
        char *error = "unknown error";
        char *transport = "unknown transport";

        const struct ethhdr *eh = (struct ethhdr *)packet;
        const struct iphdr *iph = (struct iphdr *)(packet + ETH_HLEN);

        long long diff_usec = atimer_timeval_diff_usec(&pcap->ts, prev_ts);

        // transport portion is under ethernet and IP header (ignoring options..)
        int transport_len = pcap->len - (14+20);
        int transport_caplen = pcap->caplen - (14+20);
        const u_char *transport_header = packet + (14+20);

        // we need at least eth and IP headers
        BUG_ON(transport_caplen < 0);

        if (eh->h_proto != htons(ETHERTYPE_IP)) {
                error = "not IP";
                goto skip_error;
        }

        if (iph->protocol == IPPROTO_TCP) {
                const struct tcphdr *hdr = (struct tcphdr *)transport_header;
                BUG_ON(transport_caplen < sizeof(struct tcphdr));

                transport = "tcp";
                sport = ntohs(hdr->source);
                dport = ntohs(hdr->dest);
                if (gb.filter_mgrp_probes)
                        goto skip;

        } else if (iph->protocol == IPPROTO_UDP) {
                const struct udphdr *hdr = (struct udphdr *)transport_header;
                BUG_ON(transport_caplen < sizeof(struct udphdr));

                transport = "udp";
                sport = ntohs(hdr->source);
                dport = ntohs(hdr->dest);
                if (gb.filter_mgrp_probes)
                        goto skip;

        } else if (iph->protocol == IPPROTO_MGRP) {

                const struct bhtphdr *hdr = (struct bhtphdr *)transport_header;
                BUG_ON(transport_caplen < sizeof(struct bhtphdr));

                transport = "mgrp";
                sport = 0;
                dport = 0;

                if (gb.filter_mgrp_probes && !bhtp_pkt_is_probe(hdr, transport_len))
                        goto skip;

        } else if (iph->protocol == IPPROTO_ICMP) {
                transport = "imcp";
                goto skip;

        } else {
                error = "IP protocol not handled";
                goto skip_error;
        }

        while (!packet_burst_add_pkt(&gb.burst, pcap)) {
                // create a new burst
                packet_burst_finish(&gb.burst);
                packet_burst_log(&gb.burst);
                memset(&gb.burst, 0, sizeof(gb.burst));
        }

        if (gb.debug) {
                LOG("%ld.%.6ld %8lld %5d %s:%d:%d",
                    pcap->ts.tv_sec,
                    pcap->ts.tv_usec,
                    diff_usec,
                    pcap->len,
                    transport, sport, dport);
        }

        return 1;

skip_error:
        LOG_ERR("skipping: %s: %ld.%.6ld %d",
                error,
                pcap->ts.tv_sec,
                pcap->ts.tv_usec,
                pcap->len);
skip:
        return 0;
}

static inline void
detect_bursts(pcap_t *trace)
{
        struct pcap_pkthdr pcap;
        const u_char *packet = NULL;
        struct timeval prev_ts = { 0 };

        while ((packet = pcap_next(trace, &pcap)) != NULL) {
                if (process_packet(&pcap, packet, &prev_ts))
                        prev_ts = pcap.ts;
        }

        // for the last burst
        packet_burst_finish(&gb.burst);
        packet_burst_log(&gb.burst);
}

int
main(int argc, char **argv)
{
        pcap_t *trace = NULL;
        char err[NET_PCAP_ERRBUF_SIZE];

	parse_cmdline(argc, argv);

        trace = net_pcap_open_offline_filter(gb.fname, gb.filter,
                                             err, sizeof(err));
        if (trace == NULL)
                ERR("error: pcap trace: %s", err);

        log_prologue();
        detect_bursts(trace);
        pcap_close(trace);

        return 0;
}

void
check_params(void)
{
        if (strlen(gb.fname) == 0) {
                printf("error: no pcap file specified with the -r option\n");
                printf("error: use option '-h' for help\n\n");
                exit(2);
        }
}

void
log_cmdline()
{
        LOG("%s -r %s -f '%s' -U %d -P %d -I %d",
            "./spitburst",
            gb.fname,
            gb.filter,
            gb.max_burst_usec,
            gb.max_burst_pkts,
            gb.max_burst_pkt_interval_usec);
}

void
parse_cmdline(int argc, char **argv)
{
        int found = 0;
	char short_options[] = "hdr:f:pP:U:I:";
	char c = '?';

	/* parse command line options */

	while (1) {

		c = getopt(argc, argv, short_options);

		if (c == EOF)
		    break;

		switch (c) {

                case 'd':
                        gb.debug = 1;
                        break;

                case 'r':
                        snprintf(gb.fname, sizeof(gb.fname), "%s", optarg);
                        break;

                case 'f':
                        snprintf(gb.filter, sizeof(gb.filter), "%s", optarg);
                        break;

                case 'p':
                        gb.filter_mgrp_probes = 1;
                        break;

                case 'U':
                        found = sscanf(optarg, "%u", &gb.max_burst_usec);
                        ERR_ON(!found, "failed to parse max_burst_usec: %s", optarg);
                        break;

                case 'P':
                        found = sscanf(optarg, "%u", &gb.max_burst_pkts);
                        ERR_ON(!found, "failed to parse max_burst_pkts: %s", optarg);
                        break;

                case 'I':
                        found = sscanf(optarg, "%u", &gb.max_burst_pkt_interval_usec);
                        ERR_ON(!found, "failed to parse packet interval: %s", optarg);
                        break;

		case 'h':
			usage();
			exit(0);
			break;

		    
		case '?':
			break;
		    
		default:
			break;
		}
	}

	check_params();
}

void
usage(void)
{
	printf("\n");
	printf("    spitburst -- identifies packet bursts in a libpcap file\n");
	printf("\n");

	printf("USAGE\n");
	printf("\n");
	printf("    spitburst -r <pcap-file> [options]\n");
	printf("\n");

	printf("OPTIONS\n");
	printf("        -r <pcap-file>  file to open with libpcap (use - for stdin)\n");
	printf("        -f <filter-str> tcpdump-type filter string\n");
	printf("        -p              filter only mgrp probes (on top of -f option)");
	printf("\n");
	printf("        -P <pkts>       maximum number of packets in burst\n");
        printf("                        default: %d\n", gb.max_burst_pkts);
	printf("        -U <usec>       maximum burst duration\n");
        printf("                        default: %d\n", gb.max_burst_usec);
	printf("        -I <usec>       maximum interval between packets inside a burst\n");
        printf("                        default: %d\n", gb.max_burst_pkt_interval_usec);
	printf("\n");
	printf("        -d              debug; one line per packet\n");
	printf("\n");

	exit(1);
}


