/*
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Tue Jan  8 12:48:41 EST 2008
 *
 * Transforms the text log of tracereceiver into INFO entries and
 * sends them to the local estimator.
 *
 * Expects at STDIN the output of medianet/src/apps/tracereceiver.
 * Every line from STDIN is parsed and converted into an INFO entry
 * that is sent to the local estimator.
 */

#include <stdio.h>

#include "net.h"
#include "libestimator.h"

struct global_state {
        char hostid[EST_INFO_IDLEN];

        int estimator_count;
        struct est_estimator estimator;

        int debug;
} gb = {
        .estimator_count        = 0,
        .debug                  = 0,
};

int
parse_traceline_info(const char *line, struct est_info_trace *trace)
{
        int parsed = 0;
        int sec = 0;
        int usec = 0;
        float bps = 0;
        char *last_space = NULL;
        char *decode_str = NULL;
        struct timeval timestamp;

        est_info_trace_init(trace);

        // find the decode code at the end of the string
        last_space = strrchr(line, ' ');
        if (last_space == NULL)
                goto error;
        decode_str = last_space + 1;

        // newline still there, compare only the code length
        if (strncmp(decode_str, "OK", strlen("OK")) == 0)
                trace->log.decode = EST_TRACE_DECODE_OK;
        else if (strncmp(decode_str, "BAD", strlen("BAD")) == 0)
                trace->log.decode = EST_TRACE_DECODE_BAD;
        else
                trace->log.decode = EST_TRACE_DECODE_UNDEF;

        parsed = sscanf(line,
                        "port %d seq %d size %d time %d.%d b/w %g %c %d",
                        &trace->log.port,
                        &trace->log.seq,
                        &trace->log.bytes,
                        &sec, &usec,
                        &bps,
                        &trace->log.frame_type,
                        &trace->log.temporal_place);
        if (parsed != 8) {
                LOG("error: expecting %d got %d", 8, parsed);
                goto error;
        }

        trace->log.bps = bps;

        timestamp.tv_sec = sec;
        timestamp.tv_usec = usec;
        est_fill_info_timestamp(&trace->hdr, gb.hostid, &timestamp);

        return 0;
error:
        return -1;
}

void
read_and_send_trace(FILE *f)
{
        char line[200];

        while (1) {
                int rc = 0;
                struct est_info info;
                char *res = fgets(line, sizeof(line), f);
                char infostr[200];

                if (res == NULL)
                        break;
                rc = parse_traceline_info(line, &info.trace);
                if (rc) {
                        LOGNR("error: cannot parse line: %s", line);
                        continue;
                }
                
                est_info_oneline_str(infostr, sizeof(infostr), &info);
                rc = est_push_info_entries(&gb.estimator.ch, &info, 1);
                BUG_ON(rc);
                if (gb.debug) {
                        LOG("---");
                        LOGNR("trace: %s", line);
                        LOG("info : %s", infostr);
                } else {
                        LOG("%s", infostr);
                }
        }
}

void usage();
void parse_cmdline(int argc, char **argv);

int
main(int argc, char **argv)
{
        int rc = 0;

        est_get_host_id(gb.hostid, sizeof(gb.hostid));
	parse_cmdline(argc, argv);

        rc = est_open_channel(&gb.estimator.ch, &gb.estimator.addr);
        if (rc) {
                LOG("error: failed to open channel to estimator on %s (%s)",
                    gb.estimator.unparsed, gb.estimator.addrstr);
                goto done;
        }

        // main loop
        read_and_send_trace(stdin);

        est_close_channel(&gb.estimator.ch);
done:
        return 0;
}

void
check_params(void)
{
	if (gb.estimator_count != 1) {
		printf("error: no estimator specified with the -P option\n");
		printf("error: use option '-h' for help\n\n");
		exit(2);
	}
}

void
parse_cmdline(int argc, char **argv)
{
	char short_options[] = "hP:d";
	char c = '?';

	/* parse command line options */

	while (1) {

		c = getopt(argc, argv, short_options);

		if (c == EOF)
		    break;

		switch (c) {

		case 'P':
                        gb.estimator_count++;
                        if (gb.estimator_count > 1)
                                ERR("error: too many -P options");
                        est_parse_estimator(optarg, &gb.estimator);
			break;

		case 'd':
                        gb.debug = 1;
			break;
		    
		case 'h':
			usage();
			exit(0);
			break;
		    
		case '?':
			break;
		    
		default:
			break;
		}
	}

	check_params();
}

void
usage(void)
{
	printf("\n");
	printf("    traceinfo -- converts and sends INFO/trace entries to an estimator\n");
	printf("\n");

	printf("DESCRIPTION\n");
	printf("    Converts the text output of tracereceiver into INFO/trace entries\n");
	printf("    and sends them to an estimator process.\n");
	printf("\n");

	printf("USAGE\n");
	printf("\n");
	printf("    cat tracereceiver.output | traceinfo -P <estimator[:port]> [options]\n");
	printf("\n");

	printf("OPTIONS\n");
	printf("        -d      debug info\n");
	printf("\n");
	exit(1);
}

