/*
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Mon Oct  8 16:16:03 EDT 2007
 *
 * Will use the estimator library to connect to the estimator at source:port
 * and request measurements to estimator on dest:port.
 */

#include <stdio.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <unistd.h>

#include "mgrp/bhtp.h"
#include "mgrp/udpex.h"

/* libcommon */
#include "alog.h"
#include "net.h"
#include "tick.h"
#include "cmp.h"
#include "parse.h"

#include "libestimator.h"

#define MAX_SESSIONS	1

struct global_state {

	int iterate;
	struct sockaddr_in addr_src[MAX_SESSIONS];
	struct sockaddr_in addr_dst[MAX_SESSIONS];

	int sessions;
	struct est_trafgen_session *ses[MAX_SESSIONS];
	char names[MAX_SESSIONS][200];

	int fdmax;
	fd_set fdreads;
	
	int replies;

	int trafgen_action;

	int create_traffic;
	int start_traffic;
	int stop_traffic;
#ifdef TRAFGEN_REPORT
	int start_reports;
#endif
	struct est_trafgen_traffic traffic;

} gb = {
	/* UDP traffic with packet size 1000 bytes and rate 1Kbits/sec */
	.traffic	= {
		.type		= EST_TRAFGEN_UDP_NORMAL,
		.pktlen		= 1000,
		.bps		= 1000,
	},
};

void usage(void);
void parse_cmdline(int argc, char **argv);

static inline void
include_rfd(int fd)
{
	gb.fdmax = cmp_max(gb.fdmax, fd + 1);
	FD_SET(fd, &gb.fdreads);
}

static inline void
exclude_rfd(int fd)
{
	FD_CLR(fd, &gb.fdreads);
	/* find the new max */
	while (gb.fdmax > 0) {
		if (FD_ISSET(gb.fdmax - 1, &gb.fdreads))
			break;
		gb.fdmax--;
	}
}

#ifdef TRAFGEN_REPORT
void
collect_reports(void)
{
	int i = 0;
	int rc = 0;
	int timeout = 0;
	int interrupt = 0;
	int fdready = 0;
	int total_fdready = 0;
	fd_set rfds = gb.fdreads;

	rc = select(gb.fdmax, &rfds, NULL, NULL, NULL);

	timeout = (rc == 0);
	interrupt = (rc == -1 && errno == EINTR);
	fdready = (rc > 0 ? rc : 0);

	total_fdready = fdready;

	for (i = 0; i < gb.sessions && fdready > 0; i++) {
		int final = 0;
		struct est_trafgen_report report;
		int sock = est_trafgen_session_socket(gb.ses[i]);
		
		if (!FD_ISSET(sock, &rfds))
			continue;
		
		rc = est_next_trafgen_report(gb.ses[i], &report, &final);
		/* remove the channel on error */
		ERR_ON(rc, "next estimate error: %d", rc);

		LOG(">>>>>>>>>>>>> [%s] GOT PROBE REPLY %d",
			gb.names[i], ++gb.replies);
		est_trafgen_logreport(&report);

		fdready--;
	}

	if (fdready > 0) {
		ERR("failed to process all %d sockets: %d left",
			total_fdready, fdready);
	}
}
#endif

void
trafgen_send_traffic_cmd(int action)
{
	int i = 0;
	int rc = 0;

	gb.fdmax = 0;
	FD_ZERO(&gb.fdreads);

	/* create trafgen sessions on src nodes and start traffic */
	/* traffic starts immediately but no traffic reports are generated */

	for (i = 0; i < gb.sessions; i++) {

		gb.ses[i] = est_create_trafgen_session(
				&gb.addr_src[i],
				&gb.addr_dst[i],
				EST_TRAFGEN_SRC);
		ERR_ON(gb.ses[i] == NULL, "session %d failed", i);

		/* Send TRAFGEN_TRAFFIC command to @ses->src with @action
		 *
		 * The @gb.traffic is common across sessions and describes
		 * the traffic to generate (rate and packet length).
		 * Used only for the EST_TRAFGEN_START action.
		 */
		rc = est_send_trafgen_traffic_to_src(gb.ses[i], action,
						     &gb.traffic);
		if (rc) {
		    ERR("session %d failed to send traffic action %d: %s",
		    	i, action, est_strerror(rc));
		}

		/* closing the session does not stop the traffic */
		est_destroy_trafgen_session(gb.ses[i]);
	}
}

#ifdef TRAFGEN_REPORT
void
trafgen_reports(void)
{
	int i = 0;
	int rc = 0;

	gb.fdmax = 0;
	FD_ZERO(&gb.fdreads);

	/* create trafgen sessions on dst nodes to get traffic reports */
	/* traffic has already been started with sessions on src nodes */

	for (i = 0; i < gb.sessions; i++) {

		gb.ses[i] = est_create_trafgen_session(
				&gb.addr_src[i],
				&gb.addr_dst[i],
				EST_TRAFGEN_DST);
		ERR_ON(gb.ses[i] == NULL, "session %d failed", i);

		/* instructs @dst to start sending traffic reports back */
		rc = est_start_trafgen_reports(gb.ses[i]);
		if (rc) {
		    ERR("session %d failed to start reports: %s",
		    	i, est_strerror(rc));
		}

		include_rfd(est_trafgen_session_socket(gb.ses[i]));
	}

	/* collect reports from all sessions */
	while(1)
		collect_reports();

	/* destroy probe sessions */
	for (i = 0; i < gb.sessions; i++) {
		exclude_rfd(est_trafgen_session_socket(gb.ses[i]));
		/* closing the session does not stop the traffic */
		/* only the reports stop */
		est_destroy_trafgen_session(gb.ses[i]);
	}
}
#endif

int
main (int argc, char **argv)
{
	parse_cmdline(argc, argv);

	if (gb.create_traffic)
		trafgen_send_traffic_cmd(EST_TRAFGEN_PREPARE);

	if (gb.start_traffic)
		trafgen_send_traffic_cmd(EST_TRAFGEN_START);
	else if (gb.stop_traffic)
		trafgen_send_traffic_cmd(EST_TRAFGEN_STOP);

#ifdef TRAFGEN_REPORT
	if (gb.start_reports)
		trafgen_reports();
#endif

	return 0;
}

void
check_params(void)
{
	int exclusive_options = gb.start_traffic +
			        gb.stop_traffic;
	
	int coexist_options = gb.create_traffic;
#ifdef TRAFGEN_REPORT
        coexist_options += gb.start_reports;
#endif

	if (gb.sessions <= 0) {
		printf("error: no sessions specified with the -P option\n");
		printf("error: see 'trafgen -h' for more details\n\n");
		exit(2);
	}

	if (exclusive_options == 0 && coexist_options == 0) {
		printf("error: at least one of  -c,-s,-z,r options must be specified\n");
		printf("error: see 'trafgen -h' for more details\n\n");
		exit(2);
	}

	if (exclusive_options > 1) {
		printf("error: only one of -s, -z options can be specified\n");
		printf("error: see 'trafgen -h' for more details\n\n");
		exit(2);
	}
}

void
parse_host_pair(const char *optstr)
{
	char tmp[100];
	int j = gb.sessions++;

	strncpy(gb.names[j], optarg, sizeof(gb.names[j]));

	parse_two_hosts_with_ports(optarg,
			&gb.addr_src[j],
			&gb.addr_dst[j],
			EST_DEFAULT_CONTROL_TCP_PORT);

	net_srcdst_str(tmp, sizeof(tmp),
			&gb.addr_src[j],
			&gb.addr_dst[j]);

	LOG("[%d] %s: %s", j, gb.names[j], tmp);
}

void
parse_cmdline(int argc, char **argv)
{
#ifdef TRAFGEN_REPORT
	char short_options[] = "rhxP:cs:zp:";
#else
	char short_options[] = "hxP:cs:zp:";
#endif

	int found = 0;
	char modifier = 'X';
	char c = '?';

	/* parse command line options */

	while (1) {

		c = getopt(argc, argv, short_options);

		if (c == EOF)
		    break;

		switch (c) {

		case 'P':
			if (gb.sessions >= MAX_SESSIONS)
				ERR("reached max sessions %d", gb.sessions);
			parse_host_pair(optarg);
			break;

		case 'c':
			gb.create_traffic = 1;
			break;

		case 's':
			gb.start_traffic = 1;
			found = sscanf(optarg, "%u%c",
				       &gb.traffic.bps, &modifier);
			if (found == 2 && modifier == 'M')
				gb.traffic.bps *= 1000000;
			else if (found == 2 && modifier == 'K')
				gb.traffic.bps *= 1000;
			else if (found == 1)
				gb.traffic.bps *= 1;
			else
			    ERR("invalid bandwidth value '%s'", optarg);
			break;

		case 'z':
			gb.stop_traffic = 1;
			break;

#ifdef TRAFGEN_REPORT
		case 'r':
			gb.start_reports = 1;
			break;
#endif

		case 'p':
			found = sscanf(optarg, "%u", &gb.traffic.pktlen);
			if (found != 1 ||
			    gb.traffic.pktlen > EST_TRAFGEN_MAX_PKTLEN)
				ERR("invalid packet length: %d",
				    gb.traffic.pktlen);
			break;

		case 'x':
                        gb.traffic.type = EST_TRAFGEN_UDP_MGRP;
			break;

		case 'h':
			usage();
			exit(0);
			break;
		    
		case '?':
			break;
		    
		default:
			break;
		}
	}

	check_params();
}

void
usage(void)
{
	printf("USAGE\n");
	printf("\ttrafgen -P <src[:port],dst[:port]> [options]\n");
	printf("\n");

	printf("DESCRIPTION\n");
	printf("\n");

	printf("Sends traffic commands to 'estimator' processes that generate\n");
	printf("traffic between src and dst and collects traffic reports.\n");
	printf("\n");

	printf("OPTIONS\n");
	printf("\n");

	printf("Session must not be already running between src -> dst.\n");
	printf("\n");
	printf("\t-c         creates traffic session (no traffic yet)\n");
	printf("\n");

	printf("Session must be already running between src -> dst.\n");
	printf("All commands except -r return immediately.\n");
	printf("\n");
	printf("\t-s #[KM]   starts sending traffic in bits/sec\n");
	printf("\t           (if traffic is already sent it modifies the rate\n");
	printf("\t-z         stops sending traffic but keeps traffic session\n");
	printf("\t-r         start printing traffic reports from @dst\n");
	printf("\n");

	printf("Other options\n");
	printf("\n");
	printf("\t-p         packet length in bytes\n");
	printf("\t-x         use MGRP for transport\n");
	printf("\t-h         help\n");
	printf("\n");

	printf("NOTES\n");
	printf("\n");

	printf("The 'estimator' process must be running on the specified ports\n");
	printf("on every host for this to work.\n");
	printf("\n");

	printf("The optional ports specified for 'src' and 'dst' in the -P option\n");
	printf("refer to the ports that the 'estimator' is running and not the ports\n");
	printf("between which the UDP traffic will be exchanged.\n");
	printf("\n");
	printf("If no port is specified, the default port %d is used.\n",
	       EST_DEFAULT_CONTROL_TCP_PORT);
	printf("\n");

	exit(1);
}


