/* for getline */
/* must be first for some reason, at least before unistd.h */
#define _GNU_SOURCE
#include <stdio.h>

/* for sockets */
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <unistd.h>

/* for open */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "mgrp/bhtp.h"
#include "mgrp/udpex.h"

#include "err.h"
#include "net.h"
#include "tick.h"

enum {
	TSEND_STDINOUT = 1,
	TSEND_FILE,
};

struct global_state {

	int sender;
	int receiver;
	int mgrp;
	int port;

	int mode;
	char filename[200];
	int delay;
	int buflen;

	struct sockaddr_in addr_receiver;

} gb = {
	.port		= 47919,
	.mode		= TSEND_STDINOUT,
	.filename	= "-",
	.buflen		= 2000,
};

void usage(void);
void parse_cmdline(int argc, char **argv);

void sender(void);
void receiver(void);

/*
 * Receiver
 */

void
keep_receiving_stdinout(int sock)
{
	size_t buflen = gb.buflen;
	char *buf = malloc(buflen);
	int bytes = 0;

	while ((bytes = recv(sock, buf, buflen-1, 0)) > 0) {
		buf[bytes] = 0;
		printf("%s", buf);
	}

	SYSERR_ON(bytes < 0);

	free(buf);
}

void
keep_receiving_file(int sock, char *filename)
{
	size_t buflen = gb.buflen;
	char *buf = malloc(buflen);
	int bytes = 0;
	int fd = -1;

	if (strcmp(filename, "-") == 0)
		fd = STDOUT_FILENO;
	else
		fd = open(filename,
			O_RDWR|O_CREAT|O_TRUNC,
			S_IRUSR|S_IWUSR|S_IROTH| S_IRGRP);
	SYSERR_ON(fd < 0);

	while ((bytes = recv(sock, buf, buflen, 0)) > 0) {
		write(fd, buf, bytes);
	}
	SYSERR_ON(bytes < 0);

	if (fd != STDOUT_FILENO)
		close(fd);
	free(buf);
}


void
receiver(void)
{
	int proto = (gb.mgrp ? IPPROTO_MGRP : IPPROTO_TCP);

	int sock = 0;
	int rc = 0;
	int on = 0;

	struct sockaddr_in addr = {
		.sin_family		= AF_INET,
		.sin_addr.s_addr	= htonl(INADDR_ANY),
		.sin_port		= htons(gb.port),
	};

	
	sock = socket(AF_INET, SOCK_STREAM, proto);
	SYSERR_ON(sock < 0);

	rc = bind(sock, (struct sockaddr *)&addr, sizeof(addr));
	SYSERR_ON(rc);

	on = 1;
	rc = setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (void*)&on, sizeof(on));
	SYSERR_ON(rc);

	rc = listen(sock, 0);
	SYSERR_ON(rc);

	fprintf(stderr, "receiver listening on port %d\n", ntohs(addr.sin_port));

	while (1) {
		struct sockaddr_in addr_accept = {};
		unsigned int addrlen = sizeof(addr_accept);
		int new_sock = 0;
		char dotted[50] = "";
		
		new_sock = accept(sock, &addr_accept, &addrlen);
		SYSERR_ON(new_sock < 0);
		BUG_ON(addrlen != sizeof(addr_accept));
		
		inet_ntop(addr_accept.sin_family,
			&addr_accept.sin_addr,
			dotted, sizeof(dotted));
		fprintf(stderr, "\n>>> accepted from %s\n", dotted);

		if (gb.mode == TSEND_STDINOUT) 
			keep_receiving_stdinout(new_sock);
		else if (gb.mode == TSEND_FILE) 
			keep_receiving_file(new_sock, gb.filename);
		else
			BUG();

		fflush(stdout);
		fprintf(stderr, "<<< disconnected\n\n");
		close(new_sock);
	}

	close(sock);
}

/*
 * Sender
 */

void
keep_sending_stdinout(int sock)
{
	FILE *fp = stdin;
	size_t len = 200;
	char *line = malloc(len);
	int chars = 0;

	BUG_ON(fp == NULL);

	while ((chars = getline(&line, &len, fp)) != -1) {
		int sent = send(sock, line, chars, 0);
		ERR_ON(sent != chars, "failed to send all bytes across");
	}

	free(line);
}

void
keep_sending_file(int sock, char *filename)
{
	size_t buflen = gb.buflen;
	char *buf = malloc(buflen);
	int bytes = 0;
	int fd = -1;

	if (strcmp(filename, "-") == 0)
		fd = STDIN_FILENO;
	else
		fd = open(filename, O_RDONLY);
	SYSERR_ON(fd < 0);

	while ((bytes = read(fd, buf, buflen)) > 0) {
		int sent = send(sock, buf, bytes, 0);
		ERR_ON(sent != bytes, "failed to send all bytes across");
		if (gb.delay)
			sleep(gb.delay);
	}
	SYSERR_ON(bytes < 0);

	if (fd != STDIN_FILENO)
		close(fd);
	free(buf);
}

void
sender(void)
{
	int sock = 0;
	int rc = 0;
	int proto = (gb.mgrp ? IPPROTO_MGRP : IPPROTO_TCP);

	struct sockaddr_in addr = {
		.sin_family		= AF_INET,
		.sin_addr.s_addr	= gb.addr_receiver.sin_addr.s_addr,
		.sin_port		= htons(gb.port),
	};

	
	sock = socket(AF_INET, SOCK_STREAM, proto);
	SYSERR_ON(sock < 0);

	rc = connect(sock, (struct sockaddr *)&addr, sizeof(addr));
	SYSERR_ON(rc);

	if (gb.mode == TSEND_STDINOUT) 
		keep_sending_stdinout(sock);
	else if (gb.mode == TSEND_FILE) 
		keep_sending_file(sock, gb.filename);
	else
		BUG();
	
	close(sock);
}


int
main (int argc, char **argv)
{
	parse_cmdline(argc, argv);
	if (gb.sender)
		sender();
	else
		receiver();
	return 0;
}

void
usage(void)
{
	printf("USAGE\n");
	printf("\t[  sender  ] tsend -s <receiver> [options]\n");
	printf("\t[ receiver ] tsend -r [options]\n");
	printf("\n");
	printf("OPTIONS\n");
	printf("\t-S           use STDIN/STDOUT (default)\n");
	printf("\t-F <file>    use file for transfer\n");
	printf("\t-x           use MGRP for transport\n");
	printf("\t-p <port>    send/receive port\n");
	printf("\t-d <delay>   delay in seconds between send() calls\n");
	printf("\t-b <buflen>  buffer size for send/receive\n");
	printf("\n");
	exit(1);
}

void
check_params(void)
{
	if (gb.sender == 0 && gb.receiver == 0)
		usage();

	if (gb.sender == 1 && gb.receiver == 1)
		ERR("error: must specify only one of -s or -r");
}

void
parse_cmdline(int argc, char **argv)
{
	char short_options[] = "hs:rxSF:p:d:b:";

	char c = '?';
	int found = 0;
	char error[200];

	/* parse command line options */

	while (1) {

		c = getopt(argc, argv, short_options);

		if (c == EOF)
		    break;

		switch (c) {

		case 'h':
			usage();
			exit(0);
			break;
		    
		case 's':
			gb.sender = 1;
			found = net_resolve_host(optarg,
					&gb.addr_receiver,
					error, sizeof(error));
			ERR_ON(!found, "failed to resolve host %s", optarg);
			break;

		case 'r':
			gb.receiver = 1;
			break;

		case 'x':
			gb.mgrp = 1;
			break;

		case 'p':
			found = sscanf(optarg, "%u", &gb.port);
			ERR_ON(!found, "failed to parse port: %s", optarg);
			break;

		case 'S':
			gb.mode = TSEND_STDINOUT;
			break;

		case 'F':
			gb.mode = TSEND_FILE;
			strncpy(gb.filename, optarg, sizeof(gb.filename));
			break;

		case 'd':
			found = sscanf(optarg, "%u", &gb.delay);
			ERR_ON(!found, "failed to parse delay: %s", optarg);
			break;
			
		case 'b':
			found = sscanf(optarg, "%u", &gb.buflen);
			ERR_ON(!found, "failed to parse buffer length: %s", optarg);
			break;
			
		case '?':
			ERR("invalid option");
			break;
		    
		default:
			ERR("unknown option %c", c);
			break;
		}
	}

	check_params();
}

