/* Pavlos Papageorgiou <pavlos@eng.umd.edu> */
/* Fri May 25 10:48:59 EDT 2007 */

#include <stdio.h>
#include <unistd.h>
#include <pthread.h>

#include "libestimator.h"

/* libcommon */
#include "cmp.h"
#include "alog.h"
#include "asignal.h"
#include "alist.h"
#include "net.h"

#include "estimator.h"
#include "pathload.h"
#include "trafgen.h"
#include "info.h"
#include "cpuload.h"
#include "probe.h"

struct estimator_global_state gb = {
	.estimator_control_port		=
			EST_DEFAULT_CONTROL_TCP_PORT,
	.local_traffic_port		=
			EST_LOCAL_TRAFFIC_PORT(EST_DEFAULT_CONTROL_TCP_PORT),
	.start_rcv_trafgen_normal	= 1,
	.start_rcv_trafgen_mgrp		= 0,
        .trafgen_report_every_period    = 1,
        .cpuload_report_every_period    = 1,
        .info_list_max_entries          = EST_INFO_MAX_ENTRIES,
        .timeout_pathload_rcv_sec       = 60,
};

void usage(void);
void parse_cmdline(int argc, char **argv);
void estimator_signal_handler(int signo);

static inline void
include_rfd(int fd)
{
	gb.fdmax = cmp_max(gb.fdmax, fd + 1);
	FD_SET(fd, &gb.fdreads);
}

static inline void
exclude_rfd(int fd)
{
	FD_CLR(fd, &gb.fdreads);
	/* find the new max */
	while (gb.fdmax > 0) {
		if (FD_ISSET(gb.fdmax - 1, &gb.fdreads))
			break;
		gb.fdmax--;
	}
}

/* to be used with lists */
void
free_channel(void *ptr)
{
	struct est_channel *ch = ptr;

	LOGFN("done with channel sock %d", ch->sock);
	exclude_rfd(ch->sock);
	est_close_channel(ch);
	free(ptr);
}

int
process_channel(struct est_channel *ch)
{
	int rc = 0;
	est_cmd cmd;
	char err[200];
        pthread_t thread;
        struct est_process_data *pdata = NULL;

	memset(&cmd, 0, sizeof(est_cmd));

	rc = est_recv_ctrlmsg(ch, &cmd, sizeof(cmd), err, sizeof(err));
	if (rc)
		goto error;
	
        // allocate the data for any thread we may start
        pdata = est_process_data_alloc_init(ch, &cmd);
        BUG_ON(pdata == NULL);
        // bump refcount for us (first)
        aptr_reserve(&pdata->ptr);

	switch (cmd.hdr.type) {
		case (EST_TYPE_PEERSTART | EST_TYPE_COMMAND):
			DEB("+++ PEERSTART type 0x%x", cmd.peerstart.type);
                        // bump refcount for the thread (second)
                        // the thread must release the memory when done
                        aptr_reserve(&pdata->ptr);
                        pthread_create(&thread, NULL,
                                       est_process_peerstart, pdata);
                        pthread_detach(thread);
			break;

		case (EST_TYPE_PROBE | EST_TYPE_COMMAND):
			DEB("+++ PROBE");
			pathload_process_probe(ch, &cmd.probe,
						gb.quiet_pathload);
			break;

		case (EST_TYPE_PROBESTART | EST_TYPE_COMMAND):
			DEB("+++ PROBESTART");
			pathload_process_probestart(ch, &cmd.probestart,
						gb.quiet_pathload);
			break;

		case (EST_TYPE_TRAFGEN_TRAFFIC | EST_TYPE_COMMAND):
			DEB("+++ TRAFGEN_TRAFFIC");
			est_trafgen_process_traffic(ch, &cmd.trafgen_traffic);
			break;

#ifdef TRAFGEN_REPORT
		case (EST_TYPE_TRAFGEN_REPORT | EST_TYPE_COMMAND):
			DEB("+++ TRAFGEN_REPORT");
			est_trafgen_process_report(ch, &cmd.trafgen_report);
			break;
#endif
                
                case (EST_TYPE_INFO | EST_TYPE_COMMAND):
                        DEB("+++ INFO");
                        est_process_info(ch, &cmd.info);
                        break;

                case (EST_TYPE_PROBESTATUS | EST_TYPE_COMMAND):
                        DEB("+++ PROBESTATUS");
                        est_process_probestatus(ch, &cmd.probestatus);
                        break;

		default:
			BUG();
			break;
	}
	
	return 0;

        // the memory will disappear when the thread releases it
        aptr_release(&pdata->ptr);
error:
	LOGTR("%s: %s", est_strerror(rc), (err));
	return rc;
}

int
process_listen_socket(fd_set *fds, int fdready)
{
	int rc = 0;
	int ok = 0;
	struct est_channel *ch = NULL;

	if (!FD_ISSET(gb.listen_sock, fds))
		return 0;
	
	ch = malloc(sizeof(struct est_channel));

	rc = est_accept_channel(gb.listen_sock, ch);
	if (rc) {
		LOGTR("%s: %s", "failed to accept channel", est_strerror(rc));
		goto error_free;
	}

	LOGFN("accepted new estimator channel on socket %d", ch->sock);

	/* channel allocated here and inserted into a list */
	/* channel will be freed when it will be removed from the list */
	/* either as part of alist_done() or an explicit removal */
	ok = alist_insert_last_r(&gb.channels, ch, ALIST_OWNER);
	if (ok != ALIST_OK) {
		ERR_ON(rc, "failed to insert channel into list");
		goto error_close;
	}

	include_rfd(ch->sock);
	return 1;

error_close:
	est_close_channel(ch);
error_free:
	free(ch);
	return 1;
}

int
process_channel_socket(fd_set *fds, int fdready)
{
	int rc = 0;
	int ok = 0;
	int fdprocessed = 0;
	struct anode *node = NULL;

	ok = alist_iterator_set(&gb.channels,
				ALIST_FORWARD, ALIST_BEFORE_FIRST, NULL);
	if (ok != ALIST_OK) {
		ERR("failed to set iterator");
		goto done;
	}

	while ((node = alist_iterator_next(&gb.channels)) != NULL) {
		struct est_channel *ch = node->data;

		if (fdprocessed >= fdready)
			break;
		
		if (!FD_ISSET(ch->sock, fds))
			continue;

		rc = process_channel(ch);
		if (rc) {
			/*
			 * Remove a channel because remote end closed.
			 *
			 * This function has multiple side effects:
			 *
			 *	- removes node from channel list
			 *	- deallocates node memory
			 *	- updates list iterator: next() still works
			 *	- free_channel() is called
			 *		- removes socket from gb.fdreads
			 *		- closes the channel
			 */
			ok = alist_remove_destroy_node(&gb.channels, node);
			ERR_ON(ok != ALIST_OK, "failed to remove node");
		}

		fdprocessed++;
	}
done:
	return fdprocessed;
}

void
open_listen_socket(void)
{
	int rc = 0;
	int on = 0;

	struct sockaddr_in addr = {
		.sin_family		= AF_INET,
		.sin_addr.s_addr	= htonl(INADDR_ANY),
		.sin_port		= htons(gb.estimator_control_port),
	};

	gb.listen_sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	SYSERR_ON(gb.listen_sock < 0);

	on = 1;
	rc = setsockopt(gb.listen_sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
	SYSERR_ON(rc);

	rc = bind(gb.listen_sock, (struct sockaddr *)&addr, sizeof(addr));
	SYSERR_ON(rc);

	rc = listen(gb.listen_sock, 0);
	SYSERR_ON(rc);

	include_rfd(gb.listen_sock);

	LOGFN("estimator listening on port %d\n", ntohs(addr.sin_port));
}

void
estimator_init(void)
{
	int rc = 0;
	pthread_t thread;
        int trafgen_receiver_count = 0;
        struct trafgen_receiver *trafgen_receiver_ptr[EST_TRAFGEN_INFO_MAX_RECV];

        // Used for INFO messages
        est_get_host_id(gb.estimator_id, sizeof(gb.estimator_id));

        LOG("starting estimator id: %s", gb.estimator_id);
	LOG("est_header         size %d bytes", sizeof(struct est_header));
        est_info_sizes_log();
        LOG("");

	gb.fdmax = 0;
	FD_ZERO(&gb.fdreads);

	if (gb.quiet) {
		alog_redirect(STDOUT_FILENO, "estimator.stdout");
		alog_redirect(STDERR_FILENO, "estimator.stderr");
	}

	pthread_mutex_init(&gb.receiver_mutex_door1, NULL);
	pthread_mutex_init(&gb.receiver_mutex_door2, NULL);

        /* info list */
        rc = alist_init(&gb.info_list, ALIST_NO_MUTEX, &free);
        if (rc) goto error;
        pthread_mutex_init(&gb.info_list_mutex, NULL);

	/* trafgen senders */
	pthread_mutex_init(&gb.trafgen_sender_mutex, NULL);

	/* trafgen receivers */
	if (gb.start_rcv_trafgen_normal) {
                struct trafgen_receiver *recv = &gb.rcv_trafgen_normal;

		LOGFN("starting normal UDP trafgen receiver");

                est_trafgen_receiver_init(recv);
		recv->type = EST_TRAFGEN_UDP_NORMAL;
		recv->local_traffic_port = gb.local_traffic_port;
		rc = pthread_create(&thread, NULL, est_trafgen_receiver_thread,
			            recv);
		if (rc) goto error;
                
                trafgen_receiver_count++;
                BUG_ON(trafgen_receiver_count > EST_TRAFGEN_INFO_MAX_RECV);
                trafgen_receiver_ptr[trafgen_receiver_count-1] = recv;
	}
	if (gb.start_rcv_trafgen_mgrp) {
                struct trafgen_receiver *recv = &gb.rcv_trafgen_mgrp;

		LOGFN("starting UDP over MGRP trafgen receiver");

                est_trafgen_receiver_init(recv);
		recv->type = EST_TRAFGEN_UDP_MGRP;
		/* UDP and UDP/MGRP ports can be the same; different proto */
		recv->local_traffic_port = gb.local_traffic_port;
		rc = pthread_create(&thread, NULL, est_trafgen_receiver_thread,
			            recv);
		if (rc) goto error;

                trafgen_receiver_count++;
                BUG_ON(trafgen_receiver_count > EST_TRAFGEN_INFO_MAX_RECV);
                trafgen_receiver_ptr[trafgen_receiver_count-1] = recv;
	}

        /* trafgen info: periodically collects info from all receiver threads */
        if (gb.trafgen_report_every_period > 0) {
                int i = 0;
                struct trafgen_info *info = &gb.rcv_trafgen_info;

                est_trafgen_info_init(info);
                for (i = 0; i < trafgen_receiver_count; i++) {
                        int index = info->recv_count;
                        info->recv_count++;
                        BUG_ON(info->recv_count > EST_TRAFGEN_INFO_MAX_RECV);
                        info->recv[index] = trafgen_receiver_ptr[i];
                }
                info->report_every_period = gb.trafgen_report_every_period;
                rc = pthread_create(&thread, NULL, est_trafgen_info_thread,
                                    info);
        }

        /* thread to collect periodically cpu statistics */
        if (gb.cpuload_report_every_period > 0) {
                struct cpuload *data = &gb.cpuload_data;

                est_cpuload_init(data);
                data->report_every_period = gb.cpuload_report_every_period;

                pthread_create(&thread, NULL, est_cpuload_thread, data);
                pthread_detach(thread);
        }

	/* install signal handlers */
	rc = asignal_set_handler(SIGTERM, 0, &estimator_signal_handler, NULL);
	if (rc) goto error;

	rc = asignal_set_handler(SIGINT, 0, &estimator_signal_handler, NULL);
	if (rc) goto error;

	/* estimator control channel list */
	rc = alist_init(&gb.channels, ALIST_MUTEX, &free_channel);
	if (rc) goto error;

        /* hash storage for pointer lookup */
        rc = aptr_storage_init(&gb.pointers);
        if (rc) goto error;

	/* ok, we are open for business */
	open_listen_socket();

	gb.keep_serving = 1;
	return;

error:
	ERR("error: failed to initialize the estimator");
	return;
}

void
estimator_done(void)
{
	int rc = 0;

	rc = asignal_set_handler(SIGTERM, 0, NULL, NULL);
	ERR_ON(rc, "failed to set SIGTERM handler");

	rc = asignal_set_handler(SIGINT, 0, NULL, NULL);
	ERR_ON(rc, "failed to set SIGINT handler");

	exclude_rfd(gb.listen_sock);
	rc = close(gb.listen_sock);
	SYSERR_ON(rc);

	rc = alist_done(&gb.channels);
	ERR_ON(rc, "failed to destroy channel list");

	rc = alist_done(&gb.info_list);
	ERR_ON(rc, "failed to destroy info list");

        aptr_storage_done(&gb.pointers);
}

void
estimator_serve(void)
{
	int rc = 0;
	int timeout = 0;
	int interrupt = 0;
	int fdready = 0;
	int total_fdready = 0;
	fd_set rfds = gb.fdreads;

	rc = select(gb.fdmax, &rfds, NULL, NULL, NULL);

	timeout = (rc == 0);
	interrupt = (rc == -1 && errno == EINTR);
	fdready = (rc > 0 ? rc : 0);

	total_fdready = fdready;

	if (fdready > 0)
		fdready -= process_listen_socket(&rfds, fdready);

	if (fdready > 0)
		fdready -= process_channel_socket(&rfds, fdready);

	if (fdready > 0) {
		ERR("failed to process all %d sockets: %d left",
			total_fdready, fdready);
	}
}

void
estimator_signal_handler(int signo)
{
    switch (signo) {

        case SIGINT:
        case SIGTERM:
	    LOG("cleaning up..");
            estimator_done();
            exit(0);
            break;

        default:
            LOGFN("do not know how to handle signal %d", signo);
            break;
    }
}


int
main (int argc, char **argv)
{
        strncpy(gb.estimator_id_iface,
                EST_DEFAULT_ID_IFACE,
                sizeof(gb.estimator_id_iface));

	parse_cmdline(argc, argv);

	estimator_init();
	while (gb.keep_serving) {
		estimator_serve();
	}
	estimator_done();

	return 0;
}

void
usage(void)
{
	printf("USAGE: TBD\n");
	printf("\n");
	exit(1);
}

void
check_params(void)
{
}

void
parse_cmdline(int argc, char **argv)
{
	char short_options[] = "hqQ";
	char c = '?';

	/* parse command line options */

	while (1) {

		c = getopt(argc, argv, short_options);

		if (c == EOF)
		    break;

		switch (c) {

		case 'h':
			usage();
			exit(0);
			break;
		    
		/* misnomer: redirects all output */
		case 'q':
			gb.quiet = 1;
			break;

		/* this makes only pathload quiet */
		case 'Q':
			gb.quiet_pathload = 1;
			break;

		case '?':
			break;
		    
		default:
			break;
		}
	}

	check_params();
}
