#include <unistd.h>
#include <string.h>
#include <pthread.h>

/* for inet_ntop */
#include <arpa/inet.h>

#include "alog.h"
#include "atimer.h"

#include "estimator.h"
#include "pathload.h"
#include "info.h"
#include "probe.h"

#include "libpathload.h"
#include "libestimator.h"
 
/* pathload_snd_options: helper struct for pathload_snd() thread */

static inline struct pathload_snd_options *
pathload_snd_options_alloc_init_register()
{
        struct pathload_snd_options *op = NULL;
        
        op = malloc(sizeof(struct pathload_snd_options));
        BUG_ON(op == NULL);

        pathload_snd_options_init(op);

        // store the pointer to @op->status in the pointer hash
        est_probe_status_reserve(&op->status);
        op->status.id = aptr_storage_store(&gb.pointers, &op->status);
        BUG_ON(op->status.id <= 0);

        return op;
}

static inline void
pathload_snd_options_unregister(struct pathload_snd_options *op)
{
        int rc = aptr_storage_remove(&gb.pointers, op->status.id);
        BUG_ON(rc);
        // @op is freed if refcount reaches zero
        est_probe_status_release(&op->status);
}

/* pathload_rcv_options: helper struct for pathload_rcv() thread */

static inline struct pathload_rcv_options *
pathload_rcv_options_alloc_init_register()
{
        struct pathload_rcv_options *op = NULL;
        
        op = malloc(sizeof(struct pathload_rcv_options));
        BUG_ON(op == NULL);

        pathload_rcv_options_init(op);

        // store the pointer to @op->status in the pointer hash
        est_probe_status_reserve(&op->status);
        op->status.id = aptr_storage_store(&gb.pointers, &op->status);
        BUG_ON(op->status.id <= 0);

        return op;
}

static inline void
pathload_rcv_options_unregister(struct pathload_rcv_options *op)
{
        int rc = aptr_storage_remove(&gb.pointers, op->status.id);
        BUG_ON(rc);
        // @op is freed if refcount reaches zero
        est_probe_status_release(&op->status);
}

/* sends PEERSTART to peer to start remote sender thread */
int
pathload_start_remote_sender(struct est_channel *ch,
			int tcp_port, int udp_port,
			int iterate, int mgrp, int piggyback, int fast,
			int *noports,
                        char *src_id, int src_id_len)
{
	int rc = 0;
	int ok = 0;
	struct est_peerstart_cmd cmd;
	struct est_peerstart_reply reply;
	char error[200];
	char *err = "";
	int sender_started = 0;

	*noports = 0;

	/* send PEERSTART */

	memset(&cmd, 0, sizeof(cmd));
	ok = est_fill_cmd_header(&cmd.hdr, EST_TYPE_PEERSTART);
	BUG_ON(!ok);

	cmd.type = EST_PEERTYPE_PATHLOAD;
        strncpy(cmd.dst_id, gb.estimator_id, sizeof(cmd.dst_id));
	cmd.pathload.peer_tcp_port = tcp_port;
	cmd.pathload.peer_udp_port = udp_port;
	cmd.pathload.peer_iterate = iterate;
	cmd.pathload.peer_mgrp = mgrp;
	cmd.pathload.peer_piggyback = piggyback;
	cmd.pathload.peer_fast = fast;

	rc = est_send_ctrlmsg(ch, &cmd, error, sizeof(error));
	if (rc) {
		err = "send PEERSTART";
		goto est_error;
	}

	/* PEERSTART reply */

	rc = est_recv_ctrlmsg(ch, &reply, sizeof(reply), error, sizeof(error));
	if (rc) {
		err = "recv PEERSTART";
		goto est_error;
	}

	ok = est_is_reply_final(&reply.hdr) &&
		(reply.hdr.type & EST_TYPE_PEERSTART);
	if (!ok) {
		err = "reply is not a PEERSTART";
		goto error;
	}

	sender_started = (reply.status == EST_PROBE_STATUS_SUCCESS);
        strncpy(src_id, reply.src_id, src_id_len);
        /* do not trust the string in the packet to be null terminated */
        src_id[src_id_len-1] = 0;

	LOGFN(">>> peer %s on ports %d, %d (status %d: %s)",
		(sender_started ? "STARTED" : "FAILED to start"),
		cmd.pathload.peer_tcp_port,
		cmd.pathload.peer_udp_port,
		reply.status,
		est_probe_status_str(reply.status));

	if (!sender_started) {
		*noports = (reply.status == EST_PROBE_STATUS_NOPORTS);
		err = "pathload did not start";
		goto error;
	}

	return 0;

est_error:
	LOGTR("%s failed: [%s] %s", err, est_strerror(rc), error);
	return -1;

error:
	LOGTR("failed: %s", err);
	return -1;
}

/* spawns one receiver thread and waits for it to complete */
void
pathload_one_receiver_run(char *remote_hostname,
		int tcp_port, int udp_port,
		int quiet, int mgrp,
		struct est_results *results)
{
        int rc = 0;
        int process_terminated = 0;
        int timeout_sec = gb.timeout_pathload_rcv_sec;

	pthread_t thread;
	struct pathload_rcv_options *op = NULL;

	LOG(">>>");
	LOG(">>> starting local rcv thread");

        /* allocate, assign ID and register with global pointer hash */
        /* this also bumps the refcount for @op (first time) */
        op = pathload_rcv_options_alloc_init_register();

	op->verbose = !quiet;
	op->rcv_tcp_port = tcp_port;
	op->rcv_udp_port = udp_port;
	strncpy(op->sender_hostname, remote_hostname, sizeof(op->sender_hostname));
	op->rcv_use_kernel_probes = mgrp;

        // The thread sends its status through the estimator control channel.
        // Then the estimator uses the cond variable to trigger the change.
        op->use_estimator_rpc = 1;
        estimator_get_loopback_addr(&op->estimator_addr);

        /* bump refcount for the thread (second time) */
        est_probe_status_reserve(&op->status);

	pthread_create(&thread, NULL, pathload_rcv, op);
	pthread_detach(thread);

	/* wait for thread to converge to an estimate */
        /* we are interested in @status->value and @status->results */
        /* timeout in msec */
        est_wait_for_probestatus(&op->status, timeout_sec * 1000);

        if (op->status.value == EST_PROBE_STATUS_TIMEOUT) {
                LOGTS("+++ completion status timed out!");
                // force the subprocess to finish
                pathload_rcv_terminate(op);
        }

	if (results != NULL)
		*results = op->status.results;

	printf("result: %s: %ld.%.6ld sec %.2f max %.2f min %.2f valid %d\n",
		(op->status.results.mgrp ? "mgrp" : "orig"),
		op->status.results.ended.tv_sec,
		op->status.results.ended.tv_usec,
		0.000001 * op->status.results.usec,
		op->status.results.max,
		op->status.results.min,
		op->status.results.valid);

        /* wait for the thread/execv to finish */
        while (process_terminated) {
                rc = asys_status_wait_for_change(
                                &op->completion_status,
                                timeout_sec * 1000,
                                PATHLOAD_COMPLETION_UNDEF);
                
                if (rc == -ETIMEDOUT) {
                        LOGTS("BUG: +++ timed out while waiting for pathload_rcv to finish");
                        LOGTS("+++ trying again to terminate pathload_rcv");
                        pathload_rcv_terminate(op);
                } else {
                        process_terminated = 1;
                }
        }

        /* we do not need @op anymore and thread is finished */

        // remove it from the global hash of pointers
        // the refcount for @op is decremented
        // @op is freed only when refcount reaches zero
        // the thread pathload_rcv() decrements the refcount when done
        pathload_rcv_options_unregister(op);
        return;

}

/* returns numbers 59010..59990 in steps of ten (last digit always zero) */
int
pathload_random_port()
{
	int base = 59000;

	/* get a random number 1..99: see man 3 rand */
	int j = 1 + (int) (99.0 * rand() / (RAND_MAX + 1.0));

	return (base + j*10);
}

struct pathload_session {
	struct est_channel *ch;
	struct est_probe_cmd cmd;

        char src_id[EST_INFO_IDLEN];
        char dst_id[EST_INFO_IDLEN];

	int tcp_port;
	int udp_port;
	int mgrp;
	int quiet;

	int sender_started;
	int receiver_running;
	int final;

        int between_runs_sec;
	
	pthread_mutex_t mutex;
};

/* must free the data argument */
void *
pathload_generate_probe_replies(void *arg)
{
	struct pathload_session *pathload = arg;
	char hostname[256];
	struct est_results results;

	int rc = 0;
	int ok = 0;
	struct est_probe_reply reply;
	char error[200];

	LOGFN("generator thread started %p", pathload);

	inet_ntop(pathload->cmd.e2addr.sin_family,
			&pathload->cmd.e2addr.sin_addr,
			hostname, sizeof(hostname));
	LOGFN("hostname for -s option: %s", hostname);

	do {
		/* ensure that only one receiver runs at a time */
		/* pathload gets confused.. */
		/* .. or turning them into a thread-aware does not work.. */

		LOGTR("waiting for receiver mutex (%d,%d)",
			pathload->tcp_port,
			pathload->udp_port);

		/* allow threads to alternate using two mutexes */

		pthread_mutex_lock(&gb.receiver_mutex_door1);
		pthread_mutex_lock(&gb.receiver_mutex_door2);
		pthread_mutex_unlock(&gb.receiver_mutex_door1);

		pathload_one_receiver_run(hostname,
			pathload->tcp_port,
			pathload->udp_port,
			pathload->quiet,
			pathload->mgrp,
			&results);

		pthread_mutex_unlock(&gb.receiver_mutex_door2);

                info_push_availbw_results(&results,
                                          pathload->src_id,
                                          pathload->dst_id,
                                          EST_INFO_AVAILBW_PATHLOAD);

		/* send reply with results to the requesting channel */

		memset(&reply, 0, sizeof(reply));
		ok = est_fill_reply_header(&reply.hdr, EST_TYPE_PROBE,
						pathload->final);
		BUG_ON(!ok);

		reply.valid_results = 1;
                reply.results = results;

		/* the channel pointer will remain valid until channel closes */
		/* stored in estimator.c:gb->channels list */

		/* XXX however, we do not check here if the channel exists */
		/* XXX it can be removed by estimator.c:process_channel_socket() */

		rc = est_send_ctrlmsg(pathload->ch, &reply,error, sizeof(error));
		if (rc) {
			LOGTR("failed to send PROBE reply: %s", error);
			goto done;
		}
		LOGFN("sent PROBE reply, final %d", pathload->final);

                if (!pathload->final)
                        sleep(pathload->between_runs_sec);

	} while (!pathload->final);
done:
	LOGTR("generator thread ended %p", pathload);
	free(pathload);
	return NULL;
}

void
pathload_start_reply_session(struct pathload_session *pathload)
{
	pthread_t thread;

	/* thread must free 'pathload' when done */

	pthread_create(&thread, NULL, pathload_generate_probe_replies, pathload);
	pthread_detach(thread);
}

void
pathload_process_probe(struct est_channel *ch,
			struct est_probe_cmd *cmd,
			int quiet)
{
	int rc = 0;
	struct est_channel peer;

	int tcp_port = 0;
	int udp_port = 0;
	int count = 0;
	int noports = 0;
	int sender_started = 0;
        char src_id[EST_INFO_IDLEN];

	struct pathload_session *pathload = NULL;

	int ok = 0;
	struct est_probe_reply reply;
	char error[200];

	BUG_ON(ch == NULL);

	/* open channel to remote estimator and start the sender */

	rc = est_open_channel(&peer, &cmd->e2addr);
	if (rc) {
		LOGTR("failed to open channel: %s", est_strerror(rc));
		goto error_send_final_reply;
	}
	LOGFN("opened channel");
	
	/* start remote thread */

	LOGFN("starting remote snd thread");
	do {
		LOGFN("attempting PEERSTART: attempt %d", ++count);

		tcp_port = pathload_random_port();
		udp_port = tcp_port + 1;
		rc = pathload_start_remote_sender(&peer,
				tcp_port, udp_port,
				cmd->iterate, cmd->mgrp, cmd->piggyback, cmd->fast,
				&noports,
                                src_id, sizeof(src_id));

		sender_started = (rc == 0);

		if (!sender_started) {
			LOGTR("PEERSTART failed; can retry [%s]",
				(noports ? "yes" : "no"));
		}

	/* keep trying with different ports, if that is the problem */
	} while (!sender_started && noports);

	est_close_channel(&peer);

	if (!sender_started) {
		LOGTR("failed to start pathload after %d attempts", count);
		goto error_send_final_reply;
	}

	/* prepare session thread data */
	/* thread will free the data */

	pathload = malloc(sizeof(*pathload));
	if (pathload == NULL)
		goto error_send_final_reply;

	/* the channel pointer will remain valid until channel closes */
	/* stored in estimator.c:gb->channels list */
	pathload->ch = ch;

	/* get a copy; you need e2addr to start pathload_rcv */
	pathload->cmd = *cmd;

	pathload->tcp_port = tcp_port;
	pathload->udp_port = udp_port;
	pathload->quiet = quiet;
	pathload->mgrp = cmd->mgrp;

	pathload->sender_started = sender_started;
	pathload->receiver_running = 0;
	pathload->final = !cmd->iterate;
        pathload->between_runs_sec = cmd->iterate_interval_sec;

	pthread_mutex_init(&pathload->mutex, NULL);

        // XXXX
        // We need to read the @src_id from sender; tbd
        strncpy(pathload->src_id, src_id, sizeof(pathload->src_id));
        strncpy(pathload->dst_id, gb.estimator_id, sizeof(pathload->dst_id));

	/* send first PROBE reply to indicate that pathload started */

	memset(&reply, 0, sizeof(reply));
	ok = est_fill_reply_header(&reply.hdr, EST_TYPE_PROBE, 0);
	BUG_ON(!ok);

	/* yeah.. we are sending them the pointer.. great coding style */
	reply.session = pathload;
	reply.session_created = sender_started;
	/* lying a tad but will start right after first reply is sent */
	reply.session_started = cmd->start_session;

	rc = est_send_ctrlmsg(ch, &reply, error, sizeof(error));
	if (rc) {
		LOGTR("failed to send first PROBE reply: %s", error);
		goto error_noreply;
	}
	LOGFN("sent first PROBE reply: created %d started %d",
		reply.session_created, reply.session_started);

	if (cmd->start_session) {
		/* starts a thread */
		pathload_start_reply_session(pathload);
		LOGFN("started pathload_reply session");
	}
	return;

error_send_final_reply:
	/* send final PROBE reply to indicate that pathload was not started */

	memset(&reply, 0, sizeof(reply));
	ok = est_fill_reply_header(&reply.hdr, EST_TYPE_PROBE, 1);
	BUG_ON(!ok);

	reply.session_created = 0;
	reply.session_started = 0;

	rc = est_send_ctrlmsg(ch, &reply, error, sizeof(error));
	if (rc) {
		LOGTR("failed to send first PROBE reply: %s", error);
		goto error_noreply;
	}

error_noreply:
	LOGFN("failed to process PROBE cmd");
	return;
}


void
pathload_process_probestart(struct est_channel *ch,
			struct est_probestart_cmd *cmd,
			int quiet)
{
	struct pathload_session *pathload = cmd->session;

	if (pathload == NULL) {
		LOGFN("failed to process PROBESTART, NULL session");
		return;
	}

	pthread_mutex_lock(&pathload->mutex);

	if (cmd->start && pathload->receiver_running) {
		LOGFN("failed to process PROBESTART, already started");
		goto done;
	}

	if (!cmd->start && !pathload->receiver_running) {
		LOGFN("failed to process PROBESTART, already stopped");
		goto done;
	}

	/* has no reply */

	if (cmd->start) {
		pathload->receiver_running = 1;
		pathload_start_reply_session(pathload);

	} else {
		/* thread pathload_generate_probe_replies() will stop on this */
		/* XXX sync issues? */
		pathload->receiver_running = 0;
		pathload->final = 1;
	}
done:
	pthread_mutex_unlock(&pathload->mutex);
	return;
}


void
pathload_process_peerstart(struct est_channel *ch,
			struct est_peerstart_cmd *cmd,
			int quiet)
{
	int ok = 0;
	int rc = 0;
	char error[200];

	pthread_t thread;
	struct pathload_snd_options *op = NULL;
	struct est_peerstart_reply reply;

        /* allocate, assign ID and register with global pointer hash */
        /* this also bumps the refcount for @op (first time) */
        op = pathload_snd_options_alloc_init_register();

	op->quiet = quiet;
	op->snd_tcp_port = cmd->pathload.peer_tcp_port;
	op->snd_udp_port = cmd->pathload.peer_udp_port;
	op->iterate = cmd->pathload.peer_iterate;
	op->snd_use_kernel_probes = cmd->pathload.peer_mgrp;
	op->snd_piggyback = cmd->pathload.peer_piggyback;
	op->snd_fast = cmd->pathload.peer_fast;

        // The thread sends its status through the estimator control channel.
        // Then the estimator uses the cond variable to trigger the change.
        op->use_estimator_rpc = 1;
        estimator_get_loopback_addr(&op->estimator_addr);

	LOGFN("sender thread starting: tcp %d udp %d iterate %d",
		op->snd_tcp_port,
		op->snd_udp_port,
		op->iterate);

        /* bump refcount for @op for the thread (second time) */
        est_probe_status_reserve(&op->status);

	pthread_create(&thread, NULL, pathload_snd, op);
	pthread_detach(thread);

	/* wait for thread to start and pass the port binding stage */
        /* we are interested in @status->value */
        est_wait_for_probestatus(&op->status, cmd->pathload.peer_timeout_msec);

        /* prepare the reply and send the status value */

	memset(&reply, 0, sizeof(reply));
	ok = est_fill_reply_header(&reply.hdr, EST_TYPE_PEERSTART, 1);
	BUG_ON(!ok);

	reply.status = op->status.value;
        strncpy(reply.src_id, gb.estimator_id, sizeof(reply.src_id));

	rc = est_send_ctrlmsg(ch, &reply, error, sizeof(error));
	ERR_ON(rc, "failed to send peerstart reply: %s", error);

        /* we do not need @op anymore; thread may still be using it though */

        // remove it from the global hash of pointers
        // the refcount for @op is decremented
        // @op is freed only when refcount reaches zero
        // the thread pathload_snd() decrements the refcount when done
        pathload_snd_options_unregister(op);

        /* pathload_snd() thread keeps running */
}

