#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>

/* for getnameinfo() */
#include <sys/socket.h>
#include <netdb.h>

#include "alog.h"
#include "net.h"
#include "cmp.h"

#include "estimator.h"
#include "trafgen.h"
#include "info.h"

#include "libestimator.h"
#include "libmgrp.h"

/* Manage the trafgen_sender array: taken from the gb.trafgen */

/* returns -1 if no more indexes are available in in gb.trafgen */
int
est_trafgen_find_empty_index(struct trafgen_sender *trafgen, int count)
{
	int i = 0;
	for (i = 0; i < count; i++) {
		if (!trafgen[i].is_valid)
			return i;
	}
	return -1;
}

/* returns -1 if index cannot be found in gb.trafgen */
int
est_trafgen_find_index(struct trafgen_sender *trafgen, int count,
		       struct sockaddr_in *src,
		       struct sockaddr_in *dst)
{
	int i = 0;
	for (i = 0; i < count; i++) {
		if (!trafgen[i].is_valid)
			continue;

		if (net_sockaddr_match_sinaddr(&trafgen[i].src, src) &&
		    net_sockaddr_match_sinaddr(&trafgen[i].dst, dst))
			return i;
	}
	return -1;
}

/* returns pointer in global array gb.trafgen */
struct trafgen_sender *
est_pick_and_init_trafgen()
{
	int index = 0;
	struct trafgen_sender *trafgen = NULL;
	
	pthread_mutex_lock(&gb.trafgen_sender_mutex);

	index = est_trafgen_find_empty_index(gb.snd_trafgen,
					     EST_MAX_TRAFGEN_SENDERS);
	if (index == -1) {
		goto done_release;
	}

	trafgen = &gb.snd_trafgen[index];
	est_trafgen_sender_init(trafgen);

done_release:
	pthread_mutex_unlock(&gb.trafgen_sender_mutex);
	return trafgen;
}

/* trafgen SENDER thread */

int
est_trafgen_sender_create_socket(struct trafgen_sender *trafgen)
{
	int rc = 0;

	struct sockaddr_in src_addr = {
		.sin_family		= AF_INET,
		.sin_addr.s_addr	= trafgen->src.sin_addr.s_addr,
		.sin_port		= htons(0),
	};

	struct sockaddr_in dst_addr = {
		.sin_family		= AF_INET,
		.sin_addr.s_addr	= trafgen->dst.sin_addr.s_addr,
		.sin_port		= htons(trafgen->remote_traffic_port),
	};

	if (trafgen->use_mgrp)
		trafgen->sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_MGRP);
	else
		trafgen->sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);

	if (trafgen->sock < 0) {
		rc = trafgen->sock;
		goto syserror;
	}

	/* We bind so that destination will receive the packets from
	 * the source address that it expects them (so that it can lookup
	 * src-dst sessions). We need the bind in case we have more than
	 * one interfaces and both can reach the destination.
	 */
	rc = bind(trafgen->sock,
		  (struct sockaddr *)&src_addr, sizeof(struct sockaddr_in));
	if (rc) goto close_syserror;

	/* We connect so that we do not have to specify destination every time */
	rc = connect(trafgen->sock,
		     (struct sockaddr *)&dst_addr, sizeof(struct sockaddr_in));
	if (rc) goto close_syserror;

	return 0;

close_syserror:
	close(trafgen->sock);
syserror:
	LOGTR("error: failed to create sender socket:\n%s", strerror(errno));
	return rc;
}

void
est_trafgen_start_sending_once_active(struct trafgen_sender *trafgen)
{
	char buf[2000];
	struct trafgen_payload *payload = (struct trafgen_payload *)buf;

	/* wait to become active by est_trafgen_traffic_startstop() */
	pthread_mutex_lock(&trafgen->traffic_mutex);
	while (!trafgen->traffic_active) {
		int rc = pthread_cond_wait(&trafgen->traffic_cond,
				           &trafgen->traffic_mutex);
		BUG_ON(rc);
	}
	pthread_mutex_unlock(&trafgen->traffic_mutex);

	/* zeroes payload as well */
	memset(buf, 0, sizeof(buf));
	BUG_ON(sizeof(struct trafgen_payload) > sizeof(buf));

	/* this sender_id is used by the remote side to fill an INFO/trafgen:
	 * @est_info_trafgen.stats.src_id
	 */
	strncpy(payload->sender_id, gb.estimator_id, sizeof(payload->sender_id));

	// traffic_active is volatile
	while (trafgen->traffic_active) {
		int bytes = 0;
		int sleep_usec = 0;
		int pktlen = 0;

		pthread_mutex_lock(&trafgen->traffic_mutex);
		sleep_usec = trafgen->sleep_usec;
		pktlen = trafgen->traffic.pktlen;
		pthread_mutex_unlock(&trafgen->traffic_mutex);

		bytes = send(trafgen->sock, buf, pktlen, 0);
		SYSERR_ON(bytes != pktlen);

		DEB("snd UDP len %d usec interval %d", pktlen, sleep_usec);

		usleep(sleep_usec);
	}
}

void *
est_trafgen_sender_thread(void *arg)
{
	int rc = 0;
	struct trafgen_sender *trafgen = arg;

	trafgen->self = pthread_self();

	rc = est_trafgen_sender_create_socket(trafgen);
	if (rc) goto error;

	while (1) {
		/* the loop restarts every time we stop the sender */
		est_trafgen_start_sending_once_active(trafgen);
	}

	close(trafgen->sock);
	return NULL;

error:
	trafgen->status = rc;
	LOGTR("error: sender thread (with socket %d) failed", trafgen->sock);
	return NULL;
}

/* trafgen RECEIVER thread
 *
 * There is one thread for all sources but we can have one thread
 * listening on normal UDP and one on UDP/MGRP
 */

int
est_trafgen_receiver_create_socket(struct trafgen_receiver *trafgen)
{
	int rc = 0;
        int on = 0;

        trafgen->listen_addr.sin_family = AF_INET;
        trafgen->listen_addr.sin_addr.s_addr = htonl(INADDR_ANY);
        trafgen->listen_addr.sin_port = htons(trafgen->local_traffic_port);

	if (trafgen->type == EST_TRAFGEN_UDP_MGRP)
		trafgen->sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_MGRP);
	else if (trafgen->type == EST_TRAFGEN_UDP_NORMAL)
		trafgen->sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
        else
                BUG();

	if (trafgen->sock < 0) {
		rc = trafgen->sock;
		goto syserror;
	}

        /* Turn on IP_PKTINFO on receive to get dest ip address
         * See est_trafgen_keep_receiving() for an example
         */
        on = 1;
        rc = setsockopt(trafgen->sock, SOL_IP, IP_PKTINFO, &on, sizeof(on));
        if (rc) goto close_syserror;

	rc = bind(trafgen->sock,
                  (struct sockaddr *)&trafgen->listen_addr,
                  sizeof(struct sockaddr_in));
	if (rc)
		goto close_syserror;

	LOG("receiver listening on UDP port %d, MGRP is %s",
	    trafgen->local_traffic_port,
	    (trafgen->type == EST_TRAFGEN_UDP_MGRP ? "ON" : "OFF"));
	
	/* turn on RECV timestamps */

	// XXX not yet
/* 	if (trafgen->use_mgrp) { */
/* 		int on = 1; */
/* 		LOGFN("turning on MGRP timestamps on the receive socket"); */
/* 		rc = setsockopt(sock, SOL_MGRP, UDPEX_TSTAMP, &on, sizeof(on)); */
/* 		if (rc) */
/* 			goto close_syserror; */
/* 	} */

	return 0;

close_syserror:
	close(trafgen->sock);
syserror:
	LOGTR("error: failed to create receiver socket:\n%s", strerror(errno));
	return rc;
}

/* called with @recv->stats_mutex held */
static inline struct est_trafgen_stats *
est_trafgen_lookup_stats(struct trafgen_receiver *recv,
                         struct sockaddr_in *src,
                         struct sockaddr_in *dst)
{
        struct anode *node = NULL;
        /* @recv->stats_list is ALIST_NO_MUTEX: no internal mutex */
        int rc = alist_iterator_set(&recv->stats_list,
                                ALIST_FORWARD,
                                ALIST_BEFORE_FIRST,
                                NULL);
        BUG_ON(rc);

        while ((node = alist_iterator_next(&recv->stats_list)) != NULL) {
                struct est_trafgen_stats *stats = node->data;

                if (net_sockaddr_match(&stats->src, src) &&
                    net_sockaddr_match(&stats->dst, dst) &&
                    stats->type == recv->type)
                        return stats;
        }
        return NULL;
}

// Fills the @src_id and @dst_id of @stats by doing a reverse DNS
// lookup on the IP addresses of @src and @dst.
//
// XXX Caution!
// We are calling this function with a lock held and
// as we are receiving UDP packets.
//
// Is it a problem that the reverse name resolution
// may take some time? (the library code probably sleeps)
//
// Should we launch a thread instead and fill it when
// we get a response? In that case we need to ensure
// that the @stats pointer is still valid when we
// assign it.
static inline void
est_trafgen_fill_src_dst_id(struct est_trafgen_stats *stats)
{
        int rc = 0;
        char hostname[200];
        struct sockaddr_in *addr = NULL;

        addr = &stats->src;
        rc = getnameinfo((struct sockaddr *)addr,
                         sizeof(struct sockaddr_in),
                         hostname, sizeof(hostname),
                         NULL, 0, 0);
        SYSERR_ON(rc);
        est_info_extract_id_from_hostname(hostname, stats->src_id,
                                          sizeof(stats->src_id));

        addr = &stats->dst;
        rc = getnameinfo((struct sockaddr *)addr,
                         sizeof(struct sockaddr_in),
                         hostname, sizeof(hostname),
                         NULL, 0, 0);
        SYSERR_ON(rc);
        est_info_extract_id_from_hostname(hostname, stats->dst_id,
                                          sizeof(stats->dst_id));

}

/* called with @recv->stats_mutex held */
static inline struct est_trafgen_stats *
est_trafgen_create_new_stats(struct trafgen_receiver *recv,
                             struct sockaddr_in *src,
                             struct sockaddr_in *dst,
                             struct trafgen_payload *payload)
{
        int rc = 0;
        struct est_trafgen_stats *stats = NULL;
        struct timeval now;
        
        stats = malloc(sizeof(struct est_trafgen_stats));
        if (stats == NULL) goto error;

        est_trafgen_stats_init(stats);
        stats->is_sender = 0;
        stats->src = *src;
        stats->dst = *dst;
        stats->type = recv->type;

        /* Copy the host IDs withouth doing any reverse DNS lookup
         * on the @src and @dst addresses.
         *
         * We used to call est_trafgen_fill_src_dst_id() for that.
         *
         * Do not trust the string to have a null terminator.
         * It comes from the network receive payload.
         */
        strncpy(stats->dst_id, gb.estimator_id, sizeof(stats->dst_id));
        stats->dst_id[sizeof(stats->dst_id)-1] = 0;
        strncpy(stats->src_id, payload->sender_id,  sizeof(stats->dst_id));
        stats->src_id[sizeof(stats->src_id)-1] = 0;

        /* zeroes counters and sets the start timestamp */
        gettimeofday(&now, NULL);
        est_trafgen_stats_reset(stats, &now);

        rc = alist_insert_last_r(&recv->stats_list, stats, ALIST_OWNER);
        if (rc) goto error_free;

        return stats;

error_free:
        free(stats);
error:
        return NULL;
}


static inline void
est_trafgen_receiver_update_stats(struct trafgen_receiver *recv,
                                  struct sockaddr_in *src,
                                  struct sockaddr_in *dst,
                                  struct trafgen_payload *payload,
                                  int pktbytes)
{
        struct est_trafgen_stats *stats = NULL;

        pthread_mutex_lock(&recv->stats_mutex);

        stats = est_trafgen_lookup_stats(recv, src, dst);
        if (stats == NULL) {
                stats = est_trafgen_create_new_stats(recv, src, dst, payload);
        }
        BUG_ON(stats == NULL);

        /* global stats */
        recv->total_packets++;
        recv->total_bytes += pktbytes;

        /* local stats */
        stats->packets++;
        stats->bytes += pktbytes;
        stats->pktlen_max = cmp_max(stats->pktlen_max, pktbytes);
        stats->pktlen_min = cmp_min(stats->pktlen_min, pktbytes);

        pthread_mutex_unlock(&recv->stats_mutex);

        DEB("rcv UDP len %d count %d total_bytes %d",
            pktbytes, recv->total_packets, recv->total_bytes);
}

int
est_trafgen_keep_receiving(struct trafgen_receiver *trafgen)
{
	int rc = 0;
	struct sockaddr_in src;
        struct sockaddr_in dst;

	char buf[2000];
	struct iovec iov = {
		.iov_base = buf,
		.iov_len = sizeof(buf),
	};

        unsigned char ctrl[CMSG_SPACE(sizeof(struct in_pktinfo))];

	struct msghdr msg = {
		.msg_name = (void *)&src,
		.msg_namelen = sizeof(struct sockaddr_in),
		.msg_iov = &iov,
		.msg_iovlen = 1,
		.msg_control = ctrl,
		.msg_controllen = sizeof(ctrl),
		.msg_flags = 0,
	};

	memset(buf, 0, sizeof(buf));
	memset(ctrl, 0, sizeof(ctrl));
	memset(&src, 0, sizeof(struct sockaddr_in));
	memset(&dst, 0, sizeof(struct sockaddr_in));

	while (1) {
                struct trafgen_payload *payload = (struct trafgen_payload *)buf;
                struct cmsghdr *cmsg = NULL;
                struct in_pktinfo *pktinfo = NULL;
		int bytes = recvmsg(trafgen->sock, &msg, 0);
		if (bytes == -1) {
			rc = bytes;
			goto error;
		}

                /*
                 * IP_PKTINFO
                 *
                 * To get this we need to setsockopt(s, SOL_IP, IP_PKTINFO, ..)
                 *
                 * We are interested in struct in_pktinfo.ipi_addr which is
                 * the destination address in the packet header.
                 *
                 * See also ip(7)
                 */

                cmsg = CMSG_FIRSTHDR(&msg);
                if (cmsg &&
                    cmsg->cmsg_level == SOL_IP &&
                    cmsg->cmsg_type == IP_PKTINFO)
                        pktinfo = (struct in_pktinfo *)CMSG_DATA(cmsg);

                /* @src already set by recvmsg() in @msg.msg_name */
                /* set now @dst: override the ip address of the listen addr */
                BUG_ON(pktinfo == NULL);
                dst = trafgen->listen_addr;
                dst.sin_addr = pktinfo->ipi_addr;

                /* payload points to the receive buffer */
                /* must have enough space */
                BUG_ON(bytes < sizeof(struct trafgen_payload));

                est_trafgen_receiver_update_stats(trafgen, &src, &dst,
                                                  payload, bytes);
	}
	return 0;

error:
	return rc;
}

void *
est_trafgen_receiver_thread(void *arg)
{
	int rc = 0;
	struct trafgen_receiver *trafgen = arg;

	trafgen->self = pthread_self();

	rc = est_trafgen_receiver_create_socket(trafgen);
	if (rc) goto error;

	rc = est_trafgen_keep_receiving(trafgen);
	if (rc) goto error_close;

	close(trafgen->sock);
	return NULL;

error_close:
	close(trafgen->sock);
error:
	trafgen->status = rc;
	LOGTR("error: receiver thread (with socket %d) failed", trafgen->sock);
	return NULL;
}

/* returns the number of stats copied to the array */
static inline int
est_trafgen_receiver_extract_stats(struct trafgen_receiver *recv,
                                   struct est_trafgen_stats *stats,
                                   int stats_maxlen)
{
        int extracted = 0;
        struct anode *node = NULL;
        struct timeval now;

        pthread_mutex_lock(&recv->stats_mutex);

        /* @recv->stats_list is ALIST_NO_MUTEX: no internal mutex */
        int rc = alist_iterator_set(&recv->stats_list,
                                ALIST_FORWARD,
                                ALIST_BEFORE_FIRST,
                                NULL);
        BUG_ON(rc);

        gettimeofday(&now, NULL);

        while (extracted < stats_maxlen &&
               (node = alist_iterator_next(&recv->stats_list)) != NULL) {

                struct est_trafgen_stats *stats_orig = node->data;
                struct est_trafgen_stats *stats_copy = &stats[extracted];

                *stats_copy = *stats_orig;
                stats_copy->tstamp_end = now;
                /* zeroes counters and sets the start timestamp */
                est_trafgen_stats_reset(stats_orig, &now);
                extracted++;
        }

        pthread_mutex_unlock(&recv->stats_mutex);

        BUG_ON(extracted == stats_maxlen);
        return extracted;
}

static inline void
est_trafgen_receiver_copy_push_info(struct trafgen_receiver *recv)
{
        int i = 0;
        int extracted = 0;
        struct est_trafgen_stats stats[EST_TRAFGEN_STATS_MAX];

        extracted = est_trafgen_receiver_extract_stats(
                                recv, stats, EST_TRAFGEN_STATS_MAX);

        for (i = 0; i < extracted; i++) {
                struct est_info info;

                /* push even non-zero stats */

                /* @info is a union; this initializes whole @info */
                est_info_trafgen_init(&info.trafgen);

                /* duplicating IDs in header */
                strncpy(info.trafgen.hdr.key.src_id, stats[i].src_id,
                        sizeof(info.trafgen.hdr.key.src_id));
                strncpy(info.trafgen.hdr.key.dst_id, stats[i].dst_id,
                        sizeof(info.trafgen.hdr.key.dst_id));

                info.trafgen.stats = stats[i];
                info_alloc_copy_push(&info);
        }
}

void *
est_trafgen_info_thread(void *arg)
{
        struct trafgen_info *trafgen_info = arg;

        trafgen_info->self = pthread_self();

        BUG_ON(trafgen_info->recv_count < 1);
        BUG_ON(trafgen_info->recv_count > EST_TRAFGEN_INFO_MAX_RECV);

        while (1) {
                int i = 0;

                sleep(trafgen_info->report_every_period);
                for (i = 0; i < trafgen_info->recv_count; i++)
                        est_trafgen_receiver_copy_push_info(trafgen_info->recv[i]);
        }

        return NULL;
}

/* TRAFGEN_TRAFFIC prepare|start|stop */

static inline int
est_trafgen_get_remote_traffic_port(struct sockaddr_in *remote_estimator,
				    int *remote_traffic_port,
				    char *error,
				    int errorlen)
{
	/* The correct way to do this is to ask the remote estimator.
	 * But in most cases (other than cmdline overrides) we know how
	 * the local traffic port is calculated.
	 *
	 * So do lazy coding and skip the network communication!
	 */
		
	*remote_traffic_port =
		EST_LOCAL_TRAFFIC_PORT(ntohs(remote_estimator->sin_port));

	LOGFN("remote port is %d", *remote_traffic_port);
	return 0;
}

/* processes PREPARE action of TRAFGEN_TRAFFIC command and builds a reply */
void
est_trafgen_traffic_prepare(struct trafgen_sender *trafgen,
			    struct est_channel *ch,
                            struct est_trafgen_traffic_cmd *cmd,
			    struct est_trafgen_traffic_reply *reply)
{
	int rc = 0;
	pthread_t thread;

	if (trafgen != NULL) {
		char srcdst[100];
	        net_srcdst_sinaddr_str(srcdst, sizeof(srcdst),
				       &cmd->src, &cmd->dst);
		snprintf(reply->error, sizeof(reply->error),
		       "traffic session already running between %s", srcdst);
		goto error;
	}

	trafgen = est_pick_and_init_trafgen();
	if (trafgen == NULL) {
		snprintf(reply->error, sizeof(reply->error),
			 "error: no more traffic sessions available");
		goto error;
	}

	// XXX temp, must read from cmd */
	trafgen->use_mgrp = 0;
	
	trafgen->src = cmd->src;
	trafgen->dst = cmd->dst;

	rc = est_trafgen_get_remote_traffic_port(&cmd->src,
				  	         &trafgen->remote_traffic_port,
						 reply->error,
						 sizeof(reply->error));
	if (rc) goto error;

	/* Thread starts inactive, so traffic is not specified yet.
	 * The thread state is in gb.trafgen[] and can be looked up
	 * using src,dst as key.
	 */

	rc = pthread_create(&thread, NULL, est_trafgen_sender_thread, trafgen);
	BUG_ON(rc);
	rc = pthread_detach(thread);
	BUG_ON(rc);

	reply->success = 1;
	return;

error:
	reply->success = 0;
	LOGTR("%s", reply->error);
	return;
}

/* processes START/STOP action of TRAFGEN_TRAFFIC command and builds a reply */
void
est_trafgen_traffic_startstop(struct trafgen_sender *trafgen,
		              int start,
		              struct est_channel *ch,
                              struct est_trafgen_traffic_cmd *cmd,
			      struct est_trafgen_traffic_reply *reply)
{
	if (trafgen == NULL) {
		char srcdst[100];

	        net_srcdst_sinaddr_str(srcdst, sizeof(srcdst),
				       &cmd->src, &cmd->dst);
		snprintf(reply->error, sizeof(reply->error),
		       "no traffic session running between %s", srcdst);

		reply->success = 0;
		LOGTR("%s", reply->error);
		return;
	}

	/* START action */
	if (start) {
		int rc = 0;
		int sleep_usec = 0;
		
		if (cmd->traffic.type != EST_TRAFGEN_UDP_NORMAL &&
		    cmd->traffic.type != EST_TRAFGEN_UDP_MGRP) {
			snprintf(reply->error, sizeof(reply->error),
			         "invalid traffic type %d: "
                                 "only UDP supported (both MGRP or normal)",
				 cmd->traffic.type);
			goto error;
		}

		if (cmd->traffic.bps <= 0) {
			snprintf(reply->error, sizeof(reply->error),
			         "invalid traffic bps %d",
				 cmd->traffic.bps);
			goto error;
		}

		if (cmd->traffic.pktlen <= 0 ||
		    cmd->traffic.pktlen > EST_TRAFGEN_MAX_PKTLEN) {
			snprintf(reply->error, sizeof(reply->error),
			         "invalid traffic pktlen %d",
				 cmd->traffic.pktlen);
			goto error;
		}

		sleep_usec = net_bps_to_usec_interval(cmd->traffic.bps,
						      cmd->traffic.pktlen);
		if (sleep_usec == -1) {
			snprintf(reply->error, sizeof(reply->error),
			         "invalid usec interval between packets %d",
				 sleep_usec);
			goto error;
		}

		LOGFN("starting bps %d pktlen %d sleep_usec %d",
		      cmd->traffic.bps, cmd->traffic.pktlen, sleep_usec);
			
		pthread_mutex_lock(&trafgen->traffic_mutex);

		/* Set the traffic rate (by calculating a sleep interval)
		 *
		 * The thread keeps reading this before it
		 * sends every packet, so that a rate change
		 * can take effect immediately.
		 */
		 trafgen->sleep_usec = sleep_usec;
		 trafgen->traffic = cmd->traffic;

		/* Signal the thread to become active.
		 *
		 * If the thread is already active then this has no effect.
		 * Otherwise the thread wakes up and starts sending traffic
		 *
		 * This happens in: est_trafgen_start_sending_once_active()
		 */
		trafgen->traffic_active = 1;
		rc = pthread_cond_signal(&trafgen->traffic_cond);
		BUG_ON(rc);

		pthread_mutex_unlock(&trafgen->traffic_mutex);
	
	/* STOP action */
	} else {
		if (!trafgen->traffic_active) {
			snprintf(reply->error, sizeof(reply->error),
				 "already stopped");
			goto error;
		}
		/* Stops the while loop (but not the thread in) in:
		 * est_trafgen_start_sending_once_active().
		 */
		trafgen->traffic_active = 0;
	}

	reply->success = 1;
	return;

error:
	reply->success = 0;
	LOGTR("%s", reply->error);
	return;
}

void
est_trafgen_process_traffic(struct est_channel *ch,
                            struct est_trafgen_traffic_cmd *cmd)
{
	int ok = 0;
	int rc = 0;
	char error[200];
	struct est_trafgen_traffic_reply reply;
	int trafgen_index = 0;
	struct trafgen_sender *trafgen = NULL;

	/* find existing traffic session between @src and @dst */
	trafgen_index = est_trafgen_find_index(gb.snd_trafgen,
					       EST_MAX_TRAFGEN_SENDERS,
					       &cmd->src, &cmd->dst);

	/* find the trafgen state to pass to est_trafgen_traffic_xxx() */
	/* can be null for the PREPARE action */
	if (trafgen_index > -1)
		trafgen = &gb.snd_trafgen[trafgen_index];

	/* prepare the reply */
	memset(&reply, 0, sizeof(reply));
	ok = est_fill_reply_header(&reply.hdr,
				   EST_TYPE_TRAFGEN_TRAFFIC,
				   EST_TYPE_REPLY_FINAL);
	BUG_ON(!ok);

	/* fill the reply and perform any side effects */
	if (cmd->action == EST_TRAFGEN_PREPARE)
		est_trafgen_traffic_prepare(trafgen, ch, cmd, &reply);
	else if (cmd->action  == EST_TRAFGEN_START)
		est_trafgen_traffic_startstop(trafgen, 1, ch, cmd, &reply);
	else if (cmd->action  == EST_TRAFGEN_STOP)
		est_trafgen_traffic_startstop(trafgen, 0, ch, cmd, &reply);
	else {
		snprintf(reply.error, sizeof(reply.error),
			 "invalid action 0x%x", cmd->action);
		reply.success = 0;
		LOGTR("%s", reply.error);
	}

	rc = est_send_ctrlmsg(ch, &reply, error, sizeof(error));
	if (rc) {
		LOGTR("failed to send TRAFGEN_TRAFFIC reply: %s", error);
		goto error_noreply;
	}
	return;

error_noreply:
	LOGFN("failed to process TRAFGEN_TRAFFIC cmd");
	return;
}

#ifdef TRAFGEN_REPORT
void
est_trafgen_process_report(struct est_channel *ch,
		           struct est_trafgen_report_cmd *cmd)
{
	int ok = 0;
	int rc = 0;
	char error[200];
	struct est_trafgen_report_reply reply;
	int trafgen_index = 0;
	struct trafgen_sender *trafgen = NULL;

	/* find existing traffic session between @src and @dst */
	trafgen_index = est_trafgen_find_index(gb.snd_trafgen,
					       EST_MAX_TRAFGEN_SENDERS,
					       &cmd->src, &cmd->dst);

	/* find the trafgen state */
	if (trafgen_index > -1)
		trafgen = &gb.snd_trafgen[trafgen_index];

	/* prepare the reply */
	memset(&reply, 0, sizeof(reply));
	ok = est_fill_reply_header(&reply.hdr,
				   EST_TYPE_TRAFGEN_REPORT,
				   EST_TYPE_REPLY_FINAL);
	BUG_ON(!ok);

	snprintf(reply.error, sizeof(reply.error), "not implemented yet");
	reply.success = 0;

	rc = est_send_ctrlmsg(ch, &reply, error, sizeof(error));
	if (rc) {
		LOGTR("failed to send TRAFGEN_TRAFFIC reply: %s", error);
		goto error_noreply;
	}
	return;

error_noreply:
	LOGFN("failed to process TRAFGEN_TRAFFIC cmd");
	return;
	LOGFN("not yet");
}
#endif
