/*
 * Created on   : Sat Jul 24 12:26:09 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * adar.c
 *
 * Abstract dynamic array
 *
 * $Id: adar.c,v 1.6 2005/03/19 02:20:18 pavlos Exp $
 */

#include "adar.h"
#include "logerr.h"
#include "mem.h"
#include "cmp.h"

int
adar_init(adarray *ar, int nelem, bytes_free_func *free_func)
{
    int res = 0;

    logerr_passert(ar);

    ar->p = NULL;
    ar->nelem = 0;
    ar->free_func = free_func;

    res = adar_resize(ar, nelem);
    return res;
}



int
adar_done(adarray *ar)
{
    int res = 0;

    logerr_passert(ar);

    res = adar_resize(ar, 0);
    return res;
}


int
adar_expand(adarray *ar, int byelements)
{
    int res = 0;

    logerr_passert(ar);
    logerr_assert_str(byelements > 0,
                "Non-positive expand value");

    res = adar_resize(ar, ar->nelem + byelements);

    return res;
}


int
adar_shrink(adarray *ar, int byelements)
{
    int res = 0;

    logerr_passert(ar);
    logerr_assert_str(byelements > 0,
                "Non-positive expand value");

    res = adar_resize(ar, cmp_max(0, ar->nelem - byelements));

    return res;
}


int
adar_resize(adarray *ar, int nelem)
{
    int copy_first =  0;
    int copy_last  = -1;

    int free_first =  0;
    int free_last  = -1;

    void **original_p = NULL;
    void **new_p = NULL;

    int i = 0;

    logerr_passert(ar);

    if (nelem < 0)
        return ADAR_EINVAL;

    if (nelem == ar->nelem)
        return ADAR_OK;

    original_p = ar->p;

    if (nelem > ar->nelem) {
        copy_first = 0;
        copy_last  = ar->nelem - 1;

    } else {
        copy_first = 0;
        copy_last  = nelem - 1;
        free_first = nelem;
        free_last  = ar->nelem - 1;
    }

    if (nelem > 0) {
        new_p = calloc(nelem, sizeof(void *));
        logerr_return_sys(new_p != NULL,
                "Failed to allocate memory for new array size",
                ADAR_ENOMEM);
    }
        
    /* free extra elements */

    if (ar->free_func != NULL) {

        for (i = free_first; i <= free_last ; i++) {
            if (original_p[i] != NULL) {
                ar->free_func(original_p[i]);
                original_p[i] = NULL;
            }
        }
    }

    /* copy elements */

    for (i = copy_first; i <= copy_last ; i++) {
        new_p[i] = original_p[i];
    }

    if (original_p != NULL) {
        free(original_p);
        original_p = NULL;
    }

    ar->nelem = nelem;
    ar->p = new_p;


    return ADAR_OK;
}

void
adar_log(int level, adarray *ar)
{
    int i = 0;

    logerr_passert(ar);

    logerr_log(level, "adarray: nelem %d", ar->nelem);
    logerr_log(level, "adarray: fre_func %p", ar->free_func);

    for (i = 0; i < ar->nelem; i++) {
        void *ptr = ar->p[i];
        if (ptr == NULL)
            logerr_log(level, "adarray:    [%d] = EMPTY", i);
        else
            logerr_log(level, "adarray:    [%d] = %p", i, ptr);
    }
}


