/*
 * Created on   : Tue Sep 28 19:12:40 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * ahash.c
 *
 * Abstract hash
 *
 * $Id: ahash.c,v 1.10 2005/12/12 16:39:55 pavlos Exp $
 */

#include "ahash.h"
#include "logerr.h"
#include "mem.h"

static int ahash_match_hash_index(void *data, void *key);


int
ahash_init(ahash *hash,
        uint level1_p2elem, uint level2_p2elem,
        cmp_search_func *search_func, cmp_hash64_func *hash_func)
{
    int res = 0;
    
    logerr_passert(hash);

    memset(hash, 0, sizeof(ahash));

    logerr_return_str(search_func != NULL && hash_func != NULL,
        "Undefined key handling functions: either search_func or hash_func",
        AHASH_ERROR);

    hash->level1_p2elem = level1_p2elem;
    hash->level1_mask   = (BIT_ONE64 >> (64 - level1_p2elem));
    hash->level1_mask <<= level2_p2elem;

    hash->level2_p2elem = level2_p2elem;
    hash->level2_mask   = (BIT_ONE64 >> (64 - level2_p2elem));

    hash->search_func = search_func;
    hash->hash_func = hash_func;

    res = adar_init(&hash->level1, 1 << hash->level1_p2elem, &ahash_node_free);
    logerr_return_str(res == ADAR_OK,
        "Failed to initialize level1 array",
        AHASH_ERROR);

    return AHASH_OK;
}

int
ahash_done(ahash *hash)
{
    int res = 0;

    res = adar_done(&hash->level1);
    logerr_return_str(res == ADAR_OK,
        "Failed to cleanup level1 array",
        AHASH_ERROR);
    
    return AHASH_OK;
}


void *
ahash_lookup_pos(ahash *hash, void *key, ahash_pos *pos)
{
    void *data = NULL;
    ahash_node *node = NULL;
    uint64_t hkey = 0;
    ahash_pos lpos;

    if (pos == NULL)
        pos = &lpos;

    logerr_passert(hash);
    logerr_passert(key);
    logerr_passert(pos);

    hkey = hash->hash_func(NULL, key);

    memset((pos), 0, sizeof(ahash_pos));

    pos->hash = hash;
    pos->level1_index = ahash_level1_index(hash, hkey);

    pos->level2_index = ahash_level2_index(hash, hkey);
    
    logerr_assert_str(pos->level1_index < (1 << hash->level1_p2elem),
        "level1 index out of bounds");
    logerr_assert_str(pos->level2_index < (1 << hash->level2_p2elem),
        "level2 index out of bounds");
/*     logerr_log(999, "hkey(%llu), i1(%u) i2(%u)", */
/*         hkey, pos->level1_index, pos->level2_index); */

    node = hash->level1.p[pos->level1_index];
    if (node == NULL)
        return NULL;

    data = node->elements.p[pos->level2_index];

    if (data != NULL && hash->search_func(data, key) != 0) {
        /* data occupies the position by does not match the key */
        pos->overflowed = 1;
        data = NULL;
    }

    /* look in overflow list */

    /* XXX this is necessary, because if there is no data at level2_index,
     * it is possible to have it in the overflow list.
     * This can be optimized to move data from overflow list whenever
     * the level2_index position becomes available
     */

    if (data == NULL) {
        pos->overflow_node = alist_search_for_node(
                &node->overflow,
                ALIST_FORWARD, ALIST_BEFORE_FIRST, NULL,
                hash->search_func, key);

        if (pos->overflow_node != NULL)
            data = pos->overflow_node->data;
    }

    if (data != NULL)
        pos->found = 1;

    return data;
}


int
ahash_insert_pos(ahash *hash, void *data, void *key, ahash_pos *pos)
{
    int res = 0;
    ahash_node *node = NULL;
    ahash_pos lpos;

    if (pos == NULL)
        pos = &lpos;

    logerr_passert(hash);
    logerr_passert(data);
    logerr_passert(key);
    logerr_passert(pos);

    ahash_lookup_pos(hash, key, pos);

    if (pos->found) {
        return AHASH_EEXISTS;
    }

    node = hash->level1.p[pos->level1_index];

    /* create level2 storage if not there */

    if (node == NULL) {

        node = malloc(sizeof(ahash_node));
        logerr_passert(node);

        res = ahash_node_init(node, (1 << hash->level2_p2elem));
        logerr_assert_str(res == AHASH_OK,
                "Failed to initialize hash node");

        hash->level1.p[pos->level1_index] = node;
    }

    if (node->elements.p[pos->level2_index] == NULL) {

        /* the slot is available; just put it in */
        node->elements.p[pos->level2_index] = data;

    } else {

        /* insert into overflow list */
        res = alist_insert_last_r(&node->overflow, data, ALIST_NOT_OWNER);
        logerr_assert_str(res == ALIST_OK,
                "Failed to insert data into overflow list");
        pos->overflowed = 1;
        pos->overflow_node = alist_last_node(&node->overflow);
    }

    pos->found = 1;

    return AHASH_OK;
}

int
ahash_delete_pos(ahash *hash, void *key, ahash_pos *pos)
{
    int res = 0;
    ahash_node *node = NULL;
    ahash_pos lpos;
    anode *overflow_node = NULL;

    if (pos == NULL)
        pos = &lpos;

    logerr_passert(hash);
    logerr_passert(key);
    logerr_passert(pos);

    ahash_lookup_pos(hash, key, pos);

    if (!pos->found) {
        return AHASH_EMISSING;
    }

    node = hash->level1.p[pos->level1_index];

    if (pos->overflowed) {
        /* remove from list */
        res = alist_remove_destroy_node(&node->overflow, pos->overflow_node);
        logerr_fatal_str(res == ALIST_OK,
                "Failed to remove and destroy the node from the oveflow list");

    } else {
        /* remove from elements array */
        node->elements.p[pos->level2_index] = NULL;

        /* does anyone in the overflow list hash to this position? */
        overflow_node = alist_search_for_node(
                &node->overflow,
                ALIST_FORWARD, ALIST_BEFORE_FIRST, NULL,
                &ahash_match_hash_index, pos);
        
        if (overflow_node != NULL) {
            node->elements.p[pos->level2_index] = overflow_node->data;

            /* XXX data owner? */
            res = alist_remove_destroy_node(&node->overflow, overflow_node);
            logerr_fatal_str(res == ALIST_OK,
                "Failed to remove and destroy the node from the oveflow list");
        }
    }
    
    return AHASH_OK;
}

static int
ahash_match_hash_index(void *data, void *key)
{
    uint64_t hkey = 0;
    ahash_pos *pos = key;

    uint level1_index;
    uint level2_index;

    logerr_passert(pos);
    logerr_passert(data);

    hkey = pos->hash->hash_func(data, NULL);
    level1_index = ahash_level1_index(pos->hash, hkey);
    level2_index = ahash_level2_index(pos->hash, hkey);

    logerr_assert_str(level1_index == pos->level1_index,
        "First level index does not match");

    return  (level2_index == pos->level2_index) ? 0 :
            (level2_index  < pos->level2_index) ? -1 : 1;
}

/* ahash_node functions */

int
ahash_node_init(ahash_node *node, int nelem)
{
    int res = 0;

    logerr_passert(node);

    res = adar_init(&node->elements, nelem, NULL);
    logerr_return_str(res == ADAR_OK,
        "Failed to initialize element array",
        AHASH_ERROR);

    res = alist_init(&node->overflow, ALIST_NO_MUTEX, NULL);
    logerr_return_str(res == ALIST_OK,
        "Failed to initialize overflow list",
        AHASH_ERROR);

    return AHASH_OK;
}

int
ahash_node_done(ahash_node *node)
{
    int res = 0;

    logerr_passert(node);

    res = adar_done(&node->elements);
    logerr_return_str(res == ADAR_OK,
        "Failed to cleanup element array",
        AHASH_ERROR);

    res = alist_done(&node->overflow);
    logerr_return_str(res == ALIST_OK,
        "Failed to cleanup overflow list",
        AHASH_ERROR);

    return AHASH_OK;
}

void
ahash_node_free(void *buf)
{
    int res = 0;
    ahash_node *node = buf;

    logerr_passert(node);

    res = ahash_node_done(node);
    logerr_fatal_str(res == AHASH_OK,
        "Failed to cleanup hash node");

    free(node);
}


/* logging functions */

typedef struct ahash_log_info {
    int level;
    uint index;
    char format[100];
    char prefix[100];
    cmp_log_func *log_func;
    uint *memptr;
} ahash_log_info;

int ahash_data_log(void *arg, void *data);

void
ahash_log(int level, ahash *hash, cmp_log_func *log_func)
{
    int i = 0;
    uint mem = 0;

    logerr_passert(hash);

    mem += sizeof(ahash);
    mem += hash->level1.nelem * sizeof(void *);

    logerr_log(level, "----- ahash(%p) ------", hash);

    logerr_log(level, "    level1_p2elem(%u) level1_mask(%.16x)",
        hash->level1_p2elem, hash->level1_mask);
    logerr_log(level, "    level2_p2elem(%u) level2_mask(%.16x)",
        hash->level2_p2elem, hash->level2_mask);
    logerr_log(level, "    search_func(%p) hash_func(%p)",
        hash->search_func, hash->hash_func);
    logerr_log(level, "    level1.nelem(%d) free_func(%p)",
        hash->level1.nelem, hash->level1.free_func);

    for (i = 0; i < hash->level1.nelem; i++) {
        ahash_node *node = hash->level1.p[i];

        if (node != NULL) {

            ahash_log_info log_info;
            int j = 0;
            int res = 0;

            logerr_log(level, "    ---- node(%d) ----", i);

            mem += sizeof(ahash_node);
            mem += node->elements.nelem * sizeof(void *);
            mem += node->overflow.elements * sizeof(anode);

            memset(&log_info, 0, sizeof(log_info));
            log_info.level = level;
            log_info.log_func = log_func;
            log_info.memptr = &mem;
            strcpy(log_info.format, "    elem[%d] : ");

            for (j = 0; j < node->elements.nelem; j++) {
                void *data = node->elements.p[j];

                if (data != NULL) {
                    log_info.index = j;
                    ahash_data_log(&log_info, data);
                }
            }

            memset(&log_info, 0, sizeof(log_info));
            log_info.level = level;
            log_info.log_func = log_func;
            log_info.memptr = &mem;
            strcpy(log_info.format, "    over[%d] : ");

            if (alist_elements(&node->overflow) > 0) {
                res = alist_foreach_r(&node->overflow,
                    ALIST_FORWARD, ALIST_BEFORE_FIRST,
                    &ahash_data_log, &log_info);
                logerr_fatal_str(res == ALIST_OK,
                    "Failed to log foreach data element of the overflow list");
            }
        }
    }

    logerr_log(level, "----- ahash(%p): total mem: %u bytes", hash, mem);
}


int
ahash_data_log(void *arg, void *data)
{
    ahash_log_info *log_info = arg;

    logerr_passert(log_info);
    logerr_passert(log_info->prefix);
    logerr_passert(log_info->log_func);

    sprintf(log_info->prefix, log_info->format, log_info->index);
    log_info->log_func(log_info->level, log_info->prefix, data,
        log_info->memptr);

    log_info->index++;
    
    return 1;
}



uint64_t
ahash_str64(unsigned char *str)
{
    /* XXX string hash function from: http://www.cs.yorku.ca/~oz/hash.html */

    uint64_t hash = 5381;
    int c;

    while ((c = *str++)) {
        /* hash * 33 + c */
        hash = ((hash << 5) + hash) + c;
    }

    return hash;
}

