/*
 * Created on   : Tue Jun 22 09:13:53 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * alist.c
 *
 * Generic list
 *
 * $Id: alist.c,v 1.28 2005/06/28 00:46:11 pavlos Exp $
 */

#include "alist.h"
#include "logerr.h"
#include "alog.h"

#include <stdio.h>

int
alist_init_long(alist *l,
        int use_mutex,
        bytes_free_func *free_func,
        cmp_cmp_func *cmp_func,
        rbuf_mgr *rbmgr)
{
    int res = 0;

    logerr_passert(l);

    l->use_mutex = use_mutex;
    l->elements = 0;
    l->first = NULL;
    l->current = NULL;
    l->free_func = free_func;
    l->cmp_func = cmp_func;
    l->idir = ALIST_FORWARD;
    l->ipos = ALIST_BEFORE_FIRST;
    l->rbmgr = rbmgr;

    if (use_mutex) {
        res = pthread_mutex_init(&l->access_mutex, NULL);
        if (res != 0) {
            logerr_str(res == 0, strerror(res));
            logerr_return_str(res == 0,
                        "Failed to initialize list mutex",
                        ALIST_ERROR);
        }
    }

    return ALIST_OK;
}


int
alist_done(alist *l)
{
    int res = 0;
    int locked = 0;

    void cleanup(void)
    {
        if (locked)
            alist_unlock(l);
    }

    logerr_passert(l);

    alist_lock(l);
    locked = 1;

    res = alist_destroy_all_nodes(l);
    logerr_creturn_str(res == ALIST_OK,
                "Failed to remove all nodes",
                ALIST_ERROR, cleanup);

    l->elements = 0;
    l->first = NULL;
    l->current = NULL;
    l->free_func = NULL;
    l->idir = ALIST_FORWARD;
    l->ipos = ALIST_BEFORE_FIRST;

    alist_unlock(l);
    locked = 0;

    if (l->use_mutex) {
        res = pthread_mutex_destroy(&l->access_mutex);
        if (res != 0) {
            logerr_str(res == 0, strerror(res));
            logerr_creturn_str(res == 0,
                        "Failed to destroy list mutex",
                        ALIST_ERROR, cleanup);
        }
    }
    l->use_mutex = 0;

    return ALIST_OK;
}


void
alist_lock(alist *l)
{
    int res = 0;

    logerr_passert(l);

    if (!l->use_mutex)
        return;

    res = pthread_mutex_lock(&l->access_mutex);
    logerr_fatal_str(res == 0, strerror(res));
}


void
alist_unlock(alist *l)
{
    int res = 0;

    logerr_passert(l);

    if (!l->use_mutex)
        return;

    res = pthread_mutex_unlock(&l->access_mutex);
    logerr_fatal_str(res == 0, strerror(res));
}


int
alist_elements_r(alist *l)
{
    int res = ALIST_ERROR;

    alist_lock(l);
    res = alist_elements(l);
    alist_unlock(l);

    return res;
}


int
alist_insert_here_r(alist *l, void *data, int data_owner,
        int where, int dir, anode *enode)
{
    int res = ALIST_OK;
    int locked = 0;
    anode *node = NULL;

    void cleanup(void)
    {
        if (locked)
            alist_unlock(l);
        if (node != NULL)
            res = anode_destroy(node, l->free_func);
            logerr_assert_str(res == ANODE_OK, "Failed to clean up.");
    }

    node = anode_create(data, data_owner, l->rbmgr);
    logerr_creturn_str(node != NULL,
                "Failed to create node to insert",
                ALIST_ERROR, cleanup);

    alist_lock(l);
    locked = 1;

    res = alist_insert_node_here(l, where, dir, enode, node);

    logerr_creturn_str(res == ALIST_OK,
                "Failed to insert node",
                res, cleanup);

    alist_unlock(l);
    locked = 0;

    return res;
}


void *
alist_remove_here_r(alist *l, int where)
{
    anode *node = NULL;
    void *data = NULL;
    int res = 0;

    alist_lock(l);

    if (where == ALIST_AT_FIRST)
        node = alist_first_node(l);

    else if (where == ALIST_AT_LAST)
        node = alist_last_node(l);

    else
        logerr_assert_str(0, "Invalid 'where' flag");

    if (node != NULL) {
        alist_detach_node(l, node);
    }

    /* no need to keep the lock any further; node is detached  */
    alist_unlock(l);

    if  (node != NULL) {
        data = node->data;
        /* make sure that data is not going to be deallocated */
        node->data_owner = 0;
        res = anode_destroy(node, NULL);
        logerr_assert_str(res == ANODE_OK, "Failed to destroy node");
    }

    return data;
}


int
alist_remove_all_r(alist *l)
{
    int res = 0;

    alist_lock(l);
    res = alist_destroy_all_nodes(l);
    alist_unlock(l);

    return res;
}


void *
alist_search_r(alist *l, int dir, int pos,
                cmp_search_func *search_func, void *key)
{
    anode *node = NULL;

    logerr_assert_str(pos != ALIST_AT_CURRENT, "Invalid position");

    alist_lock(l);
    node = alist_search_for_node(l, dir, pos, NULL, search_func, key);
    alist_unlock(l);

    return ((node != NULL) ? node->data : NULL);
}


int
alist_foreach_r(alist *l, int dir, int pos,
                cmp_foreach_func *foreach_func, void *arg)
{
    int res = 0;

    logerr_assert_str(pos != ALIST_AT_CURRENT, "Invalid position");

    alist_lock(l);
    res = alist_foreach_node(l, dir, pos, NULL, foreach_func, arg);
    alist_unlock(l);

    return res;
}


int
alist_iterator_set(alist *l, int dir, int pos, anode *enode)
{
    logerr_passert(l);
    l->idir = dir;
    l->ipos = pos;

    switch (l->ipos) {

        case ALIST_BEFORE_FIRST:
        case ALIST_AFTER_LAST:
            l->current = NULL;
            break;

        case ALIST_AT_FIRST:
            l->current = alist_first_node(l);
            break;

        case ALIST_AT_LAST:
            l->current = alist_last_node(l);
            break;

        case ALIST_AT_CURRENT:
            logerr_passert(enode);
            l->current = enode;
            break;

        default:
            logerr_assert_str(0, "Invalid iterator position");
            break;
    }

    return ALIST_OK;
}


anode *
alist_iterator_prev(alist *l)
{
    anode *node = NULL;
    int orig_idir = l->idir;

    logerr_passert(l);

    /* go back by reversing direction temporarily and going to next */

    if (l->idir == ALIST_FORWARD)
        l->idir = ALIST_BACKWARD;

    else if (l->idir == ALIST_BACKWARD)
        l->idir = ALIST_FORWARD;

    else if (l->idir == ALIST_FCIRCULAR)
        l->idir = ALIST_BCIRCULAR;

    else if (l->idir == ALIST_BCIRCULAR)
        l->idir = ALIST_FCIRCULAR;

    else logerr_bug(0);

    node = alist_iterator_next(l);

    /* restore original direction */
    l->idir = orig_idir;

    return node;
}

anode *
alist_iterator_next(alist *l)
{
    logerr_passert(l);
    logerr_assert_str(
        l->idir == ALIST_FORWARD ||
        l->idir == ALIST_BACKWARD ||
        l->idir == ALIST_FCIRCULAR ||
        l->idir == ALIST_BCIRCULAR,
        "Invalid iterator direction");

    switch (l->ipos) {

        case ALIST_BEFORE_FIRST:

            if (l->idir == ALIST_FORWARD) {
                l->current = alist_first_node(l);

            } else if (l->idir == ALIST_BACKWARD) {
                l->current = NULL;

            } else if (l->idir == ALIST_FCIRCULAR) {
                l->current = alist_first_node(l);

            } else if (l->idir == ALIST_BCIRCULAR) {
                l->current = alist_last_node(l);
            }
            break;

        case ALIST_AFTER_LAST:

            if (l->idir == ALIST_FORWARD) {
                l->current = NULL;

            } else if (l->idir == ALIST_BACKWARD) {
                l->current = alist_last_node(l);

            } else if (l->idir == ALIST_FCIRCULAR) {
                l->current = alist_first_node(l);

            } else if (l->idir == ALIST_BCIRCULAR) {
                l->current = alist_last_node(l);
            }
            break;

        case ALIST_AT_FIRST:

            if (l->idir == ALIST_FORWARD) {
                logerr_passert(l->current);
                l->current = l->current->next;
                if (l->current == alist_first_node(l))
                    l->current = NULL;

            } else if (l->idir == ALIST_BACKWARD) {
                l->current = NULL;

            } else if (l->idir == ALIST_FCIRCULAR) {
                logerr_passert(l->current);
                l->current = l->current->next;

            } else if (l->idir == ALIST_BCIRCULAR) {
                logerr_passert(l->current);
                l->current = l->current->prev;
            }
            break;

        case ALIST_AT_LAST:

            if (l->idir == ALIST_FORWARD) {
                l->current = NULL;

            } else if (l->idir == ALIST_BACKWARD) {
                logerr_passert(l->current);
                l->current = l->current->prev;
                if (l->current == alist_last_node(l))
                    l->current = NULL;

            } else if (l->idir == ALIST_FCIRCULAR) {
                logerr_passert(l->current);
                l->current = l->current->next;

            } else if (l->idir == ALIST_BCIRCULAR) {
                logerr_passert(l->current);
                l->current = l->current->prev;
            }
            break;

        case ALIST_AT_CURRENT:

            logerr_passert(l->current);

            if (l->idir == ALIST_FORWARD) {
                l->current = l->current->next;
                if (l->current == alist_first_node(l))
                    l->current = NULL;

            } else if (l->idir == ALIST_BACKWARD) {
                l->current = l->current->prev;
                if (l->current == alist_last_node(l))
                    l->current = NULL;

            } else if (l->idir == ALIST_FCIRCULAR) {
                logerr_passert(l->current);
                l->current = l->current->next;

            } else if (l->idir == ALIST_BCIRCULAR) {
                logerr_passert(l->current);
                l->current = l->current->prev;
            }
            break;

        default:
            logerr_assert_str(0, "Invalid iterator position");
            break;
    }

    l->ipos = ALIST_AT_CURRENT;

    if (l->current == NULL) {
        l->ipos = (l->idir == ALIST_FORWARD) ? ALIST_AFTER_LAST :
                                               ALIST_BEFORE_FIRST;

    } else if (l->current == alist_first_node(l)) {
        l->ipos = ALIST_AT_FIRST;

    } else if (l->current == alist_last_node(l)) {
        l->ipos = ALIST_AT_LAST;
    }

    return l->current;
}


int
alist_insert_node_here(alist *l, int where, int dir,
                anode *enode, anode *nnode)
{
    int res = ALIST_OK;

    switch (where) {
        
        case ALIST_BEFORE_FIRST:

            res = alist_insert_node_first(l, nnode);
            break;

        case ALIST_AFTER_FIRST:

            res = alist_insert_node_after(l, alist_first_node(l), nnode);
            break;

        case ALIST_AFTER_LAST:

            res = alist_insert_node_last(l, nnode);
            break;

        case ALIST_BEFORE_LAST:

            res = alist_insert_node_before(l, alist_last_node(l), nnode);
            break;

        case ALIST_BEFORE_CURRENT:

            res = alist_insert_node_before(l, enode, nnode);
            break;

        case ALIST_AFTER_CURRENT:

            res = alist_insert_node_after(l, enode, nnode);
            break;

        case ALIST_ORDERED:
            res = alist_insert_node_ordered(l, nnode, dir);
            break;

        default:
            logerr_assert_str(0, "Invalid insert node position");
            break;
    }

    return res;
}

int
alist_insert_node_after(alist *l, anode *enode, anode *nnode)
{
    logerr_passert(l);
    logerr_passert(nnode);

    logerr_return_str(l->elements < ALIST_MAX_ELEMENTS,
                "Reached maximum number of elements",
                ALIST_ERROR);

    if (enode == NULL) {
        /* list should be empty */
        logerr_assert_str(alist_empty(l), "List not empty");

        nnode->prev = nnode;
        nnode->next = nnode;
        l->first = nnode;

    } else {
        /* list should not be empty */
        logerr_assert_str(!alist_empty(l), "List is empty");
        logerr_passert(enode->next);

        nnode->prev = enode;
        nnode->next = enode->next;
        enode->next->prev = nnode;
        enode->next = nnode;
    }

    l->elements++;

    return ALIST_OK;
}


int
alist_insert_node_before(alist *l, anode *enode, anode *nnode)
{
    int res = 0;

    logerr_passert(l);

    if (enode == NULL) {
        /* list should be empty so it does not matter where we add */
        res = alist_insert_node_after(l, NULL, nnode);
        logerr_return_str(res == ALIST_OK, "Insert failed", res);

    } else {
        /* list should not be empty  */
        res = alist_insert_node_after(l, enode->prev, nnode);
        logerr_return_str(res == ALIST_OK, "Insert failed", res);
        if (enode == alist_first_node(l))
            l->first = nnode;
    }
    
    return res;
}


int
alist_insert_node_first(alist *l, anode *nnode)
{
    int res = 0;
    anode *enode = NULL;

    logerr_passert(l);
    enode = alist_first_node(l);
    res = alist_insert_node_before(l, enode, nnode);
    logerr_return_str(res == ALIST_OK, "Insert first failed", res);

    return res;
}


int
alist_insert_node_last(alist *l, anode *nnode)
{
    int res = 0;
    anode *enode = NULL;

    logerr_passert(l);

    enode = alist_last_node(l);

    res = alist_insert_node_after(l, enode, nnode);
    logerr_return_str(res == ALIST_OK, "Insert first failed", res);

    return res;
}


int
alist_insert_node_ordered(alist *l, anode *nnode, int dir)
{
    int res = ALIST_OK;
    anode *enode = NULL;
    int start_pos = 0;
    int insert_where = 0;
    int stop_condition = 0;
    int cmp = 0;

    logerr_passert(l);
    logerr_fatal_str(l->cmp_func != NULL,
                "No compare function specified");

    if (dir == ALIST_FORWARD) {
        start_pos = ALIST_BEFORE_FIRST;
        insert_where = ALIST_BEFORE_CURRENT;
        stop_condition = 1;
        
    } else if (dir == ALIST_BACKWARD) {
        start_pos = ALIST_AFTER_LAST;
        insert_where = ALIST_AFTER_CURRENT;
        stop_condition = -1;

    } else {
        logerr_fatal_str(0, "Invalid direction");
    }

    res = alist_iterator_set(l, dir, start_pos, NULL);
    logerr_assert_str(res == ALIST_OK, "Iterator set failed");

    while ((enode = alist_iterator_next(l)) != NULL) {

        cmp = l->cmp_func(enode->data, nnode->data);
        logerr_fatal_str(abs(cmp) <= 1,
                "Compare function returned invalid code");

        if (cmp == stop_condition)
            break;
    }

    if (enode == NULL) {
        /* we went through the whole list but towards which end ? */
        insert_where = (start_pos == ALIST_BEFORE_FIRST) ? ALIST_AFTER_LAST   :
                       (start_pos == ALIST_AFTER_LAST)   ? ALIST_BEFORE_FIRST :
                       logerr_assert_str(0, "Invalid start position");
    }

    res = alist_insert_node_here(l, insert_where, 0, enode, nnode);

    logerr_return_str(res == ALIST_OK,
                "Failed to insert ordered node",
                res);

    return res;
}


anode *
alist_detach_node(alist *l, anode *enode)
{
    logerr_passert(l);
    logerr_passert(enode);
    logerr_passert(enode->prev);
    logerr_passert(enode->next);


    /* update iterator: first pass */

    if (l->current == enode) {
        alist_iterator_prev(l);
    }

    /* update pointers of adjacent nodes */

    enode->prev->next = enode->next;
    enode->next->prev = enode->prev;

    /* update list state */

    l->elements--;

    if (enode == alist_first_node(l)) {
        l->first = (l->elements == 0) ? NULL : enode->next;
    }

    /* update iterator: second pass */

    if (l->current == enode)
        l->current = NULL;

    l->ipos = ALIST_AT_CURRENT;

    if (l->current == NULL) {
        l->ipos = (l->idir == ALIST_FORWARD) ? ALIST_AFTER_LAST :
                                               ALIST_BEFORE_FIRST;

    } else if (l->current == alist_first_node(l)) {
        l->ipos = ALIST_AT_FIRST;

    } else if (l->current == alist_last_node(l)) {
        l->ipos = ALIST_AT_LAST;
    }

    return enode;
}


anode *
alist_detach_node_here(alist *l, int where)
{
    anode *node = NULL;

    logerr_passert(l);

    if (l->elements > 0) {

        if (where == ALIST_AT_FIRST)
            node = alist_first_node(l);

        else if (where == ALIST_AT_LAST)
            node = alist_last_node(l);

        else
            logerr_assert_str(0, "Invalid 'where' flag");

        alist_detach_node(l, node);
    }

    return node;
}

int
alist_remove_destroy_node(alist *l, anode *enode)
{
    int res = 0;

    logerr_passert(l);
    logerr_passert(enode);

    alist_detach_node(l, enode);

    res = anode_destroy(enode, l->free_func);
    logerr_assert_str(res == ANODE_OK, "Failed to destroy node");
    
    return ALIST_OK;
}

int
alist_destroy_all_nodes(alist *l)
{
    anode *tmp = NULL;
    anode *last = NULL;
    int elements_left = 0;

    void cleanup(void) {
        /* leave a proper list */
        if (last != NULL) {
            last->prev = tmp;
            l->first = tmp;
            l->elements = elements_left;
        }
    }

    logerr_passert(l);
    last = alist_last_node(l);
    elements_left = alist_elements(l);

    tmp = alist_first_node(l);
    while (tmp != NULL) {

        anode *next = tmp->next;
        int res = 0;

        res = anode_destroy(tmp, l->free_func);
        logerr_creturn_str(res == ANODE_OK,
                        "Failed to destroy all nodes",
                        ALIST_ERROR, cleanup);
        elements_left--;
        tmp = next;
        if (tmp == alist_first_node(l))
            break;
    }

    logerr_assert_str(elements_left == 0, "Number of elements should be 0");
    l->elements = 0;
    l->first = NULL;

    return ALIST_OK;
}


anode *
alist_search_for_node(alist *l, int dir, int pos, anode *enode,
                cmp_search_func *search_func, void *key)
{
    int found = 0;
    int res = 0;
    anode *node = NULL;

    logerr_passert(l);
    logerr_passert(search_func);

    if (l->elements == 0)
        return NULL;

    res = alist_iterator_set(l, dir, pos, enode);
    logerr_assert_str(res == ALIST_OK, "Iterator set failed");

    while ((node = alist_iterator_next(l)) != NULL) {
        if (search_func(node->data, key) == 0) {
            found = 1;
            break;
        }
    }

    return (found) ? node : NULL;
}


int
alist_foreach_node(alist *l, int dir, int pos, anode *enode,
                cmp_foreach_func *foreach_func, void *arg)
{
    int res = 0;
    anode *node = NULL;

    logerr_passert(l);
    logerr_passert(foreach_func);

    if (l->elements == 0)
        return ALIST_OK;

    res = alist_iterator_set(l, dir, pos, enode);
    logerr_assert_str(res == ALIST_OK, "Iterator set failed");

    while ((node = alist_iterator_next(l)) != NULL) {

        res = foreach_func(arg, node->data);
        logerr_return_str(res >= 0,
                          "Failed foreach_function",
                          res);

        if (res == 0)
            break;
    }

    return (res != -1) ? ALIST_OK : ALIST_ERROR;
}

/* see cmp_foreach_func() requirements */
static int
alist_append_notowner_r(void *lptr, void *data)
{
        int rc = 0;
        alist *l = lptr;

        rc = alist_insert_last_r(l, data, ALIST_NOT_OWNER);
        BUG_ON(rc);

        /* success, keep going */
        return 1;
}

int
alist_copy_data_pointers_r(alist *src, alist *dst)
{
        return alist_foreach_r(src, ALIST_FORWARD, ALIST_BEFORE_FIRST,
                               alist_append_notowner_r, dst);
}

