/*
 * Created on   : Tue Jun 22 09:13:53 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * alist.h
 *
 * Generic list
 *
 * $Id: alist.h,v 1.25 2005/06/28 00:46:11 pavlos Exp $
 */

#ifndef __ALIST_H
#define __ALIST_H

#include <pthread.h>

#include "bytes.h"
#include "anode.h"
#include "cmp.h"

#define ALIST_MAX_ELEMENTS      ((unsigned int)-1)

/* DO NOT CHANGE: must conform to system return values 0, -1 */
#define ALIST_OK                0
#define ALIST_ERROR             -1

#define alist_elements(L)       ((L)->elements)
#define alist_first_node(L)     ((L)->first)
#define alist_last_node(L)      ((L)->first == NULL ? NULL : (L)->first->prev)
#define alist_empty(L)          ((L)->first == NULL && (L)->elements == 0)
#define alist_noncirc_next(L,N) \
        (((N)->next == (L)->first) ? NULL : (N)->next)

#define alist_peek_first_data(L) \
        ((L)->first == NULL ? NULL : (L)->first->data)

#define alist_peek_last_data(L) \
        ((L)->first == NULL ? NULL : (L)->first->prev->data)

#define ALIST_MUTEX             1
#define ALIST_NO_MUTEX          0

#define ALIST_LOCK              1
#define ALIST_NOLOCK            0

#define ALIST_OWNER             1
#define ALIST_NOT_OWNER         0

enum {
    ALIST_FORWARD,
    ALIST_BACKWARD,
    ALIST_FCIRCULAR,
    ALIST_BCIRCULAR
};

enum {
    ALIST_AT_FIRST,
    ALIST_BEFORE_FIRST,
    ALIST_AFTER_FIRST,
    ALIST_AT_LAST,
    ALIST_BEFORE_LAST,
    ALIST_AFTER_LAST,
    ALIST_AT_CURRENT,
    ALIST_BEFORE_CURRENT,
    ALIST_AFTER_CURRENT,
    ALIST_ORDERED
};

typedef struct alist {
    int use_mutex;
    pthread_mutex_t access_mutex;
    unsigned int elements;
    anode *first;
    anode *current;
    bytes_free_func *free_func;
    cmp_cmp_func *cmp_func;
    int idir;
    int ipos;
    rbuf_mgr *rbmgr;
} alist;


/* functions that initialize / destroy list */

#define alist_init(l,use_mutex,free_func) \
        (alist_init_long(l, use_mutex, free_func, NULL, NULL))

int alist_init_long(alist *l,
        int use_mutex,
        bytes_free_func *free_func,
        cmp_cmp_func *cmp_func,
        rbuf_mgr *rbmgr);

int alist_done(alist *l);

/* functions that are reentrant by locking the list mutex */

int alist_elements_r(alist *l);

int alist_insert_here_r(alist *l, void *data, int data_owner,
                int where, int dir, anode *enode);

#define alist_insert_first_r(l,data,data_owner) \
        (alist_insert_here_r(l, data, data_owner, ALIST_BEFORE_FIRST, 0, NULL))

#define alist_insert_last_r(l,data,data_owner) \
        (alist_insert_here_r(l, data, data_owner, ALIST_AFTER_LAST, 0, NULL))

#define alist_insert_ordered_r(l, data, data_owner, dir) \
        (alist_insert_here_r(l, data, data_owner, ALIST_ORDERED, dir, NULL))

#define alist_discard_first(l) \
        (alist_remove_destroy_node((l), alist_first_node(l)))

#define alist_discard_last(l) \
        (alist_remove_destroy_node((l), alist_last_node(l)))

void * alist_remove_here_r(alist *l, int where);
int    alist_remove_all_r(alist *l);

#define alist_remove_first_r(l) \
        (alist_remove_here_r(l, ALIST_AT_FIRST))

#define alist_remove_last_r(l) \
        (alist_remove_here_r(l, ALIST_AT_LAST))

void * alist_search_r(alist *l, int dir, int pos,
                cmp_search_func *search_func, void *key);

int    alist_foreach_r(alist *l, int dir, int pos,
                cmp_foreach_func *foreach_func, void *arg);


/* functions that lock and unlock the list mutex */

void alist_lock(alist *l);
void alist_unlock(alist *l);

/* functions that do NOT lock the list mutex; mutex must be locked manually  */

int     alist_iterator_set(alist *l, int dir, int pos, anode *enode);
anode * alist_iterator_prev(alist *l);
anode * alist_iterator_next(alist *l);

static inline void *
alist_iterator_next_data(alist *l)
{
        anode *node = alist_iterator_next(l);
        if (node == NULL)
                return NULL;
        return node->data;
}

int alist_insert_node_here(alist *l, int where, int dir,
                anode *enode, anode *nnode);
int alist_insert_node_after(alist *l, anode *enode, anode *nnode);
int alist_insert_node_before(alist *l, anode *enode, anode *nnode);
int alist_insert_node_first(alist *l, anode *nnode);
int alist_insert_node_last(alist *l, anode *nnode);
int alist_insert_node_ordered(alist *l, anode *nnode, int dir);

anode * alist_detach_node(alist *l, anode *enode);
anode * alist_detach_node_here(alist *l, int where);

#define alist_detach_first_node(l) \
            (alist_detach_node_here(l, ALIST_AT_FIRST))

#define alist_detach_last_node(l) \
            (alist_detach_node_here(l, ALIST_AT_LAST))

int alist_remove_destroy_node(alist *l, anode *enode);

int alist_destroy_all_nodes(alist *l);

anode * alist_search_for_node(alist *l, int dir, int pos, anode *enode,
                cmp_search_func *search_func, void *key);

int     alist_foreach_node(alist *l, int dir, int pos, anode *enode,
                cmp_foreach_func *foreach_func, void *arg);

int alist_copy_data_pointers_r(alist *src, alist *dst);

#endif /* __ALIST_H */

