/*
 * err.h	Auxiliary utilities for error handling.
 *
 * Created:	Fri Dec  9 10:00:10 EST 2005
 *
 * Author:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 */

#ifndef _ALOG_H
#define _ALOG_H

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>

/* for open, dup, dup2, close */
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


static inline void
__syserr(int cond, int sys_errno, const char *file, int line)
{
	if (!cond)
		return;

	fprintf(stderr,
		"%s:%d: cond %d: syscall errno %d: %s\n",
		file, line, cond, sys_errno,
		strerror(sys_errno));
	exit(1);
}

static inline void
__abort(int cond, const char *abortstr, const char *file, int line)
{
	if (!cond)
		return;

	fprintf(stderr, "%s:%d: %s\n", file, line,
		(abortstr ? abortstr : "aborting.."));
	exit(1);
}


static inline void
__abort_f(int cond, const char *format, ...)
{
	va_list ap;

	if (!cond)
		return;

	va_start(ap, format);
	vfprintf(stderr, format, ap);
	va_end(ap);

	exit(1);
}

#define SYSERR() \
	__syserr(1, errno, __FILE__, __LINE__)

#define SYSERR_ON(cond) \
	__syserr((cond), ((cond) == -1 ? errno : abs(cond)), __FILE__, __LINE__)

#define BUG() \
	__abort(1, "bug", __FILE__, __LINE__);

#define BUG_ON(cond) \
	__abort((cond), "bug_on(" #cond ")", __FILE__, __LINE__);

#define ERR(format, ...) \
	__abort_f(1, "----\n%s:%d (%s):\n" format "\n", __FILE__, __LINE__, __FUNCTION__, ##__VA_ARGS__)

#define ERRFN(format, ...) \
        __abort_f(1, "%s: error: " format "\n", __FUNCTION__, ##__VA_ARGS__)

#define ERR_ON(cond, format, ...) \
	__abort_f((cond), "----\n%s:%d (%s):\n" format "\n", __FILE__, __LINE__, __FUNCTION__, ##__VA_ARGS__)

#define LOG(format, ...) \
        printf(format "\n", ## __VA_ARGS__)

#define LOGTS(format, ...) \
        do { \
                struct timeval __now; \
                int __sec, __usec; \
                gettimeofday(&__now, NULL); \
                __sec = __now.tv_sec; \
                __usec = __now.tv_usec; \
                printf("%d.%.6d: " format "\n", __sec, __usec, ## __VA_ARGS__); \
        } while (0)

#define LOGNR(format, ...) \
        printf(format, ## __VA_ARGS__)

#define LOG_ERR(format, ...) \
        fprintf(stderr, format "\n", ## __VA_ARGS__)

#define LOGTR(format, ...) \
        printf("----\n%s:%d (%s):\n" format "\n", __FILE__, __LINE__, __FUNCTION__, ##__VA_ARGS__)

#define LOGFN(format, ...) \
        printf("%s:%d: " format "\n", __FUNCTION__, __LINE__, ##__VA_ARGS__)

#define LOGERRFN(format, ...) \
        fprintf(stderr, "%s:%d: " format "\n", __FUNCTION__, __LINE__, ##__VA_ARGS__)

#define LOGFN_ERR(format, ...) \
        fprintf(stderr, "%s: " format "\n", __FUNCTION__, ##__VA_ARGS__)

#ifdef DEBUG
#define DEB(format, ...) \
	printf("debug: %s:%d: " format "\n", __FILE__, __LINE__, ##__VA_ARGS__)
#else
#define DEB(format, ...)      ((void)(0))
#endif


static inline int
alog_redirect(int fd, char *filename)
{
	int oldfd = -1;
	int filefd = -1;
	int rc = 0;
	
	filefd = open(filename, O_RDWR|O_CREAT|O_TRUNC, 0600);
	SYSERR_ON(filefd < 0);

	oldfd = dup(fd);

	rc = dup2(filefd, fd);
	SYSERR_ON(rc < 0);

	close(filefd);
	return oldfd;
}

/*
 * Prints out a buffer in hex using LOG()
 *
 * inspired by the kernel source: audit.c:audit_log_hex()
 * silently reduces the bytes per line if too many
 */
static inline void
alog_hex(const void *buf, int buflen,
         int bytes_per_line, const char *prefix)
{
        int i = 0;
        unsigned char line[100];
        unsigned char *ptr = line;
        const unsigned char *cbuf = buf;
        int buffered = 0;
        static const char *hex = "0123456789abcdef";

        // each byte takes up 3 characters
        // silently decrease @bytes_per_line to fit @line
        while (bytes_per_line > 0 && bytes_per_line * 3 >= sizeof(line))
                bytes_per_line--;

        for (i = 0; i < buflen; i++) {
                if (buffered == bytes_per_line) {
                        *ptr = 0;
                        LOG("%s%s", prefix, line);
                        buffered = 0;
                        ptr = line;
                }
                *ptr++ = ' ';
                *ptr++ = hex[(cbuf[i] & 0xF0)>>4]; /* Upper nibble */
                *ptr++ = hex[cbuf[i] & 0x0F];	  /* Lower nibble */
                buffered++;
        }

        if (buffered == bytes_per_line) {
                *ptr = 0;
                LOG("%s%s", prefix, line);
        }
}

#endif
