/*
 * aptr -- a hash-based storage of pointers (thread-safe)
 *
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Fri Dec  7 11:41:46 MST 2007
 *
 */

#ifndef _APTR_H
#define _APTR_H

#include <pthread.h>

#include "ahash.h"
#include "alog.h"

// Generic copy/diff macros

#define APTR_FIELD_COPY(dst, src, field) \
        (dst)->field = (src)->field

#define APTR_FIELD_DIFF(end, begin, field) \
        ((end)->field - (begin)->field)

#define APTR_FIELD_DIFF_STORE(diff, end, begin, field) \
        (diff)->field = (end)->field - (begin)->field

// Pointer with reference counter and mutex

struct aptr {
        int refcount;
        pthread_mutex_t refcount_mutex;
        void *refptr;
};

static inline void
aptr_init(struct aptr *ptr, void *refptr)
{
        memset(ptr, 0, sizeof(struct aptr));
        pthread_mutex_init(&ptr->refcount_mutex, NULL);
        ptr->refptr = refptr;
}

// Returns the @refcount
static inline int
aptr_reserve(struct aptr *ptr)
{
        int refcount = 0;

        /* slow, but do we care? */
        pthread_mutex_lock(&ptr->refcount_mutex);
        ptr->refcount++;
        refcount = ptr->refcount;
        pthread_mutex_unlock(&ptr->refcount_mutex);

        return refcount;
}

// Decrements the reference count of @ptr using mutexes
// If the count reaches 0 then the referenced pointer @refptr
// is freed (using free()) if it is non-NULL.
//
// Returns the @refcount, if 0 then the pointer was freed (if non-NULL)
static inline int
aptr_release(struct aptr *ptr)
{
        int refcount = 0;

        /* slow, but do we care? */
        pthread_mutex_lock(&ptr->refcount_mutex);
        ptr->refcount--;
        refcount = ptr->refcount;
        pthread_mutex_unlock(&ptr->refcount_mutex);

        BUG_ON(refcount < 0);

        if (refcount == 0 && ptr->refptr != NULL) {
                free(ptr->refptr);
                ptr->refptr = NULL;
        }

        return refcount;
}

struct aptr_storage {
        ahash hash;
        pthread_mutex_t mutex;
};

int aptr_storage_init(struct aptr_storage *storage);
void aptr_storage_done(struct aptr_storage *storage);

static inline int
aptr_ptrkey(void *ptr)
{
        /* return the pointer as an integer in a portable way */
        union {
                int key;
                void *ptr;
        } both = { .ptr = ptr };
        
        return abs(both.key);
}

/* returns a non-negative ID (>= 0) that can be used for lookup */
/* returns -1 on error */
static inline int
aptr_storage_store(struct aptr_storage *storage, void *ptr)
{
        int rc = 0;
        int key = -1;
        int ptrkey = aptr_ptrkey(ptr);

        /* do not accept NULL pointers; complicates lookup */
        BUG_ON(ptr == NULL);

        pthread_mutex_lock(&storage->mutex);

        rc = ahash_insert(&storage->hash, ptr, &ptrkey);
        if (rc == 0)
                key = ptrkey;

        pthread_mutex_unlock(&storage->mutex);
        return key;
}

/* returns the pointer that corresponds to the ID/key */
/* returns NULL if not found */
static inline void *
aptr_storage_lookup(struct aptr_storage *storage, int key)
{
        void *ptr = NULL;

        pthread_mutex_lock(&storage->mutex);
        ptr = ahash_lookup(&storage->hash, &key);
        pthread_mutex_unlock(&storage->mutex);

        return ptr;
}

/* returns 0 on success */
static inline int
aptr_storage_remove(struct aptr_storage *storage, int key)
{
        int rc = 0;

        pthread_mutex_lock(&storage->mutex);
        rc = ahash_delete(&storage->hash, &key);
        pthread_mutex_unlock(&storage->mutex);

        return rc;
}

#endif
