/*
 * Created on   : Fri Jun 25 13:03:45 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * aqueue.c
 *
 * Generic queue
 *
 * $Id: aqueue.c,v 1.19 2005/06/17 23:51:29 pavlos Exp $
 */

#include "logerr.h"
#include "anode.h"
#include "aqueue.h"

int
aqueue_init(aqueue *q, int mode,
        int bounded, int recycle, int prealloc, size_t size,
        bytes_free_func free_func,
        cmp_cmp_func *cmp_func,
        rbuf_mgr *rbmgr)
{
    int res = 0;
    int stage = 0;

    void cleanup()
    {
        if (stage > 0) {
            res = pthread_mutex_destroy(&q->access_mutex);
            logerr_assert_str(res == 0, "Failed to clean up");
        }

        if (stage > 1) {
            res = alist_done(&q->available);
            logerr_assert_str(res == ALIST_OK, "Failed to clean up");
        }

        if (stage > 2) {
            res = alist_done(&q->occupied);
            logerr_assert_str(res == ALIST_OK, "Failed to clean up");
        }
    }

    logerr_passert(q);

    memset(q, 0, sizeof(aqueue));

    q->mode = mode;
    q->rbmgr = rbmgr;

    if (bit_test(q->mode, AQUEUE_COPY_DATA)) {
        q->data_size = size;
        free_func = &free;

    } else {
        q->data_size = 0;
    }

    if (bit_test(q->mode, AQUEUE_SORTED)) {
        logerr_creturn_str(cmp_func != NULL,
                    "NULL compare function",
                    AQUEUE_ERROR, cleanup);

    } else {
        cmp_func = NULL;
    }

    if (bit_test(q->mode, AQUEUE_LOCK)) {
        res = pthread_mutex_init(&q->access_mutex, NULL);
        if (res != 0) {
            logerr_str(res == 0, strerror(res));
            logerr_creturn_str(res == 0,
                        "Failed to initialize queue mutex",
                        AQUEUE_ERROR, cleanup);
        }
    }

    /* Stage 1: mutex initialized */
    stage = 1;

    res = alist_init(&q->available, ALIST_NOLOCK, free_func);
    logerr_creturn_str(res == ALIST_OK,
                "Failed to initialize available list",
                AQUEUE_ERROR, cleanup);

    q->max_available = 0;

    /* Stage 2: mutex and available list initialized */
    stage = 2;

    res = alist_init_long(&q->occupied, ALIST_NOLOCK, free_func, cmp_func, NULL);
    logerr_creturn_str(res == ALIST_OK,
                "Failed to initialize occupied list",
                AQUEUE_ERROR, cleanup);

    q->max_occupied = AQUEUE_NOLIMIT;

    /* Stage 3: mutex, available and occupied lists initialized */
    stage = 3;

    if (bit_test(q->mode, AQUEUE_BOUNDED) && bounded >= 0) {
        q->max_occupied = bounded;
    }

    if (bit_test(q->mode, AQUEUE_RECYCLE) && recycle >= 0) {
        q->max_available = recycle;
    }

    if (bit_test(q->mode, AQUEUE_PREALLOC) && prealloc >= 0) {

        int i = 0;
        void *data = NULL;
        int data_owner = 0;

        /* insert into the available list the free elements */
        for (i = 0; i < prealloc; i++) {

            if (bit_test(q->mode, AQUEUE_COPY_DATA)) {

                logerr_creturn_str(q->data_size > 0,
                            "Invalid data size",
                            AQUEUE_ERROR, cleanup);

                data = malloc(q->data_size);
                logerr_creturn_str(data != NULL,
                            "Failed to allocate data memory",
                            AQUEUE_ERROR, cleanup);

                data_owner = 1;
            }

            res = alist_insert_last_r(&q->available, data, data_owner);
            logerr_creturn_str(res == ALIST_OK,
                            "Failed to preallocate elements",
                            AQUEUE_ERROR, cleanup);
        }
    }

    return AQUEUE_OK;
}


int
aqueue_done(aqueue *q)
{
    int res = 0;
    int stage = 0;
    int locked = 0;

    void cleanup(void)
    {
        if (stage > 0) {
            res = alist_init(&q->available, 0, NULL);
            logerr_assert_str(res == ALIST_OK, "Failed to clean up");
        }

        if (stage > 1) {
            res = alist_init(&q->occupied, 0, NULL);
            logerr_assert_str(res == ALIST_OK, "Failed to clean up");
        }

        if (locked)
            aqueue_unlock(q);
    }

    logerr_passert(q);

    aqueue_lock(q);
    locked = 1;

    res = alist_destroy_all_nodes(&q->available);
    logerr_creturn_str(res == ALIST_OK,
                "Failed to remove all nodes from available list",
                ALIST_ERROR, cleanup);

    res = alist_destroy_all_nodes(&q->occupied);
    logerr_creturn_str(res == ALIST_OK,
                "Failed to remove all nodes from occupied list",
                ALIST_ERROR, cleanup);

    res = alist_done(&q->available);
    logerr_creturn_str(res == ALIST_OK,
                "Failed to destroy available list",
                ALIST_ERROR, cleanup);

    stage = 1;

    res = alist_done(&q->occupied);
    logerr_creturn_str(res == ALIST_OK,
                "Failed to destroy occupied list",
                ALIST_ERROR, cleanup);

    stage = 2;

    aqueue_unlock(q);
    locked = 0;

    if (bit_test(q->mode, AQUEUE_LOCK)) {
        res = pthread_mutex_destroy(&q->access_mutex);
        if (res != 0) {
            logerr_str(res == 0, strerror(res));
            logerr_creturn_str(res == 0,
                        "Failed to destroy queue mutex",
                        ALIST_ERROR, cleanup);
        }
    }

    return AQUEUE_OK;
}


void
aqueue_lock(aqueue *q)
{
    int res = 0;

    logerr_passert(q);

    if (!bit_test(q->mode, AQUEUE_LOCK))
        return;

    res = pthread_mutex_lock(&q->access_mutex);
    logerr_fatal_str(res == 0, strerror(res));
}

void
aqueue_unlock(aqueue *q)
{
    int res = 0;

    logerr_passert(q);

    if (!bit_test(q->mode, AQUEUE_LOCK))
        return;

    res = pthread_mutex_unlock(&q->access_mutex);
    logerr_fatal_str(res == 0, strerror(res));
}


int
aqueue_push_here(aqueue *q, void *dptr, int downer, int where)
{
    int res = 0;

    struct iovec vector;

    vector.iov_base = dptr;
    vector.iov_len = (bit_test(q->mode, AQUEUE_COPY_DATA)) ?
                q->data_size : (size_t)-1;

    res = aqueue_pushv_here(q, &vector, 1, downer, where);
    return res;
}


int
aqueue_pushv_here(aqueue *q, const struct iovec *vector, int count,
                int downer, int where)
{
    int res = 0;
    int locked = 0;
    int can_push = 0;
    int recycle_node = 0;
    void *data = NULL;
    int data_owner = 0;
    anode *node = 0;

    void cleanup(void)
    {
        if (locked)
            aqueue_unlock(q);

        if (data != NULL && data_owner) {
            free(data);
        }

        if (node != NULL) {
            node->data = NULL;
            node->data_owner = 0;
            res = anode_destroy(node, NULL);
            logerr_assert_str(res == ANODE_OK, "Failed to clean up");
        }
    }

    logerr_passert(q);
    logerr_passert(vector);
    logerr_assert_str(count > 0, "Invalid buffer vector count");

    aqueue_lock(q);
    locked = 1;

    can_push     = ( alist_elements(&q->occupied) < q->max_occupied );
    recycle_node = ( alist_elements(&q->available) > 0 );

    if (!can_push) {
        cleanup();
        /* XXX how should this error be handled? */
        return AQUEUE_FULL;
    }

    if (recycle_node) {
        node = alist_detach_first_node(&q->available);

    } else {

        if (bit_test(q->mode, AQUEUE_COPY_DATA)) {

            logerr_bug(q->data_size > 0);

            data = malloc(q->data_size);
            logerr_creturn_str(data != NULL,
                        "Failed to allocate data memory",
                        AQUEUE_ERROR, cleanup);

            data_owner = 1;
        }

        node = anode_create(data, data_owner, q->rbmgr);
        logerr_creturn_str(node != NULL,
                    "Failed to create node",
                    AQUEUE_ERROR, cleanup);

    }

    logerr_passert(node);

    if (bit_test(q->mode, AQUEUE_COPY_DATA)) {

        /* vector holds pointers to data that needs to be copied */

        logerr_passert(node->data);

        res = mem_copy_from_vector(node->data, q->data_size, vector, count);

        logerr_creturn_str(res >= 0,
                    "Failed to copy data completely",
                    AQUEUE_ERROR, cleanup);

    } else {

        logerr_bug(count == 1);

        /* this is a pointer to the data */
        node->data = vector[0].iov_base;
        node->data_owner = downer;
    }

    if (bit_test(q->mode, AQUEUE_SORTED))
        res = alist_insert_node_ordered(&q->occupied, node, ALIST_FORWARD);

    else if (where == AQUEUE_FRONT)
        res = alist_insert_node_first(&q->occupied, node);

    else if (where == AQUEUE_BACK)
        res = alist_insert_node_last(&q->occupied, node);

    else
        logerr_assert_str(0, "Invalid 'where' flag");

    logerr_creturn_str(res == ALIST_OK,
                "Failed to insert node in occupied list",
                AQUEUE_ERROR, cleanup);

    aqueue_unlock(q);
    locked = 0;
    
    return AQUEUE_OK;
}

void *
aqueue_peek_here(aqueue *q, int where)
{
    anode *node = NULL;
    void *data = NULL;
    int locked = 0;

    void cleanup(void)
    {
        if (locked)
            aqueue_unlock(q);

    }

    aqueue_lock(q);
    locked = 1;

    if (where == AQUEUE_FRONT)
        node = alist_first_node(&q->occupied);

    else if (where == AQUEUE_BACK)
        node = alist_last_node(&q->occupied);

    else
        logerr_assert_str(0, "Invalid 'where' flag");

    if (node != NULL)
        data = node->data;

    aqueue_unlock(q);
    locked = 0;

    /* XXX note that this has the problem that data pointer may become
     *     invalid, since queue is not locked any more -- needs fix
     */
    return data;
}


int
aqueue_pop_here(aqueue *q, void *ptr, int where)
{
    int res = 0;

    struct iovec vector;

    vector.iov_base = ptr;
    vector.iov_len = (bit_test(q->mode, AQUEUE_COPY_DATA)) ?
                (size_t)-1 : sizeof(void *);

    res = aqueue_popv_here(q, &vector, 1, where);
    return res;
}


int
aqueue_popv_here(aqueue *q, const struct iovec *vector, int count, int where)
{
    int res = 0;
    int locked = 0;
    anode *node = 0;
    int recycle_node = 0;

    void cleanup(void)
    {
        if (locked)
            aqueue_unlock(q);

        if (node != NULL) {
            res = anode_destroy(node, NULL);
            logerr_assert_str(res == ANODE_OK, "Failed to clean up");
        }
    }

    logerr_passert(q);
    logerr_passert(vector);
    logerr_assert_str(count > 0, "Invalid buffer vector count");

    aqueue_lock(q);
    locked = 1;

    if (where == AQUEUE_FRONT)
        node = alist_detach_first_node(&q->occupied);

    else if (where == AQUEUE_BACK)
        node = alist_detach_last_node(&q->occupied);

    else
        logerr_assert_str(0, "Invalid 'where' flag");

    if (node == NULL) {
        cleanup();
        return AQUEUE_EMPTY;
    }

    if (bit_test(q->mode, AQUEUE_COPY_DATA)) {

        /* this is a pointer to memory that holds the data */
        res = mem_copy_to_vector(node->data, q->data_size, vector, count);

        logerr_creturn_str(res == q->data_size,
                "Failed to copy data completely",
                AQUEUE_ERROR, cleanup);

    } else {

        /* this is a pointer to a pointer to the data */
        void **ptrptr = vector[0].iov_base;

        logerr_assert_str(vector[0].iov_len >= sizeof(void *),
                "Not enough space to copy pointer to data");

        *ptrptr = node->data;

        node->data = NULL;
        node->data_owner = 0;
    }

    recycle_node = ( alist_elements(&q->available) < q->max_available );

    if (recycle_node) {

        res = alist_insert_node_last(&q->available, node);
        logerr_creturn_str(res == ALIST_OK,
                    "Failed to insert node in available list",
                    AQUEUE_ERROR, cleanup);

    } else {

        res = anode_destroy(node, &free);
        node = NULL;
        logerr_creturn_str(res == ANODE_OK,
                    "Failed to destroy node",
                    AQUEUE_ERROR, cleanup);
    }

    aqueue_unlock(q);
    locked = 0;

    return AQUEUE_OK;
}


