/*
 * Created on   : Fri Jun 25 13:03:45 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * aqueue.h
 *
 * Generic queue
 *
 * $Id: aqueue.h,v 1.15 2005/06/17 23:51:29 pavlos Exp $
 */

/*
 * TODO
 *      + add condition variable so that pop/push can block
 */

#ifndef __AQUEUE_H
#define __AQUEUE_H

#include <pthread.h>

#include "bytes.h"
#include "alist.h"
#include "mem.h"
#include "rbufx.h"

#define AQUEUE_EMPTY             3
#define AQUEUE_FULL              2
#define AQUEUE_OK                0
#define AQUEUE_ERROR            -1

#define AQUEUE_NOLIMIT           ((unsigned int)-1)

#define AQUEUE_NONE             0x0000
#define AQUEUE_BOUNDED          0x0001
#define AQUEUE_RECYCLE          0x0002
#define AQUEUE_PREALLOC         0x0004
#define AQUEUE_SORTED           0x0008
#define AQUEUE_LOCK             0x0010
#define AQUEUE_COPY_DATA        0x0020

#define AQUEUE_OWNER            1
#define AQUEUE_NOT_OWNER        0

#define aqueue_elements(q)      (alist_elements(&(q)->occupied))

enum {
    AQUEUE_FRONT,
    AQUEUE_BACK
};

typedef struct aqueue {
    int mode;
    size_t data_size;
    pthread_mutex_t access_mutex;
    alist available;
    alist occupied;
    unsigned int max_occupied;
    unsigned int max_available;
    rbuf_mgr *rbmgr;
} aqueue;


/*
 * mode == AQUEUE_BOUNDED
 *     Create a queue with bounded number of elements
 *     Variable 'bounded' indicates the maximum number of queue elements.
 *     Set to -1 for infinite.
 *
 * mode == AQUEUE_RECYCLE
 *      Recycle free nodes. During a pop operation do not throw
 *      away memory for the freed node; keep and reuse.
 *      Variable 'recycle' indicates the maximum number of free nodes
 *      to keep allocated at any given time.
 *      Set to -1 for infinite
 *
 * mode == AQUEUE_PREALLOC
 *      Preallocate the free node structures [not the data ones!]
 *      Variable 'prealloc' indicates the number of free nodes to preallocate
 *
 * mode == AQUEUE_SORTED
 *      Use a sorted list. cmp_func must be a valid compare function.
 *      (see anode.h comments)
 *
 * mode == AQUEUE_LOCK
 *      Lock the queue whenever it is accessed/modifed.
 *      This provides reentrant code.
 *
 * mode == AQUEUE_COPY_DATA
 *      There are two modes for the queue:
 *
 *          1. Push a pointer to data           (no flag)
 *          2. Push the contents of the data    (AQUEUE_COPY_DATA)
 *
 *      In (1) no memory is ever allocated to hold the user data.
 *      It is assumed that the user allocates the data and passes
 *      the pointer to the queue during a push operations. This is
 *      the same pointer that is returned by the pop_ptr operation.
 *      Optionally, if a free_func is specified, then the user
 *      data will be freed on queue destruction.
 *
 *      In (2) the queue allocates and owns the memory to hold the user
 *      data. In a push operation the user passes a pointer whose memory
 *      contents are copied into the queue node.  During a pop_data
 *      operation, the user must provide a pointer into which the data
 *      is copied.  In this case, the user is required to provide the
 *      size of data during init, therefore allowing only fixed data
 *      elements.  free_func is ignored in this case.
 *
 */    
int aqueue_init(aqueue *q, int mode,
                int bounded, int recycle, int prealloc, size_t size,
                bytes_free_func free_func,
                cmp_cmp_func *cmp_func,
                rbuf_mgr *rbmgr);

#define aqueue_data_init(q, lock, data_size ) \
        (aqueue_init((q), \
                ((lock) ? AQUEUE_LOCK|AQUEUE_COPY_DATA : AQUEUE_COPY_DATA), \
                -1, -1, -1, (data_size), NULL, NULL, NULL))

#define aqueue_ptr_init(q, lock, free_func) \
        (aqueue_init((q), \
                ((lock) ? AQUEUE_LOCK : 0), \
                -1, -1, -1, 0, (free_func), NULL, NULL))

int aqueue_done(aqueue *q);


int aqueue_push_here(aqueue *q, void *dptr, int downer, int where);
int aqueue_pushv_here(aqueue *q, const struct iovec *vector, int count,
                int downer, int where);

#define aqueue_push_ptr_front(q, dptr, downer) \
        (aqueue_push_here((q), (dptr), (downer), AQUEUE_FRONT))

#define aqueue_push_ptr_back(q, dptr, downer) \
        (aqueue_push_here((q), (dptr), (downer), AQUEUE_BACK))

#define aqueue_push_data_front(q, dptr) \
        (aqueue_push_here((q), (dptr), 0, AQUEUE_FRONT))

#define aqueue_push_data_back(q, dptr) \
        (aqueue_push_here((q), (dptr), 0, AQUEUE_BACK))

#define aqueue_pushv_data_front(q, vector, count) \
        (aqueue_pushv_here((q), (vector), (count), 0, AQUEUE_FRONT))

#define aqueue_pushv_data_back(q, vector, count) \
        (aqueue_pushv_here((q), (vector), (count), 0, AQUEUE_BACK))



void * aqueue_peek_here(aqueue *q, int where);
int aqueue_pop_here(aqueue *q, void *ptr, int where);
int aqueue_popv_here(aqueue *q, const struct iovec *vector, int count,
                int where);

#define aqueue_pop_ptr_front(q, ptrptr) \
        (aqueue_pop_here((q), (ptrptr), AQUEUE_FRONT))

#define aqueue_pop_ptr_back(q, ptrptr) \
        (aqueue_pop_here((q), (ptrptr), AQUEUE_BACK))

#define aqueue_pop_data_front(q, dptr) \
        (aqueue_pop_here((q), (dptr), AQUEUE_FRONT))

#define aqueue_pop_data_back(q, dptr) \
        (aqueue_pop_here((q), (dptr), AQUEUE_BACK))

#define aqueue_popv_data_front(q, vector, count) \
        (aqueue_popv_here((q), (vector), (count), AQUEUE_FRONT))

#define aqueue_popv_data_back(q, vector, count) \
        (aqueue_popv_here((q), (vector), (count), AQUEUE_BACK))

#define aqueue_peek_front(q) \
        (aqueue_peek_here((q), AQUEUE_FRONT))

#define aqueue_peek_back(q) \
        (aqueue_peek_here((q), AQUEUE_BACK))

void aqueue_lock(aqueue *q);
void aqueue_unlock(aqueue *q);


#endif /* __AQUEUE_H */

