/*
 * Created on   : Sun Jul 18 18:43:57 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * asignal.c
 *
 * Signal related routines.
 * 
 * $Id: asignal.c,v 1.12 2004/08/06 19:34:34 pavlos Exp $
 */

#include "asignal.h"

int
asignal_set_handler(int signo, int flags,
                asignal_handler_func *nfunc,
                asignal_handler_func **ofunc_p)
{
    struct sigaction naction;
    struct sigaction oaction;

    naction.sa_handler = nfunc;
    sigemptyset(&naction.sa_mask);
    naction.sa_flags = flags;

    logerr_return_rsys(
        sigaction(signo, &naction, &oaction) != -1,
        "Failed to install new signal handler",
        ASIGNAL_EHANDLER);

    if (ofunc_p != NULL)
        *ofunc_p = oaction.sa_handler;

    return ASIGNAL_OK;
}

int
asignal_select(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
                struct timeval *timeout, const sigset_t *unblock_sigmask)
{
    int res = 0;
    int sel = 0;

    res = sigprocmask(SIG_UNBLOCK, unblock_sigmask, NULL);
    logerr_assert_sys(res != -1, "Failed to unblock signals in sigmask");

    logerr_log(999, "asignal_select: unblocked alarm: %s",
        (sigismember(unblock_sigmask, SIGALRM) ? "yes" : "no"));


    sel = select(n, readfds, writefds, exceptfds, timeout);

    res = sigprocmask(SIG_BLOCK, unblock_sigmask, NULL);
    logerr_assert_sys(res != -1, "Failed to block signals in sigmask");

    logerr_log(999, "asignal_select: blocked alarm: %s",
        (sigismember(unblock_sigmask, SIGALRM) ? "yes" : "no"));

    return sel;
}

int
asignal_block(int signum, int *was_blocked)
{
    int res = 0;
    int was_blocked_value = 0;
    sigset_t new_sigmask;
    sigset_t old_sigmask;

    res = sigemptyset(&new_sigmask);
    logerr_assert_sys(res != -1, "Failed to empty new sigmask");

    res = sigaddset(&new_sigmask, signum);
    logerr_assert_sys(res != -1, "Failed to set signal in new sigmask");

    res = sigprocmask(SIG_BLOCK, &new_sigmask, &old_sigmask);
    logerr_assert_sys(res != -1, "Failed to block signal");

    was_blocked_value = sigismember(&old_sigmask, signum);
    logerr_assert_sys(res != -1, "Failed to query signal set");

    if (was_blocked != NULL)
        *was_blocked = was_blocked_value;

    logerr_log(999, "asignal_block: blocked signal (%d) was_blocked(%d)",
                signum, was_blocked_value);

    return ASIGNAL_OK;
}


int
asignal_unblock(int signum, int *was_blocked)
{
    int res = 0;
    int was_blocked_value = 0;
    sigset_t new_sigmask;
    sigset_t old_sigmask;

    res = sigemptyset(&new_sigmask);
    logerr_assert_sys(res != -1, "Failed to empty new sigmask");

    res = sigaddset(&new_sigmask, signum);
    logerr_assert_sys(res != -1, "Failed to set signal in new sigmask");

    res = sigprocmask(SIG_UNBLOCK, &new_sigmask, &old_sigmask);
    logerr_assert_sys(res != -1, "Failed to unblock signal");

    was_blocked_value = sigismember(&old_sigmask, signum);
    logerr_assert_sys(res != -1, "Failed to query signal set");

    if (was_blocked != NULL)
        *was_blocked = was_blocked_value;

    logerr_log(999, "asignal_unblock: unblocked signal (%d) was_blocked(%d)",
                signum, was_blocked_value);

    return ASIGNAL_OK;
}


