/*
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Wed Jan 23 07:31:12 MST 2008
 * 
 * astring.h
 *
 * String helper functions
 */

#ifndef __ASTRING_H
#define __ASTRING_H

#include <string.h>
#include <string.h>
#include <stdarg.h>

#define ASTRING_TOKENS_MAX              20
#define ASTRING_TOKEN_STRING_MAX        200

// astring

struct astring {
        int maxlen;
        int len;
        char *cstr;
};

static inline int
astring_init(struct astring *str, int maxbuflen)
{
        memset(str, 0, sizeof(*str));
        str->cstr = malloc(maxbuflen);
        if (str->cstr == NULL)
                return -1;
        memset(str->cstr, 0, maxbuflen);
        // to ensure that there is always \0 at the end
        str->maxlen = maxbuflen-1;
        return 0;
}

static inline void
astring_done(struct astring *str)
{
        if (str->cstr != NULL)
                free(str->cstr);
        memset(str, 0, sizeof(*str));
}

static inline int
astring_append(struct astring *str, const char *format, ...)
{
        va_list ap;
        int written = 0;
        char *availstr = str->cstr + str->len;
        int availbuflen = (str->maxlen - str->len) + 1;

	va_start(ap, format);
	written = vsnprintf(availstr, availbuflen, format, ap);
	va_end(ap);

        if (written >= availbuflen) {
                str->len = str->maxlen;
                str->cstr[str->maxlen] = '\0';
                return -1;
        }

        str->len += written;
        return 0;
}

// astring_tokens

struct astring_tokens {
        int num;
        char *token[ASTRING_TOKENS_MAX];
        char str[ASTRING_TOKEN_STRING_MAX];
};

static inline void
astring_tokens_init(struct astring_tokens *tokens,
                    const char *token_str,
                    const char *token_delim)
{
        int i = 0;
        int was_delimiter = 1;
        int token_strlen = strlen(token_str);

        BUG_ON(strlen(token_str) >= ASTRING_TOKEN_STRING_MAX);
        BUG_ON(strlen(token_delim) == 0);

        memset(tokens, 0, sizeof(struct astring_tokens));
        strncpy(tokens->str, token_str, ASTRING_TOKEN_STRING_MAX);

        for (i = 0; i < token_strlen; i++) {
                int is_delimiter = (strchr(token_delim, tokens->str[i]) != NULL);

                if (is_delimiter)
                        tokens->str[i] = '\0';

                if (was_delimiter && !is_delimiter) {
                        BUG_ON(tokens->num >= ASTRING_TOKENS_MAX);
                        tokens->token[tokens->num] = &tokens->str[i];
                        tokens->num++;
                }

                was_delimiter = is_delimiter;
        }
}

static inline char *
astring_strncpy(char *dst, const char *src, int maxcopy)
{
        strncpy(dst, src, maxcopy);
        dst[maxcopy-1] = 0;
        return dst;
}

// Extract contiguous substring that fits in the asterisk(*) in
// the string pattern.
//
// Example:     @src            "foo.testing1234.png"
//              @pattern        "foo.*.png"
//         -->  @infix          "testing1234"
//
// Returns 0 on success, -1 on error (infix not found, bad pattern)
static inline int
astring_extract_infix(char *infix, int len, const char *src, const char *pattern)
{
        int i = 0;
        int infixlen = 0;
        char *prefix = NULL;
        char *suffix = NULL;
        char *asterisk = NULL;
        char pattern_copy[100];
        int src_suffix_offset = 0;
        
        if (len <= 0)
                return -1;
        
        BUG_ON(strlen(pattern) >= sizeof(pattern_copy));
        strcpy(pattern_copy, pattern);

        asterisk = strchr(pattern_copy, '*');
        if (asterisk == NULL)
                return -1;

        // break up the pattern
        *asterisk = '\0';
        prefix = pattern_copy;
        suffix = asterisk + 1;

        // do we have enough characters?
        infixlen = strlen(src) - (strlen(prefix) + strlen(suffix));
        // prefix and suffix overlap
        if (infixlen < 0)
                return -1;
        // infix string does not have enough space
        if (infixlen >= len)
                return -1;

        // match prefix
        for (i = 0; i < strlen(prefix); i++) {
                if (src[i] != prefix[i])
                        return -1;
        }

        // match suffix
        src_suffix_offset = strlen(src) - strlen(suffix);
        for (i = 0; i < strlen(suffix); i++) {
                if (src[src_suffix_offset + i] != suffix[i])
                        return -1;
        }

        // success: prefix and suffix match and we have enough space
        if (infixlen > 0) {
                const char *src_infix = src + strlen(prefix);
                strncpy(infix, src_infix, infixlen);
        }
        infix[infixlen] = '\0';
        return 0;
} 
#endif
