/*
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Fri Nov 16 11:48:10 EST 2007
 * 
 * asys.h
 *
 * System related functions.
 */

#ifndef __ASYS_H
#define __ASYS_H

#include <stdlib.h>
#include <stdarg.h>
#include <pthread.h>

#include <sys/types.h>
#include <sys/wait.h>

#include "anet.h"
#include "alog.h"
#include "atimer.h"

// Check to see if we got a SIGINT or SIGQUIT during a system() call
// (system() ignores SIGINT and SIGQUIT). See system(3) manpage.
static inline int
asys_system_got_sigint_sigquit(int rc)
{
        int got_signal = 0;

        if (WIFSIGNALED(rc) &&
            (WTERMSIG(rc) == SIGINT || WTERMSIG(rc) == SIGQUIT))
                got_signal = 1;
        
        return got_signal;
}

/* Use exactly like system(3) but with printf type of arguments */
static inline int
asys_system(const char *format, ...)
{
    int rc = 0;
    int count = 0;
    va_list ap;
    char cmd[200];

    va_start(ap, format);
    count = vsnprintf(cmd, sizeof(cmd), format, ap);
    BUG_ON(count >= sizeof(cmd));
    rc = system(cmd);
    va_end(ap);

    if (rc != -1 && rc != 0) {
        int child_status = WEXITSTATUS(rc);
        LOG("command exited with non-zero status: %d", child_status);
        LOG("cmd: <%s>", cmd);
        rc = -1;
    }

    return rc;
}

#define ASYS_ARGV_MAX   30

struct asys_args {
        int argc_max;
        int argv_bytes;

        int argc;
        char **argv;
};

static inline int
asys_args_count(const char *args_str)
{
        int pos = 0;
        int argc = 0;
        int in_arg = 0;

        while (args_str[pos] != 0) {
                int is_whitespace = (args_str[pos] == ' ');
                int transition = (!is_whitespace && !in_arg) ||
                                 (is_whitespace && in_arg);

                if (transition) {
                        in_arg = ~in_arg;
                        // count the arguments on the transition from
                        // whitespace to non-whitespace
                        if (in_arg)
                                argc += 1;
                }

                pos++;
                // sanity check
                BUG_ON(pos > 9999);
        }
        return argc;
}

static inline void
asys_args_append_option_partial(struct asys_args *args,
                                const char *token_start,
                                int token_len)
{
        char *token_dst = NULL;

        // the last entry is reserved for the NULL terminator
        BUG_ON(args->argc >= args->argc_max - 1);

        // added +1 for null byte
        token_dst = malloc(token_len + 1);

        args->argv[args->argc] = token_dst;
        args->argc += 1;

        // @token_start does not have null byte at the end of the token
        // use @token_len to stop
        strncpy(token_dst, token_start, token_len);
        // no null byte copied; fix it now
        token_dst[token_len] = 0;
}

static inline void
asys_args_append_options(struct asys_args *args, const char *options)
{
        int pos = 0;
        int last_pos = 0;
        int in_arg = 0;

        while (options[pos] != 0) {
                int is_whitespace = (options[pos] == ' ');
                int transition = (!is_whitespace && !in_arg) ||
                                 (is_whitespace && in_arg);

                if (transition) {
                        in_arg = ~in_arg;
                        // add an option after a transition from
                        // non-whitespace to whitespace
                        if (!in_arg) {
                                int token_len = pos - last_pos;
                                const char *token_start = options + last_pos;

                                asys_args_append_option_partial(
                                        args, token_start, token_len);
                        }
                        last_pos = pos;
                }

                pos++;
                // sanity check
                BUG_ON(pos > 9999);
        }

        // add the last option; no whitespace at the end of the string
        // to trigger transition inside the while loop
        if (in_arg) {
                int token_len = pos - last_pos;
                const char *token_start = options + last_pos;
                asys_args_append_option_partial(args, token_start, token_len);

                in_arg = ~in_arg;
                last_pos = pos;
                pos++;
        }
}

static inline void
asys_args_init(struct asys_args *args, int max_options, const char *path)
{

        memset(args, 0, sizeof(struct asys_args));

        // add one for the path at argv[0]
        // add one for the NULL terminator
        args->argc_max = max_options + 2;
        args->argv_bytes = args->argc_max * sizeof(char *);

        args->argv = malloc(args->argv_bytes);
        memset(args->argv, 0, args->argv_bytes);

        args->argv[0] = malloc(strlen(path) + 1);
        strcpy(args->argv[0], path);
        args->argc = 1;

}

static inline void
asys_args_done(struct asys_args *args)
{
        int i = 0;

        for (i = 0; i < args->argc_max; i++)
                if (args->argv[i] != NULL)
                        free(args->argv[i]);
        if (args->argv != NULL)
                free(args->argv);
        memset(args, 0, sizeof(args));
}

static inline void
asys_args_log(struct asys_args *args)
{
        int i = 0;

        LOG("args: argc_max %d, argv_bytes %d, argc %d, argv %p",
            args->argc_max,
            args->argv_bytes,
            args->argc,
            args->argv);

        for (i = 0; i < args->argc_max; i++) {
                LOG("args: argv[%d]: <%s> %p", i, args->argv[i], args->argv[i]);
                // print only the first null terminator
                if (args->argv[i] == NULL)
                        break;
        }
}

struct asys_status {
        int value;
        pthread_cond_t cond;
        pthread_mutex_t mutex;
};

static inline void
asys_status_init(struct asys_status *status)
{
        memset(status, 0, sizeof(struct asys_status));
        pthread_cond_init(&status->cond, NULL);
        pthread_mutex_init(&status->mutex, NULL);
}

static inline void
asys_status_signal_change(struct asys_status *status,
                          int new_status)
{
        pthread_mutex_lock(&status->mutex);
        status->value = new_status;
        pthread_cond_signal(&status->cond);
        pthread_mutex_unlock(&status->mutex);
}

// Waits for a conditional variable to change status.
//
// Returns 0 if status changed (normal return)
// Returns -ETIMEDOUT on timeout
// Returns negative error code on timeout (see pthread_cond_timedwait manpage)
int
asys_status_wait_for_change(struct asys_status *status,
                            int timeout_msec,
                            int current_status);

#endif
