/*
 * Created on   : Sun Jul 18 19:05:43 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * atimer.c
 *
 * Provides per-process timers
 * 
 * $Id: atimer.c,v 1.20 2005/06/17 23:51:29 pavlos Exp $
 */

#include "atimer.h"
#include "mem.h"

static struct {

    int alarm_count;

    int pending;
    struct timeval pending_alarm;

    int recurring;
    struct timeval recurring_alarm;

} atimer_globals = {0, 0, {0, 0}};

static void atimer_alarm_handler(int signo);

int
atimer_init(atimer_group *tgroup, size_t data_size,
        size_t recycle_hint, rbuf_mgr *rbmgr)
{
    int res = 0;

    logerr_assert_str(data_size > 0, "Invalid data size");
    logerr_passert(tgroup);

    memset(tgroup, 0, sizeof(atimer_group));

    tgroup->data_size = data_size;
    tgroup->orig_alarm_handler = NULL;
    tgroup->rbmgr = rbmgr;

    /* create the queue to hold the timer entries */
    res = aqueue_init(&tgroup->timers,
                AQUEUE_COPY_DATA |
                AQUEUE_RECYCLE |
                AQUEUE_SORTED,
                -1,
                (recycle_hint > 0 ? recycle_hint : ATIMER_DEF_RECYCLE),
                -1,
                sizeof(atimer_header) + tgroup->data_size,
                NULL,
                &atimer_timeval_cmp,
                tgroup->rbmgr);

    logerr_return_str(res == AQUEUE_OK,
                "Failed to initialize timer queue",
                ATIMER_ERROR);

    /* set the SIGALRM handler */
    res = asignal_set_handler(SIGALRM, SA_RESTART,
                &atimer_alarm_handler, &tgroup->orig_alarm_handler);

    logerr_return_str(res == ASIGNAL_OK,
                "Failed to set the SIGALRM handler",
                ATIMER_ERROR);

    return ATIMER_OK;
}


int
atimer_done(atimer_group *tgroup)
{
    int res = 0;

    logerr_passert(tgroup);

    res = aqueue_done(&tgroup->timers);
    logerr_return_str(res == AQUEUE_OK,
                "Failed to cleanup timer queue",
                ATIMER_ERROR);

    return ATIMER_OK;
}


int
atimer_set_timer_subsec(atimer_group *tgroup, void *data,
                long long val, int units)
{
    int res = 0;
    struct timeval trel;
    long long nsec = 0;

    logerr_assert_str(val >= 0, "Negative value for subsecond units");

    switch (units) {
    
        case ATIMER_SEC:
            trel.tv_sec  = val;
            trel.tv_usec = 0;
            break;

        case ATIMER_MSEC:
            trel.tv_sec  = (val / 1000);
            trel.tv_usec = (val % 1000) * 1000;
            break;

        case ATIMER_USEC:
            trel.tv_sec  = (val / 1000000);
            trel.tv_usec = (val % 1000000);
            break;

        case ATIMER_NSEC:
            trel.tv_sec  = (val / 1000000000);
            nsec         = (val % 1000000000);
            trel.tv_usec = (nsec / 1000);

            /* XXX Nanoseconds rounded to highest millisecond */
            if (nsec % 1000) {
                trel.tv_usec += 1;
                /* now tv_usec may overflow */
                trel.tv_sec  += trel.tv_usec / 1000000000;
                trel.tv_usec  = trel.tv_usec % 1000000000;
            }
            break;


        default:
            logerr_fatal_str(0, "Time unit not supported");
            break;
    }

    res = atimer_set_timer(tgroup, data, NULL, &trel, ATIMER_REL);
    return res;
}


int
atimer_set_timer(atimer_group *tgroup, void *data,
        struct timeval *absolute, struct timeval *relative, int type)
{
    int res = 0;
    atimer_header header;
    struct iovec vector[2];
    int count = 2;

    logerr_passert(data);

    logerr_return_str(type == ATIMER_ABS || type == ATIMER_REL,
                "Invalid timer type",
                ATIMER_EINVAL);

    logerr_return_str(absolute != NULL || type == ATIMER_REL,
                "NULL absolute pointer",
                ATIMER_EINVAL);

    logerr_return_str(relative != NULL || type == ATIMER_ABS,
                "NULL relative pointer",
                ATIMER_EINVAL);

    if (absolute == NULL)
        absolute = &header.expires;

    if (relative == NULL)
        relative = &header.duration;

    res = atimer_set_alarm(absolute, relative, type, ATIMER_ALL);
    logerr_return_str(res >= ATIMER_OK,
                "Failed to set the alarm",
                ATIMER_EALARM);

    header.expires  = *absolute;
    header.duration = *relative;

    /* insert the timer entry in the queue */

    vector[0].iov_base = &header;
    vector[0].iov_len  = sizeof(header);

    vector[1].iov_base = data;
    vector[1].iov_len  = tgroup->data_size;

    res = aqueue_pushv_data_back(&tgroup->timers, vector, count);

    logerr_log(999, "atimer_set_timer: setting a timer");
    atimer_header_log(999, "    ", &header);

    return ATIMER_OK;
}

/*
 * RETURN VALUE
 *
 *     On success:
 *
 *         ATIMER_EMPTY
 *         ATIMER_NONE
 *         ATIMER_OK
 *
 *     On error:
 *
 *         ATIMER_EEXTRACT
 *         ATIMER_EALARM
 *
 */
int
atimer_get_expired(atimer_group *tgroup, void *data,
                atimer_header *current, atimer_header *next,
                const struct timeval *now)
{
    int res = 0;
    struct iovec vector[2];
    int count = 2;
    int found_expired = 0;
    int more_timers = 0;

    atimer_header local_current;
    atimer_header local_next;
    struct timeval local_now;

    if (current == NULL)
        current = &local_current;

    if (next == NULL)
        next = &local_next;

    memset(next, 0, sizeof(atimer_header));

    logerr_passert(tgroup);

    if (aqueue_elements(&tgroup->timers) == 0)
        return ATIMER_EMPTY;

    /* get timestamp only if not provided */

    if (now == NULL) {
        res = gettimeofday(&local_now, NULL);
        logerr_assert_sys(res != -1, "gettimeofday failed");
        now = &local_now;
    }

    /* has the first timer expired? - we know there is at least one */

    res = atimer_next_expires(tgroup, &current->expires);
    logerr_assert_str(res == ATIMER_OK, "Failed to get next timer event");

    if (atimer_timeval_cmp(&current->expires, now) <= 0) {

        /* timer has expired */

        found_expired = 1;

        /* remove the timer entry */

        vector[0].iov_base = current;
        vector[0].iov_len  = sizeof(atimer_header);

        vector[1].iov_base = data;
        vector[1].iov_len  = tgroup->data_size;

        res = aqueue_popv_data_front(&tgroup->timers, vector, count);

        logerr_return_str(res == AQUEUE_OK,
                    "Failed to extract timer entry from queue",
                    ATIMER_EEXTRACT);
    }

    /*
     * set alarm for next timer event
     *
     * Needed because being in this function means that we
     * received a SIGALRM and before we exit we must reschedule.
     *
     * Note that this works even with multiple timer groups
     * because when we set the alarm, only the closest timer
     * will actually be scheduled, no matter how many times
     * we invoke atimer_set_alarm().
     */

    /* peek at the next timer event if any */

    res = atimer_next_expires(tgroup, &next->expires);
    logerr_assert_str(res >= ATIMER_OK, "Failed to get next timer event");

    more_timers = (res == ATIMER_OK);

    if (more_timers) {
        res = atimer_set_alarm(&next->expires, &next->duration,
                    ATIMER_ABS, ATIMER_ALL);
        logerr_return_str(res >= ATIMER_OK,
                    "Failed to set the alarm for the next timer entry",
                    ATIMER_EALARM);
    }

    return (found_expired) ? ATIMER_OK : ATIMER_NONE;
}


int
atimer_next_expires(atimer_group *tgroup, struct timeval *when)
{
    atimer_header *header = NULL;

    if (aqueue_elements(&tgroup->timers) == 0)
        return ATIMER_EMPTY;

    if (when == NULL)
        return ATIMER_EINVAL;

    header = aqueue_peek_front(&tgroup->timers);
    logerr_passert(header);

    memcpy(when, &header->expires, sizeof (struct timeval));

    return ATIMER_OK;
}


/* functions that operate on global structure */

int
atimer_set_alarm(struct timeval *absolute, struct timeval *relative,
                int type, int only_future)
{
    int res = 0;
    int scheduled = 0;
    int was_blocked = 0;
    struct itimerval titimer;

    struct timeval tnow;
    struct timeval tabs;
    struct timeval *prel = &titimer.it_value;

    /* set recurring alarm if requested */
    if (atimer_globals.recurring) {
        titimer.it_interval = atimer_globals.recurring_alarm;
    } else {
        titimer.it_interval.tv_sec  = 0;
        titimer.it_interval.tv_usec = 0;
    }

    res = gettimeofday(&tnow, NULL);
    logerr_assert_sys(res != -1, "gettimeofday failed");

    if (type == ATIMER_ABS) {
    
        logerr_passert(absolute);

        tabs = *absolute;
        atimer_timeval_get_offset(&tabs, &tnow, prel);
        
        if (relative != NULL)
            *relative = *prel;

    } else if (type == ATIMER_REL) {
    
        logerr_passert(relative);

        *prel = *relative;
        atimer_timeval_add_offset(&tabs, &tnow, prel);

        if (absolute != NULL)
            *absolute = tabs;

    } else {
        logerr_assert_str(0, "Invalid timeval type");
    }

    res = asignal_block(SIGALRM, &was_blocked);
    logerr_fatal_str(res == ASIGNAL_OK, "Failed to block SIGALRM");

    /* schedule a new alarm if needed */

    if (atimer_globals.pending &&
        atimer_timeval_cmp(&atimer_globals.pending_alarm, &tabs) <= 0) {

        /* an alarm is already scheduled to be delivered before our due time */

        if (!was_blocked) {
            res = asignal_unblock(SIGALRM, NULL);
            logerr_fatal_str(res == ASIGNAL_OK, "Failed to unblock SIGALRM");
        }

        return ATIMER_NONEED;
    }

    /* we need to reset the timer */
    atimer_globals.pending = 1;
    atimer_globals.pending_alarm = tabs;

    if (atimer_positive_toffset(&titimer.it_value)) {
        /* schedule an alarm signal */
        res = setitimer(ITIMER_REAL, &titimer, NULL);
        logerr_assert_sys(res != -1, "Failed to set timer with setitimer");
        scheduled = 1;

    } else if (!only_future) {
        /* generate an alarm signal right away; timer has already expired */
        res = kill(getpid(), SIGALRM);
        logerr_assert_sys(res != -1, "Failed to send SIGALRM to process");
        scheduled = 1;
    }

    if (!was_blocked) {
        res = asignal_unblock(SIGALRM, NULL);
        logerr_fatal_str(res == ASIGNAL_OK, "Failed to unblock SIGALRM");
    }

    return (scheduled) ? ATIMER_OK : ATIMER_EXPIRED;
}


int
atimer_set_recurring_alarm(struct timeval *interval)
{
    int res = 0;

    logerr_passert(interval);

    atimer_globals.pending = 0;

    atimer_globals.recurring_alarm = *interval;
    atimer_globals.recurring = (interval->tv_sec || interval->tv_usec);

    res = atimer_set_alarm(NULL, interval, ATIMER_REL, ATIMER_ALL);
    logerr_fatal_str(res == ATIMER_OK, "Failed to set alarm");

    return ATIMER_OK;
}



static void
atimer_alarm_handler(int signo)
{
    logerr_assert_str(signo == SIGALRM,
                "Did not receive SIGALRM signal");

    logerr_assert_str(atimer_globals.alarm_count >= 0,
                "Global alarm counter is negative");

    logerr_log(999, "atimer_alarm_handler: SIGALRM received");

    atimer_globals.alarm_count++;
    atimer_globals.pending = 0;
}


/* auxiliary functions */

void
atimer_timeval_add_offset(struct timeval *dst,
                const struct timeval *src, const struct timeval *offs)
{
    long usec  = 0;
    long carry = 0;

    logerr_passert(dst);
    atimer_assert_valid_tvalue(src);
    atimer_assert_valid_toffset(offs);

    usec = src->tv_usec + offs->tv_usec;

    if (usec < 0) {
        carry -= 1;
        usec += 1000000;
    }

    carry += (usec / 1000000);
    usec   = (usec % 1000000);

    dst->tv_sec  = src->tv_sec  + offs->tv_sec + carry;
    dst->tv_usec = usec;

    atimer_assert_valid_tvalue(dst);
}

void
atimer_timeval_get_offset(const struct timeval *dst,
                const struct timeval *src, struct timeval *offs)
{
    long usec  = 0;
    long carry = 0;

    atimer_assert_valid_tvalue(dst);
    atimer_assert_valid_tvalue(src);
    logerr_passert(offs);

    usec = dst->tv_usec - src->tv_usec;

    if (usec < 0) {
        carry -= 1;
        usec += 1000000;
    }

    carry += (usec / 1000000);
    usec   = (usec % 1000000);

    offs->tv_sec  = dst->tv_sec - src->tv_sec + carry;
    offs->tv_usec = usec;

    atimer_assert_valid_toffset(offs);
}


int
atimer_timeval_cmp(const void *a, const void *b)
{
    const struct timeval *ta = a;
    const struct timeval *tb = b;
    
    logerr_passert(ta);
    logerr_passert(tb);

    if (ta->tv_sec < tb->tv_sec)
        return -1;

    if (ta->tv_sec > tb->tv_sec)
        return 1;

    if (ta->tv_usec < tb->tv_usec)
        return -1;

    if (ta->tv_usec > tb->tv_usec)
        return 1;

    return 0;
}

void
atimer_timeval_log(int level, const char *prefix, const struct timeval *t)
{
    logerr_passert(t);
    logerr_log(level, "%stimeval sec(%d) usec(%d)",
                prefix, t->tv_sec, t->tv_usec);
}

void
atimer_header_log(int level, const char *prefix, const atimer_header *h)
{
    logerr_passert(h);
    logerr_log(level, "%sexpires  : sec(%d) usec(%d)",
                prefix, h->expires.tv_sec, h->expires.tv_usec);
    logerr_log(level, "%sduration : sec(%d) usec(%d)",
                prefix, h->duration.tv_sec, h->duration.tv_usec);
}

void
atimer_now_log(int level, const char *prefix)
{
    int res = 0;
    struct timeval now;

    res = gettimeofday(&now, NULL);
    logerr_assert_sys(res != -1, "gettimeofday failed");
    
    atimer_timeval_log(level, prefix, &now);
}

inline void
atimer_gettime_usec(long *sec, long *usec)
{
    int res = 0;
    struct timeval now;

    res = gettimeofday(&now, NULL);
    logerr_assert_sys(res != -1, "gettimeofday failed");

    *sec = now.tv_sec;
    *usec = now.tv_usec;
}

inline void
atimer_gettime_nsec(long *sec, long *nsec)
{
    int res = 0;
    struct timeval now;

    res = gettimeofday(&now, NULL);
    logerr_assert_sys(res != -1, "gettimeofday failed");

    *sec = now.tv_sec;

    logerr_assert_str(sizeof(long) >= 4,
        "usec will overflow");

    /* multiply by 1000 using only shifts and addition */
    *nsec = (now.tv_usec << 10) - (now.tv_usec << 4) - (now.tv_usec << 3);
}


