/*
 * Created on   : Sun Jul 18 19:05:43 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * atimer.h
 *
 * Provides per-process timers
 * 
 * $Id: atimer.h,v 1.19 2005/11/02 00:22:04 pavlos Exp $
 */

#ifndef __ATIMER_H
#define __ATIMER_H

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "aqueue.h"
#include "logerr.h"
#include "asignal.h"
#include "rbufx.h"

#define ATIMER_MSEC_PER_SEC     1000L
#define ATIMER_USEC_PER_SEC     1000000L
#define ATIMER_NSEC_PER_SEC     1000000000L

#define ATIMER_USEC_PER_MSEC    1000L
#define ATIMER_NSEC_PER_MSEC    1000000L

#define ATIMER_EXPIRED          4
#define ATIMER_NONEED           3
#define ATIMER_EMPTY            2
#define ATIMER_NONE             1
#define ATIMER_OK               0
#define ATIMER_ERROR           -1
#define ATIMER_EEXTRACT        -2
#define ATIMER_EALARM          -4
#define ATIMER_EINVAL          -5

#define ATIMER_DEF_RECYCLE      10

#define ATIMER_ONLY_FUTURE      1
#define ATIMER_ALL              0

#define ATIMER_SEC              0
#define ATIMER_MSEC             3
#define ATIMER_USEC             6
#define ATIMER_NSEC             9

#define ATIMER_BUSY_THRESHOLD  40000

#define atimer_assert_valid_tvalue(v) \
        (logerr_passert(v) || \
         logerr_assert_str((v)->tv_sec >= 0, "Negative timeval sec") || \
         logerr_assert_str((v)->tv_usec >= 0, "Negative timeval usec") || \
         logerr_assert_str((v)->tv_usec < 10000000, "timeval usec overflows"))

#define atimer_assert_valid_toffset(v) \
        (logerr_passert(v) || \
         logerr_assert_str((v)->tv_usec < 10000000, "timeval usec overflows"))


#define atimer_positive_toffset(v) \
        (((v)->tv_sec > 0) || ((v)->tv_sec == 0 && (v)->tv_usec > 0))

#define atimer_zero_timeval(v) \
        ((v)->tv_sec == 0 && (v)->tv_usec == 0)

#define atimer_timeval_usec_diff(dst, src, ptr) \
    do { \
        *(ptr) = (dst)->tv_sec - (src)->tv_sec; \
        if (*(ptr)) \
            *(ptr) = *(ptr) * 1000000; \
        *(ptr) += (dst)->tv_usec; \
        *(ptr) -= (src)->tv_usec; \
    } while (0)

static inline long long
atimer_timeval_diff_usec(const struct timeval *dst, const struct timeval *src)
{
        long long diff = (dst->tv_sec - src->tv_sec) * ATIMER_USEC_PER_SEC;
        diff += dst->tv_usec;
        diff -= src->tv_usec;
        return diff;
}

#define atimer_usleep(usec) \
    do { \
        long _usec = (usec); \
        long _slept = 0; \
        long _still = 0; \
        struct timeval _now; \
        struct timeval _start; \
        struct timeval _timeout; \
        if ((_usec) <= 0) break; \
        gettimeofday(&_start, NULL); \
        while (1) { \
            gettimeofday(&_now, NULL); \
            atimer_timeval_usec_diff(&_now, &_start, &_slept); \
            _still = _usec - _slept; \
            if (_still < 1) break; \
            if (_still > ATIMER_BUSY_THRESHOLD) { \
                _still -= ATIMER_BUSY_THRESHOLD; \
                _timeout.tv_sec  = _still / 1000000; \
                _timeout.tv_usec = _still % 1000000; \
                select(0, NULL, NULL, NULL, &_timeout); \
            } \
        } \
    } while(0)
    
enum {
    ATIMER_REL,
    ATIMER_ABS
};

typedef struct atimer_group {
    aqueue timers;
    size_t data_size;
    asignal_handler_func *orig_alarm_handler;
    rbuf_mgr *rbmgr;
} atimer_group;

typedef struct atimer_header {
    struct timeval expires;
    struct timeval duration;
} atimer_header;

/* sleep function using select (at least microsecond precision) */



#define atimer_sleep_sec(sec)   atimer_usleep((sec)  * 1000000)
#define atimer_sleep_msec(msec) atimer_usleep((msec) * 1000)

/* functions that operate on a timer group */

int atimer_init(atimer_group *tgroup, size_t data_size,
        size_t recycle_hint, rbuf_mgr *rbmgr);
int atimer_done(atimer_group *t);

int atimer_set_timer_subsec(atimer_group *tgroup, void *data,
                long long val, int units);
int atimer_set_timer(atimer_group *tgroup, void *data,
        struct timeval *absolute, struct timeval *relative, int type);

#define atimer_set_timer_sec(tgroup, data, sec) \
        atimer_set_timer_subsec((tgroup), (data), (sec), ATIMER_SEC);

#define atimer_set_timer_msec(tgroup, data, msec) \
        atimer_set_timer_subsec((tgroup), (data), (msec), ATIMER_MSEC);

#define atimer_set_timer_usec(tgroup, data, usec) \
        atimer_set_timer_subsec((tgroup), (data), (usec), ATIMER_USEC);

#define atimer_set_timer_nsec(tgroup, data, nsec) \
        atimer_set_timer_subsec((tgroup), (data), (nsec), ATIMER_NSEC);

int atimer_get_expired(atimer_group *tgroup, void *data,
                atimer_header *current, atimer_header *next,
                const struct timeval *now);

int atimer_next_expires(atimer_group *tgroup, struct timeval *when);


/* functions that operate on global structure */

int atimer_set_alarm(struct timeval *absolute, struct timeval *relative,
                int type, int only_future);
int atimer_set_recurring_alarm(struct timeval *interval);

/* auxiliary functions */

void atimer_timeval_add_offset(struct timeval *dst,
                const struct timeval *src, const struct timeval *offs);

void atimer_timeval_get_offset(const struct timeval *dst,
                const struct timeval *src, struct timeval *offs);

int atimer_timeval_cmp(const void *a, const void *b);

void atimer_timeval_log(int level, const char *prefix, const struct timeval *t);
void atimer_header_log(int level, const char *prefix, const atimer_header *h);
void atimer_now_log(int level, const char *prefix);

static inline void
atimer_timespec_add_offset(struct timespec *dst,
                           const struct timespec *src,
                           const struct timespec *offs)
{
    long nsec  = 0;
    long carry = 0;

    nsec = src->tv_nsec + offs->tv_nsec;

    if (nsec < 0) {
        carry -= 1;
        nsec += 1000000000;
    }

    carry += (nsec / 1000000000);
    nsec   = (nsec % 1000000000);

    dst->tv_sec  = src->tv_sec  + offs->tv_sec + carry;
    dst->tv_nsec = nsec;
}

static inline void
atimer_timeval_to_timespec(struct timespec *dst,
                           const struct timeval *src)
{
        dst->tv_sec = src->tv_sec;
        dst->tv_nsec = src->tv_usec;
        dst->tv_nsec *= 1000;
}

static inline void
atimer_nsec_to_timespec(struct timespec *tsp, const long long int nsec)
{
        tsp->tv_sec = nsec / ATIMER_NSEC_PER_SEC;
        tsp->tv_nsec = nsec % ATIMER_NSEC_PER_SEC;
}

static inline void
atimer_msec_to_timespec(struct timespec *tsp, const long long int msec)
{
        return atimer_nsec_to_timespec(tsp, msec * ATIMER_NSEC_PER_MSEC);
}

static inline int
atimer_timespec_is_zero(const struct timespec *tsp)
{
        return (tsp->tv_sec == 0 && tsp->tv_nsec == 0);
}

static inline double
atimer_timespec_to_sec(const struct timespec *tsp)
{
        return (1.0 * tsp->tv_sec) + (1.0 * tsp->tv_nsec / ATIMER_NSEC_PER_SEC);
}

// equivalent to system command `date` without a newline
static inline const char *
atimer_datestr()
{
        // STATIC variable
        static char datestr[100];
        int datelen = 0;
        time_t now;

        now = time(NULL);
        strncpy(datestr, ctime(&now), sizeof(datestr));
        datelen = strlen(datestr);
        if (datelen > 0 && datestr[datelen-1] == '\n')
                datestr[datelen-1] = 0;

        return datestr;
}

#endif /* __ATIMER_H */

