/*
 * Created on   : Sat Dec  4 17:53:07 EST 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * bqueue.c
 *
 * Byte Queue
 *
 * $Id: bqueue.c,v 1.3 2005/03/19 02:20:18 pavlos Exp $
 */

#include "bqueue.h"
#include "logerr.h"
#include "mem.h"
#include "cmp.h"

int
bqueue_init(bqueue *bq, int bufsize)
{
    logerr_passert(bq);
    logerr_assert_str(bufsize > 0, "Invalid bqueue size");
    
    memset(bq, 0, sizeof(bqueue));

    bq->buf = malloc(bufsize);
    bq->bufsize = bufsize;

    bqueue_flush(bq);

    return BQUEUE_OK;
}


int
bqueue_done(bqueue *bq)
{
    logerr_passert(bq);
    
    if (bq->bufsize > 0) {
        free(bq->buf);
        bq->buf = NULL;
        bq->bufsize = 0;
    }

    return BQUEUE_OK;
}


int
bqueue_flush(bqueue *bq)
{
    logerr_passert(bq);
    logerr_passert(bq->buf);

    bq->first = NULL;
    bq->last = NULL;
    bq->bytes = 0;

    memset(bq->buf, 0, bq->bufsize);

    return BQUEUE_OK;
}

int
bqueue_reserve(bqueue *bq, int dsize, int where,
        struct iovec *iov, unsigned int *iovlen)
{
    int free_contiguous = 0;

    void *chunk1 = NULL;
    int c1size = 0;

    void *chunk2 = NULL;
    int c2size = 0;

    int reserve_bytes = 0;

    logerr_passert(bq);
    logerr_bug(where == BQUEUE_FRONT || where == BQUEUE_BACK);
    logerr_bug(iov != NULL && iovlen != NULL && *iovlen >= 2);
    logerr_bug(bq->bytes == 0 ||
                (bq->buf != NULL && bq->first != NULL && bq->last != NULL));

    *iovlen = 0;

    if (dsize == 0)
        return BQUEUE_ENOTHING;

    reserve_bytes = dsize;

    if ((bq->bufsize - bq->bytes) < reserve_bytes)
        return BQUEUE_ENOSPACE;

    if (bq->bytes == 0) {

        /* queue is empty; fill contiguous space from beginning */

        c1size = reserve_bytes;
        chunk1 = bq->buf;
        bq->first = chunk1;
        bq->last  = bytes_add(chunk1, reserve_bytes - 1);

    } else if (where == BQUEUE_FRONT) {

        if (bq->first <= bq->last) {
            /* free space wraps around; content contiguous */
            free_contiguous = bq->first - bq->buf;
        } else {
            /* free space contiguous; content wraps around */
            free_contiguous = bq->first - bq->last - 1;
        }

        if (free_contiguous < reserve_bytes) {
            
            /*
             * insert two chunks into the buffer:
             *
             *     + one from bg->first backwards
             *     + one from the end of the buffer backwards
             */

            c2size = free_contiguous;
            chunk2 = bq->buf;
            c1size = reserve_bytes - c2size;
            chunk1 = bytes_add(bq->buf, bq->bufsize - c1size);
            bq->first = chunk1;

        } else {
            
            /* insert into one contiguous chunk in front of bq->first */

            c1size = reserve_bytes;
            chunk1 = bytes_sub(bq->first, c1size);
            bq->first = chunk1;
        }
        
    } else if (where == BQUEUE_BACK) {

        if (bq->first <= bq->last) {
            /* free space wraps around; content contiguous */
            free_contiguous = bq->bufsize - (bq->last - bq->buf + 1);
        } else {
            /* free space contiguous; content wraps around */
            free_contiguous = bq->first - bq->last - 1;
        }

        if (free_contiguous < reserve_bytes) {
            
            /*
             * insert two chunks into the buffer:
             *
             *     + one from bg->last onwards
             *     + one from the begining of the buffer onwards
             */

            c1size = free_contiguous;
            chunk1 = bytes_add(bq->last, 1);
            c2size = reserve_bytes - c1size;
            chunk2 = bq->buf;
            bq->last = bytes_add(chunk2, c2size - 1);

        } else {
            
            /* insert into one contiguous chunk after of bq->last */

            c1size = reserve_bytes;
            chunk1 = bytes_add(bq->last, 1);
            bq->last = bytes_add(chunk1, c1size - 1);
        }
    }

    logerr_bug(c1size + c2size == reserve_bytes);
    logerr_bug(c1size >= 0 && c2size >= 0);

    *iovlen = 0;

    /* c1size and c2size can be non-zero or zero in any combination */

    if (c1size > 0) {
        iov[*iovlen].iov_base = chunk1;
        iov[*iovlen].iov_len  = c1size;
        bq->bytes += c1size;
        (*iovlen)++;
    }

    if (c2size > 0) {
        iov[*iovlen].iov_base = chunk2;
        iov[*iovlen].iov_len  = c2size;
        bq->bytes += c2size;
        (*iovlen)++;
    }

    return reserve_bytes;
}


int
bqueue_push(bqueue *bq, void *data, int dsize, int where)
{
    int res = 0;
    struct iovec iov[2];
    size_t iovlen = 2;
    int remaining = dsize;

    logerr_bug(dsize == 0 || data != NULL);

    res = bqueue_reserve(bq, dsize, where, iov, &iovlen);

    if (iovlen == 0)
        return res;

    logerr_bug(res == dsize);
    logerr_bug(iovlen <= 2);

    if (iovlen >= 1) {
        memcpy(iov[0].iov_base, data, iov[0].iov_len);
        remaining -= iov[0].iov_len;
    }

    if (iovlen >= 2) {
        memcpy(iov[1].iov_base, (data + iov[0].iov_len), iov[1].iov_len);
        remaining -= iov[1].iov_len;
    }

    logerr_bug(remaining == 0);

    return res;
}


int
bqueue_peek(bqueue *bq, int dsize, int from, struct iovec *iov, int *iovlen)
{
    int occupied_contiguous = 0;

    void *chunk1 = NULL;
    int c1size = 0;

    void *chunk2 = NULL;
    int c2size = 0;

    int peek_bytes = 0;

    logerr_passert(bq);
    logerr_bug(from == BQUEUE_FRONT || from == BQUEUE_BACK);
    logerr_bug(dsize > 0);
    logerr_bug(iovlen != NULL && *iovlen >= 2);
    logerr_bug(bq->bytes == 0 ||
                (bq->buf != NULL && bq->first != NULL && bq->last != NULL));

    *iovlen = 0;
    peek_bytes = cmp_min(dsize, bq->bytes);

    if (bq->bytes == 0)
        return BQUEUE_EEMPTY;

    if (from == BQUEUE_FRONT) {

        if (bq->first <= bq->last) {
            /* free space wraps around; content contiguous */
            occupied_contiguous = (bq->last - bq->first + 1);
        } else {
            /* free space contiguous; content wraps around */
            occupied_contiguous = bq->bufsize - (bq->first - bq->buf);
        }

        if (occupied_contiguous < peek_bytes) {

            /*
             * extract two buffer chunks: one from bq->first onwards
             * and one from the beginning of the buffer onwards
             */

            c1size = occupied_contiguous;
            chunk1 = bq->first;
            c2size = peek_bytes - c1size;
            chunk2 = bq->buf;

        } else {

            /* copy only one contiguous chunk from bq->first onwards */

            c1size = peek_bytes;
            chunk1 = bq->first;
        }

    } else if (from == BQUEUE_BACK) {
        
        if (bq->first <= bq->last) {
            /* free space wraps around; content contiguous */
            occupied_contiguous = (bq->last - bq->first + 1);
        } else {
            /* free space contiguous; content wraps around */
            occupied_contiguous = (bq->last - bq->buf + 1);
        }

        if (occupied_contiguous < peek_bytes) {

            /*
             * copy from two chunks: one from bq->last backwards
             * and one from the end of the buffer backwards
             */

            c2size = occupied_contiguous;
            chunk2 = bq->buf;
            c1size = peek_bytes - c2size;
            chunk1 = bytes_add(bq->buf, bq->bufsize - c1size);

        } else {

            /* copy only one contiguous chunk from bq->last backwards */

            c1size = peek_bytes;
            chunk1 = bytes_sub(bq->last, c1size - 1);
        }
    }

    logerr_bug(c1size + c2size == peek_bytes);
    logerr_bug(c1size > 0 && c2size >= 0);

    /* set the peek buffer vectors; always c1size > 0 */

    iov[*iovlen].iov_base = chunk1;
    iov[*iovlen].iov_len  = c1size;
    (*iovlen)++;

    if (c2size > 0) {
        iov[*iovlen].iov_base = chunk2;
        iov[*iovlen].iov_len  = c2size;
        (*iovlen)++;
    }

    return peek_bytes;
}


int
bqueue_pop(bqueue *bq, void *data, int dsize, int from)
{
    int occupied_contiguous = 0;

    void *chunk1 = NULL;
    int c1size = 0;

    void *chunk2 = NULL;
    int c2size = 0;

    int pop_bytes = 0;

    logerr_passert(bq);
    logerr_bug(from == BQUEUE_FRONT || from == BQUEUE_BACK);
    logerr_bug(bq->bytes == 0 ||
                (bq->buf != NULL && bq->first != NULL && bq->last != NULL));

    pop_bytes = cmp_min(dsize, bq->bytes);

    if (bq->bytes == 0)
        return BQUEUE_EEMPTY;

    if (from == BQUEUE_FRONT) {

        if (bq->first <= bq->last) {
            /* free space wraps around; content contiguous */
            occupied_contiguous = (bq->last - bq->first + 1);
        } else {
            /* free space contiguous; content wraps around */
            occupied_contiguous = bq->bufsize - (bq->first - bq->buf);
        }

        if (occupied_contiguous < pop_bytes) {

            /*
             * extract two buffer chunks: one from bq->first onwards
             * and one from the beginning of the buffer onwards
             */

            c1size = occupied_contiguous;
            chunk1 = bq->first;
            c2size = pop_bytes - c1size;
            chunk2 = bq->buf;

            /* pointer has wrapped around */
            bq->first = bytes_add(chunk2, c2size);

        } else {

            /* copy only one contiguous chunk from bq->first onwards */

            c1size = pop_bytes;
            chunk1 = bq->first;

            /* may need to wrap around -- adjust later for empty buffer */
            bq->first = (c1size == occupied_contiguous) ?
                        bq->buf : bytes_add(bq->first, c1size);
        }

    } else if (from == BQUEUE_BACK) {
        
        if (bq->first <= bq->last) {
            /* free space wraps around; content contiguous */
            occupied_contiguous = (bq->last - bq->first + 1);
        } else {
            /* free space contiguous; content wraps around */
            occupied_contiguous = (bq->last - bq->buf + 1);
        }

        if (occupied_contiguous < pop_bytes) {

            /*
             * copy from two chunks: one from bq->last backwards
             * and one from the end of the buffer backwards
             */

            c2size = occupied_contiguous;
            chunk2 = bq->buf;
            c1size = pop_bytes - c2size;
            chunk1 = bytes_add(bq->buf, bq->bufsize - c1size);

            /* pointer has wrapped around */
            bq->last = bytes_sub(chunk1, 1);

        } else {

            /* copy only one contiguous chunk from bq->last backwards */

            c1size = pop_bytes;
            chunk1 = bytes_sub(bq->last, c1size - 1);

            /* may need to wrap around -- adjust later for empty buffer */
            bq->last = (c1size == occupied_contiguous) ?
                        bytes_add(bq->buf, bq->bufsize - 1) :
                        bytes_sub(chunk1, 1);
        }
    }

    logerr_bug(c1size + c2size == pop_bytes);
    logerr_bug(c1size > 0 && c2size >= 0);

    /* if data == NULL, then no bytes are copied but bytes are released */

    if (data != NULL)
        memcpy(data, chunk1, c1size);
    bq->bytes -= c1size;

    /* XXX zeroing unused memory */
    memset(chunk1, 0, c1size);

    if (c2size > 0) {
        if (data != NULL)
            memcpy(bytes_add(data, c1size), chunk2, c2size);
        bq->bytes += c2size;

        /* XXX zeroing unused memory */
        memset(chunk2, 0, c2size);
    }

    logerr_bug(bq->bytes >= 0);

    /* adjust for empty buffer */

    if (bq->bytes == 0) {
        bq->first = NULL;
        bq->last  = NULL;
    }

    return pop_bytes;
}

void
bqueue_log(int level, bqueue *bq)
{
    logerr_passert(bq);

    logerr_log(level, "ByteQueue - bufsize(%d) bytes(%d) first(%p) last(%p)",
        bq->bufsize, bq->bytes, bq->first, bq->last);

    if (bq->bufsize > 0) {

        u_char *byte = NULL;
        int i = 0;

        logerr_passert(bq->buf);

        logerr_log_more(level, LOGERR_FIRST, "");

        for (i = 0, byte = bq->buf; i < bq->bufsize; i++, byte++) {

            char before = ' ';
            char after  = ' ';

            if (i > 0 && i % 8 == 0) {
                logerr_log_more(level, LOGERR_LAST, "");
                logerr_log_more(level, LOGERR_FIRST, "");
            }

            if (byte == bq->first)
                before = '[';

            if (byte == bq->last)
                after = ']';

            logerr_log_more(level, LOGERR_MIDDLE,
                "%c%.2x%c", before, *byte, after);
        }

        logerr_log_more(level, LOGERR_LAST, "");
    }
}

