/*
 * Created on   : Sat Dec 11 16:43:29 EST 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * buf.c
 *
 * Buffer related utilities
 * 
 * $Id: buf.c,v 1.4 2005/01/03 23:15:01 pavlos Exp $Buffer related utilities
 */

#include "buf.h"
#include "logerr.h"

/* bufmem */

int
bufmem_init_alloc(bufmem *mem, size_t alloc_size, int view_count)
{
    logerr_passert(mem);

    memset(mem, 0, sizeof(bufmem));
    
    /* allocate the memory */

    mem->ptr_size = alloc_size;
    mem->ptr = malloc(mem->ptr_size);
    mem->ptr_owner = 1;
    logerr_fatal_str(mem->ptr != NULL, "Out of memory");

    bufmem_reset(mem, view_count);

    return BUF_OK;
}

int
bufmem_init_ptr(bufmem *mem, uint8_t *ptr, size_t ptr_size,
        int ptr_owner, int view_count)
{
    logerr_passert(mem);
    logerr_passert(ptr);

    memset(mem, 0, sizeof(bufmem));
    
    /* do not allocate memory but if owner, will be responsible to free */

    mem->ptr_size = ptr_size;
    mem->ptr = ptr;
    mem->ptr_owner = ptr_owner;

    bufmem_reset(mem, view_count);

    return BUF_OK;
}

int
bufmem_done(bufmem *mem)
{
    logerr_passert(mem);
    
    if (mem->ptr_owner) {
        logerr_passert(mem->ptr);
        free(mem->ptr);
        mem->ptr = NULL;
        mem->ptr_owner = 0;
    }

    memset(mem, 0, sizeof(bufmem));
    
    return BUF_OK;
}


void
bufmem_free(void *mem)
{
    logerr_passert(mem);

    bufmem_done((bufmem *)mem);
    free(mem);
}


int
bufmem_reset(bufmem *mem, int view_count)
{
    logerr_passert(mem);
    logerr_passert(mem->ptr);
    logerr_bug(mem->ptr_size > 0);
    logerr_bug(view_count >= 0);
    logerr_bug(mem->ptr_size > view_count * sizeof(bufview));

    /* first reset the buffer */

    mem->inuse = 0;
    memset(mem->ptr, 0, mem->ptr_size);
    
    /* set aside memory for view structures */

    if (view_count > 0) {

        mem->view_count = view_count;
        mem->views = bytes_add(mem->ptr,
                mem->ptr_size - (mem->view_count * sizeof(bufview)));
        
        /* normally we should call init; XXX hack (view->attached == 0) */
        memset(mem->views, 0, mem->view_count * sizeof(bufview));

    } else {
        mem->views = 0;
        mem->view_count = 0;
    }

    mem->exported_size = mem->ptr_size - (mem->view_count * sizeof(bufview));

    return BUF_OK;
}


int
bufmem_cmp(const void *mem1, const void *mem2)
{
   return cmp_ascend(((bufmem *)mem1)->ptr_size, ((bufmem *)mem2)->ptr_size);
}


/* bufview */

int
bufview_init(bufview *view)
{
    logerr_passert(view);
    
    memset(view, 0, sizeof(bufview));
    view->attached = 0;

    return BUF_OK;
}

int
bufview_done(bufview *view)
{
    bufview_reset(view);
    return BUF_OK;
}

int
bufview_reset(bufview *view)
{
    logerr_passert(view);

    if (view->attached) {
        logerr_passert(view->mem);
        bufmem_release(view->mem);
        memset(view, 0, sizeof(bufview));
    }
    
    return BUF_OK;
}

int
bufview_attach(bufview *view, bufmem *mem, int offset, int size)
{
    logerr_passert(view);
    logerr_passert(mem);
    
    memset(view, 0, sizeof(bufview));

    bufmem_use(mem);
    view->mem = mem;
    view->ptr = mem->ptr;
    view->size = mem->exported_size;

    view->attached = 1;

    return BUF_OK;
}


/* bufmgr */

int
bufmgr_init(bufmgr *mgr)
{
    int res = 0;

    logerr_passert(mgr);

    memset(mgr, 0, sizeof(bufmgr));

    res = aqueue_init(&mgr->avail,
                AQUEUE_SORTED | AQUEUE_RECYCLE,
                0, AQUEUE_NOLIMIT, 0, 0,
                &bufmem_free, &bufmem_cmp);
    logerr_assert_str(res == AQUEUE_OK, "Failed to create queue");

    res = aqueue_init(&mgr->inuse,
                AQUEUE_RECYCLE,
                0, AQUEUE_NOLIMIT, 0, 0,
                &bufmem_free, NULL);
    logerr_assert_str(res == AQUEUE_OK, "Failed to create queue");

    return BUF_OK;
}


int
bufmgr_done(bufmgr *mgr)
{
    int res = 0;
    logerr_passert(mgr);

    res = aqueue_done(&mgr->avail);
    logerr_assert_str(res == ALIST_OK, "Failed to clean-up queue");

    res = aqueue_done(&mgr->inuse);
    logerr_assert_str(res == ALIST_OK, "Failed to clean-up queue");

    return BUF_OK;
}


bufmem *
bufmgr_getmem(bufmgr *mgr, size_t size)
{
    int res = 0;
    size_t desired_size = 0;
    size_t alloc_size = 0x40;
    int view_count = 0;
    bufmem *mem = NULL;

    logerr_passert(mgr);

    /* allocate in steps of 64, 128, 256, 512, 1024, ... */

    logerr_fatal_str(size <= (0x01 << (sizeof(size_t) - 1)),
        "Unsupported bufmem size: too large");

    /* decide how many views to set aside in the buffer */
    /* XXX hard coded for now */
    view_count = 8;

    desired_size = size + (view_count * sizeof(bufview));

    while (alloc_size < desired_size)
        alloc_size <<= 1;

    /* first look into the inuse queue */

    while (aqueue_elements(&mgr->inuse) > 0) {
        
        mem = aqueue_peek_front(&mgr->inuse);

        if (mem->inuse) {
            /* no available buffer here */
            mem = NULL;
            break;
        }

        /* found available bufmem, but is it the right size? */

        res = aqueue_pop_ptr_front(&mgr->inuse, &mem);
        logerr_fatal_str(res == AQUEUE_OK && mem != NULL,
                "Failed to pop bufmem from queue");

        bufmem_reset(mem, view_count);

        if (mem->ptr_size == alloc_size) {
            /* bufmem found */
            break;
        }

        /* bufmem not of right size; recycle it */

        res = aqueue_push_ptr_back(&mgr->avail, &mem, AQUEUE_OWNER);
        logerr_fatal_str(res == AQUEUE_OK,
                "Failed to push bufmem into queue");

        mem = NULL;
    }

    if (mem == NULL) {

        /* allocate a new bufmem */

        mem = malloc(sizeof(bufmem));
        logerr_fatal_str(mem != NULL, "Out of memory");

        res = bufmem_init_alloc(mem, alloc_size, view_count);
        logerr_fatal_str(res == BUF_OK, "Failed to init bufmem");
    }

    /* insert it into inuse list */

    res = aqueue_push_ptr_back(&mgr->inuse, &mem, AQUEUE_OWNER);
    logerr_fatal_str(res == AQUEUE_OK,
            "Failed to push bufmem into queue");

    return mem;
}


