#include "ahash.h"
#include "logerr.h"

typedef struct connection {
    uint16_t port;
    int id;
} connection;


uint64_t
hash_port(void *data, void *key)
{
    uint64_t hkey = 0;
    uint16_t port = 0;

    logerr_assert_str(
        (data == NULL && key != NULL) ||
        (data != NULL && key == NULL),
        "Exactly one of 'data' or 'key' field must be non-null");
    
    if (data != NULL) {
        port = ((connection *)data)->port;
    } else {
        port = *(uint16_t *)key;
    }

    /* bits [63..16] = 10101010... */
    /* bits [15..0]  = port */
    hkey = 0xaaaaaaaaaaaa0000LL | port;
    return ahash_mix64(hkey);
}

int cmp_connection(void *a, void *b)
{
    connection *acon = a;
    connection *bcon = b;

    logerr_passert(acon);
    logerr_passert(bcon);

    return (acon->port == bcon->port) ? 0 : (acon->port < bcon->port) ? -1 : 1;
}

int search_connection(void *data, void *key)
{
    connection *con = data;
    uint16_t *port  = key;

    logerr_passert(con);
    logerr_passert(port);

    return (con->port == *port) ? 0 : (con->port < *port) ? -1 : 1;
}

void log_connection(int level, char *prefix, void *data, void *arg)
{
    connection *con = data;

    logerr_passert(con);

    logerr_log(level, "%sport(%u)", prefix, con->port);
}

int
main(int argc, char **argv)
{
    int res = 0;
    uint16_t p = 0;
    connection *con = NULL;
    connection conar[50000];

    ahash hash;

    /* first level = 2^10, second level = 2^4 elements */
    res = ahash_init(&hash, 10, 4, &search_connection, &hash_port);
    logerr_fatal_str(res == AHASH_OK, "Failed to initialize hash");


    for (p = 10; p < 400; p++) {
        conar[p].port = p;
        conar[p].id = -p;
        res = ahash_insert(&hash, &conar[p], &p);
        logerr_fatal_str(res == AHASH_OK, "Failed to insert in hash");
        if (p % 100 == 0)
            logerr_log(1, "Added up to %u", p);
    }

    p = 14;
    res = ahash_delete(&hash, &p);
    logerr_fatal_str(res == AHASH_OK, "Failed to delete from hash");

    p = 15;
    res = ahash_delete(&hash, &p);
    logerr_fatal_str(res == AHASH_OK, "Failed to delete from hash");

    for (p = 3; p < 7; p++) {
        conar[p].port = p;
        conar[p].id = -p;
        res = ahash_insert(&hash, &conar[p], &p);
        logerr_fatal_str(res == AHASH_OK, "Failed to insert in hash");
    }

    p = 17;
    res = ahash_delete(&hash, &p);
    logerr_fatal_str(res == AHASH_OK, "Failed to delete from hash");


    for (p = 12; p < 19; p++) {
        logerr_log(1, "lookup for connection with port %u", p);
        con = ahash_lookup(&hash, &p);
        if (con == NULL) {
            logerr_log(1, "    NOT FOUND");
        } else {
            logerr_log(1, "    FOUND (%p) port(%u) id(%d)",
                con, con->port, con->id);
        }
    }

    ahash_log(1, &hash, &log_connection);

    res = ahash_done(&hash);
    logerr_fatal_str(res == AHASH_OK, "Failed to cleanup hash");

    return 0;
}

