
#define IDSTR               "server"
#define ACTIONSTR           "receiving"

#include "ipc_ex.h"

void
read_multi_in_separate_buffers(ipc_channel *c)
{
    int res = 0;
    int resA, resB, resC;
    rec_A ra;
    rec_B rb;
    rec_C rc;

    logerr_log(2, "Waiting to recv [read_multi_in_separate_buffers]");

    resA = ipc_recv(c, &ra, sizeof(rec_A));
    logerr_fatal_str(resA > 0, "server failed");
    logerr_log(1, "server: received record A with %d bytes", resA);

    resB = ipc_recv(c, &rb, sizeof(rec_B));
    logerr_fatal_str(resB > 0, "server failed");
    logerr_log(1, "server: received record B with %d bytes", resB);

    resC = ipc_recv(c, &rc, sizeof(rec_C));
    logerr_fatal_str(resC > 0, "server failed");
    logerr_log(1, "server: received record C with %d bytes", resC);

    res = resA + resB + resC;
    logerr_log(1, "server: received record with %d bytes", res);

    prec_parts(&ra, &rb, &rc);
}

void
read_once_in_separate_buffers(ipc_channel *c)
{
    int res = 0;
    rec_A ra;
    rec_B rb;
    rec_C rc;
    struct iovec bvec[3];

    bvec[0].iov_base = &ra;
    bvec[0].iov_len  = sizeof(rec_A);

    bvec[1].iov_base = &rb;
    bvec[1].iov_len  = sizeof(rec_B);

    bvec[2].iov_base = &rc;
    bvec[2].iov_len  = sizeof(rec_C);

    logerr_log(2, "Waiting to recv [read_once_in_separate_buffers]");

    res = ipc_recvv(c, bvec, 3, NULL);
    logerr_fatal_str(res > 0, "server failed");

    logerr_log(1, "server: received record with %d bytes", res);

    prec_parts(&ra, &rb, &rc);
}

void
read_once_in_hybrid_buffers(ipc_channel *c)
{
    int res = 0;
    rec_A ra;
    rec_bottom *rbottom;
    struct iovec bvec[2];

    bvec[0].iov_base = &ra;
    bvec[0].iov_len  = sizeof(rec_A);

    bvec[1].iov_base = NULL;
    bvec[1].iov_len  = 0;

    logerr_log(2, "Waiting to recv [read_once_in_hybrid_buffers]");

    res = ipc_recvv(c, bvec, 2, NULL);
    logerr_fatal_str(res > 0, "server failed");

    logerr_log(1, "server: received record with %d bytes", res);

    rbottom = bvec[1].iov_base;
    logerr_passert(rbottom);

    prec_parts(&ra, &rbottom->rb, &rbottom->rc);

    free(rbottom);
}

void
read_once_and_alloc_buffer(ipc_channel *c)
{
    int res = 0;
    struct iovec bvec[1];

    bvec[0].iov_base = NULL;
    bvec[0].iov_len  = 0;

    logerr_log(2, "Waiting to recv [read_once_and_alloc_buffer]");

    res = ipc_recvv(c, bvec, 1, NULL);
    logerr_fatal_str(res > 0, "server failed");

    logerr_log(1, "server: received record with %d bytes", res);

    logerr_passert(bvec[0].iov_base);
    prec((rec *)bvec[0].iov_base);
    free(bvec[0].iov_base);
}

void
read_in_one_exact_buffer(ipc_channel *c)
{
    int res = 0;
    rec r;

    logerr_log(2, "Waiting to recv [read_in_one_exact_buffer]");

    res = ipc_recv(c, &r, sizeof(rec));
    logerr_fatal_str(res > 0, "server failed");
    logerr_log(1, "server: received record with %d bytes", res);

    prec(&r);
}

void
server(void)
{
    int res = 0;
    ipc_channel client;

    res = ipc_init(&client);
    logerr_fatal_str(res == IPC_OK, "server failed");

    res = ipc_create_fifo(FIFO_SEND, 0700, IPC_CREATE|IPC_USE_EXISTING);
    logerr_fatal_str(res == IPC_OK, "client failed");

    res = ipc_open_read_fifo(&client, FIFO_SEND, 1);
    logerr_fatal_str(res == IPC_OK, "server failed");

    while (1) {

        read_multi_in_separate_buffers(&client);
        read_in_one_exact_buffer(&client);
        read_once_in_separate_buffers(&client);
        read_once_and_alloc_buffer(&client);
        read_once_in_hybrid_buffers(&client);

        sleep(SERVER_SLEEP);
    }

    res = ipc_done(&client);
    logerr_fatal_str(res == IPC_OK, "server failed");
}

void
doalarm (int signo)
{
    logerr_log(1, "Alarm! signal(%d)\n", signo);
}

void
doalarm2 (int signo)
{
    logerr_log(1, "Alarm SECOND! signal(%d)\n", signo);
}

int
main(void)
{
    struct itimerval t;
    asignal_handler_func *old_alarm = NULL;

    printf("Begin %s\n", IDSTR);

    asignal_set_handler(SIGALRM, SA_RESTART, &doalarm, &old_alarm);

    printf("OLD %p\n", old_alarm);

    t.it_interval.tv_sec  = 1;
    t.it_interval.tv_usec = 0;
    t.it_value.tv_sec  = 0;
    t.it_value.tv_usec = 1;
    setitimer(ITIMER_REAL, &t, NULL);


    server();

    printf("Done\n");
    return 0;
}
