#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include "logerr.h"
#include "asignal.h"

#define PKTSIZE_MAX     1

static struct {
    int is_server;
    int pktlen;
} global;

void main_parse_cmdline(int argc, char **argv);

void server();
void client();

void main_done(void);
void main_signal_handler(int signo);


/* server acts as TCP sink */

void
server()
{
    int listen_sock = -1;

    int res = 0;
    int on = 0;

    uint8_t pkt[PKTSIZE_MAX];

    struct sockaddr_un addr;

    logerr_log(0, "server: wating for tcp packets");

    /* create and bind the listen socket */

    addr.sun_family = AF_LOCAL;
    strncpy(addr.sun_path, "/tmp/sync.server", sizeof(addr.sun_path));

    unlink("/tmp/sync.server");

    listen_sock = socket(AF_LOCAL, SOCK_STREAM, 0);
    logerr_bug(listen_sock != -1);

    on = 1;
    res = setsockopt(listen_sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
    logerr_bug(res != -1);

    res = bind(listen_sock, (struct sockaddr *)&addr, sizeof(addr));
    logerr_bug(res != -1);

    res = listen(listen_sock, 0);
    logerr_bug(res != -1);


    /* keep serving */

    while (1) {

        int serve_sock = -1;
        int pktsize = 0;
        int pktcnt = 0;

        serve_sock = accept(listen_sock, NULL, 0);
        logerr_bug(res != -1);

        pktsize = PKTSIZE_MAX;

        while (pktsize > 0) {

            do {
                res = recv(serve_sock, pkt, PKTSIZE_MAX, 0);
            } while (res == -1 && errno == EINTR);

            logerr_bug(res != -1);
            pktsize = res;

            if (pktsize == 0)
                break;

            pktcnt++;

            if (pktcnt % 1 == 0)
                sleep(2);

            logerr_log(-1, "recv pktcnt(%d) pktsize(%d)",
                pktcnt, pktsize);
        }

        logerr_log(-1, "sender closed");

        res = close(serve_sock);
        logerr_bug(res != -1);
    }

    res = close(listen_sock);
    logerr_bug(res != -1);
}

void
client_keep_sending(int sock)
{
    int res = 0;
    int pktcnt = 0;

    uint8_t pkt[PKTSIZE_MAX];

    logerr_log(0, "client");

    while (1) {

        fd_set wfds;

        FD_ZERO(&wfds);
        FD_SET(sock, &wfds);

        res = select(sock + 1, NULL, &wfds, NULL, NULL);
        logerr_bug(res == 1);

        do {
            res = send(sock, pkt, global.pktlen, 0);
        } while (res == -1 && errno == EINTR);

        logerr_bug(res != -1);

        pktcnt++;
        logerr_log(-1, "recv pktcnt(%d) pktsize(%d)", pktcnt, global.pktlen);
    }
}

void
client()
{
    int res = 0;
    int sock = -1;

    struct sockaddr_un addr;

    addr.sun_family = AF_LOCAL;
    strncpy(addr.sun_path, "/tmp/sync.server", sizeof(addr.sun_path));

    /* create and connect client socket */

    sock = socket(AF_LOCAL, SOCK_STREAM, 0);
    logerr_bug(res != -1);

    res = connect(sock, (struct sockaddr *)&addr, sizeof(addr));
    logerr_bug(res != -1);

    client_keep_sending(sock);

    res = close(sock);
    logerr_assert_sys(res != -1, "socket error");
}



int
main (int argc, char **argv)
{
    int res = 0;

    /* default global values */

    memset(&global, 0, sizeof(global));
    global.is_server = 1;
    global.pktlen = 1;       /* 10000 bits */

    /* cmd line options */
    main_parse_cmdline(argc, argv);

    /* install SIGTERM handler */

    res = asignal_set_handler(SIGTERM, 0, &main_signal_handler, NULL);
    logerr_fatal_str(res == ASIGNAL_OK, "failed to install SIGTERM handler");

    if (global.is_server)
        server();
    else
        client();

    return 0;
}


void
main_parse_cmdline(int argc, char **argv)
{
    /*
     * Valid Options:
     *
     * -S                       Server
     * -C                       Client
     * -p <packet-size-bytes>   Packet size used
     */

    static char short_options[] = "SCp:";

    char c = 0;
    char *endptr = NULL;
    long larg = 0;

    /* parse command line options */

    while (1) {

        c = getopt(argc, argv, short_options);

        if (c == EOF)
            break;

        switch (c) {

        case 'S':
            global.is_server = 1;
            logerr_log(0, "I am the server.");
            break;
            
        case 'C':
            global.is_server = 0;
            logerr_log(0, "I am the client.");
            break;
            
        case 'p':
            larg = (long int)strtoul(optarg, &endptr, 0);
            if (endptr == optarg) 
                logerr_abort(0,
                    "Non numeric packet size '%s': expecting (<size> bytes)",
                    optarg);
            if (larg <= 0)
                logerr_abort(0, "Invalid packet size: %ld", larg);
            global.pktlen = larg;
            
            logerr_log(0, "UDP packet size: %d bytes", global.pktlen);
            break;
            
        case '?':

            logerr_abort(0, "failed to parse command line");
            break;
            
        default:
            logerr_abort(0, "do not know hot to process arg '%c'", c);
            break;

        }
    }
}

void
main_done(void)
{
}

void
main_signal_handler(int signo)
{
    switch (signo) {

        case SIGINT:
        case SIGTERM:
            main_done();
            exit(0);
            break;

        default:

            logerr_fatal_str(0, "Do not know how to handle signal");
            break;
    }
}

