#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>

#include "logerr.h"

#define DEF_PATH        "/tmp/ustream.control"
#define DEF_CLIENT      "/tmp/ustream.client.control"

typedef struct global_options {
    int server;
    char path[500];
} global_options;


void
server(global_options *op)
{
    int retry = 3;
    int res = 0;
    int sockfd = -1;
    int connfd = -1;
    int socket_flags;
    int seq = 0;
    fd_set rfds;
    int ready = 0;
    struct sockaddr_un addr;
    unsigned int len = sizeof(addr);
    unsigned char buf[100];

    logerr_log(1, "Server");

    sockfd = socket(AF_LOCAL, SOCK_STREAM, 0);
    logerr_fatal_sys(sockfd != -1, "");

    unlink(op->path);
    memset(&addr, 0, sizeof(addr));
    addr.sun_family = AF_LOCAL;
    strncpy(addr.sun_path, op->path, sizeof(addr.sun_path));

    res = bind(sockfd, (struct sockaddr *)&addr, sizeof(addr));
    logerr_fatal_sys(res != -1, "");
    
    res = listen(sockfd, 0);
    logerr_fatal_sys(res != -1, "");

    while (retry > 0) {

        len = sizeof(addr);
        memset(&addr, 0, len);
        connfd = accept(sockfd, (struct sockaddr *)&addr, &len);

        if (connfd == -1 && errno == EINTR) {
            retry--;
            logerr_log(1, "EINTR caught. Retrying for %d more times", retry);

        } else {
            socket_flags = fcntl(connfd, F_GETFL, 0);
            logerr_fatal_sys(socket_flags != -1, "");

            res = fcntl(connfd, F_SETFL, socket_flags | O_NONBLOCK);
            logerr_fatal_sys(res != -1, "");

            break;
        }
    }

    logerr_fatal_sys(connfd != -1, "");

    logerr_log(1, "Accepted connection from len(%d) orig(%d), %s",
        len, sizeof(addr), addr.sun_path);

    seq = 0;
    while (1) {
        
        FD_ZERO(&rfds);
        FD_SET(connfd, &rfds);

        ready = select(connfd+1, &rfds, NULL, NULL, NULL);

        if (ready < 1)
            continue;

        res = read(connfd, buf, sizeof(buf));
        logerr_fatal_sys(res >= 0, "read error");
        
        if (res == 0) {
            logerr_log(1, "Peer closed");
            break;
        }

        seq += res;
        logerr_log(1, "read bytes(%d) seq(%d)", res, seq);

    } 

    res = close(sockfd);
    logerr_fatal_sys(res != -1, "");
}

void
client(global_options *op)
{
    int i = 0;
    int seq = 0;
    int res = 0;
    int sockfd = -1;
    struct sockaddr_un addr;
    unsigned char buf[1000];

    logerr_log(1, "Client");

    memset(buf, 0xf0, sizeof(buf));

    sockfd = socket(PF_LOCAL, SOCK_STREAM, 0);
    logerr_fatal_sys(sockfd != -1, "");


    unlink(DEF_CLIENT);

    memset(&addr, 0, sizeof(addr));
    addr.sun_family = AF_LOCAL;
    strncpy(addr.sun_path, DEF_CLIENT, sizeof(addr.sun_path));

    res = bind(sockfd, (struct sockaddr *)&addr, sizeof(addr));
    logerr_fatal_sys(res != -1, "");
    
    res = unlink(DEF_CLIENT);
    logerr_log(0, "Unlinked client(%s) res(%d)", DEF_CLIENT, res);

    memset(&addr, 0, sizeof(addr));
    addr.sun_family = AF_LOCAL;
    strncpy(addr.sun_path, op->path, sizeof(addr.sun_path));

    res = connect(sockfd, (struct sockaddr *)&addr, sizeof(addr));
    logerr_fatal_sys(res != -1, "");

    logerr_log(1, "Established connection.");

    seq = 0;
    for (i = 0; i < 10; i++) {

        res = write(sockfd, buf, sizeof(buf));
        logerr_fatal_sys(res >= 0, "write error");

        seq += res;
        logerr_log(1, "wrote bytes(%d) seq(%d)", res, seq);

        sleep(1);
    }

    res = close(sockfd);
    logerr_fatal_sys(res != -1, "");
}

void
main_parse_cmdline(global_options *op, int argc, char **argv)
{
    static char short_options[] = "scp:";
    char c = 0;

    logerr_passert(op);

    memset(op, 0, sizeof(global_options));

    /* initialize options */

    op->server = -1;
    strncpy(op->path, DEF_PATH, sizeof(op->path));

    /* parse command line options */

    while (1) {

        c = getopt(argc, argv, short_options);

        if (c == EOF)
            break;

        switch (c) {

        case 's':
            op->server = 1;
            break;

        case 'c':
            op->server = 0;
            break;

        case 'p':

            strncpy(op->path, optarg, sizeof(op->path));
            logerr_log(1, "Using pathname '%s'", optarg);

            break;

        case '?':

            logerr_abort(0, "failed to parse command line");
            break;
            
        default:
            logerr_abort(0, "do not know hot to process arg '%c'", c);
            break;

        }
    }
    
}

int
main (int argc, char **argv)
{
    global_options options;

    logerr_log(1, "PID(%d)", getpid());

    main_parse_cmdline(&options, argc, argv);

    if (options.server == 1)
        server(&options);
    else if (options.server == 0)
        client(&options);

    exit (0);
}


