/*
 * Created on   : Wed Jul  7 14:31:25 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * ipc.c
 *
 * Interprocess Communication Utilities
 *
 * $Id: ipc.c,v 1.12 2004/09/15 23:44:07 pavlos Exp $
 */


#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <signal.h>


#include "ipc.h"
#include "logerr.h"
#include "atimer.h"

static struct {
    int sighup_received;
} ipc_glb = {0};

static int ipc_copy_vector(ipc_channel *c, const struct iovec *vector,
                int count);
static int ipc_read_next_fragment_header(ipc_channel *c);
static void ipc_header_log(int level, ipc_header *h);

static void ipc_sighup_handler(int signo);


int
ipc_create_fifo(const char *pathname, mode_t mode, int flags)
{
    int res = 0;
    int can_exist = bit_test(flags, IPC_REPLACE|IPC_USE_EXISTING);
    mode_t orig_mode = (mode_t)-1;

    void cleanup(void)
    {
        if (orig_mode != (mode_t)-1)
            umask(orig_mode);
    }

    logerr_passert(pathname);

    orig_mode = umask(0000);

    res = mkfifo(pathname, mode);
    logerr_creturn_sys(
                res != -1 || (errno == EEXIST && can_exist),
                "Failed to create fifo",
                IPC_EMKFIFO, cleanup);

    if (res == -1 && bit_test(flags, IPC_REPLACE)) {

        res = unlink(pathname);
        logerr_creturn_sys(res != -1,
                    "Failed to unlink fifo",
                    IPC_ERMFIFO, cleanup);

        res = mkfifo(pathname, mode);
        logerr_creturn_sys(res != -1,
                    "Failed to create fifo",
                    IPC_EMKFIFO, cleanup);
    }

    umask(orig_mode);

    return IPC_OK;
}


int
ipc_init(ipc_channel *c)
{
    void cleanup(void)
    {
        logerr_assert_str(0, "cleanup not implemented");
    }

    logerr_passert(c);


    c->flags = 0;
    c->vector = NULL;
    c->vcount = 0;

    c->read.fd = -1;
    c->read.dummyfd = -1;
    c->read.atomic_bytes = 0;
    c->read.malloc_func = NULL;

    memset(&c->read.msg_header, 0, sizeof(ipc_header));
    memset(&c->read.frg_header, 0, sizeof(ipc_header));
    c->read.msg_remaining = 0;
    c->read.frg_remaining = 0;

    c->write.fd = -1;
    c->write.atomic_bytes = 0;
    c->write.id = 0;

    return IPC_OK;
}


int
ipc_done(ipc_channel *c)
{
    int res = 0;
    int what = 0;

    if (bit_test(c->flags, IPC_READ|IPC_READ_OPEN))
        what |= IPC_READ;
    
    if (bit_test(c->flags, IPC_WRITE|IPC_WRITE_OPEN))
        what |= IPC_WRITE;

    res = ipc_close(c, what);
    logerr_return_str(res == IPC_OK, "Failed to close sub channels", res);

    if (c->vector != NULL) {
        free(c->vector);
        c->vector = NULL;
        c->vcount = 0;
    }

    c->flags = 0;

    return IPC_OK;
}


int
ipc_open_read_fifo(ipc_channel *c, const char *pathname, int block)
{
    int open_flags = 0;
    int dummy_child = -1;
    int dummy_status = -1;
    int dummy_exit_status = -1;
    int res = 0;

    void cleanup(void)
    {
        logerr_assert_str(0, "cleanup not implemented");
    }

    logerr_passert(c);
    logerr_passert(pathname);

    c->flags |= IPC_READ;

    /* open read end of fifo */

    open_flags |= O_RDONLY;
    open_flags |= (block) ? 0 : O_NONBLOCK;

    logerr_log(4, "Opening read fifo [read end]: %s block(%d)",
                pathname, block);

    /* fork a child and open read channel for writing (blocking) */
    /* This way, the following open in the parent won't block */

    dummy_child = fork();
    logerr_creturn_sys(dummy_child != -1,
                "Failed to fork child for dummy open",
                IPC_EFORK, cleanup);

    if (dummy_child == 0) {
        /* this is the dummy child */
        
        int dres = 0;
        ipc_channel dserver;

        logerr_set_idstr("child");

        dres = asignal_set_handler(SIGHUP, SA_RESTART,
                        &ipc_sighup_handler, NULL);
        
        logerr_fatal_str(dres == ASIGNAL_OK,
                "Failed to set the SIGHUP handler");

        dres = ipc_init(&dserver);

        if (dres != IPC_OK) {
            logerr_abort(1,
                "child[%d]: failed to initialize ipc channel",
                getpid(), pathname);
        }

        /* this should block until parent opens for read */
        dres = ipc_open_write_fifo(&dserver, pathname, 1, 0);

        if (dres != IPC_OK) {
            logerr_abort(1,
                "child[%d]: failed to open write end of fifo %s",
                getpid(), pathname);
        }

        while (!ipc_glb.sighup_received)
            atimer_sleep_msec(100);

        dres = ipc_done(&dserver);
        logerr_log(4, "Closing write fd");

        if (dres != IPC_OK) {
            logerr_abort(1,
                "child[%d]: failed to close write end of fifo %s",
                getpid(), pathname);
        }

        logerr_log(4, "Exiting. My job is done");
        exit(0);
    }

    /* back to the parent */

    c->read.fd = open(pathname, open_flags);
    logerr_creturn_sys(c->read.fd != -1,
                "Failed to open fifo for reading",
                IPC_EOPEN, cleanup);

    logerr_log(4, "    --> Opened");

    /* open non-blocking write end of fifo */

    logerr_log(4, "Opening read fifo [dummy write end]: %s block(%d)",
                pathname, 0);

    c->read.dummyfd = open(pathname, O_NONBLOCK|O_WRONLY);
    logerr_creturn_sys(c->read.dummyfd != -1,
                "Failed to open fifo for writing",
                IPC_EOPEN, cleanup);

    logerr_log(4, "    --> Opened");

    c->flags |= IPC_READ_OPEN;

    errno = 0;
    c->read.atomic_bytes = fpathconf(c->read.fd, _PC_PIPE_BUF);
    logerr_creturn_sys(c->read.atomic_bytes != -1,
                "Failed to get PIPE_BUF for read fifo",
                IPC_EPIPEBUF, cleanup);

    logerr_log(4, "    --> Atomic bytes (%d)", c->read.atomic_bytes);

    /* terminate and reap the child process */

    res = kill(dummy_child, SIGHUP);
    logerr_fatal_sys(res != -1,
                "Failed to send SIGHUP to child process");

    res = waitpid(dummy_child, &dummy_status, 0);

    logerr_creturn_sys(res != -1,
                "Failed while waiting for dummy child process",
                IPC_EWAITPID, cleanup);

    dummy_exit_status = WEXITSTATUS(dummy_status);
    logerr_creturn_str(dummy_exit_status == 0,
                "Dummy child process returned non-zero error code",
                IPC_EWAITPID, cleanup);

    return IPC_OK;
}


int
ipc_open_write_fifo(ipc_channel *c, const char *pathname,
                int block, uint16_t writeid)
{
    int open_flags = 0;

    void cleanup(void)
    {
        logerr_assert_str(0, "cleanup not implemented");
    }

    logerr_passert(c);
    logerr_passert(pathname);

    c->flags |= IPC_WRITE;

    open_flags |= O_WRONLY;
    open_flags |= (block) ? 0 : O_NONBLOCK;


    logerr_log(4, "Opening write fifo [write end]: %s block(%d)",
                pathname, block);

    c->write.fd = open(pathname, open_flags);
    logerr_creturn_sys(c->write.fd != -1,
                    "Failed to open fifo for writing",
                    IPC_EOPEN, cleanup);

    c->flags |= IPC_WRITE_OPEN;

    logerr_log(4, "    --> Opened");

    errno = 0;
    c->write.atomic_bytes = fpathconf(c->write.fd, _PC_PIPE_BUF);
    logerr_creturn_sys(c->write.atomic_bytes != -1,
                "Failed to get PIPE_BUF for write fifo",
                IPC_EPIPEBUF, cleanup);

    logerr_log(4, "    --> Atomic bytes (%d)", c->write.atomic_bytes);

    c->write.id = writeid;

    return IPC_OK;
}


int
ipc_close(ipc_channel *c, int what)
{
    int res = 0;

    if (bit_test(what, IPC_WRITE)) {
    
        logerr_return_str(bit_test(c->flags, IPC_WRITE|IPC_WRITE_OPEN),
                        "Write subchannel not open",
                        IPC_EWNOTOPEN);

        res = close(c->write.fd);
        logerr_assert_sys(res != -1, "syscall failed");

    }

    if (bit_test(what, IPC_READ)) {
    
        logerr_return_str(bit_test(c->flags, IPC_READ|IPC_READ_OPEN),
                        "Read subchannel not open",
                        IPC_ERNOTOPEN);

        res = close(c->read.fd);
        logerr_assert_sys(res != -1, "syscall failed");

        res = close(c->read.dummyfd);
        logerr_assert_sys(res != -1, "syscall failed");

    }

    return IPC_OK;
}


int
ipc_send(ipc_channel *c, void *buf, size_t len)
{
    int res = 0;

    struct iovec vector;

    vector.iov_base = buf;
    vector.iov_len  = len;

    res = ipc_sendv(c, &vector, 1);
    return res;
}


int
ipc_sendv(ipc_channel *c, const struct iovec *vector, int count)
{
    int i = 0;
    int res = 0;
    int from = 0;
    ipc_header ihead;
    uint32_t total_msg_len = 0;
    int is_fragmented = 0;
    int frg_count = 0;
    fd_set wfds;

    logerr_passert(c);
    logerr_passert(vector);
    logerr_assert_str(count > 0, "Invalid vector count");
    logerr_assert_str(bit_test(c->flags, IPC_WRITE|IPC_WRITE_OPEN),
                "Write sub-channel not open");

    /* copy input vector to channel vector starting at pos [1 .. count] */

    res = ipc_copy_vector(c, vector, count);
    logerr_return_str(res == IPC_OK, "Failed to copy buffer vector", res);

    /* pos 0 is reserved for the ipc_header buffer */
    from = 1;

    for (i = from; i <= count; i++) {
        total_msg_len += c->vector[i].iov_len;
    }

    logerr_log(999, "ipc_sendv: total_msg_len(%d)", total_msg_len);

    is_fragmented = total_msg_len + sizeof(ipc_header) > c->write.atomic_bytes;
    frg_count = 0;

    /* repeat until all vectors are sent */
    while (from <= count) {

        /* each while loop will generate one writev on the ipc channel */

        size_t max_available = c->write.atomic_bytes - sizeof(ipc_header);
        size_t cur_available = max_available;
        size_t upto_left_bytes = 0;

        int upto = from;
        int is_last = 0;
        int is_first = 0;
        int hpos = 0;

        /* find the last index that fits into one atomic write */

        i = from;
        while (i <= count) {

            if (cur_available < c->vector[i].iov_len) {
                upto_left_bytes = c->vector[i].iov_len - cur_available;
                c->vector[i].iov_len = cur_available;
                cur_available = 0;
                upto = i;
                break;
            }

            cur_available -= c->vector[i].iov_len;
            upto = i;
            i++;
        }
        
        /* construct the ipc_header (position from-1 of the buffer vector) */

        logerr_assert_str(upto <= count, "Loop bug");

        frg_count++;

        is_first = (frg_count == 1);
        is_last  = (upto_left_bytes == 0 && upto == count);

        if (is_first) {
            /* initialize ipc_header */
            ihead.id            = c->write.id;
            ihead.total_len     = total_msg_len;
            ihead.offset        = 0;
            ihead.len           = 0;
        }

        if (is_fragmented) {
            logerr_assert_str(ihead.offset < IPC_LAST_OFFSET,
                        "Reached the last fragment offset");

            ihead.offset = (is_last) ? IPC_LAST_OFFSET : ihead.offset + 1;
        }

        ihead.len = max_available - cur_available;

        hpos = from - 1;
        logerr_assert_str(hpos >= 0 && hpos < count, "vector index bug");

        c->vector[hpos].iov_base = (void *)&ihead;
        c->vector[hpos].iov_len  = sizeof(ipc_header);

        logerr_log(999,
                "ipc_sendv: sending msg: fragmented(%d) first(%d) last(%d)",
                is_fragmented, is_first, is_last);
        ipc_header_log(999, &ihead);


        /* wait for the file descriptor to be ready for writing */

        logerr_log(991, "waiting in select");

        FD_ZERO(&wfds);
        FD_SET(c->write.fd, &wfds);

        logerr_return_rsys(
                (res =
                select(c->write.fd+1, NULL, &wfds, NULL, NULL)) != -1,
                "Failed to wait for file descriptor",
                errno);

         logerr_return_str(res == 1,
                "File descriptor not ready for writing",
                IPC_ESELECT);

        /* issue the writev command */

        logerr_log(999, "before: from(%d) upto(%d)", from, upto);
        logerr_return_rsys(
                (res =
                writev(c->write.fd, c->vector + hpos, upto - hpos + 1)) != -1,
                "Failed to writev buffer vector",
                errno);

        if (upto_left_bytes > 0) {

            /* adjust the last buffer vector that was partially sent */

            c->vector[upto].iov_base =
                (char *)c->vector[upto].iov_base + c->vector[upto].iov_len;
            c->vector[upto].iov_len = upto_left_bytes;

            upto_left_bytes = 0;
            from = upto;

        } else {
            from = upto + 1;
        }
        logerr_log(999, "after: from(%d) upto(%d)", from, upto);
    }

    return IPC_OK;
}


int
ipc_recv(ipc_channel *c, void *buf, size_t len)
{
    int res = 0;

    struct iovec vector;

    vector.iov_base = buf;
    vector.iov_len  = len;

    res = ipc_recvv(c, &vector, 1, NULL);
    return res;
}


/*
 *  + If last buffer in vector is NULL then allocate appropriately
 *    to hold the rest of the record.
 *  + Caller assumes memory ownership and has to free.
 *  + Each record can be fragmented in multiple pieces.
 *  + If no NULL buffer then just return as much as possible.
 *    Next read will return the rest.
 */
int
ipc_recvv(ipc_channel *c, struct iovec *vector, int count, void *marg)
{
    int res = 0;
    int total_bytes = 0;
    int bytes = 0;
    int recv_buf_needed = 0;
    int recv_buf_available = 0;
    int first = 0;
    int last = 0;
    int i = 0;
    int frg_first = -1;
    int frg_last = -1;
    size_t frg_last_extra_len  = 0;
    void *frg_last_extra_base  = NULL;


    logerr_passert(c);
    logerr_passert(vector);
    logerr_assert_str(count > 0, "Invalid vector count");
    logerr_assert_str(bit_test(c->flags, IPC_READ|IPC_READ_OPEN),
                "Read sub-channel not open");

    /* make sure that the first header is read (for size) */
    res = ipc_read_next_fragment_header(c);
    if (res != IPC_OK)
        return res;

    /* calculate available space in user buffers and last index */


    recv_buf_available = 0;
    for (i = 0; i < count; i++) {
        recv_buf_available += vector[i].iov_len;

        last = i;

        if (vector[i].iov_base == NULL) {
            logerr_assert_str(vector[i].iov_len == 0, "malformed length");
            break;
        }
    }

    logerr_return_str(last == count-1,
                "At most one NULL buffer allowed in the buffer vector",
                IPC_EVECNULL);

    /* allocate extra space if requested by user */

    recv_buf_needed = c->read.msg_remaining - recv_buf_available;
    if (recv_buf_needed > 0 && vector[last].iov_base == NULL) {

        void *buf = NULL;

        /* use the user-specified allocation function if present */
        if (c->read.malloc_func == NULL)
            buf = malloc(recv_buf_needed);
        else
            buf = c->read.malloc_func(marg, recv_buf_needed);

        logerr_return_sys(buf != NULL,
                        "Failed to allocate buffer",
                        IPC_ENOMEM);

        logerr_log(999, "ipc_recvv: allocated buffer(%p) size(%d) vector(%d)",
                buf, recv_buf_needed, last);
                        
        vector[last].iov_base = buf;
        vector[last].iov_len  = recv_buf_needed;
        recv_buf_available += vector[last].iov_len;
        recv_buf_needed = 0;
    }

    /* copy input vector to channel vector starting at pos [1 .. count] */

    res = ipc_copy_vector(c, vector, count);
    logerr_return_str(res == IPC_OK, "Failed to copy buffer vector", res);

    /* in the copy, user specified buffers start at index 1 NOT at index 0 */

    first = 1;
    last  = count;

    /* read as many fragments as you can in the user buffers */

    frg_first = first - 1;
    frg_last  = first - 1;
    frg_last_extra_len  = 0;

    while (recv_buf_available > 0) {

        /* each while loop will generate one readv on the ipc channel */

        int frg_buf_available = 0;
        int frg_buf_needed = 0;

        /* reclaim any space left to the previous last buffer or use next */
        if (frg_last_extra_len > 0) {

            c->vector[frg_last].iov_len  = frg_last_extra_len;
            c->vector[frg_last].iov_base = frg_last_extra_base;

            frg_first = frg_last;

        }  else {
            frg_first = frg_last+1;
        }


        /* make sure buffer vector has space for at most as much as fragment */

        frg_last  = last;
        logerr_assert_str(frg_first <= frg_last, "bug");
        for (i = frg_first; i <= frg_last; i++) {
            frg_buf_needed = c->read.frg_remaining - frg_buf_available;

            if (frg_buf_needed > c->vector[i].iov_len) {
                frg_buf_available +=  c->vector[i].iov_len;

            } else {

                /*
                 * We  must truncate current buffer and set as last.
                 * because we should not read more bytes and step over
                 * next fragment header.
                 *
                 * This is a copy; we are not messing it up for caller
                 * and we are reclaiming the space at next iteration.
                 */

                frg_last_extra_len  =  c->vector[i].iov_len - frg_buf_needed;
                frg_last_extra_base =  (char *)c->vector[i].iov_base
                        + frg_buf_needed;

                c->vector[i].iov_len = frg_buf_needed;
                frg_buf_available += c->vector[i].iov_len;
                frg_buf_needed = 0;
                frg_last = i;
                break;
            }
        }

        logerr_log(999, "ipc_recvv: reading next fragment");

        logerr_return_rsys(
                (bytes =
                readv(c->read.fd, c->vector + frg_first,
                                frg_last - frg_first + 1))
                != -1 || errno == EAGAIN,
                "Failed to read fragment header",
                IPC_EREAD);

        if (bytes == -1 && errno == EAGAIN)
            return IPC_EAGAIN;

        logerr_fatal_str(bytes == frg_buf_available,
                "Did not read expected number of bytes in fragment");

        total_bytes += bytes;

        c->read.msg_remaining -= bytes;
        c->read.frg_remaining -= bytes;
        recv_buf_available -= bytes;

        logerr_assert_str(c->read.msg_remaining >= 0, "bug");
        logerr_assert_str(c->read.frg_remaining >= 0, "bug");

        if (c->read.msg_remaining == 0)
            return total_bytes;

        res = ipc_read_next_fragment_header(c);
        if (res != IPC_OK)
            return res;
    }

    return total_bytes;
}


static int
ipc_read_next_fragment_header(ipc_channel *c)
{
    int bytes = 0;
    int fragmented = 0;
    int first_fragment = 0;
    int last_fragment = 0;

    if (c->read.msg_remaining == 0) {
        /* all fragments of previous msg done */

        /* read the next message header */
        logerr_return_rsys(
                (bytes =
                read(c->read.fd, &c->read.msg_header, sizeof(ipc_header)))
                != -1 || errno == EAGAIN,
                "Failed to read msg header",
                IPC_EREAD);
        
        if (bytes == -1 && errno == EAGAIN)
            return IPC_EAGAIN;

        logerr_log(999, "ipc_read_next_fragment_header: reading msg header");
        ipc_header_log(999, &c->read.msg_header);

        logerr_fatal_str(bytes == sizeof(ipc_header),
                        "Read partial message header");

        fragmented     = c->read.msg_header.offset > 0;
        first_fragment = c->read.msg_header.offset == 1;
        last_fragment  = bit_match(c->read.msg_header.offset, IPC_LAST_OFFSET);

        logerr_fatal_str(!fragmented || first_fragment,
                        "Invalid offset for first fragment");

        memcpy(&c->read.frg_header, &c->read.msg_header, sizeof(ipc_header));
        c->read.msg_remaining = c->read.msg_header.total_len;
        c->read.frg_remaining = c->read.frg_header.len;

        logerr_fatal_str(c->read.msg_remaining > 0,
                        "Empty message");
    }

    if (c->read.frg_remaining == 0) {
        /* previous msg fragment done, i.e. all fragment data read */

        logerr_fatal_str(c->read.frg_header.offset != IPC_LAST_OFFSET,
                        "Cannot read past the last fragment");

        /* read the next fragment header */
        logerr_return_rsys(
                (bytes =
                read(c->read.fd, &c->read.frg_header, sizeof(ipc_header)))
                != -1 || errno == EAGAIN,
                "Failed to read fragment header",
                IPC_EREAD);
        
        if (bytes == -1 && errno == EAGAIN)
            return IPC_EAGAIN;

        logerr_log(999, "ipc_read_next_fragment_header: reading frg header");
        ipc_header_log(999, &c->read.frg_header);

        logerr_fatal_str(bytes == sizeof(ipc_header),
                        "Read partial fragment header");

        logerr_fatal_str(c->read.frg_header.offset > 1,
                        "Invalid offset for subsequent fragment");

        c->read.frg_remaining = c->read.frg_header.len;
    }

    logerr_fatal_str(c->read.frg_remaining > 0,
                "Empty fragment");

    return IPC_OK;
}


static int
ipc_copy_vector(ipc_channel *c, const struct iovec *vector, int count)
{
    int required = count + 1;

    logerr_passert(c);
    logerr_passert(vector);
    logerr_assert_str(count > 0, "Invalid vector count");

    if (c->vector != NULL && c->vcount < required) {
        free(c->vector);
        c->vector = NULL;
        c->vcount = 0;
    }

    if (c->vector == NULL) {

        c->vector = malloc(sizeof(struct iovec) * required);
        logerr_return_sys(c->vector != NULL,
                        "Failed to allocate new buffer vector array",
                        IPC_ENOMEM);

        c->vcount = required;
    }

    /* first position is reserved for ipc_header buffer */

    memcpy((c->vector + 1), vector, sizeof(struct iovec) * count);

    return IPC_OK;
}

static void
ipc_header_log(int level, ipc_header *h)
{
    logerr_log(level, "IPC HEADER");
    logerr_log(level, "    id(%d)", h->id);
    logerr_log(level, "    offset(%x)", h->offset);
    logerr_log(level, "    len(%d)", h->len);
    logerr_log(level, "    total_len(%d)", h->total_len);
}

static void
ipc_sighup_handler(int signo)
{
    ipc_glb.sighup_received = 1;
}
