/* Created on   : Wed Jul  7 14:31:25 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * ipc.h
 *
 * Interprocess Communication Utilities
 *
 * $Id: ipc.h,v 1.13 2005/11/02 00:19:17 pavlos Exp $
 */

#ifndef __IPC_H
#define __IPC_H

#include <stdio.h>
#include <sys/uio.h>

#include "bytes.h"

/* positive return codes reserved for errno */

#define IPC_OK                  0
#define IPC_ERROR              -1
#define IPC_ENOMEM             -2
#define IPC_EMKFIFO            -3
#define IPC_ERMFIFO            -3
#define IPC_EPIPEBUF           -4
#define IPC_EOPEN              -5
#define IPC_EWNOTOPEN          -6
#define IPC_ERNOTOPEN          -7
#define IPC_EREAD              -8
#define IPC_EWRITE             -9
#define IPC_EAGAIN             -10
#define IPC_EVECNULL           -11
#define IPC_EFORK              -12
#define IPC_EWAITPID           -13
#define IPC_ESELECT            -14

#define IPC_LAST_OFFSET         0xFFFF

#define IPC_BLOCK               1
#define IPC_NONBLOCK            0

#define IPC_CREATE              0x0001
#define IPC_REPLACE             0x0002
#define IPC_USE_EXISTING        0x0004

#define IPC_READ                0x0001
#define IPC_WRITE               0x0002
#define IPC_RW                  (IPC_READ|IPC_WRITE)
#define IPC_READ_OPEN           0x1000
#define IPC_WRITE_OPEN          0x2000

#define IPC_PATH_MAX            108

typedef void * (*ipc_malloc_func)       (void *, size_t);
typedef void (*ipc_free_func)           (void *, void *);

/*
 * offset       0       : msg not fragmented
 *              1..FFFE : msg fragmented and this is intermediate fragment
 *              FFFF    : msg fragmented and this is last fragment
 *
 * len          data length (not header) of current fragment
 *
 * total_len    data length (not header) of entire msg
 */ 
typedef struct ipc_header {

    uint16_t id;
    uint16_t offset;
    uint16_t len;
    uint16_t __pad; /* useless */
    uint32_t total_len;

} ipc_header;

typedef struct ipc_read_channel {

    int fd;
    int dummyfd;
    size_t atomic_bytes;
    ipc_malloc_func malloc_func;

    ipc_header msg_header;
    ipc_header frg_header;
    uint32_t msg_remaining;
    uint32_t frg_remaining;

} ipc_read_channel;

typedef struct ipc_write_channel {

    int fd;
    size_t atomic_bytes;
    uint16_t id;
    uint16_t __pad; /* useless */

} ipc_write_channel;

typedef struct ipc_channel {

    int flags;
    struct iovec *vector;
    int vcount;

    ipc_read_channel read;
    ipc_write_channel write;

} ipc_channel;

int ipc_create_fifo(const char *pathname, mode_t mode, int flags);

int ipc_init(ipc_channel *c);
int ipc_done(ipc_channel *c);

int ipc_open_read_fifo(ipc_channel *c, const char *pathname, int block);

int ipc_open_write_fifo(ipc_channel *c, const char *pathname,
                int block, uint16_t writeid);

int ipc_close(ipc_channel *c, int what);

int ipc_send(ipc_channel *c, void *buf, size_t len);
int ipc_sendv(ipc_channel *c, const struct iovec *iov, int iovcnt);

int ipc_recv(ipc_channel *c, void *buf, size_t len);
int ipc_recvv(ipc_channel *c, struct iovec *vector, int count, void *marg);

#endif /* __IPC_H */

