/*
 * Created on   : Tue Jun 15 15:04:04 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * logerr.h
 *
 * Error logging routines
 *
 * $Id: logerr.h,v 1.22 2005/12/12 16:39:56 pavlos Exp $
 */

#ifndef __LOGERR_H
#define __LOGERR_H

#include <errno.h>
#include <string.h>

#include "stdio.h"
#include "bytes.h"

#define LOGERR_MAX_IDSTR_LEN    52

#define LOGERR_FIRST            0x01
#define LOGERR_MIDDLE           0x02
#define LOGERR_LAST             0x04
#define LOGERR_SINGLE           LOGERR_FIRST | LOGERR_LAST

#define LOGERR_STR              0x8000
#define LOGERR_ERRNO            0x0100
#define LOGERR_RETRY_EINTR      0x0101

#define LOGERR_SNULL_PTR        "Unexpected Null Pointer"

#define LOGERR_OFFSET           10000
#define logerr_extract_errno(e) (abs(e) - LOGERR_OFFSET)
#define logerr_encode_errno(e)  (0 - LOGERR_OFFSET - (e))

/* we need to export this global struct so that macros can access it */
/* performance considerations */

typedef struct logerr_global_state {
	int  min_level;
	int  max_level;
	int  idlen;
	char idstr[LOGERR_MAX_IDSTR_LEN];
} logerr_global_state;

extern logerr_global_state logerr_globals;

#define logerr(EXP,FLAGS,PTR) \
    do { \
        while (!(EXP)) { \
            if (logerr_aux((FLAGS), (PTR), #EXP, __FILE__, __LINE__)) \
                break; \
        } \
    } while (0);

#define logerr_return(EXP,FLAGS,PTR,RVAL) \
    do { \
        while (!(EXP)) { \
            if (logerr_aux((FLAGS), (PTR), #EXP, __FILE__, __LINE__)) \
                return (RVAL); \
        } \
    } while (0);

#define logerr_creturn(EXP,FLAGS,PTR,RVAL,CLEANUP) \
    do { \
        while (!(EXP)) { \
            if (logerr_aux((FLAGS), (PTR), #EXP, __FILE__, __LINE__)) { \
                CLEANUP(); \
                return (RVAL); \
            } \
        } \
    } while (0);

#define logerr_fatal(EXP,FLAGS,PTR) \
    do { \
        while (!(EXP)) { \
            if (logerr_aux((FLAGS), (PTR), #EXP, __FILE__, __LINE__)) \
                logerr_fatal_aux(); \
        } \
    } while (0);

/* XXX  logerr_assert* can be called as a function */
/*      returns 0 if no assert (i.e., if valid) */

#ifdef NDEBUG

#define logerr_assert(EXP,FLAGS,PTR)    ((void)0)

#else  /* NDEBUG */

#define logerr_assert(EXP,FLAGS,PTR) \
    logerr_assert_aux((EXP), (FLAGS), (PTR), #EXP, __FILE__, __LINE__)

#endif /* NDEBUG */

/* Using error string */

#define logerr_str(EXP,STR) \
        logerr(EXP,LOGERR_STR,STR)

#define logerr_return_str(EXP,STR,RVAL) \
        logerr_return(EXP,LOGERR_STR,STR,RVAL)

#define logerr_creturn_str(EXP,STR,RVAL,CLEANUP) \
        logerr_creturn(EXP,LOGERR_STR,STR,RVAL,CLEANUP)

#define logerr_fatal_str(EXP,STR) \
        logerr_fatal(EXP,LOGERR_STR,STR)

#define logerr_assert_str(EXP,STR) \
        logerr_assert(EXP,LOGERR_STR,STR)

/* System Call - Use errno to get error string */

#define logerr_sys(EXP,STR) \
        logerr(EXP,LOGERR_ERRNO|LOGERR_STR,STR)

#define logerr_return_sys(EXP,STR,RVAL) \
        logerr_return(EXP,LOGERR_ERRNO|LOGERR_STR,STR,RVAL)

#define logerr_creturn_sys(EXP,STR,RVAL,CLEANUP) \
        logerr_creturn(EXP,LOGERR_ERRNO|LOGERR_STR,STR,RVAL,CLEANUP)

#define logerr_assert_sys(EXP,STR) \
        logerr_assert(EXP,LOGERR_ERRNO|LOGERR_STR,STR)

#define logerr_fatal_sys(EXP,STR) \
        logerr_fatal(EXP,LOGERR_ERRNO|LOGERR_STR,STR)

/* System Call - Retry while EINTR; otherwise use errno to get error string */

#define logerr_rsys(EXP,STR) \
        logerr(EXP,LOGERR_RETRY_EINTR|LOGERR_STR,STR)

#define logerr_return_rsys(EXP,STR,RVAL) \
        logerr_return(EXP,LOGERR_RETRY_EINTR|LOGERR_STR,STR,RVAL)

#define logerr_creturn_rsys(EXP,STR,RVAL,CLEANUP) \
        logerr_creturn(EXP,LOGERR_RETRY_EINTR|LOGERR_STR,STR,RVAL,CLEANUP)

#define logerr_fatal_rsys(EXP,STR) \
        logerr_fatal(EXP,LOGERR_RETRY_EINTR|LOGERR_STR,STR)

#define logerr_passert(PTR) \
        logerr_assert_str((PTR) != NULL, "Invalid pointer")

#define logerr_bug(EXP) \
	logerr_assert_str((EXP), "Bug in the code")

void logerr_set_idstr(const char *str);

int logerr_aux(uint16_t flags, const void *p, const char *source,
                const char *file, unsigned int line);

int logerr_assert_aux(int valid, uint16_t flags, const void *p,
                const char *source, const char *file, unsigned int line);

void logerr_fatal_aux(void);


#define logerr_loglevel_permits(LEVEL) \
        (logerr_globals.min_level <= (LEVEL) \
            && (LEVEL) <= logerr_globals.max_level)

#define logerr_log(LEVEL, FORMAT, ARGS...) \
        do { \
            if (logerr_loglevel_permits((LEVEL))) \
                logerr_log_aux((LEVEL), (LOGERR_SINGLE), \
                    __FILE__,  __LINE__, (FORMAT) , ## ARGS); \
        } while (0)

#define logerr_log_more(LEVEL, POS, FORMAT, ARGS...) \
        do { \
            if (logerr_loglevel_permits((LEVEL))) \
                logerr_log_aux((LEVEL), (POS), \
                    __FILE__, __LINE__, (FORMAT) , ## ARGS); \
        } while (0)

void logerr_log_aux(int level, int pos,
        const char *file, unsigned int line,
        const char *format, ...);

void logerr_abort(int exitcode, const char *format, ...);

void logerr_setlevel(int level);
void logerr_setlevel_only(int level);
void logerr_setlevel_range(int min, int max);

#define bug() \
        logerr_assert_str(0, "BUG!")

#define bug_on(EXP) \
        logerr_assert_str(!(EXP), "BUG!")

#define bug_on_null(EXP) \
        logerr_assert_str((EXP) != NULL, "BUG!")

#define sysbug_on(EXP) \
        logerr_assert_sys(!(EXP), "BUG!: syscall error:")

#define syserr_on(EXP) \
        logerr_assert_sys(!(EXP), "syscall error:")


#endif /* __LOGERR_H */

