/*
 * Created on   : Tue Jul 20 15:28:32 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * mem.c
 *
 * Wrappers for memory allocation functions to keep
 * track of allocated memory.
 * 
 * $Id: mem.c,v 1.7 2004/12/07 22:22:58 pavlos Exp $
 */

#include "mem.h"
#include "logerr.h"
#include "cmp.h"

void *
mem_malloc_aux(size_t size, const char *file, unsigned int line)
{
    void *ptr = malloc(size);
    int oerrno = errno;
    
    if (ptr != NULL) 
        logerr_log(MEM_LOG_LEVEL,
                "%-15s:%4d: Memory Allocated: ptr(%p) bytes(%d)",
                file, line, ptr, size);
    
    else
        logerr_log(MEM_LOG_LEVEL,
                "%-15s:%4d: Memory Allocated: Failed",
                file, line);
    
    errno = oerrno;
    return ptr;
}

void *
mem_calloc_aux(size_t nelem, size_t elsize, const char *file, unsigned int line)
{
    void *ptr = calloc(nelem, elsize);
    int oerrno = errno;
    
    if (ptr != NULL) 
        logerr_log(MEM_LOG_LEVEL,
        "%-15s:%4d: Memory Allocated: ptr(%p) bytes(%d) nelem(%d) elsize(%d)",
                file, line, ptr, nelem*elsize, nelem, elsize);
    
    else
        logerr_log(MEM_LOG_LEVEL,
                "%-15s:%4d: Memory Allocated: Failed",
                file, line);
    
    errno = oerrno;
    return ptr;
}

void
mem_free_aux(void *ptr, const char *file, unsigned int line)
{
    logerr_passert(ptr);

    free(ptr);

    logerr_log(MEM_LOG_LEVEL,
            "%-15s:%4d: Memory Deallocated: ptr(%p)",
            file, line, ptr);

}

void
mem_free_func(void *ptr)
{
    mem_free_aux(ptr, "mem_free_func", 0);
}

int
mem_copy_from_vector(void *buf, size_t size,
                const struct iovec *vector, int count)
{
    size_t available = size;
    size_t copy = 0;
    void *tmp = buf;
    int i = 0;

    logerr_assert_str(buf != NULL || size == 0, "Invalid buffer");
    logerr_assert_str(vector != NULL || count == 0, "Invalid vector buffer");

    for (i = 0; i < count; i++) {

        logerr_assert_str(vector[i].iov_base != NULL || vector[i].iov_len == 0,
                "Invalid vector buffer");

        copy = cmp_min(available, vector[i].iov_len);
        memcpy(tmp, vector[i].iov_base, copy);
        available -= copy;
        if (copy < vector[i].iov_len)
            return MEM_ENOSPACE;
        tmp = bytes_add(tmp, copy);
    }

    return (size - available);
}

int
mem_copy_to_vector(void *buf, size_t size,
                const struct iovec *vector, int count)
{
    size_t remaining = size;
    size_t copy = 0;
    void *tmp = buf;
    int i = 0;

    logerr_assert_str(buf != NULL || size == 0, "Invalid buffer");
    logerr_assert_str(vector != NULL || count == 0, "Invalid vector buffer");

    for (i = 0; i < count; i++) {

        logerr_assert_str(vector[i].iov_base != NULL || vector[i].iov_len == 0,
                "Invalid vector buffer");

        copy = cmp_min(vector[i].iov_len, remaining);
        memcpy(vector[i].iov_base, tmp, copy);
        remaining -= copy;
        tmp = bytes_add(tmp, copy);
    }

    if (remaining > 0)
        return MEM_ENOSPACE;

    return size;
}


