/*
 * Created on   : Tue Jun 29 16:49:17 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * mshared.c
 *
 * Utilities for sharing memory among processes
 * Front-end to POSIX shared memory
 *
 * $Id: mshared.c,v 1.2 2004/06/30 20:13:54 pavlos Exp $
 */

#include "mshared.h"
#include "logerr.h"

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>

static int mshared_init(mshared *mem, char *name);
static int mshared_name_valid(const char *name);

int
mshared_create(mshared *mem, char *name, mode_t mode, size_t size)
{
    int res = 0;
    int created = 0;

    void cleanup()
    {
        if (created) {
            res = mshared_unlink(mem);
            logerr_assert_str(res == MSHARED_OK, "Failed to clean-up");
            created = 0;
        }

        res = mshared_done(mem);
        logerr_assert_str(res == MSHARED_OK, "Failed to clean-up");
    }

    res = mshared_init(mem, name);
    logerr_creturn_str(res == MSHARED_OK,
                "Failed to initialize shared memory structure",
                MSHARED_EINIT, cleanup);

    mem->fd = shm_open(mem->name, O_RDWR|O_CREAT|O_EXCL, mode);
    logerr_creturn_sys(mem->fd != -1,
                "Failed to create shared memory",
                MSHARED_EOPEN, cleanup);

    created = 1;
    mem->open = 1;

    res = ftruncate(mem->fd, size);
    logerr_creturn_sys(res != -1,
                "Failed to resize shared memory",
                MSHARED_ERESIZE, cleanup);

    mem->size = size;

    mem->ptr = mmap(NULL, size, PROT_READ|PROT_WRITE, MAP_SHARED, mem->fd, 0);
    logerr_creturn_sys(res != -1,
                "Failed to map shared memory",
                MSHARED_EMAP, cleanup);

    mem->mapped = 1;

    res = close(mem->fd);
    logerr_creturn_sys(res != -1,
                "Failed to close shared memory file descriptor",
                MSHARED_ECLOSE, cleanup);

    mem->open = 0;

    return MSHARED_OK;
}


int
mshared_open(mshared *mem, char *name, int read_only)
{
    int res = 0;
    int oflag = 0;
    int prot = 0;
    mode_t mode = 0;
    struct stat mstat;

    void cleanup()
    {
        res = mshared_done(mem);
        logerr_assert_str(res == MSHARED_OK, "Failed to clean-up");
    }

    oflag = (read_only) ? O_RDONLY : O_RDWR;
    prot  = (read_only) ? PROT_READ : (PROT_READ|PROT_WRITE);
    mode  = (read_only) ? 0400 : 0600;


    res = mshared_init(mem, name);
    logerr_creturn_str(res == MSHARED_OK,
                "Failed to initialize shared memory structure",
                MSHARED_EINIT, cleanup);

    mem->fd = shm_open(mem->name, oflag, mode);
    logerr_creturn_sys(mem->fd != -1,
                "Failed to open shared memory file descriptor",
                MSHARED_EOPEN, cleanup);

    mem->open = 1;

    res = fstat(mem->fd, &mstat);
    logerr_creturn_sys(res != -1,
                "Failed to get size of shared memory",
                MSHARED_EFSTAT, cleanup);

    mem->size = mstat.st_size;

    mem->ptr = mmap(NULL, mem->size, prot, MAP_SHARED, mem->fd, 0);
    logerr_creturn_sys(res != -1,
                "Failed to map shared memory",
                MSHARED_EMAP, cleanup);

    mem->mapped = 1;

    res = close(mem->fd);
    logerr_creturn_sys(res != -1,
                "Failed to close shared memory file descriptor",
                MSHARED_ECLOSE, cleanup);

    mem->open = 0;

    return MSHARED_OK;
}


int
mshared_unlink(mshared *mem)
{
    int res = 0;

    logerr_passert(mem);
    logerr_passert(mem->name);


    res = shm_unlink(mem->name);
    logerr_return_sys(res != -1,
                "Failed to unlink shared memory",
                MSHARED_EUNLINK);

    mem->unlinked = 1;

    return MSHARED_OK;
}


int
mshared_done(mshared *mem)
{
    int res = 0;

    logerr_passert(mem);

    if (mem->mapped) {
        logerr_passert(mem->ptr);
        res = munmap(mem->ptr, mem->size);
        logerr_return_sys(res != -1,
                    "Failed to unmap shared memory",
                    MSHARED_EUNMAP);
        mem->mapped = 0;
    }

    if (mem->open) {
        res = close(mem->fd);
        logerr_return_sys(res != -1,
                    "Failed to close shared memory file descriptor",
                    MSHARED_ECLOSE);

        mem->open = 0;
    }

    if (mem->name != NULL) {
        free(mem->name);
    }

    return MSHARED_OK;
}


/* Static functions */

static int
mshared_init(mshared *mem, char *name)
{
    size_t size = 0;

    logerr_passert(mem);

    mem->fd = 0;
    mem->size = 0;
    mem->open = 0;
    mem->mapped = 0;
    mem->unlinked = 0;
    mem->ptr = NULL;
    mem->name = NULL;

    if (!mshared_name_valid(name))
        logerr_return_str(0,
                    "Invalid name for shared memory region",
                    MSHARED_ENAME);

    size = strlen(name) + 1;
    mem->name = malloc(size);
    logerr_return_sys(mem->name != NULL,
                "Failed to allocate name string",
                MSHARED_ENOMEM);

    strncpy(mem->name, name, size);

    return MSHARED_OK;
}

static int
mshared_name_valid(const char *name)
{
    if (name == NULL)
        return 0;

    return 1;
}


