/*
 * net.h	net related functions
 *
 * Created:	Mon Nov 28 13:02:03 EST 2005
 *
 * Author:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 */

#ifndef _COMMON_NET_H
#define _COMMON_NET_H

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

/* for protocol headers */
#include <net/ethernet.h>
#include <netinet/ip.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>

/* for host resolution */
#include <stdio.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>

/* libpcap */
#include <pcap.h>

/* ioctl SIOCGIFADDR */
#include <sys/ioctl.h>
#include <net/if.h>

#include "alog.h"
#include "astring.h"
#include "atimer.h"
#include "aptr.h"

#define NET_HOSTNAME_MAX	100

#define NET_PORT_RANDOM		0xffff0000

#define NET_PORT_WELL_KNOWN	0x01000000
#define NET_PORT_REGISTERED	0x02000000
#define NET_PORT_DYNAMIC	0x04000000
#define NET_PORT_NOT_WELL_KNOWN	(NET_PORT_REGISTERED | NET_PORT_DYNAMIC)
#define NET_PORT_ALL		0x0f000000

#define NET_ADDR_SPOOFED	0x10000000
#define NET_ADDR_RANDOM		0x20000000

#define NET_ADDR_UNICAST	0x01000000
#define NET_ADDR_ALL		0x0f000000

static inline uint32_t
net_random_addr(int type)
{
	uint32_t addr = 0;
	uint8_t *octets = (uint8_t *)&addr;

	/* XXX -- expand for other types too */
	BUG_ON(type != NET_ADDR_UNICAST);

	/* http://www.iana.org/assignments/ipv4-address-space */

	/* pick an address between 0.0.0.1 - 223.254.254.254 */

	octets[0] = rand() % 224;
	octets[1] = rand() % 255;
	octets[2] = rand() % 255;
	octets[3] = 1 + rand() % 254;

	/* already at network order */

	return addr;
}

static inline int
net_valid_port(int type, uint16_t port)
{
	/* http://www.iana.org/assignments/port-numbers */

	if ((type & NET_PORT_WELL_KNOWN) && port <= 1023)
		return 1;

	if ((type & NET_PORT_REGISTERED) && port >= 1024 && port <= 49151)
		return 1;

	if ((type & NET_PORT_DYNAMIC) && port >= 49152)
		return 1;
	
	return 0;
}

static inline uint16_t
net_random_port(int type)
{
	uint16_t port;

	do {
		port = random() % 65536;
	} while (!net_valid_port(type, port));

	return port;
}

static inline int
net_addr_empty(struct sockaddr_in *addr)
{
	int empty = (addr->sin_addr.s_addr == 0) &&
		(addr->sin_port == 0) &&
		(addr->sin_family == 0);

	return empty;
}

/* returns boolean success value: 0 failure, 1 success */
static inline int
net_resolve_host(const char *name, struct sockaddr_in *addr,
			char *error, int errlen)
{
	struct hostent *hptr = NULL;
	
	hptr = gethostbyname(name);

	if (hptr == NULL) {
		strncpy(error, hstrerror(h_errno), errlen);
		return 0;
	}

	if (hptr->h_addrtype != AF_INET) {
		snprintf(error, errlen,
			"only IPv4(%d) supported: unsupported type: %d",
			AF_INET, hptr->h_addrtype);
		return 0;
	}

	memset(addr, 0, sizeof(*addr));
	addr->sin_family = AF_INET;
	memcpy(&addr->sin_addr, hptr->h_addr_list[0], sizeof(addr->sin_addr));

	return 1;
}

/* parses <host:port> and resolves host and stores into addr */
static inline int
net_resolve_host_port(const char *hostport, struct sockaddr_in *addr,
			char *error, int errlen)
{
	int found = 0;
	int port = 0;
	char str[200];
	char *name = str;
	char *port_str = NULL;
	char *colon = NULL;

	strncpy(str, hostport, sizeof(str));
	colon = strchr(str, ':');

	if (colon != NULL) {
		*colon = 0;
		port_str = colon + 1;
		found = sscanf(port_str, "%u", &port);
		if (!found) {
			snprintf(error, errlen,
				"failed to parse port '%s' in '%s'",
				port_str, hostport);
			return 0;
		}
	}

	found = net_resolve_host(name, addr, error, errlen);
	if (!found)
		return found;

	addr->sin_port = htons(port);
	return found;
}

/*	Retrieves the IP address of an interface.
 *
 *	@iface		Name of the interface ("eth0", ..)
 *	@addr		Address assigned to the interface.
 *
 *	Returns 0 on success and -1 on error and sets errno.
 */
static inline int
net_get_iface_addr(const char *iface, struct sockaddr_in *addr)
{
	int res = 0;
	int tmpsock = 0;
	int sys_errno = 0;
	int sys_res = 0;
	struct ifreq req;
	struct sockaddr_in *iface_addr = NULL;

	tmpsock = socket(AF_INET, SOCK_DGRAM, 0);
	SYSERR_ON(tmpsock < 0);

	memset(&req, 0, sizeof(req));
	strncpy(req.ifr_name, iface, sizeof(req.ifr_name));

	res = ioctl(tmpsock, SIOCGIFADDR, &req);
	if (res) {
		sys_errno = errno;
		sys_res = res;
	}
		
	res = close(tmpsock);
	SYSERR_ON(res);

	if (sys_errno) {
		errno = sys_errno;
		return sys_res;
	}
		
	iface_addr = (struct sockaddr_in *)(&req.ifr_ifru.ifru_addr);
	BUG_ON(iface_addr->sin_family != AF_INET);

	*addr = *iface_addr;

	return 0;
}

static inline int
net_sockaddr_match_sinaddr(struct sockaddr_in *lhs, struct sockaddr_in *rhs)
{
	return (lhs->sin_addr.s_addr == rhs->sin_addr.s_addr);
}

static inline int
net_sockaddr_match(struct sockaddr_in *lhs, struct sockaddr_in *rhs)
{
	int addrs_match = (lhs->sin_addr.s_addr == rhs->sin_addr.s_addr);
	int ports_match = (lhs->sin_port == rhs->sin_port);

        return addrs_match && ports_match;
}

static inline char *
net_sinaddr_str(char *str, int len, const struct sockaddr_in *addr)
{
    const void *rptr = NULL;

    BUG_ON(addr->sin_family != AF_INET);
    rptr = inet_ntop(AF_INET, &addr->sin_addr, str, len);
    BUG_ON(rptr == NULL);
    return str;
}

static inline char *
net_sockaddr_str(char *str, int len, const struct sockaddr_in *addr)
{
    char tmp[INET_ADDRSTRLEN];

    net_sinaddr_str(tmp, sizeof(tmp), addr);
    snprintf(str, len, "%s:%d", tmp, ntohs(addr->sin_port));
    return str;
}

static inline char *
net_srcdst_sinaddr_str(char *str, int len,
		       const struct sockaddr_in *src,
		       const struct sockaddr_in *dst)
{
	char stmp[INET_ADDRSTRLEN];
	char dtmp[INET_ADDRSTRLEN];

        net_sinaddr_str(stmp, sizeof(stmp), src);
        net_sinaddr_str(dtmp, sizeof(dtmp), dst);
	snprintf(str, len, "src %s > dst %s", stmp, dtmp);
        return str;
}

static inline char *
net_srcdst_str(char *str, int len,
	       const struct sockaddr_in *src,
	       const struct sockaddr_in *dst)
{
	char stmp[100];
	char dtmp[100];

        net_sockaddr_str(stmp, sizeof(stmp), src);
        net_sockaddr_str(dtmp, sizeof(dtmp), dst);
        snprintf(str, len, "src %s > dst %s", stmp, dtmp);
        return str;
}

static inline void
net_sockaddr_in_log(const char *prefix, const struct sockaddr_in *addr)
{
    char tmp[100];
    net_sockaddr_str(tmp, sizeof(tmp), addr);
    LOG("%sip:port %s", prefix, tmp);
}

/* Returns the the interval between packets (in microsec)
 * for a given bitrate and a given packet length.
 */
static inline int
net_bps_to_usec_interval(long long bps, int pktlen_bytes)
{
	double pps = 0;
	const int usec_per_sec = 1000000;

	if (pktlen_bytes <= 0 || bps <= 0)
		return -1;

	BUG_ON(pktlen_bytes == 0);
	pps = 1.0 * bps / (pktlen_bytes * 8);
	return (1.0 * usec_per_sec / pps);
}

static inline double
net_calculate_rate_per_second(long long end_val, long long start_val, long long interval_usec)
{
        if (interval_usec == 0)
                return 0;
        
        return (1.0 * end_val - start_val) / (0.000001 * interval_usec);
}

#define NET_RECV_TCP_BLOCK(s, buf, buflen, bytes_expected) \
        __net_recv_tcp_block((s), (buf), (buflen), \
                             (bytes_expected), \
                             __FILE__, __LINE__)
                
static inline int
__net_recv_tcp_block(int sock, unsigned char *buf, int buflen,
                     int bytes_expected,
                     char *filename, int line)
{
        int bytes_accumulated = 0;
        int count = 0;

        BUG_ON(buflen < bytes_expected);

        DEB("+++ blocked on recv expecting %d bytes", bytes_expected);

        while (bytes_accumulated < bytes_expected) {
                unsigned char *tmpbuf = buf + bytes_accumulated;
                int tmpbuflen = bytes_expected - bytes_accumulated;
                int bytes_recv = recv(sock, tmpbuf, tmpbuflen, 0);

                if (bytes_recv <= 0)
                        return bytes_recv;

                bytes_accumulated += bytes_recv;
                count++;
                DEB("+++ got %d bytes on %d recv", bytes_recv, count);
        }

        if (count > 1) {
                DEB("required %d passes to get %d bytes",
                    count, bytes_accumulated);
        }

        if (bytes_accumulated != bytes_expected) {
                ERR("%s:%d: "
                    "expecting exactly %d bytes in this TCP block: got %d",
                    filename, line, bytes_expected, bytes_accumulated);
        }
        
        return bytes_accumulated;
}

// Netcounters

#define NET_DEFAULT_DEVTAG              "notag"
#define NET_DEFAULT_DEVSRC              ""
#define NET_DEFAULT_DEVDST              ""

// http://sd.wareonearth.com/~phil/net/overhead/
//
// Ethernet overhead bytes:
// 12 gap + 8 preamble + 14 header + 4 trailer = 38 bytes/packet w/o 802.1q
// 12 gap + 8 preamble + 18 header + 4 trailer = 42 bytes/packet with 802.1q
#define NET_ETHERNET_OVERHEAD_BYTES     (12 + 8 + 14 + 4)

#define NET_DEVNAME_LEN                 20
#define NET_DEVTAG_LEN                  50
#define NET_HOSTNAME_LEN                50

struct net_rate {
        double rxbps;
        double rxpps;
        double txbps;
        double txpps;
};

struct net_counters {
        struct timeval timestamp;

        long long rx_bytes;
        long long rx_pkts;

        long long tx_bytes;
        long long tx_pkts;
};

struct net_device {
        char name[NET_DEVNAME_LEN];
        char src[NET_HOSTNAME_LEN];
        char dst[NET_HOSTNAME_LEN];
        char tag[NET_DEVTAG_LEN];

        char name_pattern[NET_DEVNAME_LEN+1];

        struct net_counters counts_prev;
        struct net_counters counts;

        struct net_rate avg_rate;
};

static inline void
LOG_NETCOUNTERS(struct net_counters *counts)
{
        LOG("RX %lld %lld TX %lld %lld",
            counts->rx_bytes,
            counts->rx_pkts,
            counts->tx_bytes,
            counts->tx_pkts);
}

// 0 success, -1 error
static inline int
net_device_init(struct net_device *netdev, const char *netdev_name)
{
        struct astring_tokens tokens;

        memset(netdev, 0, sizeof(struct net_device));

        // extract name and tag (if any)
        // format: <ethname>[:<src>:<dst>:<tag>]

        astring_tokens_init(&tokens, netdev_name, ":");
        if (tokens.num < 1 || tokens.num > 4) {
                LOG("error: cannot parse device: tokens: %d devname: %s",
                    tokens.num, netdev_name);
                return -1;
        }

        astring_strncpy(netdev->name, tokens.token[0], sizeof(netdev->name));
        if (tokens.num > 1)
                astring_strncpy(netdev->src, tokens.token[1], sizeof(netdev->src));
        else
                astring_strncpy(netdev->src, NET_DEFAULT_DEVSRC, sizeof(netdev->src));
        if (tokens.num > 2)
                astring_strncpy(netdev->dst, tokens.token[2], sizeof(netdev->dst));
        else
                astring_strncpy(netdev->dst, NET_DEFAULT_DEVDST, sizeof(netdev->dst));
        if (tokens.num > 3)
                astring_strncpy(netdev->tag, tokens.token[3], sizeof(netdev->tag));
        else
                astring_strncpy(netdev->tag, NET_DEFAULT_DEVTAG, sizeof(netdev->tag));

        // pattern used to match relevant line from /proc/net/dev
        snprintf(netdev->name_pattern, sizeof(netdev->name_pattern),
                 "%s:", netdev->name);

        return 0;
}

static inline void
net_counters_extract(struct net_counters *counts, const char *procnet_entry)
{
        struct astring_tokens tokens;

        memset(counts, 0, sizeof(struct net_counters));
        gettimeofday(&counts->timestamp, NULL);

        astring_tokens_init(&tokens, procnet_entry, " :");
        BUG_ON(tokens.num != 17);

        // see /proc/net/dev
        counts->rx_bytes = atoll(tokens.token[1]);
        counts->rx_pkts = atoll(tokens.token[2]);

        counts->tx_bytes = atoll(tokens.token[9]);
        counts->tx_pkts = atoll(tokens.token[10]);
}

static inline int
net_rate_rollover(const struct net_rate *rate)
{
        return (rate->rxbps < 0 ||
                rate->rxpps < 0 ||
                rate->txbps < 0 ||
                rate->txpps < 0);
}

// Calculates network rate from net_counters @start, @end
// On success, returns 0, and fills @rate
// On failure (counter rollover), returns -1, and leaves @rate unchanged
static inline int
net_rate_calculate(struct net_rate *rate,
                   const struct net_counters *start,
                   const struct net_counters *end)
{
        struct net_rate tmprate;
        long long interval_usec = atimer_timeval_diff_usec(&end->timestamp,
                                                           &start->timestamp);

        memset(&tmprate, 0, sizeof(tmprate));

        tmprate.rxbps = net_calculate_rate_per_second(8 * end->rx_bytes,
                                                      8 * start->rx_bytes,
                                                      interval_usec);
        tmprate.rxpps = net_calculate_rate_per_second(end->rx_pkts,
                                                      start->rx_pkts,
                                                      interval_usec);
        tmprate.txbps = net_calculate_rate_per_second(8 * end->tx_bytes,
                                                      8 * start->tx_bytes,
                                                      interval_usec);
        tmprate.txpps = net_calculate_rate_per_second(end->tx_pkts,
                                                      start->tx_pkts,
                                                      interval_usec);

        if (net_rate_rollover(&tmprate)) {
                return -1;
        } else {
                *rate = tmprate;
                return 0;
        }
}

static inline void
net_device_update_rate_from_buf(struct net_device *netdev, const char *buf)
{
        // must match the device name in the /proc/net/dev line
        BUG_ON(strstr(buf, netdev->name_pattern) == NULL);

        netdev->counts_prev = netdev->counts;
        net_counters_extract(&netdev->counts, buf);

        // does not update on rollover
        net_rate_calculate(&netdev->avg_rate,
                           &netdev->counts_prev,
                           &netdev->counts);
}

static inline void
net_device_update_rate_from_proc(struct net_device *netdev)
{
        char buf[300];
        FILE *procdev = NULL;

        procdev = fopen("/proc/net/dev", "r");
        SYSERR_ON(procdev == NULL);

        while (fgets(buf, sizeof(buf), procdev) != NULL) {
                // match the device name in the /proc/net/dev line
                if (strstr(buf, netdev->name_pattern) != NULL) {
                        net_device_update_rate_from_buf(netdev, buf);
                        break;
                }
        }
        fclose(procdev);
}

// net_tcpinfo

// The system's tcp_info is defined in:
//
//      /usr/include/netinet/tcp.h      (userland)
//      <linux-src>/include/linux/tcp.h (kernel)
//
// We redefine it here under a different name so that
// we can build our code without the need to modify
// the include files in /usr/include.
//
// However this only works if the kernel has been built
// with the proper tcp_info. But any changes to tcp_info
// always extend the structure so that it is backward
// compatible. The code in net_get_tcpinfo() checks
// if the returned structure is big enough.
//
struct net_system_tcp_info
{
  u_int8_t	tcpi_state;
  u_int8_t	tcpi_ca_state;
  u_int8_t	tcpi_retransmits;
  u_int8_t	tcpi_probes;
  u_int8_t	tcpi_backoff;
  u_int8_t	tcpi_options;
  u_int8_t	tcpi_snd_wscale : 4, tcpi_rcv_wscale : 4;

  u_int32_t	tcpi_rto;
  u_int32_t	tcpi_ato;
  u_int32_t	tcpi_snd_mss;
  u_int32_t	tcpi_rcv_mss;

  u_int32_t	tcpi_unacked;
  u_int32_t	tcpi_sacked;
  u_int32_t	tcpi_lost;
  u_int32_t	tcpi_retrans;
  u_int32_t	tcpi_fackets;

  /* Times. */
  u_int32_t	tcpi_last_data_sent;
  u_int32_t	tcpi_last_ack_sent;	/* Not remembered, sorry.  */
  u_int32_t	tcpi_last_data_recv;
  u_int32_t	tcpi_last_ack_recv;

  /* Metrics. */
  u_int32_t	tcpi_pmtu;
  u_int32_t	tcpi_rcv_ssthresh;
  u_int32_t	tcpi_rtt;
  u_int32_t	tcpi_rttvar;
  u_int32_t	tcpi_snd_ssthresh;
  u_int32_t	tcpi_snd_cwnd;
  u_int32_t	tcpi_advmss;
  u_int32_t	tcpi_reordering;

  u_int32_t     tcpi_rcv_rtt;
  u_int32_t	tcpi_rcv_space;

  u_int32_t	tcpi_total_retrans;

  /* MGRP Extensions to tcp_info of 2.6.25 */
  u_int64_t	tcpi_total_retrans_bytes;
};

static inline void
net_system_tcp_info_diff(struct net_system_tcp_info *diff,
                         const struct net_system_tcp_info *end,
                         const struct net_system_tcp_info *begin)
{
        memset(diff, 0, sizeof(struct net_system_tcp_info));

        // Variable values are taken from @end
        APTR_FIELD_COPY(diff, end, tcpi_rto);
        APTR_FIELD_COPY(diff, end, tcpi_ato);
        APTR_FIELD_COPY(diff, end, tcpi_snd_mss);
        APTR_FIELD_COPY(diff, end, tcpi_rcv_mss);

        APTR_FIELD_COPY(diff, end, tcpi_pmtu);
        APTR_FIELD_COPY(diff, end, tcpi_rcv_ssthresh);
        APTR_FIELD_COPY(diff, end, tcpi_rtt);
        APTR_FIELD_COPY(diff, end, tcpi_rttvar);
        APTR_FIELD_COPY(diff, end, tcpi_snd_ssthresh);
        APTR_FIELD_COPY(diff, end, tcpi_snd_cwnd);
        APTR_FIELD_COPY(diff, end, tcpi_advmss);
        APTR_FIELD_COPY(diff, end, tcpi_reordering);

        APTR_FIELD_COPY(diff, end, tcpi_rcv_rtt);
        APTR_FIELD_COPY(diff, end, tcpi_rcv_space);

        APTR_FIELD_COPY(diff, end, tcpi_unacked);
        APTR_FIELD_COPY(diff, end, tcpi_sacked);
        APTR_FIELD_COPY(diff, end, tcpi_lost);
        APTR_FIELD_COPY(diff, end, tcpi_retrans);
        APTR_FIELD_COPY(diff, end, tcpi_fackets);

        // Diff the counters
        APTR_FIELD_DIFF_STORE(diff, end, begin, tcpi_total_retrans);
        APTR_FIELD_DIFF_STORE(diff, end, begin, tcpi_total_retrans_bytes);

        // The remaining fields are ignored and set to zero
}

/*
 * Adapted from the iproute2 project:
 *
 * http://www.linuxfoundation.org/en/Net:Iproute2
 * git clone git://git.kernel.org/pub/scm/linux/kernel/git/shemminger/iproute2.git
 *
 * from iproute2/misc/ss.c:tcp_show_info()
 */
static inline void
net_system_tcp_info_oneliner(char *str, int len,
                             const struct net_system_tcp_info *info)
{
        double rtt = 0;
        struct astring astr;

        astring_init(&astr, len);

        if (info->tcpi_options & TCPI_OPT_TIMESTAMPS)
                astring_append(&astr, " ts");
        if (info->tcpi_options & TCPI_OPT_SACK)
                astring_append(&astr, " sack");
        if (info->tcpi_options & TCPI_OPT_ECN)
                astring_append(&astr, " ecn");

        if (info->tcpi_options & TCPI_OPT_WSCALE)
                astring_append(&astr, " wscale:%d,%d",
                               info->tcpi_snd_wscale,
                               info->tcpi_rcv_wscale);
        if (info->tcpi_rto && info->tcpi_rto != 3000000)
                astring_append(&astr, " rto:%g", (double)info->tcpi_rto/1000);
        if (info->tcpi_rtt)
                astring_append(&astr, " rtt:%g/%g",
                               (double)info->tcpi_rtt/1000,
                               (double)info->tcpi_rttvar/1000);
        if (info->tcpi_ato)
                astring_append(&astr, " ato:%g", (double)info->tcpi_ato/1000);
        if (info->tcpi_snd_cwnd != 2)
                astring_append(&astr, " cwnd:%d", info->tcpi_snd_cwnd);
        if (info->tcpi_snd_ssthresh < 0xFFFF)
                astring_append(&astr, " ssthresh:%d", info->tcpi_snd_ssthresh);

        rtt = (double) info->tcpi_rtt;

        if (rtt > 0 && info->tcpi_snd_mss && info->tcpi_snd_cwnd) {
                double send_bps = ((double) info->tcpi_snd_cwnd *
                                  (double) info->tcpi_snd_mss * 8000000.) / rtt;
                astring_append(&astr, " send %.2fbps", send_bps);
        }

        if (info->tcpi_rcv_rtt)
                astring_append(&astr, " rcv_rtt:%g",
                               (double) info->tcpi_rcv_rtt/1000);
        if (info->tcpi_rcv_space)
                astring_append(&astr, " rcv_space:%d", info->tcpi_rcv_space);
        
        strncpy(str, astr.cstr, len);
        astring_done(&astr);
}

static inline void
net_log_tcp_info_oneliner(const struct net_system_tcp_info *info)
{
        char buf[400];

        net_system_tcp_info_oneliner(buf, sizeof(buf), info);
        LOG("%s", buf);
}

// net_tcpinfo
//
// Wrapper around tcp_info that also records timestamp.

struct net_tcpinfo {
        long long usec_diff;
        struct timeval tstamp;
        struct net_system_tcp_info tcp_info;
};

static inline void
net_get_tcpinfo(int sock, struct net_tcpinfo *tcpinfo)
{
        int rc = 0;
        uint tcp_info_len = sizeof(tcpinfo->tcp_info);

        gettimeofday(&tcpinfo->tstamp, NULL);

        rc = getsockopt(sock, SOL_TCP, TCP_INFO,
                        (void *)&tcpinfo->tcp_info,
                        &tcp_info_len);
        SYSERR_ON(rc);
        if (tcp_info_len != sizeof(tcpinfo->tcp_info))
                ERR("tcp_info size mismatch: "
                    "we passed %d, kernel returned %d",
                    sizeof(tcpinfo->tcp_info), tcp_info_len);
}

static inline void
net_get_tcpinfo_diff(struct net_tcpinfo *diff,
                     const struct net_tcpinfo *end,
                     const struct net_tcpinfo *begin)
{
        memset(diff, 0, sizeof(struct net_tcpinfo));

        diff->usec_diff = atimer_timeval_diff_usec(&end->tstamp, &begin->tstamp);
        net_system_tcp_info_diff(&diff->tcp_info, &end->tcp_info, &begin->tcp_info);
}

#define NET_PCAP_ERRBUF_SIZE    (PCAP_ERRBUF_SIZE+200)

// pcap helper functions
static inline pcap_t *
net_pcap_open_offline_filter(const char *pcap_name, char *filter,
                             char *err, int errlen)
{
        pcap_t *trace = NULL;
        char err2[PCAP_ERRBUF_SIZE];

        trace = pcap_open_offline(pcap_name, err2);
        if (trace == NULL) {
                snprintf(err, errlen, "failed to open trace: %s", err2);
                return NULL;
        }

        if (filter != NULL && strlen(filter) > 0) {
                int rc = 0;
                struct bpf_program fp;

                rc = pcap_compile(trace, &fp, filter, 0, 0);
                if (rc) {
                        snprintf(err, errlen,
                                 "failed to compile filter '%s': %s",
                                 filter, pcap_geterr(trace));
                        return NULL;
                }

                rc = pcap_setfilter(trace, &fp);
                if (rc) {
                        snprintf(err, errlen,
                                "failed to install filter '%s': %s",
                                filter, pcap_geterr(trace));
                        return NULL;
                }
        }

        return trace;
}

#endif /* _NET_H */
