/*
 * Created on   : Sat Mar 19 18:21:48 EST 2005
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * rbuf.c
 *
 * Buffer with reference counting
 * 
 * $Id: rbuf.c,v 1.4 2005/06/28 00:46:12 pavlos Exp $Buffer related utilities
 */

#include "rbuf.h"

/* rbuf */

int
rbuf_init_alloc(rbuf *buf, size_t size)
{
    memset(buf, 0, sizeof(rbuf));

    buf->size = size;
    buf->ptr = malloc(buf->size);
    buf->owner = 1;

    if (buf->ptr == NULL)
        return RBUF_ENOMEM;

    return RBUF_OK;
}


int
rbuf_init_ptr(rbuf *buf, void *ptr, size_t size, int owner)
{
    if (ptr == NULL && size != 0)
        return RBUF_EINVAL;

    memset(buf, 0, sizeof(rbuf));

    buf->size = size;
    buf->ptr = ptr;
    buf->owner = owner;

    return RBUF_OK;
}


int
rbuf_done(rbuf *buf)
{
    if (buf->inuse !=0) {
        logerr_log(RBUF_LOG_INUSE,
            "rbuf still in use -- buf(%p) inuse(%d)", buf, (buf)->inuse);
        logerr_bug(buf->inuse == 0);
    }

    if (buf->owner && buf->ptr != NULL)
        free(buf->ptr);

    memset(buf, 0, sizeof(rbuf));

    return RBUF_OK;
}


void
rbuf_free(void *buf_ptr)
{
    int res = 0;

    logerr_passert(buf_ptr);

    res = rbuf_done((rbuf *)buf_ptr);
    logerr_bug(res == RBUF_OK);

    free(buf_ptr);
}

/* rbuf_view */

int
rbuf_view_init(rbuf_view *view)
{
    memset(view, 0, sizeof(rbuf_view));
    return RBUF_OK;
}


int
rbuf_view_done(rbuf_view *view)
{
    if (view->rbowner != NULL) {
        rbuf_release(view->rbowner);
    }
    memset(view, 0, sizeof(rbuf_view));
    return RBUF_OK;
}


void
rbuf_view_free(void *view_ptr)
{
    int res = 0;

    logerr_passert(view_ptr);

    res = rbuf_view_done((rbuf_view *)view_ptr);
    logerr_bug(res == RBUF_OK);

    free(view_ptr);
}


/* rbuf_mgr */

int
rbuf_mgr_init(rbuf_mgr *mgr, size_t min_size_pow, size_t max_size_pow)
{
    int res = 0;
    int i = 0;

    memset(mgr, 0, sizeof(rbuf_mgr));

    logerr_bug(min_size_pow >= 0 && min_size_pow <= max_size_pow);

    mgr->min_size_pow = min_size_pow;
    mgr->max_size_pow = max_size_pow;

    mgr->qcount = mgr->max_size_pow - mgr->min_size_pow + 1;

    mgr->rbufs = malloc(mgr->qcount * sizeof(aqueue));
    if (mgr->rbufs == NULL)
        return RBUF_ENOMEM;
    
    /* mgr->views is a queue of rbuf_view pointers */

    res = aqueue_init(&mgr->views, AQUEUE_RECYCLE,
            AQUEUE_NOLIMIT, AQUEUE_NOLIMIT, 0,
            0,  rbuf_view_free, NULL, NULL);

    logerr_bug(res == AQUEUE_OK);

    /* mgr->rbufs[i] is a queue of rbuf pointers
     * each holding memory of size 2^(i+min_size_pow) bytes
     */

    for (i = 0; i < mgr->qcount; i++) {

        /* create an unbounded queue that recycles node pointers */

        res = aqueue_init(mgr->rbufs + i, AQUEUE_RECYCLE,
                AQUEUE_NOLIMIT, AQUEUE_NOLIMIT, 0,
                0,  rbuf_free, NULL, NULL);

        logerr_bug(res == AQUEUE_OK);
    }

    return RBUF_OK;
}


int
rbuf_mgr_done(rbuf_mgr *mgr)
{
    int res = 0;
    int i = 0;

    res = aqueue_done(&mgr->views);
    logerr_bug(res == AQUEUE_OK);

    for (i = 0; i < mgr->qcount; i++) {
        res = aqueue_done(mgr->rbufs + i);
        logerr_bug(res == AQUEUE_OK);
    }

    free(mgr->rbufs);

    memset(mgr, 0, sizeof(rbuf_mgr));

    return RBUF_OK;
}


/* no need to recycle; manager keeps track of buffers and will reuse */

rbuf *
rbuf_mgr_alloc_aux(rbuf_mgr *mgr, size_t size, char *file, int line)
{
    int res = 0;
    size_t actual_size = 1;
    size_t actual_pow = 0;

    aqueue *bufq = NULL;
    rbuf *buf = NULL;

    /* allocate in steps of powers of two */

    while (actual_size < size || actual_pow < mgr->min_size_pow) {
        actual_size <<= 1;
        actual_pow++;
    }

    /* size must be within bounds */

    if (actual_pow > mgr->max_size_pow) {
        logerr_log(-1, "requested size(%d) too large: up to (2^%d) supported",
                size, mgr->max_size_pow);
        logerr_bug(actual_pow <= mgr->max_size_pow);
    }

    /* first pick the queue to look at */

    bufq = mgr->rbufs + (actual_pow - mgr->min_size_pow);

    /* peek the first rbuf; is it still in use? */

    buf = aqueue_peek_front(bufq);

    if (buf != NULL && !buf->inuse) {

        /* pop the rbuf from the queue; we will reinsert it later */

        res = aqueue_pop_ptr_front(bufq, &buf);
        logerr_bug(res == AQUEUE_OK);

        logerr_log(RBUF_LOG_ALLOC,
                "rbuf_mgr: size(%d) actual(%d) reuse(%d) from(%s:%d)",
                size, actual_size, 1, file, line);
    } else {

        /* first element not available; allocate new */

        buf = malloc(sizeof(rbuf));
        logerr_bug(buf != NULL);

        /* initialize and allocate actual memory block */

        res = rbuf_init_alloc(buf, actual_size);
        logerr_bug(res == RBUF_OK);

        logerr_log(RBUF_LOG_ALLOC,
                "rbuf_mgr: size(%d) actual(%d) reuse(%d) from(%s:%d)",
                size, actual_size, 0, file, line);
    }

    logerr_bug(buf != NULL && buf->size == actual_size && !buf->inuse);

    /* reserve it before you insert it into the queue */

    rbuf_reserve(buf);

    /* insert into queue so that it can be recycled when no longer in use */

    res = aqueue_push_ptr_back(bufq, buf, AQUEUE_OWNER);
    logerr_bug(res == AQUEUE_OK);

    /* user must release when no longer needed */

    return buf;
}


/* view needs to be recycled explicitly; mgr does not keep track of it */

rbuf_view *
rbuf_mgr_alloc_view(rbuf_mgr *mgr)
{
    int res = 0;
    rbuf_view *view = NULL;

    res = aqueue_pop_ptr_front(&mgr->views, &view);
    logerr_bug(res == AQUEUE_OK || res == AQUEUE_EMPTY);

    if (res == AQUEUE_EMPTY) {
        view = malloc(sizeof(rbuf_view));
    }

    logerr_bug(view != NULL);

    res = rbuf_view_init(view);
    logerr_bug(res == RBUF_OK);
    
    return view;
}

int
rbuf_mgr_recycle_view(rbuf_mgr *mgr, rbuf_view *view)
{
    int res = 0;

    res = rbuf_view_done(view);
    logerr_bug(res == RBUF_OK);

    res = aqueue_push_ptr_back(&mgr->views, view, AQUEUE_OWNER);
    logerr_bug(res == AQUEUE_OK);

    return RBUF_OK;
}



