/*
 * Created on   : Sat Mar 19 18:21:48 EST 2005
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * rbuf.h
 *
 * Buffer with reference counting
 * 
 * $Id: rbuf.h,v 1.5 2005/06/29 00:08:42 pavlos Exp $Buffer related utilities
 */

#ifndef __RBUF_H
#define __RBUF_H

#include <string.h>

#include "aqueue.h"
#include "logerr.h"

#include "rbufx.h"

#define RBUF_OK                 0
#define RBUF_ENOMEM             -1
#define RBUF_EINVAL             -2
#define RBUF_EQUEUE             -3

#define RBUF_LOG_INUSE          99
#define RBUF_LOG_ALLOC          99

struct rbuf {
    void *ptr;
    size_t size;
    int owner;
    int inuse;
};

struct rbuf_view {
    void *base;
    size_t len;
    rbuf *rbowner;
};

struct rbuf_mgr {

    size_t min_size_pow;
    size_t max_size_pow;

    aqueue views;

    int qcount;
    aqueue *rbufs;
};


/* rbuf */

#define rbuf_reserve(buf) \
        do { \
            logerr_bug((buf)->inuse >= 0); \
            (buf)->inuse++; \
            logerr_log(RBUF_LOG_INUSE, \
                "reserve     -- buf(%p) inuse(%d) from(%s:%d)", \
                buf, (buf)->inuse, __FILE__, __LINE__); \
        } while (0)

#define rbuf_release(buf) \
        do { \
            logerr_bug((buf)->inuse >= 0); \
            (buf)->inuse--; \
            logerr_log(RBUF_LOG_INUSE, \
                "--> release -- buf(%p) inuse(%d) from(%s:%d)", \
                buf, (buf)->inuse, __FILE__, __LINE__); \
            logerr_bug((buf)->inuse >= 0); \
        } while (0)

int rbuf_init_alloc(rbuf *buf, size_t size);
int rbuf_init_ptr(rbuf *buf, void *ptr, size_t size, int owner);
int rbuf_done(rbuf *buf);
void rbuf_free(void *buf_ptr);

/* rbuf_view */

#define rbuf_view_reserve(v) \
        do { \
            logerr_bug((v)->rbowner != NULL); \
            rbuf_reserve((v)->rbowner); \
        } while(0)

#define rbuf_view_release(v) \
        do { \
            if ((v)->rbowner != NULL) { \
                rbuf_release((v)->rbowner); \
                (v)->rbowner = NULL; \
            } \
        } while(0)

int rbuf_view_init(rbuf_view *view);
int rbuf_view_done(rbuf_view *view);
void rbuf_view_free(void *view_ptr);

/* rbuf_mgr */

int rbuf_mgr_init(rbuf_mgr *mgr, size_t min_size_pow, size_t max_size_pow);
int rbuf_mgr_done(rbuf_mgr *mgr);

#define rbuf_mgr_alloc(m, s) \
        (rbuf_mgr_alloc_aux((m), (s), __FILE__, __LINE__))

rbuf *rbuf_mgr_alloc_aux(rbuf_mgr *mgr, size_t size, char *file, int line);

rbuf_view *rbuf_mgr_alloc_view(rbuf_mgr *mgr);
int rbuf_mgr_recycle_view(rbuf_mgr *mgr, rbuf_view *view);

/* misc */

#define rbuf_fill_iov_from_views(iov, views, vlen) \
        do { \
            int _i = 0; \
            for (_i = 0; _i < (vlen); _i++) { \
                (iov)[_i].iov_base = (views)[_i].base; \
                (iov)[_i].iov_len  = (views)[_i].len; \
            } \
        } while(0)


#endif /* __RBUF_H */


