/*
 * ptime.h	Packet Train -- time related routines
 *
 * Created:	Tue Nov  1 19:17:21 EST 2005
 *
 * Version:	$Id: tick.h,v 1.9 2005/12/03 00:10:08 pavlos Exp $
 *
 * Author:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 */

#ifndef _TICK_H
#define _TICK_H

#include <stdint.h>
#include <errno.h>
#include <sys/time.h>
#include <time.h>
#include <assert.h>

#include "alog.h"

#define TICK_USLEEP_MAX			0xffffffff
#define TICK_MSLEEP_MAX			0x003fffff

#define TICK_MICROSLEEP_MAX		0xffffffff
#define TICK_MILLISLEEP_MAX		0x003fffff
#define TICK_SLEEP_MAX			0x00000fff

#define TICK_NANOSLEEP_PRECISION_USEC	100

#define TICK_USEC_PER_SEC               1000000L

/*
 *	Increments @tv by @usec microseconds.
 */
static inline void
tick_tv_add_usec(struct timeval *tv, uint64_t usec)
{
	usec += tv->tv_usec;
	tv->tv_sec += (usec / 1000000);
	tv->tv_usec = (usec % 1000000);
}

/*
 *	Converts @tv into microseconds and returns the value.
 */
static inline uint64_t
tick_tv_to_usec(const struct timeval *tv)
{
	uint64_t usec = tv->tv_sec;
	usec *= 1000000;
	usec += tv->tv_usec;

	return usec;
}

/*
 *	Converts @usec into struct timeval and stores in @tv.
 */
static inline void
tick_usec_to_tv(struct timeval *tv, const uint64_t usec)
{
	tv->tv_sec = usec / 1000000;
	tv->tv_usec = usec % 1000000;
}

/*
 *	Returns gettimeofday(2) in microseconds.
 */
static inline uint64_t
tick_now_usec()
{
	int res = 0;
	struct timeval now;

	res = gettimeofday(&now, NULL);
	BUG_ON(res);

	return tick_tv_to_usec(&now);
}

/*
 *	Calculate the time difference between @to and @from.
 *
 *	@to		Destination time.
 *	@from		Origin time.
 *
 *	Returns in microseconds the difference @to - @from.
 *	If @to > @from then the return value is positive.
 */
static inline int64_t
tick_tv_udiff(const struct timeval *to, struct timeval *from)
{
	int64_t usec = to->tv_sec;

	usec -= from->tv_sec;
	if (usec)
		usec *= 1000000;

	usec += to->tv_usec;
	usec -= from->tv_usec;

	return usec;
}

/*
 *	Calculate the time difference between @time and gettimeofday(2)
 *
 *	@time		Time in the past or in the future.
 *
 *	@now		Output parameter. Overwrite with current time from
 *			gettimeofday(2). Set to NULL if you do not care.
 *			We provide this output parameter to save on the
 *			invocations to gettimeofday(2).
 *
 *	Returns in microseconds the difference @time - @gettimeofday(2)
 *	and stores @gettimeofday(2) in @now if it is non-NULL.
 *
 *	If the difference is positive, then @time is a time in the future.
 *	If the difference is negative, then @time is a time in the past.
 */
static inline int64_t
tick_now_udiff_future(const struct timeval *time, struct timeval *now)
{
	int res = 0;
	int64_t usec = 0;

	struct timeval time_now;

	res = gettimeofday(&time_now, NULL);
	BUG_ON(res);

	usec = tick_tv_udiff(time, &time_now);

	if (now != NULL)
		*now = time_now;

	return usec;
}

static inline int64_t
tick_now_udiff_past(const struct timeval *past, struct timeval *now)
{
	return -tick_now_udiff_future(past, now);
}

/*
 *	Sleep with microsecond precision.
 *
 *	Uses a combination of nanosleep(2) and busy waiting to sleep
 *	with precision equal to the precision of gettimeofday(2).
 *
 *	@usec		Number of usec to sleep for.
 *
 *	@start_sleep	The time from which to start counting @usec.
 *			This may be used to increase precision when
 *			sleeping in a loop between successive invocations
 *			of this function. Can be NULL, in which case
 *			gettimeofday(2) is used.
 *
 *	@end_sleep	The time returned from gettimeofday(2) just before
 *			this function returns. Set to NULL if you do not
 *			need this value. Useful in a loop when used in
 *			conjunction with @start_sleep.
 *			For example, @start_sleep can be set equal to
 *			@end_sleep of the previous loop.
 *	
 *	Returns actual number of microseconds elapsed.
 */
static inline uint64_t
tick_usleep_w_feedback(uint64_t usec,
                       struct timeval *start_sleep,
                       struct timeval *end_sleep,
                       int *usec_effective_p,
                       int *usec_slept_p,
                       int *busywait_loops_p)
{
	int res = 0;
	int64_t usec_total = usec;
	int64_t usec_passed = 0;
	int64_t usec_left = 0;

	struct timeval before;
	struct timeval now;
	
	BUG_ON(usec > TICK_USLEEP_MAX);

        if (usec_effective_p != NULL)
                *usec_effective_p = 0;
        if (usec_slept_p != NULL)
                *usec_slept_p = 0;
        if (busywait_loops_p != NULL)
                *busywait_loops_p = 0;

	if (start_sleep != NULL) {
		before = *start_sleep;
		usec_passed = tick_now_udiff_past(&before, &now);

	} else {
		res = gettimeofday(&now, NULL);
		BUG_ON(res);

		before = now;
		usec_passed = 0;
	}

	if (usec <= 0)
		goto out;

	/* use nanosleep while over the threshold */

	usec_left = usec_total - usec_passed;

        if (usec_effective_p != NULL)
                *usec_effective_p = usec_left;

	while (usec_left > TICK_NANOSLEEP_PRECISION_USEC) {

		struct timespec req;
		int64_t usec_sleep = usec_left - TICK_NANOSLEEP_PRECISION_USEC;

		req.tv_sec = usec_sleep / 1000000;
		req.tv_nsec = (usec_sleep % 1000000) * 1000;

                if (usec_slept_p != NULL)
                        *usec_slept_p += usec_sleep;

		res = nanosleep(&req, NULL);
		BUG_ON(res != 0 && errno != EINTR);

		/*
		 * We update @now with the current time and calculate
		 * the difference in seconds from @before.
		 */
		usec_passed = tick_now_udiff_past(&before, &now);
		usec_left = usec_total - usec_passed;
	}

	/* use busy waiting */

	while ((usec_left = usec_total - usec_passed) > 0) {
                if (busywait_loops_p != NULL)
                        *busywait_loops_p += 1;
		usec_passed = tick_now_udiff_past(&before, &now);
        }
out:
	if (end_sleep != NULL)
		*end_sleep = now;

	BUG_ON(usec_passed < 0);
	return usec_passed;
}

static inline uint64_t
tick_usleep_w(uint64_t usec,
		struct timeval *start_sleep,
		struct timeval *end_sleep)
{
        return tick_usleep_w_feedback(usec, start_sleep, end_sleep,
                                      NULL, NULL, NULL);
}

static inline uint64_t
tick_usleep(uint64_t usec)
{
	return tick_usleep_w(usec, NULL, NULL);
}

static inline uint64_t
tick_msleep(uint64_t msec)
{
	BUG_ON(msec > TICK_MSLEEP_MAX);
	return tick_usleep_w(msec*1000, NULL, NULL);
}

/*
 *	Sleep using nanosleep(2).
 *
 *	Same limitations and precision as nanosleep(2).
 *	Does not affect signals.
 *
 *	@nsec			Nanoseconds to sleep for.
 *	@defer_signals		Keep sleeping even if you catch a signal.
 *
 *	If @defer_signals is 0 then it behaves exactly as nanosleep(2).
 *	In case of an error returns -1 instead of 0 and sets errno
 *	appropriately.
 *
 *	If @defer_signals is 1 then it ignores any signals and sleeps for
 *	the whole duration. It returns the number of signals caught or
 *	-1 on error with errno set (which will never be EINTR).
 */
static inline int
tick_nanosleep(uint64_t nsec, int defer_signals)
{
	int res = 0;
	int signals = 0;
	struct timespec req;
	struct timespec rem;

	rem.tv_sec = nsec / 1000000000;
	rem.tv_nsec = nsec % 1000000000;

	if (nsec == 0)
		return 0;

	do {
		req = rem;
		res = nanosleep(&req, &rem);
		if (res == -1) {
			if (errno != EINTR || !defer_signals)
				return res;
			signals++;
		}
	} while (res != 0);

	return signals;
}

static inline int
tick_microsleep(uint64_t usec, int defer_signals)
{
	BUG_ON(usec > TICK_MICROSLEEP_MAX);
	return tick_nanosleep(usec * 1000, defer_signals);
}

static inline int
tick_millisleep(uint64_t msec, int defer_signals)
{
	BUG_ON(msec > TICK_MILLISLEEP_MAX);
	return tick_nanosleep(msec * 1000000, defer_signals);
}

static inline int
tick_sleep(uint64_t sec, int defer_signals)
{
	BUG_ON(sec >  TICK_SLEEP_MAX);
	return tick_nanosleep(sec * 1000000000, defer_signals);
}

#endif /* _TICK_H */
