/* Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * Fri May 25 12:43:40 EDT 2007
 */

#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/ip.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

/* for inet_ntop */
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "libestimator.h"

/* libcommon */
#include "alog.h"
#include "net.h"

#include <string.h>
#include <unistd.h>

int
est_send_ctrlmsg(struct est_channel *ch,
		void *ctrlmsg,
		char *err_str, int err_len)
{
	int rc = 0;
	int bytes = 0;
	struct est_header *hdr = ctrlmsg;
	char err[200] = {0};

	EST_LOGHDR_SND(hdr);

	BUG_ON(ctrlmsg == NULL);
	BUG_ON(hdr->size < sizeof(struct est_header));

	/* send control message (cmd or reply) */

	bytes = send(ch->sock, ctrlmsg, hdr->size, 0);

	if (bytes < 0) {
		rc = EST_ESEND_CTRL;
		strerror_r(-bytes, err, sizeof(err));
		goto error;
	}

	if (bytes != hdr->size) {
		rc = EST_ESEND_CTRL;
		snprintf(err, sizeof(err),
			"tried to send %d bytes: sent only %d",
			hdr->size, bytes);
		goto error;
	}

	return 0;
error:
	if (err_str != NULL)
		strncpy(err_str, err, err_len);
	return rc;
}

static inline int
est_validate_ctrlmsg(struct est_header *hdr,
		char *err_str, int err_len)
{
	int rc = 0;
	int ok = 0;
	int csize = 0;
	int rsize = 0;
	char err[200] = {0};

	ok = est_valid_type_and_sizes(hdr->type, &csize, &rsize);
	if (!ok) {
		rc = EST_EINVAL_CTRL;
		snprintf(err, sizeof(err),
			"type is not valid in header %x",
			hdr->type);
		goto error;
	}

	if (est_is_reply(hdr)) {

		if (hdr->size != rsize) {
			snprintf(err, sizeof(err),
				"reply expects %d bytes: got %d",
				rsize, hdr->size);
			goto error;
		}

	} else if (est_is_cmd(hdr)) {

		if (hdr->size != csize) {
			snprintf(err, sizeof(err),
				"cmd expects %d bytes: got %d",
				csize, hdr->size);
			goto error;
		}

	} else {
		rc = EST_EINVAL_CTRL;
		snprintf(err, sizeof(err),
			"message is not a reply or a cmd %x",
			hdr->type);
		goto error;
	}

	return 0;

error:
	if (err_str != NULL)
		strncpy(err_str, err, err_len);
	return rc;
}


int
est_recv_ctrlmsg(struct est_channel *ch,
		void *ctrlmsg,
		int ctrllen,
		char *err_str, int err_len)
{
	int rc = 0;
	int bytes = 0;
	char err[200] = {0};

	struct est_header *hdr = ctrlmsg;
	unsigned char *body = (unsigned char *)ctrlmsg + sizeof(*hdr);
	int bodylen = ctrllen - sizeof(*hdr);
	int bodylen_expected = 0;

	BUG_ON(ctrlmsg == NULL);
	BUG_ON(ctrllen < sizeof(struct est_header));

	/* receive header */

        bytes = NET_RECV_TCP_BLOCK(ch->sock,
                                   (unsigned char *)hdr, sizeof(*hdr),
                                   sizeof(*hdr));
	if (bytes < 0) {
		rc = EST_ERECV_CTRL;
		strerror_r(-bytes, err, sizeof(err));
		goto error;
	}

	if (bytes == 0) {
		rc = EST_ECLOSED_REMOTE;
		snprintf(err, sizeof(err), "no control message received");
		goto error;
	}

	/* receive body */

	bodylen_expected = hdr->size - sizeof(*hdr);

	if (bodylen_expected == 0)
		goto received;

	if (bodylen_expected < 0) {
		rc = EST_ERECV_CTRL;
		snprintf(err, sizeof(err),
			"invalid size in header: %d",
			hdr->size);
		goto error;
	}

	if (bodylen_expected > bodylen) {
		rc = EST_ERECV_CTRL;
		snprintf(err, sizeof(err),
			"not enough space for %d body bytes: space only for %d",
			bodylen_expected, bodylen);
		goto error;
	}

        bytes = NET_RECV_TCP_BLOCK(ch->sock, body, bodylen, bodylen_expected);
        if (bytes < 0) {
                rc = EST_ERECV_CTRL;
                strerror_r(-bytes, err, sizeof(err));
                goto error;
        }

received:
	rc = est_validate_ctrlmsg(hdr, err_str, err_len);
	if (rc) {
		goto error;
	}

	EST_LOGHDR_RCV(hdr);
	
	return 0;

error:
	if (err_str != NULL)
		strncpy(err_str, err, err_len);
	return rc;
}

static inline int
est_hello_exchange_server(struct est_channel *ch)
{
	int rc = 0;
	int ok = 0;
	est_cmd cmd;
	est_reply reply;
	char err[200];

	memset(&cmd, 0, sizeof(est_cmd));
	memset(&reply, 0, sizeof(est_reply));

	/* receive hello command */

	rc = est_recv_ctrlmsg(ch, &cmd, sizeof(cmd), err, sizeof(err));
	if (rc)
		goto error;

	ok = est_is_cmd(&cmd.hdr) &&
		(cmd.hdr.type & EST_TYPE_HELLO);
	if (!ok) {
		rc = EST_EHELLO;
		snprintf(err, sizeof(err), "cmd was not a HELLO");
		goto error;
	}

        // remote timestamp
        ch->remote_hello_timestamp = cmd.hello.now;

        // local timestamp
        gettimeofday(&ch->local_hello_timestamp, NULL);
        reply.hello.now = ch->local_hello_timestamp;

	/* send hello reply */

	ok = est_fill_reply_header(&reply.hdr, EST_TYPE_HELLO, 1);	
	BUG_ON(!ok);

	rc = est_send_ctrlmsg(ch, &reply, err, sizeof(err));
	if (rc)
		goto error;

	return 0;

error:
	LOGTR("%s: %s", est_strerror(rc), (err));
	return rc;
}

static inline int
est_hello_exchange_client(struct est_channel *ch)
{
	int rc = 0;
	int ok = 0;
	est_cmd cmd;
	est_reply reply;
	char err[200];

	memset(&cmd, 0, sizeof(est_cmd));
	memset(&reply, 0, sizeof(est_reply));

	/* send hello command */

	ok = est_fill_cmd_header(&cmd.hdr, EST_TYPE_HELLO);	
	BUG_ON(!ok);

        // local timestamp
        gettimeofday(&ch->local_hello_timestamp, NULL);
        cmd.hello.now = ch->local_hello_timestamp;

	rc = est_send_ctrlmsg(ch, &cmd, err, sizeof(err));
	if (rc)
		goto error;

	/* receive hello reply */

	rc = est_recv_ctrlmsg(ch, &reply, sizeof(reply), err, sizeof(err));
	if (rc)
		goto error;

	ok = est_is_reply_final(&reply.hdr) &&
		(reply.hdr.type & EST_TYPE_HELLO);
	if (!ok) {
		rc = EST_EHELLO;
		snprintf(err, sizeof(err), "reply was not a HELLO");
		goto error;
	}
		
        // remote timestamp
        ch->remote_hello_timestamp = reply.hello.now;

	return 0;
error:
	LOGTR("%s: %s", est_strerror(rc), (err));
	return rc;
}

static inline void
est_init_channel(struct est_channel *ch)
{
	memset(ch, 0, sizeof(struct est_channel));
	ch->state = EST_CHANNEL_CLOSED;
}

int
est_open_channel(struct est_channel *ch, struct sockaddr_in *remote)
{
	int rc = 0;
	int on = 0;
	char buf[100];

	if (ch == NULL || remote == NULL) {
		rc = EST_EINVAL;
		goto error;
	}
	est_init_channel(ch);

	ch->remote = *remote;
	if (ch->remote.sin_port == 0) {
		LOGTR("error: no estimator port specified");
		rc = EST_EINVAL;
		goto error;
	}

	/* open control socket */

	ch->sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (ch->sock < 0) {
		rc = ch->sock;
		LOGTR("failed to open control socket");
		goto error;
	}
	ch->state = EST_CHANNEL_SOCKET_OPEN;

	on = 1;
	rc = setsockopt(ch->sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
	if (rc) {
		LOGTR("failed to set SO_REUSEADDR socket option");
		goto close_socket;
	}

	rc = connect(ch->sock, (struct sockaddr *)&ch->remote, sizeof(ch->remote));
	if (rc) {
		LOGTR("failed to connect to control socket");
		goto close_socket;
	}
	ch->state = EST_CHANNEL_HELLO_PENDING;

	/* initiate hello exchange */

	rc = est_hello_exchange_client(ch);
	if (rc) {
		LOGTR("hello exchange failed: client side");
		goto close_socket;
	}
	ch->state = EST_CHANNEL_OPEN;

	LOGFN("connected to server %s:%d",                                    
		inet_ntop(ch->remote.sin_family,
			&ch->remote.sin_addr,
			buf, sizeof(buf)),                       
		ntohs(ch->remote.sin_port));                          

	return 0;

close_socket:
	close(ch->sock);
	ch->state = EST_CHANNEL_CLOSED;
error:
	return rc;
}

int
est_accept_channel(int sock, struct est_channel *newch)
{
	int rc = 0;
	char buf[100];
	unsigned int addrlen = 0;

	if (newch == NULL) {
		rc = EST_EINVAL;
		goto error;
	}
	est_init_channel(newch);

	addrlen = sizeof(newch->remote);
	newch->sock = accept(sock, (struct sockaddr *)&newch->remote, &addrlen);
	BUG_ON(addrlen != sizeof(newch->remote));
	if (newch->sock < 0) {
		rc = newch->sock;
		LOGTR("failed to accept control socket");
		goto error;
	}
	newch->state = EST_CHANNEL_HELLO_PENDING;

	/* wait for hello exchange */

	rc = est_hello_exchange_server(newch);
	if (rc) {
		LOGTR("hello exchange failed: server side");
		goto close_socket;
	}
	newch->state = EST_CHANNEL_OPEN;

	LOGFN("connected to client %s:%d",                                    
		inet_ntop(newch->remote.sin_family,
			&newch->remote.sin_addr,
			buf, sizeof(buf)),                       
		ntohs(newch->remote.sin_port));                          

	return 0;

close_socket:
	close(newch->sock);
	newch->state = EST_CHANNEL_CLOSED;
error:
	return rc;
}

void
est_close_channel(struct est_channel *ch)
{
	int rc = 0;

	if (ch->state == EST_CHANNEL_CLOSED)
		return;

	rc = close(ch->sock);
	SYSERR_ON(rc);

	est_init_channel(ch);
}

/* pathload probing */

int
est_create_pathload_probe_session(struct est_channel *dst,
			struct sockaddr_in *src,
			int iterate,
                        int iterate_interval_sec,
			int mgrp,
                        int piggyback,
                        int fast,
			int start_session,
			void **session_p)
{
	int rc = 0;
	int ok = 0;
	struct est_probe_cmd cmd;
	struct est_probe_reply reply;
	char error[200];

	/* send PROBE */

	memset(&cmd, 0, sizeof(cmd));
	ok = est_fill_cmd_header(&cmd.hdr, EST_TYPE_PROBE);
	BUG_ON(!ok);

	cmd.e2addr = *src;
	cmd.iterate = iterate;
        cmd.iterate_interval_sec = iterate_interval_sec;
	cmd.mgrp = mgrp;
        cmd.piggyback = piggyback;
        cmd.fast = fast;
	cmd.start_session = start_session;

	rc = est_send_ctrlmsg(dst, &cmd, error, sizeof(error));
	if (rc) {
		LOGTR("failed to send PROBE: %s", error);
		goto error;
	}

	/* PROBE reply */

	rc = est_recv_ctrlmsg(dst, &reply, sizeof(reply), error, sizeof(error));
	if (rc) {
		LOGTR("failed to recv PROBE: %s", error);
		goto error;
	}

	ok = (reply.hdr.type & EST_TYPE_PROBE);
	if (!ok) {
		rc = EST_EINVAL;
		LOGTR("reply was not a PROBE");
		goto error;
	}

	if (!reply.session_created) {
		rc = EST_ENOTSTARTED;
		goto error;
	}

	if (start_session && !reply.session_started) {
		rc = EST_ENOTSTARTED;
		goto error;
	}

	/* opaque id for session */
	*session_p = reply.session;

	return 0;
error:
	return rc;
}


int
est_pathload_probestart(struct est_channel *dst, void *session, int start)
{
	int rc = 0;
	int ok = 0;
	struct est_probestart_cmd cmd;
	char error[200];

	/* PROBESTART */

	memset(&cmd, 0, sizeof(cmd));
	ok = est_fill_cmd_header(&cmd.hdr, EST_TYPE_PROBESTART);
	BUG_ON(!ok);

	cmd.start = start;
	cmd.session = session;

	LOGFN("sending back PROBESTART, start %d session %p",
		cmd.start, cmd.session);

	rc = est_send_ctrlmsg(dst, &cmd, error, sizeof(error));
	if (rc) {
		LOGTR("failed to send PROBESTART: %s", error);
		goto error;
	}

	return 0;
error:
	return rc;
}


struct est_probe_session *
est_create_probe_session(struct sockaddr_in *src,
			   struct sockaddr_in *dst,
			   int iterate,
                           int iterate_interval_sec,
			   int mgrp,
                           int piggyback,
                           int fast)
{
	int rc = 0;
	struct est_probe_session *ses = NULL;

	ses = malloc(sizeof *ses);
	if (ses == NULL)
		goto error;

	memset(ses, 0, sizeof(*ses));
	ses->src = *src;
	ses->dst = *dst;

	rc = est_open_channel(&ses->ch, &ses->dst);
	if (rc)
		goto error_free;
	
	rc = est_create_pathload_probe_session(&ses->ch,
			&ses->src,
                        iterate, iterate_interval_sec,
                        mgrp, piggyback, fast, 0,
			&ses->session);
	if (rc)
		goto error_close;
	
	return ses;

error_close:
	est_close_channel(&ses->ch);
error_free:
	free(ses);
	LOGTR("libestimator failed: %s", est_strerror(rc));
error:
	return NULL;
}

int
est_start_probe_session(struct est_probe_session *ses)
{
	int rc = 0;

	rc = est_start_pathload_probe_session(&ses->ch, ses->session);
	if (rc) {
		LOGTR("failed to start pathload probe session %p", ses->session);
		goto error;
	}
	return 0;
error:
	LOGTR("libestimator failed: %s", est_strerror(rc));
	return -1;
}
