#ifndef _LIBESTIMATOR_H
#define _LIBESTIMATOR_H

#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <netinet/in.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>

/* libcommon */
#include "atimer.h"
#include "alog.h"
#include "net.h"
#include "ahash.h"
#include "aptr.h"
#include "astring.h"

/* libsystat */
#include "libsysstat.h"

#define EST_DEFAULT_CONTROL_TCP_PORT		9090

#define EST_MAGIC		0xabcd

#define EST_TYPE_HELLO		        (1<<0)
#define EST_TYPE_PEERSTART	        (1<<1)
#define EST_TYPE_PROBE		        (1<<2)
#define EST_TYPE_PROBESTART	        (1<<3)
#define EST_TYPE_TRAFGEN_TRAFFIC	(1<<4)
#ifdef TRAFGEN_REPORT
#define EST_TYPE_TRAFGEN_REPORT		(1<<5)
#endif
#define EST_TYPE_INFO		        (1<<6)
#define EST_TYPE_PROBESTATUS            (1<<7)

#define EST_TYPE_MASK		        0x0fff

#define EST_TYPE_COMMAND	        (1<<12)
#define EST_TYPE_REPLY_FINAL	        (1<<13)
#define EST_TYPE_REPLY_MORE	        (1<<14)

#define EST_INFO_SUBTYPE_LEN            8
// max typelen depends on est_info_header_type_str(): <type> + '-' + <subtype>
#define EST_INFO_TYPELEN                (15 + 1 + EST_INFO_SUBTYPE_LEN)

#define EST_INFO_IDLEN                  8
#define EST_INFO_DEVNAME_LEN            8
#define EST_INFO_DEVTAG_LEN             24
#define EST_INFO_POST_CMDLEN            80
#define EST_INFO_STATS_TAGLEN           20

static inline const char *
est_type_name_str(int type)
{
        if ((type & EST_TYPE_MASK) == EST_TYPE_HELLO)
                return "HELLO";
        if ((type & EST_TYPE_MASK) == EST_TYPE_PEERSTART)
                return "PEERSTART";
        if ((type & EST_TYPE_MASK) == EST_TYPE_PROBE)
                return "PROBE";
        if ((type & EST_TYPE_MASK) == EST_TYPE_PROBESTART)
                return "PROBESTART";
        if ((type & EST_TYPE_MASK) == EST_TYPE_TRAFGEN_TRAFFIC)
                return "TRAFGEN_TRAFFIC";
#ifdef TRAFGEN_REPORT
        if ((type & EST_TYPE_MASK) == EST_TYPE_TRAFGEN_REPORT)
                return "TRAFGEN_REPORT";
#endif
        if ((type & EST_TYPE_MASK) == EST_TYPE_INFO)
                return "INFO";
        if ((type & EST_TYPE_MASK) == EST_TYPE_PROBESTATUS)
                return "PROBESTATUS";

        return "UNKNOWN-TYPE";
}

static inline const char *
est_type_cat_str(int type)
{
        if (type & EST_TYPE_COMMAND)
                return "cmd";
        if (type & EST_TYPE_REPLY_FINAL)
                return "reply [final]";
        if (type & EST_TYPE_REPLY_MORE)
                return "reply [more]";
        
        return "UNKNOWN-CATEGORY";
}

enum est_peertype {
	EST_PEERTYPE_NONE = 0,
	EST_PEERTYPE_PATHLOAD,
};

enum est_channel_state {
	EST_CHANNEL_CLOSED = 0,
	EST_CHANNEL_SOCKET_OPEN,
	EST_CHANNEL_HELLO_PENDING,
	EST_CHANNEL_OPEN,
};

enum est_errors {
	EST_ERROR = 1,
	EST_EINVAL,
	EST_EINVAL_CTRL,
	EST_ENOTOPEN,
	EST_EHELLO,
	EST_ESEND_CTRL,
	EST_ERECV_CTRL,
	EST_ECLOSED_REMOTE,
	EST_ETYPE_MISMATCH,
	EST_ENOTSTARTED,
	EST_ENOTFOUND,
	EST_EEXISTS,
	EST_ERROR_MAX,
};

struct est_channel {
	int state;
	int sock;
	struct sockaddr_in local;
	struct sockaddr_in remote;
	struct timeval timeout;

        struct timeval local_hello_timestamp;
        struct timeval remote_hello_timestamp;
};

static inline int
est_channel_socket(struct est_channel *ch)
{
	if (ch->state < EST_CHANNEL_SOCKET_OPEN)
		return -1;
	return ch->sock;
}

/* TRAFGEN */

enum est_create_trafgen_session_where {
	EST_TRAFGEN_NOWHERE = 0,
	EST_TRAFGEN_SRC,
	EST_TRAFGEN_DST,
};

struct est_trafgen_session {
	struct est_channel ch;
	struct sockaddr_in src;
	struct sockaddr_in dst;
};


/* describe traffic to generate */

/* this is the UDP data packet len */
#define EST_TRAFGEN_MAX_PKTLEN	(1500-20-8)

enum est_trafgen_traffic_type {
	EST_TRAFGEN_NOTYPE = 0,
	EST_TRAFGEN_UDP_NORMAL,
	EST_TRAFGEN_UDP_MGRP,
};

struct est_trafgen_traffic {
	enum est_trafgen_traffic_type type;
	int pktlen;
	int bps;
};

#ifdef TRAFGEN_REPORT
/* report about traffic received */
struct est_trafgen_report {
	int type;
	int is_running;
};
#endif


/* PROBE */

struct est_results {
	int valid;
	double min;
	double max;
	long usec;
	int mgrp;
        int piggyback;
	struct timeval ended;
        int probe_bytes;
        int probe_pkts;
        double adjust_stream;
        double adjust_fleet;
        double orig_min;
        double orig_max;
};

static inline double
est_adjust(double val, double adjust)
{
        double newval = val - adjust;
        if (newval < 0)
                newval = 0;
        return newval;
}

static inline int
est_availbw_results_oneline_str(char *str, int len,
                                struct est_results *results)
{
        return snprintf(str, len,
                        "%d %.2f %.2f %ld mgrp%s p%s %d %d "
                        "adj_stream %.2f %.2f "
                        "adj_fleet %.2f %.2f "
                        "adjust %.2f %.2f "
                        "orig %.2f %.2f ",
                        results->valid,
                        results->min,
                        results->max,
                        results->usec,
                        (results->mgrp ? "ON" : "OFF"),
                        (results->piggyback ? "ON" : "OFF"),
                        results->probe_bytes,
                        results->probe_pkts,

                        est_adjust(results->orig_min, results->adjust_stream),
                        est_adjust(results->orig_max, results->adjust_stream),

                        est_adjust(results->orig_min, results->adjust_fleet),
                        est_adjust(results->orig_max, results->adjust_fleet),

                        results->adjust_stream,
                        results->adjust_fleet,

                        results->orig_min,
                        results->orig_max);
}

struct est_probe_session {
	struct est_channel ch;
	struct sockaddr_in src;
	struct sockaddr_in dst;
	void *session;
};


/* PROBE_STATUS */

enum est_probe_status_value {
        EST_PROBE_STATUS_TERMINATED = 1,
        EST_PROBE_STATUS_SUCCESS = 0,
        EST_PROBE_STATUS_FAILED = -1,
        EST_PROBE_STATUS_TIMEOUT = -2,
        EST_PROBE_STATUS_NOPORTS = -3,
        EST_PROBE_STATUS_ALARM = -4,
        EST_PROBE_STATUS_UNDEF = -999,
};

static inline const char *
est_probe_status_str(enum est_probe_status_value value)
{
        if (value == EST_PROBE_STATUS_TERMINATED)
                return "PROBE_TERMINATED";
        if (value == EST_PROBE_STATUS_SUCCESS)
                return "PROBE_SUCCESS";
        if (value == EST_PROBE_STATUS_FAILED)
                return "PROBE_FAILED";
        if (value == EST_PROBE_STATUS_TIMEOUT)
                return "PROBE_TIMEOUT";
        if (value == EST_PROBE_STATUS_NOPORTS)
                return "PROBE_NOPORTS";
        if (value == EST_PROBE_STATUS_ALARM)
                return "PROBE_ALARM";
        if (value == EST_PROBE_STATUS_UNDEF)
                return "PROBE_UNDEF";

        return "UNKOWN_PROBE_STATUS";
}

struct est_probe_status {
        // ID for aptr_storage
        int id;

        // @value changes are protected by @mutex and signalled by @cond
	enum est_probe_status_value value;
        struct est_results results;

	pthread_mutex_t mutex;
	pthread_cond_t cond;

        // reference counting
        struct aptr ptr;
};

// @refptr is optional
//
// If non-null then it refers to the pointer that is protected by
// the reserve/release methods of @status->ptr. This means that
// when the refcount of @status->ptr reaches zero, @refptr will
// be freed.
static inline void
est_probe_status_init(struct est_probe_status *status, void *refptr)
{
        memset(status, 0, sizeof(struct est_probe_status));

        status->value = EST_PROBE_STATUS_UNDEF;
        pthread_mutex_init(&status->mutex, NULL);
        pthread_cond_init(&status->cond, NULL);

        aptr_init(&status->ptr, refptr);
}

static inline int
est_probe_status_reserve(struct est_probe_status *status)
{
        return aptr_reserve(&status->ptr);
}

static inline int
est_probe_status_release(struct est_probe_status *status)
{
        return aptr_release(&status->ptr);
}

/* COMMANDS and REPLIES */

struct est_header {
	int type;
	int size;
	int magic;
};

/* HELLO */

struct est_hello_cmd {
	/* must be first */
	struct est_header hdr;
        struct timeval now;
};

struct est_hello_reply {
	/* must be first */
	struct est_header hdr;
        struct timeval now;
};

/* PEERSTART */

struct est_peerstart_pathload {
	int peer_tcp_port;
	int peer_udp_port;
	int peer_iterate;
	int peer_mgrp;
        int peer_piggyback;
        int peer_fast;
	long peer_timeout_msec;
};

struct est_peerstart_cmd {
	/* must be first */
	struct est_header hdr;
	enum est_peertype type;
        char dst_id[EST_INFO_IDLEN];

	/* note that this is a union */
	union {
		struct est_peerstart_pathload pathload;
	};
};

struct est_peerstart_reply {
	/* must be first */
	struct est_header hdr;
	int status;
        char src_id[EST_INFO_IDLEN];
};

/* PROBE */

/* requests pathload between e2 -> e1, command sent to e1 */
struct est_probe_cmd {
	/* must be first */
	struct est_header hdr;
	struct {
		struct sockaddr_in e2addr;
		int iterate;
                int iterate_interval_sec;
		int mgrp;
                int piggyback;
                int fast;
		int start_session;
	};
};

struct est_probe_reply {
	/* must be first */
	struct est_header hdr;
	struct {
		void *session;
		int session_created;
		int session_started;
		int valid_results;
		struct est_results results;
	};
};

/* PROBESTART -- no reply */

/* XXX only one START and only one STOP please; shaky command */
/* Thread disappears after STOP, session pointer invalid */

struct est_probestart_cmd {
	/* must be first */
	struct est_header hdr;
	struct {
		int start;
		void *session;
	};
};

/* PROBESTATUS -- no reply */

// Sent by a probe sub-process (like pathload_rcv) to the estimator
// to signal a change of status. On receipt, the estimator wakes up
// the thread through the condition variable. It finds the condition
// variable based on the status_id.
struct est_probestatus_cmd {
        /* must be first */
        struct est_header hdr;

        int status_id;
        enum est_probe_status_value status_value;
        struct est_results status_results;
};

/* TRAFGEN_TRAFFIC */

enum est_trafgen_traffic_cmd_action {
	EST_TRAFGEN_NOACTION = 0,
	EST_TRAFGEN_PREPARE,
	EST_TRAFGEN_START,
	EST_TRAFGEN_STOP,
};

struct est_trafgen_stats {
        struct sockaddr_in src;
        struct sockaddr_in dst;
	enum est_trafgen_traffic_type type;
        char src_id[EST_INFO_IDLEN];
        char dst_id[EST_INFO_IDLEN];

        int is_sender;

        int packets;
        int bytes;
        int pktlen_max;
        int pktlen_min;

        struct timeval tstamp_start;
        struct timeval tstamp_end;
};

void est_trafgen_stats_oneline_str(char *str, int len,
                                   struct est_trafgen_stats *stats);

static inline void
est_info_extract_id_from_hostname(const char *hostname, char *id, int idlen)
{
        char *dot = NULL;

        /* use as id only the first part of the hostname up to the dot */
        strncpy(id, hostname, idlen);
        id[idlen-1] = 0;
        dot = strchr(id, '.');
        if (dot != NULL)
                * dot = 0;
}

static inline void
est_get_host_id(char *id, int idlen)
{
        char hostname[100];
        int rc = gethostname(hostname, sizeof(hostname));
        SYSERR_ON(rc);
        est_info_extract_id_from_hostname(hostname, id, idlen);
}

static inline void
est_trafgen_stats_reset(struct est_trafgen_stats *stats, struct timeval *now)
{
        stats->packets = 0;
        stats->bytes = 0;
        stats->pktlen_max = 0;
        stats->pktlen_min = INT_MAX;

        if (now != NULL) {
                stats->tstamp_start = *now;
                stats->tstamp_end = *now;
        } else {
                memset(&stats->tstamp_start, 0, sizeof(struct timeval));
                memset(&stats->tstamp_end, 0, sizeof(struct timeval));
        }
}

static inline void
est_trafgen_stats_init(struct est_trafgen_stats *stats)
{
        memset(stats, 0, sizeof(struct est_trafgen_stats));
        est_trafgen_stats_reset(stats, NULL);
}

static inline void
est_trafgen_stats_log(struct est_trafgen_stats *stats)
{
        double pps = 0;
        double mbps = 0;
        const char *prefix0 = "trafgen_stats:";
        const char *prefix  = "trafgen_stats:      ";

        char tmp[200];
        long long usec_diff = 0;
        int sec_diff_approx = 0;
        double sec_diff = 0;

        usec_diff = atimer_timeval_diff_usec(&stats->tstamp_end,
                                             &stats->tstamp_start);
        sec_diff = (1.0 * usec_diff) / ATIMER_USEC_PER_SEC;

        pps = (1.0 * stats->packets) / sec_diff;
        mbps = (8.0 * stats->bytes) / (sec_diff * 1000000);
        net_srcdst_str(tmp, sizeof(tmp), &stats->src, &stats->dst);

        LOG("%s %s type 0x%x", prefix0, tmp, stats->type);
        LOG("%s src_id %s dst_id %s", prefix, stats->src_id, stats->dst_id);
        LOG("%s packets %d bytes %d min %d max %d pps %f mbps %f",
            prefix,
            stats->packets,
            stats->bytes,
            stats->pktlen_min,
            stats->pktlen_max,
            pps,
            mbps);

        sec_diff_approx = usec_diff / ATIMER_USEC_PER_SEC;

        LOG("%s last updated %lld usec ago (%d sec approx)",
            prefix, usec_diff, sec_diff_approx);
}

/* requests traffic generation between @src -> @dst */
/* command sent to estimator running on @src */
struct est_trafgen_traffic_cmd {
	/* must be first */
	struct est_header hdr;
	struct sockaddr_in src;
	struct sockaddr_in dst;
	enum est_trafgen_traffic_cmd_action action;
	/* @traffic is only read if action == EST_TRAFGEN_START */
	struct est_trafgen_traffic traffic;
};

struct est_trafgen_traffic_reply {
	/* must be first */
	struct est_header hdr;
	int success;
	char error[200];
};

#ifdef TRAFGEN_REPORT
/* TRAFGEN_REPORT */

/* requests traffic generation reports for traffic between @src -> @dst */
/* command sent to estimator running on @dst */
struct est_trafgen_report_cmd {
	/* must be first */
	struct est_header hdr;
	struct sockaddr_in src;
	struct sockaddr_in dst;
};

struct est_trafgen_report_reply {
	/* must be first */
	struct est_header hdr;
	int success;
	char error[200];
	int has_report;
	struct est_trafgen_report report;
};
#endif

/* INFO */

enum est_info_header_type {
        EST_INFO_NOTYPE = 0,
        EST_INFO_TRAFGEN,
        EST_INFO_MPSTAT,
        EST_INFO_AVAILBW_PATHLOAD,
        EST_INFO_TRACE,
        EST_INFO_MEDIANET,
        EST_INFO_DEV,
        EST_INFO_MGRP,
        EST_INFO_POST,
        EST_INFO_STATS,
};

struct est_info_header_key {
        enum est_info_header_type type;
        char subtype[EST_INFO_SUBTYPE_LEN];

        /* id of node reporting this information */
        char node_id[EST_INFO_IDLEN];
        /* id of the link src */
        char src_id[EST_INFO_IDLEN];
        /* id of the link dst; empty if not a link information */
        char dst_id[EST_INFO_IDLEN];
};

static inline void
est_info_header_type_str(char *str, int len,
                         const struct est_info_header_key *key)
{
        int needed_len = 0;
        const char *type_str = "unknown-type";

        if (key->type == EST_INFO_TRAFGEN)
                type_str = "trafgen";
        else if (key->type == EST_INFO_MPSTAT)
                type_str = "mpstat";
        else if (key->type == EST_INFO_AVAILBW_PATHLOAD)
                type_str = "pathload";
        else if (key->type == EST_INFO_TRACE)
                type_str = "trace";
        else if (key->type == EST_INFO_MEDIANET)
                type_str = "medianet";
        else if (key->type == EST_INFO_DEV)
                type_str = "dev";
        else if (key->type == EST_INFO_MGRP)
                type_str = "mgrp";
        else if (key->type == EST_INFO_POST)
                type_str = "post";
        else if (key->type == EST_INFO_STATS)
                type_str = "stats";

        if (strlen(key->subtype) > 0)
                needed_len = snprintf(str, len, "%s-%s", type_str, key->subtype);
        else
                needed_len = snprintf(str, len, "%s", type_str);

        BUG_ON(needed_len >= len);
}

static inline int
est_info_header_key_cmp(struct est_info_header_key *lhs,
                        struct est_info_header_key *rhs)
{
        int rc = 0;

        if (lhs->type != rhs->type)
                return (lhs->type < rhs->type ? -1 : 1);

        if ((rc = strcmp(lhs->subtype, rhs->subtype)) != 0)
                return rc;
        if ((rc = strcmp(lhs->node_id, rhs->node_id)) != 0)
                return rc;
        if ((rc = strcmp(lhs->src_id, rhs->src_id)) != 0)
                return rc;
        if ((rc = strcmp(lhs->dst_id, rhs->dst_id)) != 0)
                return rc;

        return 0;
}

static inline uint64_t
est_info_header_key_hash(const struct est_info_header_key *key)
{
        int i = 0;
        int b = 0;

        uint64_t hkey = 0;
        uint8_t *bytes = (uint8_t *)&hkey;

        /* @b keeps increasing and is always used modulo 8 */
        for (b = 0; b < 8; b++)
                bytes[b % 8] = key->type << b;

        for (i = 0; i < 4; i++) {
                int c = 0;

                /* rotate over the strings in the key */
                const char *str = (i == 0 ? key->subtype :
                                   i == 1 ? key->node_id :
                                   i == 2 ? key->src_id :
                                   i == 3 ? key->dst_id :
                                   0);

                /* use only character before the \0 NULL character */
                while (str[c] != 0 && c < EST_INFO_IDLEN) {
                        bytes[b % 8] |= str[c];
                        c++;
                        b++;
                }
        }

        hkey = ahash_mix64(hkey);
        DEB("+++ hkey 0x%llx hkeyp %p bytes %p", hkey, &hkey, bytes);
        return hkey;
}

struct est_info_header {
        int valid;
        struct est_info_header_key key;

        char timestamp_id[EST_INFO_IDLEN];
	int timestamp_valid;
	struct timeval timestamp;
};

// If @timestamp is NULL, fill with current time
static inline void
est_fill_info_timestamp(struct est_info_header *hdr,
                        const char *hostid, const struct timeval *timestamp)
{
        if (timestamp ==  NULL)
                gettimeofday(&hdr->timestamp, NULL);
        else
                hdr->timestamp = *timestamp;

        astring_strncpy(hdr->timestamp_id, hostid, sizeof(hdr->timestamp_id));

        hdr->timestamp_valid = 1;
}

static inline void
est_fill_info_timestamp_now(struct est_info_header *hdr, const char *hostid)
{
        est_fill_info_timestamp(hdr, hostid, NULL);
}

static inline void
est_fill_info_srcdst(struct est_info_header *hdr,
                     const char *src, const char *dst)
{
        int len = 0;
        char *str = NULL;

        str = hdr->key.src_id;
        len = sizeof(hdr->key.src_id);
        strncpy(str, src, len);
        str[len-1] = 0;

        str = hdr->key.dst_id;
        len = sizeof(hdr->key.dst_id);
        strncpy(str, dst, len);
        str[len-1] = 0;
}

/* INFO/trafgen (traffic generation) */

struct est_info_trafgen {
        /* must be first */
        struct est_info_header hdr;
        struct est_trafgen_stats stats;
};

static inline void
est_info_trafgen_init(struct est_info_trafgen *info)
{
        memset(info, 0, sizeof(struct est_info_trafgen));
        info->hdr.valid = 1;
        info->hdr.key.type = EST_INFO_TRAFGEN;
        est_trafgen_stats_init(&info->stats);
}

/* INFO/mpstat (cpu load) */

struct est_info_mpstat {
        /* must be first */
        struct est_info_header hdr;
        struct mp_stats_ratio_int iratio;
};

static inline void
est_info_mpstat_init(struct est_info_mpstat *info)
{
        memset(info, 0, sizeof(struct est_info_mpstat));
        info->hdr.valid = 1;
        info->hdr.key.type = EST_INFO_MPSTAT;
}

/* INFO/availbw (uses for all the EST_INFO_AVAILBW_xxxx types */
struct est_info_availbw {
        /* must be first */
        struct est_info_header hdr;
        struct est_results results;
};


static inline void
est_info_availbw_init(struct est_info_availbw *info,
                      enum est_info_header_type type)
{
        /* append one check for each type */
        int valid_type = (type == EST_INFO_AVAILBW_PATHLOAD);

        BUG_ON(!valid_type);

        memset(info, 0, sizeof(struct est_info_availbw));
        info->hdr.valid = 1;
        info->hdr.key.type = type;
}

/* INFO/trace */

enum est_trace_log_decode_type {
        EST_TRACE_DECODE_OK = 1,
        EST_TRACE_DECODE_BAD = 0,
        EST_TRACE_DECODE_UNDEF = -1,
};

static inline const char *
est_trace_log_decode_str(enum est_trace_log_decode_type decode)
{
        if (decode == EST_TRACE_DECODE_OK)
                return "ok";
        if (decode == EST_TRACE_DECODE_BAD)
                return "bad";
        if (decode == EST_TRACE_DECODE_UNDEF)
                return "undef";

        // unknown decode result
        BUG();
        return "BUG";
}

struct est_trace_log {
        // See code that prints packet info in main() in:
        // medianet/media-net/src/apps/tracereceiver.cyc
        // Look for "seq " (with the trailing space).
        int port;
        int seq;
        int bytes;
        int temporal_place;
        enum est_trace_log_decode_type decode;
        double bps;
        char frame_type;
};

static inline int
est_trace_log_oneline_str(char *str, int len, struct est_trace_log *log)
{
        return snprintf(str, len,
                        "%d %d %d %.2f %c %d %s",
                        log->port,
                        log->seq,
                        log->bytes,
                        log->bps,
                        log->frame_type,
                        log->temporal_place,
                        est_trace_log_decode_str(log->decode));
}

struct est_info_trace {
        /* must be first */
        struct est_info_header hdr;
        struct est_trace_log log;
};

static inline void
est_info_trace_init(struct est_info_trace *info)
{
        memset(info, 0, sizeof(struct est_info_trace));
        info->hdr.valid = 1;
        info->hdr.key.type = EST_INFO_TRACE;
}

/* INFO/medianet */

enum est_medianet_results_type {
        EST_MEDIANET_CAPACITY = 1,
        EST_MEDIANET_PATHLOAD,
        EST_MEDIANET_LSUPDATE,
};

static inline const char *
est_medianet_results_type_str(enum est_medianet_results_type type)
{
        if (type == EST_MEDIANET_CAPACITY)
                return "capacity";
        if (type == EST_MEDIANET_PATHLOAD)
                return "pathload";
        if (type == EST_MEDIANET_LSUPDATE)
                return "lsupdate";

        // unknown
        BUG();
        return "BUG";
}

struct est_medianet_results {
        enum est_medianet_results_type type;
        char name[2 * EST_INFO_IDLEN];
        // this is the avail bw on the network
        union {
                struct {
                        double mbps;
                } capacity;

                struct {
                        double succ_mbps;
                        double try_mbps;
                } lsupdate;

                struct {
                        double min_mbps;
                        double max_mbps;
                } pathload;
        };
};

static inline int
est_medianet_results_oneline_str(char *str, int len,
                                 struct est_medianet_results *results)
{
        const char *type_str = est_medianet_results_type_str(results->type);

        if (results->type == EST_MEDIANET_CAPACITY)
                return snprintf(str, len,
                                "%s %s %.2f 0.00",
                                results->name,
                                type_str,
                                results->capacity.mbps);

        if (results->type == EST_MEDIANET_LSUPDATE)
                return snprintf(str, len,
                                "%s %s %.2f %.2f",
                                results->name,
                                type_str,
                                results->lsupdate.succ_mbps,
                                results->lsupdate.try_mbps);

        if (results->type == EST_MEDIANET_PATHLOAD)
                return snprintf(str, len,
                                "%s %s %.2f %.2f",
                                results->name,
                                type_str,
                                results->pathload.min_mbps,
                                results->pathload.max_mbps);
        BUG();
        return 0;
}

struct est_info_medianet {
        /* must be first */
        struct est_info_header hdr;
        struct est_medianet_results results;
};

static inline void
est_info_medianet_init(struct est_info_medianet *info)
{
        memset(info, 0, sizeof(struct est_info_medianet));
        info->hdr.valid = 1;
        info->hdr.key.type = EST_INFO_MEDIANET;
}

/* INFO/dev */

struct est_dev_stats {
        char name[EST_INFO_DEVNAME_LEN];
        char tag[EST_INFO_DEVTAG_LEN];
        double rxbps;
        double rxpps;
        double txbps;
        double txpps;
};

static inline int
est_dev_stats_oneline_str(char *str, int len, struct est_dev_stats *stats)
{
        return snprintf(str, len,
                        "%s %s RX %12.2f %8.2f TX %12.2f %8.2f",
                        stats->name,
                        stats->tag,
                        stats->rxbps,
                        stats->rxpps,
                        stats->txbps,
                        stats->txpps);
}

struct est_info_dev {
        /* must be first */
        struct est_info_header hdr;
        struct est_dev_stats stats;
};

static inline void
est_info_dev_init(struct est_info_dev *info)
{
        memset(info, 0, sizeof(struct est_info_dev));
        info->hdr.valid = 1;
        info->hdr.key.type = EST_INFO_DEV;
}

/* INFO/mgrp */

struct est_mgrp_stats {
        char dst[EST_INFO_IDLEN];

        float mgrp_pps;
        float mgrp_bps_header;
        float mgrp_bps_payload;

        float tcp_pps;
        float tcp_bps;

        float udp_pps;
        float udp_bps;

        float probe_pps;
        float probe_bps_actual;
        float probe_bps_orig;

        float probe_pps_reused;
        float probe_bps_nonpad;
        float probe_bps_pad_unused;
        float probe_bps_pad_reused;

        struct {
                float mgrp_pps;
                float mgrp_lost_pps;
                float mgrp_missing_pps;
                float mgrp_outoforder_pps;
        } recv;
};

static inline int
est_mgrp_stats_oneline_str(char *str, int len, struct est_mgrp_stats *stats)
{
        return snprintf(str, len,
                        "%s "
                        "mgrp %8.2f %12.2f %12.2f "
                        "tcp %8.2f %12.2f "
                        "udp %8.2f %12.2f "
                        "probe %8.2f %12.2f %12.2f "
                        "pad %8.2f %12.2f %12.2f %12.2f "
                        "recv %8.2f %8.2f %8.2f %8.2f", 

                        stats->dst,

                        stats->mgrp_pps,
                        stats->mgrp_bps_header,
                        stats->mgrp_bps_payload,

                        stats->tcp_pps,
                        stats->tcp_bps,

                        stats->udp_pps,
                        stats->udp_bps,

                        stats->probe_pps,
                        stats->probe_bps_actual,
                        stats->probe_bps_orig,

                        stats->probe_pps_reused,
                        stats->probe_bps_nonpad,
                        stats->probe_bps_pad_unused,
                        stats->probe_bps_pad_reused,

                        stats->recv.mgrp_pps,
                        stats->recv.mgrp_lost_pps,
                        stats->recv.mgrp_missing_pps,
                        stats->recv.mgrp_outoforder_pps);
}

struct est_info_mgrp {
        /* must be first */
        struct est_info_header hdr;
        struct est_mgrp_stats stats;
};

static inline void
est_info_mgrp_init(struct est_info_mgrp *info)
{
        memset(info, 0, sizeof(struct est_info_mgrp));
        info->hdr.valid = 1;
        info->hdr.key.type = EST_INFO_MGRP;
}

/* INFO/post */

struct est_info_post {
        /* must be first */
        struct est_info_header hdr;
        char post_cmd[EST_INFO_POST_CMDLEN];
};

static inline void
est_info_post_init(struct est_info_post *info)
{
        memset(info, 0, sizeof(struct est_info_post));
        info->hdr.valid = 1;
        info->hdr.key.type = EST_INFO_POST;
}

static inline int
est_post_oneline_str(char *str, int len, struct est_info_post *post)
{
        return snprintf(str, len, "cmd: %s", post->post_cmd);
}

/* INFO/stats */
/* Generic INFO structure to hold counters and ratios */

#define EST_INFO_STATS_STRLEN \
        (cmp_max(sizeof(struct timeval), \
         cmp_max(sizeof(long long), \
                 sizeof(double))))

union est_stat {
        long long int ival;
        double dval;
        char str[EST_INFO_STATS_STRLEN];
        struct timeval tval;
        // this consumes the remaining fields in the stats array
        char bigstr[0];
};

#define EST_INFO_STATS_MAX      16

struct est_info_stats {
        /* must be first */
        struct est_info_header hdr;

        char tag[EST_INFO_STATS_TAGLEN];

        unsigned short imask;
        unsigned short dmask;
        unsigned short tmask;
        unsigned short smask;
        unsigned short has_bigstr;
        unsigned short bigstr_pos;

        union est_stat stats[EST_INFO_STATS_MAX];
};

static inline void
est_info_stats_init(struct est_info_stats *stats, const char *subtype)
{
        memset(stats, 0, sizeof(struct est_info_stats));
        stats->hdr.valid = 1;
        stats->hdr.key.type = EST_INFO_STATS;

        BUG_ON(strlen(subtype) >= sizeof(stats->hdr.key.subtype));
        strncpy(stats->hdr.key.subtype, subtype, sizeof(stats->hdr.key.subtype));
}

static inline int
est_info_stats_pos_is_int(struct est_info_stats *stats, int pos)
{
        unsigned short posmask = (1 << pos);
        return (stats->imask & posmask);
}

static inline int
est_info_stats_pos_is_double(struct est_info_stats *stats, int pos)
{
        unsigned short posmask = (1 << pos);
        return (stats->dmask & posmask);
}

static inline int
est_info_stats_pos_is_timeval(struct est_info_stats *stats, int pos)
{
        unsigned short posmask = (1 << pos);
        return (stats->tmask & posmask);
}

static inline int
est_info_stats_pos_is_str(struct est_info_stats *stats, int pos)
{
        unsigned short posmask = (1 << pos);
        return (stats->smask & posmask);
}

// returns 1 for every position the bigstr occupies
static inline int
est_info_stats_pos_is_bigstr(struct est_info_stats *stats, int pos)
{
        return (stats->has_bigstr && pos >= stats->bigstr_pos);
}

static inline int
est_info_stats_pos_occupied(struct est_info_stats *stats, int pos)
{
        return est_info_stats_pos_is_int(stats, pos)    ||
               est_info_stats_pos_is_double(stats, pos) ||
               est_info_stats_pos_is_timeval(stats, pos)||
               est_info_stats_pos_is_str(stats, pos)    ||
               est_info_stats_pos_is_bigstr(stats, pos);
}

//
// NOTE1:
// The number of values is not maxpos+1, since there can
// be gaps in the values (not typical, but could be for
// reserved values).
//
// NOTE2:
// If there is a bigstr, the max_occupied_pos is the position
// where the bigstr starts. There is an inconsitency with
// est_info_stats_pos_occupied() that will return 1 even
// for positions > maxpos when a bigstr is present.
//
static inline int
est_info_stats_max_occupied_pos(struct est_info_stats *stats)
{
        int maxpos = -1;
        unsigned int mask_with_maxpos = stats->imask |
                                        stats->dmask |
                                        stats->tmask |
                                        stats->smask;
        
        while (mask_with_maxpos > 0) {
                maxpos++;
                mask_with_maxpos >>= 1;
        }

        if (stats->has_bigstr) {
                BUG_ON(maxpos >= stats->bigstr_pos);
                maxpos = stats->bigstr_pos;
        }

        BUG_ON(maxpos >= EST_INFO_STATS_MAX);
        return maxpos;
}

// first valid @pos is 0, max valid @pos is EST_INFO_STATS_MAX-1
static inline void
est_info_stats_setint(struct est_info_stats *stats, int pos, int val)
{
        unsigned short posmask = (1 << pos);

        BUG_ON(pos >= 8*sizeof(unsigned short));
        BUG_ON(est_info_stats_pos_occupied(stats, pos));

        stats->imask |= posmask;
        stats->stats[pos].ival = val;
}

// first valid @pos is 0, max valid @pos is EST_INFO_STATS_MAX-1
static inline void
est_info_stats_setdouble(struct est_info_stats *stats, int pos, double val)
{
        unsigned short posmask = (1 << pos);

        BUG_ON(pos >= 8*sizeof(unsigned short));
        BUG_ON(est_info_stats_pos_occupied(stats, pos));

        stats->dmask |= posmask;
        stats->stats[pos].dval = val;
}

// first valid @pos is 0, max valid @pos is EST_INFO_STATS_MAX-1
static inline void
est_info_stats_settimeval(struct est_info_stats *stats, int pos,
                          struct timeval *val)
{
        unsigned short posmask = (1 << pos);

        BUG_ON(pos >= 8*sizeof(unsigned short));
        BUG_ON(est_info_stats_pos_occupied(stats, pos));

        stats->tmask |= posmask;
        stats->stats[pos].tval = *val;
}

// first valid @pos is 0, max valid @pos is EST_INFO_STATS_MAX-1
static inline void
est_info_stats_setstr(struct est_info_stats *stats, int pos, const char *str)
{
        unsigned short posmask = (1 << pos);

        BUG_ON(pos >= 8*sizeof(unsigned short));
        BUG_ON(est_info_stats_pos_occupied(stats, pos));

        stats->smask |= posmask;
        snprintf(stats->stats[pos].str, EST_INFO_STATS_STRLEN, "%s", str);
}

static inline int
est_info_stats_bigstr_maxlen(int pos)
{
        // bigstr occupies all positions from pos..EST_INFO_STATS_MAX-1
        int occupied_positions = EST_INFO_STATS_MAX - pos;
        return (occupied_positions * sizeof(union est_stat));
}

// first valid @pos is 0, max valid @pos is EST_INFO_STATS_MAX-1
static inline void
est_info_stats_setbigstr(struct est_info_stats *stats, int pos, const char *bigstr)
{
        int maxlen = est_info_stats_bigstr_maxlen(pos);

        BUG_ON(pos >= 8*sizeof(unsigned short));
        BUG_ON(est_info_stats_pos_occupied(stats, pos));

        stats->has_bigstr = 1;
        stats->bigstr_pos = pos;
        snprintf(stats->stats[pos].bigstr, maxlen, "%s", bigstr);
}

static inline int
est_stats_oneline_str(char *str, int len, struct est_info_stats *stats)
{
        int n = 0;
        int i = 0;
        char *buf = str;
        int remaining = len;
        int maxpos = est_info_stats_max_occupied_pos(stats);

        n = snprintf(buf, remaining, "%s", stats->tag);
        BUG_ON(n >= remaining);
        buf += n;
        remaining -= n;

        for (i = 0; i <= maxpos; i++) {
                if (est_info_stats_pos_is_int(stats, i))
                        n = snprintf(buf, remaining, " %lld",
                                     stats->stats[i].ival);

                else if (est_info_stats_pos_is_double(stats, i))
                        n = snprintf(buf, remaining, " %.2lf",
                                     stats->stats[i].dval);

                else if (est_info_stats_pos_is_timeval(stats, i))
                        n = snprintf(buf, remaining, " %ld.%.6ld",
                                     stats->stats[i].tval.tv_sec,
                                     stats->stats[i].tval.tv_usec);

                else if (est_info_stats_pos_is_str(stats, i))
                        n = snprintf(buf, remaining, " %s",
                                     stats->stats[i].str);

                else if (est_info_stats_pos_is_bigstr(stats, i))
                        n = snprintf(buf, remaining, " %s",
                                     stats->stats[i].bigstr);

                else
                        n = snprintf(buf, remaining, " nil");

                BUG_ON(n >= remaining);
                buf += n;
                remaining -= n;
        }

        return (len - remaining);
}

/* INFO union */

/* note that this is a union! All structs share the same header */
struct est_info {
        union {
                struct est_info_header hdr;
                struct est_info_trafgen trafgen;
                struct est_info_mpstat mpstat;
		struct est_info_availbw availbw;
                struct est_info_trace trace;
                struct est_info_medianet medianet;
                struct est_info_dev dev;
                struct est_info_mgrp mgrp;
                struct est_info_post post;
                struct est_info_stats stats;
        };
};

static inline void
est_info_sizes_log() {
        LOG("");
        LOG(">>> sizes of INFO structures");
        LOG("est_info (union)      size %d bytes", sizeof(struct est_info));
        LOG("est_info_header       size %d bytes", sizeof(struct est_info_header));
        LOG("est_info_trafgen      size %d bytes", sizeof(struct est_info_trafgen));
        LOG("est_info_mpstat       size %d bytes", sizeof(struct est_info_mpstat));
        LOG("est_info_availbw      size %d bytes", sizeof(struct est_info_availbw));
        LOG("est_info_trace        size %d bytes", sizeof(struct est_info_trace));
        LOG("est_info_medianet     size %d bytes", sizeof(struct est_info_medianet));
        LOG("est_info_dev          size %d bytes", sizeof(struct est_info_dev));
        LOG("est_info_mgrp         size %d bytes", sizeof(struct est_info_mgrp));
        LOG("est_info_post         size %d bytes", sizeof(struct est_info_post));
        LOG("est_info_stats        size %d bytes", sizeof(struct est_info_stats));
        LOG("est_info_stats/str    size %d bytes", EST_INFO_STATS_STRLEN);
        LOG("est_info_stats/bigstr size %d bytes", est_info_stats_bigstr_maxlen(0));
}

static inline int
est_info_timestamp_str(char *str, int len, struct est_info *info)
{
        int written = 0;

        if (info->hdr.timestamp_valid) {
                char *idstr = info->hdr.timestamp_id;
                int sec = info->hdr.timestamp.tv_sec;
                int usec = info->hdr.timestamp.tv_usec;

                written = snprintf(str, len, "%s %d.%.6d", idstr, sec, usec);
        } else {
                written = snprintf(str, len, "0.0.0.0 0.000000");
        }
        return written;
}

static inline int
est_info_header_str(char *str, int len, struct est_info *info)
{
        int written = 0;
        int str_used = 0;

        written = est_info_timestamp_str(&str[str_used], len - str_used, info);
        BUG_ON(written >= (len - str_used));
        str_used += written;

        written = snprintf(&str[str_used], len - str_used,
                          " node_id <%s>"
                          " src_id <%s>"
                          " dst_id <%s>",
                          info->hdr.key.node_id,
                          info->hdr.key.src_id,
                          info->hdr.key.dst_id);

        return written;
}

static inline int
est_info_header_key_str(char *str, int len, struct est_info_header_key *key)
{
        char type_str[EST_INFO_TYPELEN];

        est_info_header_type_str(type_str, sizeof(type_str), key);
        return snprintf(str, len,
                        " type %s"
                        " node_id <%s>"
                        " src_id <%s>"
                        " dst_id <%s>",
                        type_str,
                        key->node_id,
                        key->src_id,
                        key->dst_id);
}

static inline void
est_info_oneline_str(char *str, int len, struct est_info *info)
{
        char timestamp[100];
        char oneliner[300];

        est_info_timestamp_str(timestamp, sizeof(timestamp), info);

        if (info->hdr.key.type == EST_INFO_TRAFGEN) {
                est_trafgen_stats_oneline_str(oneliner, sizeof(oneliner),
                                              &info->trafgen.stats);

        } else if (info->hdr.key.type == EST_INFO_MPSTAT) {
                /* get only the summary entry; entry 0 */
                mpstat_ratio_int_oneline_str(oneliner, sizeof(oneliner),
                                             &info->mpstat.iratio, 0);

        } else if (info->hdr.key.type == EST_INFO_AVAILBW_PATHLOAD) {
                est_availbw_results_oneline_str(oneliner, sizeof(oneliner),
                                                &info->availbw.results);

        } else if (info->hdr.key.type == EST_INFO_TRACE) {
                est_trace_log_oneline_str(oneliner, sizeof(oneliner),
                                          &info->trace.log);

        } else if (info->hdr.key.type == EST_INFO_MEDIANET) {
                est_medianet_results_oneline_str(oneliner, sizeof(oneliner),
                                                 &info->medianet.results);

        } else if (info->hdr.key.type == EST_INFO_DEV) {
                est_dev_stats_oneline_str(oneliner, sizeof(oneliner),
                                          &info->dev.stats);

        } else if (info->hdr.key.type == EST_INFO_MGRP) {
                est_mgrp_stats_oneline_str(oneliner, sizeof(oneliner),
                                           &info->mgrp.stats);

        } else if (info->hdr.key.type == EST_INFO_POST) {
                est_post_oneline_str(oneliner, sizeof(oneliner), &info->post);

        } else if (info->hdr.key.type == EST_INFO_STATS) {
                est_stats_oneline_str(oneliner, sizeof(oneliner), &info->stats);


        } else {
                // unknown key.type
                BUG();
        }
        
        snprintf(str, len, "%s %s", timestamp, oneliner);
}

static inline void
est_info_log(struct est_info *info)
{
        char header[200];
        char tmp[200];

        est_info_header_key_str(header, sizeof(header), &info->hdr.key);
        est_info_oneline_str(tmp, sizeof(tmp), info);

        LOG("%s %s", header, tmp);
}


#define EST_INFO_ENTRIES_PER_MSG        5

enum est_info_cmd_action {
        EST_INFO_NOACTION = 0,
        EST_INFO_GET,
        EST_INFO_PUT,
};

struct est_info_cmd {
	/* must be first */
	struct est_header hdr;

        enum est_info_cmd_action action;
        /* used only when action is a PUT */
        struct est_info info[EST_INFO_ENTRIES_PER_MSG];
        int info_entries;
        /* used only when action is a GET */
        int info_max_entries_to_get;
};

struct est_info_reply {
	/* must be first */
	struct est_header hdr;

        enum est_info_cmd_action action;
        /* used only when action is a GET */
        struct est_info info[EST_INFO_ENTRIES_PER_MSG];
        int info_entries;
};

/* union of all commands and replies */

typedef union est_cmd {
	struct est_hello_cmd hello;
	struct est_peerstart_cmd peerstart;
	struct est_probe_cmd probe;
	struct est_probestart_cmd probestart;
	struct est_trafgen_traffic_cmd trafgen_traffic;
#ifdef TRAFGEN_REPORT
	struct est_trafgen_report_cmd trafgen_report;
#endif
        struct est_info_cmd info;
        struct est_probestatus_cmd probestatus;
	struct {
		struct est_header hdr;
		unsigned char data[0];
	};
} est_cmd;

typedef union est_reply {
	struct est_hello_reply hello;
	struct est_peerstart_reply peerstart;
	struct est_probe_reply probe;
	struct est_trafgen_traffic_reply trafgen_traffic;
#ifdef TRAFGEN_REPORT
	struct est_trafgen_report_reply trafgen_report;
#endif
        struct est_info_reply info;
	struct {
		struct est_header hdr;
		unsigned char data[0];
	};
} est_reply;


#ifdef DEBUG

static inline void
__est_loghdr(struct est_header *hdr, int sent,
	char *file, int line, char *function)
{
	if (sent)
		printf("<<< hdr: at %s:%d %s\n", file, line, function);
	else
		printf(">>> hdr: at %s:%d %s\n", file, line, function);
	printf("    hdr: type  0x%x %s %s\n",
		hdr->type,
		est_type_name_str(hdr->type),
		est_type_cat_str(hdr->type));
	printf("    hdr: size  %d\n", hdr->size);
	printf("    hdr: magic 0x%x\n", hdr->magic);
}

#define EST_LOGHDR_SND(_h)      __est_loghdr(_h, 1, __FILE__, __LINE__, __FUNCTION__)
#define EST_LOGHDR_RCV(_h)      __est_loghdr(_h, 0, __FILE__, __LINE__, __FUNCTION__)

#else

#define EST_LOGHDR_SND(_h)      ((void)(0))
#define EST_LOGHDR_RCV(_h)      ((void)(0))

#endif

#define EST_LOGRESULTS(_r) \
	printf("result: %s %d %s: %ld.%.6ld sec %.2f max %.2f min %.2f\n", \
		((_r)->valid ? "valid" : "INVALID"), \
		(_r)->valid, \
		((_r)->mgrp ? "mgrp" : "orig"), \
		(_r)->ended.tv_sec, \
		(_r)->ended.tv_usec, \
		0.000001 * (_r)->usec, \
		(_r)->max, \
		(_r)->min)


int est_open_channel(struct est_channel *ch, struct sockaddr_in *remote);
int est_accept_channel(int sock, struct est_channel *newch);
void est_close_channel(struct est_channel *ch);

int est_send_ctrlmsg(struct est_channel *ch,
		void *ctrlmsg,
		char *err_str, int err_len);

int est_recv_ctrlmsg(struct est_channel *ch,
		void *ctrlmsg,
		int ctrllen,
		char *err_str, int err_len);

static inline const char *
est_strerror(int errnum)
{
	const char *err = "unknown error";

	if (errnum < 0)
		/* XXX not re-entrant! */
		err = strerror(-errnum);
	else if (errnum == 0)
		err = "no error";
	else if (errnum == EST_ERROR)
		err = "error";
	else if (errnum == EST_EINVAL)
		err = "invalid argument";
	else if (errnum == EST_ENOTOPEN)
		err = "channel not open";
	else if (errnum == EST_EHELLO)
		err = "hello exchange failed";
	else if (errnum == EST_ESEND_CTRL)
		err = "failed to send control message";
	else if (errnum == EST_ERECV_CTRL)
		err = "failed to receive control message";
	else if (errnum == EST_ECLOSED_REMOTE)
		err = "remote side closed";
	else if (errnum == EST_ETYPE_MISMATCH)
		err = "message has the wrong type";
	else if (errnum == EST_ENOTSTARTED)
		err = "failed to start";
	else if (errnum == EST_ENOTFOUND)
		err = "not found";
	else if (errnum == EST_EEXISTS)
		err = "already exists";

	return err;
}


static inline int
est_is_reply(struct est_header *hdr)
{
	return(hdr->type & (EST_TYPE_REPLY_FINAL | EST_TYPE_REPLY_MORE));
}

static inline int
est_is_reply_final(struct est_header *hdr)
{
	return(hdr->type & EST_TYPE_REPLY_FINAL);
}

static inline int
est_is_reply_more(struct est_header *hdr)
{
	return(hdr->type & EST_TYPE_REPLY_MORE);
}

static inline int
est_is_cmd(struct est_header *hdr)
{
	return(hdr->type & EST_TYPE_COMMAND);
}

static inline int
est_valid_type_and_sizes(int type, int *cmd_size, int *reply_size)
{
	int valid = 0;
	int csize = 0;
	int rsize = 0;

	type &= EST_TYPE_MASK;

	if (type == EST_TYPE_HELLO) {
		valid = 1;
		csize = sizeof(struct est_hello_cmd);
		rsize = sizeof(struct est_hello_reply);

	} else if (type == EST_TYPE_PEERSTART) {
		valid = 1;
		csize = sizeof(struct est_peerstart_cmd);
		rsize = sizeof(struct est_peerstart_reply);

	} else if (type == EST_TYPE_PROBE) {
		valid = 1;
		csize = sizeof(struct est_probe_cmd);
		rsize = sizeof(struct est_probe_reply);

	} else if (type == EST_TYPE_PROBESTART) {
		valid = 1;
		csize = sizeof(struct est_probestart_cmd);
		rsize = 0;

	} else if (type == EST_TYPE_TRAFGEN_TRAFFIC) {
		valid = 1;
		csize = sizeof(struct est_trafgen_traffic_cmd);
		rsize = sizeof(struct est_trafgen_traffic_reply);

#ifdef TRAFGEN_REPORT
	} else if (type == EST_TYPE_TRAFGEN_REPORT) {
		valid = 1;
		csize = sizeof(struct est_trafgen_report_cmd);
		rsize = sizeof(struct est_trafgen_report_reply);
#endif

	} else if (type == EST_TYPE_INFO) {
		valid = 1;
		csize = sizeof(struct est_info_cmd);
		rsize = sizeof(struct est_info_reply);

	} else if (type == EST_TYPE_PROBESTATUS) {
		valid = 1;
		csize = sizeof(struct est_probestatus_cmd);
		rsize = 0;


	} else {
		valid = 0;
	}

	if (valid && cmd_size != NULL)
		*cmd_size = csize;
	if (valid && reply_size != NULL)
		*reply_size = rsize;

	return valid;
}

static inline int
est_valid_type(int type)
{
	return est_valid_type_and_sizes(type, NULL, NULL);
}

static inline int
est_fill_cmd_header(struct est_header *hdr, int type)
{
	int size = 0;
	int valid = 0;

	/* validate and get cmd size: use first size arg */
	valid = est_valid_type_and_sizes(type, &size, NULL);
	if (valid) {
		memset(hdr, 0, sizeof(*hdr));
		hdr->type = type;
		hdr->type |= EST_TYPE_COMMAND;
		hdr->size = size;
		hdr->magic = EST_MAGIC;
	}

	return valid;
}

static inline int
est_fill_reply_header(struct est_header *hdr, int type, int final)
{
	int size = 0;
	int valid = 0;

	/* validate and get cmd size: use second size arg */
	valid = est_valid_type_and_sizes(type, NULL, &size);
	if (valid) {
		memset(hdr, 0, sizeof(*hdr));
		hdr->type = type;
		hdr->type |=
			(final ? EST_TYPE_REPLY_FINAL : EST_TYPE_REPLY_MORE);
		hdr->size = size;
		hdr->magic = EST_MAGIC;
	}

	return valid;
}

/* PROBE */

static inline int
est_receive_probe_result(struct est_channel *ch,
			struct est_results *results,
			int *final)
{
	int rc = 0;
	int ok = 0;
	char error[200];
	struct est_probe_reply reply;

	rc = est_recv_ctrlmsg(ch, &reply, sizeof(reply),
				error, sizeof(error));
	if (rc)
		goto error;

	ok = (reply.hdr.type & EST_TYPE_PROBE);
	if (!ok)
		goto error;

	*results = reply.results;
	*final = est_is_reply_final(&reply.hdr);

	return 0;

error:
	return rc;
}


int est_create_pathload_probe_session(struct est_channel *dst,
			struct sockaddr_in *src,
			int iterate,
                        int iterate_interval_sec,
			int mgrp,
                        int piggyback,
                        int fast,
			int start_session,
			void **session_p);

int est_pathload_probestart(struct est_channel *dst, void *session, int start);

static inline int
est_start_pathload_probe_session(struct est_channel *dst, void *session)
{
	return est_pathload_probestart(dst, session, 1);
}
			
static inline int
est_stop_pathload_probe_session(struct est_channel *dst, void *session)
{
	return est_pathload_probestart(dst, session, 0);
}

struct est_probe_session *
est_create_probe_session(struct sockaddr_in *src,
			   struct sockaddr_in *dst,
			   int iterate, int iterate_interval_sec,
                           int mgrp, int piggyback, int fast);

int est_start_probe_session(struct est_probe_session *ses);

static inline void
est_destroy_probe_session(struct est_probe_session *ses)
{
	/* XX stop the session? not tested yet! */
	est_close_channel(&ses->ch);
	free(ses);
}

static inline int
est_probe_session_socket(struct est_probe_session *ses)
{
	if (ses->ch.state < EST_CHANNEL_SOCKET_OPEN)
		return -1;
	return ses->ch.sock;
}

static inline int
est_next_probe_session_estimate(struct est_probe_session *ses,
				struct est_results *results,
				int *final)
{
	return est_receive_probe_result(&ses->ch, results, final);
}

/* TRAFGEN */

/* TRAFGEN client side (applications) */

#ifdef TRAFGEN_REPORT
static inline void
est_trafgen_logreport(struct est_trafgen_report *report)
{
	printf("report:");
}
#endif

static inline int
est_trafgen_session_socket(struct est_trafgen_session *ses)
{
	if (ses->ch.state < EST_CHANNEL_SOCKET_OPEN)
		return -1;
	return ses->ch.sock;
}

struct est_trafgen_session *est_create_trafgen_session(
			struct sockaddr_in *src,
			struct sockaddr_in *dst,
			int where);

void est_destroy_trafgen_session(struct est_trafgen_session *ses);

/* Sends TRAFGEN_TRAFFIC command to @ses->src
 *
 * Actions supported (from enum est_trafgen_traffic_cmd_action):
 *
 *	EST_TRAFGEN_PREPARE - prepare the sender thread
 *	EST_TRAFGEN_START   - starts the traffic pattern described in @traffic
 *	EST_TRAFGEN_STOP    - stops the traffic but not the thread
 */
int est_send_trafgen_traffic_to_src(struct est_trafgen_session *ses,
				    int action,
				    struct est_trafgen_traffic *traffic);

/* sends TRAFGEN_REPORT to @ses->dst */
/* The @ses->dst will keep sending traffic reports on the channel */

int est_start_trafgen_reports(struct est_trafgen_session *ses);

#ifdef TRAFGEN_REPORT
int est_next_trafgen_report(struct est_trafgen_session *ses,
			    struct est_trafgen_report *report,
			    int *final);
#endif

/* INFO */

int est_push_info_entries(struct est_channel *ch,
                          struct est_info *info_array,
                          int info_array_entries);

int est_pull_info_entries(struct est_channel *ch,
                          int info_array_max_entries_to_get,
                          struct est_info *info_array,
                          int *info_array_entries);

/* PROBESTATUS */

static inline int
est_send_probe_status(struct est_channel *ch, int id,
                      int value, struct est_results *results)
{
        int rc = 0;
        int ok = 0;
        struct est_probestatus_cmd cmd;
	char error[200];

        /* PROBESTATUS cmd */
        memset(&cmd, 0, sizeof(cmd));
        ok = est_fill_cmd_header(&cmd.hdr, EST_TYPE_PROBESTATUS);
        BUG_ON(!ok);

        cmd.status_id = id;
        cmd.status_value = value;
        if (results != NULL)
                cmd.status_results = *results;

        rc = est_send_ctrlmsg(ch, &cmd, error, sizeof(error));
	if (rc) {
		ERR("error: failed to send PROBESTATUS: %s", error);
		return rc;
	}

        /* this command has no reply */
        return 0;
}

/* ESTIMATOR */

// used by clients to represent an estimator entry
struct est_estimator {
        int valid;
        char unparsed[200];
        char addrstr[50];
        struct sockaddr_in addr;
        struct est_channel ch;
};

static inline void
est_parse_estimator(const char *optstr, struct est_estimator *estimator)
{
        int found = 0;
        char error[100];

        memset(estimator, 0, sizeof(struct est_estimator));

        strncpy(estimator->unparsed, optstr, sizeof(estimator->unparsed));
        found = net_resolve_host_port(estimator->unparsed,
                                      &estimator->addr,
                                      error, sizeof(error));
        if (!found) {
                ERR("failed to parse estimator host:port '%s': %s",
                    estimator->unparsed, error);
        }

        if (estimator->addr.sin_port == 0)
                estimator->addr.sin_port = htons(EST_DEFAULT_CONTROL_TCP_PORT);

        net_sockaddr_str(estimator->addrstr,
                         sizeof(estimator->addrstr),
                         &estimator->addr);

	LOG("resolved estimator %s: %s",
            estimator->unparsed, estimator->addrstr);

        estimator->valid = 1;
}

#define EST_INFO_GROUP_ENTRIES  EST_INFO_ENTRIES_PER_MSG

#define EST_INFO_NOPRINT 0
#define EST_INFO_PRINT  1

struct est_info_group {
        struct est_estimator *estimator;

        int info_count;
        struct est_info info_array[EST_INFO_GROUP_ENTRIES];
};

static inline void
est_info_group_init(struct est_info_group *igroup,
                    struct est_estimator *estimator)
{
        memset(igroup, 0, sizeof(struct est_info_group));
        igroup->estimator = estimator;
}

static inline void
est_info_group_flush(struct est_info_group *igroup)
{
        BUG_ON(igroup->info_count > EST_INFO_GROUP_ENTRIES);

        // flush the info_array to the estimator
        if (igroup->estimator != NULL) {
                int rc = est_push_info_entries(&igroup->estimator->ch,
                                               igroup->info_array,
                                               igroup->info_count);
                BUG_ON(rc);
        }
        igroup->info_count = 0;

}

static inline void
est_info_group_add(struct est_info_group *igroup,
                   struct est_info *info,
                   int print)
{
        BUG_ON(igroup->info_count > EST_INFO_GROUP_ENTRIES);

        if (igroup->info_count == EST_INFO_GROUP_ENTRIES)
                est_info_group_flush(igroup);

        igroup->info_array[igroup->info_count] = *info;
        igroup->info_count++;

        if (print) {
                char infostr[300];
                est_info_oneline_str(infostr, sizeof(infostr), info);
                LOG("%s", infostr);
        }
}
#endif
