#include <unistd.h>
#include <string.h>
#include <pthread.h>

/* for inet_ntop */
#include <arpa/inet.h>

#include "alog.h"
#include "libestimator.h"

struct est_trafgen_session *
est_create_trafgen_session(struct sockaddr_in *src,
			   struct sockaddr_in *dst,
			   int where)
{
	int rc = 0;
	struct est_trafgen_session *ses = NULL;

	ses = malloc(sizeof *ses);
	if (ses == NULL)
		goto error;

	memset(ses, 0, sizeof(*ses));
	ses->src = *src;
	ses->dst = *dst;

	if (where == EST_TRAFGEN_SRC)
		rc = est_open_channel(&ses->ch, &ses->src);
	else if (where == EST_TRAFGEN_DST)
		rc = est_open_channel(&ses->ch, &ses->dst);
	else
		BUG();

	if (rc)
		goto error_free;
	
	return ses;

error_free:
	free(ses);
	ERR("libestimator failed: %s", est_strerror(rc));
error:
	return NULL;
}

void
est_destroy_trafgen_session(struct est_trafgen_session *ses)
{
	est_close_channel(&ses->ch);
	free(ses);
}

int
est_send_trafgen_traffic_to_src(struct est_trafgen_session *ses,
				int action,
				struct est_trafgen_traffic *traffic)
{
	int rc = 0;
	int ok = 0;
	struct est_trafgen_traffic_cmd cmd;
	struct est_trafgen_traffic_reply reply;
	char error[200];

	/* TRAFGEN_TRAFFIC cmd */

	memset(&cmd, 0, sizeof(cmd));
	ok = est_fill_cmd_header(&cmd.hdr, EST_TYPE_TRAFGEN_TRAFFIC);
	BUG_ON(!ok);

	/* remember: these ip/ports are where the etimators are running
	 *
	 * These are not the ports where the generated traffic will
	 * go to. The estimators internally decide which ports to use.
	 */
	cmd.src = ses->src;
	cmd.dst = ses->dst;
	cmd.action = action;
	if (traffic != NULL)
		cmd.traffic = *traffic;

	rc = est_send_ctrlmsg(&ses->ch, &cmd, error, sizeof(error));
	if (rc) {
		ERR("failed to send TRAFGEN_TRAFFIC: %s", error);
		goto error;
	}

	/* TRAFGEN_TRAFFIC reply */

	rc = est_recv_ctrlmsg(&ses->ch, &reply, sizeof(reply),
	                      error, sizeof(error));
	if (rc) {
		ERR("failed to recv TRAFGEN_TRAFFIC: %s", error);
		goto error;
	}

	ok = (reply.hdr.type & EST_TYPE_TRAFGEN_TRAFFIC);
	if (!ok) {
		rc = EST_EINVAL;
		ERR("reply was not a TRAFGEN_TRAFFIC");
		goto error;
	}

	if (!reply.success) {
		ERR("traffic session failed: %s", reply.error);
		rc = EST_ERROR;
		goto error;
	}

	return 0;
error:
	return rc;
}

#ifdef TRAFGEN_REPORT
int
est_start_trafgen_reports(struct est_trafgen_session *ses)
{
	int rc = 0;
	int ok = 0;
	struct est_trafgen_report_cmd cmd;
	struct est_trafgen_report_reply reply;
	char error[200];

	/* send TRAFGEN_REPORT */

	memset(&cmd, 0, sizeof(cmd));
	ok = est_fill_cmd_header(&cmd.hdr, EST_TYPE_TRAFGEN_REPORT);
	BUG_ON(!ok);

	cmd.src = ses->src;
	cmd.dst = ses->dst;

	rc = est_send_ctrlmsg(&ses->ch, &cmd, error, sizeof(error));
	if (rc) {
		ERR("failed to send TRAFGEN_REPORT: %s", error);
		goto error;
	}

	/* TRAFGEN_REPORT reply */

	rc = est_recv_ctrlmsg(&ses->ch, &reply, sizeof(reply),
			      error, sizeof(error));
	if (rc) {
		ERR("failed to recv TRAFGEN_REPORT: %s", error);
		goto error;
	}

	ok = (reply.hdr.type & EST_TYPE_TRAFGEN_REPORT);
	if (!ok) {
		rc = EST_EINVAL;
		ERR("reply was not a TRAFGEN_REPORT");
		goto error;
	}

	if (!reply.success) {
		ERR("traffic session failed: %s", reply.error);
		rc = EST_ERROR;
		goto error;
	}

	return 0;
error:
	return rc;
}

int
est_next_trafgen_report(struct est_trafgen_session *ses,
			struct est_trafgen_report *report,
			int *final)
{
	int rc = 0;
	int ok = 0;
	char error[200];
	struct est_trafgen_report_reply reply;
	struct est_channel *ch = &ses->ch;

	rc = est_recv_ctrlmsg(ch, &reply, sizeof(reply),
				error, sizeof(error));
	if (rc)
		goto error;

	ok = (reply.hdr.type & EST_TYPE_TRAFGEN_REPORT);
	if (!ok) {
		rc = EST_EINVAL;
		ERR("reply was not a TRAFGEN_REPORT");
		goto error;
	}

	if (!reply.has_report) {
		rc = EST_EINVAL;
		ERR("reply did not contain a report");
		goto error;
	}

	*report = reply.report;
	*final = est_is_reply_final(&reply.hdr);

	return 0;

error:
	return rc;
}
#endif

void
est_trafgen_stats_oneline_str(char *str, int len,
                              struct est_trafgen_stats *stats)
{
        long long usec_diff = atimer_timeval_diff_usec(&stats->tstamp_end,
                                                       &stats->tstamp_start);

        snprintf(str, len,
                 "%s %s %6d %12d %5d %5d %8lld",
                 stats->src_id,
                 stats->dst_id,
                 stats->packets,
                 stats->bytes,
                 stats->pktlen_min,
                 stats->pktlen_max,
                 usec_diff);
}

