#include <string.h>

#include "libpathload.h"
#include "asys.h"
#include "alog.h"

#define PATHLOAD_MAX_ARGS       20

/* pathload_snd and pathload_rcv must be on the path */
char pathload_snd_path[200]             = "pathload_snd";
char pathload_rcv_path[200]             = "pathload_rcv";

static inline void
get_pathload_estimator_options(char *options, int len,
                           const struct sockaddr_in *estimator_addr,
                           int id)
{
        int needs_len = 0;
        char estimator[INET_ADDRSTRLEN];

        // get estimator's address in ip:port format
        net_sockaddr_str(estimator, sizeof(estimator), estimator_addr);

        // see pathload_snd -h and pathload_rcv -h
        // -P <estimator[:port]> -I <status-pointer-ID>
        needs_len = snprintf(options, len, " -P %s -I %d", estimator, id);
        BUG_ON(needs_len > len);
}

void
pathload_exec_and_wait(char *path, struct asys_args *args, int *exported_pid)
{
        int rc = 0;
        int pid = 0;
        int pid_status = 0;

        LOG("++++++++++++ running pathload command:");
        LOG("path: %s", path);
        asys_args_log(args);

        pid = fork();
        SYSERR_ON(pid == -1);

        // child
        if (pid == 0) {
                rc = execvp(path, args->argv);
                // only returns on exec error
                SYSERR_ON(rc);
        }

        // parent
        if (exported_pid != NULL)
                *exported_pid = pid;
        rc = waitpid(pid, &pid_status, 0);
        SYSERR_ON(rc == -1);

        LOG("++++++++++++ done pathload command: status %d", pid_status);
}

void *
pathload_snd(void *arg)
{
        char options[200];
        struct asys_args args;

        /* @op is reserved by creator of the thread */
	struct pathload_snd_options *op = arg;

        /* see libpathload-th/pathload_snd_main.c:pathload_snd_configure() */

        asys_args_init(&args, PATHLOAD_MAX_ARGS, pathload_snd_path);

        snprintf(options, sizeof(options),
                 "-T %d -U %d -P localhost",
                 op->snd_tcp_port,
                 op->snd_udp_port);
        asys_args_append_options(&args, options);

        if (op->quiet)
                asys_args_append_options(&args, "-q");
        if (op->iterate)
                asys_args_append_options(&args, "-i");
        if (op->snd_use_kernel_probes)
                asys_args_append_options(&args, "-x");
        if (op->snd_piggyback)
                asys_args_append_options(&args, "-p");

        if (op->snd_fast)
                asys_args_append_options(&args, "-f");

        if (op->use_estimator_rpc) {
                // append options to direct the sub-process to communicate
                // with the estimator using the specific ID
                get_pathload_estimator_options(
                        options, sizeof(options),
                        &op->estimator_addr, op->status.id);
                asys_args_append_options(&args, options);
        }

        while (1) {
                // fork, exec pathload and wait for it to finish
                pathload_exec_and_wait(pathload_snd_path, &args, NULL);
                LOGTS("+++ pathload_snd DONE, relaunching..!");
        }

        asys_args_done(&args);

        /* cannot access @op after next line */
        est_probe_status_release(&op->status);

        // What we really mean is that this thread cannot access @op
        // anymore but @op may still be valid.. The caller probably
        // has @op still reserved since the return result is stored
        // in op->status.results.

	return NULL;
}

void
pathload_rcv_terminate(struct pathload_rcv_options *op)
{
        if (op->subprocess_pid > 0) {
                LOG("+++ sending SIGTERM to pathload_rcv(%d)",
                    op->subprocess_pid);
                kill(op->subprocess_pid, SIGTERM);
        }
}

void *
pathload_rcv(void *arg)
{
        char options[200];
        struct asys_args args;

        /* @op is reserved by creator of the thread */
	struct pathload_rcv_options *op = arg;

        /* see libpathload-th/pathload_rcv_main.c:pathload_rcv_configure() */

        asys_args_init(&args, PATHLOAD_MAX_ARGS, pathload_rcv_path);

        snprintf(options, sizeof(options),
                 "-s %s -T %d -U %d",
                 op->sender_hostname,
                 op->rcv_tcp_port,
                 op->rcv_udp_port);
        asys_args_append_options(&args, options);

        if (op->verbose)
                asys_args_append_options(&args, "-v");
        if (op->rcv_use_kernel_probes)
                asys_args_append_options(&args, "-x");

        if (op->use_estimator_rpc) {
                // append options to direct the sub-process to communicate
                // with the estimator using the specific ID
                get_pathload_estimator_options(
                        options, sizeof(options),
                        &op->estimator_addr, op->status.id);
                asys_args_append_options(&args, options);
        }

        // fork, exec pathload and wait for it to finish
        pathload_exec_and_wait(pathload_rcv_path, &args, &op->subprocess_pid);

        // pathload_one_receiver_run() waits for this
        asys_status_signal_change(&op->completion_status,
                                  PATHLOAD_COMPLETION_OK);

        asys_args_done(&args);

        /* cannot access @op after next line */
        est_probe_status_release(&op->status);

        // What we really mean is that this thread cannot access @op
        // anymore but @op may still be valid.. The caller probably
        // has @op still reserved since the return result is stored
        // in op->status.results.

	return NULL;
}
