/*
 This file is part of pathload.

 pathload is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 pathload is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with pathload; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*-------------------------------------------------
   pathload : an end-to-end available bandwidth 
            estimation tool
   Author   : Manish Jain ( jain@cc.gatech.edu )
              Constantinos Dovrolis (dovrolis@cc.gatech.edu )
   Release  : Ver 1.3.2
   Support  : This work was supported by the SciDAC
            program of the US department 
--------------------------------------------------*/

/*
 * $Header: /net/cvs/bwtest/pathload/pathload_rcv.c,v 1.139 2006/05/19 22:30:13 jain Exp $
 */

#include <pthread.h>

#include "pathload_gbls.h"
#include "pathload_rcv.h"
#include "libpathload.h"

pthread_key_t rcv_globals_key;

static pthread_once_t rcv_once = PTHREAD_ONCE_INIT;

static inline void
__pathload_rcv_status_set(struct rcv_pathload_globals *gb,
                          int value, struct est_results *results)
{
        if (gb->use_estimator_rpc) {
                int rc = 0;
                
                LOG("+++ sending status %d, estimator rpc", value);
                rc = est_send_probe_status(&gb->estimator_channel,
                                           gb->status->id, value, results);
                BUG_ON(rc);
        } else {
                pthread_mutex_lock(&gb->status->mutex);
                gb->status->value = value;
                pthread_cond_signal(&gb->status->cond);
                pthread_mutex_unlock(&gb->status->mutex);
        }
}

static inline void pathload_rcv_status_set(int value, struct est_results *results)
{
	struct rcv_pathload_globals *gb = rcv_globals();
	__pathload_rcv_status_set(gb, value, results);
}


void pathload_rcv_done(struct rcv_pathload_globals *gb)
{
	close(gb->sock_udp);
	close(gb->sock_tcp);
	LOGFN("sockets closed");

        /* cannot access @status after next line */
        est_probe_status_release(gb->status);
        gb->status = NULL;

        if (gb->use_estimator_rpc) {
                LOG("++++ closing estimator channel");
                est_close_channel(&gb->estimator_channel);
        }

	LOGFN("thread done");
}

static void rcv_globals_free(struct rcv_pathload_globals *gb)
{
	pathload_rcv_done(gb);
	LOGFN("free thread-local pointer %p", gb);
	free(gb);
}

static void pathload_rcv_init_once(void)
{
	int rc = 0;
	LOGFN("initializing once");
	rc = pthread_key_create(&rcv_globals_key, (void *)rcv_globals_free);
	if (rc) {
		ERRFN("pthread_key_create failed %d", rc);
		doexit(-1);
	}
}


static void pathload_rcv_openlogs(int netlog, char *netlogfile,
				  int file, char *filename, char *mode)
{
  struct rcv_pathload_globals *gb = rcv_globals();
  if (netlog)
    gb->netlog_fp = fopen(netlogfile,"a");
  if (file)
    gb->pathload_fp = fopen(filename,mode);
  else
    gb->pathload_fp = fopen("pathload.log" , "a" ) ;
  fprintf(gb->pathload_fp, "\n\n");
}

void pathload_rcv_init(void)
{
	int rc = 0;
	struct rcv_pathload_globals *gb = NULL;

	pthread_once(&rcv_once, pathload_rcv_init_once);

	LOGFN("thread init");

	/* create thread-specific global structure and assign to thread key */

	gb = malloc(sizeof(struct rcv_pathload_globals));
	if (gb == NULL) {
		ERRFN("out of memory");
		doexit(-1);
	}

	rc = pthread_setspecific(rcv_globals_key, (void *)gb);
	if (rc) {
		ERRFN("pthread_setspecific");
		doexit(-1);
	}

	memset(gb, 0, sizeof(struct rcv_pathload_globals));

	/* modification not protected by MGRP defines */
	gb->overhead = 20 + 8;	// IP + UDP

	gb->slow=0;
	gb->requested_delay = 0;
	gb->interrupt_coalescence=0;
	gb->bad_fleet_cs=0;
	gb->num_stream = NUM_STREAM;
	gb->stream_len = STREAM_LEN ;
	gb->exp_flag = 1;
	gb->num=0;
	gb->snd_time_interval=0; 

	gb->converged_gmx_rmx = 0 ;
	gb->converged_gmn_rmn = 0 ;
	gb->converged_rmn_rmx = 0 ;
	gb->counter = 0 ;
	gb->prev_actual_rate = 0;
	gb->prev_req_rate = 0 ; 
	gb->cur_actual_rate = 0 ;
	gb->cur_req_rate = 0 ;
	gettimeofday(&gb->exp_start_time, NULL);
	gb->verbose=1;
	gb->bw_resol=0;
	gb->netlog=0;
	gb->increase_stream_len=0;
	gb->lower_bound=0;

	LOGFN("thread-local pointer gb %p gb2 %p", gb,
		pthread_getspecific(rcv_globals_key));
}

void pathload_rcv_configure(struct pathload_rcv_options *op)
{
	struct rcv_pathload_globals *gb = rcv_globals();
	gb->requested_delay = op->requested_delay;
	strncpy(gb->hostname, op->sender_hostname, sizeof(gb->hostname));
	gb->bw_resol = op->bw_resol;
	gb->Verbose = op->Verbose;
	gb->verbose = op->verbose;
	gb->VVerbose = op->VVerbose;

	pathload_rcv_openlogs(op->netlog, op->netlogfile,
			      op->file, op->filename, op->mode);
#ifdef MGRP
	gb->rcv_tcp_port = op->rcv_tcp_port;
	gb->rcv_udp_port = op->rcv_udp_port;
	gb->rcv_use_kernel_probes = op->rcv_use_kernel_probes;
	LOGFN("rcv_tcp_port %d %x", get_rcv_tcp_port(), get_rcv_tcp_port());
	LOGFN("rcv_udp_port %d %x", get_rcv_udp_port(), get_rcv_udp_port());
	LOGFN("rcv_use_kernel_probes %d", gb->rcv_use_kernel_probes);
#endif

	gb->status = &op->status;

        gb->use_estimator_rpc = op->use_estimator_rpc;
        LOGFN("use_estimator_rpc %d", gb->use_estimator_rpc);
        if (gb->use_estimator_rpc) {
                int rc = 0;
                char buf[100];

                net_sockaddr_str(buf, sizeof(buf), &op->estimator_addr);
                LOGFN("openning channel with estimator %s", buf);

                rc = est_open_channel(&gb->estimator_channel,
                                          &op->estimator_addr);
                BUG_ON(rc);

                LOGFN("opened channel with estimator %s", buf);

                LOG("++++ opened estimator channel");
        }
}

int pathload_rcv_main(void)
{
  struct rcv_pathload_globals *gb = rcv_globals();
  struct hostent *host_snd;
  struct sockaddr_in snd_tcp_addr, rcv_udp_addr;
  struct utsname uts ;
  l_int32 ctr_code;
  l_int32 trend, prev_trend = 0;
  socklen_t opt_len;
  l_int32 rcv_buff_sz, mss;
  l_int32 ret_val ; 
  char ctr_buff[8], myname[50], buff[26];
  struct itimerval expireat ;
  struct sigaction sigstruct ;

  int abort_count = 0;

  if ((host_snd = gethostbyname(gb->hostname)) == 0)
  {
    /* check if the user gave ipaddr */
    if ( ( snd_tcp_addr.sin_addr.s_addr = inet_addr(gb->hostname) ) == -1 )
    {
      fprintf(stderr,"%s: unknown host\n", gb->hostname);
      doexit(-1);
    }
  }
  strncpy(buff, ctime(&(gb->exp_start_time.tv_sec)), 24);
  buff[24] = '\0';
  bzero(myname,50);
  if ( gethostname(myname ,50 ) == -1 )
  {
    if ( uname(&uts) < 0 )
      strcpy(myname , "UNKNOWN") ;
    else
      strcpy(myname , uts.nodename) ;
  }
  if (gb->verbose || gb->Verbose)
    printf("\n\nReceiver %s starts measurements at sender %s on %s \n", myname , gb->hostname, buff);
  fprintf(gb->pathload_fp,"\n\nReceiver %s starts measurements at sender %s on %s \n", myname , gb->hostname, buff);

#ifdef MGRP
  if (gb->rcv_use_kernel_probes) {
	  printf("receiver: opening a MGRP socket explicitly\n");
	  if ((gb->sock_udp = socket(AF_INET, SOCK_DGRAM, IPPROTO_MGRP)) < 0)
	  {
	    perror("ERROR :: failed to open DGRAM/MGRP socket:");
	    doexit(-1);
	  }
  } else
#endif
  if ((gb->sock_udp = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
  {
    perror("ERROR :: failed to open DGRAM socket:");
    doexit(-1);
  }
  bzero((char *)&rcv_udp_addr, sizeof(rcv_udp_addr));
  rcv_udp_addr.sin_family = AF_INET;
  rcv_udp_addr.sin_addr.s_addr = INADDR_ANY;
#ifdef MGRP
  rcv_udp_addr.sin_port = htons(get_rcv_udp_port());
  gb->overhead = get_udp_ip_overhead(gb->sock_udp);
  if (udpex_socket(gb->sock_udp)) {
	printf("** MGRP detected\n");
	if (gb->rcv_use_kernel_probes) {
		printf("** Option -x asks for kernel probes: turn on timestamps\n");
		int on = 1;
		int rc = setsockopt(gb->sock_udp, SOL_MGRP, UDPEX_TSTAMP,
					&on, sizeof(on));
		if (rc) {
			perror("turning on bhtp timestamps");
			doexit(-1);
		}
	}
  } else if (gb->rcv_use_kernel_probes) {
  	printf("Asking for -x kernel probes but socket is not UDPex socket\n");
	doexit(-1);
  }
#else
  rcv_udp_addr.sin_port = htons(UDPRCV_PORT);
#endif
  opt_len = 1;
  if (setsockopt(gb->sock_udp, SOL_SOCKET, SO_REUSEADDR, &opt_len, sizeof(opt_len)) < 0)
  {
    perror("setsockopt(sock_udp,SOL_SOCKET,SO_REUSEADDR):");
    doexit(-1);
  }
  if (bind(gb->sock_udp, (struct sockaddr *)&rcv_udp_addr, sizeof(rcv_udp_addr)) < 0)
  {
    perror("ERROR :: failed to bind DGRAM socket:");
    doexit(-1);
  }
  rcv_buff_sz = UDP_BUFFER_SZ;
  if (setsockopt(gb->sock_udp, SOL_SOCKET, SO_RCVBUF, &rcv_buff_sz, sizeof(rcv_buff_sz)) < 0)
  {
    rcv_buff_sz/=2;
    if (setsockopt(gb->sock_udp, SOL_SOCKET, SO_RCVBUF, &rcv_buff_sz, sizeof(rcv_buff_sz)) < 0)
    {
      printf("ERROR :: Unable to set socket buffer to %d .\n", l32int(rcv_buff_sz));
      doexit(-1);
    }
  }

  /* set up control channel */
  if ((gb->sock_tcp = socket(AF_INET, SOCK_STREAM, 0)) < 0)
  {
    perror("socket(AF_INET, SOCK_STREAM):");
    doexit(-1);
  }
  opt_len = 1;
  if (setsockopt(gb->sock_tcp, SOL_SOCKET, SO_REUSEADDR, &opt_len, sizeof(opt_len)) < 0)
  {
    perror("setsockopt(sock_tcp,SOL_SOCKET,SO_REUSEADDR):");
    doexit(-1);
  }
  bzero((char *)&snd_tcp_addr, sizeof(snd_tcp_addr));
  snd_tcp_addr.sin_family = AF_INET;
  memcpy((void *)&(snd_tcp_addr.sin_addr.s_addr), host_snd->h_addr, host_snd->h_length);
#ifdef MGRP
  snd_tcp_addr.sin_port = htons(get_rcv_tcp_port());
#else
  snd_tcp_addr.sin_port = htons(TCPSND_PORT);
#endif
  if (connect(gb->sock_tcp, (struct sockaddr *)&snd_tcp_addr, sizeof(snd_tcp_addr)) < 0)
  {
    perror("Make sure that pathload_snd runs at sender:");
    doexit(-1);
  }
  if (fcntl(gb->sock_tcp, F_SETFL, O_NONBLOCK) < 0)
  {
    perror("fcntl:");
    doexit(-1);
  }
  /*
    measure max_pkt_sz (based on TCP MSS).
    this is not accurate because it does not take into
    account MTU of intermediate routers.
  */
  opt_len = sizeof(mss);
  if (getsockopt(gb->sock_tcp, IPPROTO_TCP, TCP_MAXSEG, &mss, &opt_len)<0)
  {
     perror("getsockopt(sock_tcp,IPPROTO_TCP,TCP_MAXSEG):");
     doexit(-1);
  }
  gb->rcv_max_pkt_sz = mss;
  if (gb->rcv_max_pkt_sz == 0 || gb->rcv_max_pkt_sz == 1448 ) 
    gb->rcv_max_pkt_sz = 1472;   /* Make it Ethernet sized MTU */
  else
    gb->rcv_max_pkt_sz = mss+12;

#ifdef MGRP
#ifdef DEBUG
    printf("DEBUG :: restricting original max_pkt_sz to %d from %d\n",
  	  gb->rcv_max_pkt_sz, MGRP_MAX_TUN_UDP);
#endif
    /* Restrict the maximum packet size to leave space for bhtp overhead. */
    gb->rcv_max_pkt_sz = MGRP_MAX_TUN_UDP;
#endif

  sigstruct.sa_handler = sig_alrm ;
  sigemptyset(&sigstruct.sa_mask);
  sigstruct.sa_flags = 0 ;
  #ifdef SA_INTERRUPT
    sigstruct.sa_flags |= SA_INTERRUPT ;
  #endif
  sigaction(SIGALRM , &sigstruct,NULL );
  expireat.it_value.tv_sec = 60 ; /* RECEIVER ABORTS TIME */
  expireat.it_value.tv_usec = 0 ;
  expireat.it_interval.tv_sec = 0 ; 
  expireat.it_interval.tv_usec = 0 ;
  setitimer(ITIMER_REAL, &expireat,NULL);

  /* receive sender max_pkt_sz */
  while ((gb->snd_max_pkt_sz = rcv_recv_ctr_mesg(gb->sock_tcp, ctr_buff)) == -1);
  if ( gb->snd_max_pkt_sz == -2 )
  {
    printf("pathload_snd did not respond for 60 sec\n");
    close(gb->sock_tcp);
    doexit(-1);
  }
  expireat.it_value.tv_sec = 0 ; 
  expireat.it_value.tv_usec = 0 ;
  expireat.it_interval.tv_sec = 0 ; 
  expireat.it_interval.tv_usec = 0 ;
  setitimer(ITIMER_REAL, &expireat,NULL);
  
  sigstruct.sa_handler = SIG_DFL ;
  sigemptyset(&sigstruct.sa_mask);
  sigstruct.sa_flags = 0 ;
  sigaction(SIGALRM , &sigstruct,NULL );


  rcv_send_ctr_mesg(ctr_buff, gb->rcv_max_pkt_sz);
  gb->max_pkt_sz = (gb->rcv_max_pkt_sz < gb->snd_max_pkt_sz) ? gb->rcv_max_pkt_sz:gb->snd_max_pkt_sz ;
  if (gb->Verbose)
    printf("  Maximum packet size          :: %ld bytes\n",gb->max_pkt_sz);
  fprintf(gb->pathload_fp,"  Maximum packet size          :: %ld bytes\n",gb->max_pkt_sz);
  gb->rcv_latency = (l_int32) recvfrom_latency(rcv_udp_addr);
  while ((gb->snd_latency = rcv_recv_ctr_mesg(gb->sock_tcp, ctr_buff)) == -1);
  if (gb->Verbose)
  {
    printf("  send latency @sndr           :: %ld usec\n",gb->snd_latency);
    printf("  recv latency @rcvr           :: %ld usec\n",gb->rcv_latency);
  }
  fprintf(gb->pathload_fp,"  send latency @sndr           :: %ld usec\n",gb->snd_latency);
  fprintf(gb->pathload_fp,"  recv latency @rcvr           :: %ld usec\n",gb->rcv_latency);
  gb->min_time_interval=
      SCALE_FACTOR*((gb->rcv_latency>gb->snd_latency)?gb->rcv_latency:gb->snd_latency) ;
  gb->min_time_interval = gb->min_time_interval>MIN_TIME_INTERVAL?
      gb->min_time_interval:MIN_TIME_INTERVAL;
  if (gb->Verbose)
    printf("  Minimum packet spacing       :: %ld usec\n",gb->min_time_interval );
  fprintf(gb->pathload_fp,"  Minimum packet spacing       :: %ld usec\n",gb->min_time_interval );
  gb->max_rate = (gb->max_pkt_sz+gb->overhead) * 8. / gb->min_time_interval ;
  gb->min_rate = (MIN_PKT_SZ+gb->overhead) * 8./ MAX_TIME_INTERVAL ;
  if(gb->Verbose)
    printf("  Max rate(max_pktsz/min_time) :: %.2fMbps\n",gb->max_rate);
  fprintf(gb->pathload_fp,"  Max rate(max_pktsz/min_time) :: %.2fMbps\n",gb->max_rate);

  /* Estimate ADR */
  gb->adr = get_adr() ;
  if ( gb->bw_resol == 0 && gb->adr != 0 )
    gb->bw_resol = .02*gb->adr ;
  else if (gb->bw_resol == 0 )
    gb->bw_resol = 2 ; 
  if(gb->Verbose)
    printf("  Grey bandwidth resolution    :: %.2f\n",grey_bw_resolution());
  fprintf(gb->pathload_fp,"  Grey bandwidth resolution    :: %.2f\n",grey_bw_resolution());
  
  if (gb->interrupt_coalescence)
  {
    gb->bw_resol = .05*gb->adr;
    if(gb->verbose||gb->Verbose)
      printf("  Interrupt coalescion detected\n");
    fprintf(gb->pathload_fp,"  Interrupt coalescion detected\n");
  }
  
  if ( gb->adr == 0 || gb->adr > gb->max_rate || gb->adr < gb->min_rate)
    gb->tr = (gb->max_rate+gb->min_rate)/2.;
  else 
    gb->tr = gb->adr ;

#ifdef MGRP
	printf("rates max %f min %f tr %f verbose %ld\n",
		gb->max_rate, gb->min_rate, gb->tr, gb->verbose);
#endif
  /* Estimate the available bandwidth.*/
  gb->transmission_rate = (l_uint32)rint(1000000 * gb->tr);
  gb->max_rate_flag = 0 ;
  gb->min_rate_flag = 0 ; 
  fflush(gb->pathload_fp);

  sigemptyset(&sigstruct.sa_mask);
  sigstruct.sa_handler = sig_alrm ;
  sigstruct.sa_flags = 0 ;
  #ifdef SA_INTERRUPT
    sigstruct.sa_flags |= SA_INTERRUPT ;
  #endif
  sigaction(SIGALRM , &sigstruct,NULL );

  if ( gb->requested_delay )
  {
    expireat.it_value.tv_sec = gb->requested_delay ; /* RECEIVER ABORTS TIME */
    expireat.it_value.tv_usec = 0 ;
    expireat.it_interval.tv_sec = 0 ; 
    expireat.it_interval.tv_usec = 0 ;
    setitimer(ITIMER_REAL, &expireat,NULL);
  }


  while (1)
  {
    pthread_testcancel();
    if ( calc_param() == -1 )
    {
      ctr_code = TERMINATE | CTR_CODE;
      rcv_send_ctr_mesg(ctr_buff, ctr_code);
      terminate_gracefully(gb->exp_start_time) ;
    }
    rcv_send_ctr_mesg(ctr_buff, gb->transmission_rate);
    rcv_send_ctr_mesg(ctr_buff,gb->cur_pkt_sz) ;
    if ( gb->increase_stream_len )
      gb->stream_len=3*STREAM_LEN;
    else
      gb->stream_len = STREAM_LEN;
    rcv_send_ctr_mesg(ctr_buff,gb->stream_len);
    rcv_send_ctr_mesg(ctr_buff,gb->time_interval);
    ctr_code = SEND_FLEET | CTR_CODE ;
    rcv_send_ctr_mesg(ctr_buff, ctr_code);

    while (1)
    {
      ret_val = rcv_recv_ctr_mesg(gb->sock_tcp, ctr_buff);
      if ((((ret_val & CTR_CODE) >> 31) == 1) &&    
           ((ret_val & 0x7fffffff) == RECV_FLEET )) 
        break ;
      else if ( (((ret_val & CTR_CODE) >> 31) == 1) &&    
                 ((ret_val & 0x7fffffff) == FINISHED_STREAM )) 
        ret_val = rcv_recv_ctr_mesg(gb->sock_tcp, ctr_buff);
    }

    if (recv_fleet() == -1)
    {
      abort_count++;
      if (abort_count > 20) {
	 printf("++++ abort_count %d\n", abort_count);
          terminate_gracefully(gb->exp_start_time);
      }
      if ( !gb->increase_stream_len )
      {
        trend = INCREASING;
        if ( gb->exp_flag == 1 && prev_trend != 0 && prev_trend != trend)
          gb->exp_flag = 0;
        prev_trend = trend;
        if (rate_adjustment(INCREASING) == -1)
          terminate_gracefully(gb->exp_start_time);
      }
    }
    else
    {
      get_sending_rate() ;
      trend = aggregate_trend_result();
#ifdef MGRP_DEBUG
	printf("\n**** trend %d bad_fleet_cs %d retry_fleet_cnd_cs %d\n",
		trend, gb->bad_fleet_cs, gb->retry_fleet_cnt_cs);
#endif
      
      if ( trend == -1 && gb->bad_fleet_cs && gb->retry_fleet_cnt_cs >NUM_RETRY_CS )
        terminate_gracefully(gb->exp_start_time) ;
      else if(( trend == -1 && gb->bad_fleet_cs && gb->retry_fleet_cnt_cs <= NUM_RETRY_CS )) /* repeat fleet with current rate. */
        continue ;

      if (trend != GREY)
      {
        if (gb->exp_flag == 1 && prev_trend != 0 && prev_trend != trend)
          gb->exp_flag = 0;
        prev_trend = trend;
      }

      if (rate_adjustment(trend) == -1)
        terminate_gracefully(gb->exp_start_time);
    }
    fflush(gb->pathload_fp);
  }
}

void
pathload_rcv_terminate(struct pathload_rcv_options *op)
{
        // see the cancellation point in pthread_rcv_main()
        int rc = pthread_cancel(op->self);
        BUG_ON(rc);
}

void *
pathload_rcv(void *arg)
{
        /* @op is reserved by creator of the thread */
	struct pathload_rcv_options *op = arg;

        op->self = pthread_self();

	pathload_rcv_init();
	pathload_rcv_configure(op);
	pathload_rcv_main();

        // The pathload_rcv_done() is called implicitly when the thread
        // terminates (for any reason) and is invoked from rcv_globals_free().
        // In turn, rcv_globals_free() is called when the thread terminates
        // because @gb is tied to the key @rcv_globals_key.
        //
        // See pthread_key_create(3)

	LOGFN("finished");
	doexit(0);

	return NULL;
}

