/*
 This file is part of pathload.

 pathload is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 pathload is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with pathload; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*-------------------------------------------------
   pathload : an end-to-end available bandwidth 
              estimation tool
   Author   : Manish Jain (jain@cc.gatech.edu)
              Constantinos Dovrolis (dovrolis@cc.gatech.edu)
   Release  : Ver 1.3.2
   Support  : This work was supported by the SciDAC
              program of the US department 
--------------------------------------------------*/
#define LOCAL
#include "pathload_gbls.h"
#include "pathload_snd.h"
#include "libpathload.h"
#include "libestimator.h"

pthread_key_t snd_globals_key;

static pthread_once_t snd_once = PTHREAD_ONCE_INIT;

static inline void
__pathload_snd_status_set(struct snd_pathload_globals *gb, int value)
{
        if (gb->use_estimator_rpc) {
                int rc = 0;
                
                LOG("+++ sending status %d, estimator rpc", value);
                rc = est_send_probe_status(&gb->estimator_channel,
                                           gb->status->id, value, NULL);
                BUG_ON(rc);
        } else {
                pthread_mutex_lock(&gb->status->mutex);
                gb->status->value = value;
                pthread_cond_signal(&gb->status->cond);
                pthread_mutex_unlock(&gb->status->mutex);
        }
}

static inline void pathload_snd_status_set(int value)
{
	struct snd_pathload_globals *gb = snd_globals();
	__pathload_snd_status_set(gb, value);
}

void pathload_snd_done(struct snd_pathload_globals *gb)
{
	close(gb->sock_udp);
	close(gb->sock_tcp);
	LOGFN("sockets closed");

	if (gb->bind_failed)
		__pathload_snd_status_set(gb, EST_PROBE_STATUS_NOPORTS);
	else
		__pathload_snd_status_set(gb, EST_PROBE_STATUS_FAILED);

        if (gb->use_estimator_rpc) {
                est_close_channel(&gb->estimator_channel);
        }

        /* cannot access @status after next line */
        est_probe_status_release(gb->status);
        gb->status = NULL;

	LOGFN("thread done");
}

static void snd_globals_free(struct snd_pathload_globals *gb)
{
	pathload_snd_done(gb);
	LOGFN("free thread-local pointer %p", gb);
	free(gb);
}

static void pathload_snd_init_once(void)
{
	int rc = 0;
	LOGFN("initializing once");
	rc = pthread_key_create(&snd_globals_key, (void *)snd_globals_free);
	if (rc) {
		ERRFN("pthread_key_create failed %d", rc);
		doexit(-1);
	}
}

void pathload_snd_init(void)
{
	// create global here
	int rc = 0;
	struct snd_pathload_globals *gb = NULL;

	pthread_once(&snd_once, pathload_snd_init_once);

	LOGFN("thread init");

	/* create thread-specific global structure and assign to thread key */

	gb = malloc(sizeof(struct snd_pathload_globals));
	if (gb == NULL) {
		ERRFN("out of memory");
		doexit(-1);
	}

	rc = pthread_setspecific(snd_globals_key, (void *)gb);
	if (rc) {
		ERRFN("pthread_setspecific");
		doexit(-1);
	}

	memset(gb, 0, sizeof(struct snd_pathload_globals));

	LOGFN("thread-local pointer gb %p gb2 %p", gb,
		pthread_getspecific(snd_globals_key));
}

void pathload_snd_configure(struct pathload_snd_options *op)
{
	struct snd_pathload_globals *gb = snd_globals();

	gb->quiet = op->quiet;
	gb->iterate = op->iterate;
#ifdef MGRP
	gb->snd_tcp_port = op->snd_tcp_port;
	gb->snd_udp_port = op->snd_udp_port;
	gb->snd_use_kernel_probes = op->snd_use_kernel_probes;
	LOGFN("snd_tcp_port %d %x", get_snd_tcp_port(), get_snd_tcp_port());
	LOGFN("snd_udp_port %d %x", get_snd_udp_port(), get_snd_udp_port());
	LOGFN("snd_use_kernel_probes %d", gb->snd_use_kernel_probes);
#endif
	gb->status = &op->status;

        gb->use_estimator_rpc = op->use_estimator_rpc;
        LOGFN("use_estimator_rpc %d", gb->use_estimator_rpc);
        if (gb->use_estimator_rpc) {
                int rc = 0;
                char buf[100];

                net_sockaddr_str(buf, sizeof(buf), &op->estimator_addr);
                LOGFN("openning channel with estimator %s", buf);

                rc = est_open_channel(&gb->estimator_channel,
                                          &op->estimator_addr);
                BUG_ON(rc);

                LOGFN("opened channel with estimator %s", buf);
        }
}

int pathload_snd_main(void)
{
  struct snd_pathload_globals *gb = snd_globals();
  struct hostent *host_rcv;
  struct timeval tv1,tv2;
  l_uint32 snd_time ;
  l_int32 ctr_code ;
  time_t localtm;
#ifdef MGRP
  int mss;
  unsigned int opt_len;
#else
  int opt_len,mss;
#endif
  int ret_val ;
  int done=0;
  int latency[30],ord_latency[30];
  int i;
  char pkt_buf[256];
  char ctr_buff[8];

  gb->num_stream = NUM_STREAM ;
  min_sleeptime();

  /* gettimeofday latency */
#if defined(MGRP) && defined(DEBUG)
  printf("DEBUG :: calculating gettimeofday latency\n");
#endif
  for(i=0;i<30;i++)
  {
    gettimeofday(&tv1,NULL);
    gettimeofday(&tv2,NULL);
    latency[i]=tv2.tv_sec*1000000+tv2.tv_usec-tv1.tv_sec*1000000-tv1.tv_usec;
  }
  snd_order_int(latency,ord_latency,30);
  gb->gettimeofday_latency = ord_latency[15];  
#ifdef DEBUG
  printf("DEBUG :: gettimeofday_latency = %d\n",gettimeofday_latency);
#endif
  /* Control stream: TCP connection */
  if ((gb->sock_tcp=socket(AF_INET, SOCK_STREAM, 0)) < 0)
  {
    perror("socket(AF_INET,SOCK_STREAM,0):");
    doexit(-1);
  }
#if defined(MGRP) && defined(DEBUG)
  printf("DEBUG :: open TCP socket(%d), listen on %d\n",
  	sock_tcp, get_tcp_port());
#endif
  opt_len=1;
  if (setsockopt(gb->sock_tcp, SOL_SOCKET, SO_REUSEADDR, (char*)&opt_len, 
        sizeof(opt_len)) < 0)
  {
    perror("setsockopt(SOL_SOCKET,SO_REUSEADDR):");
    doexit(-1);
  }
  bzero((char*)&gb->snd_tcp_addr, sizeof(gb->snd_tcp_addr));
  gb->snd_tcp_addr.sin_family         = AF_INET;
  gb->snd_tcp_addr.sin_addr.s_addr    = htonl(INADDR_ANY);
#ifdef MGRP
  gb->snd_tcp_addr.sin_port           = htons(get_snd_tcp_port());
#else
  gb->snd_tcp_addr.sin_port           = htons(TCPSND_PORT);
#endif
  if (bind(gb->sock_tcp, (struct sockaddr*)&gb->snd_tcp_addr,sizeof(gb->snd_tcp_addr)) < 0)
  {
    gb->bind_failed = 1;
    perror("bind(sock_tcp):");
    doexit(-1);
  }
  if (listen(gb->sock_tcp,1) < 0)
  {
    perror("listen(sock_tcp,1):");
    doexit(-1);
  }
  /* Data stream: UDP socket */
#ifdef MGRP
  if (gb->snd_use_kernel_probes) {
	  printf("sender: opening a MGRP socket explicitly\n");
	  if ((gb->sock_udp = socket(AF_INET, SOCK_DGRAM, IPPROTO_MGRP)) < 0)
	  {
	    perror("socket(AF_INET,SOCK_DGRAM,IPPROTO_MGRP):");
	    doexit(-1);
	  }
  } else
#endif
  if ((gb->sock_udp=socket(AF_INET, SOCK_DGRAM, 0)) < 0)
  {
    perror("socket(AF_INET,SOCK_DGRAM,0):");
    doexit(-1);
  }
#if defined(MGRP) && defined(DEBUG)
  printf("DEBUG :: open UDP socket(%d)\n", sock_udp);
#endif
#ifdef MGRP
  /* ignore the return value; just us it to print it on stdout */
  get_udp_ip_overhead(gb->sock_udp);
#endif
  bzero((char*)&gb->snd_udp_addr, sizeof(gb->snd_udp_addr));
  gb->snd_udp_addr.sin_family         = AF_INET;
  gb->snd_udp_addr.sin_addr.s_addr    = htonl(INADDR_ANY);
  gb->snd_udp_addr.sin_port           = htons(0);
  if (bind(gb->sock_udp, (struct sockaddr*)&gb->snd_udp_addr, 
        sizeof(gb->snd_udp_addr)) < 0)
  {
    perror("bind(sock_udp):");
    doexit(-1);
  }
  gb->send_buff_sz = UDP_BUFFER_SZ;
  if (setsockopt(gb->sock_udp, SOL_SOCKET, SO_SNDBUF, (char*)&gb->send_buff_sz, 
        sizeof(gb->send_buff_sz)) < 0)
  {
    gb->send_buff_sz/=2;
    if (setsockopt(gb->sock_udp, SOL_SOCKET, SO_SNDBUF, (char*)&gb->send_buff_sz, 
        sizeof(gb->send_buff_sz)) < 0)
    {
      perror("setsockopt(SOL_SOCKET,SO_SNDBUF):");
      doexit(-1);
    }
  }
  pathload_snd_status_set(EST_PROBE_STATUS_SUCCESS);
  do
  {
    if ( !gb->quiet)
      printf("\n\nWaiting for receiver to establish control stream => ");
    fflush(stdout);
    /* 
      Wait until receiver attempts to connect, 
      starting new measurement cycle
    */
    gb->rcv_tcp_adrlen = sizeof(gb->rcv_tcp_addr);
    gb->ctr_strm = accept(gb->sock_tcp, (struct sockaddr*)&gb->rcv_tcp_addr, &gb->rcv_tcp_adrlen);
    if (gb->ctr_strm < 0)
    {
      perror("accept(sock_tcp):");
      doexit(-1);
    }  
    if ( !gb->quiet)
      printf("OK\n");
    localtm = time(NULL); 
    gethostname(pkt_buf, 256);
    host_rcv=gethostbyaddr((char*)&(gb->rcv_tcp_addr.sin_addr), sizeof(gb->rcv_tcp_addr.sin_addr), AF_INET);
    if (host_rcv!=NULL)
    {
      if ( !gb->quiet)
        printf("Receiver %s starts measurements on %s", host_rcv->h_name, ctime(&localtm));
    }
    else
      if ( !gb->quiet)
        printf("Unknown receiver starts measurements at %s", ctime(&localtm));

    /* Form receiving UDP address */
    bzero((char*)&gb->rcv_udp_addr, sizeof(gb->rcv_udp_addr));
    gb->rcv_udp_addr.sin_family         = AF_INET;
    gb->rcv_udp_addr.sin_addr.s_addr    = gb->rcv_tcp_addr.sin_addr.s_addr;
#ifdef MGRP
    gb->rcv_udp_addr.sin_port           = htons(get_snd_udp_port());
#else
    gb->rcv_udp_addr.sin_port           = htons(UDPRCV_PORT);
#endif
    /* Connect UDP socket */
    connect(gb->sock_udp , (struct sockaddr *)&gb->rcv_udp_addr , sizeof(gb->rcv_udp_addr)); 
    /* Make TCP socket non-blocking */
    if (fcntl(gb->ctr_strm, F_SETFL, O_NONBLOCK)<0)
    {
      perror("fcntl(ctr_strm, F_SETFL, O_NONBLOCK):");
      doexit(-1);
    }
    opt_len = sizeof(mss) ;
    if (getsockopt(gb->ctr_strm, IPPROTO_TCP, TCP_MAXSEG, (char*)&mss, &opt_len)<0)
    {
      perror("getsockopt(sock_tcp,IPPROTO_TCP,TCP_MAXSEG):");
      doexit(-1);
    }
    gb->snd_max_pkt_sz = mss;
    if (gb->snd_max_pkt_sz == 0 || gb->snd_max_pkt_sz== 1448) 
      gb->snd_max_pkt_sz = 1472;   /* Make it Ethernet sized MTU */
    else
      gb->snd_max_pkt_sz = mss+12;

    /* tell receiver our max packet sz */
    snd_send_ctr_mesg(ctr_buff, gb->snd_max_pkt_sz) ;
    /* receiver's maxp packet size */
    while ((gb->rcv_max_pkt_sz = snd_recv_ctr_mesg( ctr_buff)) == -1);
    gb->max_pkt_sz = (gb->rcv_max_pkt_sz < gb->snd_max_pkt_sz) ? gb->rcv_max_pkt_sz:gb->snd_max_pkt_sz ;
    if ( !gb->quiet )
      printf("Maximum packet size          :: %ld bytes\n",gb->max_pkt_sz);
    /* tell receiver our send latency */
    snd_time = (l_int32) send_latency();
    snd_send_ctr_mesg(ctr_buff, snd_time) ;

#if defined(MGRP) && defined(DEBUG)
  printf("DEBUG :: wait for receiver to start ADR measurement\n");
#endif
    /* wait for receiver to start ADR measurement */
    if((ret_val=snd_recv_ctr_mesg(ctr_buff)) == -1 )break;
    if ( (((ret_val & CTR_CODE) >> 31) == 1) && ((ret_val & 0x7fffffff) == SEND_TRAIN ) ) 
    {
      if ( !gb->quiet)
        printf("Estimating ADR to initialize rate adjustment algorithm => ");
      fflush(stdout);
      if ( send_train() == -1 ) continue ;
      if ( !gb->quiet)
        printf("Done\n");
    }
    gb->fleet_id=0;
    done=0;
#if defined(MGRP) && defined(DEBUG)
  printf("DEBUG ::  Start avail-bw measurement\n");
#endif
    /* Start avail-bw measurement */
    while(!done)
    {
      if (( ret_val  = snd_recv_ctr_mesg ( ctr_buff ) ) == -1 ) break ;
      if((((ret_val & CTR_CODE) >> 31) == 1) &&((ret_val&0x7fffffff) == TERMINATE)) 
      {
        if ( !gb->quiet)
          printf("Terminating current run.\n");
        done=1;
      }
      else
      {
        gb->transmission_rate = ret_val ;
        if ((gb->cur_pkt_sz = snd_recv_ctr_mesg( ctr_buff)) <= 0 )break;
        if ((gb->stream_len = snd_recv_ctr_mesg( ctr_buff))  <= 0 )break;
        if ((gb->time_interval = snd_recv_ctr_mesg( ctr_buff)) <= 0 )break;
        if ((ret_val = snd_recv_ctr_mesg ( ctr_buff )) == -1 )break;
        /* ret_val = SENd_FLEET */
        ctr_code = RECV_FLEET | CTR_CODE ;
        if ( snd_send_ctr_mesg(ctr_buff,  ctr_code  ) == -1 ) break;
        if(send_fleet()==-1) break ;
        if ( !gb->quiet) printf("\n");
        gb->fleet_id++ ;
      }
    }
    close(gb->ctr_strm);
  }while (gb->iterate);
  
  return 0;
}

void *
pathload_snd(void *arg)
{
        /* @op is reserved by creator of the thread */
	struct pathload_snd_options *op = arg;

	pathload_snd_init();
	pathload_snd_configure(op);
	pathload_snd_main();

        // The pathload_snd_done() is called implicitly when the thread
        // terminates (for any reason) and is invoked from snd_globals_free().
        // In turn, snd_globals_free() is called when the thread terminates
        // because @gb is tied to the key @snd_globals_key.
        //
        // See pthread_key_create(3)

	LOGFN("finished");
	doexit(0);

	return NULL;
}
