#ifndef _LIBPATHLOAD_H
#define _LIBPATHLOAD_H

#include <sys/time.h>
#include <pthread.h>

// libcommon
#include "net.h"
#include "alog.h"
#include "aptr.h"
#include "asys.h"

#include "libestimator.h"

/* so that all applications include MGRP (which is obsolete now) */
#define MGRP 1

/* Port numbers (UDP for receiver, TCP for sender) */
#define PATHLOAD_UDPRCV_PORT            55001
#define PATHLOAD_TCPSND_PORT            55002

/* snd */

struct pathload_snd_options {
        struct est_probe_status status;

	int iterate;
	int quiet;
#ifdef MGRP
	int snd_tcp_port;
	int snd_udp_port;
	int snd_use_kernel_probes;
        int snd_piggyback;
        int snd_fast;
#endif

        int use_estimator_rpc;
        struct sockaddr_in estimator_addr;
};

static inline void
pathload_snd_options_init(struct pathload_snd_options *op)
{
        memset(op, 0, sizeof(*op));
        // @op->status contains @op->status.ptr for reference counting
        // The second parameter to _init() means that @op will be freed
        // when the refcount of @op->status.ptr reaches 0
        est_probe_status_init(&op->status, op);

        op->snd_tcp_port = PATHLOAD_TCPSND_PORT;
        op->snd_udp_port = PATHLOAD_UDPRCV_PORT;
        op->quiet = 0;
}

void * pathload_snd(void *arg);

/* rcv */

#define PATHLOAD_RESULT_UNDEF		0
#define PATHLOAD_RESULT_NORMAL		1
#define PATHLOAD_RESULT_COALESCENCE	2
#define PATHLOAD_RESULT_UNKNOWN		3

static inline const char *
pathload_result_str(int result_type)
{
	const char *str = "ERROR";

	switch (result_type) {

	case PATHLOAD_RESULT_UNDEF:
		str = "undefined";
		break;

	case PATHLOAD_RESULT_NORMAL:
		str = "normal";
		break;

	case PATHLOAD_RESULT_COALESCENCE:
		str = "coalescence";
		break;

	case PATHLOAD_RESULT_UNKNOWN:
		str = "unknown";
		break;
	};

	return str;
}

struct pathload_results {
	struct timeval ended;
	int mgrp;
        int piggyback;
	int payload_bytes;
	int ip_bytes;

	int type;
	long usec;
	double max;
	double min;

        int probe_bytes;
        int probe_pkts;

        double adjust_stream;
        double adjust_fleet;

        double orig_min;
        double orig_max;
};

/* result conversion functions */

static inline void
pathload_results_to_est_results(const struct pathload_results *src,
                                struct est_results *dst)
{
        /* @src and @dst are are of different types */
        /* this avoids us having to include libpathload */
        dst->valid = src->type;
        dst->min = src->min;
        dst->max = src->max;
        dst->usec = src->usec;
        dst->mgrp = src->mgrp;
        dst->piggyback = src->piggyback;
        dst->ended = src->ended;
        dst->probe_bytes = src->probe_bytes;
        dst->probe_pkts = src->probe_pkts;
        dst->adjust_stream = src->adjust_stream;
        dst->adjust_fleet = src->adjust_fleet;
        dst->orig_min = src->orig_min;
        dst->orig_max = src->orig_max;
}


#define PATHLOAD_COMPLETION_UNDEF       0
#define PATHLOAD_COMPLETION_OK          1

struct pathload_rcv_options {
        pthread_t self;

        struct est_probe_status status;

	int requested_delay;
	char sender_hostname[256];
	int verbose;

        // BEGIN - variables used only by libpathload-th
	float bw_resol;

	int Verbose;
	int VVerbose;

	int file;
	char filename[50];
	char mode[4];
	int netlog;
	char netlogfile[50];
        // END - variables used only by libpathload-th

#ifdef MGRP
	int rcv_tcp_port;
	int rcv_udp_port;
	int rcv_use_kernel_probes;
#endif

        int use_estimator_rpc;
        struct sockaddr_in estimator_addr;

        // BEGIN -- variables used only by libpathload-pr
        int subprocess_pid;
        // END -- variables used only by libpathload-pr

        // used in pathload_one_receiver_run()
        // used to wait for thread/execv to finish
        struct asys_status completion_status;
};

static inline void
pathload_rcv_options_init(struct pathload_rcv_options *op)
{
        memset(op, 0, sizeof(*op));
        // @op->status contains @op->status.ptr for reference counting
        // The second parameter to _init() means that @op will be freed
        // when the refcount of @op->status.ptr reaches 0
        est_probe_status_init(&op->status, op);

        op->verbose = 1;
        op->rcv_tcp_port = PATHLOAD_TCPSND_PORT;
        op->rcv_udp_port = PATHLOAD_UDPRCV_PORT;

        asys_status_init(&op->completion_status);
        op->completion_status.value = PATHLOAD_COMPLETION_UNDEF;
}

void * pathload_rcv(void *arg);
void pathload_rcv_terminate(struct pathload_rcv_options *op);

// make sure that no adjustment is zero (for quick boolean checks)
enum pathload_final_result_adjust {
        PATHLOAD_ADJUST_NONE = 0,
        PATHLOAD_ADJUST_STREAM,
        PATHLOAD_ADJUST_FLEET,
};

static inline int
pathload_get_adjust_type(const char *policy)
{
        if (strcmp(policy, "none") == 0)
                return PATHLOAD_ADJUST_NONE;
        else if (strcmp(policy, "stream") == 0)
                return PATHLOAD_ADJUST_STREAM;
        else if (strcmp(policy, "fleet") == 0)
                return PATHLOAD_ADJUST_FLEET;

        ERRFN("unknown adjust policy: %s", policy);
        return -1;
}

static inline char *
pathload_get_adjust_policy(int type)
{
        if (type == PATHLOAD_ADJUST_NONE)
                return "none";
        else if (type == PATHLOAD_ADJUST_STREAM)
                return "stream";
        else if (type == PATHLOAD_ADJUST_FLEET)
                return "fleet";

        ERRFN("unknown adjust type: %d", type);
        return "error";
}

void
final_result(const int packet_size, const int packet_overhead,
             const long usec, const double max_p, const double min_p,
             const char *result,
             struct est_results *results,
             enum est_probe_status_value *status_value,
             int probe_bytes, int probe_pkts,
             int allow_piggyback,
             double adjust_stream,
             double adjust_fleet,
             enum pathload_final_result_adjust adjust,
             char *results_str, int results_strlen);
#endif
