/*
 * log.c	Packet Train -- packet logging using libpcap
 *
 * Created:	Tue Oct  4 16:24:50 EDT 2005
 *
 * Version:	$Id: log.c,v 1.22 2005/12/16 23:05:20 pavlos Exp $
 *
 * Author:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 * truncated versions of original files found in:
 * /homes/pavlos/work/phd/projects/tagent/src/ptrain
 * taken on Tue Dec 20 10:58:36 EST 2005.
 */
#include <pcap.h>
#include <net/ethernet.h>
#include <netinet/ip.h>
#include <netinet/udp.h>

/* ioctl SIOCGIFADDR */
#include <sys/ioctl.h>
#include <linux/if.h>

/* inet_ntop */
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "log.h"
#include "tick.h"

#define LOG_HEADER_SKIP		(sizeof(struct ether_header) + \
				sizeof(struct iphdr) + \
				sizeof(struct udphdr))

#define LOG_CAPLEN		(LOG_HEADER_SKIP + sizeof(pt_slog_pcap))

/* Will be truncated to kernel max bufsize. */
#define LOG_PCAP_BUFSIZE	1000000

static void log_pkt(u_char *user,
			const struct pcap_pkthdr *phdr,
			const u_char *cap);

void
log_init(log_s *log, char *iface, int port)
{
	int res = 0;
	char errbuf[PCAP_ERRBUF_SIZE];

	char bpf_str[100];
	struct bpf_program fp;

	char dotted[20];
	struct sockaddr_in *iface_addr = NULL;
	struct ifreq req;
	int tmpsock;

	unsigned int bufsize = 0;
	unsigned int bufsize_sz = sizeof(bufsize);

	BUG_ON(log == NULL);

	memset(log, 0, sizeof(*log));

	/* Retrieve the IP address of the interface. */

	tmpsock = socket(AF_INET, SOCK_DGRAM, 0);
	SYSERR_ON(tmpsock < 0);

	memset(&req, 0, sizeof(req));
	strncpy(req.ifr_name, iface, sizeof(req.ifr_name));

	res = ioctl(tmpsock, SIOCGIFADDR, &req);
	SYSERR_ON(res);

	close(tmpsock);

	iface_addr = (struct sockaddr_in *)(&req.ifr_ifru.ifru_addr);
	BUG_ON(iface_addr->sin_family != AF_INET);

	printf("iface %s address %s\n",
		iface,
		inet_ntop(iface_addr->sin_family,
				&iface_addr->sin_addr,
				dotted, sizeof(dotted)));

	/* Open pcap device for capturing outgoing packets. */

	log->tx = pcap_open_live(iface, LOG_CAPLEN, 0, 10, errbuf);
	if (log->tx == NULL)
		goto pcap_open_error;

	/* Increase the pcap buffer size. */

	bufsize = LOG_PCAP_BUFSIZE;
	res = setsockopt(pcap_fileno(log->tx),
		SOL_SOCKET, SO_RCVBUF,
		&bufsize, bufsize_sz);
	SYSERR_ON(res);

	res = getsockopt(pcap_fileno(log->tx),
		SOL_SOCKET, SO_RCVBUF,
		&bufsize, &bufsize_sz);
	SYSERR_ON(res);

	printf("pcap socket buffer size: %d bytes\n", bufsize);

	res = pcap_setnonblock(log->tx, 1, errbuf);
	if (res == -1)
		goto pcap_error_buf;
	       
	sprintf(bpf_str, "udp src port %d", port);

	res = pcap_compile(log->tx, &fp, bpf_str, 0, 0);
	if (res == -1)
		goto pcap_error;

	pcap_setfilter(log->tx, &fp);
	if (res == -1)
		goto pcap_error;

	return;

pcap_error:
	snprintf(errbuf, PCAP_ERRBUF_SIZE, "%s", pcap_geterr(log->tx));
pcap_error_buf:
	pcap_close(log->tx);
pcap_open_error:
	printf("pcap[%s]: %s\n", iface, errbuf);
	BUG();
}

void
log_done(log_s *log)
{
	pcap_close(log->tx);
}

void
log_pkt(u_char *user, const struct pcap_pkthdr *phdr, const u_char *cap)
{
	pt_slog_v *slogv = (pt_slog_v *)user;
	pt_slog_pcap *slog = (pt_slog_pcap *)(cap + LOG_HEADER_SKIP);

	int i = slog->train_seqno;
	int j = 0;

	/* XXX dirty hack to ignore packets without a slog signature */

	if (slog->magic != PT_KSTAMP_MAGIC)
		return;

	BUG_ON(i >= slogv->len);

	if (slog->src_epoch != slogv->slogs[i].src_epoch ||
		slog->src_seqno != slogv->slogs[i].src_seqno)
		return;

	/* Collect the pcap timestamp. */
	slogv->slogs[i].src_pstamp = tick_tv_to_usec(&phdr->ts) -
					slogv->slogs[i].src_base_stamp;

	/* Collect outgoing kernel timestamps */
	BUG_ON(slog->magic != PT_KSTAMP_MAGIC);

	slogv->slogs[i].kstamp_num = slog->kstamp_num;
	BUG_ON(slog->kstamp_num > PT_KSTAMP_MAX);

	for (j = 0; j < slog->kstamp_num; j++)
		slogv->slogs[i].src_kstamp[j] = slog->src_kstamp[j];
}

void
log_capture_train_tx(log_s *log, FILE *flog, pt_slog_v *slogv)
{
	int i = 0;
	int res = 0;
	int dispatched = 0;

	/* A train of packets has been sent. Information is stored
	 * in @slogs. Get the packet departure timestamps from pcap
	 * and log the results.
	 */

	while (1) {
		res = pcap_dispatch(log->tx, slogv->len,
				log_pkt, (u_char *)slogv);
		if (res == 0)
			break;
		dispatched++;
	}

	for (i = 0; i < slogv->len; i++) {

		int j = 0;
		pt_slog *slog = &slogv->slogs[i];

		fprintf(flog,
			"%u \t%u \t%u \t%hu \t%hu \t%llu \t%llu "
			"\t%hu \t%hu \t%hu \t%hu \t%hu: ",
			slog->src_epoch,
			slog->src_seqno,
			slog->train_gap,
			slog->train_len,
			slog->train_seqno,
			slog->src_load_epoch,
			slog->src_base_stamp,
			slog->src_sleep_diff,
			slog->src_ustamp1,
			slog->src_ustamp2,
			slog->src_pstamp,
			slog->kstamp_num
			);

		for (j = 0; j < slog->kstamp_num; j++) {
			int kstamp = slog->src_kstamp[j];
			fprintf(flog, "\t%u ", kstamp);
		}

		fprintf(flog, "\n");
	}
	fprintf(flog, "\n");
	fflush(flog);
}

