#ifndef _PROBE_H
#define _PROBE_H

#include <string.h>
#include "mgrp/bhtp.h"
#include "mgrp/udpex.h"

#ifdef __cplusplus
extern "C" {
#endif    

int probe_send_train(int sock, int pkts, int gap, int len);

#define PROBE_FIRST	1
#define PROBE_LAST	2
#define PROBE_MIDDLE	3

#define PROBE_PBK_ON    1
#define PROBE_PBK_OFF   0

/**
 * Call this function to schedule the transmission of a probe packet as part of the train.
 *
 * Actual probe transmission is done by
 *    tagent/modules/mgrp/udpex.c:udpex_transmit_barrier_skbs()
 *
 * Function arguments are:
 *  + sock -- file descriptor of connection to kernel
 *  + order -- must be PROBE_FIRST for first probe, PROBE_LAST for
 *             last probe, PROBE_MIDDLE for all others
 *  + gap -- gap in microseconds from prior probe to this probe
 *           (i.e. wait this long BEFORE sending this probe)
 *  + buf -- header to send; ignored if PROBE_FIRST, and used otherwise
 *  + buflen -- size of header
 *  + daddr -- the destination address, leave NULL if socket connected
 *  + pad -- number of padding bytes to add after the header in this packet
 *  + can_be_piggybacked -- 1 iff piggybacking is okay, 0 otherwise
 *
 */
static inline int
probe_sendto_train_pkt(int sock, int order, long gap,
                       char *buf, int buflen, struct sockaddr_in *daddr,
                       int pad, int can_be_piggybacked)
{
	struct udpex_probe *probe = NULL;

        struct iovec iov = { 0 };
        struct msghdr msg = { 0 };
        struct cmsghdr *cmsg = NULL;

        char ctrl[CMSG_SPACE(sizeof(struct udpex_probe))];

        iov.iov_base = (void *)buf;
        iov.iov_len = buflen;

        msg.msg_name =  daddr;
        msg.msg_namelen = (daddr == NULL ? 0 : sizeof(struct sockaddr_in));
        msg.msg_iov =  &iov;
        msg.msg_iovlen =  1;
        msg.msg_control =  (void *)ctrl;
        msg.msg_controllen =  sizeof(ctrl);
        msg.msg_flags =  0;

	memset(ctrl, 0, sizeof(ctrl));

	cmsg = CMSG_FIRSTHDR(&msg);
	cmsg->cmsg_level = SOL_MGRP;
	cmsg->cmsg_type = UDPEX_PROBE;
	cmsg->cmsg_len = CMSG_LEN(sizeof(struct udpex_probe));

	probe = (struct udpex_probe *)CMSG_DATA(cmsg);

	/* ancillary data */
	/* barrier is true for all packets except last */
	probe->barrier = !(order == PROBE_LAST);
	probe->gap_usec = (order == PROBE_FIRST ? 0 : gap);
	probe->pad = pad;
        probe->piggyback_not_allowed = !can_be_piggybacked;

        // ask always for probe results
        // only matters for the last probe
        probe->tell_results = 1;

#ifdef DEBUG_PROBE
        fprintf(stderr,
                "order 0x%x gap %ld buf %p buflen %d pad %d pbk %d\n",
                order, gap, buf, buflen, pad, can_be_piggybacked);
        fprintf(stderr,
                "barrier %d gap_usec %d cmsg_len %d\n",
                probe->barrier, probe->gap_usec, cmsg->cmsg_len);
#endif
	/* the train actually leaves with the last sendmsg */
	return sendmsg(sock, &msg, 0);
}

static inline int
probe_send_train_pkt(int sock, int order, long gap,
                     char *buf, int buflen,
                     int pad, int can_be_piggybacked)
{
        return probe_sendto_train_pkt(sock, order, gap,
                                      buf, buflen, NULL,
                                      pad, can_be_piggybacked);
}

/*
 * Receives a probe from MGRP
 *
 * @sock        - UDPex/MGRP socket
 * @buf         - buffer to store probe
 *                (this contains both probe header + padding)
 * @buflen      - length of input buffer
 * @tstamps     - if not NULL, function will store the MGRP timestamps here
 *                      -> system receive timestamp
 *                      -> MGRP receive timestamp (same as system for now)
 *                      -> MGRP send timestamp
 */
static inline int
probe_recv_train_pkt(int sock, char *buf, int buflen,
                     struct bhtp_timestamps *tstamps)
{
	int received = 0;

        struct iovec iov = { 0 };
        struct msghdr msg = { 0 };
        struct cmsghdr *cmsg = NULL;

        char ctrl[CMSG_SPACE(sizeof(struct bhtp_timestamps))];

        iov.iov_base = buf;
        iov.iov_len = buflen;

        msg.msg_name = NULL;
        msg.msg_namelen = 0;
        msg.msg_iov = &iov;
        msg.msg_iovlen = 1;
        msg.msg_control = (void *)ctrl;
        msg.msg_controllen = sizeof(ctrl);
        msg.msg_flags = 0;

        received = recvmsg(sock, &msg, 0);
        if (received == -1)
                return received;

        cmsg = CMSG_FIRSTHDR(&msg);
        if (cmsg &&
            tstamps &&
            cmsg->cmsg_level == SOL_MGRP &&
            cmsg->cmsg_type == UDPEX_TSTAMP)
                *tstamps = *(struct bhtp_timestamps *)CMSG_DATA(cmsg);

        return received;
}

#ifdef __cplusplus
}
#endif

#endif
