/*
 * mpstat: per-processor statistics
 * (C) 2000-2007 by Sebastien GODARD (sysstat <at> orange.fr)
 *
 ***************************************************************************
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published  by  the *
 * Free Software Foundation; either version 2 of the License, or (at  your *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it  will  be  useful,  but *
 * WITHOUT ANY WARRANTY; without the implied warranty  of  MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License *
 * for more details.                                                       *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                   *
 ***************************************************************************
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <ctype.h>
#include <sys/utsname.h>
#include <assert.h>

#include "mpstat.h"
#include "common.h"
#include "libsysstat.h"


#define _(string) (string)

#define SCCSID "@(#)" __FILE__ " compiled " __DATE__ " " __TIME__
char *sccsid(void) { return (SCCSID); }

unsigned long long uptime[3] = {0, 0, 0};
unsigned long long uptime0[3] = {0, 0, 0};
struct mp_stats *st_mp_cpu[3];
/* NOTE: Use array of _char_ for bitmaps to avoid endianness problems...*/
unsigned char *cpu_bitmap;	/* Bit 0: Global; Bit 1: 1st proc; etc. */
struct tm mp_tstamp[3];

/* Nb of processors on the machine */
int cpu_nr = 0;


/*
 ***************************************************************************
 * Allocate mp_stats structures and cpu bitmap
 ***************************************************************************
 */
void salloc_mp_cpu(int nr_cpus)
{
   int i;

   for (i = 0; i < 3; i++) {
      if ((st_mp_cpu[i] = (struct mp_stats *) malloc(MP_STATS_SIZE * nr_cpus)) == NULL) {
	 perror("malloc");
	 exit(4);
      }

      memset(st_mp_cpu[i], 0, MP_STATS_SIZE * nr_cpus);
   }

   if ((cpu_bitmap = (unsigned char *) malloc((nr_cpus >> 3) + 1)) == NULL) {
      perror("malloc");
      exit(4);
   }

   memset(cpu_bitmap, 0, (nr_cpus >> 3) + 1);
}


/*
 ***************************************************************************
 * Recalculate interval based on each CPU's tick count
 ***************************************************************************
 */
unsigned long long mget_per_cpu_interval(struct mp_stats *st_mp_cpu_i,
					 struct mp_stats *st_mp_cpu_j)
{
   return ((st_mp_cpu_i->cpu_user + st_mp_cpu_i->cpu_nice +
	    st_mp_cpu_i->cpu_system + st_mp_cpu_i->cpu_iowait +
	    st_mp_cpu_i->cpu_hardirq + st_mp_cpu_i->cpu_softirq +
	    st_mp_cpu_i->cpu_steal + st_mp_cpu_i->cpu_idle) -
	   (st_mp_cpu_j->cpu_user + st_mp_cpu_j->cpu_nice +
	    st_mp_cpu_j->cpu_system + st_mp_cpu_j->cpu_iowait +
	    st_mp_cpu_j->cpu_hardirq + st_mp_cpu_j->cpu_softirq +
	    st_mp_cpu_j->cpu_steal + st_mp_cpu_j->cpu_idle));
}


/*
 ***************************************************************************
 * Core function used to compute statistics
 ***************************************************************************
 */
void compute_stats_core(int prev, int curr,
                        struct mp_stats_ratio_double *ratio,
                        int ratio_num)
{
   struct mp_stats
      *smci = st_mp_cpu[curr] + 1,
      *smcj = st_mp_cpu[prev] + 1;
   unsigned long long itv, pc_itv;
   int cpu;

   assert(ratio_num >= cpu_nr + 1);

   /* Compute time interval */
   itv = get_interval(uptime[prev], uptime[curr]);

   /* Check if we want global stats among all proc */
   if (*cpu_bitmap & 1) {
       memset(&ratio[0], 0, sizeof(struct mp_stats_ratio_double));
       ratio[0].cpu_user    = ll_sp_value(st_mp_cpu[prev]->cpu_user,    st_mp_cpu[curr]->cpu_user,    itv);
       ratio[0].cpu_nice    = ll_sp_value(st_mp_cpu[prev]->cpu_nice,    st_mp_cpu[curr]->cpu_nice,    itv);
       ratio[0].cpu_system  = ll_sp_value(st_mp_cpu[prev]->cpu_system,  st_mp_cpu[curr]->cpu_system,  itv);
       ratio[0].cpu_iowait  = ll_sp_value(st_mp_cpu[prev]->cpu_iowait,  st_mp_cpu[curr]->cpu_iowait,  itv);
       ratio[0].cpu_hardirq = ll_sp_value(st_mp_cpu[prev]->cpu_hardirq, st_mp_cpu[curr]->cpu_hardirq, itv);
       ratio[0].cpu_softirq = ll_sp_value(st_mp_cpu[prev]->cpu_softirq, st_mp_cpu[curr]->cpu_softirq, itv);
       ratio[0].cpu_steal   = ll_sp_value(st_mp_cpu[prev]->cpu_steal,   st_mp_cpu[curr]->cpu_steal  , itv);
       /* Handle buggy kernels */
       if (st_mp_cpu[curr]->cpu_idle < st_mp_cpu[prev]->cpu_idle)
          ratio[0].cpu_idle = 0.0;
       else
          ratio[0].cpu_idle = ll_sp_value(st_mp_cpu[prev]->cpu_idle, st_mp_cpu[curr]->cpu_idle, itv);
   }

   /* Reduce interval value to one processor */
   if (cpu_nr > 1)
      itv = get_interval(uptime0[prev], uptime0[curr]);

   if (*cpu_bitmap & 1) {
      ratio[0].irq = ll_s_value(st_mp_cpu[prev]->irq, st_mp_cpu[curr]->irq, itv);
   }

   for (cpu = 1; cpu <= cpu_nr; cpu++, smci++, smcj++) {

       memset(&ratio[cpu], 0, sizeof(struct mp_stats_ratio_double));

      /* Check if we want stats about this proc */
      if (!(*(cpu_bitmap + (cpu >> 3)) & (1 << (cpu & 0x07))))
	 continue;

      /* Recalculate itv for current proc */
      pc_itv = mget_per_cpu_interval(smci, smcj);

      if (pc_itv) {
	 /* Current CPU is not offline */
         ratio[cpu].cpu_user    = ll_sp_value(smcj->cpu_user,    smci->cpu_user,    pc_itv);
         ratio[cpu].cpu_nice    = ll_sp_value(smcj->cpu_nice,    smci->cpu_nice,    pc_itv);
         ratio[cpu].cpu_system  = ll_sp_value(smcj->cpu_system,  smci->cpu_system,  pc_itv);
         ratio[cpu].cpu_iowait  = ll_sp_value(smcj->cpu_iowait,  smci->cpu_iowait,  pc_itv);
         ratio[cpu].cpu_hardirq = ll_sp_value(smcj->cpu_hardirq, smci->cpu_hardirq, pc_itv);
         ratio[cpu].cpu_softirq = ll_sp_value(smcj->cpu_softirq, smci->cpu_softirq, pc_itv);
         ratio[cpu].cpu_steal   = ll_sp_value(smcj->cpu_steal,   smci->cpu_steal,   pc_itv);
         if (smci->cpu_idle < smcj->cpu_idle)
            ratio[cpu].cpu_idle = 0.0;
         else
            ratio[cpu].cpu_idle = ll_sp_value(smcj->cpu_idle, smci->cpu_idle, pc_itv);
         ratio[cpu].irq = ll_s_value(smcj->irq, smci->irq, itv);
      }
   }
}



/*
 ***************************************************************************
 * Read stats from /proc/stat
 ***************************************************************************
 */
void read_proc_stat(int curr)
{
   FILE *fp;
   struct mp_stats *st_mp_cpu_i;
   static char line[80];
   unsigned long long cc_user, cc_nice, cc_system, cc_hardirq, cc_softirq;
   unsigned long long cc_idle, cc_iowait, cc_steal;
   int proc_nb;

   if ((fp = fopen(STAT, "r")) == NULL) {
      fprintf(stderr, _("Cannot open %s: %s\n"), STAT, strerror(errno));
      exit(2);
   }

   while (fgets(line, 80, fp) != NULL) {

      if (!strncmp(line, "cpu ", 4)) {
	 /*
	  * Read the number of jiffies spent in the different modes
	  * among all proc. CPU usage is not reduced to one
	  * processor to avoid rounding problems.
	  */
	 st_mp_cpu[curr]->cpu_iowait = 0;	/* For pre 2.5 kernels */
	 cc_hardirq = cc_softirq = cc_steal = 0;
	 /* CPU counters became unsigned long long with kernel 2.6.5 */
	 sscanf(line + 5, "%llu %llu %llu %llu %llu %llu %llu %llu",
		&(st_mp_cpu[curr]->cpu_user),
		&(st_mp_cpu[curr]->cpu_nice),
		&(st_mp_cpu[curr]->cpu_system),
		&(st_mp_cpu[curr]->cpu_idle),
		&(st_mp_cpu[curr]->cpu_iowait),
		&(st_mp_cpu[curr]->cpu_hardirq),
		&(st_mp_cpu[curr]->cpu_softirq),
		&(st_mp_cpu[curr]->cpu_steal));

	 /*
	  * Compute the uptime of the system in jiffies (1/100ths of a second
	  * if HZ=100).
	  * Machine uptime is multiplied by the number of processors here.
	  */
	 uptime[curr] = st_mp_cpu[curr]->cpu_user +
	                st_mp_cpu[curr]->cpu_nice +
	                st_mp_cpu[curr]->cpu_system +
	                st_mp_cpu[curr]->cpu_idle +
	                st_mp_cpu[curr]->cpu_iowait +
	    		st_mp_cpu[curr]->cpu_hardirq +
	    		st_mp_cpu[curr]->cpu_softirq +
	    		st_mp_cpu[curr]->cpu_steal;
      }

      else if (!strncmp(line, "cpu", 3)) {
	 /*
	  * Read the number of jiffies spent in the different modes
	  * (user, nice, etc.) for current proc.
	  * This is done only on SMP machines.
	  */
	 cc_iowait = cc_hardirq = cc_softirq = cc_steal = 0;
	 sscanf(line + 3, "%d %llu %llu %llu %llu %llu %llu %llu %llu",
		&proc_nb,
		&cc_user, &cc_nice, &cc_system, &cc_idle, &cc_iowait,
		&cc_hardirq, &cc_softirq, &cc_steal);

	 if (proc_nb < cpu_nr) {
	    st_mp_cpu_i = st_mp_cpu[curr] + proc_nb + 1;
	    st_mp_cpu_i->cpu_user    = cc_user;
	    st_mp_cpu_i->cpu_nice    = cc_nice;
	    st_mp_cpu_i->cpu_system  = cc_system;
	    st_mp_cpu_i->cpu_idle    = cc_idle;
	    st_mp_cpu_i->cpu_iowait  = cc_iowait;
	    st_mp_cpu_i->cpu_hardirq = cc_hardirq;
	    st_mp_cpu_i->cpu_softirq = cc_softirq;
	    st_mp_cpu_i->cpu_steal   = cc_steal;
	 }
	 /* else additional CPUs have been dynamically registered in /proc/stat */
	
	 if (!proc_nb && !uptime0[curr])
	    /*
	     * Compute uptime reduced for one proc using proc#0.
	     * Done only if /proc/uptime was unavailable.
	     */
	    uptime0[curr] = cc_user + cc_nice + cc_system + cc_idle +
	                    cc_iowait + cc_hardirq + cc_softirq + cc_steal;
      }

      else if (!strncmp(line, "intr ", 5))
	 /*
	  * Read total number of interrupts received since system boot.
	  * Interrupts counter became unsigned long long with kernel 2.6.5.
	  */
	 sscanf(line + 5, "%llu", &(st_mp_cpu[curr]->irq));
   }

   fclose(fp);
}


/*
 ***************************************************************************
 * Read stats from /proc/interrupts
 ***************************************************************************
 */
void read_interrupts_stat(int curr)
{
   FILE *fp;
   struct mp_stats *st_mp_cpu_i;
   static char *line = NULL;
   unsigned long irq = 0;
   unsigned int cpu;
   char *cp, *next;

   for (cpu = 0; cpu < cpu_nr; cpu++) {
      st_mp_cpu_i = st_mp_cpu[curr] + cpu + 1;
      st_mp_cpu_i->irq = 0;
   }

   if ((fp = fopen(INTERRUPTS, "r")) != NULL) {

      if (!line) {
	 if ((line = (char *) malloc(INTERRUPTS_LINE + 11 * cpu_nr)) == NULL) {
	    perror("malloc");
	    exit(4);
	 }
      }

      while (fgets(line, INTERRUPTS_LINE + 11 * cpu_nr, fp) != NULL) {

	 if (isdigit(line[2])) {
	
	    /* Skip over "<irq>:" */
	    if ((cp = strchr(line, ':')) == NULL)
	       continue;
	    cp++;
	
	    for (cpu = 0; cpu < cpu_nr; cpu++) {
	       st_mp_cpu_i = st_mp_cpu[curr] + cpu + 1;
	       irq = strtol(cp, &next, 10);
	       st_mp_cpu_i->irq += irq;
	       cp = next;
	    }
	 }
      }

      fclose(fp);
   }
}


/*
 ***************************************************************************
 * External interface
 ***************************************************************************
 */

void mpstat_init()
{
   struct utsname header;

   /* Get HZ */
   get_HZ();

   /* How many processors on this machine ? */
   cpu_nr = get_cpu_nr(~0);

   /*
    * cpu_nr: a value of 2 means there are 2 processors (0 and 1).
    * In this case, we have to allocate 3 structures: global, proc0 and proc1.
    */
   salloc_mp_cpu(cpu_nr + 1);

   /*
    * Set bit for every processor.
    * Also indicate to display stats for CPU 'all'.
    */
   memset(cpu_bitmap, 0xff, ((cpu_nr + 1) >> 3) + 1);

   /* Get time */
   get_localtime(&(mp_tstamp[0]));

   /* Get system name, release number and hostname */
   uname(&header);
   print_gal_header(&(mp_tstamp[0]), header.sysname, header.release,
		    header.nodename);

   /* Read initial stats */
   uptime0[0] = 0;
   readp_uptime(&(uptime0[0]));
   read_proc_stat(0);
   read_interrupts_stat(0);

   /* Save the first stats collected. Will be used to compute the average */
   mp_tstamp[2] = mp_tstamp[0];
   uptime[2] = uptime[0];
   uptime0[2] = uptime0[0];
   memcpy(st_mp_cpu[2], st_mp_cpu[0], MP_STATS_SIZE * (cpu_nr + 1));
}

void mpstat_read_stats(int curr)
{
   int cpu = 0;
   struct mp_stats *st_mp_cpu_i, *st_mp_cpu_j;

   /*
    * Resetting the structure not needed since every fields will be set.
    * Exceptions are per-CPU structures: some of them may not be filled
    * if corresponding processor is disabled (offline).
    */
   for (cpu = 1; cpu <= cpu_nr; cpu++) {
      st_mp_cpu_i = st_mp_cpu[curr] + cpu;
      st_mp_cpu_j = st_mp_cpu[!curr] + cpu;
      *st_mp_cpu_i = *st_mp_cpu_j;
   }

   /* Get time */
   get_localtime(&(mp_tstamp[curr]));

   /* Read stats */
   if (cpu_nr > 1) {
      uptime0[curr] = 0;
      readp_uptime(&(uptime0[curr]));
   }
   read_proc_stat(curr);
   read_interrupts_stat(curr);
}

void mpstat_compute_total_stats(int curr, 
                                struct mp_stats_ratio_double *ratio,
                                int ratio_num)
{
   compute_stats_core(2, curr, ratio, ratio_num);
}

void mpstat_compute_interval_stats(int curr, 
                                   struct mp_stats_ratio_double *ratio,
                                   int ratio_num)
{
   compute_stats_core(!curr, curr, ratio, ratio_num);
}

int mpstat_advance_current(int curr)
{
   /* this works because @curr oscillates between 0 and 1 */
   return curr ^ 1;
}

int mpstat_init_current()
{
   /* start with the second entry; the first entry contains the initial stats */
   return 1;
}

int mpstat_get_cpu_num()
{
        return cpu_nr;
}

static inline int mpstat_ratio_str_header(char *str, int len)
{
      int written = snprintf(str, len,
                             "cpu  "
                             "  %%user "
                             "  %%nice "
                             "   %%sys "
                             "%%iowait "
                             "   %%irq "
                             "  %%soft "
                             " %%steal "
                             "  %%idle "
                             "   intr/s"
                             "\n");
        return written;
}

void mpstat_ratio_double_str(char *str, int len, int with_header,
                             struct mp_stats_ratio_double *ratio,
                             int ratio_num)
{
   int i = 0;
   int str_used = 0;

   memset(str, 0, len);

   if (with_header) {
      int buflen_max = len - str_used;
      char *buf = &str[str_used];
      int written = mpstat_ratio_str_header(buf, buflen_max);

      if (written > buflen_max) return;
      str_used += written;
   }

   for (i = 0; i < ratio_num; i++) {
      int written = 0;
      int buflen_max = len - str_used;
      char *buf = &str[str_used];

      written = snprintf(buf, buflen_max,
                         "%3d  "
                         "%7.3f "
                         "%7.3f "
                         "%7.3f "
                         "%7.3f "
                         "%7.3f "
                         "%7.3f "
                         "%7.3f "
                         "%7.3f "
                         "%9.3f "
                         "\n",
                         i,
                         ratio[i].cpu_user,
                         ratio[i].cpu_nice,
                         ratio[i].cpu_system,
                         ratio[i].cpu_iowait,
                         ratio[i].cpu_hardirq,
                         ratio[i].cpu_softirq,
                         ratio[i].cpu_steal,
                         ratio[i].cpu_idle,
                         ratio[i].irq);

      if (written > buflen_max) return;
      str_used += written;
   }
}

void
mpstat_ratio_int_oneline_str(char *str, int len,
                             struct mp_stats_ratio_int *ratio,
                             int idx)
{
      snprintf(str, len, "%3d  "
                         "%3d.%.3d "
                         "%3d.%.3d "
                         "%3d.%.3d "
                         "%3d.%.3d "
                         "%3d.%.3d "
                         "%3d.%.3d "
                         "%3d.%.3d "
                         "%3d.%.3d "
                         "%5d.%.3d ",
                         idx,
                         ratio[idx].cpu_user / 1000,
                         ratio[idx].cpu_user % 1000,
                         ratio[idx].cpu_nice / 1000,
                         ratio[idx].cpu_nice % 1000,
                         ratio[idx].cpu_system / 1000,
                         ratio[idx].cpu_system % 1000,
                         ratio[idx].cpu_iowait / 1000,
                         ratio[idx].cpu_iowait % 1000,
                         ratio[idx].cpu_hardirq / 1000,
                         ratio[idx].cpu_hardirq % 1000,
                         ratio[idx].cpu_softirq / 1000,
                         ratio[idx].cpu_softirq % 1000,
                         ratio[idx].cpu_steal / 1000,
                         ratio[idx].cpu_steal % 1000,
                         ratio[idx].cpu_idle / 1000,
                         ratio[idx].cpu_idle % 1000,
                         ratio[idx].irq / 1000,
                         ratio[idx].irq % 1000);
}


void mpstat_ratio_int_str(char *str, int len, int with_header,
                             struct mp_stats_ratio_int *ratio,
                             int ratio_num)
{
   int i = 0;
   int str_used = 0;

   memset(str, 0, len);

   if (with_header) {
      int buflen_max = len - str_used;
      char *buf = &str[str_used];
      int written = mpstat_ratio_str_header(buf, buflen_max);

      if (written > buflen_max) return;
      str_used += written;
   }

   for (i = 0; i < ratio_num; i++) {
      int written = 0;
      int buflen_max = len - str_used;
      char *buf = &str[str_used];
      char tmp[200];

      mpstat_ratio_int_oneline_str(tmp, sizeof(tmp), ratio, i);
      written = snprintf(buf, buflen_max, "%s\n", tmp);

      if (written > buflen_max) return;
      str_used += written;
   }
}

void mpstat_downgrade_ratio_double_to_int(
                        const struct mp_stats_ratio_double *from_ratio_double,
                        struct mp_stats_ratio_int *to_ratio_int)
{
   memset(to_ratio_int, 0, sizeof(struct mp_stats_ratio_int));

   to_ratio_int->cpu_idle    = 1000 * from_ratio_double->cpu_idle;
   to_ratio_int->cpu_iowait  = 1000 * from_ratio_double->cpu_iowait;
   to_ratio_int->cpu_user    = 1000 * from_ratio_double->cpu_user;
   to_ratio_int->cpu_nice    = 1000 * from_ratio_double->cpu_nice;
   to_ratio_int->cpu_system  = 1000 * from_ratio_double->cpu_system;
   to_ratio_int->cpu_hardirq = 1000 * from_ratio_double->cpu_hardirq;
   to_ratio_int->cpu_softirq = 1000 * from_ratio_double->cpu_softirq;
   to_ratio_int->cpu_steal   = 1000 * from_ratio_double->cpu_steal;
   to_ratio_int->irq         = 1000 * from_ratio_double->irq;
}
