/*
 * Created on		Tue Feb 28 08:46:26 EST 2006
 *
 * Author		Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 * Based on an older version of a library with the same name
 *
 */

#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

/* dlsym() - retrieve symbols for socket api */
/* The __* defines seem to be needed for RTLD_NEXT. */
#define _GNU_SOURCE
#define __USE_GNU
#include <dlfcn.h>

#include "err.h"
#include "bitset.h"

#include "tsocket.h"

#include "mgrp/bhtp.h"
#include "mgrp/udpex.h"
#include "mgrp/tcp.h"

#define TSOCKET_MAX_SOCKET		4096

#define TSOCKET_HIJACK_UDP		1
#define TSOCKET_HIJACK_TCP		2

/* Functions from lib.c that we are overriding. */

/* socket related */

typedef int (tsocket_socket_func)
        (int domain, int type, int protocol);

typedef int (tsocket_close_func)
        (int s);

typedef int (tsocket_setsockopt)
	(int s, int level, int optname, const void *optval, socklen_t optlen);

/* gethostbyname, gethostbyaddr, gethostbyname_r */

typedef struct hostent *(tsocket_gethostbyname)
        (const char *name);

typedef struct hostent *(tsocket_gethostbyaddr)
        (const char *addr, int len, int type);

typedef int (tsocket_gethostbyname_r)
        (const char *name,
        struct hostent *ret, char *buf, size_t buflen,
        struct hostent **result, int *h_errnop);

/* local private state */

struct tsocket_state {

	int valid;
	int inhibited;

	struct bitset hijacked;

	struct {
		tsocket_socket_func *socket;
		tsocket_close_func *close;
		tsocket_setsockopt *setsockopt;

		tsocket_gethostbyname *gethostbyname;
		tsocket_gethostbyaddr *gethostbyaddr;
		tsocket_gethostbyname_r *gethostbyname_r;

	} orig;

	struct {
		/* options */
		int mss;
		int tcp;
		int udp;
		int reuse;
		int probe;
		int queue;
		int fraglen;
	} op;
};

static struct tsocket_state tsocket = {
	.valid = 0,
};


void tsocket_init();
void tsocket_done();

static inline int
tsocket_init_if_needed()
{
	int needed = !tsocket.valid;
	if (needed)
		tsocket_init();
	return needed;
}

static inline int
tsocket_is_hijacked(int s)
{
	tsocket_init_if_needed();
	return bitset_contains(&tsocket.hijacked, s);
}

static inline int
tsocket_intercept(int sock)
{
	tsocket_init_if_needed();
	
	return 1;
}


static inline int
tsocket_inhibit_start()
{
	tsocket_init_if_needed();
	tsocket.inhibited++;
	DEBUG("inhibit_start level %d", tsocket.inhibited);
	return tsocket.inhibited;
}


static inline int
tsocket_inhibit_stop()
{
	tsocket_init_if_needed();
	BUG_ON(!tsocket.inhibited);
	tsocket.inhibited--;
	DEBUG("inhibit_stop level %d", tsocket.inhibited);
	return tsocket.inhibited;
}

static void *
tsocket_get_symbol(char *symbol)
{
	void *handle = RTLD_NEXT;
	void *ptr = NULL;
	char *error = NULL;

	ptr = dlsym(handle, symbol);
	error = dlerror();

	ERR_ON(error != NULL, "dlsym failed: %s", error);
	ERR_ON(ptr == NULL, "symbol for '%s' is NULL", symbol);

	return ptr;
}

int
tsocket_get_option_int(char *option)
{
	int found = 0;
	int ival = 0;
	char *val = NULL;
	
	val = getenv(option);
	if (val == NULL)
		return 0;

	found = sscanf(val, "%d", &ival);
	if (!found)
		return 0;

	DEBUG("found option %s: %d", option, ival);
	return ival;
}

void
tsocket_init()
{
	DEBUG("tsocket_init");
	memset(&tsocket, 0, sizeof(tsocket));

	/* The bit set holds all the sockets that have been hijacked. */

	bitset_init(&tsocket.hijacked, TSOCKET_MAX_SOCKET);

	/* Get pointers original socket functions. */

	tsocket.orig.socket = tsocket_get_symbol("socket");
	tsocket.orig.close = tsocket_get_symbol("close");
	tsocket.orig.setsockopt = tsocket_get_symbol("setsockopt");

	tsocket.orig.gethostbyname = tsocket_get_symbol("gethostbyname");
	tsocket.orig.gethostbyname_r = tsocket_get_symbol("gethostbyname_r");
	tsocket.orig.gethostbyaddr = tsocket_get_symbol("gethostbyaddr");

	/* Pick up options from the environment TSOCKET_XXX */
	tsocket.op.mss = tsocket_get_option_int("TSOCKET_MSS");
	tsocket.op.tcp = tsocket_get_option_int("TSOCKET_TCP");
	tsocket.op.udp = tsocket_get_option_int("TSOCKET_UDP");
	tsocket.op.reuse = tsocket_get_option_int("TSOCKET_REUSE");
	tsocket.op.probe = tsocket_get_option_int("TSOCKET_PROBE");
	tsocket.op.queue = tsocket_get_option_int("TSOCKET_QUEUE");
	tsocket.op.fraglen = tsocket_get_option_int("TSOCKET_FRAGLEN");

	tsocket.valid = 1;
}

void
tsocket_done()
{
	BUG_ON(!tsocket.valid);
	DEBUG("tsocket_init");
	
	bitset_done(&tsocket.hijacked);
	memset(&tsocket, 0, sizeof(tsocket));
}

/* Functions from libc that are intercepted. */

static inline int
tsocket_hijack(int domain, int type, int protocol)
{
	tsocket_init_if_needed();

	/* User can inhibit hijacking explicitly with tsocket_inhibit_*() */
	if (tsocket.inhibited)
		return 0;
	
	/* UDP is hjacked. */
	if (tsocket.op.udp &&
	    domain == AF_INET &&
	    type == SOCK_DGRAM &&
	    (protocol == IPPROTO_IP || protocol == IPPROTO_UDP))
		return TSOCKET_HIJACK_UDP;

	/* TCP is hjacked. */
	if ((tsocket.op.tcp || tsocket.op.mss) &&
	    domain == AF_INET &&
	    type == SOCK_STREAM &&
	    (protocol == IPPROTO_IP || protocol == IPPROTO_TCP))
		return TSOCKET_HIJACK_TCP;
	
	return 0;
}

int
socket(int domain, int type, int protocol)
{
	int hijack = 0;
	int s = -1;

	DEBUG("socket");
	
	hijack = tsocket_hijack(domain, type, protocol);

	if (hijack == TSOCKET_HIJACK_UDP) {
		/* Hijack the socket call. */
		type = SOCK_DGRAM;
		protocol = IPPROTO_MGRP;

	} else if (hijack == TSOCKET_HIJACK_TCP) {
		/* nothing yet; we do setsockopt() later */

	} else {
		BUG_ON(hijack != 0);
	}

	s = tsocket.orig.socket(domain, type, protocol);

	DEBUG("Original socket returned %d", s);

	if (s == -1) {
		DEBUG("socket error %s", strerror(errno));
		return s;
	}

	if (hijack) {
		DEBUG("Hijacking socket %d", s);
		bitset_insert(&tsocket.hijacked, s);
	}

	if (hijack == TSOCKET_HIJACK_UDP) {

		if (tsocket.op.probe) {
			int rc = 0;
			/* Tell the UDPex socket that we are sending probes. */
			/* Send immediately; do not piggyback */
			rc = ioctl(s, UDPEX_IOC_PROBE);
			SYSERR_ON(rc);
			/* HACK: udpex reserves the top 100 bytes and reuses
			 * the rest */
		}

		if (tsocket.op.queue) {
			int rc = 0;
			/* Packets can be qeueued for piggybacking. */
			rc = ioctl(s, UDPEX_IOC_QUEUE);
			SYSERR_ON(rc);
		}

	} else if (hijack == TSOCKET_HIJACK_TCP) {

		int rc = 0;
		int actual_mss = 0;
		size_t mss_size = sizeof(int);
		int on = 1;

		/* limit the MSS so that it can tunnel through MGRP */

		if (tsocket.op.tcp && !tsocket.op.mss)
			tsocket.op.mss = MGRP_MAX_TUN_MSS;

		if (tsocket.op.mss > 0) {

			if (tsocket.op.mss > MGRP_MAX_TUN_MSS) {
				DEBUG("requested MSS %d too large; new MSS %d",
					tsocket.op.mss, MGRP_MAX_TUN_MSS);
				tsocket.op.mss = MGRP_MAX_TUN_MSS;
			}

			DEBUG("setsockopt setting MSS to %d", tsocket.op.mss);
			rc = tsocket.orig.setsockopt(s,
					IPPROTO_TCP, TCP_MAXSEG,
					&tsocket.op.mss,
					sizeof(tsocket.op.mss));
			SYSERR_ON(rc);

			rc = getsockopt(s, IPPROTO_TCP, TCP_MAXSEG,
					&actual_mss, &mss_size);

			SYSERR_ON(rc);
			DEBUG("getsockopt TCP_MAXSEG: asked for %d got %d",
				tsocket.op.mss, actual_mss);
		}

		/* turn on TCP over MGRP */

		if (tsocket.op.tcp) {
			DEBUG("Turning ON TCP_MGRP");
			on = 1;
			rc = setsockopt(s, IPPROTO_TCP, TCP_MGRP,
					&on, sizeof(on));
			SYSERR_ON(rc);

			/* put tcp_bhtp socket in fragmentation mode */
			if (tsocket.op.fraglen) {
				int fraglen = tsocket.op.fraglen;

				/* turn on fragmentation */
				on = 1;
				rc = setsockopt(s,
						SOL_MGRP, TCP_MGRP_FRAGMENT,
						&on, sizeof(on));
				SYSERR_ON(rc);

				/* set the maximum fragment size */
				rc = setsockopt(s,
						SOL_MGRP, TCP_MGRP_FRAG_LEN_MAX,
						&fraglen, sizeof(fraglen));
				SYSERR_ON(rc);
			}


		}
	}

	return s;
}

int
close(int s)
{
	int rc = 0;

	DEBUG("close %d", s);

	rc = tsocket.orig.close(s);

	if (rc == -1)
		return rc;

	/* No harm done if not hijacked. */
	bitset_remove(&tsocket.hijacked, s);

	return rc;
}

int
setsockopt(int s, int level, int optname, const void *optval, socklen_t optlen)
{
	int rc = 0;

	if (s == -1 && level == -1 && optname == -1) {
		int on = *(int *)optval;
		if (on)
			tsocket_inhibit_start();
		else
			tsocket_inhibit_stop();
		return rc;
	}

	if (tsocket_is_hijacked(s)) {

		if (level == IPPROTO_TCP && optname == TCP_MAXSEG) {
			int mss = *(int *)optval;
			DEBUG("setsockopt %d TCP_MAXSEG with MSS %d", s, mss);

			/* tsocket.op.mss is always set if we are hijacking TCP */
			if (tsocket.op.mss && mss > tsocket.op.mss) {
				mss = tsocket.op.mss;
				DEBUG("    --> limiting MSS to %d", mss);
			}

			rc = tsocket.orig.setsockopt(s, level, optname,
						&mss, sizeof(mss));
			return rc;
		}
	}

	rc = tsocket.orig.setsockopt(s, level, optname, optval, optlen);
	return rc;
}

/* gethostbyname, gethostbyaddr, gethostbyname_r
 *
 * We intercept the gethostby* family of functions because we do not
 * want to modify any sockets opened by the libc resolver.
 */

struct hostent *
gethostbyname(const char *name)
{
	struct hostent *hptr = NULL;

	DEBUG("gethostbyname");

	/* turn off modifications in socket() */
	tsocket_inhibit_start();

	hptr = tsocket.orig.gethostbyname(name);

	/* turn modifications in socket() back on */
	tsocket_inhibit_stop();
	
	return hptr;
}

struct hostent *
gethostbyaddr(const char *addr, int len, int type)
{
	struct hostent *hptr = NULL;

	DEBUG("gethostbyaddr");

	/* turn off modifications in socket() */
	tsocket_inhibit_start();

	hptr = tsocket.orig.gethostbyaddr(addr, len, type);

	/* turn modifications in socket() back on */
	tsocket_inhibit_stop();

	return hptr;
}

int
gethostbyname_r(const char *name,
        struct hostent *ret, char *buf, size_t buflen,
        struct hostent **result, int *h_errnop)
{
	int rc = 0;

	DEBUG("gethostbyname_r");

	/* turn off modifications in socket() */
	tsocket_inhibit_start();

	rc = tsocket.orig.gethostbyname_r(name, ret, buf, buflen,
					result, h_errnop);

	/* turn modifications in socket() back on */
	tsocket_inhibit_stop();

	return rc;
}

