/*
 * BHTP		Bottom Half Transport Protocol
 *
 * Created:	Sun Mar 26 19:36:51 EST 2006
 *
 * Author:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 */

#ifndef _BHTP_H
#define _BHTP_H

#if __KERNEL__
#include <linux/types.h>
#include <linux/in.h>
#include <linux/net.h>
#include <net/tcp.h>
#include <linux/tcp.h>

#include "tick.h"

#define ASSERT(condition) BUG_ON(!(condition))

#else
/* sys/types.h must be defined before linux/types.h
 * to avoid 'conflicting types' error.
 *
 * see for an example: http://sourceware.org/ml/newlib/2005/msg00611.html
 */
#include <stdio.h>
#include <sys/types.h>
#include <linux/types.h>
#include <assert.h>

#define ASSERT(condition) assert(condition)

#endif /* __KERNEL__ */

#ifdef __cplusplus
extern "C" {
#endif    

/* FIXME: HACK: XXX:
 *
 * IPPROTO_XXX should be be defined in linux/in.h 
 * SOL_XXX should be be defined in linux/socket.h 
 *
 * see also 'man getprotoent'
 */
#define IPPROTO_MGRP		254
#define SOL_MGRP		254

#define BHTP_MAX_SEG		2	/* power of two please */

#define BHTP_PAD_BYTE		0xaf

#define BHTP_PROTO_TCP          0x1
#define BHTP_PROTO_UDP          0x2
#define MGRP_PROTO_FRAG         0x3
#define BHTP_PROTO_MAX          0x3

#if __KERNEL__
static inline int
mgrp_to_ip_protocol(int mgrp_proto)
{
        if (mgrp_proto == BHTP_PROTO_TCP)
                return IPPROTO_TCP;
        if (mgrp_proto == BHTP_PROTO_UDP)
                return IPPROTO_UDP;

        // cannot match MGRP protocol with IP protocol
        ASSERT(0);
        return -1;
}
#endif

static inline const char *
mgrp_tun_proto_str(int tun_proto)
{
        if (tun_proto == BHTP_PROTO_TCP)
                return "tcp";
        if (tun_proto == BHTP_PROTO_UDP)
                return "udp";
        if (tun_proto == MGRP_PROTO_FRAG)
                return "frag";
        return "UNKNOWN";
}

#define MAX_BHTP_HEADER	sizeof(struct bhtphdr) + MAX_HEADER

/* Default values for module parameters */
/* #define BHTP_LATENCY_NS		(2 * NSEC_PER_MSEC) */
#define BHTP_LATENCY_NS		0
#define MGRP_MAXDELAY_NS	(2 * NSEC_PER_MSEC)
#define MGRP_MAXPBK_PERCENT     0
#define BHTP_PROBE_TIMEOUT_NS	0
#define BHTP_PROBE_ACCUMULATED	0

/* Maximum MSS for tunneled transports.
 *
 * TCP header(20) + options(12) + payload <= 1400
 * Leaves 80 bytes for BHTP + 20bytes IP = 1500
 */
#define MGRP_MAX_TUN_MSS	(1400 - 20 - 12)

/*
 * Maximum UDP payload for tunneled transports.
 *
 * UDP header(8) + payload <= 1400
 * Leaves 80 bytes for BHTP + 20bytes IP = 1500
 */
// #define MGRP_MAX_TUN_UDP	(1400 - 8)	

// magic number comes from MGRP-fixed max TCP MSS
// XXX hack because of the piggybacking bug where
// probes with sizes larger than the MGRP/TCP MSS
// are not piggybacked.
#define MGRP_MAX_TUN_UDP	1278

// MGRP sequence number

typedef __u32 mgrp_seqno_t;

#define MGRP_SEQNO_MAX          ((mgrp_seqno_t)-1)
#define MGRP_SEQNO_MID          (MGRP_SEQNO_MAX/2)

// Headers

struct bhtphdr_tun {
	__u16
              // encapsulated protocol
              proto:2,

              /*
               * Only one TUN per packet can have the padding flag set.
               *
               * The number of padding bytes can be derived from the
               * MGRP header, the total packet bytes and the bytes
               * occupied by the TUNs. See bhtp_orig_pad_bytes().
               */
              needs_padding:1,

              // portion of this TUN has already been sent
              // check before pushing up to receiving protocol
              duplicate:1,

              // unused (will be used for fragmentation)
              __unused_flag:1,

              // total length of the TUN (hdr+data)
              len:11;
};

struct bhtphdr {

        mgrp_seqno_t seqno;

        // The 48bit sndstamp represents nanosec from a base timestamp.
        //
        // The base timestamp is a per-flow attribute set to the
        // current time when the flow is created in bhtp_flow_init().
        //
        // This means that it rolls over every 78 hours.
        //
        // I broke up the 48 bits into 32bits and 16bits, to fit in the
        // header.  Probably not so smart network programming, but, hey,
        // I am just writing a thesis..
        __u32 sndstamp_msb32;
        __u16 sndstamp_lsb16;

        __u8 tuns:2,

             /*
              * When @has_unused_padding flag is set, this is what you know:
              *
              *   - there is at least one pad byte at the end of the packet
              *   - there is a probe TUN that needs padding
              *     (but the reverse is not always true, eg, fully reused pad)
              *   - if TUNs > 1, then some of the original padding was reused
              *                  and this padding is the unused padding
              *   - if TUN == 1, then this padding equals the original padding
              *   - use bhtp_unused_pad_bytes() to get the number of pad bytes
              */
             has_unused_padding:1,

             _unused_flag:1,
             _unused_proto:4;

        __u8 _unused_ackreq;

	struct bhtphdr_tun tun[BHTP_MAX_SEG];
};

// Helper comparison functions for MGRP Seqno

static inline mgrp_seqno_t
mgrp_seqno_add(mgrp_seqno_t seqno, int offset)
{
        return (seqno + offset);
}

static inline mgrp_seqno_t
mgrp_seqno_inc(mgrp_seqno_t seqno)
{
        return mgrp_seqno_add(seqno, 1);
}

// returns distance @lhs - @rhs
// =0 they are equal
// >0 lhs > rhs
// <0 lhs < rhs
static inline long long
mgrp_seqno_distance(mgrp_seqno_t lhs, mgrp_seqno_t rhs)
{
        long long l = lhs;
        long long r = rhs;
        long long dist = l-r;
        long long absdist = (dist < 0 ? -dist : dist);

        if (absdist < MGRP_SEQNO_MID) {
                // no rollover
                return dist;
        }

        // the smaller of the two has rolled over (and is really ahead)
        if (l < r) {
                // positive
                dist = (MGRP_SEQNO_MAX - r) + 1 + l;
        } else {
                // negative
                dist = (MGRP_SEQNO_MAX - l) + 1 + r;
                dist *= -1;
        }

        return dist;
}

// 1 if @lhs < @rhs
static inline int
mgrp_seqno_before(mgrp_seqno_t lhs, mgrp_seqno_t rhs)
{
        return (mgrp_seqno_distance(lhs, rhs) < 0);
}

// 1 if @lhs > @rhs
static inline int
mgrp_seqno_after(mgrp_seqno_t lhs, mgrp_seqno_t rhs)
{
        return (mgrp_seqno_distance(lhs, rhs) > 0);
}

// 1 if @lhs == @rhs
static inline int
mgrp_seqno_equal(mgrp_seqno_t lhs, mgrp_seqno_t rhs)
{
        return (mgrp_seqno_distance(lhs, rhs) == 0);
}

// sndtstamp

#define BHTP_SNDSTAMP_MASK              0x0000ffffffffffffLL
#define BHTP_SNDSTAMP_OVERFLOW_TEST     (~BHTP_SNDSTAMP_MASK)

static inline unsigned long long bhtp_get_sndstamp_ns(const struct bhtphdr *bh)
{
        unsigned long long nsec_msb32 = bh->sndstamp_msb32;
        unsigned long long nsec_lsb16 = bh->sndstamp_lsb16;

        return (nsec_msb32<<16) | nsec_lsb16;
}

// 0 when ok, -1 on overflow
static inline int bhtp_set_sndstamp_ns(struct bhtphdr *bh,
                                       unsigned long long nsec)
{
        int overflow = 0;

        // sndstamp overflow
        if (BHTP_SNDSTAMP_OVERFLOW_TEST & nsec) {
                overflow = 1;
                nsec &= BHTP_SNDSTAMP_MASK;
        }

        bh->sndstamp_msb32 = nsec>>16;
        bh->sndstamp_lsb16 = nsec & 0xffffLL;

        return overflow;
}

// tuns

static inline int
bhtp_header_tuns(const struct bhtphdr *bh)
{
        return bh->tuns;
}

static inline void
bhtp_header_tuns_set(struct bhtphdr *bh, int tuns)
{
        bh->tuns = tuns;
}

// tcp debugging

static inline void
__mgrp_add_flag(int flag, char flag_char,
                char **buf_p, int *remaining_p)
{
        int n = 0;

        if (!flag)
                return;

        n = snprintf(*buf_p, *remaining_p, "%c", flag_char);
        ASSERT(n < *remaining_p);
        *buf_p += n;
        *remaining_p -= n;
}

#if __KERNEL__

// use only in the kernel
// conflicts with incompatible redefinition of tcphdr in tcpdump/tcp.h

static inline void
mgrp_tcphdr_oneliner(char *str, int len, struct tcphdr *th, int pktlen)
{
        int n = 0;
        char *buf = str;
        int remaining = len;
        int remaining_before_flags = 0;

        int hdr_bytes = 4 * th->doff;
        int data_bytes = pktlen - hdr_bytes;
        int src_port = ntohs(th->source);
        int dst_port = ntohs(th->dest);
        int seq = ntohl(th->seq);
        int ack = ntohl(th->ack_seq);

        n = snprintf(buf, remaining, "[tcp %d] %d > %d: ",
                    pktlen, src_port, dst_port);
        ASSERT(n < remaining);
        buf += n;
        remaining -= n;

        remaining_before_flags = remaining;
        __mgrp_add_flag(th->fin, 'F', &buf, &remaining);
        __mgrp_add_flag(th->syn, 'S', &buf, &remaining);
        __mgrp_add_flag(th->rst, 'R', &buf, &remaining);
        __mgrp_add_flag(th->psh, 'P', &buf, &remaining);
        if (remaining == remaining_before_flags)
                __mgrp_add_flag(1, '.', &buf, &remaining);

        __mgrp_add_flag(1, ' ', &buf, &remaining);

        if (th->fin || th->syn || th->rst) {
                n = snprintf(buf, remaining, "%d:%d(%d): ",
                             seq, seq + data_bytes, data_bytes);
                ASSERT(n < remaining);
                buf += n;
                remaining -= n;
        }

        n = snprintf(buf, remaining, "ack %d options %d", ack, hdr_bytes-20);
        ASSERT(n < remaining);
        buf += n;
        remaining -= n;
}

#endif

// byte counting in BHTP header

static inline int
bhtp_tun_bytes(const struct bhtphdr *bh)
{
        int i = 0;
        int bytes = 0;
        int tuns = bhtp_header_tuns(bh);

        for (i = 0; i < tuns; i++)
                bytes += bh->tun[i].len;

        return bytes;
}

static inline int
bhtp_pad_sanity_checks(const struct bhtphdr *bh, int pktlen,
                       int tuns,
                       int tun_bytes,
                       int padding_tuns,
                       int unused_pad_bytes,
                       int reused_pad_byes,
                       char *error,
                       int errlen)
{
        int max_avail_tun_bytes = pktlen - sizeof(struct bhtphdr);

        if (tuns < 0) {
                if (error)
                        snprintf(error, errlen,
                                 "expecting at least one TUN got %d", tuns);
                return -1;
        }
        if (tun_bytes > max_avail_tun_bytes) {
                if (error)
                        snprintf(error, errlen,
                                 "TUN length mismatch: tuns ask for %d got only %d",
                                 tun_bytes, max_avail_tun_bytes);
                return -1;
        }
        if (unused_pad_bytes > 0 && !bh->has_unused_padding) {
                if (error)
                        snprintf(error, errlen,
                                 "MGRP pkt has %d padbytes but no header flag",
                                 unused_pad_bytes);
                return -1;
        }
        if (unused_pad_bytes == 0 && bh->has_unused_padding) {
                if (error)
                        snprintf(error, errlen,
                                 "MGRP pkt has no padbytes but the header flag is set");

                return -1;
        }
        if (padding_tuns > 1) {
                if (error)
                        snprintf(error, errlen,
                                 "at most one TUN can need padding: got %d",
                                 padding_tuns);
                return -1;
        }
        if (padding_tuns == 0 && unused_pad_bytes > 0) {
                if (error)
                        snprintf(error, errlen,
                                 "invalid MGRP packet (len %d): "
                                 "has %d padbytes but does not contain a probe",
                                 pktlen, unused_pad_bytes);
                return -1;
        }
        return 0;
}

//
// Returns the actual number of pad bytes in the packet.
//
// This represents the unused part of the padding since some
// of the padding can be reused. See notes in struct bhtphdr.
//
// @pktlen is the full length of the BHTP packet (including the BHTP header)
static inline int
bhtp_unused_pad_bytes(const struct bhtphdr *bh, int pktlen)
{
        if (!bh->has_unused_padding)
                return 0;
        // the unused padding is after all the TUNs
        return pktlen - sizeof(struct bhtphdr) - bhtp_tun_bytes(bh);
}

//
// Returns the number of the original pad bytes of the embedded probe
//
// @pktlen is the full length of the BHTP packet (including the BHTP header)
//
// If @reusedp, @unusedp are non-NULL, then they contain the byte breakdown.
// Both together total the original pad bytes.
//
// Also a good place for some header sanity checks.
// Returns -1 if the sanity checks fail, >= 0 on success.
static inline int
bhtp_orig_pad_bytes(const struct bhtphdr *bh, int pktlen,
                    int *reusedp, int *unusedp,
                    char *error, int errlen)
{
        int i = 0;
        int rc = 0;
        int tun_bytes = 0;
        int padding_tuns = 0;
        int potentially_reused_pad_bytes = 0;
        int unused_pad_bytes = 0;
        int tuns = bhtp_header_tuns(bh);

        for (i = 0; i < tuns; i++) {
                tun_bytes += bh->tun[i].len;
                if (bh->tun[i].needs_padding) {
                        padding_tuns++;
                } else {
                        potentially_reused_pad_bytes += bh->tun[i].len;
                }
        }
        unused_pad_bytes = pktlen - sizeof(struct bhtphdr) - tun_bytes;

        rc = bhtp_pad_sanity_checks(bh, pktlen,
                                    tuns,
                                    tun_bytes,
                                    padding_tuns,
                                    unused_pad_bytes,
                                    potentially_reused_pad_bytes,
                                    error, errlen);
        if (rc)
                return -1;

        if (padding_tuns == 0)
                return 0;

        if (reusedp != NULL)
                *reusedp = potentially_reused_pad_bytes;
        if (unusedp != NULL)
                *unusedp = unused_pad_bytes;

        return (potentially_reused_pad_bytes + unused_pad_bytes);
}

// Returns 1 if MGRP packet is probe
//
// Header errors are silently ignored.
// If you want to catch errors, call bhtp_orig_pad_bytes() instead.
static inline int
bhtp_pkt_is_probe(const struct bhtphdr *bh, int pktlen)
{
        int pad_orig = bhtp_orig_pad_bytes(bh, pktlen, NULL, NULL, NULL, 0);

        // errors silently ignored
        if (pad_orig == -1)
                return 0;

        return (pad_orig > 0);
}

#define BHTP_TSTAMP_WALLCLOCK		(1<<0)
#define BHTP_TSTAMP_MONOTONIC		(1<<1)

#define BHTP_TSTAMP_LOCAL		(1<<2)
#define BHTP_TSTAMP_REMOTE		(1<<3)

#define BHTP_TSTAMP_TAKEN_BELOW_BHTP	(1<<6)
#define BHTP_TSTAMP_TAKEN_AT_BHTP	(1<<7)

struct bhtp_tstamp {
	int type;
	unsigned long long nsec;
};

struct bhtp_timestamps {
	struct bhtp_tstamp sys;
	struct bhtp_tstamp rcv;
	struct bhtp_tstamp snd;
};

#ifdef __KERNEL__
static inline void
bhtp_tstamp_to_timeval(struct bhtp_tstamp *tstamp, struct timeval *tv)
{
	*tv = ns_to_timeval(tstamp->nsec);
}
#else
static inline void
bhtp_tstamp_to_timeval(struct bhtp_tstamp *tstamp, struct timeval *tv)
{
	unsigned long long sec =  (tstamp->nsec / 1000000000L);
	unsigned long long nsec = (tstamp->nsec % 1000000000L);
	unsigned long long usec = nsec / 1000;
	if (nsec % 1000 > 500)
		usec += 1;
	tv->tv_sec = sec;
	tv->tv_usec = usec;
}
#endif

#ifndef __KERNEL__
#include <stdio.h>
static inline void
bhtp_print_tv_diff(struct timeval *tv, struct timeval *prev)
{
	unsigned long long diff = (unsigned long long)-1;
	if (prev) {
		diff = tv->tv_sec;
		diff -= prev->tv_sec;
		diff *= 1000000;
		diff += tv->tv_usec;
		diff -= prev->tv_usec;
	}

	printf("sec %-10ld usec %-10ld diff %-10lld",
		tv->tv_sec, tv->tv_usec, diff);
}

static inline void
bhtp_print_stamp_diff(struct bhtp_timestamps *stamps,
		      struct bhtp_timestamps *prev)
{
	if (prev != NULL) {
		printf("dsys %-10llu ", stamps->sys.nsec - prev->sys.nsec);
		printf("drcv %-10llu ", stamps->rcv.nsec - prev->rcv.nsec);
		printf("dsnd %-10llu",  stamps->snd.nsec - prev->snd.nsec);
	}
}
static inline void
bhtp_print_stamp_diff_usec(struct bhtp_timestamps *stamps,
		           struct bhtp_timestamps *prev)
{
	if (prev != NULL) {
		printf("dsys %-10llu ", (stamps->sys.nsec - prev->sys.nsec)/1000);
		printf("drcv %-10llu ", (stamps->rcv.nsec - prev->rcv.nsec)/1000);
		printf("dsnd %-10llu",  (stamps->snd.nsec - prev->snd.nsec)/1000);
	}
}

static inline void
bhtp_print_stamp(struct bhtp_timestamps *stamps)
{
	printf("sys %llu rcv %llu snd %llu",
		stamps->sys.nsec, stamps->rcv.nsec, stamps->snd.nsec);
}
#endif

/* pkt points to the top of the bhtp header; whole packet in contiguous memory area */
static inline const unsigned char *
bhtp_get_tun_chunk(const struct bhtphdr *bh, int tindex, const unsigned char *pkt)
{
	const unsigned char *tdata = pkt;
	int above = 0;
	int t = 0;

	for (t = 0; t < tindex; t++)
		above += bh->tun[t].len;

	tdata += sizeof(struct bhtphdr);
	tdata += above;

	return tdata;
}

static inline int
u8int(__u8 u)
{
	int i = u;
	return i;
}

static inline int
u16int(__u16 u)
{
	int i = u;
	return i;
}

static inline int
u32int(__u32 u)
{
	int i = u;
	return i;
}

static inline int
u64int(__u64 u)
{
	int i = u;
	return i;
}


#define u2i(u) \
	(sizeof(u) == 1 ? u8int(u) : \
	 sizeof(u) == 2 ? u16int(u) : \
	 sizeof(u) == 4 ? u32int(u) : \
	 sizeof(u) == 8 ? u64int(u) : (u))

#ifdef __cplusplus
}
#endif    


#endif /* _BHTP_H */

