/*
 * BHTP		Bottom Half Transport Protocol
 *
 * Created:	Sun Mar 26 19:36:51 EST 2006
 *
 * Author:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 */

#ifndef _BHTP_ABI_H
#define _BHTP_ABI_H

#include "log.h"
#include "bhtp.h"
#include "tick.h"
#include "proto.h"
#include "mgrp_common.h"
#include "mgrp_counters.h"

#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>

#define MGRP_PBK_DEFAULT_POLICY         MGRP_PBK_FRAGMENT_SENDONCE

enum mgrp_pbk_policy {

        MGRP_PBK_INVALID = 0,

        /*
         * The _STANDALONE_ policies piggyback transport data chunks
         * that are valid standalone fragments. The fragments are
         * generated through the ops->fragment callback.
         */

        // sends every packet once
        MGRP_PBK_STANDALONE_SENDONCE,

        // sends every packet once piggybacked and once normally
        MGRP_PBK_STANDALONE_REDUNDANT,

        /*
         * The _FRAGMENT_ policies perform fragmentation and reassembly
         * at the MGRP layer. These policies do not produce piggybacked
         * chunks that are valid standalone packets.
         *
         * These policies add a fragmenation header to all TUN chunks.
         * The MGRP protocol is set to MGRP_PROTO_FRAG pseudo protocol
         * and the fragmenation header contains the next "real" protocol.
         */

        // sends every packet once
        MGRP_PBK_FRAGMENT_SENDONCE,

        // sends every packet once piggybacked and once normally
        MGRP_PBK_FRAGMENT_REDUNDANT,

        MGRP_PBK_MAX
};

static inline const char *
mgrp_pbk_policy_str(int policy) {

        const char *mgrp_pbk_policy_string[MGRP_PBK_MAX] = {
                [MGRP_PBK_INVALID]              = "MGRP_PBK_INVALID",
                [MGRP_PBK_STANDALONE_SENDONCE]  = "MGRP_PBK_STANDALONE_SENDONCE",
                [MGRP_PBK_STANDALONE_REDUNDANT] = "MGRP_PBK_STANDALONE_REDUNDANT",
                [MGRP_PBK_FRAGMENT_SENDONCE]    = "MGRP_PBK_FRAGMENT_SENDONCE",
                [MGRP_PBK_FRAGMENT_REDUNDANT]   = "MGRP_PBK_FRAGMENT_REDUNDANT",
        };

        if (policy < 0 || policy >= MGRP_PBK_MAX)
                return "ERROR: UNKNOWN POLICY";
        
        return mgrp_pbk_policy_string[policy];
}

struct mgrp_probe_feedback {
        int mgrp_bytes;
        int pbk_bytes;
        ktime_t rider_orig_send_time;
};

static inline void
mgrp_probe_feedback_init(struct mgrp_probe_feedback *feedback, int mgrp_bytes)
{
        memset(feedback, 0, sizeof(*feedback));
        feedback->mgrp_bytes = mgrp_bytes;
}

// http://sd.wareonearth.com/~phil/net/overhead/
//
// Ethernet overhead bytes:
// 12 gap + 8 preamble + 14 header + 4 trailer = 38 bytes/packet w/o 802.1q
// 12 gap + 8 preamble + 18 header + 4 trailer = 42 bytes/packet with 802.1q
#define MGRP_OVERHEAD_ETHERNET  (12 + 8 + 14 + 4)
#define MGRP_OVERHEAD_NETWORK   (MGRP_OVERHEAD_ETHERNET + sizeof(struct iphdr))
#define MGRP_OVERHEAD_TRANSPORT (MGRP_OVERHEAD_NETWORK + sizeof(struct bhtphdr))

/* bhtp_ut state flags */
#define BHTP_TX_SCHED          0x0001

#define BHTP_FLOW_ACK_ALL	(1<<0)
#define BHTP_FLOW_ONE_DST	(1<<1)
#define BHTP_FLOW_SHARED	(1<<2)
#define BHTP_FLOW_PKT_SHARE	(1<<3)

/*
 * These are byte lengths of the packet below the UDP header.
 *
 * So, for example when sending a UDP packet send(buf:1000bytes),
 * then the packet lengths:
 *
 *	tun_len		= UDPex(8) + probe_payload()
 *	tun_frag_max_len= maximum length of tun chunk
 *			  (if max_len < tun_len then bhtp needs to fragment)
 *
 *	tun_hdr		= UDPex(8)
 *	tun_data	= probe_payload()
 *	tun_data_orig	= tun_data + tun_pad
 *	tun_data_frag	= n/a (portion of top data already fragmented)
 *	tun_len		= tun_hdr + tun_data
 *	tun_pad		= 1000 - tun_data
 *
 *	** non-bhtp case:
 *	ip_packet_len	= IP(20) + UDP(8) + 1000
 *
 *	** bhtp-case (no piggybacking):
 *	ip_packet_len	= IP(20) + BHTP() + tun_hdr + tun_data + tun_pad
 *
 *	** bhtp_case (piggybacking):
 *	ip_packet_len	= IP(20) + BHTP() +
 *		(tun_hdr + tun_data) +			// udp probe
 *		(tun2_hdr + tun2_data) +		// piggybacked
 *		(tun_pad - tund2_hdr - tun2_data)	// leftover padding
 */

#define BHTP_QSKB_MAGIC		0xaf
#define BHTP_QSKB_MAGIC_POISON	0x11

#define BHTP_QSKB_PROBE		(1<<0)
#define BHTP_QSKB_QUEUE		(1<<1)
#define BHTP_QSKB_RECALC_HDR	(1<<2)
#define BHTP_QSKB_FRAGMENT	(1<<3)
#define BHTP_QSKB_NEVER_QUEUE	(1<<4)
#define BHTP_QSKB_VESSEL	(1<<5)
#define BHTP_QSKB_FRAG_EXHAUSTED        (1<<6)
#define MGRP_QSKB_DUPLICATE             (1<<7)
#define MGRP_QSKB_FRAGMENT_CHUNK        (1<<8)
#define MGRP_QSKB_FRAGID                (1<<9)

struct bhtp_flow;

// MGRP state inside an skb (stored in the control buffer: skb.cb[48])
struct bhtp_qskb {
	struct bhtp_tun *tun;

        u16 type;
        u16 magic;
        u32 fragid;

        // used by mgrp_fragment_chunk() to create fragments that
        // are not valid standalone fragments
        // does not require cooperation from transport protocol
	unsigned short tun_len;
	unsigned short tun_chunk_frag;

        // used by ops->fragment() to create valid standalone fragments
        // requires cooperation from transport protocol
	unsigned short tun_frag_len_max;
	unsigned short tun_hdr;
	unsigned short tun_data;
	unsigned short tun_data_orig;
	unsigned short tun_data_frag;

        union {
                struct {
                        // first probe is 1
                        u8 probe_pos_in_train;
                        u8 probes_in_train;
                } vessel;
                // unused variable taking space to pad structure
                // to 32-bit boundary
                unsigned short __unused;
        };

	short tun_pad;
        short tun_piggyback;
	int tun_gap;

	union {
		struct {
			struct bhtphdr *bh;
		} recv;
		struct {
			struct bhtp_flow_req *req;
		} send;
	};

	ktime_t kt;

        // When the module is loaded bhtp_is_safe() makes sure that this
        // structure fits inside 48 bytes (the skb control buffer size)
};

#define MAX_INT_VALUE(intvar)   ((1 << (8*sizeof(intvar))) - 1)

static inline int
bhtp_qskb_is_safe(void)
{
	struct sk_buff tmp;
        struct bhtp_qskb dummy_qskb;
	int safe = sizeof(struct bhtp_qskb) <= sizeof(tmp.cb);

	LOG("bhtp_qskb %s safe: bhtp_qskb_size %d cb_size %d max_len %d",
		(safe ? "is" : "is NOT"),
		sizeof(struct bhtp_qskb),
		sizeof(tmp.cb),
                MAX_INT_VALUE(dummy_qskb.tun_len));

	return safe;
}

static inline int
bhtp_qsk_valid(struct bhtp_qskb *qskb) {
	return (qskb->magic == BHTP_QSKB_MAGIC);
}

static inline struct bhtp_qskb *
bhtp_qsk_novalidate(struct sk_buff *skb) {
	struct bhtp_qskb *qskb = (struct bhtp_qskb *)skb->cb;
	BUG_ON(sizeof(struct bhtp_qskb) > sizeof(skb->cb));
	return qskb;
}

static inline struct bhtp_qskb *
__bhtp_qsk_trace(struct sk_buff *skb, const char *filename, int line)
{
        struct bhtp_qskb *qskb = bhtp_qsk_novalidate(skb);
        if (!bhtp_qsk_valid(qskb)) {
                ERRFN("%s:%d: invalid qsk in skb %p len %d magic %d%s",
                      filename, line, skb, skb->len, qskb->magic,
                      (qskb->magic == BHTP_QSKB_MAGIC_POISON ? " (poisoned)" : ""));
                BUG();
        }
        return qskb;
}

#define bhtp_qsk(_skb)  __bhtp_qsk_trace((_skb), __FILE__, __LINE__)

static inline struct bhtp_qskb *
bhtp_qsk_init(struct sk_buff *skb) {
	struct bhtp_qskb *qskb = bhtp_qsk_novalidate(skb);
	memset(qskb, 0, sizeof(*qskb));
	qskb->magic = BHTP_QSKB_MAGIC;
	return qskb;
}

static inline void
mgrp_qsk_invalidate(struct sk_buff *skb) {
	struct bhtp_qskb *qskb = bhtp_qsk_novalidate(skb);
	memset(qskb, 0, sizeof(*qskb));
        qskb->magic = BHTP_QSKB_MAGIC_POISON;
}

/* Same as bhtp_qsk_init but clones @from_qskb instead of zeroing it out */
static inline struct bhtp_qskb *
bhtp_qsk_clone(struct sk_buff *skb, struct bhtp_qskb *from_qskb) {
	struct bhtp_qskb *qskb = bhtp_qsk_novalidate(skb);
	memcpy(qskb, from_qskb, sizeof(*qskb));
	return qskb;
}

static inline void
mgrp_tskb_type_unset(struct sk_buff *skb, int type)
{
        u32 type_bits = type;
        bhtp_qsk(skb)->type &= ~type_bits;
}

static inline void
mgrp_tskb_type_set(struct sk_buff *skb, int type)
{
        bhtp_qsk(skb)->type |= type;
}

static inline int
mgrp_tskb_has_type(struct sk_buff *skb, int type)
{
        return ((bhtp_qsk(skb)->type & type) > 0);
}

static inline int
mgrp_tskb_fragid_get(struct sk_buff *skb, u32 *fragid)
{
	struct bhtp_qskb *qskb = bhtp_qsk(skb);
        if (!(qskb->type & MGRP_QSKB_FRAGID))
                return -1;
        if (fragid != NULL)
                *fragid = qskb->fragid;
        return 0;
}

static inline void
mgrp_tskb_fragid_set(struct sk_buff *skb, u32 fragid)
{
	struct bhtp_qskb *qskb = bhtp_qsk(skb);

        // you must fragid_unset before overwriting fragid
        if ((qskb->type & MGRP_QSKB_FRAGID) && (fragid != qskb->fragid))
                ERRFN("overwriting fragid %d with %d", qskb->fragid, fragid);

        qskb->fragid = fragid;
        qskb->type |= MGRP_QSKB_FRAGID;
}

static inline void
mgrp_tskb_fragid_unset(struct sk_buff *skb)
{
	struct bhtp_qskb *qskb = bhtp_qsk(skb);
        qskb->fragid = 0;
        mgrp_tskb_type_unset(skb, MGRP_QSKB_FRAGID);
}

static inline int
mgrp_policy_redundant(enum mgrp_pbk_policy policy)
{
        return (policy == MGRP_PBK_STANDALONE_REDUNDANT ||
                policy == MGRP_PBK_FRAGMENT_REDUNDANT);
}

static inline int
mgrp_policy_standalone(enum mgrp_pbk_policy policy)
{
        return (policy == MGRP_PBK_STANDALONE_SENDONCE ||
                policy == MGRP_PBK_STANDALONE_REDUNDANT);
}

static inline int
mgrp_policy_sendonce(enum mgrp_pbk_policy policy)
{
        return (policy == MGRP_PBK_STANDALONE_SENDONCE ||
                policy == MGRP_PBK_FRAGMENT_SENDONCE);
}

static inline int
mgrp_policy_fragment(enum mgrp_pbk_policy policy)
{
        return (policy == MGRP_PBK_FRAGMENT_SENDONCE ||
                policy == MGRP_PBK_FRAGMENT_REDUNDANT);
}

/* a BHTP flow */
struct bhtp_flow {

	u32 id;
	u32 flags;
        // 0 or >1
	atomic_t users;
        // 0 or 1
        atomic_t recv_user;

	spinlock_t lock;

	u32 daddr;

	/* part of the flow list */
	struct list_head flist;

	/* points to the poll_list of transports */
	struct list_head tuns;

	struct sock *sk;
	struct dst_entry *dst;
        int dst_dev_speed_mbps;

        enum mgrp_pbk_policy pbk_policy;

	/* list of queued skbs that are waiting to be sent */
	struct sk_buff_head tosend;
	long long tosend_accumulated;
	spinlock_t tosend_lock;
	ktime_t tosend_latency;

        // list of queued skbs after they have been fully piggybacked
        struct sk_buff_head tosend_pbk;

        atomic_t fragid;
	atomic_t reqid;

        /* counters */
        struct mgrp_counters counters;

        /* base timestamp for bhtphdr.sndstamp */
        unsigned long long sndstamp_base_ns;

};


/* upper transport unit */

/* _ops and _proto are defined once per upper transport protocol */
struct bhtp_tun_ops {

	/* rcv: process a received protocol segment: hdr + data */
	int (*rcv)(struct sk_buff *skb, struct sockaddr_in *remote_addr);

	/*
	 * snd: bite a fragment from the main protocol segment
	 *
	 *	Callee must keep track of data already fragmented
	 *	and make sure that qskb length fields are valid.
	 *	(see qskb->tun_data_frag)
	 */
	struct sk_buff * (*fragment)(struct sk_buff *skb,
                                     int frag_len, int copy_data);

	/*
	 * mark the packet so that the receiver can tell it apart
         *
         * This is a valid packet that either came from the transport
         * protocol or was generated with ops->fragment().
	 */
	void (*mark)(struct sk_buff *skb);

        /*
         * Returns 1 if there are any fragments left in the skb.
         *
         * Also returns:
         *
         *      @fraglen_min    min length of next fragment
         *      @fralen_max     max length of next fragment
         *
         *      @has_more_frags 1 or 0, to indicate if it has more
         *                      fragments beyond the current fragment
         *
         * If no fragments are left, the input/output arguments
         * remain unchanged. Any of them can be NULL if the return
         * values are not needed.
         */
	int (*fraglen_left)(struct sk_buff *skb,
                            int *fraglen_min, int *fraglen_max,
                            int *has_more_frags);
};

struct bhtp_tun_prot {
	u8 proto;
	struct bhtp_tun_ops *ops;
};

/* This is an instance, associate with each socket for example */
struct bhtp_tun {
	struct bhtp_tun_prot *prot;
	unsigned long state;
/* 	struct list_head poll_list; */
	unsigned int barrier_pkts;
	unsigned int barrier_total;
	unsigned int barrier_wait_usec;
	unsigned int barrier_min;
	unsigned int barrier_max;
};

#define BHTP_TUN_PROT_HAS_RCV(_prot) \
	((_prot) != NULL && \
	 (_prot)->ops != NULL && \
	 (_prot)->ops->rcv != NULL)

#define BHTP_TUN_PROT_HAS_FRAG(_prot) \
	((_prot) != NULL && \
	 (_prot)->ops != NULL && \
	 (_prot)->ops->fragment != NULL && \
	 (_prot)->ops->fraglen_left != NULL)

#define MGRP_OPS_HAS_FUNC(_ops, _func) \
	 ((_ops)->_func != NULL)

static inline int
bhtp_flow_matches_addr(struct bhtp_flow *bflow, struct sockaddr_in *usin)
{
	return (bflow->daddr == usin->sin_addr.s_addr);
}

static inline int
bhtp_flow_addr(struct bhtp_flow *bflow, int count)
{
	u8 *octets = (u8 *)&bflow->daddr;
	BUG_ON(count <= 0 || count > 4);
	return octets[count-1];
}


/* timestamps */

static inline void
bhtp_recv_get_sysstamp(struct sock *sk, struct sk_buff *skb,
		       struct bhtp_tstamp *tstamp)
{
	// HERE
	/* based on sock_recv_timestamp() */
	struct timeval stamp;

	tstamp->type |= BHTP_TSTAMP_LOCAL;
	tstamp->type |= BHTP_TSTAMP_WALLCLOCK;

	skb_get_timestamp(skb, &stamp);
	if (stamp.tv_sec == 0) {
                __net_timestamp(skb);
                skb_get_timestamp(skb, &stamp);
		tstamp->type |= BHTP_TSTAMP_TAKEN_AT_BHTP;
	} else {
		tstamp->type |= BHTP_TSTAMP_TAKEN_BELOW_BHTP;
	}

	tstamp->nsec = timeval_to_ns(&stamp);

	/* FIXME: is this needed? It is in sock_recv_timestamp() */
/* 	sk->sk_stamp = stamp; */
}

static inline void
bhtp_recv_get_sndstamp(struct sock *sk, struct sk_buff *skb,
		       struct bhtp_tstamp *tstamp)
{
	struct bhtp_qskb *qskb = bhtp_qsk(skb);

	tstamp->type |= BHTP_TSTAMP_REMOTE;
	tstamp->type |= BHTP_TSTAMP_MONOTONIC;
	tstamp->type |= BHTP_TSTAMP_TAKEN_AT_BHTP;

	tstamp->nsec = bhtp_get_sndstamp_ns(qskb->recv.bh);
}

static inline void
bhtp_recv_get_rcvstamp(struct sock *sk, struct sk_buff *skb,
		       struct bhtp_tstamp *tstamp)
{
	tstamp->type |= BHTP_TSTAMP_LOCAL;
	tstamp->type |= BHTP_TSTAMP_MONOTONIC;
        tstamp->nsec = ktime_to_ns(skb->tstamp);

	if (tstamp->nsec > 0) {
		tstamp->type |= BHTP_TSTAMP_TAKEN_BELOW_BHTP;
	} else {
		tstamp->type |= BHTP_TSTAMP_TAKEN_AT_BHTP;
		tstamp->nsec = tick_now_ns();
	}
}

static inline void
bhtp_recv_get_timestamps(struct sock *sk, struct sk_buff *skb,
			 struct bhtp_timestamps *tstamps)
{
	memset(tstamps, 0, sizeof(*tstamps));

	bhtp_recv_get_sysstamp(sk, skb, &tstamps->sys);
	bhtp_recv_get_rcvstamp(sk, skb, &tstamps->rcv);
	bhtp_recv_get_sndstamp(sk, skb, &tstamps->snd);
}

/* start/stop payload buffer */
int bhtp_drain_tosend(struct bhtp_flow *bflow, int drop);

static inline ktime_t
mgrp_global_flow_latency(struct bhtp_flow *bflow)
{
	return ktime_add_ns(ktime_set(0, 0), bhtp.latency);
}

static inline ktime_t
mgrp_local_flow_latency(struct bhtp_flow *bflow)
{
	return bflow->tosend_latency;
}


static inline int
mgrp_flow_payload_buffer_start(struct bhtp_flow *bflow)
{
	if (ktime_to_ns(mgrp_global_flow_latency(bflow)) > 0)
		return 0;
	
	bflow->tosend_latency = ktime_add_ns(ktime_set(0, 0), bhtp.maxdelay);
	return 1;
}

static inline void
mgrp_flow_payload_buffer_stop(struct bhtp_flow *bflow)
{
	int drained = 0;
	if (ktime_to_ns(mgrp_local_flow_latency(bflow)) == 0)
		return;

	bflow->tosend_latency = ktime_set(0, 0);
	drained = bhtp_drain_tosend(bflow, 0);
	if (drained >= 0)
		DEBUG("flow_payload_buffer_stop: drained buffer: %d skbs",
			drained);
}

/* Returns in nanosec the transmission delay of @mgrp_pktsize bytes at @mbps.
 *
 * The @mgrp_pktsize is the size of the packet with the MGRP header.
 *
 * transmission(sec)  = wiresize(bits) / bps(bits/sec)
 * transmission(nsec) = 10^3 * wiresize(bits) / mbps(Mbits/sec)
 */
static inline uint
mgrp_transmission_delay(uint mgrp_pktsize, uint mbps)
{
        uint wire_bits = 8 * (MGRP_OVERHEAD_NETWORK + mgrp_pktsize);
        return (1000 * wire_bits) / mbps;
}

int bhtp_register_tun_prot(struct bhtp_tun_prot *tun_prot);
int bhtp_unregister_tun_prot(struct bhtp_tun_prot *tun_prot);

struct bhtp_flow *bhtp_find_attach_flow(struct sockaddr_in *usin, int flags);
int bhtp_detach_flow(struct bhtp_flow *bflow, struct bhtp_tun *tun);

int bhtp_schedule_skb(struct bhtp_flow *bflow, struct sk_buff *tun_skb,
                      struct mgrp_probe_feedback *feedback);
                
int bhtp_send_barrier_skbs(struct bhtp_flow *bflow,
		struct sock *sk, struct bhtp_tun *tun,
		unsigned long long *twait_p,
		unsigned long long *tsend_p);

int mgrp_flush_piggybacking(struct bhtp_flow *bflow,
                            struct mgrp_probe_feedback *feedback);

#endif /* _BHTP_ABI_H */

