/*
 * Fragmenation Protocol for MGRP
 *
 * Sat Aug  9 16:12:43 EDT 2008 <pavlos@cs.umd.edu>
 *
 * This is a pseudo-protocol to perform fragmentation and reassembly in MGRP.
 * It is not a full transport protocol like TCP or UDP, which also interact
 * with the socket and IP layer.
 *
 * By using a pseudo-protocol, bhtp_rcv() and bhtp_tun_rcv() remain simple
 * and need no special code to handle fragments. It is up to the receive
 * code of the pseudo-protocol (mgrp_frag_tun_recv) to retain state and
 * perform the reassembly. Eventually, the pseudo-protocol will have to
 * pass a packet to one of the "real" transport protocols like TCP or UDP.
 *
 * Using a pseudo-protocol is appropriate because reassembly needs a
 * special header that then points to the next protocol in the chain.
 *
 * Entry points for reassembly code (receiver side):
 *      mgrp_frag_tun_rcv()
 *      mgrp_frag_process_fragment()
 *
 * Entry point for fragmentation code (sender side):
 *      mgrp_fragment_chunk()
 */

// Must come before log.h. Otherwise we get compiler warning
#define LOG_PREFIX "frag: "
#include "log.h"

#include "frag.h"
#include "proto.h"

/******************************************************
 * mgrp_frag_state
 *
 * Global state for fragmentation pseudo-protocol
 *
 * Holds the receive reassembly hash and the operation callbacks
 * for the fragmentation pseudo-protocol.
 *
 * Initialized statically lower down.
 */

struct mgrp_frag_state {
        struct bhtp_tun_prot tun_prot;
        struct bhtp_tun_ops tun_ops;
        struct bhtp_tun tun;

        struct mgrp_frag_hash fhash;            // fragment reassembly
#ifdef LOG_FRAGHASH
        struct debug_fraghash hs;
#endif
} frag;


/******************************************************
 ******************************************************
 * FRAGMENTATION -- receiver side
 *
 * Reassembly of fragments using reassembly hash
 */

/******************************************************
 * mgrp_frag_bucket
 *
 * Each bucket represents a packet to be reassembled.
 * It holds an skb_queue for all the packet fragments.
 */

static inline void
mgrp_frag_key_init(struct mgrp_frag_key *key,
                   struct mgrp_fraghdr *fh,
                   struct sockaddr_in *source_addr)
{
        memset(key, 0, sizeof(*key));
        key->source_addr = source_addr->sin_addr.s_addr;
        key->fragid = fh->fragid;
        key->proto = fh->proto;
}

// @size == -1 indicates open-ended gap
static inline void
mgrp_frag_gap_init(struct mgrp_frag_gap *gap, int offset, int size,
                   struct sk_buff *before, struct sk_buff *after)
{
        memset(gap, 0, sizeof(*gap));
        gap->offset = offset;
        gap->size = size;
        INIT_LIST_HEAD(&gap->glist);
        gap->before = before;
        gap->after = after;
}

static inline void
mgrp_frag_bucket_init(struct mgrp_frag_bucket *bucket,
                      struct mgrp_frag_key *key)
{
        struct mgrp_frag_gap *gap = NULL;

        memset(bucket, 0, sizeof(*bucket));
        bucket->key = *key;
        skb_queue_head_init(&bucket->fragments);
        INIT_LIST_HEAD(&bucket->gaps);

        // insert one open-ended gap
        gap = kmalloc(sizeof(struct mgrp_frag_gap), GFP_ATOMIC);
        BUG_TRAP(gap != NULL);
        mgrp_frag_gap_init(gap, 0, -1, NULL, NULL);
        list_add(&gap->glist, &bucket->gaps);
}

static inline int
mgrp_frag_bucket_oneliner_header(char *str, int len)
{
        return snprintf(str, len,
                        "fragid "
                        "flags "
                        "FRAGS "
                        "recv "
                        "drops_sent "
                        "drops_overlap "
                        "PKTS "
                        "recv "
                        "drops_sent "
                        "drops_overlap");
}

static inline int
mgrp_frag_bucket_oneliner(char *str, int len, struct mgrp_frag_bucket *bucket)
{
        return snprintf(str, len,
                        "%-6d %c%c%c%c%c "
                        "FRAGS %-4hd %-10hd %-13hd "
                        "PKTS %-4hd %-10hd %-13hd"
#ifdef LOG_FRAGHASH
                        "HASH %.16llx %.16llx %.8x"
#endif
                        ,

                        bucket->key.fragid,
                        (bucket->released_fragments     ? 'R' : '.'),
                        (bucket->got_unfragmented       ? 'U' : '.'),
                        (bucket->already_sent           ? 'S' : '.'),
                        (bucket->used_frags_for_sent    ? 'F' : '.'),
                        (bucket->used_packet_for_sent   ? 'P' : '.'),

                        bucket->frags.received,
                        bucket->frags.dropped_because_already_sent,
                        bucket->frags.dropped_because_overlapping,

                        bucket->pkts.received,
                        bucket->pkts.dropped_because_already_sent,
                        bucket->pkts.dropped_because_overlapping
#ifdef LOG_FRAGHASH
                        ,
                        bucket->hs.unhashed,
                        bucket->hs.hashed,
                        bucket->hs.index
#endif
        );
}

static inline void
mgrp_frag_bucket_release_fragments_and_gaps(struct mgrp_frag_bucket *bucket)
{
        struct sk_buff *skb = NULL;
        struct mgrp_frag_gap *gap = NULL;
        struct mgrp_frag_gap *gap_tmp = NULL;

        if (bucket->released_fragments)
                return;

        // free all gaps first because they point to the skbs
        list_for_each_entry_safe(gap, gap_tmp, &bucket->gaps, glist) {
                list_del(&gap->glist);
                kfree(gap);
        }

        // drop all fragments
        while ((skb = skb_dequeue(&bucket->fragments)) != NULL)
                kfree_skb(skb);

        // we need a flag because we emptied the gap list
        bucket->released_fragments = 1;
}

static inline void
mgrp_frag_bucket_release_reassembly_memory(struct mgrp_frag_bucket *bucket)
{
        mgrp_frag_bucket_release_fragments_and_gaps(bucket);
        if (bucket->got_unfragmented) {
                kfree(bucket->unfragmented);
                bucket->got_unfragmented = 0;
                bucket->unfragmented = NULL;
        }
}

static inline void
mgrp_frag_bucket_done(struct mgrp_frag_bucket *bucket)
{
        mgrp_frag_bucket_release_reassembly_memory(bucket);
        memset(bucket, 0, sizeof(*bucket));
}

static inline int
mgrp_frag_bucket_key_matches(struct mgrp_frag_bucket *bucket,
                             struct mgrp_frag_key *other_key)
{
        return (bucket->key.proto == other_key->proto &&
                bucket->key.source_addr == other_key->source_addr &&
                bucket->key.fragid == other_key->fragid);
}

static inline void
mgrp_frag_bucket_insert_fragment_before(struct mgrp_frag_bucket *bucket,
                                        struct sk_buff *oldskb,
                                        struct sk_buff *newskb)
{
        if (oldskb == NULL)
                skb_queue_tail(&bucket->fragments, newskb);
        else
                skb_insert(oldskb, newskb, &bucket->fragments);
}
                                     
static inline void
mgrp_frag_bucket_insert_fragment_after(struct mgrp_frag_bucket *bucket,
                                       struct sk_buff *oldskb,
                                       struct sk_buff *newskb)
{
        if (oldskb == NULL)
                skb_queue_head(&bucket->fragments, newskb);
        else
                skb_append(oldskb, newskb, &bucket->fragments);
}

/******************************************************
 * mgrp_grag_bucket -- gap management in buckets
 */

enum {
        MGRP_FRAG_OVERLAP_NONE = 0,
        MGRP_FRAG_OVERLAP_LEFTSIDE,
        MGRP_FRAG_OVERLAP_MIDDLE,
        MGRP_FRAG_OVERLAP_RIGHTSIDE,
};

static inline int
mgrp_frag_gap_overlaps(struct mgrp_frag_gap *gap,
                       int fragoff, int fraglen, int islast)
{
        // the offset of the rightmost (inclusive) byte
        int right_fragoff = fragoff + fraglen - 1;

        if (gap->size == -1) {
                // open-ended gap
                if (right_fragoff < gap->offset)
                        return MGRP_FRAG_OVERLAP_NONE;
                // the overlap can go all the way past the right side
                if (fragoff <= gap->offset)
                        return MGRP_FRAG_OVERLAP_LEFTSIDE;
                // the overlap starts after the left side of the gap and
                // since it is the last fragment it defines the right side
                if (islast)
                        return MGRP_FRAG_OVERLAP_RIGHTSIDE;
        } else {
                // close-ended gap
                int right_gapoff = gap->offset + gap->size - 1;

                // outside of gap
                if (right_fragoff < gap->offset || fragoff > right_gapoff)
                        return MGRP_FRAG_OVERLAP_NONE;

                // the overlap can go all the way past the right side
                if (fragoff <= gap->offset)
                        return MGRP_FRAG_OVERLAP_LEFTSIDE;

                // the overlap starts after the left side of the gap
                // and must include (or go beyond) the right side of the gap
                if (right_fragoff >= right_gapoff)
                        return MGRP_FRAG_OVERLAP_RIGHTSIDE;
        }

        return MGRP_FRAG_OVERLAP_MIDDLE;
}

static inline void
mgrp_frag_bucket_remove_filled_gaps(struct mgrp_frag_bucket *bucket,
                                    int fragoff, int fraglen)
{
        ERRFN("not implemented");
        BUG();
}

static inline void
mgrp_frag_bucket_adjust_gap_leftside(struct mgrp_frag_bucket *bucket,
                                     struct mgrp_frag_gap *gap,
                                     struct sk_buff *fragment,
                                     int *removed_gap)
{
        struct mgrp_fraghdr *fh = (struct mgrp_fraghdr *)fragment->transport_header;
        int fragoff = fh->fragoff;
        int fraglen = fragment->len;

        // we only need the part of the fragment that starts
        // at the left side of the gap
        int usable_size = fraglen - (gap->offset - fragoff);
        int gap_exceeded = 0;
        int advance_offset = 0;
        int gap_newsize = 0;

        // the fragment covers the left side of the gap
        // (but may extend beyond the right side)
        BUG_TRAP(gap->offset >= fragoff);
        BUG_TRAP(usable_size > 0);
        BUG_TRAP(removed_gap != NULL);

        *removed_gap = 0;

        if (gap->size == -1) {
                // we found the open-ended gap
                // usual case when fragments come in order
                advance_offset = usable_size;
                gap_newsize = (fh->more_fragments ? -1 : 0);

        } else if (gap->size > usable_size) {
                // we found an intermediate gap
                // this fragment fills the gap only partially
                advance_offset = usable_size;
                gap_newsize = gap->size - usable_size;

        } else if (gap->size <= usable_size) {
                // we found an intermediate gap
                // this fragment fills the gap completely
                // and may go beyond the gap
                advance_offset = gap->size;
                gap_newsize = 0;
                gap_exceeded = (gap->size < usable_size);
        }

        // insert the fragment in the fragment list
        mgrp_frag_bucket_insert_fragment_after(bucket, gap->before, fragment);
        gap->before = fragment;
        gap->offset += advance_offset;
        gap->size = gap_newsize;

        DEBUG_REAS("modified gap o:%d l:%d", gap->offset, gap->size);

        // no more need for this gap
        if (gap->size == 0) {
                DEBUG_REAS("removing gap");
                list_del(&gap->glist);
                *removed_gap = 1;
        }
        // remove any other gaps that are covered by this fragment
        if (gap_exceeded)
                mgrp_frag_bucket_remove_filled_gaps(bucket, fragoff, fraglen);
}
                                      
static inline void
mgrp_frag_bucket_adjust_gap_rightside(struct mgrp_frag_bucket *bucket,
                                      struct mgrp_frag_gap *gap,
                                      struct sk_buff *fragment)
{
        struct mgrp_fraghdr *fh = (struct mgrp_fraghdr *)fragment->transport_header;
        int fragoff = fh->fragoff;
        int fraglen = fragment->len;
        int gap_exceeded = 0;
        int gap_newsize = 0;

        // the fragment covers the right side of the gap
        // (but may not include the left side)
        BUG_TRAP(gap->offset < fragoff);
        BUG_TRAP(fraglen > 0);

        if (gap->size == -1) {
                // open-ended gap will become close-ended
                // must be the last fragment to be on rightside
                BUG_TRAP(!fh->more_fragments);
                gap_newsize = fragoff - gap->offset;

        } else {
                int offset_beyond_gap = gap->offset + gap->size;
                int gap_expects_filled = offset_beyond_gap - fragoff;

                // fragment must overlap and big enough to cover right side
                BUG_TRAP(gap_expects_filled > 0);
                BUG_TRAP(fraglen >= gap_expects_filled);

                gap_newsize = gap->size - gap_expects_filled;
                gap_exceeded = (fraglen > gap_expects_filled);
        }
        
        // insert fragment in fragment list
        mgrp_frag_bucket_insert_fragment_before(bucket, gap->after, fragment);
        gap->after = fragment;
        gap->size = gap_newsize;

        DEBUG_REAS("modified gap o:%d l:%d", gap->offset, gap->size);

        // remove any other gaps that are covered by this fragment
        if (gap_exceeded)
                mgrp_frag_bucket_remove_filled_gaps(bucket, fragoff, fraglen);
}

static inline void
mgrp_frag_bucket_adjust_gap_middle(struct mgrp_frag_bucket *bucket,
                                   struct mgrp_frag_gap *gap,
                                   struct sk_buff *fragment)
{
        struct mgrp_fraghdr *fh = (struct mgrp_fraghdr *)fragment->transport_header;
        int fragoff = fh->fragoff;
        int fraglen = fragment->len;
        int gap_newsize = 0;

        // we always create a new gap (old gap is divided into two)
        struct mgrp_frag_gap *newgap = kmalloc(sizeof(*newgap), GFP_ATOMIC);

        // the fragment should not touch the left side
        BUG_TRAP(gap->offset < fragoff);
        BUG_TRAP(newgap != NULL);

        if (gap->size == -1) {
                // the fragment should not touch the right side
                BUG_TRAP(fh->more_fragments);

                // the open-ended gap becomes close-ended
                gap_newsize = fragoff - gap->offset;

                // create a new open-ended gap for the right side
                mgrp_frag_gap_init(newgap,
                                   fragoff + fraglen, -1, 
                                   fragment, NULL);

        } else {
                int offset_beyond_gap = gap->offset + gap->size;
                int max_frag_could_fill = offset_beyond_gap - fragoff;
                int newgap_size = max_frag_could_fill - fraglen;

                // fragment must overlap  and should not touch the right side
                BUG_TRAP(max_frag_could_fill > 0);
                BUG_TRAP(fraglen < max_frag_could_fill);

                // reduce the size of the old gap
                gap_newsize = fragoff -  gap->offset;

                // create a new gap for the right side of the gap
                mgrp_frag_gap_init(newgap,
                                   fragoff + fraglen, newgap_size,
                                   fragment, gap->after);
        }

        // insert fragment in fragment list
        mgrp_frag_bucket_insert_fragment_before(bucket, gap->after, fragment);
        gap->after = fragment;
        gap->size = gap_newsize;

        DEBUG_REAS("modified gap o:%d l:%d", gap->offset, gap->size);
        DEBUG_REAS("added gap o:%d l:%d", newgap->offset, newgap->size);

        // add the new gap after the existing gap
        list_add(&newgap->glist, &gap->glist);
}

static inline void
mgrp_frag_bucket_fill_gaps_and_store_fragment(struct mgrp_frag_bucket *bucket,
                                              struct sk_buff *fragment)
{
        struct mgrp_frag_gap *gap = NULL;
        struct mgrp_fraghdr *fh = (struct mgrp_fraghdr *)fragment->transport_header;
        int fragoff = fh->fragoff;
        int fraglen = fragment->len;
        int islast = !fh->more_fragments;
        int overlap = MGRP_FRAG_OVERLAP_NONE;

        // find the first gap that can be filled (at least partially)
        // gaps are ordered in the list by increasing offset
        list_for_each_entry(gap, &bucket->gaps, glist) {
                DEBUG_REAS("gap o:%d l:%d", gap->offset, gap->size);
                overlap = mgrp_frag_gap_overlaps(gap, fragoff, fraglen, islast);
                if (overlap != MGRP_FRAG_OVERLAP_NONE)
                        break;
        }

        // no gap found for the any of the data in the fragment
        // data for this fragment has already been received
        if (overlap == MGRP_FRAG_OVERLAP_NONE) {
                // drop fragment, no longer needed
                DEBUG_REAS("no gap overlap fragid:%d fragoff:%d fraglen:%d",
                           fh->fragid, fragoff, fraglen);
                bucket->frags.dropped_because_overlapping++;
                kfree(fragment);
                return;
        }

        // since gaps are ordered in increasing offset,
        // that means that @gap is the first gap that
        // can be filled (at least partially) by @fragment

        if (overlap == MGRP_FRAG_OVERLAP_LEFTSIDE) {
                // gap may be completely filled and removed from gap list
                int removed_gap = 0;
                DEBUG_REAS("leftside overlap fragid:%d fragoff:%d fraglen:%d",
                           fh->fragid, fragoff, fraglen);
                mgrp_frag_bucket_adjust_gap_leftside(bucket,
                                                     gap, fragment,
                                                     &removed_gap);
                if (removed_gap) {
                        kfree(gap);
                        gap = NULL;
                }

        } else if (overlap == MGRP_FRAG_OVERLAP_RIGHTSIDE) {
                // gap becomes smaller but never completely filled
                DEBUG_REAS("rightside overlap fragid:%d fragoff:%d fraglen:%d",
                           fh->fragid, fragoff, fraglen);
                mgrp_frag_bucket_adjust_gap_rightside(bucket, gap, fragment);

        } else if (overlap == MGRP_FRAG_OVERLAP_MIDDLE) {
                // gap is divided into two (one new created and added to list)
                DEBUG_REAS("middle overlap fragid:%d fragoff:%d fraglen:%d",
                           fh->fragid, fragoff, fraglen);
                mgrp_frag_bucket_adjust_gap_middle(bucket, gap, fragment);

        } else {
                ERRFN("unknown overlap type %d", overlap);
                BUG();
        }
}

/******************************************************
 * mgrp_grag_bucket -- fragment storing and reassembly
 */

static inline void
mgrp_frag_bucket_add_fragment(struct mgrp_frag_bucket *bucket,
                              struct sk_buff *skb)
{
        struct mgrp_fraghdr *fh = (struct mgrp_fraghdr *)skb->transport_header;

        DEBUG_REAS("new frag %d fragid:%d fragoff:%d len:%d",
                   fh->fragmented, fh->fragid, fh->fragoff, skb->len);

        if (fh->fragmented) {
                // normal fragment
                mgrp_frag_bucket_fill_gaps_and_store_fragment(bucket, skb);

        } else if (bucket->got_unfragmented) {
                // we got the whole packet in one fragment
                // duplicate, drop skb, we do not need it
                bucket->pkts.dropped_because_overlapping++;
                kfree_skb(skb);

        } else {
                // we got the whole packet in one fragment
                bucket->unfragmented = skb;
                bucket->got_unfragmented = 1;
                // we can reassemble the packet immediately
                // no need to keep around fragments
                mgrp_frag_bucket_release_fragments_and_gaps(bucket);
                DEBUG_FRAGHEX(skb, fh->fragid, fh->fragoff, fh,
                              FRAGHDR_NOT_INCLUDED, ">>> unfragmented");
        }

}

static inline int
mgrp_frag_bucket_can_reassemble(struct mgrp_frag_bucket *bucket)
{
        if (bucket->got_unfragmented)
                return 1;
        return !bucket->released_fragments && list_empty(&bucket->gaps);
}

/*
 * Prepare the reassembled skb so that it can be passed to ops->rcv():
 *
 *      - copy MAC and IP headers from the last fragment
 *      - copy skb->dst, skb->dev, skb->pkt_type
 *
 * The problem: when we combine fragment skbs into one new skb,
 * we do not have a valid skb that came from the network stack.
 * So the skb is not attached to a device, does not contain
 * IP or MAC headers, etc. The problem is that TCP (among others)
 * uses the IP header (mainly for checksum verification) and other
 * pointers (skb->dev, skb->dst, skb->pkt_type, see tcp_v4_rcv()).
 *
 * We copy fields and headers from the last fragment. This is not
 * entirely correct (for example the IP header length is off) but
 * it should work (since we need only a subset of those fields).
 *
 * Used core/skbuff.c:skb_segment() as a guide.
 *
 * Ugly, I know, I really do. But I am just writing a thesis :)
 */
static inline void
mgrp_frag_emulate_fragment_skb(struct sk_buff *tskb, struct sk_buff *fragment)
{
        struct mgrp_fraghdr *fh = (struct mgrp_fraghdr *)fragment->transport_header;

        // We need to extract the IP header in the fragment skb.
        // 
        // The pointer fragment->network_header is still valid,
        // but since we have moved the fragment->transport_header
        // (to point to the fragmentation header inside the MGRP TUN)
        // we cannot use skb_network_header_len() to get the length.
        // 
        // However, we have stored a pointer to the original MGRP
        // header (just below IP) in the qskb, so we can use that.
        unsigned char *mgrphdr = (unsigned char *)bhtp_qsk(fragment)->recv.bh;
        int network_hdrlen = mgrphdr - fragment->network_header;
        int mac_hdrlen = fragment->mac_len;

        void *original_data = tskb->data;
        int original_len = tskb->len;

        // fix the skb fields
        tskb->dev = fragment->dev;
        skb_copy_queue_mapping(tskb, fragment);
        tskb->priority = fragment->priority;
        tskb->dst = dst_clone(fragment->dst);
        tskb->pkt_type = fragment->pkt_type;

        // do not use the protocol of the fragment skb because that would
        // be "MGRP"; instead replace for the "real" protocol that was
        // encapsulated inside MGRP (need to translate because MGRP and
        // IP use different protocol numbers due to size restrictions)

        tskb->protocol = mgrp_to_ip_protocol(fh->proto);

        // fix the skb headers (add headers backwards)

        // the transport header is included in the data we copied
        // from the fragments; just fix the pointer
        skb_reset_transport_header(tskb);

        DEBUG_REAS("added transport header ?? skb_len %d", tskb->len);

        // copy IP header

        // XXX the length field in the header is wrong! do we care?
        // XXX (contains the length of the MGRP packet of the last fragment)
        skb_push(tskb, network_hdrlen);
        skb_reset_network_header(tskb);
        memcpy(tskb->network_header, fragment->network_header, network_hdrlen);
        DEBUG_REAS("added network header %d skb_len %d",
                   network_hdrlen, tskb->len);

        // copy MAC header
        // XXX (is this needed? does it make sense?)
        skb_push(tskb, mac_hdrlen);
        skb_reset_mac_header(tskb);
        memcpy(tskb->mac_header, fragment->mac_header, mac_hdrlen);
        tskb->mac_len = mac_hdrlen;
        DEBUG_REAS("added mac header %d skb_len %d",
                   fragment->mac_len, tskb->len);

        // undo the changes to the skb->data and skb->len by adding the headers
        // remember, we just needed to place them inside the skb memory
        // and reset the header pointers, but skb_push() modified the skb
        // every time we added a new header so now we need to revert those
        // changes because the protocol code expects skb->data to point to
        // the transport header (ie, the TCP, UDP, headers)
        skb_pull(tskb, mac_hdrlen);
        skb_pull(tskb, network_hdrlen);

        // this leaves the skb exactly as it came in
        // those bug traps below would have saved me a day of debugging!
        BUG_TRAP(original_len == tskb->len);
        BUG_TRAP(original_data == tskb->data);
        BUG_TRAP(tskb->transport_header == tskb->data);
}

// Appends a portion of the @src skb to the @dst skb
static inline int
mgrp_skb_append_portion(struct sk_buff *dst, struct sk_buff *src,
                        int src_offset, int src_len)
{
        int rc = 0;
        if (skb_tailroom(dst) < src_len) {
                ERRFN("dst skb does not fit %d bytes; only %d left",
                      src_len, skb_tailroom(dst));
                return -1;
        }
        rc = skb_copy_bits(src, src_offset, skb_put(dst, src_len), src_len);
        if (rc) {
                ERRFN("failed to copy %d bits from offset %d", src_len, src_offset);
                return -1;
        }
        return 0;
}

static inline struct sk_buff *
mgrp_frag_bucket_combine_fragments(struct mgrp_frag_bucket *bucket)
{
        int found_last = 0;
        int expecting_offset = 0;
        struct sk_buff *fragment = NULL;
        struct sk_buff *tskb = alloc_skb(MGRP_FRAG_MAXSIZE, GFP_ATOMIC);

        if (tskb == NULL)
                goto error;

        // reserve header space so that we can copy later the MAC and IP
        // headers of the last fragment in the reassembled packet
        skb_reserve(tskb, MAX_HEADER);

        while ((fragment = skb_dequeue(&bucket->fragments)) != NULL) {
                struct mgrp_fraghdr *fh = (struct mgrp_fraghdr *)
                                          fragment->transport_header;

                int islast = !fh->more_fragments;
                int fragoff = fh->fragoff;
                int fraglen = fragment->len;
                int right_fragoff = fragoff + fraglen - 1;

                DEBUG_FRAGHEX(fragment, bucket->key.fragid, fragoff, fh,
                              FRAGHDR_NOT_INCLUDED, ">>> combine_frags");
                BUG_TRAP(fh->fragmented);
                
                // We allow for the reassembly skbs to overlap, but
                // they should be ordered so that there are no gaps
                if (fragoff > expecting_offset) {
                        ERRFN("gap in the reassembly skbs: expected offset %d got %d",
                              expecting_offset, fragoff);
                        goto error_requeue;
                }

                // append the new data from the fragment to the reassembly skb
                if (right_fragoff >= expecting_offset) {
                        int rc = 0;
                        int offset = expecting_offset - fragoff;
                        int len = fraglen - offset;

                        rc = mgrp_skb_append_portion(tskb, fragment, offset, len);
                        if (rc)
                                goto error_requeue;
                        expecting_offset += len;
                }

                if (islast) {
                        BUG_TRAP(expecting_offset == (fragoff + fraglen));
                        BUG_TRAP(expecting_offset == tskb->len);
                        found_last = 1;
                        DEBUG_FRAGHEX(tskb, bucket->key.fragid, 0, NULL,
                                      FRAGHDR_NOT_INCLUDED, ">>> proemulate");
                        mgrp_frag_emulate_fragment_skb(tskb, fragment);
                }

                // we have copied the fragment data, we no longer need it
                kfree_skb(fragment);
                if (islast)
                        break;
        }

        DEBUG_FRAGHEX(tskb, bucket->key.fragid, 0, NULL,
                      FRAGHDR_NOT_INCLUDED, ">>> reassembled");
        BUG_TRAP(found_last);

        // we have removed the fragment skbs
        // (some overlapping framents may still be there)
        //
        // we release the fragments so that we can clear the gaps,
        // otherwise the gap list will be in an invalid state with
        // pointers to skbs that have been freed
        mgrp_frag_bucket_release_fragments_and_gaps(bucket);
        return tskb;

error_requeue:
        // put the current fragment skb back
        skb_queue_head(&bucket->fragments, fragment);
        // free the new skb
        kfree(tskb);
error:
        return NULL;
}

#ifdef ABANDONED
static inline struct sk_buff *
__abandoned_mgrp_frag_bucket_combine_unfragmented(struct mgrp_frag_bucket *bucket)
{
        struct sk_buff *tskb = NULL;

        // we already have the whole packet; just clone it
        tskb = skb_clone(bucket->unfragmented, GFP_ATOMIC);
        DEBUG_FRAGHEX(tskb, bucket->key.fragid, 0, NULL,
                      FRAGHDR_NOT_INCLUDED,
                      ">>> combine_pkt");
        // XXX wrong!
        // XXX this overwrites the original skb headers, which means
        // it overwrites the MGRP header and wrecks havoc later!!
        mgrp_frag_emulate_fragment_skb(tskb, bucket->unfragmented);

        return tskb;
}
#endif

static inline struct sk_buff *
mgrp_frag_bucket_combine_unfragmented(struct mgrp_frag_bucket *bucket)
{
        // need to allocate and not clonne because we are going
        // to copy IP and MAC headers for tcp_bhtp_rcv()

        // XXX Tried to use skb_clone() and not do the _emulate_ fix,
        // XXX but kept getting problems, could not solve, abandoned..
        // XXX (look above abandoned code)

        int rc = 0;
        struct sk_buff *tskb = alloc_skb(MGRP_FRAG_MAXSIZE, GFP_ATOMIC);

        if (tskb == NULL)
                return NULL;

        // reserve header space so that we can copy later the MAC and IP
        // headers of the last fragment in the reassembled packet
        skb_reserve(tskb, MAX_HEADER);

        // copy the bits from the unfragmented (duplicate) skb in the bucket
        rc = mgrp_skb_append_portion(tskb, bucket->unfragmented,
                                     0, bucket->unfragmented->len);
        if (rc)
                goto error;

        DEBUG_FRAGHEX(tskb, bucket->key.fragid, 0, NULL,
                      FRAGHDR_NOT_INCLUDED,
                      ">>> combine_pkt");

        mgrp_frag_emulate_fragment_skb(tskb, bucket->unfragmented);
        return tskb;

error:
        kfree(tskb);
        return NULL;
}

static inline struct sk_buff *
mgrp_frag_bucket_reassemble(struct mgrp_frag_bucket *bucket, int *used_frags)
{
        struct bhtp_qskb *qskb = NULL;
        struct sk_buff *tskb = NULL;
        
        if (!mgrp_frag_bucket_can_reassemble(bucket))
                return NULL;

        // create a new skb to hold the reassembled packet
        if (bucket->got_unfragmented) {
                BUG_TRAP(bucket->unfragmented != NULL);
                // allocates a new skb and copies the unfragmented packet
                tskb = mgrp_frag_bucket_combine_unfragmented(bucket);
                *used_frags = 0;
        } else {
                // newly allocates skb + data with all fragments copied
                // skb has been made to "look like" it came from IP
                tskb = mgrp_frag_bucket_combine_fragments(bucket);
                *used_frags = 1;
        }

        if (tskb == NULL)
                return NULL;

        // fix the qskb (MGRP control buffer) to look like it came from
        // bhtp_tun_rcv() [only used by UDPex which should not have fragments]
        // (XXX needed?)
        //
	// overloading the control buffer in sk_buff for MGRP private data
	// only a few fields filled; TUN rcv code should do the rest
        qskb = bhtp_qsk_init(tskb);
        qskb->tun_len = tskb->len;
        qskb->tun_pad = 0;
        qskb->recv.bh = NULL;
        mgrp_tskb_fragid_set(tskb, bucket->key.fragid);

        return tskb;
}

static inline int
mgrp_frag_bucket_reassemble_and_send(struct mgrp_frag_bucket *bucket,
                                     struct mgrp_frag_hash *fhash)
{
        int rc = 0;
        struct sk_buff *tskb = NULL;
        struct bhtp_tun_prot *prot = NULL;
        int used_frags = 0;
        struct sockaddr_in remote_addr = {
                .sin_family = AF_INET,
                .sin_addr.s_addr = bucket->key.source_addr,
        };

        // find the protocol that expects the reassembled packet
        prot = bhtp_lookup_proto(bucket->key.proto);
        if (!BHTP_TUN_PROT_HAS_RCV(prot)) {
                ERRFN("no rcv() for proto %d for reassembled packet fragid %d",
                      bucket->key.proto, bucket->key.fragid);
                rc = -1;
                goto done;
        }

        // reassemble the packet in a new skb and fix the qskb like
        // bhtp_tun_rcv() so that we can pass directly to ops->rcv()
        tskb = mgrp_frag_bucket_reassemble(bucket, &used_frags);
        if (tskb == NULL) {
                ERRFN("cannot reassemble bucket proto %d fragid %d",
                      bucket->key.proto, bucket->key.fragid);
                rc = -1;
                goto done;
        }

        DEBUG_REAS("reassembled fragid:%d and sending up len %d",
                   bucket->key.fragid, tskb->len);

        // hand to the next protocol (see bhtp_tun_rcv)
        rc = prot->ops->rcv(tskb, &remote_addr);
        if (rc) {
                // on error we need to drop the @tskb that we just created
                ERRFN("dropped reassembled packet len %d: ops->rcv() failed fragid %d proto %d",
                      tskb->len, bucket->key.fragid, bucket->key.proto);
                kfree(tskb);
                tskb = NULL;
                rc = -1;
                goto done;
        }

        // counters and flags
        bucket->already_sent = 1;
        if (used_frags) {
                bucket->used_frags_for_sent = 1;
                fhash->count.reassembled_from_fragments++;
        } else {
                bucket->used_packet_for_sent = 1;
                fhash->count.reassembled_from_packet++;
        }

        // fragments no longer needed
        mgrp_frag_bucket_release_reassembly_memory(bucket);
done:
        return rc;
}

/******************************************************
 * mgrp_frag_overflow_bucket
 *
 * Overflow bucket in the overflow list attached to a fixed bucket
 * Allocated individually on demand
 * Contains one fixed reassembly bucket
 * Part of a list of overflow reassembly buckets
 */

static inline void
mgrp_frag_overflow_bucket_init(struct mgrp_frag_overflow_bucket *obucket,
                               struct mgrp_frag_key *key)
{
        memset(obucket, 0, sizeof(*obucket));
        mgrp_frag_bucket_init(&obucket->bucket, key);
        INIT_LIST_HEAD(&obucket->olist);
}

static inline void
mgrp_frag_overflow_bucket_done(struct mgrp_frag_overflow_bucket *obucket)
{
        mgrp_frag_bucket_done(&obucket->bucket);
}

/******************************************************
 * mgrp_frag_fixed_bucket
 *
 * Fixed bucket in the hash array.
 * Allocated with all the buckets in the hash array.
 * Contains one fixed reassembly bucket
 * Contains a list of overflow reassembly buckets
 */

static inline void
mgrp_frag_fixed_bucket_init(struct mgrp_frag_fixed_bucket *fbucket)
{
        memset(fbucket, 0, sizeof(*fbucket));
        INIT_LIST_HEAD(&fbucket->overflow_buckets);
        fbucket->initialized = 1;
        // the internal bucket is not initialized yet, only after we occupy it
}

static inline void
mgrp_frag_fixed_bucket_done(struct mgrp_frag_fixed_bucket *fbucket)
{
        struct mgrp_frag_overflow_bucket *obucket = NULL;
        struct mgrp_frag_overflow_bucket *obucket_tmp = NULL;

        if (!fbucket->initialized)
                return;
        
        if (fbucket->occupied)
                mgrp_frag_bucket_done(&fbucket->bucket);

        list_for_each_entry_safe(obucket, obucket_tmp,
                                 &fbucket->overflow_buckets,
                                 olist) {
                list_del(&obucket->olist);
                mgrp_frag_overflow_bucket_done(obucket);
                kfree(obucket);
        }
        memset(fbucket, 0, sizeof(*fbucket));
}

/******************************************************
 * mgrp_frag_hash
 *
 * The reassembly hash.
 *
 * Each hash entry contains a fixed reassemby bucket (struct bucket)
 * and alist of overflow buckets.
 */

static inline int
mgrp_frah_hash_actual_buckets(int max_buckets, int *bits)
{
        int actual_buckets = 0;
        int actual_bits = 1;

        // some sanity check, this is a packet hash, it cannot be too small
        if (max_buckets < MGRP_FRAG_HASH_BUCKETS_MIN) {
                LOGFN("num_of_buckets %d is too low; increasing to %d",
                      max_buckets, MGRP_FRAG_HASH_BUCKETS_MIN);
                max_buckets = MGRP_FRAG_HASH_BUCKETS_MIN;
        }

        // we need the number of buckets to be power of two so that we can
        // just use a simple mask to get the index from the hash value

        // upgrade to the next power of two
        // makes hash indexing simpler
        while (((max_buckets-1)>>actual_bits) > 0)
                actual_bits++;
        actual_buckets = (1<<actual_bits);

        if (max_buckets != actual_buckets)
                LOGFN("asked for %d buckets but bumping up to %d buckets (power of two)",
                      max_buckets, actual_buckets);

        *bits = actual_bits;
        return actual_buckets;
}

static inline int
mgrp_frag_hash_init(struct mgrp_frag_hash *fhash, unsigned int max_buckets)
{
        int rc = 0;
        int alloc_bytes = 0;
        int bits = 0;
        int buckets = mgrp_frah_hash_actual_buckets(max_buckets, &bits);

        memset(fhash, 0, sizeof(struct mgrp_frag_hash));

        // allocate the hash array (hash key indexes into this array)
        alloc_bytes = buckets * sizeof(struct mgrp_frag_fixed_bucket);
        fhash->fbuckets = vmalloc(alloc_bytes);
        if (fhash->fbuckets == NULL) {
                ERRFN("failed to allocate %d bytes for %d buckets",
                      alloc_bytes, buckets);
                rc = -ENOMEM;
                goto error;
        }
        BUG_TRAP(alloc_bytes == buckets * sizeof(fhash->fbuckets[0]));
        // this memset is necessary to mark all fixed bucket "not initialized"
        memset(fhash->fbuckets, 0, alloc_bytes);
        fhash->fbuckets_alloc_bytes = alloc_bytes;

        // do lazy initialization for the fixed buckets (ie, do nothing now)
        // the memset() above takes care to zero all the 'initialized' flags
        // so that we can initialize buckets as we touch them

        fhash->index_mask64 = (1LL<<bits) - 1;
        fhash->index_bits = bits;
        fhash->num_of_buckets = buckets;

        // used at the single hash entry point (mgrp_frag_process_fragment())
	spin_lock_init(&fhash->lock);

        LOGFN("allocated frag hash %d buckets %d bytes %d index_bits",
              fhash->num_of_buckets, alloc_bytes, fhash->index_bits);

        return 0;
error:
        return rc;
}

// Clears all fixed hash entries and all overflow lists
// Hash remains allocated but empty, just like after an _init()
static inline void
__mgrp_frag_hash_reset(struct mgrp_frag_hash *fhash)
{
        int i = 0;
        for (i = 0; i < fhash->num_of_buckets; i++)
                mgrp_frag_fixed_bucket_done(fhash->fbuckets + i);
        if (fhash->fbuckets_alloc_bytes > 0)
                memset(fhash->fbuckets, 0, fhash->fbuckets_alloc_bytes);
        memset(&fhash->count, 0, sizeof(fhash->count));
}

void
mgrp_frag_hash_reset(struct mgrp_frag_hash *fhash)
{
	spin_lock_bh(&fhash->lock);
        __mgrp_frag_hash_reset(fhash);
	spin_unlock_bh(&fhash->lock);
}

static inline void
mgrp_frag_hash_done(struct mgrp_frag_hash *fhash)
{
        __mgrp_frag_hash_reset(fhash);
        if (fhash->fbuckets_alloc_bytes > 0)
                vfree(fhash->fbuckets);
        memset(fhash, 0, sizeof(*fhash));
}

static inline unsigned int
mgrp_frag_hash_index(struct mgrp_frag_hash *fhash,
                     struct mgrp_frag_key *key)
{
        u64 fragid64 = key->fragid;
        unsigned int index = (fragid64 & fhash->index_mask64);
#ifdef LOG_FRAGHASH
        // use global debug_fraghash so that we can store it in the bucket
        // when we first create it in mgrp_frag_hash_reassembly_bucket()
        frag.hs.unhashed = fragid64;
        frag.hs.hashed = fhash->index_mask64;
        frag.hs.index = index;
#endif
        return index;
}

/*
 * Returns the fixed bucket in the hash array where the key hashes.
 *
 * This may not contain the reassembly bucket that corresponds to @key.
 * The reassembly bucket can be internal to the fixed bucket or in one
 * of the overflow buckets. The caller has to do that work.
 *
 * Always returns a non-NULL and initialized fixed bucket.
 */
static inline struct mgrp_frag_fixed_bucket *
mgrp_frag_hash_fixed_bucket(struct mgrp_frag_hash *fhash,
                            struct mgrp_frag_key *key)
{
        int index = mgrp_frag_hash_index(fhash, key);
        struct mgrp_frag_fixed_bucket *fbucket = &fhash->fbuckets[index];

        BUG_TRAP(index < fhash->num_of_buckets);

        if (!fbucket->initialized)
                mgrp_frag_fixed_bucket_init(fbucket);
        return fbucket;
}

static inline struct mgrp_frag_bucket *
mgrp_frag_hash_find_existing_reassembly_bucket(
                        struct mgrp_frag_hash *fhash,
                        struct mgrp_frag_key *key,
                        struct mgrp_frag_fixed_bucket **fbucket_p)
{
        struct mgrp_frag_fixed_bucket *fbucket = NULL;
        struct mgrp_frag_overflow_bucket *obucket = NULL;

        // jump to the hash array bucket based on the key
        fbucket = mgrp_frag_hash_fixed_bucket(fhash, key);
        if (fbucket_p != NULL)
                *fbucket_p = fbucket;
        
        // look in the internal bucket
        if (fbucket->occupied &&
            mgrp_frag_bucket_key_matches(&fbucket->bucket, key))
                return &fbucket->bucket;

        // look in the overflow list
        list_for_each_entry(obucket, &fbucket->overflow_buckets, olist) {
                if (mgrp_frag_bucket_key_matches(&obucket->bucket, key))
                        return &obucket->bucket;
        }

        return NULL;
}

#define MGRP_FRAG_INSERT_BUCKET_IFNOTTHERE      1
#define MGRP_FRAG_EXISTING_BUCKET_ONLY          0

static inline struct mgrp_frag_bucket *
mgrp_frag_hash_reassembly_bucket(struct mgrp_frag_hash *fhash,
                                 struct mgrp_frag_key *key,
                                 int insert_if_not_there,
                                 int *inserted_new)
{
        struct mgrp_frag_bucket *bucket = NULL;
        struct mgrp_frag_fixed_bucket *fbucket = NULL;

        if (inserted_new != NULL)
                *inserted_new = 0;

        bucket = mgrp_frag_hash_find_existing_reassembly_bucket(
                                fhash, key, &fbucket);
        if (bucket != NULL)
                return bucket;
        if (bucket == NULL && !insert_if_not_there)
                return NULL;

        // no bucket found, insert a new one (we already know the fixed bucket)
        if (fbucket->occupied) {
                // add a new bucket to the overflow list
                struct mgrp_frag_overflow_bucket *obucket = NULL;
                obucket = kmalloc(sizeof(*obucket), GFP_ATOMIC);
                if (obucket == NULL)
                        return NULL;
                mgrp_frag_overflow_bucket_init(obucket, key);
                list_add(&obucket->olist, &fbucket->overflow_buckets);
                bucket = &obucket->bucket;
                fhash->count.overflow_buckets_used++;
        } else {
                // use the internal bucket
                mgrp_frag_bucket_init(&fbucket->bucket, key);
                fbucket->occupied = 1;
                bucket = &fbucket->bucket;
                fhash->count.fixed_buckets_used++;
        }
        fhash->count.needed_reassembly++;
#ifdef LOG_FRAGHASH
        // mgrp_frag_hash_index has stored debug_fraghash in global
        bucket->hs = frag.hs;
#endif

        if (inserted_new != NULL)
                *inserted_new = 1;
        return bucket;
}

/******************************************************
 * Processes a new fragment.
 *
 * Find its reassembly bucket in the reassembly hash. If all fragments
 * have been received, create a new skb to hold all fragments and pass
 * to next protocol. Otherwise, queue in reassembly buffer.
 *
 * @key                         fragment key (addr,fragid)
 *
 * @skb                         holds a fragment
 * @skb->transport_header       fragmentation header (included in the skb)
 * @skb->data                   data fragment (fragmentation header removed)
 * @skb->len                    length of the data fragment
 *
 * bhtp_qsk(skb), qskb          invalid, do not use
 *
 * On success return 0 and take ownership of the skb (drop, queue or pass on)
 * On error return -1. Caller will free the skb.
 */
static inline int
mgrp_frag_process_fragment(struct sk_buff *skb, struct mgrp_frag_key *key)
{
        // use the global fragmentation hash
        struct mgrp_frag_hash *fhash = &frag.fhash;

        int rc = 0;
        int inserted_new_bucket = 0;
        struct mgrp_frag_bucket *bucket = NULL;
        struct mgrp_fraghdr *fh = (struct mgrp_fraghdr *)skb->transport_header;

        DEBUG_REAS(">>>..................................................");

        spin_lock_bh(&fhash->lock);

        // hash lookup (will create new bucket if not there)
        bucket = mgrp_frag_hash_reassembly_bucket(
                                fhash, key,
                                MGRP_FRAG_INSERT_BUCKET_IFNOTTHERE,
                                &inserted_new_bucket);
        if (bucket == NULL) {
                ERRFN("no reassembly bucket found for key src(%d) fragid(%d)",
                      key->source_addr, key->fragid);
                // rc is non-zero, caller will drop skb
                rc = -1;
                goto done;
        }

        if (fh->fragmented)
                bucket->frags.received++;
        else
                bucket->pkts.received++;

        if (bucket->already_sent) {
                // this is a duplicate
                if (fh->fragmented)
                        bucket->frags.dropped_because_already_sent++;
                else
                        bucket->pkts.dropped_because_already_sent++;
                // drop fragment, no longer needed
                kfree(skb);
                goto done;
        }

        // In case we have released all fragments (perhaps because we
        // got the whole unfragmented packet) but not already sent,
        // then we try to add the fragment we will drop it as a duplicate
        // because we wont find any overlapping gap (the gap list is empty).
        // So: no need to treat that case specially
                
        // add fragment skb in sorted skb queue and update gap list
        mgrp_frag_bucket_add_fragment(bucket, skb);

        // at this point @skb has been consumed (dropped or queued)
        // we should not return non-zero code to caller after this point
        // (because then the caller would free the skb)

        // reassemble and send up if all fragments are here
        if (mgrp_frag_bucket_can_reassemble(bucket)) {
                rc = mgrp_frag_bucket_reassemble_and_send(bucket, fhash);
                if (rc) {
                        ERRFN("reassembly or protocol send failed proto %d fragid %d",
                              key->proto, key->fragid);
                        // rc is zero, no skb drops
                        rc = 0;
                        goto done;
                }
        }
done:
        spin_unlock_bh(&fhash->lock);
        // caller will drop if non-zero
        return rc;
}

/******************************************************
 * mgrp_frag_tun_rcv()
 *
 * Receive callback for the fragmentation pseudo-protocol.
 *
 * Only entry point to reassembly hash.
 */

int
mgrp_frag_tun_rcv(struct sk_buff *skb, struct sockaddr_in *remote_addr)
{
        int rc = 0;
        struct mgrp_frag_key key;
	struct bhtp_qskb *qskb = bhtp_qsk(skb);
        struct mgrp_fraghdr *fh = NULL;

        // expecting fragmentation header to be on top
	if (!pskb_may_pull(skb, sizeof(struct mgrp_fraghdr))) {
                ERRFN("dropped fragment TUN: cannot extract fragmentation header from skb len %d",
                      skb->len);
                goto drop;
        }

        // reposition the skb to contain only the fragment data
        skb->transport_header = skb->data;
        fh = (struct mgrp_fraghdr *)skb->transport_header;
        skb_pull(skb, sizeof(struct mgrp_fraghdr));

        // fragments must have at least one byte
        if (skb->len == 0) {
                ERRFN("TUN fragment has no payload");
                goto drop_verbose;
        }

        /* The control buffer in sk_buff is overloaded for MGRP private data.
         *
         * The qskb is partially filled in bhtp_tun_rcv():
         *      qskb->tun_len
         *      qskb->tun_pad
         *      qskb->recv.bh
         * Perform sanity checks.
         */
        BUG_TRAP(qskb->tun_len == sizeof(struct mgrp_fraghdr) + skb->len);
        if (qskb->tun_pad != 0) {
                ERRFN("TUN fragment has non zero padding %d", qskb->tun_pad);
                goto drop_verbose;
        }

        /*
         * UPDATE: the below is no longer true. We need qskb->recv.bh
         *         in mgrp_frag_emulate_fragment_skb() in order to
         *         calculate the length of the network header (since we
         *         have moved skb->transport_header to point to the
         *         fragmentation header). Keeping just for discussion.
         *
         * Invalidate the qskb because we are not going to use it for
         * the fragmentation. The fragmentation layer is a pseudo-protocol
         * that will eventually call a tun_rcv() for another "real" protocol.
         *
         * At that point we will create a new skb to hold all the fragments
         * and we are going to create a valid qskb, to look as if it was
         * coming from bhtp_tun_rcv().
         *
         */

        // do not invalidate the skb
        // mgrp_qsk_invalidate(skb);

        DEBUG_FRAGHEX(skb, fh->fragid, fh->fragoff, fh,
                      FRAGHDR_NOT_INCLUDED, ">>> fragment");

        // process fragment
        // callee will either queue or will reassemble and deliver packet
        mgrp_frag_key_init(&key, fh, remote_addr);
        rc = mgrp_frag_process_fragment(skb, &key);
        if (rc)
                goto drop_verbose;

        return 0;

drop_verbose:
        ERRFN("dropped fragment TUN skb len %d fragid %d fragoff %d",
              skb->len, fh->fragid, fh->fragoff);
        log_hex(fh, sizeof(struct mgrp_fraghdr), 4, "fraghdr: ");
drop:
        kfree(skb);
        // we always handle the skb even if there is an error
        // so errors never propagate to caller
        return 0;
}

int
__dummy_mgrp_frag_tun_rcv(struct sk_buff *skb) {
        static long long __dummy_counter = 0;

        __dummy_counter++;

        // placeholder code, just dropping for now

        /*
         * Since our variable is 64bit, the modulo (division) on 32bit
         * machines causes the module not to load:
         *
         * $ insmod mgrp
         * insmod: error inserting 'mgrp.ko': -1 Unknown symbol in module
         * mgrp: Unknown symbol __moddi3
         *
         * The problem is that gcc translates this into a libgcc call
         * that is not defined in the kernel because it is expensive.
         *
         * http://www.mail-archive.com/linux-kernel@vger.kernel.org/msg185853.html
         *
         * Since we do not care about the exact number we use for the modulo,
         * we use 1024, a power of two, whose module becomes a simple mask.
         */
        if (__dummy_counter % 1024 == 0) {
                ERRFN("placeholder code: TUN fragment %lld dropped len %d",
                      __dummy_counter, skb->len);
        }
        kfree_skb(skb);
        return 0;
}


/******************************************************
 ******************************************************
 * FRAGMENTATION -- sender side
 *
 * Fragmentation of chunks for the _FRAGMENT_ policies
 * Protocol agnostic (does not need to be TCP like _STANDALONE_ policies.
 */

static inline void
mgrp_fragment_chunk_fix_qskb(struct sk_buff *chunk, int chunk_len, u32 fragid)
{
        // not valid yet, needs to be initialized (do not call bhtp_qsk yet)
        struct bhtp_qskb *qskb = NULL;

        // required because it sets the MAGIC number for sanity checks
        bhtp_qsk_init(chunk);

        // ! change of protocol
        // ! original protocol should be in fragmenation header
        // ! new protocol is MGRP_PROTO_FRAG
        qskb = bhtp_qsk(chunk);
        qskb->tun_len = chunk_len;
        qskb->tun_hdr = sizeof(struct mgrp_fraghdr);
        qskb->tun_data = chunk_len - sizeof(struct mgrp_fraghdr);
        qskb->tun = &frag.tun;

        // XXX this may not be necessary for the chunks
        // XXX where do we use fragid and the flag in the qskb?
        mgrp_tskb_fragid_set(chunk, fragid);
        mgrp_tskb_type_set(chunk, MGRP_QSKB_FRAGMENT_CHUNK);

        DEBUG_FRAG("len %d hdr %d data %d fragid %d",
                   qskb->tun_len,
                   qskb->tun_hdr,
                   qskb->tun_data,
                   qskb->fragid);
}

struct sk_buff *
mgrp_fragment_chunk(struct sk_buff *skb, int chunk_len, int copy_data)
{
        // @copy_data is ignored since we always copy the data

        // *data* in this context is the whole packet we receive from
        // the transport protocol, for example including TCP headers
        // and TCP payload. The *header* is the fragmentation header.

        int rc = 0;
        struct bhtp_qskb *qskb = bhtp_qsk(skb);

        int skb_proto = qskb->tun->prot->proto;
        int skb_dataleft = qskb->tun_len - qskb->tun_chunk_frag;
        int skb_dataoffset = qskb->tun_chunk_frag;

        int chunk_hdrlen = sizeof(struct mgrp_fraghdr);
        int chunk_datalen = chunk_len - chunk_hdrlen;
        // no need to allocate more, because MGRP will copy the TUN
        int chunk_alloc = chunk_len;
        int chunk_islast = (skb_dataleft == chunk_datalen);

        struct sk_buff *chunk = NULL;
        struct mgrp_fraghdr *chunkhdr = NULL;
        void *chunkdat = NULL;
        u32 fragid = 0;

        DEBUG_FRAG(">>>");
        DEBUG_FRAG("chunk_len %d dataleft %d dataoffset %d first %d last %d",
                   chunk_len,
                   skb_dataleft,
                   skb_dataoffset,
                   chunk_isfirst,
                   chunk_islast);

        if (skb_dataleft <= 0) {
                ERRFN("no data left in chunk");
                goto error;
        }
        if (chunk_datalen == 0) {
                ERRFN("no room for any data in chunk");
                goto error;
        }
        if (chunk_datalen < 0) {
                ERRFN("chunk header %d is larger than chunk length %d",
                      chunk_hdrlen, chunk_len);
                goto error;
        }
        if (chunk_datalen > skb_dataleft) {
                ERRFN("not enough data in skb %d to fill chunk %d",
                      skb_dataleft, chunk_datalen);
                goto error;
        }

        // The original skb must contain in its qskb a fragmentation ID
        // Will be used to fill the fragmenation header of the chunk
        rc = mgrp_tskb_fragid_get(skb, &fragid);
        if (rc) {
                ERRFN("original skb does not have a FRAGID");
                goto error;
        }

        // new skb will hold the new fragment
        chunk = alloc_skb(chunk_alloc, GFP_ATOMIC);
        if (chunk == NULL) {
                ERRFN("no skb buffers for %d bytes", chunk_alloc);
                goto error;
        }

        // reset the MGRP private data of the new skb
        // major change: the TUN changes protocol (becomes FRAG)
        // the chunk skb does not have the full length yet
        mgrp_fragment_chunk_fix_qskb(chunk, chunk_len, fragid);
        
        // adding fragmentation header for reassembly
        skb_reserve(chunk, chunk_hdrlen);
        chunk->transport_header = skb_push(chunk, chunk_hdrlen);
        chunkhdr = (struct mgrp_fraghdr *)chunk->transport_header;
        DEBUG_FRAG("added frag header, chunk-skb %d", chunk->len);

        // fill fragmentation header (used by receiver for reassembly)
        mgrp_fraghdr_init(chunkhdr);
        chunkhdr->proto = skb_proto;
        // duplicate flag may be changed by caller based on redundant policy
        chunkhdr->duplicate = 0;
        // in the case that the skb fits fully into one fragment,
        // still set fragmented flag to distinguish from sending
        // a duplicate skb with a fragmented header (matters to
        // the reassembly hash on the receiver).
        chunkhdr->fragmented = 1;
        chunkhdr->more_fragments = !chunk_islast;
        chunkhdr->fragoff = skb_dataoffset;
        chunkhdr->fragid = fragid;
        DEBUG_FRAGHDR(chunkhdr);

        // copy the payload chunk
        chunkdat = skb_put(chunk, chunk_datalen);
        rc = skb_copy_bits(skb, skb_dataoffset, chunkdat, chunk_datalen);
        if (rc) {
         ERRFN("skb_copy_bits failed from original skb: rc %d", rc);
                goto error_freechunk;
        }
        DEBUG_FRAG("added frag payload, chunk-skb %d", chunk->len);

        // update the portion of the original skb that we have removed
        // !! critical step, modifies state so that next mgrp_fragment_chunk()
        // !! will get the next part of the original skb
        qskb->tun_chunk_frag += chunk_datalen;

        // sanity check
        if (qskb->tun_chunk_frag > qskb->tun_len) {
                ERRFN("BUG? tun_chunk_frag %d exceeded tun_len %d",
                      qskb->tun_chunk_frag, qskb->tun_len);
                goto error_freechunk;
        }

        DEBUG_FRAGHEX(chunk, fragid, chunkhdr->fragoff, chunkhdr,
                      FRAGHDR_INCLUDED, ">>> fragment");
        if (chunk_islast)
                DEBUG_FRAGHEX(skb, fragid, 0, NULL,
                              FRAGHDR_NOT_INCLUDED, ">>> original");

        // no need for special flags to signal an exhausted skb
        // as in the case of tcp_tun_fragment(). In the case of
        // TCP we needed it because we distinguished between
        // TCP header and data and we could not tell if a packet
        // was exhausted or just contained the header.

        return chunk;

error_freechunk:
        kfree_skb(chunk);
error:
        return NULL;
}

/*
 * Encapsulate a TUN of protocol protoA inside a fragmentation TUN
 * that has protocol MGRP_PROTO_FRAG with @fragid.
 *
 * Expects a tskb as sent to bhtp_transmit_lone_skb()
 */
int
mgrp_fragment_chunk_encapsulate(struct sk_buff *tskb, int fragid)
{
        struct mgrp_fraghdr *fh = NULL;
        struct bhtp_qskb *qskb = bhtp_qsk(tskb);
        int tskb_proto = qskb->tun->prot->proto;
        int tskb_type = qskb->type;
        int fraglen = sizeof(struct mgrp_fraghdr) + qskb->tun_len;

        DEBUG_FRAG("fragid %d len %d", fragid, tskb->len);

        if (skb_headroom(tskb) < sizeof(struct mgrp_fraghdr)) {
                ERRFN("need %d for fraghdr, got only %d, fragid %d len %d",
                      sizeof(struct mgrp_fraghdr),
                      skb_headroom(tskb),
                      fragid,
                      tskb->len);
                return -1;
        }

        tskb->transport_header = skb_push(tskb, sizeof(struct mgrp_fraghdr));
        fh = (struct mgrp_fraghdr *)tskb->transport_header;

        // fill fragmentation header (used by receiver for reassembly)
        mgrp_fraghdr_init(fh);
        fh->proto = tskb_proto;
        fh->duplicate = 0;
        fh->fragmented = 0;
        fh->more_fragments = 0;
        fh->fragoff = 0;
        fh->fragid = fragid;
        DEBUG_FRAGHDR(fh);

        // reset the MGRP private data of the new skb
        // major change: the TUN changes protocol (becomes FRAG)
        mgrp_fragment_chunk_fix_qskb(tskb, fraglen, fragid);
        qskb = bhtp_qsk(tskb);
        // propagate the duplicate flag
        if (tskb_type & MGRP_QSKB_DUPLICATE)
                qskb->type |= MGRP_QSKB_DUPLICATE;

        DEBUG_FRAGHEX(tskb, fragid, 0, fh, FRAGHDR_INCLUDED, ">>> encapsulate");

        return 0;
}

/*******************************************************
 *******************************************************
 * Global state for fragmenation pseudo-protocol
 */

struct mgrp_frag_state frag = {
        .tun_prot = {
                .proto          = MGRP_PROTO_FRAG,
                .ops            = &frag.tun_ops,
        },

        .tun_ops = {
                .rcv            = mgrp_frag_tun_rcv,
        },

        .tun = {
                .prot           = &frag.tun_prot,
        },
        // hash is initialized in mgrp_frag_init()
};

int __init
mgrp_frag_init(int max_frag_buckets)
{
        int rc = 0;

        LOG("loading");
        
        // Do not zero the global state since we have initialized it statically.
        // Initialize the rest here.
        rc = mgrp_frag_hash_init(&frag.fhash, max_frag_buckets);
        if (rc)
                goto error;

        // register with MGRP
        rc = bhtp_register_tun_prot(&frag.tun_prot);
        if (rc)
                goto error_freehash;
        LOG("registered with MGRP");

        LOG("struct mgrp_fraghdr %d bytes", sizeof(struct mgrp_fraghdr));
        LOG("loaded");
        return 0;

error_freehash:
        mgrp_frag_hash_done(&frag.fhash);
error:
        ERR("loading aborted");
        return rc;
}

void
mgrp_frag_exit(void)
{
	LOG("unloading");

	bhtp_unregister_tun_prot(&frag.tun_prot);
	LOG("de-registered from BHTP");

        mgrp_frag_hash_done(&frag.fhash);
	LOG("removed reassembly hash");

	LOG("unloaded");
}

///////////////////////////////////////////////
// seq_file interface for /proc/net/mgrp_frag_buckets

/*
 * The kernel throttles how many times _next is called in a single
 * /proc operation even if the elements are not exhauted.
 *
 * That means that doing:
 *
 *      cat /proc/net/<entry>
 *
 * results in the following calls:
 *
 *      _start, _next, _next, ... [until NULL or threshold], _stop
 *      _start, _next, _next, ... [until NULL or threshold], _stop
 *      ......
 *      _start, _next, _next, ... [until NULL or threshold], _stop
 *
 * until all the elements are exhausted (ie, NULL is returned).
 */
static void *mgrp_frag_buckets_seq_start(struct seq_file *seq, loff_t *pos)
{
        // use the global fragmentation hash
        struct mgrp_frag_hash *fhash = &frag.fhash;
        char str[300];

        // header
        // (only appears if there is at least one element)
        // (only appears the first time _start is invoked in a batch)
        if (*pos == 0) {
                mgrp_frag_bucket_oneliner_header(str, sizeof(str));
                seq_printf(seq, "#index   ovflow %s\n", str);
        }

        // LOCK
        spin_lock_bh(&fhash->lock);
        if (*pos >= fhash->num_of_buckets)
                return NULL;
        return &fhash->fbuckets[*pos];
}

static void *mgrp_frag_buckets_seq_next(struct seq_file *seq,
                                        void *v, loff_t *pos)
{
        // use the global fragmentation hash
        struct mgrp_frag_hash *fhash = &frag.fhash;
        (*pos)++;
        if (*pos >= fhash->num_of_buckets)
                return NULL;
        return &fhash->fbuckets[*pos];
}

static void mgrp_frag_buckets_seq_stop(struct seq_file *seq, void *v)
{
        // use the global fragmentation hash
        struct mgrp_frag_hash *fhash = &frag.fhash;
        // UNLOCK
	spin_unlock_bh(&fhash->lock);
}

static int mgrp_frag_buckets_seq_show(struct seq_file *seq, void *v)
{
        // use the global fragmentation hash
        struct mgrp_frag_hash *fhash = &frag.fhash;

        struct mgrp_frag_fixed_bucket *fbucket = v;
        struct mgrp_frag_overflow_bucket *obucket = NULL;
        int index = mgrp_frag_hash_fixed_bucket_index(fhash, fbucket);
        
        int overflowed = 0;
        char str[300];

        if (!fbucket->initialized)
                // no lines for unused buckets
                return 0;

        if (!fbucket->occupied && list_empty(&fbucket->overflow_buckets))
                // no lines for empty buckets
                return 0;

        // fixed bucket
        if (fbucket->occupied)
                mgrp_frag_bucket_oneliner(str, sizeof(str), &fbucket->bucket);
        else
                snprintf(str, sizeof(str), "unoccupied");
        // the constants depend on the header in _seq_start
        seq_printf(seq, "%-8d %-6d %s\n", index, 0, str);

        // overflow buckets
        list_for_each_entry(obucket, &fbucket->overflow_buckets, olist) {
                overflowed++;
                mgrp_frag_bucket_oneliner(str, sizeof(str), &obucket->bucket);
                // the constants depend on the header in _seq_start
                seq_printf(seq, "%-8d %-6d %s\n", index, overflowed, str);
        }

        return 0;
}

static const struct seq_operations mgrp_frag_buckets_seqops = {
        .start = mgrp_frag_buckets_seq_start,
        .next = mgrp_frag_buckets_seq_next,
        .stop = mgrp_frag_buckets_seq_stop,
        .show = mgrp_frag_buckets_seq_show,
};

static int mgrp_frag_buckets_open(struct inode *inode, struct file *file)
{
        return seq_open(file, &mgrp_frag_buckets_seqops);
}

const struct file_operations mgrp_frag_buckets_fops = {
        .owner = THIS_MODULE,
        .open = mgrp_frag_buckets_open,
        .read = seq_read,
        .llseek = seq_lseek,
        .release = seq_release,
};

///////////////////////////////////////////////
// /proc/mgrp_frag_hash_stats
int 
mgrp_frag_hash_counters_show(char *buffer, char **start, off_t offset,
                             int length, int *eof, void *data)
{
        // use the global fragmentation hash
        struct mgrp_frag_hash *fhash = &frag.fhash;
        int written = 0;

        if (offset != 0)
                return 0;

	spin_lock_bh(&fhash->lock);
        written = snprintf(buffer, length,
                           "#BUCKETS "
                           "total    "
                           "fixed    "
                           "overflow "
                           "    "
                           "REASSEMBLE "
                           "total    "
                           "okfrag   "
                           "okpkt\n"

                           " BUCKETS "
                           "%-8d "
                           "%-8lld "
                           "%-8lld "
                           "    "
                           "REASSEMBLE "
                           "%-8lld "
                           "%-8lld "
                           "%-8lld\n",

                           fhash->num_of_buckets,
                           fhash->count.fixed_buckets_used,
                           fhash->count.overflow_buckets_used,
                           fhash->count.needed_reassembly,
                           fhash->count.reassembled_from_fragments,
                           fhash->count.reassembled_from_packet);
	spin_unlock_bh(&fhash->lock);
        return written;
}

int 
mgrp_frag_reset(char *buffer, char **start, off_t offset,
                int length, int *eof, void *data)
{
        // use the global fragmentation hash
        struct mgrp_frag_hash *fhash = &frag.fhash;
        int written = 0;

        if (offset != 0)
                return 0;

        mgrp_frag_hash_reset(fhash);
        mgrp_reset_fragid_in_all_flows();

        written = snprintf(buffer, length, "fragmentation send/receive reset");

        return written;
}


#ifdef LOG_FRAGHEX

// number of skbs queued by the debug function before spitting the hex out
#define MGRP_FRAG_DEBUG_FRAGHEX_MAX     40

struct debug_fraghex_state {
        int fragid;
        int offset;
        struct mgrp_fraghdr *fh;
        char prefix[30];
};

static inline void
__debug_fraghex_now(struct sk_buff *skb,
                    int fragid, int offset, char *prefix,
                    struct mgrp_fraghdr *fh)
{
        LOGFN("%s %d offset %d len %d", prefix, fragid, offset, skb->len);
        if (fh != NULL)
                __debug_fragment_hdr(fh, prefix);
        log_hex(skb->data, skb->len, 20, "    ");
}

/*
 * Three modes: (uses static data to keep state across invocations)
 *
 * 1. queues skb clones up until max (overrides cb to store local state)
 * 2. when max is reached it prints hexdumps of the skb data
 * 3. a noop over max
 */
void
__debug_fraghex(struct sk_buff *skb, int fragid, int offset,
                struct mgrp_fraghdr *fh,
                int includes_fraghdr,
                char *prefix)
{
        // !!! these variables are static !!!
        static int max = MGRP_FRAG_DEBUG_FRAGHEX_MAX;
        static int debug_count = 0;
        static struct sk_buff_head todebug;

        // init
        if (debug_count == 0) {
                skb_queue_head_init(&todebug);
        }

        debug_count++;

        // stoppped debugging
        if (debug_count > max)
                return;

        // storing to debug later
        if (debug_count <= max) {
                struct sk_buff *fragskb = skb_clone(skb, GFP_ATOMIC);
                struct debug_fraghex_state *state = (void *)&fragskb->cb[0];

                BUG_TRAP(fragskb != NULL);
                BUG_TRAP(sizeof(struct debug_fraghex_state) <=
                         sizeof(fragskb->cb));

                if (includes_fraghdr)
                        skb_pull(fragskb, sizeof(struct mgrp_fraghdr));
                state->fragid = fragid;
                state->offset = offset;
                state->fh = fh;
                strncpy(state->prefix, prefix, sizeof(state->prefix));
                skb_queue_tail(&todebug, fragskb);
        }

        // debug now
        if (debug_count == max) {
                struct sk_buff *fragskb = NULL;
                while ((fragskb = skb_dequeue(&todebug)) != NULL) {
                        struct debug_fraghex_state *state = (void *)&fragskb->cb[0];
                        __debug_fraghex_now(fragskb,
                                            state->fragid,
                                            state->offset,
                                            state->prefix,
                                            state->fh);
                        kfree_skb(fragskb);
                }
        }
}
#endif // LOG_FRAGHEX

