/*
 * Fragmenation Protocol for MGRP
 *
 * Sat Aug  9 16:12:43 EDT 2008 <pavlos@cs.umd.edu>
 *
 * This is a pseudo-protocol to perform fragmentation and reassembly in MGRP.
 * It is not a full transport protocol like TCP or UDP, which also interact
 * with the socket and IP layer.
 *
 * By using a pseudo-protocol, bhtp_rcv() and bhtp_tun_rcv() remain simple
 * and need no special code to handle fragments. It is up to the receive
 * code of the pseudo-protocol (mgrp_frag_tun_recv) to retain state and
 * perform the reassembly. Eventually, the pseudo-protocol will have to
 * pass a packet to one of the "real" transport protocols like TCP or UDP.
 *
 * Using a pseudo-protocol is appropriate because reassembly needs a
 * special header that then points to the next protocol in the chain.
 *
 */
#ifndef __MGRP_FRAG__
#define __MGRP_FRAG__

#if __KERNEL__
#include <linux/in.h>
#include <linux/net.h>
#include <linux/ioctl.h>
#include <linux/skbuff.h>
#include <linux/hash.h>

#include "bhtp_abi.h"

#define MGRP_FRAG_HASH_BUCKETS_DEFAULT  (1<<18)
#define MGRP_FRAG_HASH_BUCKETS_MIN      64      // sanity check

#define MGRP_FRAG_MAXSIZE               (MAX_HEADER + 1500)

#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <linux/types.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <unistd.h>
#endif

// The header is exposed to user-space too (for tcpdump)

/******************************************************
 * Fragmentation header
 *
 * This header is used when a TUN contains a fragment.
 *
 * The MGRP TUN header will contain proto=MGRP_PROTO_FRAG and looks
 * to MGRP as any other transport protocol. It needs to define its
 * own TUN callbacks (struct bhtp_tun_prot) and register with MGRP.
 *
 * Similar to IPv6 fragmentation header but not identical
 * http://www.tcpipguide.com/free/t_IPv6DatagramExtensionHeaders-5.htm#Figure_108
 *
 * Differences:
 *
 *      - We are using a full 16bit fragmentation offset so that
 *      we do not have to deal with padding to 8-byte boundaries
 *
 *      - We use the reserved bits to set the "more fragment" flag
 */
struct mgrp_fraghdr {
        // The protocol for the next header that must be registered with
        // MGRP so that the fragmentation protocol can deliver the packet
        // (the values are taken from BHTP_PROTO_XXX in bhtp.h)
        // (use bhtp_lookup_proto() to get protocol callbacks)
        __u8 proto;

        __u8
             // TUN payload (fragmented or not) is duplicate
             duplicate:1,

             // TUN payload has been fragmented
             fragmented:1,

             // more fragments are following (only valid when fragmented)
             more_fragments:1,

             __unused_flags:5;

        // byte offset of this fragment from the original TUN datagram
        // (only valid when fragmented, 0 otherwise)
        __u16 fragoff;

        // unique ID for every TUN datagram that was fragmented or duplicated
        // (always valid even when not fragmented; used also for duplicates)
        __u32 fragid;
};

#if __KERNEL__

static inline void
mgrp_fraghdr_init(struct mgrp_fraghdr *hdr)
{
        memset(hdr, 0, sizeof(struct mgrp_fraghdr));
}

static inline int
mgrp_fragment_chunk_started(struct sk_buff *skb)
{
        return (bhtp_qsk(skb)->tun_chunk_frag > 0);
}

struct sk_buff * mgrp_fragment_chunk(struct sk_buff *skb,
                                    int chunk_len, int copy_data);
int mgrp_fragment_chunk_encapsulate(struct sk_buff *tskb, int fragid);

// Includes the fragmentation header
static inline int
mgrp_fragment_chunk_left(struct sk_buff *skb)
{
	struct bhtp_qskb *qskb = bhtp_qsk(skb);
        // original protocol payload minus portion already fragmented
        int skb_dataleft = qskb->tun_len - qskb->tun_chunk_frag;

        BUG_ON(skb_dataleft < 0);
        if (skb_dataleft == 0)
                return 0;
        // Even if there is one byte, it needs a fragmenation header
        return sizeof(struct mgrp_fraghdr) + skb_dataleft;
}


/******************************************************
 * mgrp_frag_bucket
 *
 * Each bucket represents a packet to be reassembled.
 * It holds an skb_queue for all the packet fragments.
 */

struct mgrp_frag_key {
        u8 proto;
        u32 source_addr;
        u32 fragid;
};

struct mgrp_frag_gap {
        int offset;
        int size;
        struct list_head glist;
        // pointers to skbs before and after the gap (may be null)
        // the skbs are in the skb queue bucket.fragments
        struct sk_buff *before;
        struct sk_buff *after;
};

struct mgrp_frag_counters {
        short received;
        short dropped_because_already_sent;
        short dropped_because_overlapping;
};

#ifdef LOG_FRAGHASH
struct debug_fraghash {
        u64 unhashed;
        u64 hashed;
        unsigned int index;
};
#endif

struct mgrp_frag_bucket {
        struct mgrp_frag_counters frags;
        struct mgrp_frag_counters pkts;

        u8 released_fragments:1,
           got_unfragmented:1,
           already_sent:1,
           used_frags_for_sent:1,
           used_packet_for_sent:1,
           __unused_flags3;

        struct mgrp_frag_key key;
        struct sk_buff *unfragmented;
        struct sk_buff_head fragments;          // unassembled fragments
        struct list_head gaps;
#ifdef LOG_FRAGHASH
        struct debug_fraghash hs;
#endif
};

/******************************************************
 * mgrp_frag_overflow_bucket
 *
 * Overflow bucket in the overflow list attached to a fixed bucket
 * Allocated individually on demand
 * Contains one fixed reassembly bucket
 * Part of a list of overflow reassembly buckets
 */

struct mgrp_frag_overflow_bucket {
        struct mgrp_frag_bucket bucket;         // reassembly bucket
        struct list_head olist;                 // part of the overflow list
};

/******************************************************
 * mgrp_frag_fixed_bucket
 *
 * Fixed bucket in the hash array.
 * Allocated with all the buckets in the hash array.
 * Contains one fixed reassembly bucket
 * Contains a list of overflow reassembly buckets
 */

struct mgrp_frag_fixed_bucket {
        u8 initialized:1,
            occupied:1,
            __unused_flags:6;

        struct mgrp_frag_bucket bucket;         // reassembly bucket
        struct list_head overflow_buckets;      // mgrp_frag_overflow_bucket list
};

/******************************************************
 * mgrp_frag_hash
 *
 * The reassembly hash.
 *
 * Each hash entry contains a fixed reassemby bucket (struct bucket)
 * and alist of overflow buckets.
 */

struct mgrp_frag_hash_counters {
       // counters never decremented

       // each bucket (fixed or overflow) corresponds to one
       // packet that needs reassembly
       long long fixed_buckets_used;
       long long overflow_buckets_used;

       // number of packets that at some point needed reassembly
       long long needed_reassembly;

       // sum is the number of packets that were reassembled
       // should be equal or less than the counter above
       long long reassembled_from_fragments;
       long long reassembled_from_packet;
};

struct mgrp_frag_hash {
       int index_bits;                          // number of bits used for index
       u64 index_mask64;
       int num_of_buckets;                      // power of two please
       int fbuckets_alloc_bytes;
       struct mgrp_frag_fixed_bucket *fbuckets;
       spinlock_t lock;
       struct mgrp_frag_hash_counters count;
};

static inline int
mgrp_frag_hash_fixed_bucket_index(struct mgrp_frag_hash *fhash,
                                  struct mgrp_frag_fixed_bucket *fbucket)
{
        unsigned char *first = (unsigned char *)fhash->fbuckets;
        unsigned char *cur = (unsigned char *)fbucket;
        return (cur - first) / sizeof(struct mgrp_frag_fixed_bucket);
}

void mgrp_frag_hash_reset(struct mgrp_frag_hash *fhash);

int __init mgrp_frag_init(int max_frag_buckets);
void mgrp_frag_exit(void);

// proc filesystem
extern const struct file_operations mgrp_frag_buckets_fops;
int mgrp_frag_hash_counters_show(char *buffer, char **start, off_t offset,
                                 int length, int *eof, void *data);
int mgrp_frag_reset(char *buffer, char **start, off_t offset,
                    int length, int *eof, void *data);

/******************************************************
 * debug code for fragmentation headers
 */

static inline void
__debug_fragment_hdr(struct mgrp_fraghdr *hdr, const char *prefix)
{
        int proto = hdr->proto;
        LOG("%s fraghdr proto %d flags:%s%s%s fragoff %d fragid %d",
            prefix,
            proto,
            (hdr->duplicate      ? "D" : "."),
            (hdr->fragmented     ? "F" : "."),
            (hdr->more_fragments ? "M" : "."),
            hdr->fragoff,
            hdr->fragid);
}
static inline void
__debug_tunhdr(struct bhtphdr_tun *hdr, const char *prefix)
{
        int proto = hdr->proto;
        int len = hdr->len;
        if (proto != MGRP_PROTO_FRAG)
                return;
        LOG("%s tundhr proto %d flags:%s%s len %d",
            prefix,
            proto,
            (hdr->needs_padding ? "P" : "."),
            (hdr->duplicate     ? "D" : "."),
            len);
}


// fragmentation debugging (sender side)
#ifdef LOG_FRAGMENT
#define DEBUG_FRAG(format, ...)         LOGFN(format, ## __VA_ARGS__)
#define DEBUG_FRAGHDR(hdr)              __debug_fragment_hdr(hdr, __FUNCTION__ ":")
#define DEBUG_FRAGTUN(tun)              __debug_tunhdr(tun, __FUNCTION__ ":")
#define DEBUG_LONESKB(s) \
        if (mgrp_tskb_has_type((s), MGRP_QSKB_FRAGID)) \
                LOGFN("before lone: len %d fragid %d", \
                      (s)->len, bhtp_qsk((s))->fragid)
#else
#define DEBUG_FRAG(format, ...)         ((void)0)
#define DEBUG_FRAGHDR(hdr)              ((void)0)
#define DEBUG_FRAGTUN(tun)              ((void)0)
#define DEBUG_LONESKB(s)                ((void)0)
#endif  // LOG_FRAGMENT

// reassembly debugging (receiver side)
#ifdef LOG_REASSEMBLY
#define DEBUG_REAS(format, ...)         LOGFN(format, ## __VA_ARGS__)
#define DEBUG_RFRAGHDR(hdr)             __debug_fragment_hdr(hdr, __FUNCTION__ ": ")
#define DEBUG_RFRAGTUN(tun)             __debug_tunhdr(tun, __FUNCTION__ ": ")
#else
#define DEBUG_REAS(format, ...)         ((void)0)
#define DEBUG_RFRAGHDR(hdr)             ((void)0)
#define DEBUG_RFRAGTUN(tun)             ((void)0)
#endif  // LOG_REASSEMBLY

// fragmentation/reassembly debugging
// (sender side:   hex dump of originalsa and fragments created)
// (receiver side: hex dump of fragments and reassembled)
#define FRAGHDR_INCLUDED                1
#define FRAGHDR_NOT_INCLUDED            0

#ifdef LOG_FRAGHEX
void __debug_fraghex(struct sk_buff *skb,
                     int fragid,
                     int offset,
                     struct mgrp_fraghdr *fh,
                     int includes_fraghdr,
                     char *prefix);
#define DEBUG_FRAGHEX(s, ...)           __debug_fraghex((s), ## __VA_ARGS__)
#else
#define DEBUG_FRAGHEX(s, ...)           ((void)0)
#endif // LOG_FRAGHEX

#endif  // KERNEL
#endif  // MGRP_FRAG
