/*
 * Simple logging macros.
 *
 * Created:	Sun Mar 26 19:36:51 EST 2006
 *
 * Author:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 *
 */

#ifndef _MGRP_LOG_H
#define _MGRP_LOG_H

#ifdef LOG_PREFIX
#define PREFIX	LOG_PREFIX
#else
#define PREFIX "log: "
#warning no LOG_PREFIX defined, ok
#endif

#include <linux/string.h>

static inline char *
log_basename(char *str)
{
	int len = strlen(str);
	while (len-- > 0) {
		if (str[len] == '/')
			break;
	}
	return str+len+1;
}

#if __KERNEL__
#include "rcerr.h"

#include <linux/kernel.h>

#define PRINT(format, ...) \
	printk(KERN_CRIT PREFIX format "\n", ## __VA_ARGS__)

#define PRINT_TR(format, ...) \
	printk(KERN_CRIT PREFIX "%s:%d " format "\n", \
		__FILE__, __LINE__,  ## __VA_ARGS__)


#define ERR_ON(condition, format, ...) \
	do {\
	if (unlikely((condition)!=0)) PRINT(format, ## __VA_ARGS__);\
	} while(0)

#else

#include <stdio.h>

#define PRINT(format, ...) \
	printf(PREFIX format "\n", ## __VA_ARGS__)

#define PRINT_TR(format, ...) \
	printf(PREFIX "%s:%d " format "\n", \
		__FILE__, __LINE__,  ## __VA_ARGS__)

#define ERR_ON(condition, format, ...) \
	do {\
	if ((condition)!=0) PRINT(format, ## __VA_ARGS__);\
	} while(0)

#endif // __KERNEL__


#define LOG(format, ...) \
        PRINT(format, ## __VA_ARGS__)

#define LOGFN(format, ...) \
        PRINT("%s:%d: " format, __FUNCTION__, __LINE__, ## __VA_ARGS__)

#define LOG_TR(format, ...) \
        PRINT_TR(format, ## __VA_ARGS__)

#define ERR(format, ...) \
        PRINT("error: " format, ## __VA_ARGS__)

#define ERRFN(format, ...) \
        PRINT("%s:%d: error: " format, __FUNCTION__, __LINE__, ## __VA_ARGS__)

#define ERRBUG(format, ...) \
        PRINT("BUG " format, ## __VA_ARGS__)


#define ERR_TR(format, ...) \
        PRINT_TR("error: " format, ## __VA_ARGS__)



#ifdef LOG_DEBUG

#define DEBUG_ON	1

#define DEBUG(format, ...) \
        PRINT(format, ## __VA_ARGS__)

#define DEBUGFN(format, ...) \
        PRINT("%s:%d: " format, __FUNCTION__, __LINE__, ## __VA_ARGS__)

#define DEBUG_TR(format, ...) \
        PRINT_TR(format, ## __VA_ARGS__)

#else

#define DEBUG_ON	0

#define DEBUG(format, ...)		((void)0)
#define DEBUGFN(format, ...)		((void)0)
#define DEBUG_TR(format, ...)		((void)0)

/* #warning "--NO-DEBUG-PRINTK--" */
#endif

#ifdef DEBUG_MEDIANET
#define MDEBUG(format, ...) \
        PRINT("%s:%d: " format, __FUNCTION__, __LINE__, ## __VA_ARGS__)
#else
#define MDEBUG(format, ...)             ((void)0)
#endif

static inline int
debug_on(void)
{
        return DEBUG_ON;
}

#if __KERNEL__

#include <linux/skbuff.h>

int log_tcphdr_count(unsigned char *buf, int buflen);

static inline void
log_print_buf_bytes(unsigned char *buf, int buflen)
{
	int l = 0;
	/* print only up to a multiple of 4 */
	int lines = buflen / 4;

	for (l = 0; l < lines; l++) {
		int i = l * 4;
		int b1 = buf[i];
		int b2 = buf[i+1];
		int b3 = buf[i+2];
		int b4 = buf[i+3];
		LOG("    %.2x %.2x %.2x %.2x", b1, b2, b3, b4);
	}
}

/* returns number of tcp headers found in buf (hardcoded dest port) */
/* return negative on error */
static inline int
log_tcphdr_in_skb(struct sk_buff *skb, char *file, int line)
{
	int rc = 0;
	int hdrs = 0;
	int buflen = skb->len;
	void *buf = NULL;

	if (buflen <= 0)
		return 0;

	buf = kmalloc(buflen, GFP_ATOMIC);
	if (buf == NULL) {
		ERR("tcphdrs_contained: out of mem");
		rc = -ENOMEM;
		goto error;
	}

	rc = skb_copy_bits(skb, 0, buf, buflen);
	if (rc) {
		ERR("tcphdrs_contained: skb_copy_bits failed");
		goto error;
	}

	hdrs = log_tcphdr_count(buf, buflen);
	if (file != NULL)
		LOG(">>> %s:%d: headers %d skb %d", file, line, hdrs, skb->len);
	else
		LOG(">>> headers %d skb %d", hdrs, skb->len);

	kfree(buf);
	return hdrs;

error:
	return rc;
}

#ifdef LOG_TCPHDR
#define LOG_TCPHDR_SKB(_skb) \
	log_tcphdr_in_skb((_skb), log_basename(__FILE__), __LINE__)
#else
#define LOG_TCPHDR_SKB(_skb)	((void)0)

#endif

#endif /* __KERNEL__ */

/*
 * Prints out a buffer in hex using LOG()
 *
 * inspired by audit.c:audit_log_hex()
 * silently reduces the bytes per line if too many
 */
static inline void
log_hex(const void *buf, int buflen,
        int bytes_per_line, const char *prefix)
{
        int i = 0;
        unsigned char line[100];
        unsigned char *ptr = line;
        const unsigned char *cbuf = buf;
        int buffered = 0;
        static const unsigned char *hex = "0123456789abcdef";

        // each byte takes up 3 characters
        // silently decrease @bytes_per_line to fit @line
        while (bytes_per_line > 0 && bytes_per_line * 3 >= sizeof(line))
                bytes_per_line--;

        for (i = 0; i < buflen; i++) {
                if (buffered == bytes_per_line) {
                        *ptr = 0;
                        LOG("%s%s", prefix, line);
                        buffered = 0;
                        ptr = line;
                }
                *ptr++ = ' ';
                *ptr++ = hex[(cbuf[i] & 0xF0)>>4]; /* Upper nibble */
                *ptr++ = hex[cbuf[i] & 0x0F];	  /* Lower nibble */
                buffered++;
        }

        if (buffered > 0) {
                *ptr = 0;
                LOG("%s%s", prefix, line);
        }
}


/* -- Important!
 *
 * #define LOG_PREFIX	"XXX: "
 * Must define LOG_PREFIX with your own prefix.
 */

#endif
