/*
 * BHTP		Bottom Half Transport Protocol
 *
 * Created:	Sun Mar 26 19:36:51 EST 2006
 *
 * Author:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 */


#include <linux/delay.h>

#include "main.h"
#include "tick.h"
#include "sched.h"
#include "proc.h"

/*
 * MGRP TUN protocols
 *
 * See BHTP_PROTO_XXX constants and BHTP_PROTO_MAX.
 * See bhtp_lookup_proto() for protocol callbacks.
 */
#include "tcp.h"
#include "udpex.h"
#include "frag.h"

extern int __init udpex_init(void);
extern void udpex_exit(void);

extern int __init tcp_bhtp_init(void);
extern void tcp_bhtp_exit(void);

// Test the MGRP seqno helper functions (ad-hoc sanity checks)

static inline void
mgrp_seqno_test_pair(mgrp_seqno_t lhs, mgrp_seqno_t rhs)
{
        LOG("    lhs %u rhs %u dist %lld bf %s af %s eq %s inc %u %u",
              lhs, rhs,
              mgrp_seqno_distance(lhs, rhs),
              (mgrp_seqno_before(lhs, rhs) ? "YES" : "NO"),
              (mgrp_seqno_after(lhs, rhs) ? "YES" : "NO"),
              (mgrp_seqno_equal(lhs, rhs) ? "YES" : "NO"),
              mgrp_seqno_inc(lhs),
              mgrp_seqno_inc(rhs));
}

static inline void
mgrp_seqno_test(void)
{
        long long seqno = MGRP_SEQNO_MAX;
        long long seqno1 = mgrp_seqno_inc(seqno);
        long long seqno2 = seqno+1;


	LOGFN("");
        LOG("    size %u max %u mid %u",
              sizeof(mgrp_seqno_t),
              MGRP_SEQNO_MAX, MGRP_SEQNO_MID);

        // test the rollover on incrementing variables of
        // a larger int size
        LOG("    test inc of %lld, inc %lld (correct), plus %lld (incorrect)",
              seqno, seqno1, seqno2);

        mgrp_seqno_test_pair(100, 90);
        mgrp_seqno_test_pair(90, 100);
        mgrp_seqno_test_pair(MGRP_SEQNO_MAX, 100);
        mgrp_seqno_test_pair(100, MGRP_SEQNO_MAX);
}

static inline int
bhtp_is_safe(void)
{
	int rc = 0;

	if (!bhtp_qskb_is_safe()) {
		ERR("private bhtp_qskb does not fit in skb control buffer");
		ERR("aborting");
		rc = -EINVAL;
		goto done;
	}
done:
	return rc;

}

static int __init bhtp_init(void)
{
	int rc = 0;

	LOG("loading");

        mgrp_seqno_test();

	rc = bhtp_is_safe();
	if (rc)
		goto error;

	tick_init();

        // MGRP needs network timestamps taken at netif_receive_skb()
        net_enable_timestamp();

	rc = bhtp_state_init();
	if (rc)
		goto tick_done;

/* 	LOG("starting scheduler thread"); */
/* 	bhtp.sched_thread = kthread_run(bhtp_sched_thread, NULL, "bhtpd"); */

	/* register with the IP layer */
	rc = inet_add_protocol(&bhtp.protocol, IPPROTO_MGRP);
	if (rc < 0) {
		ERR("failed to register BHTP with IP layer");
		rc = -EINVAL;
		goto error_bhtp_state;
	}

	LOG("registered with IP layer");

	LOG("loading datagram transport: UDPex");
	rc = udpex_init();
	if (rc < 0) {
		ERR("failed to load UDPex");
		goto error_protocol;
	}
	LOG("UDPex loaded");

	LOG("loading stream transport: TCP");
	rc = tcp_bhtp_init();
	if (rc < 0) {
		ERR("failed to load TCP");
		goto error_udpex;
	}
	LOG("TCP loaded");

	LOG("loading FRAGMENTATION protocol");
	rc = mgrp_frag_init(bhtp.buckets);
	if (rc < 0) {
		ERR("failed to load FRAG");
		goto error_tcp;
	}
	LOG("FRAGMENATION protocol loaded");

	LOG("loading /proc filesystem");
	rc = mgrp_proc_init();
	if (rc < 0) {
		ERR("failed to load /proc filesystem");
		goto error_frag;
	}

        LOG("MGRP OVERHEAD");
        LOG("    ethernet  (ETH + PHY)   %d", MGRP_OVERHEAD_ETHERNET);
        LOG("    network   (IP header)   %d", MGRP_OVERHEAD_NETWORK - MGRP_OVERHEAD_ETHERNET);
        LOG("    transport (MGRP header) %d", MGRP_OVERHEAD_TRANSPORT - MGRP_OVERHEAD_NETWORK);
        LOG("    total     (ETH/IP/MGRP) %d", MGRP_OVERHEAD_TRANSPORT);

	LOG("loaded with piggybacking policy %s",
            mgrp_pbk_policy_str(bhtp.policy));
	return 0;

error_frag:
        mgrp_frag_exit();
error_tcp:
        tcp_bhtp_exit();
error_udpex:
        udpex_exit();
error_protocol:
        inet_del_protocol(&bhtp.protocol, IPPROTO_MGRP);
error_bhtp_state:
        bhtp_state_done();
tick_done:
	tick_done();
error:
	LOG("loading aborted");
	return rc;
	
}

static void __exit
bhtp_exit(void)
{

	int rc = 0;

	bhtp.done = 1;

	LOG("unloading");

        mgrp_proc_done();

/* 	LOG("stopping scheduler thread"); */
/* 	wake_up(&bhtp.req_wq); */
/* 	rc = kthread_stop(bhtp.sched_thread); */
/* 	LOG("    thread stopped rc %d", rc); */

        mgrp_frag_exit();
        LOG("FRAG unloaded");

	tcp_bhtp_exit();
	LOG("TCP unloaded");

	udpex_exit();
	LOG("UDPex unloaded");

	rc = inet_del_protocol(&bhtp.protocol, IPPROTO_MGRP);
	if (rc < 0)
		ERR("failed to delete BHTP from IP layer");
	LOG("de-registered from IP layer");

	bhtp_state_done();
        // Network timestamps no longer needed
        net_disable_timestamp();
	tick_done();
	LOG("unloaded");
}

module_init(bhtp_init);
module_exit(bhtp_exit);

MODULE_AUTHOR("Pavlos Papageorgiou <pavlos@eng.umd.edu>");
MODULE_DESCRIPTION("Kernel support for MGRP");
MODULE_LICENSE("GPL");

/* module parameters */
/* see LDDv3 p35 */

/* parameters permissions: read/write by everyone */
/* #define MGRP_PARAM_PERM		(S_IRUGO|S_IWUGO) */

/*
 * XXX does not work for new kernel (worked for <2.6.18), bug?
 *
 * 	main.c:155: error: size of array 'type name' is negative
 *
 * For now, do rw for root, r for others and chmod from load script.
 */
#define MGRP_PARAM_PERM		(S_IRUGO|S_IWUSR)

module_param_named(latency, bhtp.latency, int, MGRP_PARAM_PERM);
MODULE_PARM_DESC(latency, "Queue latency in nanosec");

module_param_named(maxpbk, bhtp.maxpbk_percent, int, MGRP_PARAM_PERM);
MODULE_PARM_DESC(maxpbk, "Maximum percentage for packet train piggybacking (0..100)");

module_param_named(maxdelay, bhtp.maxdelay, int, MGRP_PARAM_PERM);
MODULE_PARM_DESC(maxdelay, "Maximum delay when latency=0 and probes wait");

module_param_named(probe_show, udpex_op.probe_show, int, MGRP_PARAM_PERM);
MODULE_PARM_DESC(probe_show, "When delaying probes, print stats");

module_param_named(probe_timeout, udpex_op.probe_timeout, int, MGRP_PARAM_PERM);
MODULE_PARM_DESC(probe_timeout, "Probe timeout in nanosec (but millisec granularity)");

module_param_named(probe_accumulated, udpex_op.probe_accumulated, int, MGRP_PARAM_PERM);
MODULE_PARM_DESC(probe_accumulated, "Bytes of accumulated data before we send probe");

module_param_named(debug_port_unreachable, udpex_op.debug_port_unreachable, int, MGRP_PARAM_PERM);
MODULE_PARM_DESC(debug_port_unreachable, "Print debug info when UDPex port not found and packet is dropped");

module_param_named(buckets, bhtp.buckets, int, MGRP_PARAM_PERM);
MODULE_PARM_DESC(buckets, "Number of buckets in the reassembly hash (receive side)");

module_param_named(policy, bhtp.policy, int, MGRP_PARAM_PERM);
MODULE_PARM_DESC(policy, "Piggybacking policy (see enum mgrp_pbk_policy)");

