/*
 * MGRP		Measurement Manager Protocol
 *
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Mon Feb  4 08:05:06 MST 2008
 *
 * Userspace header file
 */

#ifndef __MGRP_H
#define __MGRP_H

#if __KERNEL__
error "This header file should be included only in userspace!"
#endif

#include <assert.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

#include "mgrp_common.h"

/* Maximum TUN size for tunneled transports.
 * IP(20) + MGRP(16) + TUN <= 1500
 */
#define MGRP_MAX_TUN_SIZE	(1500 - SIZEOF_IPHDR - SIZEOF_MGRPHDR - SIZEOF_FRAGHDR)

struct mgrp_socket_conf {
        // unused
        int flags;
};

static inline int
mgrp_tcp_socket(int domain, int type, int protocol,
                struct mgrp_socket_conf *conf)
{
        int s = 0;
        int on = 0;
        int rc = 0;

        // TCP considers part of the MSS all header options and
        // payload but not the fixed header. So the maximum packet
        // that it will produce (to be fed to MGRP) is:
        //
        //      MSS + <fixed portion of tcphdr (20)>
        //
        // So @mss limits the packet size that the TCP stack can produce.
        // This value is set with the original TCP stack.
        //
        // Since we use TSO, the TCP stack can pass down the stack an skb
        // that holds multiple packets, each up to MSS+hdr bytes.
        int mss = MGRP_MAX_TUN_SIZE - SIZEOF_TCPHDR;

        // Original TCP passes skbs to MGRP that each may hold multiple
        // MSS worth of packets. Since the original TCP does not communicate
        // to MGRP the value of the MSS, we need to set it explicitly.
        //
        // So @fraglen_max is the maximum portion of the MGRP packet that
        // a TCP packet can hold. MGRP calls back TCP to perform the
        // fragmentation.
        int fraglen_max = MGRP_MAX_TUN_SIZE;

        // original socket
        s = socket(domain, type, protocol);
        if (s == -1)
                return s;

        // limit MSS so that it can tunnel over MGRP
        rc = setsockopt(s, IPPROTO_TCP, TCP_MAXSEG, &mss, sizeof(mss));
        if (rc)
                goto error;

        // turn on TCP over MGRP
        on = 1;
        rc = setsockopt(s, IPPROTO_TCP, TCP_MGRP, &on, sizeof(on));
        if (rc)
                goto error_mgrp;

        // turn on fragmentation and set the maximum allowed fragment size
        on = 1;
        rc = setsockopt(s, SOL_MGRP, TCP_MGRP_FRAGMENT, &on, sizeof(on));
        assert(rc == 0);

        rc = setsockopt(s, SOL_MGRP, TCP_MGRP_FRAG_LEN_MAX,
                        &fraglen_max, sizeof(fraglen_max));
        if (rc)
                goto error_mgrp;

        // configure MGRP/UDP socket based on conf
        // XXX not yet; see tsocket.c (hardcoded for now)

        fprintf(stderr, ">>> mgrp.h: mgrp_tcp_socket: socket %d\n", s);
        return s;

error_mgrp:
        fprintf(stderr,
                "%s:%d: error: MGRP setsockopt failed: is MGRP loaded?\n",
                __FILE__, __LINE__);
error:
        return -1;
        
}

static inline int
mgrp_udp_socket(int domain, int type, int protocol,
                struct mgrp_socket_conf *conf)
{
        int s = 0;

        // override protocol type
        type = SOCK_DGRAM;
        protocol = IPPROTO_MGRP;

        // original socket
        s = socket(domain, type, protocol);
        if (s == -1)
                return s;

        // configure MGRP/UDP socket based on conf
        // XXX not yet; see tsocket.c

        fprintf(stderr, ">>> mgrp.h: mgrp_udp_socket: socket %d\n", s);
        return s;
}

static inline int
mgrp_socket(int domain, int type, int protocol,
            struct mgrp_socket_conf *conf)
{
        if (type == SOCK_DGRAM)
                return mgrp_udp_socket(domain, type, protocol, conf);

        if (type == SOCK_STREAM)
                return mgrp_tcp_socket(domain, type, protocol, conf);

        return socket(domain, type, protocol);
}

#endif
