/*
 * MGRP		Measurement Manager Protocol
 *
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Fri Feb 15 10:37:12 MST 2008
 *
 * MGRP counters
 *
 * Common header file (both userspace and kernel)
 */

#ifndef __MGRP_COUNTERS_H
#define __MGRP_COUNTERS_H

struct mgrp_counters {
        long epoch;

        long long mgrp_pkts;
        long long mgrp_bytes_header;
        long long mgrp_bytes_payload;

        // does not include udp packets inside fragments
        // (depending on the policy all piggybacked packets may be fragments)
        long long tcp_pkts;
        long long tcp_bytes;

        // does not include udp packets that are probes
        // does not include udp packets inside fragments
        // (depending on the policy all piggybacked packets may be fragments)
        long long udp_pkts;
        long long udp_bytes;

        // the chunks are fragments of protocol packets
        long long frag_chunks;
        long long frag_bytes;

        // fragmentation stats
        long long frag_chunks_onedup;
        long long frag_chunks_onefrag;
        long long frag_chunks_firstonly;
        long long frag_chunks_middleonly;
        long long frag_chunks_lastonly;

        // you need to inspect the fragmentation header in
        // mgrp_counters_update_send_side() to get fragment
        // counts for each protocol
        // eg. to get frag_tcp_chunks, frag_udp_chunks, ..

        // udp packets that have nonzero 'pad' bytes
        long long probe_pkts;
        // actual = nonpad + pad_unused
        long long probe_bytes_actual;
        // orig   = nonpad + pad_unused + pad_reused
        long long probe_bytes_orig;

        // reuse counters
        long long probe_pkts_reused;
        long long probe_bytes_nonpad;
        long long probe_bytes_pad_unused;
        long long probe_bytes_pad_reused;

        // recv counters

        struct {

        int valid_expected;

        // not strictly a counter, but used to compute the counters below
        long long mgrp_pkt_expected_seqno;

        long long mgrp_pkts;

        // The lost_packets can be approximated by:
        //      mgrp_pkts_missing - mgrp_pkts_outoforder
        //
        // We say "approximate" because the packets may
        // still come in the future as out-of-order.
        long long mgrp_pkts_lost;

        // pkts that were missing from proper sequence
        long long mgrp_pkts_missing;
        // pkts that came out of order
        long long mgrp_pkts_outoforder;
        
        } recv;
};

static inline int
mgrp_counters_oneliner(char *str, int len, struct mgrp_counters *counters)
{
        return snprintf(str, len,
                        "epoch %ld "
                        "mgrp %lld %lld %lld "
                        "tcp %lld %lld "
                        "udp %lld %lld "
                        "probe %lld %lld %lld "
                        "pad %lld %lld %lld %lld "
                        "recv %lld %lld %lld %lld "
                        "frag %lld %lld %lld %lld %lld %lld %lld",

                        counters->epoch,

                        counters->mgrp_pkts,
                        counters->mgrp_bytes_header,
                        counters->mgrp_bytes_payload,

                        counters->tcp_pkts,
                        counters->tcp_bytes,

                        counters->udp_pkts,
                        counters->udp_bytes,

                        counters->probe_pkts,
                        counters->probe_bytes_actual,
                        counters->probe_bytes_orig,

                        counters->probe_pkts_reused,
                        counters->probe_bytes_nonpad,
                        counters->probe_bytes_pad_unused,
                        counters->probe_bytes_pad_reused,

                        counters->recv.mgrp_pkts,
                        counters->recv.mgrp_pkts_lost,
                        counters->recv.mgrp_pkts_missing,
                        counters->recv.mgrp_pkts_outoforder,

                        counters->frag_chunks,
                        counters->frag_bytes,
                        counters->frag_chunks_onedup,
                        counters->frag_chunks_onefrag,
                        counters->frag_chunks_firstonly,
                        counters->frag_chunks_middleonly,
                        counters->frag_chunks_lastonly);
}

static inline int
mgrp_counters_parse_oneliner(struct mgrp_counters *counters,
                             const char *srcstr)
{
        int parsed = sscanf(srcstr,
                            "epoch %ld "
                            "mgrp %lld %lld %lld "
                            "tcp %lld %lld "
                            "udp %lld %lld "
                            "probe %lld %lld %lld "
                            "pad %lld %lld %lld %lld "
                            "recv %lld %lld %lld %lld "
                            "frag %lld %lld %lld %lld %lld %lld %lld",

                            &counters->epoch,

                            &counters->mgrp_pkts,
                            &counters->mgrp_bytes_header,
                            &counters->mgrp_bytes_payload,

                            &counters->tcp_pkts,
                            &counters->tcp_bytes,

                            &counters->udp_pkts,
                            &counters->udp_bytes,

                            &counters->probe_pkts,
                            &counters->probe_bytes_actual,
                            &counters->probe_bytes_orig,

                            &counters->probe_pkts_reused,
                            &counters->probe_bytes_nonpad,
                            &counters->probe_bytes_pad_unused,
                            &counters->probe_bytes_pad_reused,

                            &counters->recv.mgrp_pkts,
                            &counters->recv.mgrp_pkts_lost,
                            &counters->recv.mgrp_pkts_missing,
                            &counters->recv.mgrp_pkts_outoforder,

                            &counters->frag_chunks,
                            &counters->frag_bytes,
                            &counters->frag_chunks_onedup,
                            &counters->frag_chunks_onefrag,
                            &counters->frag_chunks_firstonly,
                            &counters->frag_chunks_middleonly,
                            &counters->frag_chunks_lastonly);

        if (parsed == 26)
                return 0;
        else
                return -1;
}
#endif
