/*
 * MGRP		Measurement Manager Protocol
 *
 * Sun Sep 14 17:59:38 EDT 2008 <pavlos@cs.umd.edu>
 */

#ifndef __MGRPHDR_H
#define __MGRPHDR_H

#include "bhtp.h"
#include "frag.h"

// @pkt should point to the MGRP header
// @caplen is the length of @pkt that was actually captured
static inline const struct mgrp_fraghdr *
mgrp_get_fraghdr(const struct bhtphdr *bh, const u_char *pkt, int pkt_caplen)
{
        int t = 0;
        int tuns = bhtp_header_tuns(bh);

        for (t = 0; t < tuns; t++) {
                const struct bhtphdr_tun *tun = &bh->tun[t];
                const u_char *tdata = bhtp_get_tun_chunk(bh, t, pkt);

                if (tun->proto == MGRP_PROTO_FRAG) {
                        int len_before_fh = 0;
                        // frag header is at the beginning of the chunk
                        const struct mgrp_fraghdr *fh = (struct mgrp_fraghdr *)tdata;

                        // make sure the header has been captured
                        len_before_fh = (const u_char *)fh - pkt;
                        if (pkt_caplen - len_before_fh >= sizeof(struct mgrp_fraghdr))
                                return fh;
                        else
                                return NULL;
                }
        }

        return NULL;
}

enum mgrp_header_error_codes {
        MGRP_HEADER_ERROR_NONE = 0,
        MGRP_HEADER_ERROR_FRAGHDR = -1,
        MGRP_HEADER_ERROR_MGRPHDR = -2,
};

// returns 0 on success
// returns negative on error
//      MGRP_HEADER_ERROR
static inline int
mgrp_header_detect_transports(const struct bhtphdr *bh, int pktlen,
                              const struct mgrp_fraghdr *fh,
                              int *is_udp, int *udp_len, int *udp_len_orig,
                              int *is_tcp, int *tcp_len,
                              char *errstr, int errlen)
{
        int i = 0;
        int mgrp_tuns = bhtp_header_tuns(bh);
        int pad_orig = bhtp_orig_pad_bytes(bh, pktlen, NULL, NULL, NULL, 0);

        for (i = 0; i < mgrp_tuns; i++) {
                const struct bhtphdr_tun *tun = &bh->tun[i];

                if (tun->proto == MGRP_PROTO_FRAG) {
                        // The FRAG protocol contains another protocol inside it
                        // Padding not supported for this chunk
                        BUG_ON(tun->needs_padding);

                        if (fh == NULL) {
                                if (errstr != NULL)
                                        snprintf(errstr, errlen, "error: fraghdr not captured!");
                                return MGRP_HEADER_ERROR_FRAGHDR;

                        } else if (fh->proto == BHTP_PROTO_TCP) {
                                *is_tcp = 1;
                                *tcp_len = tun->len - sizeof(struct mgrp_fraghdr);

                        } else if (fh->proto == BHTP_PROTO_UDP) {
                                *is_udp = 1;
                                *udp_len = tun->len - sizeof(struct mgrp_fraghdr);

                        } else {
                                if (errstr != NULL)
                                        snprintf(errstr, errlen, "error: unknown FRAG protocol %d", fh->proto);
                                return MGRP_HEADER_ERROR_FRAGHDR;
                        }

                } else if (tun->proto == BHTP_PROTO_TCP) {
                        *is_tcp = 1;
                        *tcp_len = tun->len;

                } else if (tun->proto == BHTP_PROTO_UDP) {
                        *is_udp = 1;
                        *udp_len = tun->len;
                        if (tun->needs_padding)
                                *udp_len_orig = tun->len + pad_orig;

                } else {
                        if (errstr != NULL)
                                snprintf(errstr, errlen, "error: unknown MGRP protocol %d", tun->proto);
                        return MGRP_HEADER_ERROR_MGRPHDR;
                }
        }
        return 0;
}

// Oneline representation of MGRP header
//
// @pktlen is the full length of the BHTP packet (including the BHTP header)
//
// The oneliner has the following format:
//
// mgrp:len:tuns:pad_orig:pad_reused:pad_unused>tcp:len>udp:len:origlen
// 
static inline void
mgrp_header_oneliner(const struct bhtphdr *bh, int pktlen,
                     const struct mgrp_fraghdr *fh,
                     char *str, int len)
{
        int i = 0;
        int n = 0;
        char *buf = str;
        int remaining = len;

        int mgrp_len = pktlen;
        int mgrp_tuns = bhtp_header_tuns(bh);
        int pad_reused = 0;
        int pad_unused = 0;
        int pad_orig = bhtp_orig_pad_bytes(bh, pktlen,
                                           &pad_reused, &pad_unused,
                                           NULL, 0);
        
        n = snprintf(buf, remaining, "mgrp:%d:%d:%d:%d:%d",
                     mgrp_len, mgrp_tuns, pad_orig, pad_reused, pad_unused);
        if (n >= remaining)
                goto done;
        buf += n;
        remaining -= n;

        for (i = 0; i < mgrp_tuns; i++) {
                const struct bhtphdr_tun *tun = &bh->tun[i];

                if (tun->needs_padding)
                        n = snprintf(buf, remaining, ">%s:%d:%d",
                                     mgrp_tun_proto_str(tun->proto),
                                     tun->len, tun->len + pad_orig);

                else if (tun->proto == MGRP_PROTO_FRAG && fh != NULL)
                        n = snprintf(buf, remaining, ">%s:%d:frag",
                                     mgrp_tun_proto_str(fh->proto),
                                     tun->len - sizeof(struct mgrp_fraghdr));

                else
                        n = snprintf(buf, remaining, ">%s:%d",
                                     mgrp_tun_proto_str(tun->proto), tun->len);

                if (n >= remaining)
                        goto done;
                buf += n;
                remaining -= n;
        }
done:
        str[len-1] = 0;
        return;
}
#endif
