/*
 * BHTP		Bottom Half Transport Protocol
 *
 * Created:	Sun Mar 26 19:36:51 EST 2006
 *
 * Author:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 */

#ifndef _BHTP_ABI_H
#define _BHTP_ABI_H

#include "log.h"
#include "bhtp.h"
#include "tick.h"
#include "proto.h"
#include "mgrp_common.h"
#include "mgrp_counters.h"

#include <linux/skbuff.h>

/* bhtp_ut state flags */
#define BHTP_TX_SCHED          0x0001

#define BHTP_FLOW_ACK_ALL	(1<<0)
#define BHTP_FLOW_ONE_DST	(1<<1)
#define BHTP_FLOW_SHARED	(1<<2)
#define BHTP_FLOW_PKT_SHARE	(1<<3)

/*
 * These are byte lengths of the packet below the UDP header.
 *
 * So, for example when sending a UDP packet send(buf:1000bytes),
 * then the packet lengths:
 *
 *	tun_len		= UDPex(8) + probe_payload()
 *	tun_frag_max_len= maximum length of tun chunk
 *			  (if max_len < tun_len then bhtp needs to fragment)
 *
 *	tun_hdr		= UDPex(8)
 *	tun_data	= probe_payload()
 *	tun_data_orig	= tun_data + tun_pad
 *	tun_data_frag	= n/a (portion of top data already fragmented)
 *	tun_len		= tun_hdr + tun_data
 *	tun_pad		= 1000 - tun_data
 *
 *	** non-bhtp case:
 *	ip_packet_len	= IP(20) + UDP(8) + 1000
 *
 *	** bhtp-case (no piggybacking):
 *	ip_packet_len	= IP(20) + BHTP() + tun_hdr + tun_data + tun_pad
 *
 *	** bhtp_case (piggybacking):
 *	ip_packet_len	= IP(20) + BHTP() +
 *		(tun_hdr + tun_data) +			// udp probe
 *		(tun2_hdr + tun2_data) +		// piggybacked
 *		(tun_pad - tund2_hdr - tun2_data)	// leftover padding
 */

#define BHTP_QSKB_MAGIC		0xaf

#define BHTP_QSKB_PROBE		(1<<0)
#define BHTP_QSKB_QUEUE		(1<<1)
#define BHTP_QSKB_RECALC_HDR	(1<<2)
#define BHTP_QSKB_FRAGMENT	(1<<3)
#define BHTP_QSKB_NEVER_QUEUE	(1<<4)
#define BHTP_QSKB_VESSEL	(1<<5)

struct bhtp_flow;

struct bhtp_flow_req {
	struct sk_buff *skb;
	struct bhtp_flow *bflow;
	struct list_head list;
	atomic_t sent;
	long long idg;
	long long idf;
	ktime_t latency;
	ktime_t send_kt;
};

// Each skb contains one counter for each consumer type

// Convenience enum. MUST start from 0 and go up.
// It is used as index in the bhtp_qskb.consumer[] array.
enum mgrp_consumer_type {
        MGRP_CONSUMER_TRANSPORT = 0,
        MGRP_CONSUMER_PBK,
        MGRP_CONSUMER_PBK_REDUNDANT,
        MGRP_CONSUMER_MAX,
};

struct bhtp_qskb {
	struct bhtp_tun *tun;
	short magic;
	unsigned char type;

        // Bitfield, one bit for every consumer type.
        //
        // If set then the consumer has exhausted the fragments
        // Cannot keep track only with tun_data_frag because
        // we may have packets with only a header, in which
        // case tun_data_frag is always zero.
        //
        // See tcp_tun_fraglen_left() and tcp_tun_fragment()
	unsigned char tun_data_frag_done;

        struct {
                unsigned short tun_data_frag;
        } consumer[MGRP_CONSUMER_MAX];

	unsigned short tun_len;
	unsigned short tun_frag_len_max;
	unsigned short tun_hdr;
	unsigned short tun_data;
	unsigned short tun_data_orig;
	short tun_pad;
        short tun_piggyback;
	int tun_gap;
	union {
		struct {
			struct bhtphdr *bh;
		} recv;
		struct {
			struct bhtp_flow_req *req;
		} send;
	};
	ktime_t kt;
	/* reached its limit on 32-bit arch: 48 bytes */
};

static inline int
mgrp_consumer_type_done(struct bhtp_qskb *qskb, enum mgrp_consumer_type type)
{
        unsigned char done_bitmask = (1 << type);
        int frag_done = (qskb->tun_data_frag_done & done_bitmask);
        return frag_done;
}

static inline void
mgrp_consumer_type_done_set(struct bhtp_qskb *qskb, int done,
                            enum mgrp_consumer_type type)
{
        unsigned char done_bitmask = (1 << type);

        qskb->tun_data_frag_done &= ~done_bitmask;
        if (done)
                qskb->tun_data_frag_done &= done_bitmask;
}

static inline void
mgrp_consumer_type_update_frag(struct bhtp_qskb *qskb,
                               int frag_len, int frag_data, 
                               enum mgrp_consumer_type type)
{
}

#define MAX_INT_VALUE(intvar)   ((1 << (8*sizeof(intvar))) - 1)

static inline int
bhtp_qskb_is_safe(void)
{
	struct sk_buff tmp;
        struct bhtp_qskb dummy_qskb;
	int safe = sizeof(struct bhtp_qskb) <= sizeof(tmp.cb);

	LOG("bhtp_qskb %s safe: bhtp_qskb_size %d cb_size %d max_len %d",
		(safe ? "is" : "is NOT"),
		sizeof(struct bhtp_qskb),
		sizeof(tmp.cb),
                MAX_INT_VALUE(dummy_qskb.tun_len));

	LOG("bhtp_qskb MGRP_CONSUMER_MAX %d", MGRP_CONSUMER_MAX);

	return safe;
}

static inline int
bhtp_qsk_valid(struct bhtp_qskb *qskb) {
	return (qskb->magic == BHTP_QSKB_MAGIC);
}

static inline struct bhtp_qskb *
bhtp_qsk_novalidate(struct sk_buff *skb) {
	struct bhtp_qskb *qskb = (struct bhtp_qskb *)skb->cb;
	BUG_ON(sizeof(struct bhtp_qskb) > sizeof(skb->cb));
	return qskb;
}

static inline int
bhtp_skb_has_valid_qsk(struct sk_buff *skb) {
	struct bhtp_qskb *qskb = bhtp_qsk_novalidate(skb);
	return bhtp_qsk_valid(qskb);
}

#define bhtp_qsk(_skb) \
	(bhtp_skb_has_valid_qsk((_skb)) ? \
		bhtp_qsk_novalidate((_skb)) : \
		(ERR_TR("invalid qsk in skb %p", (_skb)), \
		(struct bhtp_qskb *)mgrp_panic()))

static inline struct bhtp_qskb *
bhtp_qsk_init(struct sk_buff *skb) {
	struct bhtp_qskb *qskb = bhtp_qsk_novalidate(skb);
	memset(qskb, 0, sizeof(*qskb));
	qskb->magic = BHTP_QSKB_MAGIC;
	return qskb;
}

/* Same as bhtp_qsk_init but clones @from_qskb instead of zeroing it out */
static inline struct bhtp_qskb *
bhtp_qsk_clone(struct sk_buff *skb, struct bhtp_qskb *from_qskb) {
	struct bhtp_qskb *qskb = bhtp_qsk_novalidate(skb);
	memcpy(qskb, from_qskb, sizeof(*qskb));
	return qskb;
}

/* a BHTP flow */
struct bhtp_flow {

	u32 id;
	u32 flags;
        // 0 or >1
	atomic_t users;
        // 0 or 1
        atomic_t recv_user;

	spinlock_t lock;

	u32 daddr;

	/* part of the flow list */
	struct list_head flist;

	/* points to the poll_list of transports */
	struct list_head tuns;

	struct sock *sk;
	struct dst_entry *dst;

	/* list of queued skbs that are waiting to be sent */
	struct sk_buff_head tosend;
	long long tosend_accumulated;
	spinlock_t tosend_lock;
	ktime_t tosend_latency;

	atomic_t reqid;

        /* counters */
        struct mgrp_counters counters;

        /* base timestamp for bhtphdr.sndstamp */
        unsigned long long sndstamp_base_ns;
};


/* upper transport unit */

/* _ops and _proto are defined once per upper transport protocol */
struct bhtp_tun_ops {

	/* rcv: process a received protocol segment: hdr + data */
	int (*rcv)(struct sk_buff *skb);

	/*
	 * snd: bite a fragment from the main protocol segment
	 *
	 *	Callee must keep track of data already fragmented
	 *	and make sure that qskb length fields are valid.
	 *	(see qskb->tun_data_frag)
	 */
	struct sk_buff * (*fragment)(struct sk_buff *skb, int frag_len,
                                     enum mgrp_consumer_type type);

	/*
	 * snd: return the length (hdr + data) of the packet remaining
	 *
	 * Usually used during fragmentation. If the packet has not been
	 * fragmented, this is equivalent to skb->len.
	 */
	int (*fraglen_left)(struct sk_buff *skb, enum mgrp_consumer_type type);
};

struct bhtp_tun_prot {
	u8 proto;
	struct bhtp_tun_ops *ops;
};

/* This is an instance, associate with each socket for example */
struct bhtp_tun {
	struct bhtp_tun_prot *prot;
	unsigned long state;
/* 	struct list_head poll_list; */
	unsigned int barrier_pkts;
	unsigned int barrier_total;
	unsigned int barrier_wait_usec;
	unsigned int barrier_min;
	unsigned int barrier_max;
};

#define BHTP_TUN_PROT_HAS_RCV(_prot) \
	((_prot) != NULL && \
	 (_prot)->ops != NULL && \
	 (_prot)->ops->rcv != NULL)

#define BHTP_TUN_PROT_HAS_FRAG(_prot) \
	((_prot) != NULL && \
	 (_prot)->ops != NULL && \
	 (_prot)->ops->fragment != NULL && \
	 (_prot)->ops->fraglen_left != NULL)

static inline int
bhtp_flow_matches_addr(struct bhtp_flow *bflow, struct sockaddr_in *usin)
{
	return (bflow->daddr == usin->sin_addr.s_addr);
}

static inline int
bhtp_flow_addr(struct bhtp_flow *bflow, int count)
{
	u8 *octets = (u8 *)&bflow->daddr;
	BUG_ON(count <= 0 || count > 4);
	return octets[count-1];
}


/* timestamps */

static inline void
bhtp_recv_get_sysstamp(struct sock *sk, struct sk_buff *skb,
		       struct bhtp_tstamp *tstamp)
{
	// HERE
	/* based on sock_recv_timestamp() */
	struct timeval stamp;

	tstamp->type |= BHTP_TSTAMP_LOCAL;
	tstamp->type |= BHTP_TSTAMP_WALLCLOCK;

	skb_get_timestamp(skb, &stamp);
	if (stamp.tv_sec == 0) {
                __net_timestamp(skb);
                skb_get_timestamp(skb, &stamp);
		tstamp->type |= BHTP_TSTAMP_TAKEN_AT_BHTP;
	} else {
		tstamp->type |= BHTP_TSTAMP_TAKEN_BELOW_BHTP;
	}

	tstamp->nsec = timeval_to_ns(&stamp);

	/* FIXME: is this needed? It is in sock_recv_timestamp() */
/* 	sk->sk_stamp = stamp; */
}

static inline void
bhtp_recv_get_sndstamp(struct sock *sk, struct sk_buff *skb,
		       struct bhtp_tstamp *tstamp)
{
	struct bhtp_qskb *qskb = bhtp_qsk(skb);

	tstamp->type |= BHTP_TSTAMP_REMOTE;
	tstamp->type |= BHTP_TSTAMP_MONOTONIC;
	tstamp->type |= BHTP_TSTAMP_TAKEN_AT_BHTP;

	tstamp->nsec = bhtp_get_sndstamp_ns(qskb->recv.bh);
}

static inline void
bhtp_recv_get_rcvstamp(struct sock *sk, struct sk_buff *skb,
		       struct bhtp_tstamp *tstamp)
{
	tstamp->type |= BHTP_TSTAMP_LOCAL;
	tstamp->type |= BHTP_TSTAMP_MONOTONIC;
        tstamp->nsec = ktime_to_ns(skb->tstamp);

	if (tstamp->nsec > 0) {
		tstamp->type |= BHTP_TSTAMP_TAKEN_BELOW_BHTP;
	} else {
		tstamp->type |= BHTP_TSTAMP_TAKEN_AT_BHTP;
		tstamp->nsec = tick_now_ns();
	}
}

static inline void
bhtp_recv_get_timestamps(struct sock *sk, struct sk_buff *skb,
			 struct bhtp_timestamps *tstamps)
{
	memset(tstamps, 0, sizeof(*tstamps));

	bhtp_recv_get_sysstamp(sk, skb, &tstamps->sys);
	bhtp_recv_get_rcvstamp(sk, skb, &tstamps->rcv);
	bhtp_recv_get_sndstamp(sk, skb, &tstamps->snd);
}

/* start/stop payload buffer */
int bhtp_drain_tosend(struct bhtp_flow *bflow, int drop);

static inline ktime_t
mgrp_global_flow_latency(struct bhtp_flow *bflow)
{
	return ktime_add_ns(ktime_set(0, 0), bhtp.latency);
}

static inline ktime_t
mgrp_local_flow_latency(struct bhtp_flow *bflow)
{
	return bflow->tosend_latency;
}


static inline int
mgrp_flow_payload_buffer_start(struct bhtp_flow *bflow)
{
	if (ktime_to_ns(mgrp_global_flow_latency(bflow)) > 0)
		return 0;
	
	bflow->tosend_latency = ktime_add_ns(ktime_set(0, 0), bhtp.maxdelay);
	return 1;
}

static inline void
mgrp_flow_payload_buffer_stop(struct bhtp_flow *bflow)
{
	int drained = 0;
	if (ktime_to_ns(mgrp_local_flow_latency(bflow)) == 0)
		return;

	bflow->tosend_latency = ktime_set(0, 0);
	drained = bhtp_drain_tosend(bflow, 0);
	if (drained >= 0)
		DEBUG("flow_payload_buffer_stop: drained buffer: %d skbs",
			drained);
}


int bhtp_register_tun_prot(struct bhtp_tun_prot *tun_prot);
int bhtp_unregister_tun_prot(struct bhtp_tun_prot *tun_prot);

struct bhtp_flow *bhtp_find_attach_flow(struct sockaddr_in *usin, int flags);
int bhtp_detach_flow(struct bhtp_flow *bflow, struct bhtp_tun *tun);

int bhtp_schedule_skb(struct bhtp_flow *bflow, struct sk_buff *tun_skb);

int bhtp_send_barrier_skbs(struct bhtp_flow *bflow,
		struct sock *sk, struct bhtp_tun *tun,
		unsigned long long *twait_p,
		unsigned long long *tsend_p);

#endif /* _BHTP_ABI_H */

