/*
 * BHTP		Bottom Half Transport Protocol
 *
 * Created:	Sun Mar 26 19:36:51 EST 2006
 *
 * Author:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 */

#include <net/route.h>		/* for ip_route_XXX */

#include "main.h"
#include "proto.h"
#include "tick.h"
#include "sysfs.h"

/* hack, used to be in skbuff.h */
/* try to use skb_copy_bits() instead */
#include "/src/kernel/linux/net/core/kmap_skb.h"

#ifdef BHTP_FRAGS
#error BHTP_FRAGS does not work yet
#endif

static inline int bhtp_enqueue_req(struct bhtp_flow *bflow, struct sk_buff *skb,
				ktime_t latency);

static inline void bhtp_requeue_req(struct bhtp_flow_req *req);
static inline struct bhtp_flow_req *bhtp_dequeue_req(void);


/* dummy socket prot to be attached to bflow, needed by ip_queue_xmit */

struct dummy_sock {
	/* inet_sock has to be the first member */
	struct inet_sock inet;
};

int
dummy_init_sock(struct sock *sk)
{
	LOG("dummy_init_sock");
	return 0;
}

int
dummy_destroy_sock(struct sock *sk)
{
	LOG("dummy_destroy_sock");
	return 0;
}

struct proto dummy_prot = {
	.name		= "dummy",
	.owner		= THIS_MODULE,
	.init		= dummy_init_sock,
	.destroy	= dummy_destroy_sock,
	.obj_size	= sizeof(struct dummy_sock),
};

static inline void
bhtp_header_fill_sndstamp(struct bhtphdr *bh, const unsigned long long base_ns)
{
        int overflow = bhtp_set_sndstamp_ns(bh, tick_now_ns() - base_ns);
        if (overflow)
                LOGFN("sndstamp overflow");
}

/* A B I */

int
bhtp_register_tun_prot(struct bhtp_tun_prot *tun_prot)
{
	int rc = -EINVAL;

	/* FIXME: sanity checks? */

	if (tun_prot->ops == NULL)
		goto err;
	if (tun_prot->ops->rcv == NULL)
		goto err;

	spin_lock_bh(&bhtp.lock);

	/* FIXME: tun_prot->proto is u8 so it is never out of bounds. */
	if (bhtp.tun_prots[tun_prot->proto] != NULL)
		/* Proto is already registered. */
		goto err_unlock;

	bhtp.tun_prots[tun_prot->proto] = tun_prot;

	spin_unlock_bh(&bhtp.lock);
	return 0;

err_unlock:
	spin_unlock_bh(&bhtp.lock);
err:
	return rc;
}

int
bhtp_unregister_tun_prot(struct bhtp_tun_prot *tun_prot)
{
	int rc = -EINVAL;

	spin_lock_bh(&bhtp.lock);

	if (tun_prot != bhtp.tun_prots[tun_prot->proto])
		goto err_unlock;

	bhtp.tun_prots[tun_prot->proto] =  NULL;

	spin_unlock_bh(&bhtp.lock);
	return 0;

err_unlock:
	spin_unlock_bh(&bhtp.lock);
	return rc;
}

static struct dst_entry *
bhtp_get_dst(struct sockaddr_in *daddr, struct sockaddr_in *saddr)
{
	/* based on net/sctp/protocol.c:sctp_v4_get_dst() */
	/* Also check LNI, c21, p478-9, Setting the route. */

	struct rtable *rt = NULL;
	struct dst_entry *dst = NULL;
	struct flowi fl = {
		.proto = IPPROTO_MGRP,
		.fl4_dst = daddr->sin_addr.s_addr,
	};

	if (saddr)
		fl.fl4_src = saddr->sin_addr.s_addr;

	LOG("looking up route dst " NIPQUAD_FMT " src " NIPQUAD_FMT,
	      NIPQUAD(fl.fl4_dst), NIPQUAD(fl.fl4_src));
	      
	if (!ip_route_output_key(&init_net, &rt, &fl))
		dst = &rt->u.dst;

	return dst;
}

static inline struct dst_entry *
bhtp_dst_check(struct bhtp_flow *bflow)
{
	/* based on net/sock.h:sk_dst_check() */
	struct dst_entry *dst = bflow->dst;

	if (dst && dst->obsolete) {
		dst_release(dst);
		bflow->dst = NULL;
		return NULL;
	}

	return dst;
}

void
bhtp_sk_destruct(struct sock *sk)
{
	LOG("DESTRUCT bhtp_sk_destruct");
}

/* 0 ok, -1 error */
static inline int
bhtp_flow_init(struct bhtp_flow *bflow, int bflow_id,
               struct net *net, struct sockaddr_in *usin, int flags)
{
	struct dst_entry *dst = NULL;
	struct sockaddr_in dst_addr = {
		.sin_addr.s_addr = usin->sin_addr.s_addr,
	};

        // bflow has been allocated and zeroed by caller; now we need to init

	spin_lock_init(&bflow->lock);
	spin_lock_init(&bflow->tosend_lock);

	INIT_LIST_HEAD(&bflow->flist);
	INIT_LIST_HEAD(&bflow->tuns);

	skb_queue_head_init(&bflow->tosend);

	bflow->id = bflow_id;
	bflow->flags = flags;
	bflow->daddr = usin->sin_addr.s_addr;

	bflow->sk = sk_alloc(net, AF_INET, GFP_ATOMIC, &dummy_prot);
	if (bflow->sk == NULL)
		goto error;
	/* this seems to be the only field that ip_queue_xmit needs */
	/* do we need to do a sock_hold() ? */
	sock_hold(bflow->sk);
	/* HACK so that ip_queue_xmit will pick them up */
	bflow->sk->sk_protocol = IPPROTO_MGRP;
	bflow->sk->sk_destruct = &bhtp_sk_destruct;
	inet_sk(bflow->sk)->uc_ttl = 162;

	/* FIXME: We are ignoring source address at this point.
	 * This is ok when our machine has one IP but not for
	 * multihomed machines.
	 */

	dst = bhtp_get_dst(&dst_addr, NULL);
	if (dst) {
		bflow->dst = dst_clone(dst);
		dst_release(dst);
		LOG("got dst");
	}

        bflow->counters.epoch = bflow->id;

        // All MGRP packets contain a send timestamp (bhtphdr.sndstamp)
        // that is an offset to this base timestamp.
        bflow->sndstamp_base_ns = tick_now_ns();
        LOGFN("sndstamp_base_ns %lld", bflow->sndstamp_base_ns);

	LOG("created new bhtp flow: id %d, addr " NIPQUAD_FMT,
		bflow->id, NIPQUAD(bflow->daddr));

	return 0;
error:
	LOG_TR("out of mem");
	return -1;
}


void
bhtp_flow_done(struct bhtp_flow *bflow)
{
        // flow was marked by bhtp_rcv() in bhtp_find_recv_flow()
        if (atomic_add_unless(&bflow->recv_user, -1, 0)) {
                atomic_dec(&bflow->users);
        }

	if (atomic_read(&bflow->users) != 0) {
		LOG("BUG: Flow is still used. Forcing removal.");
        }

	if (bflow->dst)
		dst_release(bflow->dst);
	/* XXX FIXME check if also still part of a list in flist */

	bhtp_drain_tosend(bflow, 1);

	LOG("FLOW refcount %d", atomic_read(&bflow->sk->sk_refcnt));

	/* FIXME A hack because we get the following error on the console:
	 *
	 *	Attempt to release TCP socket in state 1 f6eb4980
	 *
	 * which is in af_inet.c:inet_sock_desctruct().
	 *
	 * This could also be why after we open/close a TCP BHTP socket
	 * we cannot unload the bhtp module. FIXME
	 */
	bflow->sk->sk_state = TCP_CLOSE;
	sock_put(bflow->sk);
	bflow->sk = NULL;
}

/*
 * Destroys bhtp flows and removes them from the global list:
 *
 *      if @only_this_flow == NULL, destroys all flows
 *      if @only_this_flow != NULL, destroys only the matching flow
 * 
 * Returns the number of flows removed.
 *
 * Holds bhtp.lock
 */
static inline int
mgrp_destroy_flows(struct bhtp_flow *only_this_bflow)
{
        int count = 0;
        struct bhtp_flow *bflow = NULL;
        struct bhtp_flow *tmp_bflow = NULL;

        spin_lock_bh(&bhtp.lock);

        list_for_each_entry_safe(bflow, tmp_bflow, &bhtp.flows, flist) {
                if ((only_this_bflow == NULL) || (bflow == only_this_bflow)) {
                        list_del(&bflow->flist);
                        bhtp_flow_done(bflow);
                        kfree(bflow);
                        count++;
                        if (only_this_bflow != NULL)
                                break;
                }
        }

        spin_unlock_bh(&bhtp.lock);
        return count;
}

// Returns number of flows destroyed
int
mgrp_destroy_all_flows(void)
{
        int count = mgrp_destroy_flows(NULL);
        LOG("destroyed %d flows", count);
        return count;
}

// Returns -1 on error, 0 on success
static inline int
mgrp_destroy_one_flow(struct bhtp_flow *bflow)
{
        int count = mgrp_destroy_flows(bflow);
        BUG_ON(count > 1);
        return (count == 1) ? 0 : -1;
}

int
bhtp_detach_flow(struct bhtp_flow *bflow, struct bhtp_tun *tun)
{
	int can_destroy = 0;
	int users = 0;
        int rc = 0;
	
	can_destroy = atomic_dec_and_test(&bflow->users);
	users = atomic_read(&bflow->users);

	if (users < 0) {
		ERR("cannot detach flow: negative user count: %d", users);
		return -EINVAL;
	}

	if (!can_destroy) {
		DEBUG("not detaching flow %d: positive user count: %d",
                      bflow->id, users);
		/* but not an error */
		return 0;
	}

        rc = mgrp_destroy_one_flow(bflow);
        if (rc) {
		ERR("cannot detach flow: failed to destroy it: %d", rc);
                return -EINVAL;
        }
	return 0;
}

// must be called with bhtp.lock held
static inline struct bhtp_flow *
_bhtp_flow_lookup(struct sockaddr_in *usin, int flags)
{
	struct bhtp_flow *bflow = NULL;

        // XXX also add a hash table for fast lookup
        // XXX not a problem for our setup where we have 1-2 flows max

	list_for_each_entry(bflow, &bhtp.flows, flist) {
		/* XXX should we compare flags too? */
		if (bhtp_flow_matches_addr(bflow, usin))
			return bflow;
	}
        return NULL;
}

static inline struct bhtp_flow *
bhtp_find_create_flow(struct sockaddr_in *usin, int flags)
{
        int rc = 0;
	struct bhtp_flow *bflow = NULL;
	struct bhtp_flow *new_bflow = NULL;
        int new_bflow_id = 0;

	spin_lock_bh(&bhtp.lock);
        bflow = _bhtp_flow_lookup(usin, flags);
        if (bflow != NULL) {
                goto flow_found_unlock;
        } else {
                // reserve a flow id for the flow that we will create
                bhtp.last_flowid++;
                new_bflow_id = bhtp.last_flowid;
        }
	spin_unlock_bh(&bhtp.lock);

        // flow does not exist
        // create it outside of the lock (due to mem allocations)

	new_bflow = kzalloc(sizeof(struct bhtp_flow), GFP_ATOMIC);
	if (new_bflow == NULL) {
                LOG_TR("out of mem");
		goto error;
        }

        // XXX should the net namespace propagate higher?
        // XXX ie, to the callers of bhtp_find_attach_flow?
        rc = bhtp_flow_init(new_bflow, new_bflow_id,
                            &init_net, usin, flags);
        if (rc)
                goto err_free_new_bflow;

        // insert the new flow while holding lock

	spin_lock_bh(&bhtp.lock);

        // check again; another bhtp_find_attach_flow() may have sneaked it in
        bflow = _bhtp_flow_lookup(usin, flags);
        if (bflow != NULL)
                goto flow_found_unlock;

        list_add(&new_bflow->flist, &bhtp.flows);
        bflow = new_bflow;

flow_found_unlock:
	spin_unlock_bh(&bhtp.lock);
	return bflow;

err_free_new_bflow:
        bhtp_flow_done(new_bflow);
        kfree(new_bflow);
        new_bflow = NULL;
error:
	return NULL;
}

struct bhtp_flow *
bhtp_find_attach_flow(struct sockaddr_in *usin, int flags)
{
        struct bhtp_flow *bflow = bhtp_find_create_flow(usin, flags);
        if (bflow != NULL)
                atomic_inc(&bflow->users);
        return bflow;
}

static inline struct bhtp_flow *
bhtp_find_mark_recv_flow(struct sockaddr_in *usin, int flags)
{
        struct bhtp_flow *bflow = bhtp_find_create_flow(usin, flags);

        if (bflow != NULL) {
                int already_marked = atomic_read(&bflow->recv_user);
                if (!already_marked) {
                        int marked = atomic_add_unless(&bflow->recv_user, 1, 1);
                        LOGFN("marked flow");
                        if (marked) {
                                // it is decremented in bhtp_flow_done()
                                atomic_inc(&bflow->users);
                        }
                }
        }
        return bflow;
}


/* FIXME: Is bhtp_shared_info still needed?
 *
 * Space for it is reserved in bhtp_alloc_container_skb().
 * All bhtp private data is saved in to the qskb which is
 * stored in the skb->cb[].
 *
 * Note that the big difference between these two is during
 * skb_clone(). The qskb is cloned as well (since it is part
 * of the sk_buff structure itself) while the bhtp_shared_info
 * is not (it is part of the skb head buffer).
 */

/* private bhtp data */
struct bhtp_shared_info {
	int len;
};

inline struct bhtp_shared_info *
bhtp_shinfo(struct sk_buff *skb)
{
	struct bhtp_shared_info *info = (struct bhtp_shared_info *)skb->head;
	return info;
}

void
bhtp_container_skb_destructor(struct sk_buff *skb)
{
/* 	struct bhtp_shared_info *info = bhtp_shinfo(skb); */
/* 	DEBUG("skb_destructor len %d infor->len %d", skb->len, info->len); */
}

inline struct sk_buff *
bhtp_alloc_container_skb(int size)
{
	struct sk_buff *skb = NULL;
	int reserve =  SKB_DATA_ALIGN(sizeof(struct bhtp_shared_info));
	
	size += reserve;
	skb = alloc_skb(size, GFP_ATOMIC);
	if (skb == NULL)
		return NULL;
	
	skb_reserve(skb, reserve);
	memset(bhtp_shinfo(skb), 0, reserve);

	skb->destructor = bhtp_container_skb_destructor;
	return skb;
}

	/* no fragments supported yet FIXME */
/* 	DEBUG("tun_len %d tun_data_len %d nr_frags %d", */
/* 		tun_skb->len, tun_skb->data_len, skb_shinfo(tun_skb)->nr_frags); */
/* 	DEBUG("nonlinear %d headlen %d pagelen %d", */
/* 		skb_is_nonlinear(tun_skb), */
/* 		skb_headlen(tun_skb), */
/* 		skb_pagelen(tun_skb)); */

/* returns negative error code or >= valid index */
static inline int
bhtp_reserve_tun_in_header(struct bhtphdr *bh,
                           int proto, int len, int needs_padding)
{
	int tuns = bhtp_header_tuns(bh);

	if (tuns < 0 || tuns >= BHTP_MAX_SEG) {
		ERR("reserve_tun: no more TUNs left: tuns %d max %d",
			tuns, BHTP_MAX_SEG);
		return -ENOBUFS;
	}

	bh->tun[tuns].proto = proto;
	bh->tun[tuns].len = len;
	bh->tun[tuns].needs_padding = needs_padding;

        bhtp_header_tuns_set(bh, tuns+1);

	/* index for newly created TUN entry */
	return tuns;
}

/* skb->transport_header must point to BHTP header */
static inline int
bhtp_append_tun_pad(struct sk_buff *skb, int padding)
{
	int rc = 0;
	struct bhtphdr *bh = (struct bhtphdr *)skb->transport_header;
	void *padptr = NULL;

	if (padding <= 0) {
		rc = -EINVAL;
		goto error;
	}
	
	DEBUG("append_pad: inserting padding: %d bytes", padding);

        if (bh->has_unused_padding) {
                ERR("MGRP pkt has already padding");
                rc = -EINVAL;
		goto error;
        }

	/* does it fit? */
	if (unlikely(skb_tailroom(skb) < padding)) {
		ERR("append_tun: no space for %d padding, only %d left",
			padding, skb_tailroom(skb));
		rc = -ENOBUFS;
		goto error;
	}

	padptr = skb_put(skb, padding);
	memset(padptr, BHTP_PAD_BYTE, padding);
        bh->has_unused_padding = 1;

	return 0;
error:
	return rc;
}

/* skb->transport_header must point to BHTP header */
/* tun_skb->cb[] must point to a valid bhtp_qskb structure */
static inline int
bhtp_append_tun_skb(struct sk_buff *skb, struct sk_buff *tun_skb)
{
	/* reference: ip_append_page() and LNI c21, p504 */
	/* assert: BHTP header must be valid */

	int rc = 0;

	int ilen = 0;
	void *iptr = NULL;
	int i = 0;
	int j = 0;

	struct bhtphdr *bh = (struct bhtphdr *)skb->transport_header;
	struct bhtp_qskb *qskb = NULL;

	int appended = 0;
	int orig_len = skb->len;

	DEBUG("bhtp_append_tun_skb");

	if (tun_skb == NULL) {
		rc =  -EINVAL;
		goto error;
	}

	DEBUG("append_tun: tun-len (main) %d tun-data-len (frags) %d",
		tun_skb->len, tun_skb->data_len);

	qskb = bhtp_qsk(tun_skb);

	/* does it fit? */
	if (unlikely(skb_tailroom(skb) < tun_skb->len)) {
		ERR("append_tun: no space for %d bytes, only %d left",
			tun_skb->len, skb_tailroom(skb));
		rc = -ENOBUFS;
		goto error;
	}

	/* append the main skb buffer (no fragments) */
	ilen = skb_headlen(tun_skb);
	iptr = tun_skb->data;
	memcpy(skb_put(skb, ilen), iptr, ilen);
	appended += ilen;
	DEBUG("append_tun: copied headlen %d bytes total %d", ilen, appended);

	/* append the skb fragments */
	for (i = 0; i < skb_shinfo(tun_skb)->nr_frags; i++) {

		skb_frag_t *frag = &skb_shinfo(tun_skb)->frags[i];
		/* requires a hack to include kmap_skb.h */
		/* try to use skb_copy_bits() instead */
		u8 *vaddr =  kmap_skb_frag(frag);

		ilen = frag->size;
		iptr = vaddr + frag->page_offset;
		memcpy(skb_put(skb, ilen), iptr, ilen);
		appended += ilen;
		DEBUG("append_tun: copied frag(%d) %d bytes total %d",
				i, ilen, appended);
		kunmap_skb_frag(vaddr);
	}

	/* the tun_skb->len must agree with tun_lengths */
	/* this means that the padding is not part of the tun_skb->len */

	/* tun_len must always be valid */
	BUG_ON(qskb->tun_len != qskb->tun_hdr + qskb->tun_data);
	/* skb should be trimmed from padding */
	BUG_ON(tun_skb->len != qskb->tun_len);
	/* must append the whole skb */
	BUG_ON(appended != tun_skb->len);

	/* update the BHTP header with the new TUN */
	j = bhtp_reserve_tun_in_header(bh,
                                       qskb->tun->prot->proto,
                                       appended,
                                       qskb->tun_pad > 0);

	if (j < 0) {
		rc = j;
		goto restore_container_skb;
	}

	DEBUG("append_tun: appended TUN %d, proto %d, len %d, needs_padding %d",
		j, bh->tun[j].proto, bh->tun[j].len, bh->tun[j].needs_padding);

	return 0;

restore_container_skb:
	skb_trim(skb, orig_len);
error:
	return rc;
}


#ifdef BHTP_FRAGS
#error OUTDATED code from bhtp_append_tun
	int len = 0;
	int offset = 0;

	struct page *page = NULL;
	i = skb_shinfo(skb)->nr_frags;
	BUG_ON(i >= MAX_SKB_FRAGS);

	page = virt_to_page(tun);
	offset = offset_in_page(tun);
	len = seglen;
	/* cannot span more pages yet FIXME */
	BUG_ON(offset + len > PAGE_SIZE);

	get_page(page);
	skb_fill_page_desc(skb, i, page, offset, len);
	skb->len += len;
	skb->data_len += len;
	skb->truesize += len;

	DEBUG("fill_page_desc offset %d len %d", offset, len);

	j = bh->tuns;
	BUG_ON(j >= BHTP_MAX_SEG);
	DEBUG("appended tun skb frags %d pos %d proto %d len %d",
		skb_shinfo(skb)->nr_frags, j, tun_proto, seglen);
#endif

static inline void
mgrp_counters_update_recv_side(struct mgrp_counters *counters,
                               struct bhtphdr *bh, int pktlen)
{
        counters->recv.mgrp_pkts++;

        if (counters->recv.valid_expected) {
                mgrp_seqno_t cur = bh->seqno;
                mgrp_seqno_t expected = counters->recv.mgrp_pkt_expected_seqno;

                // some packets in the middle are missing
                if (mgrp_seqno_after(cur, expected)) {
                        counters->recv.mgrp_pkts_missing +=
                                mgrp_seqno_distance(cur, expected);

                // we got a previously missing packet
                } else if (mgrp_seqno_before(cur, expected)) {
                        counters->recv.mgrp_pkts_outoforder++;
                }
        }

        counters->recv.mgrp_pkts_lost =
                counters->recv.mgrp_pkts_missing -
                counters->recv.mgrp_pkts_outoforder;

        // the expected_seqno increases always monotonically
        // but may jump over missing packets
        counters->recv.mgrp_pkt_expected_seqno = mgrp_seqno_inc(bh->seqno);
        counters->recv.valid_expected = 1;
}

static inline void
mgrp_counters_update_send_side(struct mgrp_counters *counters,
                               struct bhtphdr *bh, int pktlen)
{
        int t = 0;
        int tuns = bhtp_header_tuns(bh);

        int probe_pkts = 0;
        int probe_nonpad_bytes = 0;
        int nonprobe_bytes = 0;

        counters->mgrp_pkts++;
        counters->mgrp_bytes_header += sizeof(struct bhtphdr);

        for (t = 0; t < tuns; t++) {
                struct bhtphdr_tun *tun = &bh->tun[t];

                // padding is no longer a TUN, must include later
                counters->mgrp_bytes_payload += tun->len;

                if (tun->proto == BHTP_PROTO_TCP) {
                        counters->tcp_pkts++;
                        counters->tcp_bytes += tun->len;
                        nonprobe_bytes += tun->len;

                } else if (tun->proto == BHTP_PROTO_UDP && !tun->needs_padding) {
                        counters->udp_pkts++;
                        counters->udp_bytes += tun->len;
                        nonprobe_bytes += tun->len;

                } else if (tun->proto == BHTP_PROTO_UDP && tun->needs_padding) {
                        // defer updating probe stats until all TUNs visited
                        probe_pkts++;
                        probe_nonpad_bytes += tun->len;

                } else {
                        ERR_TR("error: unknown proto %d", tun->proto);
                }
        }

        if (probe_pkts == 1) {
                int pad_reused = 0;
                int pad_unused = 0;
                char error[200];
                int pad_orig = bhtp_orig_pad_bytes(bh, pktlen,
                                                   &pad_reused, &pad_unused,
                                                   error, sizeof(error));

                if (pad_orig == -1) {
                        ERR_TR("error: invalid header: %s", error);
                        goto done;
                }

                counters->probe_pkts++;
                if (pad_reused > 0)
                        counters->probe_pkts_reused++;

                // include the padding in the payload
                counters->mgrp_bytes_payload += pad_unused;

                counters->probe_bytes_actual += probe_nonpad_bytes + pad_unused;
                counters->probe_bytes_orig += probe_nonpad_bytes + pad_orig;

                counters->probe_bytes_nonpad += probe_nonpad_bytes;
                counters->probe_bytes_pad_unused += pad_unused;
                counters->probe_bytes_pad_reused += pad_orig - pad_unused;

        } else if (probe_pkts > 1) {
                ERR_TR("error: expecting at most 1 probe but got %d",
                       probe_pkts);
        }
done:
        return;
}

/* SKB is headerless - no BHTP header yet */
/* may return NET_XMIT_CN which is not an error */
/*
 * On success, tun_skb is freed.
 * On failure, tun_skb is not freed and a return code rc returned.
 *
 * Called with bflow->tosend_lock held (XXX please verify).
 */
static inline int
bhtp_transmit_skb(struct bhtp_flow *bflow,
			int append_padbytes, 
			struct sk_buff *tskbs[], int tuns)
{
	int rc = 0;
	int i = 0;
	struct bhtphdr *bh = NULL;
	struct dst_entry *dst = NULL;
	struct sk_buff *skb =NULL;
	
	BUG_ON(bflow == NULL);

	/*
	 * STEP: create the container skb
	 *
	 * We need to create a BHTP skb that holds the header
	 * and just uses the buffer from the tun_skb.
	 */

#ifdef BHTP_FRAGS
	skb = bhtp_alloc_container_skb(MAX_BHTP_HEADER);
#else
	skb = bhtp_alloc_container_skb(MAX_BHTP_HEADER + 2000);
#endif
	if (unlikely(skb == NULL)) {
		ERR("xmit no buffers");
		rc = -ENOBUFS;
		goto error;
	}

	skb_reserve(skb, MAX_BHTP_HEADER);
	skb->transport_header = skb_push(skb, sizeof(struct bhtphdr));
	bh = (struct bhtphdr *)skb->transport_header;

	/* do we need sock_hold here? if yes, who does the sock_put? */
	/* ip_queue_xmit expects skb->sk->.. */
	skb->sk = bflow->sk;

	/* Build Header */
	memset(bh, 0, sizeof(struct bhtphdr));
	/* seqno is u64, larger than the int on 32-bit machines, FIXME */
	bh->seqno = atomic_add_return(1, &bhtp.seqno);
	/* timestamp is filled later, just before we hand to IP layer */


	/*
	 * STEP: combine multiple TUN skbs in the container skb
	 *
	 * Then you need to be able to combine multiple
	 * skb buffers into this one (for reuse).
	 */
	 for (i = 0; i < tuns; i++) {
	 	rc = bhtp_append_tun_skb(skb, tskbs[i]);
		if (rc) {
			ERR("xmit_skb: failed to append_tun_skb");
			goto free_container_skb;
		}
	 }
	
	/*
	 * STEP: add the padding
	 */
	 if (append_padbytes > 0) {
	 	rc = bhtp_append_tun_pad(skb, append_padbytes);
		if (rc) {
			ERR("xmit_skb: failed to append %d padbytes",
                            append_padbytes);
			goto free_container_skb;
		}
	 }

	/* HACK: make sure packet length is below MSS */
	/*	 1480 = 1500 - IP(20) */
	if (skb->len > 1480) {
		ERR("xmit: skb too large: %d", skb->len);
		rc = -EINVAL;
		goto free_container_skb;
	}

	/*
	 * STEP: find packet destination and hand over to IP
	 *
	 * The destination is a per-flow attribute. All the TUNs
	 * that were appended go to the same dest.
	 */

	/* check if the cached destination is valid */
	dst = bhtp_dst_check(bflow);

	/* lookup the destination */
	if (!dst) {
		struct sockaddr_in dst_addr = {
			.sin_addr.s_addr = bflow->daddr,
		};
		dst = bhtp_get_dst(&dst_addr, NULL);
		if (!dst) {
			ERR("transmit_skb: aborted: no route to dst " NIPQUAD_FMT,
				NIPQUAD(bflow->daddr));
			/* FIXME: correct error? */
			rc = -ENETUNREACH;
			goto free_container_skb;
		}
		bflow->dst = dst_clone(dst);
		DEBUG("OK, routed dst " NIPQUAD_FMT, NIPQUAD(bflow->daddr));
	}

	/*
	 * STEP: free the TUN skbs
	 *
	 * No need for them anymore since they are part of the
	 * container skb now. Wait until this late to ensure that
	 * no errors occured.
	 */
	 for (i = 0; i < tuns; i++)
	 	kfree_skb(tskbs[i]);

	/* FIXME: does dst need to be cloned earlier? */
	skb->dst = dst_clone(bflow->dst);

	DEBUG("xmit: ip_queue_xmit");
	bhtp_header_fill_sndstamp(bh, bflow->sndstamp_base_ns);
        mgrp_counters_update_send_side(&bflow->counters, bh, skb->len);
        if (debug_on()) {
                int loglen = min_t(int, skb->len, 48);
                log_hex((const char *)bh, loglen, 4, "snd: ");
        }
	rc = ip_queue_xmit(skb, 0);

	DEBUG("    ip_queue_xmit done rc %d", rc);
	if (rc < 0) {
		ERR("xmit: ip_queue_xmit failed rc %d", rc);
		/* ip_queue_xmit took care of the kfree_skb() */
		goto error;
	}

	return rc;

free_container_skb:
	kfree_skb(skb);
	ERR("xmit: container packet dropped");
error:
	return rc;

}

/*
 *	Fragments an skb and transmits its fragments.
 *
 *	@bflow		BHTP flow
 *	@tskb		TUN skb with upper transport (fe TCP) header
 *	@max_frag_len	No fragment will be larger than this (hdr + data)
 *
 *	This function is generic and is protocol agonstic. It assumes
 *	that the contained protocol (TUN) provides fragment operations
 *	(ops->fragment and ops->fraglen_left). For length bookkeeping
 *	(see length fields of bhtp_qskb) BHTP assumes that the
 *	protocol packet is comprised of a protocol header and data.
 *	When the packet is fragmented, data is split among packets
 *	and a new header is generated for each fragment. BHTP does
 *	not care how this is done.
 *
 *	The function does not make any implicit assumptions about
 *	the nature of the transport protocol (fe, byte-oriented vs
 *	message-oriented). TCP supports fragmentation naturally while
 *	UDP does not. However, another datagram transport protocol
 *	could explicitly support fragmentation.
 */
static inline int
bhtp_fragment_transmit_skb(struct bhtp_flow *bflow, struct sk_buff *tskb,
			   int max_frag_len)
{
	int rc = 0;
	int left = 0;
	int frag_count = 0;
	
	struct bhtp_qskb *qskb = bhtp_qsk(tskb);
	struct bhtp_tun_ops *ops = qskb->tun->prot->ops;

        enum mgrp_consumer_type type = MGRP_CONSUMER_TRANSPORT;

	LOG_TCPHDR_SKB(tskb);

	DEBUG("frag_xmit: fragmenting tun skb len %d max_frag_len %d",
		tskb->len, max_frag_len);

	if (max_frag_len <= 0) {
		ERR("frag_xmit: max_frag_len %d too small", max_frag_len);
		rc = -EINVAL;
		goto error;
	}

	while ((left = ops->fraglen_left(tskb, type)) > 0) {

		int frag_len = (left > max_frag_len ? max_frag_len : left);
		struct sk_buff *frag = ops->fragment(tskb, frag_len, type);

		/* FIXME: optimize when no fragment is needed */
		/* DONE: never comes in here if does not need fragmentation */
		/* see bhtp_transmit_lone_skb */

		if (frag == NULL) {
			ERR("frag_xmit: failed to create fragment %d from skb len %d",
				frag_len, tskb->len);
			rc = -EINVAL;
			goto error;
		}

		LOG_TCPHDR_SKB(frag);

		frag_count += 1;

		/* protect against runaway code */
		BUG_ON(frag_count > 100);

		DEBUG("frag_xmit: created fragnent %d len %d, transmitting",
				frag_count, frag->len);

		rc = bhtp_transmit_skb(bflow, 0, &frag, 1);
		if (rc < 0) {
			ERR("frag_xmit: failed to send fragment %d len %d",
				frag_count, frag->len);
			goto error;
		}
	}

	kfree_skb(tskb);

error:
	return rc;
}

/* TUN skb fragmentation
 *
 * When a TUN skb is fragmented, skb->len never changes.
 * All the information is kept in qskb->tun_data_frag.
 *
 * For example, if the skb has already been fragmented
 * but is not further fragmentable:
 *
 *	bhtp_skb_already_fragmented(skb) &&
 *	bhtp_skb_can_fragment_further(skb)
 *
 * you cannot just send the skb, but you have to create the last
 * fragment (because skb->len is the whole unfragmented skb).
 */

/*
 * Does the skb belong to a TUN that allows fragmentation?
 * The skb may not be fragmentable further.
 */
static inline int
bhtp_skb_is_fragmentable(struct sk_buff *skb,
                         enum mgrp_consumer_type type)
{
	struct bhtp_qskb *qskb = bhtp_qsk(skb);

        // has padding
        if (qskb->tun_pad > 0)
                return 0;

        // fragmentation disabled
        if (qskb->tun_frag_len_max == 0)
                return 0;

        // no fragment callback
        if (!BHTP_TUN_PROT_HAS_FRAG(qskb->tun->prot))
                return 0;

        // consumer type does not play a role
	return 1;
}

/* Has the skb already given up a fragment? */
static inline int
bhtp_skb_already_fragmented(struct sk_buff *skb,
                            enum mgrp_consumer_type type)
{
	struct bhtp_qskb *qskb = bhtp_qsk(skb);
        // pick the right tun_data_frag counter
	return (qskb->consumer[type].tun_data_frag > 0);
}

/* Can the skb give up more fragments? */
static inline int
bhtp_skb_can_fragment_further(struct sk_buff *skb, int frag_len,
                              enum mgrp_consumer_type type)
{
        // XXX HERE consumer type should play a role; not yet

	struct bhtp_qskb *qskb = bhtp_qsk(skb);
	int can_fragment =
		frag_len <= qskb->tun_frag_len_max &&
		frag_len >= 0;

	return can_fragment;
}

/* Is the skb too big to send as is? */
static inline int
bhtp_skb_must_fragment_further(struct sk_buff *skb,
                               enum mgrp_consumer_type type)
{
	struct bhtp_qskb *qskb = bhtp_qsk(skb);
	int fraglen_left = qskb->tun->prot->ops->fraglen_left(skb, type);
	int must_fragment = fraglen_left > qskb->tun_frag_len_max;

        BUG_ON(type != MGRP_CONSUMER_TRANSPORT);

	return must_fragment;
}

/* Does the skb need to be transmitted using bhtp_skb_fragment_transmit()?
 *
 * This does not necessarily mean that we are breaking up the
 * skb further. It just means that we have to use
 * bhtp_fragment_transmit_skb() and not bhtp_transmit_skb().
 * (for example when sending the last fragment).
 */
static inline int
bhtp_skb_needs_fragment_transmit(struct sk_buff *skb)
{
        enum mgrp_consumer_type type = MGRP_CONSUMER_TRANSPORT;

	if (!bhtp_skb_is_fragmentable(skb, type))
		return 0;

	if (bhtp_skb_already_fragmented(skb, type))
		return 1;
	
	if (bhtp_skb_must_fragment_further(skb, type))
		return 1;
	
	return 0;
}

static inline int
bhtp_fraglen_left(struct sk_buff *skb,
                  enum mgrp_consumer_type type)
{
	int left = skb->len;

	if (bhtp_skb_already_fragmented(skb, type))
		left = bhtp_qsk(skb)->tun->prot->ops->fraglen_left(skb, type);
	
	return left;
}


/* Queueing functions for bflow->tosend queue
 *
 * Versions with preceeding underscores __ need to be protected by
 * bflow->tosend_lock.
 */

static inline int
__bhtp_enqueue_skb(struct bhtp_flow *bflow, struct sk_buff *tskb,
		ktime_t latency)
{
	int rc = 0;

	skb_queue_tail(&bflow->tosend, tskb);
	bflow->tosend_accumulated += tskb->len;
	rc = bhtp_enqueue_req(bflow, tskb, latency);
	if (rc)
		goto rollback;

	DEBUG("enqueue_skb: added %d acc %lld",
		tskb->len,
		bflow->tosend_accumulated);

	DEBUG("skb QUEUED %p len %d accumulated %lld rc %d elapsed %lld",
		tskb, tskb->len,
		bflow->tosend_accumulated, rc,
		ktime_to_ns(ktime_sub(tick_now(), bhtp_qsk(tskb)->kt)));

	return 0;

rollback:
	bflow->tosend_accumulated -= tskb->len;
	skb_dequeue_tail(&bflow->tosend);
	return rc;
	
}

static inline int
bhtp_enqueue_skb(struct bhtp_flow *bflow, struct sk_buff *tskb,
		ktime_t latency)
{
	int rc = 0;
	spin_lock_bh(&bflow->tosend_lock);
	rc = __bhtp_enqueue_skb(bflow, tskb, latency);
	spin_unlock_bh(&bflow->tosend_lock);
	return rc;
}


static inline struct sk_buff *
__bhtp_dequeue_skb_only(struct bhtp_flow *bflow, struct sk_buff *this_only)
{
	struct sk_buff *skb = NULL;
	int removed = 0;

	skb = skb_dequeue(&bflow->tosend);
	if (skb == NULL)
		goto done;

	if (this_only && skb != this_only)
		goto put_back;

	removed = bhtp_fraglen_left(skb, MGRP_CONSUMER_TRANSPORT);
	bflow->tosend_accumulated -= removed;
	DEBUG("dequeue_skb: removed %d acc %lld",
		removed,
		bflow->tosend_accumulated);
	if (bflow->tosend_accumulated < 0)
		ERR("*** bhtp_dequeue_skb: negative accumulation %lld len %d",
			bflow->tosend_accumulated, skb->len);
done:
	return skb;

put_back:
	ERR("dequeue_skb_only: asking for %p %lld got %p %lld accumulated %lld",
		this_only, bhtp_qsk(this_only)->send.req->idg,
		skb, bhtp_qsk(skb)->send.req->idg,
		bflow->tosend_accumulated);
	skb_queue_head(&bflow->tosend, skb);
	skb = NULL;
	goto done;
}

static inline struct sk_buff *
bhtp_dequeue_skb_only(struct bhtp_flow *bflow, struct sk_buff *this_only)
{
	struct sk_buff *skb = NULL;

	spin_lock_bh(&bflow->tosend_lock);
	skb = __bhtp_dequeue_skb_only(bflow, NULL);
	spin_unlock_bh(&bflow->tosend_lock);

	return skb;
}

static inline struct sk_buff *
__bhtp_dequeue_skb(struct bhtp_flow *bflow)
{
	return __bhtp_dequeue_skb_only(bflow, NULL);
}

static inline struct sk_buff *
bhtp_dequeue_skb(struct bhtp_flow *bflow)
{
	struct sk_buff *skb = NULL;

	spin_lock_bh(&bflow->tosend_lock);
	skb = __bhtp_dequeue_skb(bflow);
	spin_unlock_bh(&bflow->tosend_lock);

	return skb;
}

static inline void
__bhtp_requeue_skb(struct bhtp_flow *bflow, struct sk_buff *skb)
{
	int added = bhtp_fraglen_left(skb, MGRP_CONSUMER_TRANSPORT);
	skb_queue_head(&bflow->tosend, skb);
	bflow->tosend_accumulated += added;
	DEBUG("requeue_skb: added %d acc %lld skblen %d",
		added,
		bflow->tosend_accumulated,
		skb->len);
}


static inline int
bhtp_transmit_lone_skb(struct bhtp_flow *bflow, struct sk_buff *tun_skb)
{
	int rc = 0;
	struct bhtp_qskb *qskb = bhtp_qsk(tun_skb);
	int append_padbytes = qskb->tun_pad;

	DEBUG("schedule:    frag_max %d len %d pad %d ops %d",
			qskb->tun_frag_len_max,
			tun_skb->len,
			append_padbytes,
			BHTP_TUN_PROT_HAS_FRAG(qskb->tun->prot));

	if (bhtp_skb_needs_fragment_transmit(tun_skb))
		rc = bhtp_fragment_transmit_skb(bflow, tun_skb,
				qskb->tun_frag_len_max);
	else
		rc = bhtp_transmit_skb(bflow, append_padbytes, &tun_skb, 1);

	return rc;
}

// Returns 1 if the @skb can be a rider inside a probe.
//
// This means that we can bite out a chunk of up to @aval_pad bytes
// from the skb. The skb contains one counter for each consumer type.

static inline int
mgrp_skb_can_be_rider(struct sk_buff *skb, int avail_pad,
                      enum mgrp_consumer_type type)

{
        // assumes that the skb is willing to be a rider

        // rejected: rider needs padding
        if (bhtp_qsk(skb)->tun_pad > 0)
                return 0;

         // rider not fragmentable: accept only if it fits inside padding
         if (!bhtp_skb_is_fragmentable(skb, type))
                return (skb->len <= avail_pad);

        /* rider is fragmentable from this point on */

        /* rejected: we cannot fragment rider further to fit in padding */
        if (!bhtp_skb_can_fragment_further(skb, avail_pad, type))
                return 0;

        /* accepted: skb may fit whole or may need fragmenting */
        return 1;
}

static inline struct sk_buff *
bhtp_find_rider_skb(struct bhtp_flow *bflow,
                    struct sk_buff *vessel, int *padding)
{
	struct sk_buff *rider = NULL;
	struct sk_buff *skb = NULL;
	int pad = bhtp_qsk(vessel)->tun_pad;
	int rider_left = 0;
	int fraglen = 0;

        enum mgrp_consumer_type consumer = MGRP_CONSUMER_PBK_REDUNDANT;

        // hey, we are in a presidential campaign :) Go Obama!
        int candidate_vetted = 0;
        struct sk_buff *candidate = NULL;
        
	spin_lock_bh(&bflow->tosend_lock);

        // traverse @bflow->tosend skb queue to find skb that has
        // a fragment left for piggybacking. As we piggyback on top
        // of a packet train, the initial part of the skb queue will
        // get used up. A little inefficient, but we have to do this
        // every time because TCP may send out the skbs in the meantime
        //
        // see code in skb_peek
        // be very careful
        // MUST be protected by @bflow->tosend_lock
        candidate = ((struct sk_buff *)&bflow->tosend)->next;
        while (candidate != (struct sk_buff *)&bflow->tosend) {

                // stop once you hit an skb that cannot be a rider
                // cannot go past that skb to look for candidates
                if (!mgrp_skb_can_be_rider(candidate, pad, consumer))
                        break;

                if (bhtp_fraglen_left(candidate, consumer) > 0) {
                        candidate_vetted = 1;
                        break;
                }

                // XXX question?
                // if the skb contains only a header (such as an ack)
                // then bhtp_fraglen_left always returns non-zero
                //
                // How do we differentiate the case were it has already
                // been piggybacked?
        }

        if (!candidate_vetted)
                goto done;

        rider_left = bhtp_fraglen_left(candidate, consumer);
        fraglen = min_t(int, rider_left, pad);

	rider =  bhtp_qsk(skb)->tun->prot->ops->fragment(skb, fraglen, consumer);
	if (rider) {
		BUG_TRAP(fraglen == rider->len);
		pad -= fraglen;
	}

        // even if we exhausted the skb, it is up to the TCP request
        // to remove it from the queue

	DEBUG("find_rider_skb: req %lld rider_left %d fraglen %d pad %d",
		req->idg, rider_left, fraglen, pad);
done:
	BUG_TRAP(pad >= 0);
	*padding = pad;
	spin_unlock_bh(&bflow->tosend_lock);

	return rider;
}

#if 0
static inline struct sk_buff *
__OLD__bhtp_find_rider_skb(struct bhtp_flow *bflow,
		struct sk_buff *vessel, int *padding)
{
	struct sk_buff *rider = NULL;
	struct sk_buff *skb = NULL;
	int pad = bhtp_qsk(vessel)->tun_pad;
	int rider_left = 0;
	int fraglen = 0;
	struct bhtp_flow_req *req = NULL;
	int sent = 0;

	spin_lock_bh(&bflow->tosend_lock);

	skb = __bhtp_dequeue_skb(bflow);
	if (!skb)
		goto done;
	
	req = bhtp_qsk(skb)->send.req;

	/* is the skb suitable to be a rider? */

	/* rejected: rider needs padding */
	if (bhtp_qsk(skb)->tun_pad > 0)
		goto done_requeue;

	 if (!bhtp_skb_is_fragmentable(skb)) {
		/* accepted: rider not fragmentable but fits inside padding */
		if (skb->len <= pad)
			goto use_asis;

		/* rejected: rider does not fit */
		goto done_requeue;
	}

	/* accepted: no need to start fragmenting if it fits whole */
	if (!bhtp_skb_already_fragmented(skb) && skb->len <= pad)
		goto use_asis;
	
	/* rejected: we cannot fragment rider to fit in padding */
	if (!bhtp_skb_can_fragment_further(skb, pad))
		goto done_requeue;
	
	/* accepted: need to fragment packet to fit in padding */
	
	rider_left = bhtp_fraglen_left(skb);
	fraglen = min_t(int, rider_left, pad);

	rider =  bhtp_qsk(skb)->tun->prot->ops->fragment(skb, fraglen);
	if (rider) {
		BUG_TRAP(fraglen == rider->len);
		pad -= fraglen;
	}

	/* does the skb still have fragments? */
	if (bhtp_fraglen_left(skb) > 0)
		goto done_requeue;

	DEBUG("find_rider_skb: req %lld discard skb %p %d",
		req->idg, skb, skb->len);
	DEBUG("find_rider_skb: req %lld rider_left %d fraglen %d pad %d",
		req->idg, rider_left, fraglen, pad);

	/* completely used up skb */
	kfree_skb(skb);
	goto done_used;

use_asis:
	rider = skb;
	pad -= skb->len;
	DEBUG("find_rider_skb: req %lld skb %p use_asis", req->idg, req->skb);
	goto done_used;

done_requeue:
	DEBUG("find_rider_skb: req %lld skb %p done_requeue", req->idg, req->skb);
	__bhtp_requeue_skb(bflow, skb);
	goto done;

done_used:
	/* This is ok, even if done after skb is freed or used.
	 * We are holding the bflow->tosend_lock which also protects
	 * bhtp_dequeue_skb_only() which is used to send packets.
	 */
	sent = atomic_inc_return(&req->sent);
	DEBUG("find_rider_skb: req %lld skb %p sent %d done_used",
		req->idg, req->skb, sent);
	if (sent != 1)
		ERR("find_rider_skb: req %lld expects req->sent to be 1, got %d",
			req->idg, sent);

done:
	BUG_TRAP(pad >= 0);
	*padding = pad;
	if (req)
		DEBUG("find_rider_skb: req %lld skb %p rider %p %d pad %d done",
			req->idg, req->skb, rider, (rider ? rider->len : 0), pad);
	else
		DEBUG("find_rider_skb: req <none> skb <none> rider %p %d pad %d done",
			rider, (rider ? rider->len : 0), pad);
	spin_unlock_bh(&bflow->tosend_lock);

	return rider;
}
#endif


static inline int
bhtp_transmit_vessel_skb(struct bhtp_flow *bflow, struct sk_buff *vessel)
{
	int rc = 0;
	int append_padbytes = 0;
	struct sk_buff *rider = NULL;

	int tuns = 0;
	struct sk_buff *tskbs[2];

	/* the vessel must come first */
	/* quirk of the implementation */
	/* see udpex_tun_rcv() /left in rcv skb/ */
	tskbs[tuns] = vessel;
	tuns++;

        // The vessel skb has some padding that can be reused to
        // accomadate a rider skb. The @append_padbytes is the portion
        // of the padding that was not reused (unused) and needs to
        // be appended to the MGRP packet.
	rider = bhtp_find_rider_skb(bflow, vessel, &append_padbytes);

	if (rider) {
		tskbs[tuns] = rider;
		tuns++;
	}

	rc = bhtp_transmit_skb(bflow, append_padbytes, tskbs, tuns);
	if (rc < 0)
		/* the vessel is not dropped; handled by the caller */
		goto drop_rider;
	
	return rc;

drop_rider:
	ERR("xmit_vessel_skb: DROPPED rider skb %d", rider->len);
	kfree_skb(rider);
	return rc;
}


static inline void
bhtp_send_req(struct bhtp_flow_req *req)
{
	int rc = 0;
	int sent = 0;
	struct sk_buff *skb = NULL;

	spin_lock_bh(&req->bflow->tosend_lock);

	sent = atomic_inc_return(&req->sent);
	if (sent != 1) {
		/* probably ok: skb swept under our feet */
		/* either by a probe or by draining the queue */
		DEBUG("send_req: req %lld already sent %d skb %p ",
			req->idg, sent, req->skb);
		goto done;
	}

	skb = __bhtp_dequeue_skb_only(req->bflow, req->skb);
	if (!skb) {
		/* not ok: front of queue does not have skb that we expect */
		ERRBUG("send_req: req %lld for missing skb %p sent %d",
			req->idg, req->skb, atomic_read(&req->sent));
		goto done;
	}
	
	DEBUG("send_req: sending req %lld skb %p sent %d",
		req->idg, req->skb, atomic_read(&req->sent));

	rc = bhtp_transmit_lone_skb(req->bflow, req->skb);
	if (rc < 0) {
		ERR("send_req: req %lld dropping skb %p %d rc %d",
			req->idg, req->skb, req->skb->len, rc);
		kfree_skb(skb);
	} else if (rc > 0) {
		/* FIXME */
		ERR("send_req: req %lld skb %p %d rc %d when transmitting, how to handle?",
			req->idg, req->skb, req->skb->len, rc);
	}

done:
	spin_unlock_bh(&req->bflow->tosend_lock);
	return;
}

static inline ktime_t
bhtp_req_remaining(struct bhtp_flow_req *req)
{
	ktime_t now_kt = tick_now();

	if (ktime_to_ns(req->send_kt) > ktime_to_ns(now_kt))
		return ktime_sub(req->send_kt, now_kt);
	else
		return ktime_set(0, 0);
}

static inline unsigned long
bhtp_req_remaining_jiffies(struct bhtp_flow_req *req)
{
	unsigned long long nsec = ktime_to_ns(bhtp_req_remaining(req));
	unsigned long usec = 0;
	unsigned long j = 0;

	/* approximation gives us less usecs: it is ok */
	usec = nsec;
	usec /= 1000;
	j = usecs_to_jiffies(usec);

	DEBUG("remaining nsec %lld usec %ld jiffies %ld",
		nsec, usec, j);

	return j;
}

static inline int
bhtp_req_timer_reschedule(struct bhtp_flow_req *req)
{
	unsigned long j = jiffies;
	unsigned long interval = 0;

	if (timer_pending(&bhtp.req_timer))
		return 0;
	
	interval = bhtp_req_remaining_jiffies(req);
	bhtp.req_timer.expires = j + interval;
	add_timer(&bhtp.req_timer);
	DEBUG("reschedule timer in jiffies %ld msec %d usec %d",
		interval,
		jiffies_to_msecs(interval), 
		jiffies_to_usecs(interval));

	return 1;
}

static inline int
bhtp_enqueue_req(struct bhtp_flow *bflow, struct sk_buff *skb, ktime_t latency)
{
	int rc = 0;
	int must_schedule = 0;
	struct bhtp_flow_req *req = NULL;
	struct bhtp_qskb *qskb = bhtp_qsk(skb);
	
	req = kzalloc(sizeof(struct bhtp_flow_req), GFP_ATOMIC);
	if (!req) {
		ERR("enqueue_req: out of mem");
		rc = -ENOMEM;
		goto done;
	}
	/* backpointer to request inside the skb */
	qskb->send.req = req;

	req->skb = skb;
	req->bflow = bflow;
	req->idg = atomic_inc_return(&bhtp.reqid);
	req->idf = atomic_inc_return(&bflow->reqid);
	req->latency = latency;
	req->send_kt = ktime_add(qskb->kt, latency);

	spin_lock_bh(&bhtp.req_lock);
	must_schedule = list_empty(&bhtp.req_list);
	list_add_tail(&req->list, &bhtp.req_list);
	spin_unlock_bh(&bhtp.req_lock);

	atomic_inc(&bhtp.req_waiting);

	if (must_schedule)
		bhtp_req_timer_reschedule(req);
done:
	return rc;
}

static inline void
bhtp_requeue_req(struct bhtp_flow_req *req)
{
	spin_lock_bh(&bhtp.req_lock);
	list_add(&req->list, &bhtp.req_list);
	spin_unlock_bh(&bhtp.req_lock);

	atomic_inc(&bhtp.req_waiting);
}

static inline struct bhtp_flow_req *
bhtp_dequeue_req(void)
{
	struct bhtp_flow_req *req = NULL;


	spin_lock_bh(&bhtp.req_lock);
	if (!list_empty(&bhtp.req_list)) {
		req = list_entry(bhtp.req_list.next,
			struct bhtp_flow_req, list);
		list_del(&req->list);
	}
	spin_unlock_bh(&bhtp.req_lock);

	if (req)
		atomic_dec(&bhtp.req_waiting);

	return req;
}


void
bhtp_drain_req(void)
{
	struct bhtp_flow_req *req = NULL;
	ktime_t remaining = ktime_set(0, 0);

	while ((req = bhtp_dequeue_req()) != NULL) {
		remaining = bhtp_req_remaining(req);
		if (ktime_to_ns(remaining) > TICK_NSEC)
			goto send_later;
		bhtp_send_req(req);
		kfree(req);
	}

	return;

send_later:
	/* reschedule the timer */
	bhtp_requeue_req(req);
	bhtp_req_timer_reschedule(req);
	return;
}

void
bhtp_drop_req(void)
{
	struct list_head *l = NULL;
	struct list_head *tmp = NULL;
	struct bhtp_flow_req *req = NULL;

	spin_lock_bh(&bhtp.req_lock);
	list_for_each_safe(l, tmp, &bhtp.req_list) {
		req = list_entry(l, struct bhtp_flow_req, list);
		list_del(&req->list);
		atomic_dec(&bhtp.req_waiting);
		LOG("drop_req: DROPPING request, waiting %d",
			atomic_read(&bhtp.req_waiting));
		kfree(req);
	}
	spin_unlock_bh(&bhtp.req_lock);
}

void
bhtp_req_timer(unsigned long arg)
{
	bhtp_drain_req();
}

ktime_t
bhtp_process_req(void)
{
	struct bhtp_flow_req *req = NULL;
	ktime_t delay = ktime_set(0, 0);

	while ((req = bhtp_dequeue_req()) != NULL) {
		delay = bhtp_req_remaining(req);
		while (ktime_to_ns(delay) > NSEC_PER_MSEC) {
			LOG("SLEEP %lld", ktime_to_ns(delay));
			msleep(1);
		}
/* 		if (ktime_to_ns(delay) > 1) { */
/* 			LOG("ndelay %lld", ktime_to_ns(delay)); */
/* 			ndelay(ktime_to_ns(delay)); */
/* 		} */
		bhtp_send_req(req);
		kfree(req);
	}

	return ktime_set(0, 0);

/* put_back: */
/* 	bhtp_requeue_req(req); */
/* 	return delay; */
}

static inline ktime_t
mgrp_flow_latency(struct bhtp_flow *bflow)
{
	ktime_t global_latency = mgrp_global_flow_latency(bflow);
	ktime_t local_latency = mgrp_local_flow_latency(bflow);

	/* only one latency should be defined */
	if (ktime_to_ns(global_latency) > 0) {
		if (ktime_to_ns(local_latency) > 0)
			ERR("flow_latency: "
			    "both latencies defined: global %lld local %lld ns",
			    ktime_to_ns(global_latency),
			    ktime_to_ns(local_latency));
		return global_latency;
	}

	return local_latency;
}


/* SKB is headerless - no BHTP header yet */
/* may return NET_XMIT_CN which is not an error */
/* may return BHTP_QUEUED  which is not an error */
/*
 * >0	tun_skb is queued at the BHTP layer, tun_skb is not freed
 * 0	tun_skb transmitted (passed to IP) and tun_skb is freed.
 * <0	tun_skb is not freed and a return code rc returned.
 */
int
bhtp_schedule_skb(struct bhtp_flow *bflow, struct sk_buff *tun_skb)
{
	int rc = 0;
	struct bhtp_qskb *qskb = bhtp_qsk(tun_skb);
	int is_vessel = qskb->type & BHTP_QSKB_VESSEL;
	ktime_t latency = mgrp_flow_latency(bflow);

	int send_now = ktime_to_ns(latency) == 0 ||
			is_vessel ||
			qskb->type & BHTP_QSKB_NEVER_QUEUE;

	LOG_TCPHDR_SKB(tun_skb);

	qskb->kt = tick_now();

/* 	XXX HERE */
/* 	When probe is waiting, set the flow_latency */
/* 	when the probe is sent (nothing waiting) */
/* 		if bhtp.latency non zero, empty the payload queue */
/* 		set the flow_latency to zero */

	if (!send_now)
		rc = bhtp_enqueue_skb(bflow, tun_skb, latency);
	else if (is_vessel)
		rc = bhtp_transmit_vessel_skb(bflow, tun_skb);
	else
		rc = bhtp_transmit_lone_skb(bflow, tun_skb);

	return rc;
}


int
bhtp_drain_tosend(struct bhtp_flow *bflow, int drop)
{
	int skb_count = 0;
	struct sk_buff *skb;

	spin_lock_bh(&bflow->tosend_lock);

	if (skb_queue_empty(&bflow->tosend))
		goto done;

	while ((skb = __bhtp_dequeue_skb(bflow)) != NULL) {
		int rc = 0;
		int sent = 0;
		struct bhtp_flow_req *req = bhtp_qsk(skb)->send.req;

		skb_count++;
		
		if (drop) {
			ERR("drain_tosend: dropping skb %d", skb->len);
			kfree_skb(skb);
			continue;
		}

		/* mark request as "sent"; see bhtp_send_req() */
		sent = atomic_inc_return(&req->sent);
		if (sent != 1) {
			/* not ok: already sent out: must be bug */
			ERRBUG("drain_tosend: req %lld already sent %d skb %p",
				req->idg, sent, req->skb);
			continue;
		}

		rc = bhtp_transmit_lone_skb(bflow, skb);
		if (rc) {
			ERR("drain_tosend: dropping skb %d rc %d",
				skb->len, rc);
			kfree_skb(skb);
			continue;
		}
	}

done:
	spin_unlock_bh(&bflow->tosend_lock);
	return skb_count;
}

int
bhtp_send_barrier_skbs(struct bhtp_flow *bflow,
		struct sock *sk, struct bhtp_tun *tun,
		unsigned long long *twait_p,
		unsigned long long *tsend_p)

{
	int rc = 0;

	unsigned long long started = tick_now_ns();
	unsigned long long twait = 0;
	unsigned long long tsend = 0;
	unsigned long long canwait = 0;

	int keep_waiting = 0;
	int scheduled = 0;
	int not_worthit =
		tun->barrier_total < tun->barrier_pkts * 400 ||
		tun->barrier_max < 400;

	canwait = tun->barrier_wait_usec;
	canwait *= 1000;

	LOG("send_barrier_skbs: wait %d pkts %d total %d min %d max %d",
		tun->barrier_wait_usec,
		tun->barrier_pkts,
		tun->barrier_total,
		tun->barrier_min,
		tun->barrier_max);

	if (not_worthit)
		goto send_now;

	/* FIXME: what about if two processes do this? */
	BUG_ON(1);
/* 	atomic_set(&bflow->queue_pkts, 1); */

	do {
		schedule();
		scheduled++;
		twait = tick_now_ns() - started;
		keep_waiting = 
			(bflow->tosend_accumulated < (tun->barrier_total/3)) &&
			(twait < canwait);
		DEBUG("twait %llu canwait %llu", twait, canwait);
	} while (keep_waiting);

	LOG("send_barrier_skbs: scheduled %d twait %lld accumulated %lld",
		scheduled, twait, bflow->tosend_accumulated);
send_now:
	tsend = tick_now_ns();
/* 	rc = tun->prot->ops->send_barrier_skbs(sk); */
	// XXX work is done in udpex_send_barrier_skbs
	// XXX renamed to udpex_transmit_barrier_skbs
	BUG();
	tsend = tick_now_ns() - tsend;
	
	if (rc || not_worthit)
		goto done;
	
	BUG_ON(1);
/* 	rc = bhtp_send_queued_pkts(bflow); */
done:
	if (twait_p)
		*twait_p = twait;
	if (tsend_p)
		*tsend_p = tsend;

	return rc;
}

#define BHTP_NOT_NULL_DROP(p) \
	do \
		if ((p) == NULL) { \
			ERR_TR("pointer %s is NULL", #p); \
			goto drop; \
		} \
	while (0)

static inline struct bhtp_tun_prot *
bhtp_lookup_proto(int proto)
{
	if (proto < 0 || proto > BHTP_PROTO_MAX)
		return NULL;
	
	return bhtp.tun_prots[proto];
}


static inline int
bhtp_tun_rcv(struct sk_buff *skb, int trim_above, int orig_padbytes,
		struct bhtphdr *bh, struct bhtphdr_tun *tun)
{
	struct sk_buff *tskb = NULL;
	struct bhtp_tun_prot *prot = NULL;
	struct bhtp_qskb *qskb = NULL;
	int rc = 0;

        BHTP_NOT_NULL_DROP(skb);

	/*
	 * Create a new skb that contains only the TUN segment:
	 *
	 *	- clone the bhtp skb; bhtp header has been already pulled
	 *	- remove any TUN segments above our segment
	 *	- keep only our TUN segments by removing segments below
	 *	- reposition ->transport_header header
	 *	- hand to upper transport (UDPex, TCP, etc).
	 *
	 * Be careful to only operate on the cloned skb fields.
	 * Do not overwrite any of the skb head data or any of the
	 * fragments since all TUNS share the data. That is, even
	 * if it is beyond skb->tail, another TUN may use it.
	 * Very true when TUNs are stacked inside a BHTP packet.
	 *
	 * See more in udpex_tun_rcv() notes about "recreate the padding":
	 *
	 *	cd bhtp
	 *	svn log -r261 .
	 */

        DEBUG_TR("cloning skb %d len", skb->len);
	tskb = skb_clone(skb, GFP_ATOMIC);
	if (trim_above)
		skb_pull(tskb, trim_above);
	skb_trim(tskb, tun->len);
	tskb->transport_header  = tskb->data;

	DEBUG("tun_rcv: proto %d len %d needs_padding %d",
		tun->proto, tun->len, tun->needs_padding);

	/* Demux the TUN segment based on L4 protocol and process it. */
	prot = bhtp_lookup_proto(tun->proto);

	if (!BHTP_TUN_PROT_HAS_RCV(prot)) {
		ERR("tun_rcv: unknown proto %d or no rcv ops for prot %p",
			tun->proto, prot);
		goto drop_verbose;
	}

	/* overloading the control buffer in sk_buff for bhtp private data */
	qskb = bhtp_qsk_init(tskb);

	/* only a few fields filled; TUN rcv code should do the rest */
	qskb->tun_len = tskb->len;

        // The padding in a BHTP packet must be assigned to exactly one
        // of the TUNs. It is up to the TUN protocol to deal with the padding.
        if (tun->needs_padding) {
                if (orig_padbytes == 0) {
                        ERR("tun_rcv: TUN needs padding flag but no padbytes");
                        goto drop_verbose;
                }
                qskb->tun_pad = orig_padbytes;
        }

	/* CARE the bhtphdr pointer is assumed to point inside the skb */
	/* CARE it must remain valid while qskb is valid (ie within BHTP) */
	/* CARE This is used to retrieve timestamp in bhtp_recv_get_rcvstamp() */
	qskb->recv.bh = bh;

	/* hand over to the TUN rcv code */
	rc = prot->ops->rcv(tskb);
	if (rc) {
		ERR("tun_rcv: rcv ops error: %d", rc);
		goto drop_verbose;
	}

	return 0;

drop_verbose:
        log_hex(bh, sizeof(struct bhtphdr), 4, "bhtphdr: ");
        LOG("tun_rcv: ------------------");
        log_hex(tskb->transport_header, min(20, 21), 4, "tundata: ");
        ERR("tun_rcv: TUN segment DROPPED, proto %d, len %d, trim_above %d",
            tun->proto, tun->len, trim_above);
drop:
	kfree_skb(tskb);
	return 0;
}

int
bhtp_rcv(struct sk_buff *skb)
{
	struct bhtphdr *bh = NULL;
	int processed = 0;
	int t = 0;
	int tuns = 0;
        int bhtp_pktlen = 0;
        int bhtp_payload_len = 0;
        const char *errstr = "unknown error";
        const struct iphdr *iph= (struct iphdr *)skb->network_header;

        struct bhtp_flow *bflow = NULL;
        struct sockaddr_in remote_addr = {
                .sin_family = AF_INET,
                .sin_addr.s_addr = iph->saddr,
        };

        int pad_reused = 0;
        int pad_unused = 0;
        int pad_orig = 0;

        char error[200];

        DEBUG("************************************");
	DEBUG("bhtp_rcv skb %p len %d", skb, skb->len);
        if (debug_on()) {
                int loglen = min_t(int, skb->len, 48);
                log_hex((const char *)skb->transport_header, loglen, 4, "rcv: ");
        }

	if (!pskb_may_pull(skb, sizeof(struct bhtphdr))) {
		ERR("rcv: cannot pull BHTP header: skb_len %d (needs %d)",
			skb->len, sizeof(struct bhtphdr));
                errstr = "pkt too small for BHTP header";
		goto drop;
	}

	bh = (struct bhtphdr *)skb->transport_header;
	BHTP_NOT_NULL_DROP(bh);

        bhtp_pktlen = skb->len;
        pad_orig = bhtp_orig_pad_bytes(bh, bhtp_pktlen,
                                       &pad_reused, &pad_unused,
                                       error, sizeof(error));
        if (pad_orig == -1) {
                ERR("error: invalid header: %s", error);
                errstr = "failed to extract padding information from header";
                goto drop_verbose;
        }

        DEBUG("rcv: BHTP pkt from IP len %d pad_unused %d pad_reused %d",
              bhtp_pktlen, pad_unused, pad_reused);

	if (ktime_to_ns(skb->tstamp) == 0)
		skb->tstamp = tick_now();

	/* reposition the L4 pointer below the BHTP header */
	skb->transport_header = skb_pull(skb, sizeof(struct bhtphdr));
	BHTP_NOT_NULL_DROP(skb->transport_header);

        // removed bhtp header, only bhtp payload left
        bhtp_payload_len = skb->len;

        // look for the flow (needed for receive stats)
        //
        //      - create if not there
        //      - mark it (bflow->recv_user) the first time

        bflow = bhtp_find_mark_recv_flow(&remote_addr, 0);
        if (bflow == NULL) {
                ERR("rcv: cannot create bhtp_flow for src: "
                    NIPQUAD_FMT,
                    NIPQUAD(remote_addr.sin_addr.s_addr));
                errstr = "failed to create bhtp_flow";
                goto drop_verbose;
        }

        // XXX change this lock to a new lock protecting the counters!
	spin_lock_bh(&bflow->tosend_lock);
        mgrp_counters_update_recv_side(&bflow->counters, bh, skb->len);
	spin_unlock_bh(&bflow->tosend_lock);

	/* extract and process each TUN segment */
	processed = 0;

	tuns = bhtp_header_tuns(bh);
	if (tuns <= 0 || tuns > BHTP_MAX_SEG) {
                ERR("rcv: invalid header_tuns %d", tuns);
                errstr = "invalid tuns in header";
                goto drop_verbose;
	}

        DEBUG("rcv: processing %d tuns", tuns);
	for (t = 0; t < tuns; t++) {
                int remaining = bhtp_payload_len - processed - pad_unused;

		/* sanity check */
		if (bh->tun[t].len > remaining) {
                        ERR("rcv: invalid len %d for TUN entry %d, only %d left",
                            bh->tun[t].len, t, remaining);
                        errstr = "invalid TUN len";
                        goto drop_verbose;
		}
                if (bh->tun[t].needs_padding) {
                        // the TUN needs to know how to reconstruct the probe
                        bhtp_tun_rcv(skb, processed, pad_orig, bh, &bh->tun[t]);
                } else {
                        bhtp_tun_rcv(skb, processed, 0, bh, &bh->tun[t]);
                }
		processed += bh->tun[t].len;
	}

	/* The whole packet needs to be accounted for through TUNs. */
	if (processed + pad_unused != bhtp_payload_len) {
                ERR("rcv: BHTP pkt not processed fully: "
                    "payloadbytes %d processed %d padbytes %d",
                    bhtp_payload_len, processed, pad_unused);
                errstr = "not all BHTP payload processed";
                goto drop_verbose;
	}

        // no need to check if padding fields make sense
        // already done in bhtp_orig_pad_bytes() higher up

	/* no more need for container skb; tun skbs are cloned */
	kfree_skb(skb);
	/* always zero */
	return 0;

drop_verbose:
        log_hex((const char *)bh, min_t(int, 48, skb->len), 4, "drop: ");
        if (bhtp_pktlen == 20 || bhtp_pktlen == 32) {
                // looks like a TCP packet without a MGRP header
                // this is a know bug of MGRP
                char str[200];
                struct tcphdr *th = (struct tcphdr *)bh;
                mgrp_tcphdr_oneliner(str, sizeof(str), th, skb->len);
                DEBUG("%s", str);
        }
drop:
	ERR("rcv: BHTP packet DROPPED: %s", errstr);
	kfree_skb(skb);
	/* always zero */
	return 0;
}


void
bhtp_err(struct sk_buff *skb, __u32 info)
{
	LOG("bhtp_err");
}

int
bhtp_state_init(void)
{
	int rc = 0;
	unsigned long j = jiffies;

	/* do not zero state: it is initialized later at declaration */

	/* FIXME: This is ok and enough because tun->proto is u8 */
	bhtp.tun_prots = kzalloc(sizeof(struct bhtp_tun) * 256, GFP_KERNEL);
	if (bhtp.tun_prots == NULL)
		return -ENOBUFS;

	spin_lock_init(&bhtp.lock);
	spin_lock_init(&bhtp.req_lock);

	INIT_LIST_HEAD(&bhtp.req_list);
	init_waitqueue_head(&bhtp.req_wq);

	/* sysfs */
/* 	rc = bhtp_sysfs_register(); */
	if (rc)
		goto free_tun_prots;


	init_timer(&bhtp.req_timer);
	bhtp.req_timer.function = bhtp_req_timer;
	bhtp.req_timer.expires = j + msecs_to_jiffies(5000);
	add_timer(&bhtp.req_timer);

	return 0;

free_tun_prots:
	kfree(bhtp.tun_prots);
	return rc;
}

void
bhtp_state_done(void)
{
	mgrp_destroy_all_flows();
	kfree(bhtp.tun_prots);
	bhtp.tun_prots = NULL;

/* 	bhtp_sysfs_unregister(); */

	del_timer_sync(&bhtp.req_timer);
	bhtp_drop_req();
}

struct bhtp_state bhtp = {

	/* for IP layer registration */
	.protocol = {
		.handler	= bhtp_rcv,
		.err_handler	= bhtp_err,
		.no_policy	= 1,
	},
	.flows = LIST_HEAD_INIT(bhtp.flows),
	.latency = BHTP_LATENCY_NS,
	.maxdelay = MGRP_MAXDELAY_NS,
};

/* BHTP ABI */
EXPORT_SYMBOL(bhtp_register_tun_prot);
EXPORT_SYMBOL(bhtp_unregister_tun_prot);
EXPORT_SYMBOL(bhtp_find_attach_flow);
EXPORT_SYMBOL(bhtp_detach_flow);
EXPORT_SYMBOL(bhtp_transmit_skb);

