/*
 * MGRP		Measurement Manager Protocol
 *
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Tue Feb  5 09:34:08 MST 2008
 *
 * /proc/net/mgrp_* entries
 */

#include <linux/proc_fs.h>
#include <linux/seq_file.h>

#include "main.h"
#include "frag.h"
#include "proto.h"

/* For more on the seq_file interface, LDD, ed3, chap4, p87 */

///////////////////////////////////////////////
// seq_file interface for /proc/mgrp_counters

static void *mgrp_proc_counters_seq_start(struct seq_file *seq, loff_t *pos)
{
        int i = 0;
        struct bhtp_flow *bflow = NULL;

	spin_lock_bh(&bhtp.lock);

        list_for_each_entry(bflow, &bhtp.flows, flist) {
                if (i == *pos)
                        return bflow;
                i++;
        }

        return NULL;
}

static void *mgrp_proc_counters_seq_next(struct seq_file *seq,
                                         void *v, loff_t *pos)
{
        struct bhtp_flow *bflow = v;

        ++*pos;
        list_for_each_entry_continue(bflow, &bhtp.flows, flist) {
                return bflow;
        }
        return NULL;
}

static void mgrp_proc_counters_seq_stop(struct seq_file *seq, void *v)
{
	spin_unlock_bh(&bhtp.lock);
}

static int mgrp_proc_counters_seq_show(struct seq_file *seq, void *v)
{
        struct bhtp_flow *bflow = v;
        char counters[300];

        // put counters into a string
        spin_lock_bh(&bflow->tosend_lock);
        mgrp_counters_oneliner(counters, sizeof(counters), &bflow->counters);
        spin_unlock_bh(&bflow->tosend_lock);

        // flow destination address + counter representation
        seq_printf(seq, NIPQUAD_FMT " %s\n", NIPQUAD(bflow->daddr), counters);
        return 0;
}

static const struct seq_operations mgrp_proc_counters_seqops = {
        .start = mgrp_proc_counters_seq_start,
        .next = mgrp_proc_counters_seq_next,
        .stop = mgrp_proc_counters_seq_stop,
        .show = mgrp_proc_counters_seq_show,
};

static int mgrp_proc_counters_open(struct inode *inode, struct file *file)
{
        return seq_open(file, &mgrp_proc_counters_seqops);
}

const struct file_operations mgrp_proc_counters_fops = {
        .owner = THIS_MODULE,
        .open = mgrp_proc_counters_open,
        .read = seq_read,
        .llseek = seq_lseek,
        .release = seq_release,
};

static int 
mgrp_proc_counters_reset(char *buffer, char **start, off_t offset,
                         int length, int *eof, void *data)
{
        int written = 0;

        if (offset != 0)
                return 0;

        mgrp_reset_counters_in_all_flows();
        written = snprintf(buffer, length, "mgrp_counters_reset");
        return written;
}

#ifdef MGRP_TSKBSTATS
///////////////////////////////////////////////
// seq_file interface for /proc/mgrp_tskbstats

/*
 * The kernel throttles how many times _next is called in a single
 * /proc operation even if the elements are not exhauted.
 *
 * That means that doing:
 *
 *      cat /proc/net/mgrp_tskbstats
 *
 * results in the following calls:
 *
 *      _start, _next, _next, ... [until NULL or threshold], _stop
 *      _start, _next, _next, ... [until NULL or threshold], _stop
 *      ......
 *      _start, _next, _next, ... [until NULL or threshold], _stop
 *
 * until all the elements are exhausted (ie, NULL is returned).
 */
static void *mgrp_proc_tskbstats_seq_start(struct seq_file *seq, loff_t *pos)
{
        struct mgrp_tskbstats *stats = bhtp.tskbstats;
        char str[300];

        // header
        // (only appears if there is at least one element)
        // (only appears the first time _start is invoked in a batch)
        if (*pos == 0) {
                mgrp_tskbstats_oneliner_header(str, sizeof(str));
                seq_printf(seq, "%s\n", str);
        }

        // LOCK
	spin_lock_bh(&stats->lock);
        if (*pos >= stats->count)
                return NULL;
        return stats->entries + *pos;
}

static void *mgrp_proc_tskbstats_seq_next(struct seq_file *seq,
                                         void *v, loff_t *pos)
{
        struct mgrp_tskbstats *stats = bhtp.tskbstats;
        (*pos)++;
        if (*pos >= stats->count)
                return NULL;
        return stats->entries + *pos;
}

static void mgrp_proc_tskbstats_seq_stop(struct seq_file *seq, void *v)
{
        struct mgrp_tskbstats *stats = bhtp.tskbstats;
        // UNLOCK
	spin_unlock_bh(&stats->lock);
}

static int mgrp_proc_tskbstats_seq_show(struct seq_file *seq, void *v)
{
        struct mgrp_tskbstats_entry *entry = v;
        char str[300];

        // put stats into a string
        mgrp_tskbstats_oneliner(str, sizeof(str), entry);
        seq_printf(seq, "%s\n", str);
        return 0;
}

static const struct seq_operations mgrp_proc_tskbstats_seqops = {
        .start = mgrp_proc_tskbstats_seq_start,
        .next = mgrp_proc_tskbstats_seq_next,
        .stop = mgrp_proc_tskbstats_seq_stop,
        .show = mgrp_proc_tskbstats_seq_show,
};

static int mgrp_proc_tskbstats_open(struct inode *inode, struct file *file)
{
        return seq_open(file, &mgrp_proc_tskbstats_seqops);
}

const struct file_operations mgrp_proc_tskbstats_fops = {
        .owner = THIS_MODULE,
        .open = mgrp_proc_tskbstats_open,
        .read = seq_read,
        .llseek = seq_lseek,
        .release = seq_release,
};

///////////////////////////////////////////////
// /proc/mgrp_tskbstats_reset
// XXX dummy proc entry that resets tskbstats

static int 
mgrp_proc_tskbstats_reset(char *buffer, char **start, off_t offset,
                          int length, int *eof, void *data)
{
        int written = 0;
        int reset_entries = 0;
        struct mgrp_tskbstats *stats = bhtp.tskbstats;

        if (offset != 0)
                return 0;

	spin_lock_bh(&stats->lock);
        reset_entries = stats->count;
        mgrp_tskbstats_reset(stats);
	spin_unlock_bh(&stats->lock);

        written = snprintf(buffer, length,
                           "resetting %d tskbstats entries\n",
                           reset_entries);
        LOGFN("%d entries", reset_entries);

        return written;
}
#endif

///////////////////////////////////////////////
// /proc/mgrp_destroy_flows
// XXX dummy proc entry that destroys all flows (so that it can unload)

static int 
mgrp_proc_destroy_flows(char *buffer, char **start, off_t offset,
                        int length, int *eof, void *data)
{
        int written = 0;
        int flows_destroyed = 0;

        if (offset != 0)
                return 0;

        flows_destroyed = mgrp_destroy_all_flows();
        written = snprintf(buffer, length, "destroyed %d flows\n",
                           flows_destroyed);
        return written;
}

/*
 * Install all /proc entries
 */

int __init
mgrp_proc_init(void)
{
        struct proc_dir_entry *proc = NULL;

        /* Install /proc/net/mgrp_counters on default net namespace */
        proc = proc_net_fops_create(&init_net,
                                    "mgrp_counters", S_IRUGO,
                                    &mgrp_proc_counters_fops);
        if (proc == NULL)
                goto error;

        proc = create_proc_read_entry("mgrp_counters_reset", 0, NULL,
                                      mgrp_proc_counters_reset, NULL);

        if (proc == NULL)
                goto error_remove_counters;


        proc = create_proc_read_entry("mgrp_destroy_flows", 0, NULL,
                                      mgrp_proc_destroy_flows, NULL);
        if (proc == NULL)
                goto error_remove_counters_reset;

        // proc entries related to fragmentation
        proc = proc_net_fops_create(&init_net,
                                    "mgrp_frag_buckets", S_IRUGO,
                                    &mgrp_frag_buckets_fops);
        if (proc == NULL)
                goto error_remove_destroy_flows;

        proc = create_proc_read_entry("mgrp_frag_reset", 0, NULL,
                                      mgrp_frag_reset, NULL);

        if (proc == NULL)
                goto error_remove_frag_buckets;

        proc = create_proc_read_entry("mgrp_frag_hash_counters", 0, NULL,
                                      mgrp_frag_hash_counters_show, NULL);

        if (proc == NULL)
                goto error_remove_frag_reset;

#ifdef MGRP_TSKBSTATS
        /* Install /proc/net/mgrp_tskbstats on default net namespace */
        proc = proc_net_fops_create(&init_net,
                                    "mgrp_tskbstats", S_IRUGO,
                                    &mgrp_proc_tskbstats_fops);
        if (proc == NULL)
                goto error_remove_frag_hash_counters;

        proc = create_proc_read_entry("mgrp_tskbstats_reset", 0, NULL,
                                      mgrp_proc_tskbstats_reset, NULL);
        if (proc == NULL)
                goto error_remove_tskbstats;

#endif

        return 0;

#ifdef MGRP_TSKBSTATS
error_remove_tskbstats:
        proc_net_remove(&init_net, "mgrp_tskbstats");
error_remove_frag_hash_counters:
        remove_proc_entry("mgrp_frag_hash_counters", NULL);
#endif
error_remove_frag_reset:
        remove_proc_entry("mgrp_frag_reset", NULL);
error_remove_frag_buckets:
        proc_net_remove(&init_net, "mgrp_frag_buckets");
error_remove_destroy_flows:
        remove_proc_entry("mgrp_destroy_flows", NULL);
error_remove_counters_reset:
        remove_proc_entry("mgrp_counters_reset", NULL);
error_remove_counters:
        proc_net_remove(&init_net, "mgrp_counters");
error:
        return -1;
}

void __exit
mgrp_proc_done(void)
{
        proc_net_remove(&init_net, "mgrp_counters");
        remove_proc_entry("mgrp_counters_reset", NULL);
        remove_proc_entry("mgrp_destroy_flows", NULL);
        proc_net_remove(&init_net, "mgrp_frag_buckets");
        remove_proc_entry("mgrp_frag_reset", NULL);
        remove_proc_entry("mgrp_frag_hash_counters", NULL);
#ifdef MGRP_TSKBSTATS
        proc_net_remove(&init_net, "mgrp_tskbstats");
        remove_proc_entry("mgrp_tskbstats_reset", NULL);
#endif
}
