/*
 * BHTP		Bottom Half Transport Protocol
 *
 * Created:	Sun Mar 26 19:36:51 EST 2006
 *
 * Author:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 */

#include <linux/ethtool.h>      // to retrieve device speeds
#include <net/route.h>		/* for ip_route_XXX */

#include "main.h"
#include "proto.h"
#include "tick.h"
#include "sysfs.h"
#include "frag.h"
#include "tcp.h"
#include "req.h"

/* hack, used to be in skbuff.h */
/* try to use skb_copy_bits() instead */
#include </usr/src/linux/net/core/kmap_skb.h> /* "../../kernel-compiled/linux/net/core/kmap_skb.h" */

#ifdef LOG_DUPLICATE
static inline void
__debug_duplicate_skb(const char *prefix, struct sk_buff *skb,
                      const char *function)
{
        if (mgrp_tskb_duplicate(skb)) {
                u32 fragid = 0;
                int rc = mgrp_tskb_fragid_get(skb, &fragid);
                LOG("%s: %s skb %p len %d duplicate %d fragid %s %d",
                    function, prefix,
                    skb, skb->len,
                    mgrp_tskb_duplicate(skb),
                    (rc ? "NOT SET" : "SET"), fragid);
        }
}

#define DEBUG_DUPSKB(_prefix, _skb) \
        __debug_duplicate_skb(_prefix, _skb, __FUNCTION__)

#define DEBUG_DUP(format, ...) \
        LOGFN(format, ## __VA_ARGS__)

#else
#define DEBUG_DUPSKB(_prefix, _skb)     ((void)0)
#define DEBUG_DUP(format, ...)          ((void)0)
#endif

/* dummy socket prot to be attached to bflow, needed by ip_queue_xmit */

struct dummy_sock {
	/* inet_sock has to be the first member */
	struct inet_sock inet;
};

int
dummy_init_sock(struct sock *sk)
{
	LOG("dummy_init_sock");
	return 0;
}

int
dummy_destroy_sock(struct sock *sk)
{
	LOG("dummy_destroy_sock");
	return 0;
}

struct proto dummy_prot = {
	.name		= "dummy",
	.owner		= THIS_MODULE,
	.init		= dummy_init_sock,
	.destroy	= dummy_destroy_sock,
	.obj_size	= sizeof(struct dummy_sock),
};

static inline void
bhtp_header_fill_sndstamp(struct bhtphdr *bh, const unsigned long long base_ns)
{
        int overflow = bhtp_set_sndstamp_ns(bh, tick_now_ns() - base_ns);
        if (overflow)
                LOGFN("sndstamp overflow");
}

/* A B I */

int
bhtp_register_tun_prot(struct bhtp_tun_prot *tun_prot)
{
	int rc = -EINVAL;

	/* FIXME: sanity checks? */

	if (tun_prot->ops == NULL)
		goto err;
	if (tun_prot->ops->rcv == NULL)
		goto err;

	spin_lock_bh(&bhtp.lock);

	/* FIXME: tun_prot->proto is u8 so it is never out of bounds. */
	if (bhtp.tun_prots[tun_prot->proto] != NULL)
		/* Proto is already registered. */
		goto err_unlock;

	bhtp.tun_prots[tun_prot->proto] = tun_prot;

	spin_unlock_bh(&bhtp.lock);
	return 0;

err_unlock:
	spin_unlock_bh(&bhtp.lock);
err:
	return rc;
}

int
bhtp_unregister_tun_prot(struct bhtp_tun_prot *tun_prot)
{
	int rc = -EINVAL;

	spin_lock_bh(&bhtp.lock);

	if (tun_prot != bhtp.tun_prots[tun_prot->proto])
		goto err_unlock;

	bhtp.tun_prots[tun_prot->proto] =  NULL;

	spin_unlock_bh(&bhtp.lock);
	return 0;

err_unlock:
	spin_unlock_bh(&bhtp.lock);
	return rc;
}

static struct dst_entry *
bhtp_get_dst(struct sockaddr_in *daddr, struct sockaddr_in *saddr)
{
	/* based on net/sctp/protocol.c:sctp_v4_get_dst() */
	/* Also check LNI, c21, p478-9, Setting the route. */

	struct rtable *rt = NULL;
	struct dst_entry *dst = NULL;
	struct flowi fl = {
		.proto = IPPROTO_MGRP,
		.fl4_dst = daddr->sin_addr.s_addr,
	};

	if (saddr)
		fl.fl4_src = saddr->sin_addr.s_addr;

	DEBUGFN("looking up route dst " NIPQUAD_FMT " src " NIPQUAD_FMT,
	        NIPQUAD(fl.fl4_dst), NIPQUAD(fl.fl4_src));
	      
	if (!ip_route_output_key(&init_net, &rt, &fl))
		dst = &rt->u.dst;

	return dst;
}

static inline struct dst_entry *
bhtp_dst_check(struct bhtp_flow *bflow)
{
	/* based on net/sock.h:sk_dst_check() */
	struct dst_entry *dst = bflow->dst;

	if (dst && dst->obsolete) {
		dst_release(dst);
		bflow->dst = NULL;
		return NULL;
	}

	return dst;
}

void
bhtp_sk_destruct(struct sock *sk)
{
	DEBUGFN("DESTRUCT bhtp_sk_destruct");
}

static inline int
mgrp_dst_dev_speed(struct dst_entry *dst)
{
        int err = 0;
        struct ethtool_cmd cmd = { .cmd = ETHTOOL_GSET };

        if (!dst->dev->ethtool_ops->get_settings) {
                // loopback (for example)
                return 0;
        }

        err = dst->dev->ethtool_ops->get_settings(dst->dev, &cmd);
        if (err < 0) {
                ERRFN("ethtool get_settings failed (%d)", err);
                return 0;
        }

        DEBUGFN("%s speed: %d", dst->dev->name, cmd.speed);
        BUG_TRAP(cmd.speed == 10 || cmd.speed == 100 || cmd.speed == 1000);

        return cmd.speed;
}

/* 0 ok, -1 error */
static inline int
bhtp_flow_init(struct bhtp_flow *bflow, int bflow_id,
               struct net *net, struct sockaddr_in *usin, int flags)
{
	struct dst_entry *dst = NULL;
	struct sockaddr_in dst_addr = {
		.sin_addr.s_addr = usin->sin_addr.s_addr,
	};

        // bflow has been allocated and zeroed by caller; now we need to init

	spin_lock_init(&bflow->lock);
	spin_lock_init(&bflow->tosend_lock);

	INIT_LIST_HEAD(&bflow->flist);
	INIT_LIST_HEAD(&bflow->tuns);

        bflow->pbk_policy = bhtp.policy;
	skb_queue_head_init(&bflow->tosend);
	skb_queue_head_init(&bflow->tosend_pbk);

	bflow->id = bflow_id;
	bflow->flags = flags;
	bflow->daddr = usin->sin_addr.s_addr;

	bflow->sk = sk_alloc(net, AF_INET, GFP_ATOMIC, &dummy_prot);
	if (bflow->sk == NULL)
		goto error;
	/* this seems to be the only field that ip_queue_xmit needs */
	/* do we need to do a sock_hold() ? */
	sock_hold(bflow->sk);
	/* HACK so that ip_queue_xmit will pick them up */
	bflow->sk->sk_protocol = IPPROTO_MGRP;
	bflow->sk->sk_destruct = &bhtp_sk_destruct;
	inet_sk(bflow->sk)->uc_ttl = 162;

	/* FIXME: We are ignoring source address at this point.
	 * This is ok when our machine has one IP but not for
	 * multihomed machines.
	 */

	dst = bhtp_get_dst(&dst_addr, NULL);
	if (dst) {
		bflow->dst = dst_clone(dst);
                bflow->dst_dev_speed_mbps = mgrp_dst_dev_speed(bflow->dst);
		dst_release(dst);
		DEBUGFN("got dst");
	}

        bflow->counters.epoch = bflow->id;

        // All MGRP packets contain a send timestamp (bhtphdr.sndstamp)
        // that is an offset to this base timestamp.
        bflow->sndstamp_base_ns = tick_now_ns();
        DEBUGFN("sndstamp_base_ns %lld", bflow->sndstamp_base_ns);

	DEBUGFN("created new bhtp flow: id %d, addr " NIPQUAD_FMT,
		bflow->id, NIPQUAD(bflow->daddr));

	return 0;
error:
	LOG_TR("out of mem");
	return -1;
}


void
bhtp_flow_done(struct bhtp_flow *bflow)
{
        // flow was marked by bhtp_rcv() in bhtp_find_recv_flow()
        if (atomic_add_unless(&bflow->recv_user, -1, 0)) {
                atomic_dec(&bflow->users);
        }

	if (atomic_read(&bflow->users) != 0) {
		LOG("BUG: Flow is still used. Forcing removal.");
        }

	if (bflow->dst)
		dst_release(bflow->dst);
	/* XXX FIXME check if also still part of a list in flist */

	bhtp_drain_tosend(bflow, 1);

	DEBUGFN("FLOW refcount %d", atomic_read(&bflow->sk->sk_refcnt));

	/* FIXME A hack because we get the following error on the console:
	 *
	 *	Attempt to release TCP socket in state 1 f6eb4980
	 *
	 * which is in af_inet.c:inet_sock_desctruct().
	 *
	 * This could also be why after we open/close a TCP BHTP socket
	 * we cannot unload the bhtp module. FIXME
	 */
	bflow->sk->sk_state = TCP_CLOSE;
	sock_put(bflow->sk);
	bflow->sk = NULL;
}

/*
 * Destroys bhtp flows and removes them from the global list:
 *
 *      if @only_this_flow == NULL, destroys all flows
 *      if @only_this_flow != NULL, destroys only the matching flow
 * 
 * Returns the number of flows removed.
 *
 * Holds bhtp.lock
 */
static inline int
mgrp_destroy_flows(struct bhtp_flow *only_this_bflow)
{
        int count = 0;
        struct bhtp_flow *bflow = NULL;
        struct bhtp_flow *tmp_bflow = NULL;

        spin_lock_bh(&bhtp.lock);

        list_for_each_entry_safe(bflow, tmp_bflow, &bhtp.flows, flist) {
                if ((only_this_bflow == NULL) || (bflow == only_this_bflow)) {
                        list_del(&bflow->flist);
                        bhtp_flow_done(bflow);
                        kfree(bflow);
                        count++;
                        if (only_this_bflow != NULL)
                                break;
                }
        }

        spin_unlock_bh(&bhtp.lock);
        return count;
}

// Returns number of flows destroyed
int
mgrp_destroy_all_flows(void)
{
        int count = mgrp_destroy_flows(NULL);
        LOG("destroyed %d flows", count);
        return count;
}

// Returns -1 on error, 0 on success
static inline int
mgrp_destroy_one_flow(struct bhtp_flow *bflow)
{
        int count = mgrp_destroy_flows(bflow);
        BUG_ON(count > 1);
        return (count == 1) ? 0 : -1;
}

int
bhtp_detach_flow(struct bhtp_flow *bflow, struct bhtp_tun *tun)
{
	int can_destroy = 0;
	int users = 0;
        int rc = 0;
	
	can_destroy = atomic_dec_and_test(&bflow->users);
	users = atomic_read(&bflow->users);

	if (users < 0) {
		ERR("cannot detach flow: negative user count: %d", users);
		return -EINVAL;
	}

	if (!can_destroy) {
		DEBUG("not detaching flow %d: positive user count: %d",
                      bflow->id, users);
		/* but not an error */
		return 0;
	}

        rc = mgrp_destroy_one_flow(bflow);
        if (rc) {
		ERR("cannot detach flow: failed to destroy it: %d", rc);
                return -EINVAL;
        }
	return 0;
}

// must be called with bhtp.lock held
static inline struct bhtp_flow *
_bhtp_flow_lookup(struct sockaddr_in *usin, int flags)
{
	struct bhtp_flow *bflow = NULL;

        // XXX also add a hash table for fast lookup
        // XXX not a problem for our setup where we have 1-2 flows max

	list_for_each_entry(bflow, &bhtp.flows, flist) {
		/* XXX should we compare flags too? */
		if (bhtp_flow_matches_addr(bflow, usin))
			return bflow;
	}
        return NULL;
}

static inline struct bhtp_flow *
bhtp_find_create_flow(struct sockaddr_in *usin, int flags)
{
        int rc = 0;
	struct bhtp_flow *bflow = NULL;
	struct bhtp_flow *new_bflow = NULL;
        int new_bflow_id = 0;

	spin_lock_bh(&bhtp.lock);
        bflow = _bhtp_flow_lookup(usin, flags);
        if (bflow != NULL) {
                goto flow_found_unlock;
        } else {
                // reserve a flow id for the flow that we will create
                bhtp.last_flowid++;
                new_bflow_id = bhtp.last_flowid;
        }
	spin_unlock_bh(&bhtp.lock);

        // flow does not exist
        // create it outside of the lock (due to mem allocations)

	new_bflow = kzalloc(sizeof(struct bhtp_flow), GFP_ATOMIC);
	if (new_bflow == NULL) {
                LOG_TR("out of mem");
		goto error;
        }

        // XXX should the net namespace propagate higher?
        // XXX ie, to the callers of bhtp_find_attach_flow?
        rc = bhtp_flow_init(new_bflow, new_bflow_id,
                            &init_net, usin, flags);
        if (rc)
                goto err_free_new_bflow;

        // insert the new flow while holding lock

	spin_lock_bh(&bhtp.lock);

        // check again; another bhtp_find_attach_flow() may have sneaked it in
        bflow = _bhtp_flow_lookup(usin, flags);
        if (bflow != NULL)
                goto flow_found_unlock;

        list_add(&new_bflow->flist, &bhtp.flows);
        bflow = new_bflow;

flow_found_unlock:
	spin_unlock_bh(&bhtp.lock);
	return bflow;

err_free_new_bflow:
        bhtp_flow_done(new_bflow);
        kfree(new_bflow);
        new_bflow = NULL;
error:
	return NULL;
}

struct bhtp_flow *
bhtp_find_attach_flow(struct sockaddr_in *usin, int flags)
{
        struct bhtp_flow *bflow = bhtp_find_create_flow(usin, flags);
        if (bflow != NULL)
                atomic_inc(&bflow->users);
        return bflow;
}

static inline struct bhtp_flow *
bhtp_find_mark_recv_flow(struct sockaddr_in *usin, int flags)
{
        struct bhtp_flow *bflow = bhtp_find_create_flow(usin, flags);

        if (bflow != NULL) {
                int already_marked = atomic_read(&bflow->recv_user);
                if (!already_marked) {
                        int marked = atomic_add_unless(&bflow->recv_user, 1, 1);
                        DEBUGFN("marked flow");
                        if (marked) {
                                // it is decremented in bhtp_flow_done()
                                atomic_inc(&bflow->users);
                        }
                }
        }
        return bflow;
}


/* FIXME: Is bhtp_shared_info still needed?
 *
 * Space for it is reserved in bhtp_alloc_container_skb().
 * All bhtp private data is saved in to the qskb which is
 * stored in the skb->cb[].
 *
 * Note that the big difference between these two is during
 * skb_clone(). The qskb is cloned as well (since it is part
 * of the sk_buff structure itself) while the bhtp_shared_info
 * is not (it is part of the skb head buffer).
 */

/* private bhtp data */
struct bhtp_shared_info {
	int len;
};

inline struct bhtp_shared_info *
bhtp_shinfo(struct sk_buff *skb)
{
	struct bhtp_shared_info *info = (struct bhtp_shared_info *)skb->head;
	return info;
}

void
bhtp_container_skb_destructor(struct sk_buff *skb)
{
/* 	struct bhtp_shared_info *info = bhtp_shinfo(skb); */
/* 	DEBUG("skb_destructor len %d infor->len %d", skb->len, info->len); */
}

inline struct sk_buff *
bhtp_alloc_container_skb(int size)
{
	struct sk_buff *skb = NULL;
	int reserve =  SKB_DATA_ALIGN(sizeof(struct bhtp_shared_info));
	
	size += reserve;
	skb = alloc_skb(size, GFP_ATOMIC);
	if (skb == NULL)
		return NULL;
	
	skb_reserve(skb, reserve);
	memset(bhtp_shinfo(skb), 0, reserve);

	skb->destructor = bhtp_container_skb_destructor;
	return skb;
}

	/* no fragments supported yet FIXME */
/* 	DEBUG("tun_len %d tun_data_len %d nr_frags %d", */
/* 		tun_skb->len, tun_skb->data_len, skb_shinfo(tun_skb)->nr_frags); */
/* 	DEBUG("nonlinear %d headlen %d pagelen %d", */
/* 		skb_is_nonlinear(tun_skb), */
/* 		skb_headlen(tun_skb), */
/* 		skb_pagelen(tun_skb)); */

/* returns negative error code or >= valid index */
static inline int
bhtp_reserve_tun_in_header(struct bhtphdr *bh,
                           int proto, int len,
                           int needs_padding,
                           int duplicate)
{
	int tuns = bhtp_header_tuns(bh);

	if (tuns < 0 || tuns >= BHTP_MAX_SEG) {
		ERR("reserve_tun: no more TUNs left: tuns %d max %d",
			tuns, BHTP_MAX_SEG);
		return -ENOBUFS;
	}

	bh->tun[tuns].proto = proto;
	bh->tun[tuns].len = len;
	bh->tun[tuns].needs_padding = needs_padding;
	bh->tun[tuns].duplicate = duplicate;

        bhtp_header_tuns_set(bh, tuns+1);

	/* index for newly created TUN entry */
	return tuns;
}

/* skb->transport_header must point to BHTP header */
static inline int
bhtp_append_tun_pad(struct sk_buff *skb, int padding)
{
	int rc = 0;
	struct bhtphdr *bh = (struct bhtphdr *)skb->transport_header;
	void *padptr = NULL;

	if (padding <= 0) {
		rc = -EINVAL;
		goto error;
	}
	
	DEBUG("append_pad: inserting padding: %d bytes", padding);

        if (bh->has_unused_padding) {
                ERR("MGRP pkt has already padding");
                rc = -EINVAL;
		goto error;
        }

	/* does it fit? */
	if (unlikely(skb_tailroom(skb) < padding)) {
		ERR("append_tun: no space for %d padding, only %d left",
			padding, skb_tailroom(skb));
		rc = -ENOBUFS;
		goto error;
	}

	padptr = skb_put(skb, padding);
	memset(padptr, BHTP_PAD_BYTE, padding);
        bh->has_unused_padding = 1;

	return 0;
error:
	return rc;
}

/* skb->transport_header must point to BHTP header */
/* tun_skb->cb[] must point to a valid bhtp_qskb structure */
static inline int
bhtp_append_tun_skb(struct sk_buff *skb, struct sk_buff *tun_skb)
{
	/* reference: ip_append_page() and LNI c21, p504 */
	/* assert: BHTP header must be valid */

	int rc = 0;

	int ilen = 0;
	void *iptr = NULL;
	int i = 0;
	int j = 0;

	struct bhtphdr *bh = (struct bhtphdr *)skb->transport_header;
	struct bhtp_qskb *qskb = NULL;

	int appended = 0;
	int orig_len = skb->len;

	DEBUG("bhtp_append_tun_skb");

	if (tun_skb == NULL) {
		rc =  -EINVAL;
		goto error;
	}

	DEBUG("append_tun: tun-len (main) %d tun-data-len (frags) %d",
		tun_skb->len, tun_skb->data_len);

	qskb = bhtp_qsk(tun_skb);

	/* does it fit? */
	if (unlikely(skb_tailroom(skb) < tun_skb->len)) {
		ERR("append_tun: no space for %d bytes, only %d left",
			tun_skb->len, skb_tailroom(skb));
		rc = -ENOBUFS;
		goto error;
	}

	/* append the main skb buffer (no fragments) */
	ilen = skb_headlen(tun_skb);
	iptr = tun_skb->data;
	memcpy(skb_put(skb, ilen), iptr, ilen);
	appended += ilen;
	DEBUG("append_tun: copied headlen %d bytes total %d", ilen, appended);

	/* append the skb fragments */
	for (i = 0; i < skb_shinfo(tun_skb)->nr_frags; i++) {

		skb_frag_t *frag = &skb_shinfo(tun_skb)->frags[i];
		/* requires a hack to include kmap_skb.h */
		/* try to use skb_copy_bits() instead */
		u8 *vaddr =  kmap_skb_frag(frag);

		ilen = frag->size;
		iptr = vaddr + frag->page_offset;
		memcpy(skb_put(skb, ilen), iptr, ilen);
		appended += ilen;
		DEBUG("append_tun: copied frag(%d) %d bytes total %d",
				i, ilen, appended);
		kunmap_skb_frag(vaddr);
	}

	/* the tun_skb->len must agree with tun_lengths */
	/* this means that the padding is not part of the tun_skb->len */

	/* tun_len must always be valid */
	BUG_ON(qskb->tun_len != qskb->tun_hdr + qskb->tun_data);
	/* skb should be trimmed from padding */
	BUG_ON(tun_skb->len != qskb->tun_len);
	/* must append the whole skb */
	BUG_ON(appended != tun_skb->len);

        DEBUG_DUPSKB("tun", tun_skb);

	/* update the BHTP header with the new TUN */
	j = bhtp_reserve_tun_in_header(bh,
                                       qskb->tun->prot->proto,
                                       appended,
                                       qskb->tun_pad > 0,
                                       mgrp_tskb_has_type(tun_skb,
                                                          MGRP_QSKB_DUPLICATE));

	if (j < 0) {
		rc = j;
		goto restore_container_skb;
	}

	DEBUG_FRAGTUN(&bh->tun[j]);
	DEBUG("append_tun: appended TUN %d, proto %d, len %d, needs_padding %d",
		j, bh->tun[j].proto, bh->tun[j].len, bh->tun[j].needs_padding);

	return 0;

restore_container_skb:
	skb_trim(skb, orig_len);
error:
	return rc;
}

static inline void
mgrp_counters_update_recv_side(struct mgrp_counters *counters,
                               struct bhtphdr *bh, int pktlen)
{
        counters->recv.mgrp_pkts++;

        if (counters->recv.valid_expected) {
                mgrp_seqno_t cur = bh->seqno;
                mgrp_seqno_t expected = counters->recv.mgrp_pkt_expected_seqno;

                // some packets in the middle are missing
                if (mgrp_seqno_after(cur, expected)) {
                        counters->recv.mgrp_pkts_missing +=
                                mgrp_seqno_distance(cur, expected);

                // we got a previously missing packet
                } else if (mgrp_seqno_before(cur, expected)) {
                        counters->recv.mgrp_pkts_outoforder++;
                }
        }

        counters->recv.mgrp_pkts_lost =
                counters->recv.mgrp_pkts_missing -
                counters->recv.mgrp_pkts_outoforder;

        // the expected_seqno increases always monotonically
        // but may jump over missing packets
        counters->recv.mgrp_pkt_expected_seqno = mgrp_seqno_inc(bh->seqno);
        counters->recv.valid_expected = 1;
}

// the skb should be an MGRP packet with the header ready to be
// transmitted to the IP layer
static inline void
mgrp_counters_update_send_side(struct mgrp_counters *counters,
                               struct sk_buff *skb)
{
        int t = 0;
        struct bhtphdr *bh = (struct bhtphdr *)skb->transport_header;
        int tuns = bhtp_header_tuns(bh);
        int tun_offset = sizeof(struct bhtphdr); // where first tun starts

        int probe_pkts = 0;
        int probe_nonpad_bytes = 0;
        int nonprobe_bytes = 0;

        counters->mgrp_pkts++;
        counters->mgrp_bytes_header += sizeof(struct bhtphdr);

        for (t = 0; t < tuns; t++) {
                struct bhtphdr_tun *tun = &bh->tun[t];
                unsigned char *tundata = skb->transport_header + tun_offset;

                // padding is no longer a TUN, must include later
                counters->mgrp_bytes_payload += tun->len;

                if (tun->proto == BHTP_PROTO_TCP) {
                        counters->tcp_pkts++;
                        counters->tcp_bytes += tun->len;
                        nonprobe_bytes += tun->len;

                } else if (tun->proto == BHTP_PROTO_UDP && !tun->needs_padding) {
                        counters->udp_pkts++;
                        counters->udp_bytes += tun->len;
                        nonprobe_bytes += tun->len;

                } else if (tun->proto == BHTP_PROTO_UDP && tun->needs_padding) {
                        // defer updating probe stats until all TUNs visited
                        probe_pkts++;
                        probe_nonpad_bytes += tun->len;

                } else if (tun->proto == MGRP_PROTO_FRAG) {
                        struct mgrp_fraghdr *fh = (struct mgrp_fraghdr *)tundata;

                        counters->frag_chunks++;
                        counters->frag_bytes += tun->len;

                        if (fh->fragmented) {
                                int isfirst = (fh->fragoff == 0);
                                int islast = (!fh->more_fragments);

                                if (isfirst && islast)
                                        counters->frag_chunks_onefrag++;
                                else if (isfirst)
                                        counters->frag_chunks_firstonly++;
                                else if (islast)
                                        counters->frag_chunks_lastonly++;
                                else
                                        counters->frag_chunks_middleonly++;
                        } else {
                                // not fragmented means that it is a
                                // duplicate packet that was fragmented
                                // and sent as piggybacking
                                //
                                // this would be the whole duplicate packet
                                // that needs a fragmentation header so
                                // that it goes through the reassembly hash
                                counters->frag_chunks_onedup++;
                        }

                } else {
                        ERR_TR("error: unknown proto %d", tun->proto);
                }
                tun_offset += tun->len;
        }

        if (probe_pkts == 1) {
                int pad_reused = 0;
                int pad_unused = 0;
                char error[200];
                int pad_orig = bhtp_orig_pad_bytes(bh, skb->len,
                                                   &pad_reused, &pad_unused,
                                                   error, sizeof(error));

                if (pad_orig == -1) {
                        ERR_TR("error: invalid header: %s", error);
                        goto done;
                }

                counters->probe_pkts++;
                if (pad_reused > 0)
                        counters->probe_pkts_reused++;

                // include the padding in the payload
                counters->mgrp_bytes_payload += pad_unused;

                counters->probe_bytes_actual += probe_nonpad_bytes + pad_unused;
                counters->probe_bytes_orig += probe_nonpad_bytes + pad_orig;

                counters->probe_bytes_nonpad += probe_nonpad_bytes;
                counters->probe_bytes_pad_unused += pad_unused;
                counters->probe_bytes_pad_reused += pad_orig - pad_unused;

        } else if (probe_pkts > 1) {
                ERR_TR("error: expecting at most 1 probe but got %d",
                       probe_pkts);
        }
done:
        return;
}

/* SKB is headerless - no BHTP header yet */
/* may return NET_XMIT_CN which is not an error */
/*
 * On success, tun_skb is freed.
 * On failure, tun_skb is not freed and a return code rc returned.
 *
 * Called with bflow->tosend_lock held (XXX please verify).
 */
static inline int
bhtp_transmit_skb(struct bhtp_flow *bflow,
			int append_padbytes, 
			struct sk_buff *tskbs[], int tuns,
                        struct mgrp_probe_feedback *feedback)
{
	int rc = 0;
	int i = 0;
	struct bhtphdr *bh = NULL;
	struct dst_entry *dst = NULL;
	struct sk_buff *skb =NULL;
	
	BUG_ON(bflow == NULL);

	/*
	 * STEP: create the container skb
	 *
	 * We need to create a BHTP skb that holds the header
	 * and just uses the buffer from the tun_skb.
	 */

	skb = bhtp_alloc_container_skb(MAX_BHTP_HEADER + 2000);
	if (unlikely(skb == NULL)) {
		ERR("xmit no buffers");
		rc = -ENOBUFS;
		goto error;
	}

	skb_reserve(skb, MAX_BHTP_HEADER);
	skb->transport_header = skb_push(skb, sizeof(struct bhtphdr));
	bh = (struct bhtphdr *)skb->transport_header;

	/* do we need sock_hold here? if yes, who does the sock_put? */
	/* ip_queue_xmit expects skb->sk->.. */
	skb->sk = bflow->sk;

	/* Build Header */
	memset(bh, 0, sizeof(struct bhtphdr));
	bh->seqno = atomic_add_return(1, &bhtp.seqno);

	/* timestamp is filled later, just before we hand to IP layer */

        // XXX timestamp is broken because the ip_queue_xmit() results
        // XXX in timestamps that are smaller than transmission latencies
        // XXX The problem is that we should probably use what UDP uses

	/*
	 * STEP: combine multiple TUN skbs in the container skb
	 *
	 * Then you need to be able to combine multiple
	 * skb buffers into this one (for reuse).
	 */
	 for (i = 0; i < tuns; i++) {
                DEBUG_DUPSKB("tun", tskbs[i]);
	 	rc = bhtp_append_tun_skb(skb, tskbs[i]);
		if (rc) {
			ERRFN("failed to append_tun_skb");
			goto free_container_skb;
		}
	 }
	
	/*
	 * STEP: add the padding
	 */
	 if (append_padbytes > 0) {
	 	rc = bhtp_append_tun_pad(skb, append_padbytes);
		if (rc) {
			ERRFN("failed to append %d padbytes",
                            append_padbytes);
			goto free_container_skb;
		}
	 }

	/* HACK: make sure packet length is below MSS */
	/*	 1480 = 1500 - IP(20) */
	if (skb->len > 1480) {
		ERRFN("skb too large: %d", skb->len);
		rc = -EINVAL;
		goto free_container_skb;
	}

	/*
	 * STEP: find packet destination and hand over to IP
	 *
	 * The destination is a per-flow attribute. All the TUNs
	 * that were appended go to the same dest.
	 */

	/* check if the cached destination is valid */
	dst = bhtp_dst_check(bflow);

	/* lookup the destination */
	if (!dst) {
		struct sockaddr_in dst_addr = {
			.sin_addr.s_addr = bflow->daddr,
		};
		dst = bhtp_get_dst(&dst_addr, NULL);
		if (!dst) {
			ERR("transmit_skb: aborted: no route to dst " NIPQUAD_FMT,
				NIPQUAD(bflow->daddr));
			/* FIXME: correct error? */
			rc = -ENETUNREACH;
			goto free_container_skb;
		}
		bflow->dst = dst_clone(dst);
                bflow->dst_dev_speed_mbps = mgrp_dst_dev_speed(bflow->dst);
		DEBUGFN("OK, routed dst " NIPQUAD_FMT, NIPQUAD(bflow->daddr));
	}

	/*
	 * STEP: free the TUN skbs
	 *
	 * No need for them anymore since they are part of the
	 * container skb now. Wait until this late to ensure that
	 * no errors occured.
	 */
	 for (i = 0; i < tuns; i++)
	 	kfree_skb(tskbs[i]);

	/* FIXME: does dst need to be cloned earlier? */
	skb->dst = dst_clone(bflow->dst);

	DEBUG("xmit: ip_queue_xmit");
	bhtp_header_fill_sndstamp(bh, bflow->sndstamp_base_ns);
        mgrp_counters_update_send_side(&bflow->counters, skb);
        if (debug_on()) {
                int loglen = min_t(int, skb->len, 48);
                log_hex((const char *)bh, loglen, 4, "snd: ");
        }

        // at this point we know exactly the size of the generated MGRP packet
        if (feedback != NULL) {
                // caller can fill the remaining fields *after* we fill the size
                mgrp_probe_feedback_init(feedback, skb->len);
        }

	rc = ip_queue_xmit(skb, 0);

	DEBUG("    ip_queue_xmit done rc %d", rc);
	if (rc < 0) {
		ERR("xmit: ip_queue_xmit failed rc %d", rc);
		/* ip_queue_xmit took care of the kfree_skb() */
		goto error;
	}

	return rc;

free_container_skb:
	kfree_skb(skb);
	ERR("xmit: container packet dropped");
error:
	return rc;

}

/*
 *	Fragments an skb and transmits its fragments.
 *
 *	@bflow		BHTP flow
 *	@tskb		TUN skb with upper transport (fe TCP) header
 *	@max_frag_len	No fragment will be larger than this (hdr + data)
 *
 *	This function is generic and is protocol agonstic. It assumes
 *	that the contained protocol (TUN) provides fragment operations
 *	(ops->fragment and ops->fraglen_left). For length bookkeeping
 *	(see length fields of bhtp_qskb) BHTP assumes that the
 *	protocol packet is comprised of a protocol header and data.
 *	When the packet is fragmented, data is split among packets
 *	and a new header is generated for each fragment. BHTP does
 *	not care how this is done.
 *
 *	The function does not make any implicit assumptions about
 *	the nature of the transport protocol (fe, byte-oriented vs
 *	message-oriented). TCP supports fragmentation naturally while
 *	UDP does not. However, another datagram transport protocol
 *	could explicitly support fragmentation.
 */
static inline int
bhtp_fragment_transmit_skb(struct bhtp_flow *bflow, struct sk_buff *tskb,
			   int max_frag_len)
{
	int rc = 0;
	int frag_count = 0;
        int fraglen_max = 0;
        int fraglen_min = 0;
	
	struct bhtp_qskb *qskb = bhtp_qsk(tskb);
	struct bhtp_tun_ops *ops = qskb->tun->prot->ops;

	LOG_TCPHDR_SKB(tskb);

	DEBUG("frag_xmit: fragmenting tun skb len %d max_frag_len %d",
		tskb->len, max_frag_len);

	if (max_frag_len <= 0) {
		ERR("frag_xmit: max_frag_len %d too small", max_frag_len);
		rc = -EINVAL;
		goto error;
	}

	while (ops->fraglen_left(tskb, &fraglen_min, &fraglen_max, NULL)) {

                int fraglen = 0;
                struct mgrp_probe_feedback feedback = { 0 };
                struct sk_buff *frag = NULL;

                // Never comes in here if it does not need fragmentation.
                // So if we are only sending a header, which can never be
                // fragmented, we will never call fragment()
                // (which cannot handle it).
                //
                // see bhtp_transmit_lone_skb()

                if (fraglen_min > max_frag_len) {
                        ERR("frag_xmit: minimum fragment size %d is larger than available space %d",
                            fraglen_min, max_frag_len);
                        rc = -EINVAL;
                        goto error;
                }

                // create the largest fragment that you can
		fraglen = min_t(int, fraglen_max, max_frag_len);
		frag = ops->fragment(tskb, fraglen, 0);

		if (frag == NULL) {
			ERR("frag_xmit: failed to create fragment %d from skb len %d",
				fraglen, tskb->len);
			rc = -EINVAL;
			goto error;
		}

		LOG_TCPHDR_SKB(frag);

		frag_count += 1;

		/* protect against runaway code */
		BUG_ON(frag_count > 100);

		DEBUG("frag_xmit: created fragnent %d len %d, transmitting",
				frag_count, frag->len);

		rc = bhtp_transmit_skb(bflow, 0, &frag, 1, &feedback);
		if (rc < 0) {
			ERR("frag_xmit: failed to send fragment %d len %d",
				frag_count, frag->len);
			goto error;
		}
	}

	kfree_skb(tskb);

error:
	return rc;
}

/* TUN skb fragmentation
 *
 * When a TUN skb is fragmented, skb->len never changes.
 * All the information is kept in qskb->tun_data_frag.
 *
 * For example, if the skb has already been fragmented
 * but is not further fragmentable:
 *
 *	bhtp_skb_already_fragmented(skb) &&
 *	bhtp_skb_can_fragment_further(skb)
 *
 * you cannot just send the skb, but you have to create the last
 * fragment (because skb->len is the whole unfragmented skb).
 */

/*
 * Does the skb belong to a TUN that allows fragmentation?
 * The skb may not be fragmentable further.
 */
static inline int
bhtp_skb_is_fragmentable(struct sk_buff *skb)
{
	struct bhtp_qskb *qskb = bhtp_qsk(skb);
	int header_only = (qskb->tun_data == 0);
	int fragmentable = 
                        !header_only &&
			qskb->tun_pad == 0 &&
			qskb->tun_frag_len_max > 0 &&
			BHTP_TUN_PROT_HAS_FRAG(qskb->tun->prot);
	
	return fragmentable;
}

/* Has the skb already given up a fragment? */
static inline int
bhtp_skb_already_fragmented(struct sk_buff *skb)
{
	struct bhtp_qskb *qskb = bhtp_qsk(skb);
        if (qskb->tun_data_frag > 0 && qskb->tun_chunk_frag > 0)
                ERR("both types of fragmentation cannot be active");
	return (qskb->tun_data_frag > 0);
}

/* Can the skb give up more fragments? */
static inline int
bhtp_skb_can_fragment_further(struct sk_buff *skb, int max_frag_len)
{
        int has_frags = 0;
        int fraglen_min = 0;
        struct bhtp_tun_ops *ops = bhtp_qsk(skb)->tun->prot->ops;

        has_frags = ops->fraglen_left(skb, &fraglen_min, NULL, NULL);
        BUG_ON(!has_frags);

        // the minimum allowed fragment must fit in the availble space
	return (fraglen_min <= max_frag_len);
}

/* Is the skb too big to send as is? */
static inline int
bhtp_skb_must_fragment_further(struct sk_buff *skb)
{
        int has_frags = 0;
        int has_more_frags = 0;
        struct bhtp_tun_ops *ops = bhtp_qsk(skb)->tun->prot->ops;

        has_frags = ops->fraglen_left(skb, NULL, NULL, &has_more_frags);
        BUG_ON(!has_frags);

        // if @has_more_frags=1, then the current skb fragment is not
        // enough to send all the remaining data
        return has_more_frags;
}

/* Does the skb need to be transmitted using bhtp_skb_fragment_transmit()?
 *
 * This does not necessarily mean that we are breaking up the
 * skb further. It just means that we have to use
 * bhtp_fragment_transmit_skb() and not bhtp_transmit_skb().
 * (for example when sending the last fragment).
 */
static inline int
bhtp_skb_needs_fragment_transmit(struct sk_buff *skb)
{
	if (!bhtp_skb_is_fragmentable(skb))
		return 0;

	if (bhtp_skb_already_fragmented(skb))
		return 1;

	if (bhtp_skb_must_fragment_further(skb))
		return 1;

	return 0;
}

static inline void
mgrp_skb_reset_fragmentation(struct sk_buff *skb)
{
        struct bhtp_qskb *qskb = bhtp_qsk(skb);
        qskb->tun_chunk_frag = 0;
        qskb->tun_data_frag = 0;
        if (qskb->type & BHTP_QSKB_FRAG_EXHAUSTED)
                qskb->type ^= BHTP_QSKB_FRAG_EXHAUSTED;
}

// does not include the header
static inline int
bhtp_fragdata_left(struct sk_buff *skb)
{
        struct bhtp_qskb *qskb = bhtp_qsk(skb);
        int fragdata_left = qskb->tun_data - qskb->tun_data_frag;
        return fragdata_left;
}

// includes the header
static inline int
bhtp_max_fraglen_left(struct sk_buff *skb)
{
        int has_frags = 0;
        int fraglen_max = 0;
        struct bhtp_tun_ops *ops = bhtp_qsk(skb)->tun->prot->ops;

        has_frags = ops->fraglen_left(skb, NULL, &fraglen_max, NULL);
        return (has_frags ? fraglen_max : 0);
}

static inline int
bhtp_has_frags(struct sk_buff *skb)
{
        int has_frags = 0;
        struct bhtp_tun_ops *ops = bhtp_qsk(skb)->tun->prot->ops;

        has_frags = ops->fraglen_left(skb, NULL, NULL, NULL);
        return has_frags;
}


/* Queueing functions for bflow->tosend queue
 *
 * Versions with preceeding underscores __ need to be protected by
 * bflow->tosend_lock.
 */

static inline int
__bhtp_enqueue_skb(struct bhtp_flow *bflow, struct sk_buff *tskb,
		ktime_t latency)
{
	int rc = 0;

	skb_queue_tail(&bflow->tosend, tskb);
	bflow->tosend_accumulated += tskb->len;
        // create and enqueue a new "send" request for this skb
	rc = mgrp_enqueue_req(bflow, tskb, latency);
	if (rc)
		goto rollback;

	DEBUG("enqueue_skb: added %d acc %lld",
		tskb->len,
		bflow->tosend_accumulated);

	DEBUG("skb QUEUED %p len %d accumulated %lld rc %d elapsed %lld",
		tskb, tskb->len,
		bflow->tosend_accumulated, rc,
		ktime_to_ns(ktime_sub(tick_now(), bhtp_qsk(tskb)->kt)));

	return 0;

rollback:
	bflow->tosend_accumulated -= tskb->len;
	skb_dequeue_tail(&bflow->tosend);
	return rc;
	
}

static inline int
bhtp_enqueue_skb(struct bhtp_flow *bflow, struct sk_buff *tskb,
		ktime_t latency)
{
	int rc = 0;

	spin_lock_bh(&bflow->tosend_lock);
	rc = __bhtp_enqueue_skb(bflow, tskb, latency);
	spin_unlock_bh(&bflow->tosend_lock);
	return rc;
}


static inline struct sk_buff *
__bhtp_dequeue_skb_only(struct bhtp_flow *bflow, struct sk_buff *this_only)
{
	struct sk_buff *skb = NULL;
	int removed = 0;

	skb = skb_dequeue(&bflow->tosend);
	if (skb == NULL)
		goto done;

	if (this_only && skb != this_only)
		goto put_back;

        // not including the headers because their number depends
        // on the number of fragments that will be created
        // XXX is this byte accounting close enough?
        // XXX is it still used?
	removed = bhtp_fragdata_left(skb);
	bflow->tosend_accumulated -= removed;
	DEBUG("dequeue_skb: removed %d acc %lld",
		removed,
		bflow->tosend_accumulated);
	if (bflow->tosend_accumulated < 0)
		ERR("*** bhtp_dequeue_skb: negative accumulation %lld len %d",
			bflow->tosend_accumulated, skb->len);
done:
	return skb;

put_back:
	ERR("dequeue_skb_only: asking for %p %lld got %p %lld accumulated %lld",
		this_only, bhtp_qsk(this_only)->send.req->idg,
		skb, bhtp_qsk(skb)->send.req->idg,
		bflow->tosend_accumulated);
	skb_queue_head(&bflow->tosend, skb);
	skb = NULL;
	goto done;
}

static inline struct sk_buff *
bhtp_dequeue_skb_only(struct bhtp_flow *bflow, struct sk_buff *this_only)
{
	struct sk_buff *skb = NULL;

	spin_lock_bh(&bflow->tosend_lock);
	skb = __bhtp_dequeue_skb_only(bflow, NULL);
	spin_unlock_bh(&bflow->tosend_lock);

	return skb;
}

static inline struct sk_buff *
__bhtp_dequeue_skb(struct bhtp_flow *bflow)
{
	return __bhtp_dequeue_skb_only(bflow, NULL);
}

static inline struct sk_buff *
bhtp_dequeue_skb(struct bhtp_flow *bflow)
{
	struct sk_buff *skb = NULL;

	spin_lock_bh(&bflow->tosend_lock);
	skb = __bhtp_dequeue_skb(bflow);
	spin_unlock_bh(&bflow->tosend_lock);

	return skb;
}

static inline void
__bhtp_requeue_skb(struct bhtp_flow *bflow, struct sk_buff *skb)
{
        // not including the headers because their number depends
        // on the number of fragments that will be created
        // XXX is this byte accounting close enough?
        // XXX is it still used?
	int added = bhtp_fragdata_left(skb);

	skb_queue_head(&bflow->tosend, skb);
	bflow->tosend_accumulated += added;
	DEBUG("requeue_skb: added %d acc %lld skblen %d",
		added,
		bflow->tosend_accumulated,
		skb->len);
}

enum {
        MGRP_FLOWQ_INVALID = 0,
        MGRP_FLOWQ_TOSEND,
        MGRP_FLOWQ_TOSEND_PBK,
};

static inline struct sk_buff *
__mgrp_flow_dequeue_skb_only(struct bhtp_flow *bflow,
                             struct sk_buff *this_only,
                             int queue_type)
{
        struct sk_buff *skb = NULL;

        if (queue_type == MGRP_FLOWQ_TOSEND) {
                skb = __bhtp_dequeue_skb_only(bflow, this_only);

        } else if (queue_type == MGRP_FLOWQ_TOSEND_PBK) {
                skb = skb_dequeue(&bflow->tosend_pbk);
                if (this_only != NULL && skb != this_only) {
                        // put back
                        skb_queue_head(&bflow->tosend_pbk, skb);
                        skb = NULL;
                }

        } else {
                ERRFN("unknown queue_type: %d", queue_type);
        }
        return skb;
}

static inline void
__mgrp_flow_requeue_skb(struct bhtp_flow *bflow, struct sk_buff *skb,
                        int queue_type)
{
        if (queue_type == MGRP_FLOWQ_TOSEND) {
                __bhtp_requeue_skb(bflow, skb);

        } else if (queue_type == MGRP_FLOWQ_TOSEND_PBK) {
                skb_queue_head(&bflow->tosend_pbk, skb);

        } else {
                ERRFN("unknown queue_type: %d", queue_type);
        }
}


static inline struct sk_buff *
__mgrp_flow_dequeue_skb(struct bhtp_flow *bflow, int queue_type)
{
        return __mgrp_flow_dequeue_skb_only(bflow, NULL, queue_type);
}


// Enum that describes if a rider can piggyback on a probe

enum mgrp_rider_status {
        MGRP_RIDER_REJECTED = 0,
        MGRP_RIDER_ASIS,
        MGRP_RIDER_STANDALONE_CHUNK,
        MGRP_RIDER_FRAGMENT_CHUNK,
};

static inline enum mgrp_rider_status
mgrp_skb_can_be_standalone_rider(struct bhtp_flow *bflow,
                                 struct sk_buff *skb, int avail_pad)
{
        // assumes that the skb is willing to be a rider

        if (!bhtp_skb_is_fragmentable(skb)) {
                /* accepted: rider not fragmentable but fits inside padding */
                if (skb->len <= avail_pad)
                        return MGRP_RIDER_ASIS;

                /* rejected: rider does not fit */
                return MGRP_RIDER_REJECTED;
        }

        /* accepted: no need to start fragmenting if it fits whole */
        if (!bhtp_skb_already_fragmented(skb) && skb->len <= avail_pad)
                return MGRP_RIDER_ASIS;
        
        /* rejected: we cannot fragment rider to fit in padding */
        if (!bhtp_skb_can_fragment_further(skb, avail_pad))
                return MGRP_RIDER_REJECTED;
        
        return MGRP_RIDER_STANDALONE_CHUNK;
}

static inline enum mgrp_rider_status
mgrp_skb_can_be_fragment_rider(struct bhtp_flow *bflow,
                               struct sk_buff *skb, int avail_pad)
{
        // always create a packet with a fragment header
        // (helps with duplication)

        // even if it is one byte we can always fragment a chunk
        return MGRP_RIDER_FRAGMENT_CHUNK;
}

static inline enum mgrp_rider_status
mgrp_skb_can_be_rider(struct bhtp_flow *bflow,
                      struct sk_buff *skb, int avail_pad)
{
        /* rejected: rider needs padding */
        if (bhtp_qsk(skb)->tun_pad > 0)
                return MGRP_RIDER_REJECTED;

        if (mgrp_policy_standalone(bflow->pbk_policy))
                return mgrp_skb_can_be_standalone_rider(bflow, skb, avail_pad);

        else if (mgrp_policy_fragment(bflow->pbk_policy))
                return mgrp_skb_can_be_fragment_rider(bflow, skb, avail_pad);

        else {
                ERRFN("unknown fragment policy: %d", bflow->pbk_policy);
                return MGRP_RIDER_REJECTED;
        }
}

static inline u32
mgrp_flow_assign_get_fragid(struct bhtp_flow *bflow, struct sk_buff *skb)
{
        int rc = 0;
        u32 fragid = 0;

        rc = mgrp_tskb_fragid_get(skb, &fragid);
        if (rc) {
                fragid = atomic_inc_return(&bflow->fragid);
                mgrp_tskb_fragid_set(skb, fragid);
        }

        return fragid;
}

void
mgrp_reset_fragid_in_all_flows(void)
{
        struct bhtp_flow *bflow = NULL;

        spin_lock_bh(&bhtp.lock);
        list_for_each_entry(bflow, &bhtp.flows, flist) {
                atomic_set(&bflow->fragid, 0);
        }
        spin_unlock_bh(&bhtp.lock);
}

void
mgrp_reset_counters_in_all_flows(void)
{
        struct bhtp_flow *bflow = NULL;

        spin_lock_bh(&bhtp.lock);
        list_for_each_entry(bflow, &bhtp.flows, flist) {
                memset(&bflow->counters, 0, sizeof(bflow->counters));
                bflow->counters.epoch = bflow->id;
        }
        spin_unlock_bh(&bhtp.lock);
}


// returns 0 if not piggybacking allowed
// return 1 for packets inside a packet train that fit within maxpbk_percent
// returns 2 for lone probes, piggybacking always allowed regardless of percent
static inline int
mgrp_flow_policy_allows_piggybacking(struct bhtp_flow *bflow,
                                     struct sk_buff *vessel)
{
        int train_percent = 100;
        struct bhtp_qskb *qskb = bhtp_qsk(vessel);
        // int pad = qskb->tun_pad;

        BUG_TRAP(qskb->type & BHTP_QSKB_VESSEL);

        // user through sendmsg call disallows piggybacking
        if (!qskb->tun_piggyback)
                return 0;

        if (bhtp.maxpbk_percent == 0)
                return 0;

        // no barrier set, each probe considered its own train
        else if (qskb->vessel.probes_in_train == 0)
                return 2;

        else if (bhtp.maxpbk_percent == 100)
                return 1;

        train_percent *= qskb->vessel.probe_pos_in_train;
        train_percent /= qskb->vessel.probes_in_train;
        return (train_percent <= bhtp.maxpbk_percent);
}

static inline void
__mgrp_flow_mark_fragment_policy(struct bhtp_flow *bflow,
                                 struct sk_buff *original,
                                 struct sk_buff *pbked)
{
        if (mgrp_policy_standalone(bflow->pbk_policy)) {
                // STANDALONE policy: each TUN chunk is a valid packet

        } else if (mgrp_policy_fragment(bflow->pbk_policy)) {
                // FRAGMENT policy: all work done in mgrp_fragment_chunk()

        } else {
                ERRFN("unknown fragment policy: %d", bflow->pbk_policy);
        }
}

// Marks the "duplicate" flag of the skbs based on flow policy
static inline void
__mgrp_flow_mark_redundant_policy(struct bhtp_flow *bflow,
                                  struct sk_buff *original,
                                  struct sk_buff *pbked)
{
        if (mgrp_policy_sendonce(bflow->pbk_policy)) {
                // SENDONCE policy: do not duplicate packet
                //
                // This policy sends each transport segment exactly once,
                // either piggybacked on a probe or normally over MGRP.

                // nothing to do, leave the duplicate flag empty

        } else if (mgrp_policy_redundant(bflow->pbk_policy)) {
                u32 fragid = 0;

                // REDUNDANT policy: send again normally
                //
                // This policy sends all transport segments normally,
                // regardless of the portion of segments that went
                // out piggybacked.
                //
                // So all piggybacked segments are sent twice.

                // We need to connect the original and the piggybacked
                // duplicates with a common ID. We will use the fragid
                // that is also used in fragmenation.

                // assign an ID to the original if not already there
                fragid = mgrp_flow_assign_get_fragid(bflow, original);

                // mark both skbs as duplicates
                mgrp_tskb_type_set(original, MGRP_QSKB_DUPLICATE);
                mgrp_tskb_type_set(pbked, MGRP_QSKB_DUPLICATE);

                // @fragid already assigned to original
                mgrp_tskb_fragid_set(pbked, fragid);

                // XXX HERE neet to change the qskb->tun_proto=FRAG
                // XXX HERE and prepend and fill the fragmenation header

                DEBUG_DUPSKB("marked original:", original);
                DEBUG_DUPSKB("marked pbked   :", pbked);

                // This has two effects:
                //
                // 1. The TUN header of both the original and the piggybacked
                //    fragments are marked as "duplicate" on send.
                //
                // 2. The original skb is requeued so that it will be resent
                //    in __mgrp_flow_handle_fully_piggybacked_skb()

        } else {
                ERRFN("unknown redundancy policy: %d", bflow->pbk_policy);
        }
}

// Marks the skbs based on flow policy (the flags alter sending behavior)
static inline void
__mgrp_flow_mark_policy(struct bhtp_flow *bflow,
                        struct sk_buff *original,
                        struct sk_buff *pbked)
{
        __mgrp_flow_mark_fragment_policy(bflow, original, pbked);
        __mgrp_flow_mark_redundant_policy(bflow, original, pbked);
}

/* Deal with a transport skb (that was in the tosend queue) once it
 * has been fully piggybacked on a probe.
 *
 * Each skb has a request associated with it that is created
 * when the transport protocol hands the skb to MGRP.
 * The request and the skb are linked with backpointers.
 * All requests are stored in the global bhtp.req_list and
 * are ordered by increasing deadline. Once the deadline is
 * reached the request/skb is sent using bhtp_send_req()
 *
 * The way that this request is updated/handled determines
 * also the piggybacking policy.
 *
 * NOTE: Must be called with bflow->tosend_lock held
 * NOTE: The skb is no longer needed. If you do not use it, free it
 */
static inline void
__mgrp_flow_handle_fully_piggybacked_skb(struct bhtp_flow *bflow,
                                         struct sk_buff *skb)
{
        int sent = 0;
        int piggybacked = 0;
        struct bhtp_flow_req *req = bhtp_qsk(skb)->send.req;

        // check backpointers
        BUG_ON(req->skb != skb);

        // mark it as "fully_piggybacked"
        piggybacked = atomic_inc_return(&req->fully_piggybacked);
        if (piggybacked != 1)
                ERRFN("req %lld expects req->fully_piggybacked to be 1 but got %d",
                      req->idg, piggybacked);

        if (mgrp_tskb_has_type(skb, MGRP_QSKB_DUPLICATE)) {
                DEBUG_DUP("skb %d is duplicate, reset fragmentation", skb->len);

                // Reset the bhtp_qskb fragmentation state so that
                // fragmentation (this time to send the packet out
                // in bhtp_transmit_lone_skb) starts again from the top.
                mgrp_skb_reset_fragmentation(skb);
                
                // Add the skb to the tosend_pbk queue, which contains skbs
                // that have alread been piggybacked. When the request
                // is serviced in bhtp_send_req() it will know which
                // queue to search for the skb (based on req->piggybacked)
                skb_queue_tail(&bflow->tosend_pbk, skb);

        } else {
                // no duplicates and skb already sent (in piggybacked fragments)
                
                // discard skb, piggybacked fragments have been carved out
                kfree_skb(skb);
                req->skb = NULL;

                // cancel request to send original skb by marking it as "sent"
                // when the request is serviced, it will just ignore the skb
                sent = atomic_inc_return(&req->sent);
                if (sent != 1)
                        ERRFN("req %lld expects req->sent to be 1 but got %d",
                              req->idg, sent);
        }
}

// Dequeues an skb from one of the tosend queues that must match
// the skb in the request @req. Applies piggybacking policy.
static inline struct sk_buff *
__mgrp_flow_dequeue_req_skb(struct bhtp_flow_req *req) 
{
        struct sk_buff *skb = NULL;
        int fully_piggybacked = atomic_read(&req->fully_piggybacked);
        int queue_type = (fully_piggybacked ? MGRP_FLOWQ_TOSEND_PBK :
                                              MGRP_FLOWQ_TOSEND);

        // dequeue the skb (from the appropriate queue) only if it matches
        skb = __mgrp_flow_dequeue_skb_only(req->bflow, req->skb, queue_type);
        if (!skb) {
                /* not ok: front of queue does not have skb that we expect */
                ERRBUG("send_req: req %lld for missing skb %p sent %d",
                        req->idg, req->skb, atomic_read(&req->sent));
                return NULL;
        }

        // When we are sending a request skb, we always send all of its
        // remaining fragments. If we have already sent some fragments,
        // then this means those fragments are piggybacked.
        //
        // In case of redundant piggybacking policies reset the fragmenation
        // so that you can send the whole skb.
        if (mgrp_policy_redundant(req->bflow->pbk_policy) && !fully_piggybacked)
                mgrp_skb_reset_fragmentation(skb);

        return skb;
}

static inline struct sk_buff *
__mgrp_flow_requeue_req_skb(struct bhtp_flow_req *req) 
{
        struct sk_buff *skb = NULL;
        int fully_piggybacked = atomic_read(&req->fully_piggybacked);
        int queue_type = (fully_piggybacked ? MGRP_FLOWQ_TOSEND_PBK :
                                              MGRP_FLOWQ_TOSEND);

        // requeue the skb (in the appropriate queue)
        __mgrp_flow_requeue_skb(req->bflow, req->skb, queue_type);

        return skb;
}

static inline struct sk_buff *
mgrp_create_rider(struct bhtp_flow *bflow,
                  struct sk_buff *skb, 
                  int max_riderlen,
                  int rider_status,
                  int copy_data,
                  int *skb_exhausted)
{
        struct sk_buff *rider = NULL;

        if (rider_status == MGRP_RIDER_REJECTED) {
                // the minimum fragment does not fit in the padding
                *skb_exhausted = 0;
                goto done;

        } else if (rider_status == MGRP_RIDER_ASIS) {
                // use the whole skb as a rider
                // copy both the skb and data if you are going to mark the packet data
                // otherwise use cheaper skb_clone()
                if (copy_data)
                        rider = skb_copy(skb, GFP_ATOMIC);
                else
                        rider = skb_clone(skb, GFP_ATOMIC);
                if (rider == NULL) {
                        ERRFN("failed to duplicate (clone or copy) skb len %d", skb->len);
                        goto done;
                }
                BUG_TRAP(skb->len == rider->len);
                *skb_exhausted = 1;

        } else if (rider_status == MGRP_RIDER_STANDALONE_CHUNK) {
                struct bhtp_tun_ops *ops = bhtp_qsk(skb)->tun->prot->ops;
                // create the largest fragment that can fit inside the padding
                // each fragment has a header and is a valid standalone packet
                int fraglen = min_t(int, bhtp_max_fraglen_left(skb),
                                         max_riderlen);

                rider =  ops->fragment(skb, fraglen, copy_data);
                if (rider == NULL) {
                        ERRFN("failed to create standalone chunk skb len %d into frag len %d",
                              skb->len, fraglen);
                        goto done;
                }
                BUG_TRAP(fraglen == rider->len);

                // does the skb still have fragments after we
                // extracted the most recent fragment?
                *skb_exhausted = !bhtp_has_frags(skb);

        } else if (rider_status == MGRP_RIDER_FRAGMENT_CHUNK) {
                // create the largest fragment that can fit inside the padding
                // each fragment has a fragmenation header so that it can be
                // reassembled
                int chunklen = min_t(int, mgrp_fragment_chunk_left(skb),
                                          max_riderlen);

                // Assign a fragmentation ID if one is not already
                // assigned. This ensures that the original skb will
                // have a valid FRAGID every time we generate a fragment.
                // We ignore the return value (FRAGID) because we do not
                // need it here; we only care for the side effects.
                mgrp_flow_assign_get_fragid(bflow, skb);

                // Create the rider chunk out of the unused portion of
                // the original skb. The fragments will pull the FRAGID
                // from the original skb that we set above.
                rider = mgrp_fragment_chunk(skb, chunklen, copy_data);
                if (rider == NULL) {
                        ERRFN("failed to create fragment chunk skb len %d into frag len %d",
                              skb->len, chunklen);
                        goto done;
                }
                BUG_TRAP(chunklen == rider->len);

                // does the skb still have fragments after we
                // extracted the most recent fragment?
                *skb_exhausted = (mgrp_fragment_chunk_left(skb) == 0);

        } else {
                ERRFN("unknown rider status %d", rider_status);
                goto done;
        }
done:
        return rider;
}

static inline struct sk_buff *
mgrp_find_remaining_rider(struct bhtp_flow *bflow)
{
        struct sk_buff *rider = NULL;
        struct sk_buff *skb = NULL;
        int fully_piggybacked_skb = 0;

        // always copy the skb data for now (inefficient but simpler)
        int copy_data = 1;

        // only applies to _FRAGMENT_ policies, because fragments are
        // useless without the last one (the _STANDALONE_ policies
        // generate valid chunks and we have no problem).
        if (!mgrp_policy_fragment(bflow->pbk_policy))
                return NULL;

        spin_lock_bh(&bflow->tosend_lock);

        // get the next skb from the tosend queue
        skb = __bhtp_dequeue_skb(bflow);
        // no transport traffic to piggyback on
        if (!skb)
                goto done;
        // the skb at the head of the queue has not had any fragments taken
        if (!mgrp_fragment_chunk_started(skb)) {
                __bhtp_requeue_skb(bflow, skb);
                goto done;
        }

        // create one rider that contains the last fragment
        rider = mgrp_create_rider(bflow, skb,
                                  mgrp_fragment_chunk_left(skb),
                                  MGRP_RIDER_FRAGMENT_CHUNK,
                                  copy_data,
                                  &fully_piggybacked_skb);
        if (rider == NULL)
                goto done;

        // the original skb must be exhausted
        BUG_TRAP(fully_piggybacked_skb);

        // mark both the original skb and the rider (piggybacked fragments)
        // based on the piggybacking policy (duplicates, fragments, etc)
        __mgrp_flow_mark_policy(bflow, skb, rider);

        // the whole skb has been piggybacked:
        //      - it has already been removed from the tosend queue
        //      - depending on our piggybacking policy decide what
        //        to do with the original skb (eg. resend it normally)
        //
        // NOTE: skb no longer needed; ownership passed to the callee
        __mgrp_flow_handle_fully_piggybacked_skb(bflow, skb);

        // original skb is no longer valid beyond this point
done:
        spin_unlock_bh(&bflow->tosend_lock);
        return rider;
}

static inline struct sk_buff *
bhtp_find_rider_skb(struct bhtp_flow *bflow,
                struct sk_buff *vessel, int *padding,
                ktime_t *rider_orig_send_time)
{
        struct sk_buff *rider = NULL;
        struct sk_buff *skb = NULL;
        int pad = bhtp_qsk(vessel)->tun_pad;
        int fully_piggybacked_skb = 0;
        enum mgrp_rider_status rider_status = 0;
        struct bhtp_flow_req *req = NULL;

        // always copy the skb data for now (inefficient but simpler)
        int copy_data = 1;

        if (!mgrp_flow_policy_allows_piggybacking(bflow, vessel)) {
                DEBUGFN("no pbk for probe %d out of %d percent %d",
                        bhtp_qsk(vessel)->vessel.probe_pos_in_train,
                        bhtp_qsk(vessel)->vessel.probes_in_train,
                        bhtp.maxpbk_percent);
                goto done_unlocked;
        }

        spin_lock_bh(&bflow->tosend_lock);

        // get the next skb from the tosend queue
        skb = __bhtp_dequeue_skb(bflow);
        if (!skb)
                // no transport traffic to piggyback on
                goto done;

        req = bhtp_qsk(skb)->send.req;

        // is the skb suitable to be a rider?
        rider_status = mgrp_skb_can_be_rider(bflow, skb, pad);

        // create it depending on policy
        rider = mgrp_create_rider(bflow, skb, pad, rider_status, copy_data,
                                  &fully_piggybacked_skb);
        if (rider == NULL)
                goto done;

        // at this point we have found a rider skb and we
        // need to decide what to do with the original skb

        // Return the original send time for the packet that
        // contains the rider. This information can be used
        // by the caller to determine whether the piggybacked
        // packet would have been sent in-between the probes
        // (if it was not piggybacked).
        *rider_orig_send_time = req->send_kt;

        // account for the portion of the padding reused by the rider
        pad -= rider->len;

        // mark both the original skb and the rider (piggybacked fragments)
        // based on the piggybacking policy (duplicates, fragments, etc)
        __mgrp_flow_mark_policy(bflow, skb, rider);

        if (!fully_piggybacked_skb) {
                // put the skb back so the next probe can continue piggybacking
                __bhtp_requeue_skb(bflow, skb);

        } else {
                // the whole skb has been piggybacked:
                //      - it has already been removed from the tosend queue
                //      - depending on our piggybacking policy decide what
                //        to do with the original skb (eg. resend it normally)
                //
                // NOTE: skb no longer needed; ownership passed to the callee
                __mgrp_flow_handle_fully_piggybacked_skb(bflow, skb);

                // original skb is no longer valid beyond this point
        }
done:
        BUG_TRAP(pad >= 0);
        spin_unlock_bh(&bflow->tosend_lock);
done_unlocked:
        // return how much padding remains after the piggybacking
        *padding = pad;
        return rider;
}

static inline int
bhtp_transmit_vessel_skb(struct bhtp_flow *bflow, struct sk_buff *vessel,
                         struct mgrp_probe_feedback *feedback)
{
	int rc = 0;
	int append_padbytes = 0;
	struct sk_buff *rider = NULL;
        int rider_bytes = 0;
        ktime_t rider_orig_send_time = KTIME_ZERO;

	int tuns = 0;
	struct sk_buff *tskbs[2];

	/* the vessel must come first */
	/* quirk of the implementation */
	/* see udpex_tun_rcv() /left in rcv skb/ */
	tskbs[tuns] = vessel;
	tuns++;

        // The vessel skb has some padding that can be reused to
        // accomadate a rider skb. The @append_padbytes is the portion
        // of the padding that was not reused (unused) and needs to
        // be appended to the MGRP packet.
	rider = bhtp_find_rider_skb(bflow, vessel, &append_padbytes, 
                                    &rider_orig_send_time);

        // XXX HERE here is where we would find out from mgrp if the rider
        //     is taken ahead of time (and hence should not be counted
        //     to correct pathload accuracy)

	if (rider) {
                DEBUG_DUPSKB("rider", rider);
		tskbs[tuns] = rider;
		tuns++;
                rider_bytes = rider->len;
	}

        // bhtp_transmit_skb() resets and fills @feedback->mgrp_bytes
	rc = bhtp_transmit_skb(bflow, append_padbytes, tskbs, tuns, feedback);
	if (rc < 0)
		/* the vessel is not dropped; handled by the caller */
		goto drop_rider;
	
        if (feedback) {
                feedback->pbk_bytes = rider_bytes;
                feedback->rider_orig_send_time = rider_orig_send_time;
        }

	return rc;

drop_rider:
	ERR("xmit_vessel_skb: DROPPED rider skb %d", rider->len);
	kfree_skb(rider);
	return rc;
}

static inline int
bhtp_transmit_lone_skb(struct bhtp_flow *bflow, struct sk_buff *tun_skb,
                       struct mgrp_probe_feedback *feedback)
{
	int rc = 0;
        int fragid = 0;
	struct bhtp_qskb *qskb = bhtp_qsk(tun_skb);

	DEBUG("schedule:    frag_max %d len %d pad %d ops %d",
			qskb->tun_frag_len_max,
			tun_skb->len,
			qskb->tun_pad,
			BHTP_TUN_PROT_HAS_FRAG(qskb->tun->prot));

	if (bhtp_skb_needs_fragment_transmit(tun_skb)) {
                // skb has not been fully piggybacked yet but time to go

                // STANDALONE policy where the skb has been partially sent
                // we need to send out the remaining standalone chunk
		rc = bhtp_fragment_transmit_skb(bflow, tun_skb,
				qskb->tun_frag_len_max);

                // does not return @feedback because it may
                // send multiple packets

        } else if (mgrp_fragment_chunk_started(tun_skb)) {
                // skb has not been fully piggybacked yet but time to go

                // FRAGMENT policy
                // We have sent some parts of this packet as fragments
                // piggybacked inside probes but its deadline has been
                // reached with bhtp_send_req(). Since we are not sending
                // any duplicates, just send the remaining fragment.
                // we reset fragmentation for duplicates in:
                //      __mgrp_flow_dequeue_req_skb()


                int fully_pbked_skb = 0;
                struct sk_buff *last = NULL;
                // need to copy because we are only taking a fragment
                int copy_data = 1;

                last = mgrp_create_rider(bflow, tun_skb,
                                         mgrp_fragment_chunk_left(tun_skb),
                                         MGRP_RIDER_FRAGMENT_CHUNK,
                                         copy_data,
                                         &fully_pbked_skb);
                BUG_TRAP(last != NULL);
                BUG_TRAP(fully_pbked_skb);

                // we are done with the original skb
                kfree_skb(tun_skb);

                rc = bhtp_transmit_skb(bflow, 0, &last, 1, feedback);

	} else if (mgrp_tskb_fragid_get(tun_skb, &fragid) != -1) {
                // skb has been fragmented
                // its fragments may not have fully been sent
                // but we have a duplicate policy and we have to resend
                // the whole thing since it was not caught above

                // The skb has been assigned a fragment ID, which means
                // that it has been also sent in fragments and this is
                // probably a duplicate (instead of just the remaining
                // fragment, otherwise it would have been caught above).
                // We need to encapsulate it insidea fragmentation header
                // so that the packet will go through the reassembly code.

                DEBUG_FRAG("needs_fraghdr fragid %d len %d",
                           fragid, tun_skb->len);

                rc = mgrp_fragment_chunk_encapsulate(tun_skb, fragid);
                if (rc)
                        goto done;
		rc = bhtp_transmit_skb(bflow, 0, &tun_skb, 1, feedback);

        } else {
		rc = bhtp_transmit_skb(bflow, 0, &tun_skb, 1, feedback);
        }
done:
	return rc;
}


int
mgrp_flush_piggybacking(struct bhtp_flow *bflow,
                        struct mgrp_probe_feedback *feedback)
{
        int rc = 0;
        int append_padbytes = 0;
        struct sk_buff *rider = NULL;

	int tuns = 0;
	struct sk_buff *tskbs[2];
        
        rider = mgrp_find_remaining_rider(bflow);
        if (rider == NULL)
                goto done;

        tskbs[tuns] = rider;
        tuns++;

        DEBUG_FRAG("sending rider len %d", rider->len);

        // the rider goes out by itself, no padding, no probe
        // the idea is not to leave in the tosend queue
        // half-piggybacked skbs after we send a probe burst
        rc = bhtp_transmit_skb(bflow, append_padbytes, tskbs, tuns, feedback);
	if (rc < 0)
		goto drop_rider;

done:
	return rc;

drop_rider:
	ERRFN("DROPPED rider skb %d", rider->len);
	kfree_skb(rider);
	return rc;
}


void
bhtp_send_req(struct bhtp_flow_req *req)
{
	int rc = 0;
	int sent = 0;
	struct sk_buff *skb = NULL;
        struct mgrp_probe_feedback feedback = { 0 };

	spin_lock_bh(&req->bflow->tosend_lock);

	sent = atomic_inc_return(&req->sent);
	if (sent != 1) {
		/* probably ok: skb swept under our feet */
		/* either by a probe or by draining the queue */
		DEBUG("send_req: req %lld already sent %d skb %p ",
			req->idg, sent, req->skb);
		goto done;
	}

        // dequeue the skb (from the appropriate queue) only if it matches
        skb = __mgrp_flow_dequeue_req_skb(req);
        if (skb == NULL)
                goto done;

	DEBUG("send_req: sending req %lld skb %p sent %d",
		req->idg, req->skb, atomic_read(&req->sent));

        DEBUG_LONESKB(req->skb);
	rc = bhtp_transmit_lone_skb(req->bflow, req->skb, &feedback);

	if (rc < 0) {
		ERR("send_req: req %lld dropping skb %p %d rc %d",
			req->idg, req->skb, req->skb->len, rc);
		kfree_skb(skb);

	} else if (rc > 0) {
                /* This is probably a NET_XMIT_DROP (see dev_queue_xmit) which
                 * means that the packet is dropped by the transmit queue.
                 */
		ERR("send_req: req %lld skb %p %d rc %d when transmitting, how to handle?",
			req->idg, req->skb, req->skb->len, rc);
	}

done:
	spin_unlock_bh(&req->bflow->tosend_lock);
	return;
}


static inline ktime_t
mgrp_flow_latency(struct bhtp_flow *bflow)
{
	ktime_t global_latency = mgrp_global_flow_latency(bflow);
	ktime_t local_latency = mgrp_local_flow_latency(bflow);

	/* only one latency should be defined */
	if (ktime_to_ns(global_latency) > 0) {
		if (ktime_to_ns(local_latency) > 0)
			ERR("flow_latency: "
			    "both latencies defined: global %lld local %lld ns",
			    ktime_to_ns(global_latency),
			    ktime_to_ns(local_latency));
		return global_latency;
	}

	return local_latency;
}


/* SKB is headerless - no BHTP header yet */
/* may return NET_XMIT_CN which is not an error */
/* may return BHTP_QUEUED  which is not an error */
/*
 * >0	tun_skb is queued at the BHTP layer, tun_skb is not freed
 * 0	tun_skb transmitted (passed to IP) and tun_skb is freed.
 * <0	tun_skb is not freed and a return code rc returned.
 */
int
bhtp_schedule_skb(struct bhtp_flow *bflow, struct sk_buff *tun_skb,
                  struct mgrp_probe_feedback *feedback)
{
	int rc = 0;
	struct bhtp_qskb *qskb = bhtp_qsk(tun_skb);
	int is_vessel = qskb->type & BHTP_QSKB_VESSEL;
	ktime_t latency = mgrp_flow_latency(bflow);
        ktime_t now = tick_now();
	int send_now = ktime_to_ns(latency) == 0 ||
			is_vessel ||
			qskb->type & BHTP_QSKB_NEVER_QUEUE;
#ifdef MGRP_TSKBSTATS
        struct mgrp_tskbstats_entry stats = {
                .ts             = ktime_to_timeval(now),
                .skb_len        = tun_skb->len,
                .latency        = ktime_set(0, 0),
                .qskb           = *qskb,
                .code           = '?',
        };
#endif

	LOG_TCPHDR_SKB(tun_skb);

	qskb->kt = now;

/* 	XXX HERE */
/* 	When probe is waiting, set the flow_latency */
/* 	when the probe is sent (nothing waiting) */
/* 		if bhtp.latency non zero, empty the payload queue */
/* 		set the flow_latency to zero */

        DEBUGFN("vessel %d send_now %d len %d",
                is_vessel, send_now, tun_skb->len);

        // if the pktsize is not overwriten, the -1 will tell the
        // caller that the packet was queued and not sent
        // (that is an MGRP packet was not built and not pushed to IP)
        if (feedback != NULL)
                mgrp_probe_feedback_init(feedback, -1);

	if (!send_now)
		rc = bhtp_enqueue_skb(bflow, tun_skb, latency);
	else if (is_vessel)
		rc = bhtp_transmit_vessel_skb(bflow, tun_skb, feedback);
	else {
                DEBUG_LONESKB(tun_skb);
		rc = bhtp_transmit_lone_skb(bflow, tun_skb, feedback);
        }

#ifdef MGRP_TSKBSTATS
        // log each skb that comes in
        if (!send_now) {
                stats.code = MGRP_TSKBCODE_QUEUED;
                stats.latency = latency;
        } else if (is_vessel) {
                stats.code = MGRP_TSKBCODE_PROBENOW;
        } else {
                stats.code = MGRP_TSKBCODE_SENDNOW;
        }
        mgrp_tskbstats_add(bhtp.tskbstats, &stats);
#endif

	return rc;
}

static inline int
__mgrp_drain_tosend_queue(struct bhtp_flow *bflow, int drop, int queue_type)
{
	int skb_count = 0;
	struct sk_buff *skb = NULL;

	while ((skb = __mgrp_flow_dequeue_skb(bflow, queue_type)) != NULL) {
		int rc = 0;
		int sent = 0;
                struct mgrp_probe_feedback feedback = { 0 };
		struct bhtp_flow_req *req = bhtp_qsk(skb)->send.req;

		skb_count++;
		
		/* mark request as "sent"; see bhtp_send_req() */
		sent = atomic_inc_return(&req->sent);
		if (sent != 1) {
			/* not ok: already sent out: must be bug */
			ERRBUG("drain_tosend: req %lld already sent %d skb %p",
				req->idg, sent, req->skb);
			continue;
		}

		if (drop) {
			ERRFN("dropping skb %d", skb->len);
			kfree_skb(skb);
			continue;
		}

                DEBUG_LONESKB(skb);
		rc = bhtp_transmit_lone_skb(bflow, skb, &feedback);
		if (rc) {
			ERRFN("dropping skb %d rc %d", skb->len, rc);
			kfree_skb(skb);
			continue;
		}
	}

	return skb_count;
}


int
bhtp_drain_tosend(struct bhtp_flow *bflow, int drop)
{
	int skb_count = 0;

        spin_lock_bh(&bflow->tosend_lock);
        skb_count += __mgrp_drain_tosend_queue(bflow, drop, MGRP_FLOWQ_TOSEND_PBK);
        skb_count += __mgrp_drain_tosend_queue(bflow, drop, MGRP_FLOWQ_TOSEND);
        spin_unlock_bh(&bflow->tosend_lock);

        return skb_count;
}

int
bhtp_send_barrier_skbs(struct bhtp_flow *bflow,
		struct sock *sk, struct bhtp_tun *tun,
		unsigned long long *twait_p,
		unsigned long long *tsend_p)

{
	int rc = 0;

	unsigned long long started = tick_now_ns();
	unsigned long long twait = 0;
	unsigned long long tsend = 0;
	unsigned long long canwait = 0;

	int keep_waiting = 0;
	int scheduled = 0;
	int not_worthit =
		tun->barrier_total < tun->barrier_pkts * 400 ||
		tun->barrier_max < 400;

	canwait = tun->barrier_wait_usec;
	canwait *= 1000;

	LOG("send_barrier_skbs: wait %d pkts %d total %d min %d max %d",
		tun->barrier_wait_usec,
		tun->barrier_pkts,
		tun->barrier_total,
		tun->barrier_min,
		tun->barrier_max);

	if (not_worthit)
		goto send_now;

	/* FIXME: what about if two processes do this? */
	BUG_ON(1);
/* 	atomic_set(&bflow->queue_pkts, 1); */

	do {
		schedule();
		scheduled++;
		twait = tick_now_ns() - started;
		keep_waiting = 
			(bflow->tosend_accumulated < (tun->barrier_total/3)) &&
			(twait < canwait);
		DEBUG("twait %llu canwait %llu", twait, canwait);
	} while (keep_waiting);

	LOG("send_barrier_skbs: scheduled %d twait %lld accumulated %lld",
		scheduled, twait, bflow->tosend_accumulated);
send_now:
	tsend = tick_now_ns();
/* 	rc = tun->prot->ops->send_barrier_skbs(sk); */
	// XXX work is done in udpex_send_barrier_skbs
	// XXX renamed to udpex_transmit_barrier_skbs
	BUG();
	tsend = tick_now_ns() - tsend;
	
	if (rc || not_worthit)
		goto done;
	
	BUG_ON(1);
/* 	rc = bhtp_send_queued_pkts(bflow); */
done:
	if (twait_p)
		*twait_p = twait;
	if (tsend_p)
		*tsend_p = tsend;

	return rc;
}

#define BHTP_NOT_NULL_DROP(p) \
	do \
		if ((p) == NULL) { \
			ERR_TR("pointer %s is NULL", #p); \
			goto drop; \
		} \
	while (0)

static inline int
bhtp_tun_rcv(struct sk_buff *skb, struct sockaddr_in *remote_addr,
             int trim_above, int orig_padbytes,
             struct bhtphdr *bh, struct bhtphdr_tun *tun)
{
	struct sk_buff *tskb = NULL;
	struct bhtp_tun_prot *prot = NULL;
	struct bhtp_qskb *qskb = NULL;
	int rc = 0;

        BHTP_NOT_NULL_DROP(skb);

	/*
	 * Create a new skb that contains only the TUN segment:
	 *
	 *	- clone the bhtp skb; bhtp header has been already pulled
	 *	- remove any TUN segments above our segment
	 *	- keep only our TUN segments by removing segments below
	 *	- reposition ->transport_header header
	 *	- hand to upper transport (UDPex, TCP, etc).
	 *
	 * Be careful to only operate on the cloned skb fields.
	 * Do not overwrite any of the skb head data or any of the
	 * fragments since all TUNS share the data. That is, even
	 * if it is beyond skb->tail, another TUN may use it.
	 * Very true when TUNs are stacked inside a BHTP packet.
	 *
	 * See more in udpex_tun_rcv() notes about "recreate the padding":
	 *
	 *	cd bhtp
	 *	svn log -r261 .
	 */

        DEBUG_TR("cloning skb %d len", skb->len);
	tskb = skb_clone(skb, GFP_ATOMIC);
	if (trim_above)
		skb_pull(tskb, trim_above);
	skb_trim(tskb, tun->len);
	tskb->transport_header  = tskb->data;

	DEBUG("tun_rcv: proto %d len %d needs_padding %d duplicate %d",
		tun->proto, tun->len, tun->needs_padding, tun->duplicate);

	/* Demux the TUN segment based on L4 protocol and process it. */
	prot = bhtp_lookup_proto(tun->proto);

	if (!BHTP_TUN_PROT_HAS_RCV(prot)) {
		ERR("tun_rcv: unknown proto %d or no rcv ops for prot %p",
			tun->proto, prot);
		goto drop_verbose;
	}

	/* overloading the control buffer in sk_buff for bhtp private data */
	qskb = bhtp_qsk_init(tskb);

	/* only a few fields filled; TUN rcv code should do the rest */
	qskb->tun_len = tskb->len;

        // The padding in a BHTP packet must be assigned to exactly one
        // of the TUNs. It is up to the TUN protocol to deal with the padding.
        if (tun->needs_padding) {
                if (orig_padbytes == 0) {
                        ERR("tun_rcv: TUN needs padding flag but no padbytes");
                        goto drop_verbose;
                }
                qskb->tun_pad = orig_padbytes;
        }

	/* CARE the bhtphdr pointer is assumed to point inside the skb */
	/* CARE it must remain valid while qskb is valid (ie within BHTP) */
	/* CARE This is used to retrieve timestamp in bhtp_recv_get_rcvstamp() */
	qskb->recv.bh = bh;

        if (tun->duplicate) {
                mgrp_tskb_type_set(tskb, MGRP_QSKB_DUPLICATE);
                DEBUG_DUP("duplicate tun: proto %d len %d", tun->proto, tun->len);
        }

	/* hand over to the TUN rcv code */
	rc = prot->ops->rcv(tskb, remote_addr);
	if (rc) {
		ERR("tun_rcv: rcv ops error: %d", rc);
		goto drop_verbose;
	}

	return 0;

drop_verbose:
        log_hex(bh, sizeof(struct bhtphdr), 4, "bhtphdr: ");
        LOG("tun_rcv: ------------------");
        log_hex(tskb->transport_header, min(20, 21), 4, "tundata: ");
        ERR("tun_rcv: TUN segment DROPPED, proto %d, len %d, trim_above %d",
            tun->proto, tun->len, trim_above);
drop:
	kfree_skb(tskb);
	return 0;
}

int
bhtp_rcv(struct sk_buff *skb)
{
	struct bhtphdr *bh = NULL;
	int processed = 0;
	int t = 0;
	int tuns = 0;
        int bhtp_pktlen = 0;
        int bhtp_payload_len = 0;
        const char *errstr = "unknown error";
        const struct iphdr *iph= (struct iphdr *)skb->network_header;

        struct bhtp_flow *bflow = NULL;
        struct sockaddr_in remote_addr = {
                .sin_family = AF_INET,
                .sin_addr.s_addr = iph->saddr,
        };

        int pad_reused = 0;
        int pad_unused = 0;
        int pad_orig = 0;

        char error[200];

        DEBUG("************************************");
	DEBUG("bhtp_rcv skb %p len %d", skb, skb->len);
        if (debug_on()) {
                int loglen = min_t(int, skb->len, 48);
                log_hex((const char *)skb->transport_header, loglen, 4, "rcv: ");
        }

	if (!pskb_may_pull(skb, sizeof(struct bhtphdr))) {
                // TCP with MGRP IP protocol bug
                if (mgrp_looks_like_tcp(skb)) {
                        mgrp_sendto_tcp_directly(skb);
                        goto done;
                }
		ERRFN("cannot pull BHTP header: skb_len %d (needs %d)",
			skb->len, sizeof(struct bhtphdr));
                errstr = "pkt too small for BHTP header";
		goto drop;
	}

	bh = (struct bhtphdr *)skb->transport_header;
	BHTP_NOT_NULL_DROP(bh);

        bhtp_pktlen = skb->len;
        pad_orig = bhtp_orig_pad_bytes(bh, bhtp_pktlen,
                                       &pad_reused, &pad_unused,
                                       error, sizeof(error));
        if (pad_orig == -1) {
                // TCP with MGRP IP protocol bug
                if (mgrp_looks_like_tcp(skb)) {
                        mgrp_sendto_tcp_directly(skb);
                        goto done;
                }
                ERRFN("invalid header: %s", error);
                errstr = "failed to extract padding information from header";
                goto drop_verbose;
        }

        DEBUG("rcv: BHTP pkt from IP len %d pad_unused %d pad_reused %d",
              bhtp_pktlen, pad_unused, pad_reused);

	if (ktime_to_ns(skb->tstamp) == 0)
		skb->tstamp = tick_now();

	/* reposition the L4 pointer below the BHTP header */
	skb->transport_header = skb_pull(skb, sizeof(struct bhtphdr));
	BHTP_NOT_NULL_DROP(skb->transport_header);

        // removed bhtp header, only bhtp payload left
        bhtp_payload_len = skb->len;

        // look for the flow (needed for receive stats)
        //
        //      - create if not there
        //      - mark it (bflow->recv_user) the first time

        bflow = bhtp_find_mark_recv_flow(&remote_addr, 0);
        if (bflow == NULL) {
                ERR("rcv: cannot create bhtp_flow for src: "
                    NIPQUAD_FMT,
                    NIPQUAD(remote_addr.sin_addr.s_addr));
                errstr = "failed to create bhtp_flow";
                goto drop_verbose;
        }

        // XXX change this lock to a new lock protecting the counters!
	spin_lock_bh(&bflow->tosend_lock);
        mgrp_counters_update_recv_side(&bflow->counters, bh, skb->len);
	spin_unlock_bh(&bflow->tosend_lock);

	/* extract and process each TUN segment */
	processed = 0;

	tuns = bhtp_header_tuns(bh);
	if (tuns <= 0 || tuns > BHTP_MAX_SEG) {
                ERR("rcv: invalid header_tuns %d", tuns);
                errstr = "invalid tuns in header";
                goto drop_verbose;
	}

        DEBUG("rcv: processing %d tuns", tuns);
	for (t = 0; t < tuns; t++) {
                int remaining = bhtp_payload_len - processed - pad_unused;

		/* sanity check */
		if (bh->tun[t].len > remaining) {
                        ERR("rcv: invalid len %d for TUN entry %d, only %d left",
                            bh->tun[t].len, t, remaining);
                        errstr = "invalid TUN len";
                        goto drop_verbose;
		}
                if (bh->tun[t].needs_padding) {
                        // the TUN needs to know how to reconstruct the probe
                        bhtp_tun_rcv(skb, &remote_addr,
                                     processed, pad_orig,
                                     bh, &bh->tun[t]);
                } else {
                        bhtp_tun_rcv(skb, &remote_addr,
                                     processed, 0,
                                     bh, &bh->tun[t]);
                }
		processed += bh->tun[t].len;
	}

	/* The whole packet needs to be accounted for through TUNs. */
	if (processed + pad_unused != bhtp_payload_len) {
                ERR("rcv: BHTP pkt not processed fully: "
                    "payloadbytes %d processed %d padbytes %d",
                    bhtp_payload_len, processed, pad_unused);
                errstr = "not all BHTP payload processed";
                goto drop_verbose;
	}

        // no need to check if padding fields make sense
        // already done in bhtp_orig_pad_bytes() higher up
	/* no more need for container skb; tun skbs are cloned */
	kfree_skb(skb);
done:
	/* always zero */
	return 0;

drop_verbose:
        log_hex((const char *)bh, min_t(int, 48, skb->len), 4, "drop: ");
        if (bhtp_pktlen == 20 || bhtp_pktlen == 32) {
                // looks like a TCP packet without a MGRP header
                // this is a know bug of MGRP
                char str[200];
                struct tcphdr *th = (struct tcphdr *)bh;
                mgrp_tcphdr_oneliner(str, sizeof(str), th, skb->len);
                DEBUG("%s", str);
        }
drop:
	ERR("rcv: BHTP packet DROPPED: %s", errstr);
	kfree_skb(skb);
	/* always zero */
	return 0;
}


void
bhtp_err(struct sk_buff *skb, __u32 info)
{
	LOG("bhtp_err");
}

int
bhtp_state_init(void)
{
	int rc = 0;

	/* do not zero state: it is initialized later at declaration */

#ifdef MGRP_TSKBSTATS
        bhtp.tskbstats = vmalloc(sizeof(struct mgrp_tskbstats));
        if (bhtp.tskbstats == NULL) {
                ERRFN("failed to allocated mgrp_tskbstats %d bytes",
                      sizeof(struct mgrp_tskbstats));
                return -ENOBUFS;
        }
        mgrp_tskbstats_init(bhtp.tskbstats);
        LOGFN("allocated %d bytes for mgrp_tskbstats",
              sizeof(struct mgrp_tskbstats));
#endif

	/* FIXME: This is ok and enough because tun->proto is u8 */
	bhtp.tun_prots = kzalloc(sizeof(struct bhtp_tun) * 256, GFP_KERNEL);
	if (bhtp.tun_prots == NULL)
		return -ENOBUFS;

	spin_lock_init(&bhtp.lock);
	spin_lock_init(&bhtp.req_lock);

	INIT_LIST_HEAD(&bhtp.req_list);
	init_waitqueue_head(&bhtp.req_wq);

	/* sysfs */
/* 	rc = bhtp_sysfs_register(); */
	if (rc)
		goto free_tun_prots;


        hrtimer_init(&bhtp.req_hrtimer, CLOCK_REALTIME, HRTIMER_MODE_ABS);
        bhtp.req_hrtimer.function = mgrp_req_hrtimer;

	return 0;

free_tun_prots:
	kfree(bhtp.tun_prots);
	return rc;
}

void
bhtp_state_done(void)
{
	mgrp_destroy_all_flows();
#ifdef MGRP_TSKBSTATS
        if (bhtp.tskbstats != NULL)
                vfree(bhtp.tskbstats);
        bhtp.tskbstats = NULL;
#endif
	kfree(bhtp.tun_prots);
	bhtp.tun_prots = NULL;

/* 	bhtp_sysfs_unregister(); */

        mgrp_drop_req();
        hrtimer_cancel(&bhtp.req_hrtimer);
}

struct bhtp_state bhtp = {

	/* for IP layer registration */
	.protocol = {
		.handler	= bhtp_rcv,
		.err_handler	= bhtp_err,
		.no_policy	= 1,
	},
	.flows = LIST_HEAD_INIT(bhtp.flows),
	.latency = BHTP_LATENCY_NS,
        .maxpbk_percent = MGRP_MAXPBK_PERCENT,
	.maxdelay = MGRP_MAXDELAY_NS,
        .buckets = MGRP_FRAG_HASH_BUCKETS_DEFAULT,
        .policy = MGRP_PBK_DEFAULT_POLICY,
};

/* BHTP ABI */
EXPORT_SYMBOL(bhtp_register_tun_prot);
EXPORT_SYMBOL(bhtp_unregister_tun_prot);
EXPORT_SYMBOL(bhtp_find_attach_flow);
EXPORT_SYMBOL(bhtp_detach_flow);
EXPORT_SYMBOL(bhtp_transmit_skb);

