/*
 * Timer requests to send skbs that enter the MGRP buffer if not piggybacked.
 *
 * Wed Aug 27 07:52:53 MDT 2008
 */
#ifndef __MGRP_REQ__
#define __MGRP_REQ__

#include <linux/hrtimer.h>
#include "main.h"

struct bhtp_flow_req {
        struct sk_buff *skb;
        struct bhtp_flow *bflow;
        struct list_head list;
        atomic_t sent;
        atomic_t fully_piggybacked;
        long long idg;
        long long idf;
        ktime_t latency;
        ktime_t send_kt;
};

void mgrp_drop_req(void);
enum hrtimer_restart mgrp_req_hrtimer(struct hrtimer *timer);

// returns -1 if no request was found
// returns 0 if there is at least one request and fills @expires
// with the expriration time of the head request
static inline int
mgrp_get_expiration_of_head_req(ktime_t *expires)
{
        int rc = -1;
        struct bhtp_flow_req *req = NULL;

        spin_lock_bh(&bhtp.req_lock);
        if (!list_empty(&bhtp.req_list)) {
                req = list_entry(bhtp.req_list.next, struct bhtp_flow_req, list);
                *expires = req->send_kt;
                rc = 0;
        }
        spin_unlock_bh(&bhtp.req_lock);

        return rc;
}

// returns 0 if request has send time before the current time
static inline ktime_t
mgrp_req_remaining(struct bhtp_flow_req *req)
{
        ktime_t now_kt = tick_now();
        ktime_t remaining = ktime_set(0, 0);
        if (ktime_to_ns(req->send_kt) > ktime_to_ns(now_kt))
                remaining = ktime_sub(req->send_kt, now_kt);
        return remaining;
}

static inline int
mgrp_req_expired(struct bhtp_flow_req *req)
{
        // any timer within a microsec is considered expired
        return (ktime_to_ns(mgrp_req_remaining(req)) <= NSEC_PER_USEC);
}

// returns NULL if no requests exist or no requests expired
static inline struct bhtp_flow_req *
mgrp_dequeue_expired_req(void)
{
        struct bhtp_flow_req *req = NULL;

        spin_lock_bh(&bhtp.req_lock);
        if (!list_empty(&bhtp.req_list)) {
                req = list_entry(bhtp.req_list.next, struct bhtp_flow_req, list);
                if (mgrp_req_expired(req)) {
                        list_del(&req->list);
                        atomic_dec(&bhtp.req_waiting);
                } else {
                        req = NULL;
                }
        }
        spin_unlock_bh(&bhtp.req_lock);

        return req;
}

// returns NULL if no requests exist
static inline struct bhtp_flow_req *
mgrp_dequeue_req_and_peek_next(int *has_next, ktime_t *next_expires)
{
        struct bhtp_flow_req *req = NULL;
        struct bhtp_flow_req *next_req = NULL;

        *has_next = 0;
        spin_lock_bh(&bhtp.req_lock);

        if (list_empty(&bhtp.req_list))
                goto done_unlock;

        req = list_entry(bhtp.req_list.next, struct bhtp_flow_req, list);
        list_del(&req->list);
        atomic_dec(&bhtp.req_waiting);

        next_req = list_entry(bhtp.req_list.next, struct bhtp_flow_req, list);
        if (next_req != NULL) {
                *has_next = 1;
                *next_expires = next_req->send_kt;
        }
done_unlock:
        spin_unlock_bh(&bhtp.req_lock);

        return req;
}


// returns NULL if no requests exist or no requests expired
// does NOT dequeue the request
static inline struct bhtp_flow_req *
mgrp_peek_expired_req(void)
{
        struct bhtp_flow_req *req = NULL;

        spin_lock_bh(&bhtp.req_lock);
        if (!list_empty(&bhtp.req_list)) {
                req = list_entry(bhtp.req_list.next, struct bhtp_flow_req, list);
                if (!mgrp_req_expired(req))
                        req = NULL;
        }
        spin_unlock_bh(&bhtp.req_lock);

        return req;
}

// creates and enqueues a new request to send the skb @latency into the future
// also starts the request timer if needed
static inline int
mgrp_enqueue_req(struct bhtp_flow *bflow, struct sk_buff *skb, ktime_t latency)
{
        int rc = 0;
        int no_req_pending = 0;
        struct bhtp_flow_req *req = NULL;
        struct bhtp_qskb *qskb = bhtp_qsk(skb);
        
        req = kzalloc(sizeof(struct bhtp_flow_req), GFP_ATOMIC);
        if (!req) {
                ERR("enqueue_req: out of mem");
                rc = -ENOMEM;
                goto done;
        }
        /* backpointer to request inside the skb */
        qskb->send.req = req;

        req->skb = skb;
        req->bflow = bflow;
        req->idg = atomic_inc_return(&bhtp.reqid);
        req->idf = atomic_inc_return(&bflow->reqid);
        req->latency = latency;
        req->send_kt = ktime_add(qskb->kt, latency);

        spin_lock_bh(&bhtp.req_lock);
        no_req_pending = list_empty(&bhtp.req_list);
        list_add_tail(&req->list, &bhtp.req_list);
        atomic_inc(&bhtp.req_waiting);
        spin_unlock_bh(&bhtp.req_lock);

        // If no request is pending then it is up to us to reschedule the
        // timer. We need to first wait for the existing handler to finish.
        if (no_req_pending) {
                MDEBUG("hrtimer_cancel start");
                hrtimer_cancel(&bhtp.req_hrtimer);
                MDEBUG("hrtimer_cancel end");

                hrtimer_start(&bhtp.req_hrtimer, req->send_kt, HRTIMER_MODE_ABS);
        }

        // If there were requests pending, then we are sure that
        // the timer hanlder will reschedule the timer.
        // See also comments in mgrp_req_hrtimer().
done:
        return rc;
}
#endif
