/* Must come before log.h. Otherwise we get compiler warning  */
#define LOG_PREFIX "tcp-bhtp: "
/* #define LOG_DEBUG */
#include "log.h"

#include "tcp.h"
#include "frag.h"

int tcp_bhtp_setup(struct sock *sk);

int
tcp_bhtp_rcv(struct sk_buff *skb, struct sockaddr_in *remote_addr)
{
	int rc = 0;
        int fragid = 0;

        // uncomment if you want TCP to ignore checksum (not a good idea)
        // skb->ip_summed = CHECKSUM_UNNECESSARY;

	DEBUG("tcp_bhtp_rcv skb_len %d", skb->len);
        if (mgrp_tskb_fragid_get(skb, &fragid) != -1)
                DEBUG_FRAGHEX(skb, fragid, 0, NULL,
                              FRAGHDR_NOT_INCLUDED,
                              ">>> tcprcv");

	rc = tcp_v4_rcv(skb);

	DEBUG("tcp_bhtp_rcv returned from tcp_v4_rcv %d", rc);
	return rc;
}

inline struct tcp_bhtp_sock *
tcp_bhtp_sk(struct sock *sk)
{
	struct tcp_sock *tp = tcp_sk(sk);
	struct tcp_bhtp_sock *tbs = (struct tcp_bhtp_sock *)tp->mgrp.data;
	return tbs;
}

/* needed for any function that may be called directly from tcp kernel code */
static inline int
tcp_bhtp_setup_if_needed(struct sock *sk)
{
	int rc = 0;

	if (likely(tcp_bhtp_sk(sk)->setup))
		return 0;
		
	rc = tcp_bhtp_setup(sk);
	if (rc)
		ERR("xmit: tcp_bhtp_setup failed %d", rc);
	return rc;
}


#define BHTP_NOT_NULL(p) \
	do \
		if ((p) == NULL) { \
			ERR_TR("pointer %s is NULL", #p); \
			return -1; \
		} \
	while (0)

#define BHTP_NOT_NULL_DROP(p) \
	do \
		if ((p) == NULL) { \
			ERR_TR("pointer %s is NULL", #p); \
			rc = -1; \
			goto drop; \
		} \
	while (0)

static inline int
tcp_header_len(struct sk_buff *skb)
{
	struct tcphdr *th = (struct tcphdr *)skb->transport_header;
	/* header length in 32bit words */
	return (4 * th->doff);
}

/*
 * Fills the length fields of qskb contained in skb.
 *
 * Assumes that the skb contains a TCP packet with a valid header.
 * Assumes no padding and the data has not been fragmented.
 */

static inline void
tcp_fill_qskb_lengths(struct sk_buff *skb)
{
	struct bhtp_qskb *qskb = bhtp_qsk(skb);
	
        // Sanity check to make sure that the length
        // variables are big enough. We use shorts
        // in bhtp_qskb because the whole structure
        // should fit in sbk->cb[] which is 48 bytes.
        //
        // We only need to check for tun_len because if
        // any of the other length variables does not fit,
        // then also tun_len should not fit
        if (skb->len >= MAX_INT_VALUE(qskb->tun_len)) {
                ERRFN("skblen %d is too large, max is %d",
                      skb->len, MAX_INT_VALUE(qskb->tun_len));
        }

	qskb->tun_hdr = tcp_header_len(skb);
	qskb->tun_data_orig = skb->len - qskb->tun_hdr;
	qskb->tun_data = qskb->tun_data_orig;
	qskb->tun_data_frag = 0;        // standalone fragmentation
	qskb->tun_chunk_frag = 0;       // fragmentation without valid headers
	qskb->tun_len = qskb->tun_hdr + qskb->tun_data;
	qskb->tun_pad = 0;

        // the original skb cannot be less than what we think the header is
	if (skb->len < qskb->tun_hdr) {
		ERR("fill_lengths: skb_len %d tun_hdr %d tun_len %d",
			skb->len, qskb->tun_hdr, qskb->tun_len);
	}
}

/* attach to bhtp flow based on tcp socket's ip destination address */
int
tcp_bhtp_attach_flow_if_needed(struct sock *sk)
{
	struct tcp_bhtp_sock *tbs = tcp_bhtp_sk(sk);
	struct sockaddr_in usin = {
		.sin_family = AF_INET,
	};

	if (likely(tbs->bflow != NULL))
		return 0;

	usin.sin_addr.s_addr = inet_sk(sk)->daddr;
	tbs->bflow = bhtp_find_attach_flow(&usin, 0);

	DEBUGFN("attach_flow: attached to flow %p", tbs->bflow);
	if (tbs->bflow == NULL) {
		ERR("attach_flow: no valid flow found");
		return -1;
	}

	return 0;
}


int
tcp_bhtp_queue_xmit(struct sk_buff *skb, int ipfragok)
{
	int rc = 0;
        struct mgrp_probe_feedback feedback = { 0 };
	struct sock *sk = NULL;
	struct tcp_bhtp_sock *tbs = NULL;
	struct bhtp_qskb *qskb = NULL;

	BHTP_NOT_NULL_DROP(skb);

	sk = skb->sk;
	BHTP_NOT_NULL_DROP(sk);

	tbs = tcp_bhtp_sk(sk);
	BHTP_NOT_NULL_DROP(tbs);

        DEBUG("************************************");
	DEBUG("xmit: skb-len %d fragok %d", skb->len, ipfragok);
        if (debug_on()) {
                char str[200];
                struct tcphdr *th = (struct tcphdr *)skb->transport_header;
                mgrp_tcphdr_oneliner(str, sizeof(str), th, skb->len);
                DEBUG("%s", str);
                log_hex((const char *)th, 20, 4, "tcp: ");
        }

	BUG_ON(ipfragok);

	LOG_TCPHDR_SKB(skb);

	/* tbs (struct tcp_bhtp_sock) must be valid */
	rc = tcp_bhtp_setup_if_needed(sk);
	if (rc)
		goto drop;
	
	/* attach tbs->bflow to a valid bhtp flow that goes to ip dst */
	rc = tcp_bhtp_attach_flow_if_needed(sk);
	if (rc)
		goto drop;

	/* store bhtp metadata in the control buffer skb->cb[] */
	qskb = bhtp_qsk_init(skb);

	qskb->tun = &tbs->tun;
	tcp_fill_qskb_lengths(skb);
	if (tbs->fragment) {
		qskb->tun_frag_len_max = tbs->frag_len_max;
		DEBUG("xmit: fragmenting on purpose frag_max %d",
			qskb->tun_frag_len_max);
	}

	/* hand to BHTP to schedule based on qskb metadata */
	rc = bhtp_schedule_skb(tbs->bflow, skb, &feedback);
	if (rc  < 0) {
		ERR("xmit: bhtp_transmit_skb failed %d", rc);
		goto drop;
	}

	return 0;

drop:
	kfree_skb(skb);
	ERR("xmit: packet DROPPED");
	return rc;
}

EXPORT_SYMBOL(tcp_bhtp_queue_xmit);

static inline __u32
tcp_seq_advance(__u32 seq, unsigned int offset)
{
	seq += offset;
	return seq;
}


/*
 *	Fixes the TCP header of a fragmented packet.
 *
 *	The header to fix is @th of length @hlen. It contains a copy of the
 *	original unfragmented packet. For TCP, we need to fix the sequence
 *	number, the flags and calculate the checksum. The TCP options are
 *	copied verbatim.
 *
 *		A pointer to the @data is provided for the checksum.
 *		The @data_offset is provided for the sequence number adustment.
 *
 *	This function is skb agnostic, but normally the @th and @data pointers
 *	point to the head buffer of an skb and are assumed to be contiguous.
 */

static inline int
tcp_fix_fragment_header(struct tcphdr *th, int hlen,
		       void *data, int data_len,
		       unsigned int data_offset, int is_last)
{
	/* see tcp_fragment() tso_fragment() */

	unsigned int new_seq = 0;
	int rc = 0;
	int not_supported =
		th->syn ||
		th->rst ||
		th->urg;


	if (not_supported) {
		ERR("tcp_fix_segment_header: do not know how to segment this tcp packet");
		goto error;
	}

	/* TCP options: no change because copied verbatim in all segments */

	/* adjust sequence number */
	new_seq = tcp_seq_advance(ntohl(th->seq), data_offset);

	DEBUG("fix: seqn %d seqh %d offset %d new_seqn %d new_seqh %d",
		u2i(th->seq),
		u2i(ntohl(th->seq)),
		data_offset,
		htonl(new_seq),
		new_seq);

	th->seq = htonl(new_seq);

	DEBUG("fix:            seqn %d seqh %d",
		u2i(th->seq),
		u2i(ntohl(th->seq)));

	/* PSH and FIN should only be set in the last fragment. */
	if (!is_last) {
		th->fin = 0;
		th->psh = 0;
	}

	/* checksum */
	// FIXME: calculate checksum

	return 0;
error:
	return rc;
}

/*
 * Return a pointer to the @data_offset of the TCP skb.
 *
 * Assumes that the skb contains the TCP header and that both header
 * and data are contained in the skb head buffer.
 */
static inline unsigned char *
tcp_data_ptr(struct sk_buff *skb, int data_offset)
{
	int hdrlen = tcp_header_len(skb);
	int offset = hdrlen + data_offset;

	BUG_ON(skb->data != skb->transport_header);

	if (data_offset < 0)
		goto invalid;
	
	if (offset >= (skb->len - skb->data_len)) {
		ERR("tcp_data_ptr: not supporting fragments");
		goto invalid;
	}

	return (skb->data + offset);

invalid:
	return NULL;
}

// Returns the (unaligned) pointer to the timestamp value of the TCP header
// Returns NULL if not found
//
// code extracted from ipv4/tcp_input.c:tcp_parse_options()
// skb must contain a valid tcp packet with a valid tcp header
static inline unsigned char *
tcp_mgrp_timestamp_ptr(struct sk_buff *skb)
{
	struct tcphdr *th = tcp_hdr(skb);
	unsigned char *ptr = (unsigned char *)(th + 1);
	int length = (th->doff * 4) - sizeof(struct tcphdr);

        while (length > 0) {
                int opcode = *ptr++;
                int opsize;

                switch (opcode) {

		case TCPOPT_EOL:
			return NULL;

		case TCPOPT_NOP:	/* Ref: RFC 793 section 3.1 */
			length--;
			continue;

		default:
			opsize = *ptr++;

			if (opsize < 2) /* "silly options" */
				return NULL;

			if (opsize > length)
				return NULL; /* don't parse partial options */
                        
                        if (opcode == TCPOPT_TIMESTAMP)
                                if (opsize == TCPOLEN_TIMESTAMP)
                                        // returns pointer to timestamp value
                                        return ptr;

			ptr += opsize-2;
			length -= opsize;
                }
        }

        return NULL;
}

// Zeroes the timestamp value of the timestamp option in the TCP headaer
static inline void
tcp_mgrp_zero_timestamp(struct sk_buff *skb)
{
	struct tcphdr *th = tcp_hdr(skb);
	int thlen = (th->doff * 4);
        unsigned char *tstamp_ptr = tcp_mgrp_timestamp_ptr(skb);

        if (tstamp_ptr != NULL) {
                if (debug_on()) {
                        LOGFN("================================================");
                        log_hex(th, thlen, 4, "tcphdr before: ");
                }

                memset(tstamp_ptr, 0, 4);

                if (debug_on()) {
                        log_hex(th, thlen, 4, "tcphdr after : ");
                }
        }
}

/*
 * Returns 1 if there are any fragments left in the skb.
 *
 * Also returns:
 *
 *      @fraglen_min    min length of next fragment
 *      @fralen_max     max length of next fragment
 *
 *      @has_more_frags 1 or 0, to indicate if it has more
 *                      fragments beyond the current fragment
 *
 * If no fragments are left, the input/output arguments
 * remain unchanged. Any of them can be NULL if the return
 * values are not needed.
 *
 * This is one of the TUN fragment operations (bhtp_tun_ops->fraglen_left).
 *
 * Assumes that qskb lengths are valid. Used during fragmentation.
 * This length is the length of the TCP packet that would result if
 * we combine a TCP header with the remaining data in the skb (excluding
 * data that has been included in previous fragments).
 */
int
tcp_tun_fraglen_left(struct sk_buff *skb,
                     int *fraglen_min, int *fraglen_max,
                     int *has_more_frags)
{
        int min = 0;
        int max = 0;
        int more_frags = 0;

	struct bhtp_qskb *qskb = bhtp_qsk(skb);
	int frag_data_left = qskb->tun_data - qskb->tun_data_frag;
	int header_only = (qskb->tun_data == 0);

        // This flag is set by tcp_tun_fragment()
        int frags_exhausted = qskb->type & BHTP_QSKB_FRAG_EXHAUSTED;

        BUG_ON(frag_data_left < 0);
        BUG_ON(frags_exhausted && frag_data_left > 0);
        // we need space for at least one data byte (in the degenerate case)
        BUG_ON(qskb->tun_frag_len_max < (qskb->tun_hdr + 1));

        if (frags_exhausted)
                return 0;

        if (header_only) {
                BUG_ON(frag_data_left > 0);
                min = qskb->tun_hdr;
                max = qskb->tun_hdr;

        } else {
                int desired_fraglen_max = qskb->tun_hdr + frag_data_left;

                BUG_ON(frag_data_left == 0);
                // stream protocol, possible to send only one data byte across
                min = qskb->tun_hdr + 1;
                if (desired_fraglen_max > qskb->tun_frag_len_max) {
                        max = qskb->tun_frag_len_max;
                        more_frags = 1;
                } else {
                        max = desired_fraglen_max;
                }
        }

        BUG_ON(min > max);
                
	DEBUG("fraglen_left: hdr %d data_orig %d data_left %d min %d max %d",
		qskb->tun_hdr, qskb->tun_data, frag_data_left, min, max);
		
        if (fraglen_min != NULL)
                *fraglen_min = min;
        if (fraglen_max != NULL)
                *fraglen_max = max;
        if (has_more_frags != NULL)
                *has_more_frags = more_frags;

	return 1;
}

/*
 * Extract a TCP fragment skb out of a TCP skb packet.
 *
 * This is one of the TUN fragment operations (bhtp_tun_ops->fragment).
 *
 * Allocates a new skb to hold a TCP packet that contains a fragment of
 * the data contained in @skb. The total length of the new fragment is
 * @frag_len and includes both header and data.
 *
 * TCP keeps track of the data already fragmented (bitten out) from the
 * original skb in @qskb->tun_data_frag. This function can be used repeateadly
 * to create fragments out of a TCP skb. We assume that the original skb
 * includes a valid skb header.
 *
 * See also the kernel functions: tcp_fragment() tso_fragment() skb_split().
 * A key difference from tcp_fragment: skbs here are not headerless.
 *
 * Assumes that the skb has data beyond the header.
 *
 * This version always copies the skb data, so @copy_data is ignored.
 */
struct sk_buff * 
tcp_tun_fragment(struct sk_buff *skb, int frag_len, int copy_data)
{
	int rc = 0;

	struct bhtp_qskb *qskb = bhtp_qsk(skb);

	int frag_hdr = qskb->tun_hdr;
	int frag_data = frag_len - frag_hdr;
	int frag_alloc = MAX_TCP_HEADER + frag_data;

	int frag_data_left = qskb->tun_data - qskb->tun_data_frag;
	int frag_data_offset = qskb->tun_data_frag;
	int frag_is_last = (frag_data_left == frag_data);
	int skb_offset = frag_hdr + frag_data_offset;

	struct sk_buff *frag = NULL;
        struct tcphdr *frag_th = NULL;
	unsigned char *frag_data_ptr = NULL;

	DEBUG("tcp_tun_frag:");
	DEBUG("      hdr %d data %d alloc %d data_left %d data_offset %d is_last %d",
		frag_hdr,
		frag_data,
		frag_alloc,
		frag_data_left,
		frag_data_offset,
		frag_is_last);
	
	if (qskb->tun_data == 0) {
		ERR("tun_frag: cannot fragment a packet with no data");
		goto error;
	}

	if (frag_data == 0) {
		ERR("tun_frag: no room for data in fragment");
		goto error;
	}

	if (frag_data < 0) {
		ERR("tun_frag: header %d is larger than frag %d",
			frag_hdr, frag_len);
		goto error;
	}

	if (frag_data > frag_data_left) {
		ERR("tun_frag: asking to frag %d; only %d left",
			frag_data, frag_data_left);
		goto error;
	}

	/*
	 * frag
	 *
	 * New skb that holds the new segment.
	 * Allocate enough space to hold all L4/L3/L2 headers + data frag
	 */

	frag = alloc_skb(frag_alloc, GFP_ATOMIC);
	if (frag == NULL) {
		ERR("tun_frag: no skb buffers for %d bytes", frag_alloc);
		goto error;
	}

	/* copy the original L4 header now; fix it later */
	skb_reserve(frag, MAX_TCP_HEADER);
	frag->transport_header = skb_push(frag, frag_hdr);
	memcpy(frag->transport_header, skb->transport_header, frag_hdr);
	frag_th = (struct tcphdr *)frag->transport_header;

	LOG_TCPHDR_SKB(frag);

	/* copy the data portion of the fragment */
	frag_data_ptr = skb_put(frag, frag_data);
	rc = skb_copy_bits(skb, skb_offset, frag_data_ptr, frag_data);
	if (rc) {
		ERR("tun_frag: error copying bits from original skb %d rc", rc);
		goto free_frag;
	}

	LOG_TCPHDR_SKB(frag);

	/* fix the bhtp private data of the new skb */
	/* must do after header and data are in frag because it uses frag->len */
	bhtp_qsk_clone(frag, qskb);
	tcp_fill_qskb_lengths(frag);

	/* fix the header for the new segment */
	rc = tcp_fix_fragment_header(frag_th, frag_hdr,
				     frag_data_ptr, frag_data,
				     frag_data_offset, frag_is_last);
	if (rc) {
		ERR("tun_frag: error fixing segment header");
		goto free_frag;
	}

	/* update the portion of the original skb that is fragmented */
	qskb->tun_data_frag += frag_data;

        // sanity check
        if (qskb->tun_data_frag > qskb->tun_data) {
                ERR("tun_frag: bug? tun_data_frag %d exceeded tun_data %d",
                    qskb->tun_data_frag, qskb->tun_data);
                goto free_frag;
        }

        // flag to tell tcp_tun_fraglen_left() that we are done
        if (qskb->tun_data_frag == qskb->tun_data)
                qskb->type |= BHTP_QSKB_FRAG_EXHAUSTED;

	return frag;

free_frag:
	kfree_skb(frag);
error:
	return NULL;
}

void
tcp_tun_mark(struct sk_buff *skb)
{
        // XXX this scheme does not work for TCP!
        // XXX the sender discards the segments as old
        // XXX see rfc1323, 4.2.1 R1)

        BUG();

        // return (bflow->pbk_policy == MGRP_PBK_REDUNDANT);
        // overwrite the TCP timestamp (if it exists) to mark
        // this TCP segment and its associated ack
        tcp_mgrp_zero_timestamp(skb);
}

int
do_tcp_bhtp_getsockopt(struct sock *sk, int level, int optname,
	 	    char __user *optval, int __user *optlen)
{
	/* modeled after do_tcp_getsockopt() */

	int val = 0;
	int len = 0;
	int rc = 0;

	LOG("bhtp_getsockopt: ");

	if (get_user(len, optlen))
		return -EFAULT;

	len = min_t(unsigned int, len, sizeof(int));

	if (len < 0)
		return -EINVAL;

	rc = tcp_bhtp_setup_if_needed(sk);
	if (rc)
		return rc;

	switch (optname) {

	default:
		return -ENOPROTOOPT;
	}

	if (put_user(len, optlen))
		return -EFAULT;
	if (copy_to_user(optval, &val, len))
		return -EFAULT;

	return 0;
}


static inline int
do_tcp_bhtp_setsockopt(struct sock *sk, int level, int optname,
		    char __user *optval, int optlen)
{
	/* modeled after do_tcp_setsockopt() */

	struct tcp_bhtp_sock *tbs = tcp_bhtp_sk(sk);
	int val = 0;
	int rc = 0;

	if (optlen < sizeof(int))
		return -EINVAL;

	if (get_user(val, (int __user *)optval))
		return -EFAULT;

	rc = tcp_bhtp_setup_if_needed(sk);
	if (rc)
		goto done;

	lock_sock(sk);

	switch (optname) {

	case TCP_MGRP_FRAGMENT:
		DEBUGFN("setsockopt TCP_MGRP_FRAGMENT %d", val);
		tbs->fragment = (val ? 1 : 0);
		tcp_bhtp_frag_state_debug(tbs);
		break;
		
	case TCP_MGRP_FRAG_LEN_MAX:
		DEBUGFN("setsockopt TCP_MGRP_FRAG_LEN_MAX %d", val);
		tbs->frag_len_max = val;
		tcp_bhtp_frag_state_debug(tbs);
		break;
		
	default:
		rc = -ENOPROTOOPT;
		break;
	};

	release_sock(sk);
done:
	return rc;
}


int
tcp_bhtp_getsockopt(struct sock *sk, int level, int optname,
	 	    char __user *optval, int __user *optlen)
{
	if (level != SOL_MGRP)
		return tcp_orig_icsk_af_ops(sk)->getsockopt(sk, level, optname,
				optval, optlen);

	return do_tcp_bhtp_getsockopt(sk, level, optname, optval, optlen);
}


int
tcp_bhtp_setsockopt(struct sock *sk, int level, int optname,
		    char __user *optval, int optlen)
{
	if (level != SOL_MGRP)
		return tcp_orig_icsk_af_ops(sk)->setsockopt(sk, level, optname,
				optval, optlen);

	return do_tcp_bhtp_setsockopt(sk, level, optname, optval, optlen);
}


struct bhtp_tun_ops tcp_tun_ops = {
	.rcv		= tcp_bhtp_rcv,
	.fragment	= tcp_tun_fragment,
	.fraglen_left	= tcp_tun_fraglen_left,
	.mark	        = tcp_tun_mark,
};

struct bhtp_tun_prot tcp_tun_prot = {
	.proto		= BHTP_PROTO_TCP,
	.ops		= &tcp_tun_ops,
};


int
tcp_bhtp_setup(struct sock *sk)
{
	int rc = 0;
	struct tcp_sock *tp = NULL;
	struct tcp_bhtp_sock *tbs = NULL;

	BHTP_NOT_NULL(sk);

	tp = tcp_sk(sk);
	BHTP_NOT_NULL(tp);

	tbs = tcp_bhtp_sk(sk);
	BHTP_NOT_NULL(tbs);

	DEBUGFN("tcp_bhtp_setup %p", sk);

	if (sizeof(tp->mgrp.data) < sizeof(struct tcp_bhtp_sock)) {
		ERR("tp->mgrp.data not large enough: max %d needed %d",
			sizeof(tp->mgrp.data),
			sizeof(struct tcp_bhtp_sock));
		rc = -ENOMEM;
		goto done;
	}

	memset(tbs, 0, sizeof(*tbs));

	tbs->setup = 1;
	tbs->tun.prot = &tcp_tun_prot;
	tbs->frag_len_max = 400;
done:
	return rc;
}

EXPORT_SYMBOL(tcp_bhtp_setup);


int __init
tcp_bhtp_init(void)
{
	int rc = 0;

	LOG("attaching kernel TCP to BHTP");
	
	/* register with BHTP */
	rc = bhtp_register_tun_prot(&tcp_tun_prot);
	if (rc)
		goto error;
	LOG("registered with BHTP");

	/* registering with TCP (for TCP tunnel over BHTP) -- FIXME hack */
	ipv4_specific_mgrp.queue_xmit = tcp_bhtp_queue_xmit;
	ipv4_specific_mgrp.setsockopt = tcp_bhtp_setsockopt;
	ipv4_specific_mgrp.getsockopt = tcp_bhtp_getsockopt;

	LOG("registered with TCP");

	LOG("loaded");
	return 0;

error:
	ERR("loading aborted");
	return rc;
}


void
tcp_bhtp_exit(void)
{
	LOG("unloading");

	/* unregister with TCP (for TCP tunnel over BHTP) -- FIXME hack */
	ipv4_specific_mgrp.queue_xmit = NULL;
	LOG("unregistered with TCP");

	bhtp_unregister_tun_prot(&tcp_tun_prot);
	LOG("unregistered with BHTP");

	LOG("unloaded");
}

