/*
 * BHTP		Bottom Half Transport Protocol
 *
 * Created:	Tue Sep 12 07:30:43 MDT 2006
 *
 * Author:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 */

#define TOTAL_LOOPS	1000000

#include "main.h"

struct tick_state tick = {
        // Calibrated by tick_gapsleep_calibrate().
        // Default values will not yield best results.
        .gap_busywait_threshold_ns  = 1 * NSEC_PER_MSEC,
        .gap_busywait_adjust_percent = 100,

        // dotimer experimentation
	.started = 0,
	.nsec = 10000,
	.times = 10,
};

unsigned long tick_cyc2ns_scale __read_mostly = 0;

struct tick_stats {
	unsigned long duration;
	unsigned long long avg;
	unsigned long long min;
	unsigned long long max;
	int loops;
	int min_index;
	int max_index;
	int overpow;
	int over[10];
};

static inline void
tick_stats_print(struct tick_stats *stats)
{
	int j = 0;
	int thres = 0;

	if (stats->loops == 0) {
		ERR("tick_stats: zero loops, nothing to report");
		return;
	}

	LOG("tick_stats iter %d total %lu avg %llu nanoseconds",
		stats->loops, stats->duration, stats->avg);

	if (stats->max_index == -1 && stats->min_index == -1)
		goto done;

	LOG("tick_stats         max %8llu index %d",
		stats->max, stats->max_index);
	LOG("tick_stats         min %8llu index %d",
		stats->min, stats->min_index);
	for (j=0, thres=1; j <= stats->overpow; j++, thres*=10)
		LOG("tick_stats         over%8d %10d %d%%",
			thres, stats->over[j], 100*stats->over[j]/stats->loops);
done:
	LOG("tick_stats");
}

#define tick_measure_gettime(__ops, __loops, __overpow, __statsp) \
	do { \
		int __i, __j, __thres; \
		unsigned long long __started, __ended, __tmp, __d; \
		struct tick_stats __stats = { \
			.min_index = -1, \
			.min = (unsigned long long)-1, \
			.max_index = -1, \
			.max = 0, \
			.loops = __loops, \
			.overpow = __overpow, \
		}; \
		int __maxpow = (sizeof(__stats.over) / sizeof(__stats.over[0])) - 1; \
		if (__stats.overpow > __maxpow) \
			__stats.overpow = __maxpow; \
		LOG("measuring execution time for %s", #__ops); \
		__started = (__ops); \
		__ended = __tmp = __started; \
		for (__i = 0; __i < __loops; __i++) { \
			__ended = (__ops); \
			__d = __ended - __tmp; \
			__tmp = __ended; \
			if (__d > __stats.max) {__stats.max = __d; __stats.max_index = __i;} \
			if (__d < __stats.min) {__stats.min = __d; __stats.min_index = __i;} \
			for (__j=0, __thres=1; __j <= __stats.overpow; __j++, __thres*=10) \
				if (__d > __thres) \
					__stats.over[__j]++; \
		} \
		__d = __ended - __started; \
		__stats.duration = __d; \
		__stats.avg = __stats.duration / __stats.loops; \
		if (__statsp) \
			*(__statsp) = __stats; \
		if (__d != __stats.duration) \
			ERR("** ERROR in measurement: duration overflow"); \
	} while(0)

#define tick_measure_ops(__ops, __loops, __statsp) \
	do { \
		int __i; \
		unsigned long long __started, __ended, __tmp, __d; \
		struct tick_stats __stats = { \
			.min_index = -1, \
			.min = (unsigned long long)-1, \
			.max_index = -1, \
			.max = 0, \
			.loops = __loops, \
		}; \
		int __maxpow = (sizeof(__stats.over) / sizeof(__stats.over[0])) - 1; \
		if (__stats.overpow > __maxpow) \
			__stats.overpow = __maxpow; \
		LOG("measuring execution time for %s", #__ops); \
		__started =  ktime_to_ns(tick_now_gettimeofday()); \
		__ended = __tmp = __started; \
		for (__i = 0; __i < __loops; __i++) { \
			((void) __ops); \
		} \
		__ended =  ktime_to_ns(tick_now_gettimeofday()); \
		__d = __ended - __started; \
		__stats.duration = __d; \
		__stats.avg = __stats.duration / __stats.loops; \
		if (__statsp) \
			*(__statsp) = __stats; \
		if (__d != __stats.duration) \
			ERR("** ERROR in measurement: duration overflow"); \
	} while(0)

static inline void
tick_measure(void)
{
#define MEM_BYTES	1000
	struct tick_stats stats;
	struct timespec dummyts;
	char *mem = kzalloc(MEM_BYTES, GFP_KERNEL);

	tick_measure_ops(
		ktime_get_ts(&dummyts),
		TOTAL_LOOPS, &stats);
	tick_stats_print(&stats);

	tick_measure_ops(
		ktime_get_real_ts(&dummyts),
		TOTAL_LOOPS, &stats);
	tick_stats_print(&stats);

	tick_measure_ops(
		getnstimeofday(&dummyts),
		TOTAL_LOOPS, &stats);
	tick_stats_print(&stats);

	tick_measure_ops(
		tick_now_gettimeofday(),
		TOTAL_LOOPS, &stats);
	tick_stats_print(&stats);

	tick_measure_ops(
		ktime_to_ns(tick_now_gettimeofday()),
		TOTAL_LOOPS, &stats);
	tick_stats_print(&stats);

	tick_measure_gettime(
		ktime_to_ns(tick_now_gettimeofday()),
		TOTAL_LOOPS, 3, &stats);
	tick_stats_print(&stats);

	tick_measure_ops(
		ktime_to_ns(tick_now_cycles()),
		TOTAL_LOOPS, &stats);
	tick_stats_print(&stats);

	tick_measure_gettime(
		ktime_to_ns(tick_now_cycles()),
		TOTAL_LOOPS, 3, &stats);
	tick_stats_print(&stats);

	tick_measure_ops(
		(tick_now_cycles(), tick_now_gettimeofday()),
		TOTAL_LOOPS, &stats);
	tick_stats_print(&stats);

	tick_measure_ops(memset(mem, 0, MEM_BYTES),
		TOTAL_LOOPS, &stats);
	tick_stats_print(&stats);

	kfree(mem);
}


static inline void
tick_measure_timestamp_latency_monotonic(void)
{
	int i = TOTAL_LOOPS;

	unsigned int long long ended = 0;
	unsigned long long d;
	unsigned long long max = 0;

	int min_index = 0;
	int max_index = 0;

	unsigned long long min = 999999999;
	unsigned long duration;
	unsigned int long long started = ktime_to_ns(tick_now_cycles());
	unsigned int long long tmp = started;
	int over100 = 0;
	int over1000 = 0;
	int over10000 = 0;

	LOG("measuring monotonic timestamp latency ...........");

	while(i-- > 0) {
		ended = ktime_to_ns(tick_now_cycles());
		d = ended-tmp;
		tmp = ended;
		if (d > max) {
			max = d;
			max_index = i;
		}
		if (d < min) {
			min = d;
			min_index = i;
		}
		if (d > 100)
			over100++;
		if (d > 1000)
			over1000++;
		if (d > 10000)
			over10000++;
	}

	duration = ended-started;
	LOG("measure iter %d total %lu avg %lu nanoseconds",
		TOTAL_LOOPS, duration, duration/TOTAL_LOOPS);
	LOG("measure         u64 %d __u64 %d",
		sizeof(u64), sizeof(__u64));
	LOG("measure         max %llu index %d",
		max, max_index);
	LOG("measure         min %llu index %d",
		min, min_index);
	LOG("measure         over100 %d %d%%",
		over100, 100*over100/TOTAL_LOOPS);
	LOG("measure         over1000 %d %d%%",
		over1000, 100*over1000/TOTAL_LOOPS);
	LOG("measure         over10000 %d %d%%",
		over10000, 100*over10000/TOTAL_LOOPS);
}

enum hrtimer_restart
tick_dotimer(struct hrtimer *timer)
{
	long long elapsed = tick_now_ns() - tick.started;

	LOG("********************** tick_dotimer: wanted %ld elapsed %lld",
		tick.nsec,
		elapsed);
	
	tick.times--;

/* 	Not exported! */
/* 	http://groups.google.com/group/linux.kernel/browse_thread/thread/27d58e73e22c9c2b/1045425f049f7f7d */

/* 	hrtimer_forward(timer, tick_now(), ktime_set(0, tick.nsec)); */
	return HRTIMER_NORESTART;
}

struct timer_list t2;
static int t2_times = 10;
static long long t2_nsec = 0;

void
tick_dotimer2(unsigned long arg)
{
	struct timer_list *t = (struct timer_list *)arg;
	long long elapsed = tick_now_ns() - t2_nsec;
	unsigned long j = 0;

	LOG("+++++++++++++++ tick_dotimer2: elapsed %lld", elapsed);

	j = jiffies;
	t2_nsec = tick_now_ns();
	t->expires = j + msecs_to_jiffies(1000);
	t->expires = j + 1;
	t2_times--;
	if (t2_times > 0)
		add_timer(t);
}


static const int do_timer_example1 = 0;
static const int do_timer_example2 = 0;

void
tick_init(void)
{
	char str[100];
	struct timespec tp;

        tick_gapsleep_calibrate();

	/* initialize the tick_now_cycles() */
	tick_cyc2ns_scale = tick_get_cyc2ns_scale();

/* 	tick_measure(); */
/* 	tick_measure_timestamp_latency_monotonic(); */

	LOG("tick: using timestamp from %s",
		tick_now_timestamp_str(str, sizeof(str)));

	/* accuracy of hrtimers */

	hrtimer_get_res(CLOCK_MONOTONIC, &tp);
	LOG("tick: hrtimer resolution for CLOCK_MONOTONIC sec %ld nsec %ld",
		tp.tv_sec, tp.tv_nsec);

	hrtimer_get_res(CLOCK_REALTIME, &tp);
	LOG("tick: hrtimer resolution for CLOCK_REALTIME sec %ld nsec %ld",
		tp.tv_sec, tp.tv_nsec);
	
	if (do_timer_example1) {
		int rc = 0;
		ktime_t kt = ktime_set(0, tick.nsec);
		hrtimer_init(&tick.timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL);
		tick.timer.function = tick_dotimer;

		tick.started = tick_now_ns();
		rc = hrtimer_start(&tick.timer, kt, HRTIMER_MODE_REL);
		LOG("tick: hrtimer start rc %d nsec %lld wanted %ld",
			rc, ktime_to_ns(kt), tick.nsec);
	}
	
	if (do_timer_example2) {
		unsigned long j = jiffies;

		init_timer(&t2);
		t2_nsec = tick_now_ns();
		t2.data = (unsigned long)&t2;
		t2.function = tick_dotimer2;
		t2.expires = j + msecs_to_jiffies(5000);
		add_timer(&t2);
	}
}

void
tick_done(void)
{
        if (do_timer_example2)
                del_timer_sync(&t2);
}

/********************************
 * Copied functions from hrtimer.c because hrtimer_nanosleep() is not exported
 */

// copied from hrtimer.c:hrtimer_wakeup
static enum hrtimer_restart tick_wakeup(struct hrtimer *timer)
{
	struct hrtimer_sleeper *t =
		container_of(timer, struct hrtimer_sleeper, timer);
	struct task_struct *task = t->task;

	t->task = NULL;
	if (task)
		wake_up_process(task);

	return HRTIMER_NORESTART;
}

// copied from hrtimer.c:hrtimer_init_sleeper
void tick_init_sleeper(struct hrtimer_sleeper *sl, struct task_struct *task)
{
	sl->timer.function = tick_wakeup;
	sl->task = task;
#ifdef CONFIG_HIGH_RES_TIMERS
	sl->timer.cb_mode = HRTIMER_CB_IRQSAFE_NO_SOFTIRQ;
#endif
}

// copied from hrtimer.c:do_nanosleep
static int __sched tick_do_nanosleep(struct hrtimer_sleeper *t, enum hrtimer_mode mode)
{
	tick_init_sleeper(t, current);

	do {
		set_current_state(TASK_INTERRUPTIBLE);
		hrtimer_start(&t->timer, t->timer.expires, mode);
		if (!hrtimer_active(&t->timer))
			t->task = NULL;

		if (likely(t->task))
			schedule();

		hrtimer_cancel(&t->timer);
		mode = HRTIMER_MODE_ABS;

	} while (t->task && !signal_pending(current));

	__set_current_state(TASK_RUNNING);

	return t->task == NULL;
}

// partial re-implementation of hrtimer_nanosleep(), because it is not exported
// see http://fixunix.com/kernel/395758-better-msleep-drivers.html
long tick_nanosleep_abs(struct timespec *rqtp, const clockid_t clockid)
{
        // support only absolute timers, hrtimers has more code for _REL
        const enum hrtimer_mode mode = HRTIMER_MODE_ABS;
        struct hrtimer_sleeper t;

        hrtimer_init(&t.timer, clockid, mode);
        t.timer.expires = timespec_to_ktime(*rqtp);
        if (tick_do_nanosleep(&t, mode))
                return 0;

        return -ERESTARTNOHAND;
}
