/*
 * BHTP		Bottom Half Transport Protocol
 *
 * Created:	Tue Sep 12 07:30:43 MDT 2006
 *
 * Author:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 */

#ifndef _BHTP_TICK_H
#define _BHTP_TICK_H

#include <linux/time.h>
#include <linux/ktime.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/hrtimer.h>

#define KTIME_ZERO      ktime_set(0, 0)

extern struct tick_state tick;

struct tick_state {
        // used by tick_gapsleep for sub-microsec gap generation
        // calibratated during startup by tick_gapsleep_calibrate()
        int gap_busywait_threshold_ns;
        int gap_busywait_adjust_percent;

        // dotimer experimentation
	long long started;
	unsigned long nsec;
	int times;
	struct hrtimer timer;
};

/* #define TICK_MONOTONIC	1 */
/* Monotonic does not work, since monotonic_clock was removed from
 * the kernel and sched_clock() is not exported.
 *
 * FIXME: revisit later
 *
 *	.. or not! the gettimeofday is pretty fast now:
 *
 *	bhtp: measure iter 1000000 total 75882000 avg 75 nanoseconds
 *	bhtp: measure         u64 8 __u64 8
 *	bhtp: measure         max 5249000 index 999999
 *	bhtp: measure         min 0 index 999997
 *	bhtp: measure         over100 70558 7%
 *	bhtp: measure         over1000 18 0%
 *	bhtp: measure         over10000 1 0%
 *	bhtp: measuring monotonic timestamp latency ...........
 *	bhtp: measure iter 1000000 total 3223692240 avg 3223 nanoseconds
 *	bhtp: measure         u64 8 __u64 8
 *	bhtp: measure         max 7927394208633 index 999999
 *	bhtp: measure         min 42 index 999973
 *	bhtp: measure         over100 17 0%
 *	bhtp: measure         over1000 12 0%
 *	bhtp: measure         over10000 1 0%
 *	bhtp: using timestamp from gettimeofday
 *
 * On the same hardare (emulab:pc3000) it used to be ~500nsec for
 * gettimeofday() and ~50nsec of monotonic_clock().
 */

void tick_init(void);
void tick_done(void);

static inline ktime_t
tick_now_gettimeofday(void)
{
	struct timespec ts;
	getnstimeofday(&ts);
	return timespec_to_ktime(ts);
}

long tick_nanosleep_abs(struct timespec *rqtp, const clockid_t clockid);


// *** taken from arch/i386/kernel/tsc.c *** BEGIN ******************
//
// because kernel does not export sched_clock HACK HACK

extern unsigned long tick_cyc2ns_scale;

#define TICK_CYC2NS_SCALE_FACTOR 10 /* 2^10, carefully chosen */

static inline unsigned long tick_get_cyc2ns_scale(void)
{
	return (1000000 << TICK_CYC2NS_SCALE_FACTOR)/cpu_khz;
}

static inline unsigned long long tick_cycles_2_ns(unsigned long long cyc)
{
	return (cyc * tick_cyc2ns_scale) >> TICK_CYC2NS_SCALE_FACTOR;
}

// *** taken from arch/i386/kernel/tsc.c *** END ***********************

static inline ktime_t
tick_now_cycles(void)
{
	/* lifted from sched_clock() */

	unsigned long long this_offset = 0;

	/* read the Time Stamp Counter: */
	rdtscll(this_offset);

	/* return the value in ktime_t */
	return ktime_set(0, tick_cycles_2_ns(this_offset));
}

static inline ktime_t
tick_now(void)
{
#ifdef TICK_MONOTONIC
#error TICK_MONOTONIC not supported. Does it work?
	return tick_now_cycles();
#else
 	return tick_now_gettimeofday();
#endif
}

static inline unsigned long long
tick_now_ns(void)
{
	return ktime_to_ns(tick_now());
}

static inline struct timeval
tick_now_timeval(void)
{
	return ktime_to_timeval(tick_now());
}

static inline char *
tick_now_timestamp_str(char *str, int strlen)
{
#ifdef TICK_MONOTONIC
	snprintf(str, strlen, "monotonic");
#else
	snprintf(str, strlen, "gettimeofday");
#endif
	return str;
}

/**************************************************
 * Precise gap generation at low CPU load.
 *
 * To generate gaps at sub-microsec precision we use a combination of
 * hrtimers and busy waiting. Using hrtimers alone is not sufficient.
 * So we use hrtimers for the initial portion of the gap up until a
 * threshold, and after that we do busy waiting.
 *
 * - We use tick_nanosleep_abs() to sleep with hrtimers.
 *   This is a partial copy of hrtimer_nanosleep().
 *   We could not use the original directly because it is not exported.
 *   
 * - For busy waiting we use ndelay()
 *
 * This scheme needs two variables to be calibrated:
 *
 * - gap_busywait_threshold_ns     
 *
 *   This is the end portion of the gap that has to be generated using
 *   busywaiting. This is needed that hrtimers do not appear to be
 *   precise to our requirement of sub-microsecond.
 *
 * - gap_busywait_adjust_percent
 *
 *   After we return from the hrtimer, we need to busy wait for the
 *   remaining part of the gap. This variable reduces the actual
 *   number of remaining nanoseconds (a value of 75 reduces it by 1/4).
 *   This is needed because of the overhead associated with retrieving
 *   wallclock time and executing ndelay().
 *
 * We perform the calibration in tick_gapsleep_calibrate().
 */

static inline void
tick_gapsleep_hrtimer(ktime_t endtime, long adjust_ns)
{
        int rc = 0;
        struct timespec ts = ktime_to_timespec(ktime_sub_ns(endtime, adjust_ns));

        // if the expiration has passed then nanosleep returns immediately
        // (see code in do_nanosleep() where it checks for task)
        rc = tick_nanosleep_abs(&ts, CLOCK_REALTIME);
        if (rc) {
                ERRFN("nanosleep failed: rc %d expires %lld orig %lld adj %ld",
                      rc,
                      ktime_to_ns(timespec_to_ktime(ts)),
                      ktime_to_ns(endtime),
                      adjust_ns);
        }
}

static inline void
__tick_gapsleep(ktime_t gap,
                ktime_t from_time,
                long busywait_threshold_ns,
                long busywait_adjust_percent)
{
        ktime_t gap_ends = ktime_add(from_time, gap);
        long remaining_ns = 0;

        if (ktime_equal(gap, KTIME_ZERO))
                return;

        // use an hrtimer to sleep for initial portion of the gap
        if (ktime_to_ns(gap) > busywait_threshold_ns)
                tick_gapsleep_hrtimer(gap_ends, busywait_threshold_ns);

        // use busy wait for the remaining
        remaining_ns = ktime_to_ns(ktime_sub(gap_ends, tick_now()));
        if (remaining_ns < 0)
                return;

        // adjust due to overhead
        remaining_ns *= busywait_adjust_percent;
        remaining_ns /= 100;
        ndelay(remaining_ns);
        return;
}

static inline void
tick_gapsleep(ktime_t gap, ktime_t from_time) {
        // busywait values set by tick_gapsleep_calibrate()
        __tick_gapsleep(gap, from_time,
                        tick.gap_busywait_threshold_ns,
                        tick.gap_busywait_adjust_percent);
}

/*
 * XXX TODO: The calibration is not automatic for now.
 * We manually set values for the emulab pc3000 hosts.
 *
 * For hints how to set up auto-calibration, see mgrp/doc/tick.calibration.
 */
static inline void
tick_gapsleep_calibrate(void)
{
        LOGFN("manual calibration for emulab");
        tick.gap_busywait_threshold_ns = 40000;
        tick.gap_busywait_adjust_percent = 85;

        LOG("    busywait_threshold_ns    %d", tick.gap_busywait_threshold_ns);
        LOG("    busywait_adjust_percent  %d", tick.gap_busywait_adjust_percent);
}
#endif
