/*
 * UDPex	UDP Extended, Transport unit built on top of BHTP
 *
 * Created:	Thu Aug 18 19:15:52 EDT 2005
 *
 * Version:	$Id: proto.c,v 1.7 2005/11/23 00:10:35 pavlos Exp $
 *
 * Authors:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 */

/* Must come before log.h. Otherwise we get compiler warning  */
#define LOG_PREFIX "udpex: "
#include "log.h"

#include "udpex.h"

#include <linux/ip.h>
#include <net/inet_hashtables.h>

#include <net/protocol.h>
#include <net/ip.h>
#include <net/inet_common.h>
#include <linux/vmalloc.h>

#include "bhtp_abi.h"

#include <linux/types.h>

// I know that this is horrible. Substitute with a hash table FIXME
#define UDPEX_DPORT_MIN		49152
#define UDPEX_DPORT_MAX		65535

#define UDPEX_PORTS_DYNAMIC	(UDPEX_DPORT_MAX - UDPEX_DPORT_MIN)
#define UDPEX_PORTS		65536	/* power of two please */

#define MAX_UDPEX_HEADER	sizeof(struct udpexhdr) + MAX_HEADER

struct udpex_ports {
	int used_ports_num;
	char used_ports[UDPEX_PORTS>>3];
	u16 last_port_reserved;
	struct sock *sk[UDPEX_PORTS];
	struct sock_cached *last_sk;
};

struct udpex_state {
	struct proto prot;
	struct proto_ops ops;
	struct inet_protosw protosw;
	struct bhtp_tun_prot tun_prot;
	struct bhtp_tun_ops tun_ops;
	struct udpex_ports *ports;
	struct udpex_options *options;
};

extern struct udpex_state udpex;

// We store the stats of the first UDPEX_BARRIER_STATS_MAX packets
// in udpex_sock.barrier_stats[i] when sending in barrier mode.
// Only needed when probe_show is on (to see how well we did generating gaps).
//
// Increase this to be at least as large as the largest packet train that
// you intend to send (see udpex_transmit_barrier_skbs)
#define UDPEX_BARRIER_STATS_MAX         300

struct udpex_barrier_stats {
        ktime_t gap;            // gap for which we slept/busywaited
        ktime_t orig_gap;       // before changing to minimum valid gap
        ktime_t actual_gap;     // actual gap from timestamp diff
};

struct udpex_sock {
	/* inet_sock has to be the first member */
	struct inet_sock inet;

	/* FIXME in host order */
	u16 src_port;

	/* FIXME in network order */
	u16 dst_port;

	int connected;

	struct bhtp_flow *bflow;
	int flags;

	struct bhtp_tun tun;

	int probe;
	int probe_keep;

	int queue;

	int tstamp;

	int barrier_down;
	ktime_t barrier_lowered;
	/* list of queued skbs that are waiting for the barrier */
	struct sk_buff_head barrier;
	int barrier_count;

        // stats for probes sents within a barrier
        // declared here because they take too much space on the limited stack
        // see udpex_transmit_barrier_skbs()
        struct udpex_barrier_stats barrier_stats[UDPEX_BARRIER_STATS_MAX];

	struct udpex_probe_results probe_results;

        // helpers to compute duration of a probe train in udpex_sendmsg()
        ktime_t barrier_probe_send_start;
        ktime_t barrier_probe_send_end;
        ktime_t barrier_probe_duration;
        int barrier_probe_prev_mgrpsize;
};

// Returns the size of the UDPex/MGRP packet that would result
// if the skb is sent out by itself (takes padding into account).
//
// The size includes the MGRP and UDPex header but not the IP header.
// The skb should have a prober qskb.
static inline int
udpex_mgrp_packet_size(struct sk_buff *skb)
{
        int size = 0;
	struct bhtp_qskb *qskb = bhtp_qsk(skb);

        size += sizeof(struct bhtphdr);
        size += sizeof(struct udpexhdr);
        size += skb->len + qskb->tun_pad;

        return size;
}
 
static inline int
udpex_reserve_port(struct udpex_ports *ports, unsigned short snum)
{
	int port_taken = 1;
	int port_provided = (snum > 0);

	if (port_provided) {
		port_taken = test_and_set_bit(snum, (long *)ports->used_ports);
		if (port_taken)
			goto error_port_in_use;
	
	/* let us pick port */
	} else {
		int attempts = 0;
		snum = ports->last_port_reserved;
		do {
			if (attempts++ > UDPEX_PORTS_DYNAMIC)
				break;

			if (snum == UDPEX_DPORT_MAX)
				snum = UDPEX_DPORT_MIN;
			else
				snum++;

			port_taken = test_and_set_bit(snum,
						(long *)ports->used_ports);

		} while (port_taken);

		if (port_taken)
			goto error_no_more_ports;
	}

	if (!port_provided)
		ports->last_port_reserved = snum;

	ports->used_ports_num++;
	DEBUG("reserved port %d", snum);
	return snum;

/* Are these the correct errors? FIXME */
error_no_more_ports:
	DEBUG("failed: no more dynamic ports available");
	return -EADDRNOTAVAIL;
error_port_in_use:
	DEBUG("failed: port %u in use", snum);
	return -EADDRINUSE;
}

static inline struct udpex_sock *
udpex_sk(const struct sock *sk)
{
	return (struct udpex_sock *)sk;
}

static void
udpex_close(struct sock *sk, long timeout)
{
	DEBUG("UDPex: close");
	sk_common_release(sk);
}

int
udpex_disconnect(struct sock *sk, int flags)
{
	DEBUG("UDPex: disconnect");
	return 0;
}


static inline int
__udpex_connect(struct sock *sk, struct sockaddr *uaddr, int addr_len,
		int transient)
{
	/* instead of calling ip4_datagram_connect */

	int rc = 0;
	struct udpex_sock *up = udpex_sk(sk);
	struct sockaddr_in *usin = (struct sockaddr_in *) uaddr;
	struct bhtp_flow *bflow = NULL;
	
	if (up->connected)
		return -EISCONN;

	if (addr_len < sizeof(*usin)) 
	  	return -EINVAL;

	if (usin->sin_family != AF_INET) 
	  	return -EAFNOSUPPORT;

	/* Try to use the existing bhtp flow */
	/* XXX Optimization for repeated sendmsg without connect */
	if (up->bflow != NULL && bhtp_flow_matches_addr(up->bflow, usin))
		bflow = up->bflow;

	/* Ask BHTP for a flow based on usin->sin_addr (port ignored)
	 *
	 * BHTP will either create a new one or returning
	 * an existing one depending on the flags.
	 */
	if (bflow == NULL) {
		/* only sin_addr is used; sin_port is ignored in the flow */
		bflow = bhtp_find_attach_flow(usin, up->flags);
		if (bflow == NULL) {
			rc = -EINVAL;
			goto error;
		}
	}

	/* Ok, we found the flow */
	up->bflow = bflow;
	up->dst_port = usin->sin_port;
	if (!transient)
		up->connected = 1;

	return 0;

error:
	return rc;
}


int
udpex_connect(struct sock *sk, struct sockaddr *uaddr, int addr_len)
{
	DEBUG("UDPex: connect");
	return __udpex_connect(sk, uaddr, addr_len, 0);
}

int
udpex_ioctl(struct sock *sk, int cmd, unsigned long arg)
{
	struct udpex_sock *up = udpex_sk(sk);
	int rc = 0;

	DEBUG("ioctl cmd %d", cmd);

	if (_IOC_TYPE(cmd) != UDPEX_IOC_MAGIC)
		return -ENOTTY;

	switch(cmd) {

	case UDPEX_IOC_ISACTIVE:
		return 0;

	case UDPEX_IOC_PROBE:
		up->probe = 1;
		/* FIXME: This should read the probe_keep size; set to 100 */
		up->probe_keep = 100;
		LOG("ioctl IOC_PROBE: probe_keep = %d", up->probe_keep);
		return 0;

	case UDPEX_IOC_QUEUE:
		up->queue = 1;
		LOG("ioctl IOC_QUEUE");
		return 0;

	case UDPEX_IOC_PROBE_RESULTS:
		rc = copy_to_user(
			(struct udpex_probe_results __user *)arg,
			&up->probe_results,
			sizeof(up->probe_results));
		if (rc)
			return rc;
		return 0;

	default:
		return -ENOIOCTLCMD;
	}
	return(0);
}

static inline void
udpex_add_header(struct sock *sk, struct sk_buff *skb)
{
}

/* SKB is headerless - no UDPex header yet */
int
udpex_transmit_skb(struct sock *sk, struct sk_buff *skb,
                   struct mgrp_probe_feedback *feedback)
{
	int rc = 0;

	struct udpex_sock *up = udpex_sk(sk);
	struct udpexhdr *uh = NULL;

	struct bhtp_qskb *qskb = NULL;

	if (unlikely(skb == NULL))
		return -ENOBUFS;

/* 	DEBUG("udpex_transmit_skb"); */

	/* private UDPex data in Control Buffer */
	qskb = bhtp_qsk(skb);

	/* reserve space for the UDPex header */
	skb->transport_header = skb_push(skb, sizeof(struct udpexhdr));
	uh = (struct udpexhdr *)skb->transport_header;

	if (!skb->sk) {
		DEBUG("skb_set_owner");
		skb_set_owner_w(skb, sk);
	}

	/* skb should not be routed */
	BUG_ON(skb->dst != NULL);

	/* Build Header */
	/* FIXME src is in host order but dst is in network order */
	uh->source = htons(up->src_port);
	uh->dest = up->dst_port;

	rc = bhtp_schedule_skb(up->bflow, skb, feedback);
	if (rc)
		goto drop;

	return 0;

drop:
	kfree_skb(skb);
	ERR("transmit: packet DROPPED");
	return rc;
}

static inline int
udpex_set_dest(struct sock *sk, struct msghdr *msg, int *connected)
{
	/* Lifted almost verbatim from udp_sendmsg() */

	struct inet_sock *inet = inet_sk(sk);

	if (msg->msg_name) {
		struct sockaddr_in * usin = (struct sockaddr_in*)msg->msg_name;

		if (msg->msg_namelen < sizeof(*usin))
			return -EINVAL;
		if (usin->sin_family != AF_INET) {
			if (usin->sin_family != AF_UNSPEC)
				return -EAFNOSUPPORT;
		}

		inet->daddr = usin->sin_addr.s_addr;
		inet->dport = usin->sin_port;
		if (inet->dport == 0)
			return -EINVAL;
	} else {
		if (sk->sk_state != TCP_ESTABLISHED)
			return -EDESTADDRREQ;
		/* Open fast path for connected socket.
		   Route will not be used, if at least one option is set.
		 */
		*connected = 1;
  	}

	return 0;
}

static inline void
bhtp_recv_timestamp(struct msghdr *msg, struct sock *sk, struct sk_buff *skb)
{
	/* similar to sock_recv_timestamp() */
	struct udpex_sock *up = udpex_sk(sk);

	if (up->tstamp) {
		struct bhtp_timestamps tstamps;
		bhtp_recv_get_timestamps(sk, skb, &tstamps);
		put_cmsg(msg, SOL_MGRP, UDPEX_TSTAMP,
			sizeof(struct bhtp_timestamps), &tstamps);
	}
}


struct udpex_cmsgs {
	struct udpex_probe *probe;
};

static inline int
udpex_msghdr_parse(const struct msghdr *msg, struct udpex_cmsgs *cmsgs)
{
	/* based on sctp_msghdr_parse() called from sctp_sendmsg() */

	struct cmsghdr *cmsg = NULL;

	for (cmsg = CMSG_FIRSTHDR(msg);
	     cmsg != NULL;
	     cmsg = CMSG_NXTHDR((struct msghdr *)msg, cmsg)) {

		if (!CMSG_OK(msg, cmsg)) {
			LOG("msghdr_parse: cmsg not ok");
			LOG("    cmsg_len %d sizeof_cmsghdr %d msg_controllen %d",
				cmsg->cmsg_len,
				sizeof(struct cmsghdr),
				msg->msg_controllen);
			return -EINVAL;
		}

		if (cmsg->cmsg_level != SOL_MGRP)
			continue;

		switch (cmsg->cmsg_type) {

		case UDPEX_PROBE:
			if (cmsg->cmsg_len !=
			    CMSG_LEN(sizeof(struct udpex_probe))) {
				LOG("msghdr_parse: udpex_probe not ok");
				LOG("    cmsg_len %d sizeof_cmsghdr %d msg_controllen %d",
					cmsg->cmsg_len,
					sizeof(struct cmsghdr),
					msg->msg_controllen);
				return EINVAL;
			}
			cmsgs->probe = (struct udpex_probe *)CMSG_DATA(cmsg);
			break;

		default:
			return -EINVAL;
		}
	}
	return 0;
}


static inline struct bhtp_qskb *
udpex_fill_send_qskb(struct sock *sk, struct sk_buff *skb,
		struct udpex_cmsgs *cmsgs)
{
	struct udpex_sock *up = udpex_sk(sk);
	/* private UDPex data in Control Buffer */
	struct bhtp_qskb *qskb = bhtp_qsk_init(skb);

	qskb->tun = &up->tun;
	qskb->tun_hdr = sizeof(struct udpexhdr);
	qskb->tun_data_orig = skb->len;
	qskb->tun_data = skb->len;
	qskb->tun_len = qskb->tun_hdr + qskb->tun_data;

	if (cmsgs->probe != NULL) {
		if (DEBUG_ON)
			udpex_probe_print(cmsgs->probe);
		qskb->tun_pad = cmsgs->probe->pad;
		qskb->tun_gap = cmsgs->probe->gap_usec;
                qskb->tun_piggyback = !cmsgs->probe->piggyback_not_allowed;
		if (qskb->tun_pad > 0)
			qskb->type |= BHTP_QSKB_VESSEL;
	}

	qskb->type |= BHTP_QSKB_NEVER_QUEUE;

	/* HACK: temp */
	if (up->probe) {
		qskb->type |= BHTP_QSKB_PROBE;
		if (up->queue)
			qskb->type |= BHTP_QSKB_NEVER_QUEUE;

		/* FIXME, these lengths are hacks */
		qskb->tun_data = up->probe_keep;
		qskb->tun_pad = qskb->tun_data_orig - qskb->tun_data;
		qskb->tun_len = qskb->tun_hdr + qskb->tun_data;

		/* get rid of the padding; pad info stored in qskb->tun_pad */
		/* needed when appending the tun to a BHTP skb */
		/* see bhtp_append_tun_skb() */
		skb_trim(skb, qskb->tun_hdr + qskb->tun_data);

	} else {
		/* HACK: temp */
/* 		qskb->type |= BHTP_QSKB_NEVER_QUEUE; */
	}


	return qskb;
}

static inline ktime_t
udpex_smart_ndelay(ktime_t waitfor)
{
	long long adjusted_ns = ktime_to_ns(waitfor);

	/* FIXME
	 * Empirically we need a reduction in the sleep duration by 7%.
	 * Is this because of ndelay overhead or is this the result of
	 * BHTP overhead ?
	 *
	 * If it is not ndelay overhead, then the adhjustment should
	 * happen in udpex_waitfor_gap and not here.
	 *
	 * Also, use x/1024 instead of y/1000 for efficiency.
	 */
	adjusted_ns *= 885;
	adjusted_ns >>= 10;

	ndelay(adjusted_ns);

	return ktime_set(0, adjusted_ns);
}


static inline int
udpex_barrier_down(struct udpex_cmsgs *cmsgs)
{
	if (cmsgs == NULL)
		return 0;
	
	if (cmsgs->probe == NULL)
		return 0;

	return (cmsgs->probe->barrier != 0);
}

static inline int
udpex_can_wait(struct udpex_cmsgs *cmsgs)
{
	if (cmsgs == NULL)
		return 0;
	
	if (cmsgs->probe == NULL)
		return 0;

	return (cmsgs->probe->can_wait_for_usec);
}

static inline int
udpex_tell_results(struct udpex_cmsgs *cmsgs)
{
	if (cmsgs == NULL)
		return 0;
	
	if (cmsgs->probe == NULL)
		return 0;

	return (cmsgs->probe->tell_results);
}

void
udpex_drop_barrier_skbs(struct sock *sk)
{
	struct udpex_sock *up = udpex_sk(sk);
	struct sk_buff *skb = NULL;

	while ((skb = skb_dequeue(&up->barrier)) != NULL) {
		LOG("DROPPING barrier skb %d", skb->len);
		kfree_skb(skb);
	}
}

static inline void
udpex_barrier_stats_init(struct udpex_sock *up)
{
        int i = 0;
        for (i = 0; i < UDPEX_BARRIER_STATS_MAX; i++) {
                struct udpex_barrier_stats *stats = &up->barrier_stats[i];
                memset(stats, 0, sizeof(struct udpex_barrier_stats));
        }
}

static inline void
udpex_barrier_stats_update(struct udpex_barrier_stats *stats,
                           ktime_t cur, ktime_t prev,
                           ktime_t desired_gap,
                           ktime_t orig_gap)
{
        if (stats == NULL)
                return;
        stats->gap = desired_gap;
        stats->orig_gap = orig_gap;
        stats->actual_gap = ktime_sub(cur, prev);
}

static inline void
udpex_barrier_stats_debug(struct udpex_sock *up, int sent)
{
        int i = 0;

        for (i = 0; i < sent; i++) {
                ktime_t overrun = KTIME_ZERO;
                ktime_t underrun = KTIME_ZERO;
                struct udpex_barrier_stats *stats = &up->barrier_stats[i];

                if (i >= UDPEX_BARRIER_STATS_MAX) {
                        ERRFN("too many packets %d, truncated %d", sent, i);
                        break;
                }

                if (ktime_to_ns(stats->actual_gap) > ktime_to_ns(stats->gap))
                        overrun = ktime_sub(stats->actual_gap, stats->gap);
                else
                        underrun = ktime_sub(stats->gap, stats->actual_gap);

                LOG("barrier %d pkt %-3d over %-8lld under %-8lld actual %-8lld orig %-8lld",
                    up->barrier_count, i,
                    ktime_to_ns(overrun),
                    ktime_to_ns(underrun),
                    ktime_to_ns(stats->actual_gap),
                    ktime_to_ns(stats->orig_gap));
        }
}

// Back-to-back packets result invalid timestamps because
// packets are timestamped before they are sent.
//
// Adjust the gap based on the device transmission delay
// of the previous packet.
//
// See svn log -r2460
static inline long long
mgrp_adjusted_gap_ns(struct bhtp_flow *bflow,
                     long long gapns_orig,
                     int prev_mgrp_pktsize)
{
        long long gapns_min = 0;

        // no change, we do not know the device speed
        // (could happen for loopback interfaces)
        if (bflow->dst_dev_speed_mbps == 0)
                gapns_min = gapns_orig;

        // no change, could be the first of a packet burst
        // no need for any adjustment
        else if (prev_mgrp_pktsize == 0)
                gapns_min = gapns_orig;

        else
                gapns_min = mgrp_transmission_delay(prev_mgrp_pktsize,
                                                    bflow->dst_dev_speed_mbps);

        return max_t(long long, gapns_orig, gapns_min);
}

// returns zero for loopback or pseudo-interfaces
static inline long long
mgrp_adjusted_xmit_ns(struct bhtp_flow *bflow, int mgrp_pktsize)
{
        // use zero gap to get the transmissiom time
        return mgrp_adjusted_gap_ns(bflow, 0, mgrp_pktsize);
}

static inline int
udpex_in_barrier_window(struct udpex_sock *up, ktime_t kt)
{
        long long ns = ktime_to_ns(kt);
        long long start_ns = ktime_to_ns(up->barrier_probe_send_start);
        long long end_ns = ktime_to_ns(ktime_add(up->barrier_probe_send_start,
                                                 up->barrier_probe_duration));

        return (ns >= start_ns && ns <= end_ns);
}
                        

int
udpex_transmit_barrier_skbs(struct sock *sk, struct udpex_probe_results *res)
{
        int rc = 0;
        struct sk_buff *skb = NULL;
        int sent = 0;                   // 0,1,2,..
        int probe_pos_in_train = 0;     // 1,2,3..
        ktime_t prev = KTIME_ZERO;
        struct mgrp_probe_feedback prev_probe = { 0 };

        struct udpex_sock *up = udpex_sk(sk);
        int probe_show = udpex.options->probe_show;

        up->barrier_count++;

        if (probe_show)
                udpex_barrier_stats_init(up);

        /* drain the skb queue, barrier has been lifted */
        while ((skb = skb_dequeue(&up->barrier)) != NULL) {

                ktime_t now;
                struct bhtp_qskb *qskb = bhtp_qsk(skb);

                long long gapns_orig = bhtp_qsk(skb)->tun_gap * NSEC_PER_USEC;
                ktime_t gap_orig = ktime_add_ns(KTIME_ZERO, gapns_orig);

                long long gapns = mgrp_adjusted_gap_ns(up->bflow,
                                                    gapns_orig,
                                                    prev_probe.mgrp_bytes);
                ktime_t gap = ktime_add_ns(KTIME_ZERO, gapns);

                probe_pos_in_train++;
                qskb->vessel.probe_pos_in_train = probe_pos_in_train;
                qskb->vessel.probes_in_train = up->tun.barrier_pkts;

                if (sent == 0) {
                        // no gap for first probe
                        prev = tick_now();

                } else {
                        struct udpex_barrier_stats *stats = NULL;

                        if (probe_show && sent < UDPEX_BARRIER_STATS_MAX)
                                stats = &up->barrier_stats[sent];

                        // generate the gap
                        tick_gapsleep(gap, prev);
                        now = tick_now();
                        if (probe_show)
                                // record overruns and diffs for gap stats
                                udpex_barrier_stats_update(stats, now, prev,
                                                           gap, gap_orig);
                        prev = now;
                }

                // prepends UDPex header and sends to MGRP
                rc = udpex_transmit_skb(sk, skb, &prev_probe);
                if (rc) {
                        kfree_skb(skb);
                        ERRFN("udpex transmit failed (%d) for probe %d len %d",
                              rc, probe_pos_in_train, skb->len);
                        goto drop_barrier_skbs;
                }
                sent++;
                if (res) {
                        res->total_pkts++;
                        res->total_bytes += prev_probe.mgrp_bytes;
                        if (prev_probe.pbk_bytes > 0) {
                                res->pbk_pkts++;
                                res->pbk_bytes += prev_probe.pbk_bytes;
                                if (udpex_in_barrier_window(up, prev_probe.rider_orig_send_time)) {
                                        res->pbk_inwin_pkts++;
                                        res->pbk_inwin_bytes += prev_probe.pbk_bytes;
                                }
                        }
                }
                DEBUGFN("barrier UP, pkt %d, gapns %lld", sent, gapns);
        }

        // Tell MGRP that we are done with the probe burst so
        // that MGRP can flush any half-piggybacked packets.
        mgrp_flush_piggybacking(up->bflow, &prev_probe);

        if (probe_show)
                udpex_barrier_stats_debug(up, sent);

        return 0;

drop_barrier_skbs:
        udpex_drop_barrier_skbs(sk);
        return rc;
}

#define UDPEX_DELAY_NONE	0
#define UDPEX_DELAY_TIMEOUT	1
#define UDPEX_DELAY_ACCUMULATED	2

#define UDPEX_DELAY2STR(_code) \
	((_code) == UDPEX_DELAY_TIMEOUT ? "timeout" : \
	 (_code) == UDPEX_DELAY_ACCUMULATED ? "accumulated" : \
	 "nodelay")

int
udpex_send_barrier_skbs(struct sock *sk, struct udpex_probe_results *res)
{
	int rc = 0;
	struct udpex_sock *up = udpex_sk(sk);
        ktime_t now = tick_now();
	ktime_t mark = now;
	ktime_t twait = ktime_set(0, 0);
	ktime_t tsend = ktime_set(0, 0);

	int added_delay = UDPEX_DELAY_NONE;

	int probe_timeout = udpex.options->probe_timeout;
	int probe_accumulated = udpex.options->probe_accumulated;
	int probe_show = udpex.options->probe_show;

	int payload_buffer_on = 0;
	long long actual_accumulated = 0;

	if (!probe_timeout)
		goto send_now;

	/* code to wait until timeout for enough data */

	payload_buffer_on = mgrp_flow_payload_buffer_start(up->bflow);

	if (probe_timeout < 0) {
		ERR("send_barrier_skbs: negative probe timeout %d",
			probe_timeout);
		goto send_now;
	}

	while (1) {
		twait = ktime_sub(tick_now(), mark);

		/* timeout expired */
		if (ktime_to_ns(twait) >= probe_timeout) {
			added_delay = UDPEX_DELAY_TIMEOUT;
			break;
		}

		/* enough bytes accumulated */
		if (up->bflow->tosend_accumulated > probe_accumulated) {
			added_delay = UDPEX_DELAY_ACCUMULATED;
			actual_accumulated = up->bflow->tosend_accumulated;
			goto send_now;
		}

		msleep(1);
	}

send_now:
        now = tick_now();
	twait = ktime_sub(now, mark);
	mark = now;
        // helper used to separate piggybacking portions in @res
        up->barrier_probe_send_start = now;

	rc = udpex_transmit_barrier_skbs(sk, res);

        now = tick_now();
	tsend = ktime_sub(now, mark);
        up->barrier_probe_send_end = now;

	if (payload_buffer_on)
		mgrp_flow_payload_buffer_stop(up->bflow);

	if (rc) {
		char tmp[100];
		ERR("send_barrier_skbs: xmit failed: %s",
			rc2str(rc, tmp, sizeof(tmp)));
	}

	if (res) {
                // fill probe_results
                // res has already been partiall filled by
                //      udpex_transmit_barrier_skbs()
                // res will be partially filled later by caller (res->t1)
                res->sent = 1;
                res->t2 = ktime_to_ns(twait);
                res->t3 = ktime_to_ns(tsend);

                // computed above
                res->send_start = ktime_to_ns(up->barrier_probe_send_start);
                res->send_end = ktime_to_ns(up->barrier_probe_send_end);
                // computed earlier in the caller udpex_sendmsg()
                res->send_precomputed_duration = ktime_to_ns(up->barrier_probe_duration);
        }

	if (probe_show)
		LOG("send_barrier_skbs: "
		    "barrier %d [%s] twait %lld tsend %lld accumulated %lld",
			up->barrier_count,
			UDPEX_DELAY2STR(added_delay),
			ktime_to_ns(twait),
			ktime_to_ns(tsend),
			actual_accumulated);

	return rc;
}

int
udpex_sendmsg(struct kiocb *iocb, struct sock *sk, struct msghdr *msg,
	     size_t len)
{
	struct udpex_sock *up = udpex_sk(sk);
	int noblock = msg->msg_flags & MSG_DONTWAIT;
	struct bhtp_qskb *qskb = NULL;

	struct sk_buff *skb = NULL;
	int rc = 0;
	int size = 0;

	int wait_usec = 0;
	int barrier_was_down = 0;
	int barrier_down = 0;
	int first_time_barrier_down = 0;
	int single_send = 0;
	int available = 0;

        long long gap_ns = 0;

	struct udpex_probe_results res = {.sent = 0};
	struct udpex_cmsgs cmsgs = {
		.probe = NULL,
	};

/* 	DEBUG("sendmsg"); */
/* 	DEBUG("max_h %d len %d", sk->sk_prot->max_header, len); */
	DEBUG("put_cmsg: ORIG %p %d", msg->msg_control, msg->msg_controllen);

	lock_sock(sk);
	size = sk->sk_prot->max_header + len;
        if (!udpex_sk(sk)->connected && msg->msg_name == NULL) {
                ERRFN("Destination address required");
                rc = -EDESTADDRREQ;
                goto err_release;
        }
	if (msg->msg_name != NULL) {
		rc = __udpex_connect(sk, msg->msg_name, msg->msg_namelen, 1);
		if (rc)
			goto err_release;
	}
	release_sock(sk);

	/* parse ancillary data */
	rc = udpex_msghdr_parse(msg, &cmsgs);
	if (rc)
		return rc;
	
	skb = sock_alloc_send_skb(sk, size, noblock, &rc);

	lock_sock(sk);

	// XXX proper error?
	rc = -ENOMEM;
	if (skb == NULL)
		goto err_release;

	/* copy user data in skb */
	skb_reserve(skb, sk->sk_prot->max_header);
	rc = memcpy_fromiovec(skb_put(skb, len), msg->msg_iov, len);
	if (rc != 0)
		goto err_free;

	qskb = udpex_fill_send_qskb(sk, skb, &cmsgs);

	wait_usec = udpex_can_wait(&cmsgs);
	barrier_down = udpex_barrier_down(&cmsgs);
	barrier_was_down = !barrier_down && up->barrier_down;
	first_time_barrier_down = barrier_down && !up->barrier_down;
	single_send = !barrier_down && !barrier_was_down;
	up->barrier_down = barrier_down;

	/* normal case: send only one packet with no barrier */

        if (single_send) {
                struct mgrp_probe_feedback feedback = { 0 };
                long long gapns = qskb->tun_gap * NSEC_PER_USEC;
                if (wait_usec > 0) {
                        ERRFN("can_wait_for_usec incompatible with single send %d usec", wait_usec);
                        goto err_free;
                }
                if (gapns > 0) {
                        ERRFN("unsupported: non-zero gap %lld for single send", gapns);
                        ERRFN("use barriers instead");
                        goto err_free;
                }

		/* send the packet passed to sendmsg */
		rc = udpex_transmit_skb(sk, skb, &feedback);
		if (rc)
			goto err_release;
                // XXX feedback not copied to @res for single send
		goto done;
	}

	/* special case: barrier is down or was just lifted */

	if (first_time_barrier_down) {
		/* time that barrier was first lowered */
		up->barrier_lowered = tick_now();

                // This will record the time it takes to transmit
                // the whole probe trains (between the barriers)
                // including both transmission times and gaps
                // between packets.
                //
                // This will be used to decide which portion of the
                // piggybacked bytes would have been sent between the
                // probes if they were not piggybacked (pathload compensation)
                up->barrier_probe_send_start = KTIME_ZERO;
                up->barrier_probe_send_end = KTIME_ZERO;
                up->barrier_probe_duration = KTIME_ZERO;
                up->barrier_probe_prev_mgrpsize = 0;

		up->tun.barrier_total = 0;
		up->tun.barrier_min = (short)-1;
		up->tun.barrier_max = 0;
		up->tun.barrier_wait_usec = 0;
		up->tun.barrier_pkts = 0;
	}

        /**********************/

         /* compute duration of probe train */

        // gap from previous packet
        // needs to be adjusted based on previous packet
        // the first packet has tun_gap=0 and prev_pktsize=0, so returns 0
        //
        // same as adjustment in udpex_transmit_barrier_skbs()
        // here we only need it to calculate the duration of the probe train
        // in udpex_transmit_barrier_skbs() we actually transmit the probes
        // and need to know the gap to generate; we could adjust the gaps here
        gap_ns = mgrp_adjusted_gap_ns(up->bflow,
                                      qskb->tun_gap * NSEC_PER_USEC,
                                      up->barrier_probe_prev_mgrpsize);
        // time it takes to transmit the current probe
        // so this is the contribution of the current gap+probe to the
        // duration of the whole probe transaction
        up->barrier_probe_duration = ktime_add_ns(up->barrier_probe_duration, gap_ns);
        up->barrier_probe_prev_mgrpsize = udpex_mgrp_packet_size(skb);

        /**********************/

        /* update barrier stats */
	available = qskb->tun_pad;
	up->tun.barrier_wait_usec = wait_usec;
	up->tun.barrier_total += available;
	up->tun.barrier_min = min_t(int, available, up->tun.barrier_min);
	up->tun.barrier_max = max_t(int, available, up->tun.barrier_max);
	up->tun.barrier_pkts += 1;

	/* queue skbs for later dispatch */
	skb_queue_tail(&up->barrier, skb);

	DEBUG("sendmsg: barrier %s, pkt QUEUED",
		(barrier_down ? "DOWN" : "JUST LIFTED"));
	if (barrier_down && wait_usec > 0)
		ERR("sendmsg: can_wait_for_usec incompatible with "
		    "barrier down %d usec", wait_usec);

	if (barrier_down)
		goto done;

	/* reaches here only if barrier was just lifted (we can now send pkts) */

	res.t1 = ktime_to_ns(ktime_sub(tick_now(), up->barrier_lowered));
	if (wait_usec > 0)
		DEBUG("sendmsg: can wait for %d usec", wait_usec);

	rc = udpex_send_barrier_skbs(sk, &res);
	if (rc)
		goto err_release;

	if (udpex_tell_results(&cmsgs)) {
		DEBUG("put_cmsg: %p %d len %d",
			msg->msg_control, msg->msg_controllen,
			sizeof(struct udpex_probe));
		/* store so that it can be retrieved later by ioctl */
		up->probe_results = res;
		if (rc)
			ERR("failed to put_cmsg rc %d", rc);
	}

	if (DEBUG_ON) {
		char tmp[200];
		ERR("sendmsg: barrier %s",
			udpex_probe_results_str(tmp, sizeof(tmp), &res));
	}
done:
	release_sock(sk);
	return len;

err_free:
	ERRFN("DROPPED skb %d", skb->len);
	kfree_skb(skb);
err_release:
	release_sock(sk);
	ERRFN("failed, rc %d", rc);
	return rc;
}

static int
udpex_recvmsg(struct kiocb *iocb, struct sock *sk, struct msghdr *msg,
	     size_t len, int noblock, int flags, int *addr_len)
{
	/* Taken from udp_recvmsg. */

	struct inet_sock *inet = inet_sk(sk);
  	struct sockaddr_in *sin = (struct sockaddr_in *)msg->msg_name;
  	struct sk_buff *skb = NULL;
  	int copied = 0;
	int err = 0;


/* 	DEBUG("recvmsg"); */

	/*
	 *	Check any passed addresses
	 */
	if (addr_len)
		*addr_len=sizeof(*sin);

	if (flags & MSG_ERRQUEUE)
/* 		return ip_recv_error(sk, msg, len); */
		BUG();

	skb = skb_recv_datagram(sk, flags, noblock, &err);
	if (!skb)
		goto out;
  
  	copied = skb->len - sizeof(struct udpexhdr);
	if (copied > len) {
		copied = len;
		msg->msg_flags |= MSG_TRUNC;
	}

	err = skb_copy_datagram_iovec(skb,
			sizeof(struct udpexhdr),
			msg->msg_iov,
			copied);

	if (err)
		goto out_free;

	sock_recv_timestamp(msg, sk, skb);
	bhtp_recv_timestamp(msg, sk, skb);

	/* Copy the address. */
	if (sin)
	{
		struct iphdr *iph = (struct iphdr*) skb->network_header;
		struct udpexhdr *uh = (struct udpexhdr *)skb->transport_header;
		sin->sin_family = AF_INET;
		sin->sin_port = uh->source;
		sin->sin_addr.s_addr = iph->saddr;
		memset(sin->sin_zero, 0, sizeof(sin->sin_zero));
  	}
	if (inet->cmsg_flags)
		ip_cmsg_recv(msg, skb);

	err = copied;
	if (flags & MSG_TRUNC)
		err = skb->len - sizeof(struct udpexhdr);
  
out_free:
  	skb_free_datagram(sk, skb);
out:
  	return err;
}

static int
udpex_backlog_rcv(struct sock * sk, struct sk_buff *skb)
{
	DEBUG("UDPex: backlog_rcv");
	return 0;
}

static void
udpex_hash(struct sock *sk)
{
	DEBUG("hash");
	BUG();
}

static void
udpex_unhash(struct sock *sk)
{
	DEBUG("UDPex: unhash (inet_unhash)");
}

static int
udpex_get_port(struct sock *sk, unsigned short snum)
{
	// FIXME What about locking?
	int rc = 0;
	struct inet_sock *inet = inet_sk(sk);
	struct udpex_sock *up = udpex_sk(sk);

	DEBUG("getport, asking for %u", snum);

	rc = udpex_reserve_port(udpex.ports, snum);
	if (rc < 0)
		return 1;
	
	snum = rc;
	
	BUG_TRAP(snum != 0);

	// hack, FIXME, this should be a hash
	udpex.ports->sk[snum] = sk;

	up->src_port = snum;
	inet->num = snum;
	DEBUG("getport, ok, got %u", up->src_port);
	return 0;
}

inline struct sock *
udpex_port_to_sock(unsigned short snum)
{
/* 	DEBUG("looking for snum %d", snum); */
	return udpex.ports->sk[snum];
}

unsigned int
udpex_poll(struct file *file, struct socket *sock, poll_table *wait)
{
	return datagram_poll(file, sock, wait);
}

int
udpex_tun_rcv(struct sk_buff *skb, struct sockaddr_in *remote_addr)
{
	int rc = 0;
  	struct sock *sk = NULL;
	struct udpexhdr *uh = NULL;
	struct bhtp_qskb *qskb = NULL;

/* 	DEBUG("rcv from BHTP"); */

	if (!pskb_may_pull(skb, sizeof(struct udpexhdr)))
		BUG();

	uh = (struct udpexhdr *)skb->transport_header;

	sk = udpex_port_to_sock(ntohs(uh->dest));
	if (sk == NULL) {
                if (udpex_op.debug_port_unreachable) {
                        ERR("rcv: port_to_sock failed for sk == NULL");
                        ERR("rcv:    src,dest %hu,%hu, ntoh %hu,%hu",
                                uh->source, uh->dest,
                                ntohs(uh->source), ntohs(uh->dest));
                        goto drop;
                } else {
                        goto drop_silent;
                }
	}

	/* the control buffer in sk_buff is overloaded for bhtp private data */
	/* the qskb is partially filled in bhtp_rcv_tun() */
	qskb = bhtp_qsk(skb);

	/* fill the rest of the fields */
	/* fields that are valid: tun_len and tun_pad */
	/* use udpex_transmit_skb() for references */
	qskb->tun = &udpex_sk(sk)->tun;
	/* qskb->tun_len is valid */
	/* qksb->tun_pad is valid */
	qskb->tun_hdr = sizeof(struct udpexhdr);
	qskb->tun_data = qskb->tun_len - qskb->tun_hdr;
	qskb->tun_data_orig = qskb->tun_data + qskb->tun_pad;

	if (qskb->tun_pad > 0) {
		// FIXME: See "recreate the padding" svn log:
		//
		//	cd bhtp
		//	svn log -r261 .
		//
		// We would like to pass this as ancillary data
		// instead of extending the skb to include padding
		// that is useless anyhow.
		// Obviously we need this for UDP compatibility mode.
		// But if our receiver understands BHTP then it can
		// receive the padding as a length parameter in an
		// ancillary data in recv() instead of doing memcopies
		// between kernel and user-space.

		int padding = qskb->tun_pad;
		void *padptr = NULL;
		struct sk_buff *copy = NULL;
		
		DEBUG("tun_rcv: TUN has %d padding", qskb->tun_pad);

		// FIXME
		// Instead of creating a copy, can I just use the
		// skb fragments? How to do that?

		copy = pskb_copy(skb, GFP_ATOMIC);
		if (copy == NULL) {
			ERR("tun_rcv: failed to copy TUN skb");
			goto drop;
		}

		/* operate on the copy from now on */
		kfree_skb(skb);
		skb = copy;

		if (skb_tailroom(skb) < padding) {
			/*
			 * FIXME: this forces us to put the vessel first
			 * in the packet. This is an unecessary limitation.
			 * See bhtp_transmit_vessel_skb().
			 * Try to avoid copying.
			 *
			 * This should not happen since padded TUN is
			 * usually the first TUN, see bhtp_schedule_skb().
			 * The skb head should be large enough to include
			 * the expanded TUN (hdr + data + padding).
			 */
			ERR("tun_rcv: TUN has %d pad but only %d left in rcv skb",
				padding, skb_tailroom(skb));
			padding = skb_tailroom(skb);
		}

		/* since we have copied the head, we can overwrite the padding */
		padptr = skb_put(skb, padding);
		memset(padptr, BHTP_PAD_BYTE, padding);
	}

	/* Hand over to the socket to do the recv(). */

	/* FIXME: How do we do custom ancillary data? */
	/*	  needed to pass padding and timestamps */

	rc = sock_queue_rcv_skb(sk, skb);
	if (rc < 0) {
		ERR("rcv: sock_queue_rcv_skb failed rc %d", rc);
		goto drop;
	}

	return 0;

drop:
	ERR("rcv: packet DROPPED");
drop_silent:
	kfree_skb(skb);
	return 0;
}

int
do_udpex_getsockopt(struct sock *sk, int level, int optname,
	 	    char __user *optval, int __user *optlen)
{
	/* modeled after do_udp_getsockopt() */

	struct udpex_sock *up = udpex_sk(sk);
	int val = 0;
	int len = 0;

	if (get_user(len, optlen))
		return -EFAULT;
	
	len = min_t(unsigned int, len, sizeof(int));

	if (len < 0)
		return -EINVAL;
	
	switch (optname) {
	
	case UDPEX_TSTAMP:
		val = up->tstamp;
		break;
	
	default:
		return -ENOPROTOOPT;
	}

	if (put_user(len, optlen))
		return -EFAULT;
	if (copy_to_user(optval, &val, len))
		return -EFAULT;
	return 0;
}

 	
int
do_udpex_setsockopt(struct sock *sk, int level, int optname,
		    char __user *optval, int optlen)
{
	/* modeled after do_udp_setsockopt() */

	struct udpex_sock *up = udpex_sk(sk);
	int val = 0;
	int err = 0;

	if (optlen < sizeof(int))
		return -EINVAL;
	
	if (get_user(val, (int __user *)optval))
		return -EFAULT;
	
	switch (optname) {
	
	case UDPEX_TSTAMP:
		up->tstamp = (val != 0);
		if (up->tstamp) {
                        // XXX No need to enable timestamp
                        // MGRP turns it on in bhtp_init() when module is loaded
                }
		DEBUGFN("turning recv timestamps %s",
			(up->tstamp ? "ON" : "OFF"));
		break;
	
	default:
		return -ENOPROTOOPT;
	}

	return err;
}

int
udpex_getsockopt(struct sock *sk, int level, int optname,
		 char __user *optval, int __user *optlen)
{
	if (level != SOL_MGRP) {
		LOG("getsockopt: WARNING passing to IP level %d optname %d",
			level, optname);
		return ip_getsockopt(sk, level, optname, optval, optlen);
	}
	return do_udpex_getsockopt(sk, level, optname, optval, optlen);
}


int
udpex_setsockopt(struct sock *sk, int level, int optname,
		 char __user *optval, int optlen)
{
	if (level != SOL_MGRP) {
		LOG("setsockopt: WARNING passing to IP level %d optname %d",
			level, optname);
		return ip_setsockopt(sk, level, optname, optval, optlen);
	}
	return do_udpex_setsockopt(sk, level, optname, optval, optlen);
}

static int
udpex_init_sock(struct sock *sk)
{
	struct udpex_sock *up = udpex_sk(sk);
	u8 *priv = (u8 *)up + sizeof(up->inet);
	int priv_size = sizeof(*up) - sizeof(up->inet);

	DEBUG("init_sock");

	/* Cannot do memset here because inet structure is embedded at top */

	/* Do we need this memset? This is probably done in the socket code? */
	memset(priv, 0, priv_size);

	up->tun.prot = &udpex.tun_prot;

	/* queue for packets when barrier is down */
	skb_queue_head_init(&up->barrier);

	return 0;
}

static int
udpex_destroy_sock(struct sock *sk)
{
	struct udpex_sock *up = udpex_sk(sk);

	DEBUG("UDPex: destroy_sock");

	lock_sock(sk);

	/* XXX FIX cleanup ports. This is a mess. */
	if (up->src_port > 0) {
		DEBUG("releasing port %u", up->src_port);
		udpex.ports->sk[up->src_port] = NULL;
		test_and_clear_bit(up->src_port, (long *)udpex.ports->used_ports);
	}

	/* Clean up a referenced bind bucket. */
/* 	if (inet_csk(sk)->icsk_bind_hash != NULL) */
/* 		inet_put_port(&udpex.hashinfo, sk); */

	if (up->bflow != NULL) {
		int rc = bhtp_detach_flow(up->bflow, &up->tun);
		if (rc)
			DEBUG("failed to detach: rc = %d", rc);
	}

	udpex_drop_barrier_skbs(sk);

	release_sock(sk);
	return 0;
}

struct udpex_options udpex_op = {
	.probe_show = 0,
	.probe_timeout = BHTP_PROBE_TIMEOUT_NS,
	.probe_accumulated = BHTP_PROBE_ACCUMULATED,
	.debug_port_unreachable = 0,
};


struct udpex_state udpex = {

	/* socket layer interface */
	.prot = {
		.name		= "UDPex",
		.owner		= THIS_MODULE,
		.close		= udpex_close,
		.connect	= udpex_connect,
		.disconnect	= udpex_disconnect,
		.ioctl		= udpex_ioctl,
		.init		= udpex_init_sock,
		.destroy	= udpex_destroy_sock,
		.setsockopt	= udpex_setsockopt,
		.getsockopt	= udpex_getsockopt,
		.sendmsg	= udpex_sendmsg,
		.recvmsg	= udpex_recvmsg,
		.backlog_rcv	= udpex_backlog_rcv,
		.hash		= udpex_hash,
		.unhash		= udpex_unhash,
		.get_port	= udpex_get_port,
		.max_header	= MAX_UDPEX_HEADER,
		.obj_size	= sizeof(struct udpex_sock),
	},

	/* socket layer interface */
	.ops = {
		.family		= PF_INET,
		.owner		= THIS_MODULE,
		.release	= inet_release,
		.bind		= inet_bind,
		.connect	= inet_dgram_connect,
		.socketpair	= sock_no_socketpair,
		.accept		= sock_no_accept,
		.getname	= inet_getname,
		.poll		= udpex_poll,
		.ioctl		= inet_ioctl,
		.listen		= sock_no_listen,
		.shutdown	= inet_shutdown,
		.setsockopt	= sock_common_setsockopt,
		.getsockopt	= sock_common_getsockopt,
		.sendmsg	= inet_sendmsg,
		.recvmsg	= sock_common_recvmsg,
		.mmap		= sock_no_mmap,
		.sendpage	= sock_no_sendpage,
	},

	/* for socket layer registration */
	.protosw = {
		.type		= SOCK_DGRAM,
		.protocol	= IPPROTO_MGRP,
		.prot		= &udpex.prot,
		.ops		= &udpex.ops,
		.capability	= -1,
		.no_check	= 0,
		.flags		= 0,
	},

	/* for BHTP registration */
	.tun_prot = {
		/* backwards compatible to normal UDP */
		.proto		= BHTP_PROTO_UDP,
		.ops		= &udpex.tun_ops,
	},

	.tun_ops = {
		.rcv			= udpex_tun_rcv,
	},

	.options = &udpex_op,
};

int __init
udpex_init(void)
{
	int rc = 0;

	LOG("loading");

	LOG("allocating ports bytes %d", sizeof(struct udpex_ports));

	udpex.ports = vmalloc(sizeof(struct udpex_ports));
	if (udpex.ports == NULL) {
		rc = -ENOBUFS;
		goto out_error;
	}

	memset(udpex.ports, 0, sizeof(struct udpex_ports));
	udpex.ports->last_port_reserved = UDPEX_DPORT_MAX;
	LOG("allocated port hash");

	/* register with BHTP */
	rc = bhtp_register_tun_prot(&udpex.tun_prot);
	if (rc)
		goto out_free_port_hash;
	LOG("registered with BHTP");

	/* register with the socket layer */
	rc = proto_register(&udpex.prot, 1);
	if (rc)
		goto out_unregister_tun;
	LOG("registered with socket layer");

	/* XXX what does this register? */
	inet_register_protosw(&udpex.protosw);
	LOG("registered inet_protosw");

	LOG("struct udpexhdr %d bytes", sizeof(struct udpexhdr));
	LOG("loaded");

	return 0;

/* out_unregister_protosw: */
/* 	inet_unregister_protosw(&udpex.protosw); */
/* out_proto_unregister: */
/* 	proto_unregister(&udpex.prot); */
out_free_port_hash:
	kfree(udpex.ports);
out_unregister_tun:
	bhtp_unregister_tun_prot(&udpex.tun_prot);
out_error:
	ERR("loading aborted");
	return rc;
	
}

void
udpex_exit(void)
{
	LOG("unloading");

	inet_unregister_protosw(&udpex.protosw);
	LOG("de-registered inet_protosw");

	proto_unregister(&udpex.prot);
	LOG("de-registered from socket layer");

	bhtp_unregister_tun_prot(&udpex.tun_prot);
	LOG("de-registered from BHTP");

	vfree(udpex.ports);
	udpex.ports = NULL;
	LOG("de-allocated proto hash");

	LOG("unloaded");
}

