/*
 * $Id: badabing.cc,v 1.28 2005/12/07 16:24:12 jsommers Exp $
 */

/* 
 * Copyright (c) 2005  Joel Sommers.  All rights reserved.
 * 
 * This file is part of badabing, a one-way packet loss measurement tool.
 * 
 * Badabing is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Badabing is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Badabing; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "badabing.h"

#ifdef MGRP
#include "probe.h"
#include "mgrp/mgrp.h"
extern int debug_on;
#define DEBUGTS(format, ...) if (debug_on) LOGTS(format, ## __VA_ARGS__)
#else
#define DEBUGTS(format, ...) (void)(0)
#endif

void BadaBing::sender()
{
    measureSyscallOverhead();
    measureMinSleep();
    if (m_interval_duration < m_min_sleep)
    {
        std::cerr << "## WARNING: specified interval duration is finer than" << std::endl;
        std::cerr << "## WARNING: OS capability --- continuing anyway." << std::endl;
        std::cerr << "## WARNING: view all results with extreme suspicion!" << std::endl; 
    }

    if (m_verbose)
    {   
        std::cerr << "## sending to " << inet_ntoa(m_target_addr.sin_addr) << std::endl;
        std::cerr << "## p: " << m_p << std::endl;
        std::cerr << "## interval duration: " << m_interval_duration << " microseconds" << std::endl;
    }

    int sd = -1;
#ifdef MGRP
    if (m_use_mgrp) 
    {
      sd = mgrp_socket(AF_INET, SOCK_DGRAM, IPPROTO_MGRP, NULL);
    }
    else {
#endif
      sd = socket(AF_INET, SOCK_DGRAM, 0);
#ifdef MGRP
    }
#endif      
    
    if (sd == -1) {
      std::cerr << "socket call failed: " << errno << '/' << strerror(errno) << std::endl;
      throw -1;
    }

    struct sockaddr_in source;
    memset(&source, 0, sizeof(struct sockaddr_in));
    source.sin_family = AF_INET;
    source.sin_port = htons(SOURCE_PORT);
    if (0 != (bind(sd, (const struct sockaddr*)&source, sizeof(struct sockaddr_in)))) {
      std::cerr << "bind call failed: " << errno << '/' << strerror(errno) << std::endl;
      throw -1;
    }

    
    
    m_payload_size = std::max(m_payload_size, int(sizeof(WireBing)));

    struct timeval interval_start;
    struct timeval interval_length = { 0, m_interval_duration };
    gettimeofday(&interval_start, 0);
    timeradd(&interval_start, &interval_length, &interval_start);

    int current_time_slot = 0;
    int probe_sequence = 0;
    int skipped_slots = 0;    
    bool sent_previous = false;
    int num_probes = 2; // hardcoded for basic algorithm-send probes at slots
                        // i and i+1
    int current_probe = 0;

    int slots_remain = m_max_slots;
    try
    {
        while (1)
        {
            int slips = 0;
            bool send_now = false;

	    waitUntil(interval_start, slips);
            skipped_slots += slips;
            current_time_slot += (1 + slips);
            
            timeradd(&interval_start, &interval_length, &interval_start);

            if (!send_now) 
            {
                if (current_probe == 0)
                {
                    send_now = BadaBingDist::unifThresh(m_p);
                }
                else 
                {
                    // copy what we did last time.
                    send_now = sent_previous;
                }
            }
            
            if (send_now)
            {
                sendProbe(sd, probe_sequence, m_probe_size, current_time_slot, skipped_slots);
                ++probe_sequence;
                sent_previous = true;
                skipped_slots = 0;
            }
            else
            {
                sent_previous = false;
                skipped_slots++;
            }

            if (++current_probe == num_probes)
            {
                current_probe = 0;
            }

            if (--slots_remain == 0)
            {
                if (m_verbose)
                    std::cerr << "## reached max number of slots - stopping." << std::endl;
                break;
            }
        }
    }
    catch (...)
    {
        close (sd);
    }
    return;
}


void BadaBing::sendProbe(int sd, int probe_seq, int probe_size, int this_slot, int empty_slots)
{
    if (m_verbose > 1)
        std::cerr << "probe<" << probe_seq << ">probesize<" 
                  << probe_size << ">" << std::endl;

    struct timeval sendtime;
    char buffer[MAX_PAYLOAD];
    WireBing *wb = (WireBing*)&buffer[0];

    wb->m_slot = htonl(this_slot);
    wb->m_empty_slots = htonl(empty_slots);
    wb->m_probe_seq = htonl(probe_seq);
    wb->m_pkts = htonl(probe_size);

    for (int i = 0; i < probe_size; ++i)
    {
        wb->m_pkt_seq = htonl(i);
        gettimeofday(&sendtime, 0);
        wb->m_sec = htonl(sendtime.tv_sec);
        wb->m_usec = htonl(sendtime.tv_usec);

        ssize_t sbytes = 0;
	if (m_use_kernel) {
	  int order = (i 
		       ? ( (i == probe_size - 1) ? PROBE_LAST : PROBE_MIDDLE)
		       : PROBE_FIRST);
          printf("send probe_size %d i %d order 0x%x buflen %d pad %d pbk %d\n",
                 probe_size, i, order, sizeof(WireBing),
                 m_payload_size - sizeof(WireBing), PROBE_PBK_ON);
	  sbytes = probe_sendto_train_pkt(sd, 
					order, 
					0, 
					buffer, 
					sizeof(WireBing),
					&m_target_addr,
					m_payload_size - sizeof(WireBing),
					PROBE_PBK_ON
					);
	  if (sbytes == -1)
	  {
	    std::cerr << "error on send: " << errno << '/' << strerror(errno) << std::endl;
	    throw -1;
	  }
	  
	}
	else {
	  sbytes = sendto(sd, buffer, m_payload_size, 0, (const struct sockaddr*)&m_target_addr, sizeof(struct sockaddr_in));
	  if (sbytes != m_payload_size)
	  {
	    std::cerr << "error on send: " << errno << '/' << strerror(errno) << std::endl;
	    throw -1;
	  }
	}
	std::cerr << "send: " << sbytes << " " << errno << '/' << strerror(errno) << std::endl;

    }
}


void BadaBing::waitUntil(struct timeval &interval_start, int &slippage)
{
    slippage = 0; // keep track of time slots we've slipped because of OS
    
    struct timeval now, diff;
    gettimeofday(&now, 0);
    timersub(&interval_start, &now, &diff);
    int usec_remain = (diff.tv_sec * 1000000 + diff.tv_usec);
 
    if (usec_remain < -m_interval_duration)
    {
        slippage = -(usec_remain / m_interval_duration);
        int newsleep = abs(usec_remain % m_interval_duration);
        struct timeval remainder = {0, newsleep};
        timeradd(&now, &remainder, &interval_start);
        std::cerr << "## " << now.tv_sec << '.' << std::setw(6) 
                  << std::setfill('0') << now.tv_usec
                  << " WARNING: time interval slipped by " << slippage 
                  << " time slot(s) (" << -usec_remain << " usec).  "
                  << " sleeping an additional " << newsleep << " to resynch."
                  << std::endl;
    }

    // sleep
    timersub(&interval_start, &now, &diff);
    usec_remain = (diff.tv_sec * 1000000 + diff.tv_usec);

    int sleepy = usec_remain - m_min_sleep;
    if (sleepy > 0)
        usleep(sleepy);

    while (1)
    {
        gettimeofday(&now, 0);
        timersub(&interval_start, &now, &diff);
        usec_remain = (diff.tv_sec * 1000000 + diff.tv_usec);
        if (usec_remain < m_syscall_overhead)
            break;
    }
}
      

void BadaBing::measureSyscallOverhead()
{
    struct timeval tvli[BINGOSTIMINGSAMPLES];
    for (int i = 0; i < BINGOSTIMINGSAMPLES; ++i)
        gettimeofday(&tvli[i], 0);

    double diffsum;
    struct timeval diff;
    std::list<double> diffli;
    for (int i = 1; i < BINGOSTIMINGSAMPLES; ++i)
    {
        timersub(&tvli[i], &tvli[i-1], &diff);
        diffsum += diff.tv_sec * 1000000.0 + double(diff.tv_usec);
        diffli.push_back(diff.tv_sec * 1000000.0 + double(diff.tv_usec));
    }
    diffli.sort();

    std::list<double>::iterator it = diffli.begin();
    for (int i = 0; i < BINGOSTIMINGSAMPLES/2; ++i, ++it) ;
    double median = *it;

    double sco = diffsum / (BINGOSTIMINGSAMPLES - 1);
    if (m_verbose)
    {
        std::cout << "##syscall overhead mean: " << sco << " microseconds" << std::endl;
        std::cout << "##syscall overhead median: " << median << " microseconds" << std::endl;
    }
    m_syscall_overhead = int(sco);
}


void BadaBing::measureMinSleep()
{   
    struct timeval ts[BINGOSTIMINGSAMPLES];
    gettimeofday(&ts[0], 0);
    for (int i = 1; i < BINGOSTIMINGSAMPLES; ++i)
    {
        usleep(1);
        gettimeofday(&ts[i], 0);
    }
    double usecsum = 0.0;
    double usecsumsq = 0.0;
    int imax = 0;
    struct timeval diff;
    std::list<double> diffli;
    for (int i = 1; i < BINGOSTIMINGSAMPLES; ++i)
    {
        timersub(&ts[i], &ts[i-1], &diff);
        double usecs = diff.tv_sec * 1000000.0 + double(diff.tv_usec);
        usecsum += usecs;
        usecsumsq += pow(usecs, 2.0);
        imax = std::max(imax, int(diff.tv_sec * 1000000 + diff.tv_usec));
        diffli.push_back(diff.tv_sec * 1000000.0 + double(diff.tv_usec));
    }
    diffli.sort();
    std::list<double>::iterator it = diffli.begin();
    for (int i = 0; i < BINGOSTIMINGSAMPLES/2; ++i, ++it) ;
    double median = *it;
    double mean = usecsum / double(BINGOSTIMINGSAMPLES - 1);
    double stdev = sqrt((usecsumsq * (BINGOSTIMINGSAMPLES - 1) - pow(usecsum,2.0)) / (double(BINGOSTIMINGSAMPLES - 2) * double(BINGOSTIMINGSAMPLES - 1)));
     
    
    // m_min_sleep is the minimun amount of time (usecs) that we'll attempt
    // to sleep.  otherwise, we spin-wait.
    m_min_sleep = std::max(0, int(mean + (3 * stdev)));
    
    if (m_verbose)
    {
        std::cout << "##mean sleep: " << mean << " microseconds" << std::endl;
        std::cout << "##stdev sleep: " << stdev << " microseconds" << std::endl;
        std::cout << "##median sleep: " << median << " microseconds" << std::endl;
        std::cout << "##max sleep: " << imax << " microseconds" << std::endl;
    }
}   

    
void BadaBing::receiver()
{
    if (m_verbose)
        std::cerr << "## receiving from anyone." << std::endl;

    int sd = -1;
#ifdef MGRP
    if (m_use_mgrp) 
    {
      sd = mgrp_socket(AF_INET, SOCK_DGRAM, IPPROTO_MGRP, NULL);
    }
    else {
#endif
      sd = socket(AF_INET, SOCK_DGRAM, 0);
#ifdef MGRP
    }

    if (sd == -1) {
      std::cerr << "socket call failed: " << errno << '/' << strerror(errno) << std::endl;
    }

    int on = 0;
    int rv = -1;
    if (m_use_kernel) {
      // printf("** Using MGRP timestamps\n");
      on = 1;
      rv = setsockopt(sd, SOL_MGRP, UDPEX_TSTAMP, &on, sizeof(on));
      if (-1 == rv) {
	std::cerr << "error in setsockopt(SOL_MGRP, UDPEX_TSTAMP): " << errno << '/' << strerror(errno) << std::endl;
	throw -1;
      }
    } else if (m_use_so_timestamp) {
      // printf("** Using SO_TIMESTAMP timestamps\n");
      on = 1;
      rv = setsockopt(sd, SOL_SOCKET, SO_TIMESTAMP, &on, sizeof(on));
      if (-1 == rv) {
	std::cerr << "error in setsockopt(SOL_SOCKET, SO_TIMESTAMP): " << errno << '/' << strerror(errno) << std::endl;
	throw -1;
      }
    } else {
      //printf("** Using application timestamps\n");
    }

#endif      

    struct sockaddr_in sin;
    memset(&sin, 0, sizeof(struct sockaddr_in));
    sin.sin_family = AF_INET;
    sin.sin_port = htons(DEST_PORT);
    if (0 != (bind(sd, (const struct sockaddr*)&sin, sizeof(struct sockaddr_in)))) {
      std::cerr << "bind call failed: " << errno << '/' << strerror(errno) << std::endl;
      throw -1;
    }
    pollfd pfd[1]; 

    char buffer[MAX_PAYLOAD];
    WireBing wb;

    unsigned int prev_seq = 0xffffffff;
    try
    {
        while (1)
        {
            pfd[0].fd = sd;
            pfd[0].events = POLLIN;
            pfd[0].revents = 0;

            int np = 1;
            int rv = poll(&pfd[0], np, 0);
            if (rv == -1)
            {
                std::cerr << "error in poll(): " << errno << '/' << strerror(errno) << std::endl;
                throw -1;
            }
            else if (rv > 0)
            {
                if (pfd[0].revents & POLLIN)
                {
		  /*** see tagent/tools/pathload/pathload_recv_func.c:bhtprecv() ***/
		    ssize_t rbytes;

#ifdef MGRP
		    struct bhtp_timestamps bts;
		    if (m_use_kernel) {
			rbytes = probe_recv_train_pkt(sd, buffer, MAX_PAYLOAD, &bts);
		    } 
		    else { // TODO: so_timestamp only (blah)
#endif
			// activity on
			SOCKLEN_T fromlen = sizeof(struct sockaddr_in);
			rbytes = recvfrom(sd, buffer, MAX_PAYLOAD, 0, (struct sockaddr*)&sin, &fromlen);
#ifdef MGRP
		    }
#endif
                    if (rbytes < int(sizeof(WireBing)))
                    {
                        std::cerr << "error on receive - incomplete probe (" << errno << '/' << strerror(errno) << ")" << std::endl;
                        throw -1;
                    }

                    memcpy(&wb, &buffer[0], sizeof(WireBing));

                    struct timeval tv;
                    gettimeofday(&tv, 0);

                    struct timeval tv2;
#ifdef MGRP
		    if (m_use_kernel) {
			bhtp_tstamp_to_timeval(&bts.rcv, &tv);
			bhtp_tstamp_to_timeval(&bts.snd, &tv2);
		    } // todo: else if (m_use_so_timestamp) {
		    else {
#endif
			tv2.tv_sec = ntohl(wb.m_sec);
			tv2.tv_usec = ntohl(wb.m_usec);
#ifdef MGRP
		    }
#endif
                    /*
                     * really simple output format.  we don't do any processing
                     * in the receiver - just print stuff out.  all estimation
                     * is done in an external script.
                     */
                    std::cout << tv.tv_sec << '.' 
                              << std::setw(6) << std::setfill('0') << tv.tv_usec 
                              << ' ' << tv2.tv_sec << '.' << std::setw(6) 
                              << std::setfill('0') << tv2.tv_usec << ' ' 
                              << htonl(wb.m_slot) << ' ' << htonl(wb.m_empty_slots)
                              << ' ' << ntohl(wb.m_probe_seq) << ' ' 
                              << ntohl(wb.m_pkt_seq) << ' ' << ntohl(wb.m_pkts) 
                              << ' ' << rbytes << std::endl;

                    unsigned int seq = ntohl(wb.m_probe_seq) * ntohl(wb.m_pkts) +  ntohl(wb.m_pkt_seq);

                    if (m_verbose > 1)
                        std::cerr << "## seq,prevseq: " << seq << "," << prev_seq << std::endl;
                    prev_seq = seq;
                }
            }
        }
    } 
    catch (...)
    {
        close (sd);
    }
    return;
}

