/*
 * $Id: badabing.h,v 1.23 2005/12/07 16:24:12 jsommers Exp $
 */

/* 
 * Copyright (c) 2005  Joel Sommers.  All rights reserved.
 * 
 * This file is part of badabing, a one-way packet loss measurement tool.
 * 
 * Badabing is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Badabing is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Badabing; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef __BADABING_H__
#define __BADABING_H__

#include <iostream>
#include <iomanip>
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <poll.h>
#include <unistd.h>
#include <errno.h>
#include <string>
#include <list>
#include <math.h>
#include "config.h"


static const unsigned short SOURCE_PORT = 13887;
static const unsigned short DEST_PORT = 13889;
static const int MAX_PAYLOAD = 4096;
static const int BINGOSTIMINGSAMPLES = 100;


struct WireBing
{
    unsigned int m_slot;         //<! time slot in which this probe was sent
    unsigned int m_empty_slots;  //<! empty time slots preceeding this probe
    unsigned int m_probe_seq;    //<! sequence of this probe
    unsigned int m_pkts;         //<! number of packets within this probe
    unsigned int m_pkt_seq;      //<! sequence of this packet within the probe
    unsigned int m_sec;          //<! time this probe was sent, seconds
    unsigned int m_usec;         //<! time this probe was sent, micros
};


class BadaBing
{
public:
 BadaBing() : m_p(0.3), m_max_slots(0), m_payload_size(600), m_verbose(0), m_syscall_overhead(0), m_min_sleep(0), m_clock_tick(0), m_interval_duration(5000), m_probe_size(3)
#ifdef MGRP
        , m_use_mgrp(false), m_use_kernel(false), m_use_so_timestamp(false)
#endif
    {
        memset(&m_target_addr, 0, sizeof(struct sockaddr_in));
    }
    virtual ~BadaBing() {}

    void sender();
    void receiver();

    void setTarget(const char *dstaddr)
        {
            memset(&m_target_addr, 0, sizeof(struct sockaddr_in));
            m_target_addr.sin_family = AF_INET;
            m_target_addr.sin_port = htons(DEST_PORT);

            struct hostent *hp = gethostbyname2(dstaddr, AF_INET);
            if (hp) 
            {
                memcpy(&m_target_addr.sin_addr, hp->h_addr_list[0], hp->h_length);
            }
        }
    bool checkTarget()
        {
            if (m_target_addr.sin_addr.s_addr == 0 || m_target_addr.sin_port == 0)
                return false; 
            return true;
        }
    void setP(double &d) { m_p = d; }
    void setMaxSlots(int &i) { m_max_slots = i; }
    void setPayloadSize(int &i)  { m_payload_size = i; }
    void setVerbosity(int &i)  { m_verbose = i; }
    void setIntervalDuration(int &i) { m_interval_duration = i; }
    void setProbeSize(int &i) { m_probe_size = i; }
    void setPort(unsigned short &s) { m_target_addr.sin_port = htons(s); }
#ifdef MGRP
    void setUseMGRP(bool &u) { m_use_mgrp = u; }
    void setUseKernel(bool &u) { m_use_kernel = u; }
    void setUseSocketTimestamp(bool &u) { m_use_so_timestamp = u; }
#endif
    
private:
    void sendProbe(int, int, int, int, int);
    void waitUntil(struct timeval &, int &);
    void measureMinSleep();
    void measureSyscallOverhead();

    double m_p;
    int m_max_slots;

    struct sockaddr_in m_target_addr;
    int m_payload_size;
    int m_verbose;

    int m_syscall_overhead;
    int m_min_sleep;
    int m_clock_tick;

    int m_interval_duration;
    int m_probe_size;

#ifdef MGRP
    bool m_use_mgrp;
    bool m_use_kernel;
    bool m_use_so_timestamp;
#endif
   
};


/* 
 * a static utility class for generating deviates
 */
class BadaBingDist
{
public:
    static void init(unsigned long iseed = 0xffffffff)
        {
            if (iseed == 0xffffffff)
#if (__linux || __sun)
		srandom(time(NULL) % getpid());
#else
                srandomdev(); // freebsd, macos x
#endif
            else
                srandom(iseed);
        }

    static double expo(double mean)
        {
            return (-1.0 * mean * log1p(rzeroone()));
        }

    static int geom(int mean)
        {
            if (mean == 0 || mean == 1)
                return 1;
            double p = -1.0 / double(mean);
            return int((log1p(rzeroone()) / log1p(p)));
        }

        // return true/false, depending whether uniform rand [0,1]
        // is less than thresh
    static bool unifThresh(double thresh)
        {
            double r = rzeroone() * -1.0;
            return (r <= thresh);
        }

private:
    // disallow construction or copy construction
    BadaBingDist() {}
    BadaBingDist(const BadaBingDist &) {}

    // return random floating point number on interval [-1,0)
    static double rzeroone()
        {
            double rv = 0.0;
            do
            {
                rv = double(random()) / double(RAND_MAX);
            } while  (rv == 0.0);
            return (-1.0 * rv);
        }
};

#endif // __BADABING_H__
