/*
 * $Id: main.cc,v 1.20 2005/12/07 16:24:12 jsommers Exp $
 */

/* 
 * Copyright (c) 2005  Joel Sommers.  All rights reserved.
 * 
 * This file is part of badabing, a one-way packet loss measurement tool.
 * 
 * Badabing is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Badabing is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Badabing; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "badabing.h"

void usage(const char *proggie)
{
    std::cerr << "usage: " << proggie << std::endl;
    std::cerr << "both: " << std::endl;
    std::cerr << "  -v increase verbosity" << std::endl;
#ifdef MGRP
    std::cerr << "  -x use MGRP transport (no kernel probes unless -k also)" << std::endl;
    std::cerr << "  -k use kernel probing (implies -x)" << std::endl;
    std::cerr << "  -t use socket timestamps (SO_TIMESTAMP)" << std::endl;
#endif
    std::cerr << "  -P <dst port (" << DEST_PORT << ")>" << std::endl;
    std::cerr << "receiver: " << std::endl;
    std::cerr << "  -R " << std::endl;
    std::cerr << "sender: " << std::endl;
    std::cerr << "  -N <int> max number of time slots (default: run continuously)" << std::endl;
    std::cerr << "  -S <dst addr (127.0.0.1)> " << std::endl;
    std::cerr << "  -l <udp payload size (600 bytes)>" << std::endl;
    std::cerr << "  -p <p value (0.3)>" << std::endl;
    std::cerr << "  -s <rng seed>" << std::endl;

    std::cerr << "  -i <interval duration (5000 microsec)>" << std::endl;
    std::cerr << "  -b <probe size (3 pkts)>" << std::endl << std::endl;
    std::cerr << "NB: one of -R or -S must be set." << std::endl;
}


int main(int argc, char **argv)
{
    int c;
    unsigned long seed = 0xffffffff;

    std::string dstaddr = "127.0.0.1";
    bool sender = false;
    bool receiver = false;

#ifdef MGRP
    bool use_mgrp = false;
    bool use_kernel = false;
    bool use_so_timestamps = false;
#endif

    double p = 0.3;
    int payload_size = 600;
    int verbose = 0;

    int probe_size = 3;
    int interval_duration = 5000;

    int max_slots = 0;
    unsigned short port = DEST_PORT;

    const char *OPTIONS_STRING = "b:i:l:N:p:P:Rs:S:v"
#ifdef MGRP // let CPP do the auto-appending
      "xkt"      
#endif
      ;

    while ((c = getopt(argc, argv, OPTIONS_STRING)) != EOF)
    {
        switch(c)
        {
        case 'b':
            probe_size = atoi(optarg);
            break;
        case 'i':
            interval_duration = atoi(optarg);
            break;
        case 'l':
            payload_size = atoi(optarg);
            break;
#ifdef MGRP
	case 'x':
            use_mgrp = true;
            break;
	case 'k':
            use_kernel = true;
	    use_mgrp = true;
            break;
	case 't':
            use_so_timestamps = true;
            break;
#endif
        case 'N':
            max_slots = atoi(optarg);
            break;
        case 'p':
            p = atof(optarg);
            break;
        case 'P':
            port = atoi(optarg);
            break;
        case 'R':
            receiver = true;
            sender = false;
            break;
        case 'S':
            dstaddr = optarg;
            sender = true;
            receiver = false;
            break;
        case 's':
            seed = strtoul(optarg, NULL, 10);
            break;
        case 'v':
            verbose++;
            break;
        default:
            usage(argv[0]);
            exit (-1);
        }
    }

    BadaBingDist::init(seed);

    BadaBing bb;
    bb.setTarget(dstaddr.c_str());
    bb.setPort(port);

    bb.setP(p);
    bb.setMaxSlots(max_slots);
    bb.setPayloadSize(payload_size);
    bb.setVerbosity(verbose);
    bb.setProbeSize(probe_size);
    bb.setIntervalDuration(interval_duration);
#ifdef MGRP
    bb.setUseMGRP(use_mgrp);
    bb.setUseSocketTimestamp(use_so_timestamps);
    bb.setUseKernel(use_kernel);
#endif

    struct timeval starttime;
    gettimeofday(&starttime, 0);
    std::cerr << "## badabing started at " << starttime.tv_sec << '.'
              << std::setw(6) << std::setfill('0') << starttime.tv_usec
              << std::endl;

    if (sender)
    {
        if (!bb.checkTarget())
        {
            std::cerr << "invalid target address." << std::endl;
            return 0;
        }
        if (p <= 0.0 || p > 1.0)
        {
            std::cerr << "invalid setting of p (must be (0,1])" << std::endl;
            return 0;
        }
        if (payload_size > MAX_PAYLOAD)
        {
            std::cerr << "payload too large.  max payload is " << MAX_PAYLOAD << std::endl;
            return 0;
        }
        
        bb.sender();
    }
    else if (receiver)
    {
        bb.receiver();
    }
    else
        std::cerr << "Must use -R or -S to specify as receiver or sender." << std::endl;

    exit(0);
}

