#!/usr/bin/env python

#
# $Id: proc_badabing.py,v 1.7 2005/12/07 16:24:12 jsommers Exp $
#

#
# Copyright (c) 2005  Joel Sommers.  All rights reserved.
#
# This file is part of badabing, a one-way packet loss measurement tool.
#
# Badabing is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Badabing is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Badabing; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#


"""
this script is for processing output from badabing.  
"""

__author__ = "joel sommers"
__date__ = "january 2005"
__version__ = "$Revision: 1.7 $"

# "constants"
Lost = 1
NotLost = 0

import sys,string,math,optparse,gzip,sre,time

parser = optparse.OptionParser(usage="usage: %prog [options]",
       version="%prog $Revision: 1.7 $")
parser.add_option("-a", "--alpha",
                  help="specify alpha value [default: 0.005]",
                  default=0.005, type="float", dest="alpha")
parser.add_option("-t", "--tau",
                  help="specify tau value [default: 0.05 sec]",
                  default=0.05, type="float", dest="tau")
parser.add_option("-d", "--debug",
                  help="increase debugging verbosity [0=no debug; default: 0]",
                  default=0, action="count", dest="debug")
parser.add_option("-N", "--numslots",
                  help="maximum number of discrete time slots to process [default: no maximum]",
                  default=-1, type="int", dest="max_slot")
parser.add_option("-i", "--interval",
                  help="interval length for time discretization [default: 0.005 sec.]",
                  default=0.005, type="float", dest="interval")
parser.add_option("-o", "--outputperiod", metavar="OUTPUTPERIOD",
                  help="print periodic results every OUTPUTPERIOD sec. [default: results only at end]",
                  default=0, type="float", dest="results_interval")

(opts, args) = parser.parse_args()


isgzip = False
if not len(args):
    if opts.debug:
        print >>sys.stderr,"no input file specified --- using stdin."
    infile = 'stdin'
else:
    infile = args[0]
    if len(args) > 1:
        print >>sys.stderr,"#looks like multiple input files were specified."
        print >>sys.stderr,"#only looking at the first one."
    isgzip = (sre.search("\.gz$",infile) != None)
    
# print parameters prior to starting
print "# %s started at %s" % (sys.argv[0], time.asctime())
print "# infile <%s> alpha<%3.3f> tau<%3.3f> maxslot<%d> interval<%3.3f>" % (infile, opts.alpha, opts.tau, opts.max_slot, opts.interval)


################################################

def median(li):
    """
    calculate the median value of a list.  
    used in calculating a threshold OWD.
    """
    xli = li
    xli.sort()
    rv = 0.0
    p = len(xli) / 2
    if len(xli) % 2 == 0:
        rv = (xli[p-1] + xli[p]) / 2.0
    else:
        rv = xli[p]
    return (rv)


def add_packet(probe_dict, recv, send, slot_num, empty_slots, probe_id, pkt_id, probe_size):
    """
    incorporate the most recently seen packet into the probe dictionary/hash:
       - probe id (dictionary key)
       - keep OWD for each packet in a probe
       - number of packets in the probe
       - number of empty discrete time slots between this probe and previous
       - min send time of packets in this probe, and max recv time 
         of packets in this probe (time bounds on when packets were sent/recvd)
    """
    global opts

    # NB: synchronized timestamps are assumed...
    owd = round(recv - send, 6)

    # find record of probe, if it exists
    if probe_dict.has_key(probe_id):
        # update contents of existing probe record
        probe_rec = probe_dict[probe_id]
        probe_rec['send'] = min(send, probe_rec['send'])
        probe_rec['recv'] = max(recv, probe_rec['recv'])
        # some simple sanity checks...
        assert (probe_rec['empty'] == empty_slots)
        assert (probe_rec['psize'] == probe_size)
        pr_dict = probe_rec['pkts']
        if pr_dict.has_key(pkt_id):
            # duplicate packet - just print a warning for now.
            # note that we also don't detect out-of-order packets presently.
            print "# warning: duplicate packet detected (probe id %d, pkt id %d, send time %6.6f)" % (probe_id, pkt_id, send)
        pr_dict[pkt_id] = owd
    else:
        # create new probe record
        probe_dict[probe_id] = { 'recv':recv, 'send':send, 'empty':empty_slots, 'psize':probe_size, 'pkts':{ pkt_id:owd } }
    if opts.debug > 2:
        print "updated id %d: %s" % (probe_id, str(probe_dict[probe_id]))
    return
    

def check_thresholds(probe_rec, probe_id, max_owd_est, last_loss):
    """
    check probe_rec against tau and alpha thresholds to decide whether
    to mark probe as lost or not.  return True if probe should be marked
    lost, False otherwise.
    """
    global opts
    
    # check whether this probe is within tau of last lost probe
    # or if owd > (1-alpha)*maxowd
    # if not, then set last_loss to -1

    recv_ts = probe_rec['recv']
    pkt_dict = probe_rec['pkts']
    probe_owd = max(pkt_dict.values())

    real_loss = (not probe_rec.has_key('psize')) or (probe_rec['psize'] != len(pkt_dict)) or (probe_rec['status'] == Lost)

    tau_hit = abs(last_loss - recv_ts) <= opts.tau
    alpha_hit = probe_owd >= max_owd_est * (1.0 - opts.alpha)
    is_lost = real_loss or (tau_hit or alpha_hit) 

    if opts.debug > 2:
        print "# check thresh: id %d  probe_owd: %3.6f  recv_ts: %6.6f  tau: %d  alpha: %d  lost: %d" % (probe_id, probe_owd, recv_ts, tau_hit, alpha_hit, is_lost)

    if opts.debug and is_lost:
        print "# marked probe %d lost because of alpha<%d> or tau<%d>" % (probe_id, alpha_hit, tau_hit)

    return is_lost


def update_owd(owd_list, probe_dict, probe_id):
    """
    NB: probe_id may not exist --- use the most recent successful probe.
    (if that doesn't work,  use an arbitrarily large value, e.g., 1 second.)
    """
    global opts
    
    tmp_id = probe_id
    done = False
    while tmp_id > 0 and not done:
        if probe_dict.has_key(tmp_id):
            probe_rec = probe_dict[tmp_id]
            if probe_rec.has_key('pkts'):
                pkt_dict = probe_rec['pkts']
                for (pnum,powd) in pkt_dict.items():
                    owd_list.append(powd)
                done = True
        tmp_id -= 1

    if not done:
        owd_list.append(1.0)

    if opts.debug:
        print "# new max OWD estimate: %3.6f" % (median(owd_list))
        if opts.debug > 1:
            print "# full OWD list: ",owd_list
 
 
def check_recent_probes(probe_dict, first_probe, probe_id, owd_est, ts):
    """
    a packet within a probe was lost (with probe id probe_id).  look at
    probes in recent past to determine whether we should mark any of them
    as lost (i.e., whether they're within tau seconds of ts, or owd for
    any packet is within (1-alpha)*owd_est).
    """
    global opts
    tmp_id = probe_id
    done = False
    if opts.debug:
        print "# check recent: first: %d  id: %d  owd_est: %3.6f  ts: %6.6f" % (first_probe, probe_id, owd_est, ts)
    while not done and tmp_id >= first_probe:
        probe_rec = probe_dict[tmp_id]

        # already hit some lost probe in past---stop
        if probe_rec['status'] == Lost:
            break
            
        if check_thresholds(probe_rec, tmp_id, owd_est, ts):
            probe_rec['status'] = Lost
        else:
            done = True
        tmp_id -= 1


def mark_lost(probe_dict, stat_dict, first_probe, curr_probe, last_probe):
    """
    go through probe dictionary, marking probes as lost or not.
    
    we start with curr_probe and go to last_probe, making sure that we just
    look at pairs of probes (i.e., only "basic experiments").  return next
    probe to start with (which must be the 1st probe of a pair).  

    for each probe, we add a Lost/NotLost marker ('status').  

    for each lost probe (probe with any lost packet):
      - mark any previous probes as lost if falling within tau or 
        alpha thresholds
      - mark successive probes as lost if falling within tau or alpha 
        thresholds
      - max OWD should be set as median OWD for other (non-lost) packets of the
        probe where loss happened.

    once we go through probes, count 00,01,10,11 occurrences and calculate
    frequency and duration estimates.  also report p1 and p2 as a 
    self-validation/calibration measure (we want these values to be very close
    in order to trust the results).

    NB: successive calls of mark_lost() could cause previously checked probes 
    (toward end of previous value of 'last_probe') to be marked as Lost.  
    therefore, we just count y_i occurrences each time around --- it does not
    cost much. 
    """
    global opts
    owd_list = []
    last_loss = -1

    new_curr = curr_probe
    assert (new_curr % 2 == 0)
    
    if last_probe % 2 == 1: 
        last_probe -=  1
        assert (last_probe > curr_probe)

    # fixme: last_probe should not be at very end if we're doing a mid-expt
    # stat dump (may not have received all pkts in probe)

    if opts.debug:
        print "#mark lost- cycling through ids %d -> %d" % (new_curr, last_probe)
    while new_curr <= last_probe:
        if probe_dict.has_key(new_curr):
            # we found a record for probe new_curr
            
            pr_dict = probe_dict[new_curr]
            if not pr_dict.has_key('recv'):
                new_curr += 1
                continue

            ts = pr_dict['recv']

            if opts.debug > 2:
                print "#",new_curr,pr_dict
            
            pkt_dict = pr_dict['pkts']
            if pr_dict['psize'] != len(pkt_dict):
                if opts.debug:
                    print "# packet lost in probe %d (%d out of %d received)" % (new_curr, len(pkt_dict), pr_dict['psize'])
                update_owd(owd_list, probe_dict, new_curr)
                pr_dict['status'] = Lost
                last_loss = pr_dict['recv']
                check_recent_probes(probe_dict, first_probe, (new_curr - 1), median(owd_list), ts)
                 
            else:
                pr_dict['status'] = NotLost
                
                # we didn't lose any packets in this probe, but if there
                # was loss within last tau seconds, we mark this probe as
                # lost, or if owd is larger than (1-alpha)*median(owd_list).
                # if we aren't within either tau or alpha thresholds, then
                # set "last_loss" to -1, indicating that we're no longer in
                # a congestion period.
                if last_loss > 0:
                    if check_thresholds(pr_dict, new_curr, median(owd_list), last_loss):
                        pr_dict['status'] = Lost
                    else:
                        last_loss = -1
                        owd_list = []
            
        else:
            # there's no record to be found for probe new_curr
            
            # update owd estimate from most recent successful probe
            update_owd(owd_list, probe_dict, new_curr)

            if opts.debug:
                print "# probe %d was totally lost" % (new_curr)

            probe_dict[new_curr] = { 'status':Lost }

            # also, update last_loss based on next successful packet 
            # (recv time - interval * empty_slots)
            tmp_id = new_curr + 1
            while tmp_id <= last_probe:
                if probe_dict.has_key(tmp_id):
                    tmp_pdict = probe_dict[tmp_id]
                    next_recv = tmp_pdict['recv']
                    mt_slots = tmp_pdict['empty']
                    last_loss = next_recv - mt_slots * opts.interval
                    tmp_pdict = probe_dict[new_curr]
                    break
                tmp_id += 1

            # now, search backward, possibly marking some previous probes
            # as lost depending on alpha and tau.
            check_recent_probes(probe_dict, first_probe, (new_curr - 1), median(owd_list), last_loss)
            
        new_curr += 1
    return (new_curr - 1)


def print_estimates(probe_dict, first_probe, last_probe):
    """
    """
    global opts

    if last_probe % 2 == 1: 
        last_probe -=  1
        assert (last_probe > first_probe)
    assert (first_probe % 2 == 0)
    
    M = 0
    oneone = 0
    zeroone = 0
    onezero = 0
    
    for id in xrange(first_probe, last_probe, 2):
        probe_rec1 = probe_dict[id]
        probe_rec2 = probe_dict[id+1]
        s1 = probe_rec1['status']
        s2 = probe_rec2['status']
        if s1 == 0 and s2 == 1:
            zeroone += 1
        elif s1 == 1 and s2 == 0:
            onezero += 1
        elif s1 == 1 and s2 == 1:
            oneone += 1
        M += 1

        if opts.debug > 1:
            print "# id %d -> %d%d" % (id, s1, s2)
            print "#     ",probe_rec1
            print "#     ",probe_rec2

    print "# M:%d  01:%d  10:%d  11:%d" % (M, zeroone, onezero, oneone)
    assert (M > 0)
    
    freq_est = float(onezero + oneone) / float(M)

    R = oneone + onezero + zeroone
    S = onezero + zeroone
    r = 0.0
    dur_est = 0.0
    if S > 0:
        dur_est = 2.0 * float(R) / float(S) - 1.0
    dur_est_time = dur_est * opts.interval
    if oneone > 0:
        r = float(zeroone + onezero) / float(oneone)

    print "freq_est %3.6f" % (freq_est)
    print "dur_est(time)  %3.6f (dur_est(slots): %3.6f  R:%d  S:%d  r:%3.6f" % (dur_est_time, dur_est, R, S, r)
    if opts.debug:
        print "# for consistent duration est, 01 and 10 occurrences should be close..."
        print "# for consistent freq est, r should 'reasonably' close to 1.0..."
    
################################################

start_time = -1.0
first_probe = -1
curr_probe = -1
last_recv = -1
last_stat = -1
last_probe = -1
next_results = -1
probe_dict = dict()
stat_dict = dict()

if infile == 'stdin':
    inf = sys.stdin
else:    
    if isgzip:
        inf = gzip.GzipFile(infile)
    else:
        inf = file(infile)

while 1:
    line = inf.readline()
    if not line:
        break
    line.strip()
    fields = line.split()

    recv = float(fields[0])
    send = float(fields[1])
    slot_num = int(fields[2])
    empty_slots = int(fields[3])
    probe_id = int(fields[4])
    pkt_id = int(fields[5])
    probe_size = int(fields[6])
    pkt_size = int(fields[7])

    if start_time < 0:
        if probe_id % 2 == 1:
            # skip this probe if it isn't the start of a basic experiment
            continue

        start_time = send
        first_probe = probe_id
        curr_probe = first_probe
        last_stat = start_time
        if opts.results_interval > 0:
            next_results = start_time + opts.results_interval

    add_packet(probe_dict, recv, send, slot_num, empty_slots, probe_id, pkt_id, probe_size)
    last_recv = recv
    last_probe = max(last_probe, probe_id)
    if opts.results_interval > 0 and last_recv >= next_results:
        print "# %s periodic stats at %s: first_probe<%d> last_probe<%d> last_recv<%s>" % (sys.argv[0], time.asctime(), curr_probe, last_probe, float(last_recv))
        curr_probe = mark_lost(probe_dict, stat_dict, first_probe, curr_probe, last_probe)
        print_estimates(probe_dict, first_probe, last_probe)
        next_results += opts.results_interval

    if opts.max_slot > 0 and slot_num >= opts.max_slot:
        break
 
if not (inf == sys.stdin):
    inf.close()

print "# *** final stats ***"
curr_probe = mark_lost(probe_dict, None, first_probe, first_probe, last_probe)
print_estimates(probe_dict, first_probe, last_probe)
print "# %s done at %s" % (sys.argv[0], time.asctime())

