/*************************************************************************

 linkedlist.t
 Copyright (C) 2006, Norbert Vegh
 Norbert Vegh, ntools@norvegh.com

 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 2 of the License, or (at your
 option) any later version.

*************************************************************************/

#ifndef LINKEDLIST_T
#define LINKEDLIST_T

#include <iostream>

using namespace std;

/*
 This template class implements a linked list. The data list elements
 contain only pointers to user data, and not data itself, for efficiency.
 It means that before each insert, you have to create a new data element
 with new operator . The class supports multiple cursors to make it thread safe.
 The default is one cursor ( number 0 ), but you can create arbitrary number
 of cursors with initCursor(). Each thread must use a distinct cursor
 for manipulating the list. When you insert something into the list,
 the cursor is set to the new element. If the cursor is 0 when calling
 an insert function, it is first set to the last entry.
 endOfList() returns true when the cursor is 0 (points after the last
 element). 

 NOTE ABOUT THE INSERT FUNCTIONS:
 If you want to use this function in multi-threaded application,
 and other threads may have access to the list during the process,
 then all cursor functions must be protected with mutex or read-write lock.

 NOTE ABOUT THE REMOVE FUNCTION:
 If you want to use this function in multi-threaded application,
 and other threads may have access to the list during the process,
 then all cursor functions must be protected with mutex or read-write
 lock. This function removes an entry that can be pointed by other
 cursors. Therefore at the end of the function we check all cursors,
 and update the ones pointing to the removed entry. It is possible
 however, that before the update an other thread tries to use a cursor
 which points to the removed entry, which can cause the program to crash.

 Furthermore, keep in mind, that the data elements are not protected either,
 more threads can access the same entry, and get the pointer to the same
 data element.

*/

template <class T>
class LinkedList
{
 public:
	LinkedList();
	~LinkedList();
	void initCursors( int num );  // create the given number of cursors
	void first( int curs );  // set cursor to the first entry
	void last( int curs );  // set cursor to the last entry
	void next( int curs );  // set cursor to the next entry
	T *getFirst( int curs );  // get the first entry
	T *getNext( int curs );  // get the next entry
	T *getEntry( int curs );  // get the actual entry at cursor
	int isLast( int curs );  // returns TRUE if the cursor is on the last entry
	int endOfList( int curs );  // returns TRUE if the cursor points to no entry
	void insertBefore( T *,  int curs );  // insert an enrty before the cursor
	void insertAfter( T *, int curs );  // insert an enrty after the cursor
	void remove( int curs);  // remove the entry at the cursor
	int empty();  // returns TRUE if the list is empty
	int size();  // returns the size of the list
	void dump( int curs );  // print the content of the list
	
 private:
	struct ListEntry
	{
		T *data;
		struct ListEntry *prev;
		struct ListEntry *next;
	} *head, *tail, **cursortb;
	int cursors;
	int lsize;  // list size
};

template <class T>
LinkedList<T>::LinkedList()
{
	lsize = 0;
	head = tail = 0;
	cursortb = new ListEntry*;
	memset( cursortb, 0, sizeof( ListEntry ** ) );
	cursors = 1;
}

template <class T>
LinkedList<T>::~LinkedList()
{
	ListEntry *actual, *next;	

	actual = head;
	while( actual != 0 )
	{
		next = actual->next;
		delete actual->data;
		delete actual;
		actual = next;
		lsize--;
	}
	delete[] cursortb;
}

// Creates the given number of cursors. In multi-threaded applications each
// thread can have its own cursor to step through the list and manipulate entries.
// This function is not thread safe!

template <class T>
void LinkedList<T>::initCursors( int num )
{
	if( num < 1 )
	{
		cerr << "LinkedList: Invalid num parameter\n";
		exit( -1 );
	}
	delete[] cursortb;
	cursortb = new ListEntry *[num];
	if( cursortb == 0 )
	{
		cerr << "Cannot allocate memory\n";
		exit( -1 );
	}
	memset( cursortb, 0, num * sizeof( ListEntry ** ) );
	cursors = num;
}

template <class T>
void LinkedList<T>::first( int curs = 0 )
{
	if( curs < 0 || curs >= cursors )
	{
		cerr << "LinkedList: invalid cursor\n";
		exit( -1 );
	}
	cursortb[curs] = head;
}

template <class T>
void LinkedList<T>::last( int curs = 0 )
{
	if( curs < 0 || curs >= cursors )
	{
		cerr << "LinkedList: invalid cursor\n";
		exit( -1 );
	}
	cursortb[curs] = tail;
}

template <class T>
void LinkedList<T>::next( int curs = 0 )
{
	if( curs < 0 || curs >= cursors )
	{
		cerr << "LinkedList: invalid cursor\n";
		exit( -1 );
	}
	if( cursortb[curs] == 0 ) return; // no entries
	if( cursortb[curs]->next == 0 ) cursortb[curs] = 0; // we are at the last entry
	else cursortb[curs] = cursortb[curs]->next;  // take the next entry
}

template <class T>
T *LinkedList<T>::getFirst( int curs )
{
	if( curs < 0 || curs >= cursors )
	{
		cerr << "LinkedList: invalid cursor\n";
		exit( -1 );
	}
	cursortb[curs] = head;
	if( cursortb[curs] == 0 ) return 0;
	else return cursortb[curs]->data;
}

template <class T>
T *LinkedList<T>::getNext( int curs = 0 )
{
	if( curs < 0 || curs >= cursors )
	{
		cerr << "LinkedList: invalid cursor\n";
		exit( -1 );
	}
	if( cursortb[curs] == 0 ) return 0; // no entries
	if( cursortb[curs]->next == 0 ) return 0; // we are at the last entry
	cursortb[curs] = cursortb[curs]->next;  // take the next entry
	return cursortb[curs]->data;	// return it
}

template <class T>
T *LinkedList<T>::getEntry( int curs = 0 )
{
	if( curs < 0 || curs >= cursors )
	{
		cerr << "LinkedList: invalid cursor\n";
		exit( -1 );
	}
	if( cursortb[curs] == 0 ) return 0;  // no entry
	return cursortb[curs]->data;
}

template <class T>
int LinkedList<T>::size()
{
	return lsize;
}

template <class T>
void LinkedList<T>::insertBefore( T *newt, int curs = 0 )
{
	ListEntry *p, *prev;

	if( curs < 0 || curs >= cursors )
	{
		cerr << "LinkedList: invalid cursor\n";
		exit( -1 );
	}
	p = new ListEntry;
	p->data = newt;
	if( lsize == 0 )  // if there is no entry
	{
		p->next = 0;
		p->prev = 0;  // no previous entry
		cursortb[curs] = p;
		head = p;
		tail = p;
		lsize++;
	}
	else if( cursortb[curs] == head )  // we are at the head entry
	{
		p->next = head;
		p->prev = 0;
		cursortb[curs]->prev = p;
		head = p;
		cursortb[curs] = p;
		lsize++;
	}
	else  // otherwise we are within or at the end of the list
	{
		prev = cursortb[curs]->prev;
		prev->next = p;
		cursortb[curs]->prev = p;
		p->prev = prev;
		p->next = cursortb[curs];
		cursortb[curs] = p;
		lsize++;
	}
}

template <class T>
void LinkedList<T>::insertAfter( T *newt, int curs = 0 )
{
	ListEntry *p, *next;

	if( curs < 0 || curs >= cursors )
	{
		cerr << "LinkedList: invalid cursor\n";
		exit( -1 );
	}
	p = new ListEntry;
	p->data = newt;
	if( lsize == 0 )  // if there is no entry
	{
		p->next = 0;
		p->prev = 0;  // no previous entry
		cursortb[curs] = p;
		head = p;
		tail = p;
		lsize++;
	}
	else  // otherwise
	{
		if( cursortb[curs] == 0 )  // if we are at the end of the list
		{
			cursortb[curs] = tail;
		}
		if( cursortb[curs]->next == 0 ) tail = p;
		next = cursortb[curs]->next;
		p->next = next;
		if( next != 0 ) next->prev = p;
		p->prev = cursortb[curs];
		cursortb[curs]->next = p;
		cursortb[curs] = p;
		lsize++;
	}
}

template <class T>
int LinkedList<T>::endOfList( int curs = 0 )
{
	if( curs < 0 || curs >= cursors )
	{
		cerr << "LinkedList: invalid cursor\n";
		exit( -1 );
	}
	if( cursortb[curs] == 0 ) return -1;
	else return 0;
}

template <class T>
int LinkedList<T>::isLast( int curs = 0 )
{
	if( curs < 0 || curs >= cursors )
	{
		cerr << "LinkedList: invalid cursor\n";
		exit( -1 );
	}
	if( cursortb[curs] == 0 ) return 0;
	if( cursortb[curs]->next == 0 ) return -1;
	return 0;
}

template <class T>
int LinkedList<T>::empty()
{
	if( lsize == 0 ) return -1;
	else return 0;
}

template <class T>
void LinkedList<T>::remove( int curs = 0 )
{
	int i;
	ListEntry *prev, *next, *tmp;
	
	if( curs < 0 || curs >= cursors )
	{
		cerr << "LinkedList: invalid cursor\n";
		exit( -1 );
	}
	tmp = cursortb[curs];  // store the original pointer
	if( cursortb[curs] == 0 )  // no entry to remove
	{
		return;
	}
	if( cursortb[curs]->next == 0 ) tail = cursortb[curs]->prev;
	if( cursortb[curs] == head )  // this is the head entry
	{
		head = cursortb[curs]->next;
		delete cursortb[curs]->data;
		delete cursortb[curs];
		cursortb[curs] = head;
		if( cursortb[curs] != 0 ) cursortb[curs]->prev = 0;
		lsize--;
	}
	else
	{
		prev = cursortb[curs]->prev;
		next = cursortb[curs]->next;
		delete cursortb[curs]->data;
		delete cursortb[curs];
		if( prev != 0 ) prev->next = next;
		if( next != 0 ) next->prev = prev;
		cursortb[curs] = next;
		lsize--;
	}
	
	// now update the other cursors that pointed to the removed entry
	
	for( i = 0; i < cursors; i++ )
	{
		if( cursortb[i] == tmp )
		{
			cursortb[i] = cursortb[curs];
		}
	}
}

template <class T>
void LinkedList<T>::dump( int curs = 0 )
{
	ListEntry *pt;
	
	if( curs < 0 || curs >= cursors )
	{
		cerr << "LinkedList: invalid cursor\n";
		exit( -1 );
	}
	cout << "List size: " << lsize << endl;
	pt = head;
	while( pt != 0 )
	{
		if( pt == cursortb[curs] ) cout << "*";
		else cout << " ";
		cout << *( pt->data ) << endl;
		pt = pt->next;
	}
}

#endif
