/*
 This file is part of pathload.

 pathload is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 pathload is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 al_int32 with pathload; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*-------------------------------------------------
   pathload : an end-to-end available bandwidth 
              estimation tool
   Author   : Manish Jain ( jain@cc.gatech.edu )
              Constantinos Dovrolis (dovrolis@cc.gatech.edu )
   Release  : Ver 1.3.2
   Support  : This work was supported by the SciDAC
              program of the US department 
--------------------------------------------------*/

/*
 * $header$
 */

#ifdef MGRP
#include "probe.h"
#endif

#ifdef MGRP
extern int debug_on;
#define DEBUGTS(format, ...) if (debug_on) LOGTS(format, ## __VA_ARGS__)
#else
#define DEBUGTS(format, ...) (void)(0)
#endif



#if SIZEOF_LONG == 4
  typedef long l_int32 ;
  typedef unsigned long l_uint32 ;
#elif SIZEOF_INT == 4
  typedef int l_int32 ;
  typedef unsigned int l_uint32 ;
#endif

#ifdef LOCAL
#define EXTERN
#else
#define EXTERN extern
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <errno.h>
#include <ctype.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <strings.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#include <math.h>
#include <float.h>
#include <fcntl.h>
#include <sys/utsname.h>
#ifdef THRLIB
#include <pthread.h>
#endif

/* Code numbers sent from receiver to sender using the TCP control stream */
#define  CTR_CODE           0x80000000
#define  SEND_FLEET         0x00000001
#define  RECV_FLEET         0x00000002
#define  CONTINUE_STREAM    0x00000003
#define  FINISHED_STREAM    0x00000007
#define  TERMINATE          0x00000005
#define  ABORT_FLEET        0x00000006
#define  SEND_TRAIN         0x00000008 
#define  FINISHED_TRAIN     0x00000009
#define  BAD_TRAIN          0x0000000a
#define  GOOD_TRAIN         0x0000000b
#define  PIGGYBACK_ALLOWED  0x0000000c
#define  PIGGYBACK_NOTUSED  0x0000000d

/* Port numbers (UDP for receiver, TCP for sender) */
#define UDPRCV_PORT         55001
#define TCPSND_PORT         55002

#define UDP_BUFFER_SZ       400000    /* bytes */
#define TREND_ARRAY_LEN     50 


#define NUM_STREAM          12
#define MAX_STREAM_LEN      400
#define STREAM_LEN          100       /* # of packets */
#ifdef MGRP
/* min packet size can be lower if overhead > 28, fe when using bhtp.
 *
 * We do this so that we have the same ip packet size when
 * using small packets (bhtp and non-bhtp case), so that
 * it is easier to compare in experiments.
 */
#define MIN_PKT_SZ          (200 - overhead + 28)
#else
#define MIN_PKT_SZ          200       /* bytes */
#endif


#define MAX_TRAIN           5
#define TRAIN_LEN           50

/* Characteristics of Packet Stream */
EXTERN l_int32 time_interval ;              /* in us */
EXTERN l_uint32 transmission_rate ; /* in bps */
EXTERN l_int32 cur_pkt_sz ;                 /* in bytes */
EXTERN l_int32 max_pkt_sz ;                 /* in bytes */
EXTERN l_int32 rcv_max_pkt_sz ;             /* in bytes */
EXTERN l_int32 snd_max_pkt_sz ;             /* in bytes */
EXTERN l_int32 num_stream ;
EXTERN l_int32 stream_len ;                 /* in packets */

EXTERN l_int32 verbose ;
EXTERN l_int32 Verbose ;
EXTERN l_int32 VVerbose ;

/* modification not protected by MGRP defines */
EXTERN int overhead;

EXTERN int pathload_probe_bytes;
EXTERN int pathload_probe_pkts;

#ifdef MGRP
extern int adjust;
extern int tcp_port;
extern int udp_port;
extern int use_kernel_probes;
extern int allow_piggyback;
extern int faster_pathload;

// used only by the receiver to adjust estimates
extern long long last_stream_adj_bytes;
extern long long last_stream_adj_duration_nsec;
extern long long last_fleet_adj_bytes;
extern long long last_fleet_adj_duration_nsec;

static inline __u16
get_tcp_port()
{
	__u16 port = tcp_port;
	return port;
}

static inline __u16
get_udp_port()
{
	__u16 port = udp_port;
	return port;
}

static inline int
get_udp_ip_overhead(int sock)
{
	int overhead = 20;	// IP
	int active = udpex_socket(sock);
	FILE *fp = stdout;

	if (active)
		overhead += sizeof(struct bhtphdr) +	// MGRP
			    sizeof(struct udpexhdr);	// UDPex
	else
		overhead += 8;	// UDP


	fprintf(fp,
		"MGRP is %s for socket %d, ip/udp overhead %d bytes\n",
		(active ? "ON" : "OFF"),
		sock,
		overhead);

	return overhead;
}

static inline void
set_payload_int(char *buf, int index, l_int32 val)
{
	memcpy(buf+index*sizeof(l_int32), &val, sizeof(l_int32));
}

#ifdef MGRP
#include "libestimator.h"
#include "libpathload.h"

extern int adjust;
extern int send_estimator_info;
extern int use_estimator_rpc;
extern int estimator_rpc_id;
extern struct est_estimator estimator;

extern char local_hostid[EST_INFO_IDLEN];
extern char remote_hostid[EST_INFO_IDLEN];

// based on experimental results:
// http://macdonald.cs.umd.edu:8003/tcp4x/compare/pathload-adjust.html
#define DEFAULT_PATHLOAD_ADJUST_TYPE    PATHLOAD_ADJUST_STREAM

static inline void
pathload_mgrp_help(int is_sender)
{
        LOG("");
        LOG("MGRP OPTIONS");
        LOG(" -x                Turns on MGRP (UDPex socket instead of UDP)");
        if (is_sender) {
        LOG("                   Also turns on kernel probes (barriers)");
        LOG(" -p                Turns on piggybacking");
        }
        LOG(" -a <adjust>       Adjust measurement based on piggybacking");
        LOG("                   Valid adjust policies: 'none', 'stream', 'fleet'");
        LOG("                   Default policy: '%s'",
                                pathload_get_adjust_policy(DEFAULT_PATHLOAD_ADJUST_TYPE));
        LOG("");
        LOG("ESTIMATOR OPTIONS");
        LOG(" -P <host>[:port]  Estimator address (needs one of -I or -F)");
        LOG("");
        LOG(" -I <rpc-ID>       Use estimator RPC to send results. Use this mode when");
        LOG("                   the estimator initiates the pathload session.");
        if (!is_sender) {
        LOG(" -F                Send INFO/availbw to the estimator with the results.");
        LOG("                   Use when invoking pathload directly.");
        }
        LOG("");
        LOG("OTHER OPTIONS");
        LOG(" -T <port>         TCP control port (default %d)", tcp_port);
        LOG(" -U <port>         UDP probe port (default %d)", udp_port);
        LOG(" -d                Turn on debug statments.");
        if (is_sender) {
        LOG(" -f                Faster pathload. [sender only]");
        LOG("                   No gap between trains but has higher probing overhead");
        }
}

static inline void
pathload_estimator_init()
{
  est_get_host_id(local_hostid, sizeof(local_hostid));
  snprintf(remote_hostid, sizeof(remote_hostid), "UNKNOWN");

  LOG("local_hostid %s", local_hostid);

  LOG("faster_pathload %d", faster_pathload);
  LOG("allow_piggyback %d", allow_piggyback);
  LOG("use_estimator_rpc %d", use_estimator_rpc);
  LOG("send_estimator_info %d", send_estimator_info);
  if (estimator.valid) {
    int rc = 0;
    LOG("openning channel with estimator %s", estimator.addrstr);
    rc = est_open_channel(&estimator.ch, &estimator.addr);
    BUG_ON(rc);
  }
}

static inline void
pathload_estimator_done()
{
        if (estimator.valid) {
                LOG("closing channel with estimator %s", estimator.addrstr);
                est_close_channel(&estimator.ch);
        }
}

static inline void
rpc_status_results(int status, struct est_results *results)
{
        int rc = 0;
        if (!use_estimator_rpc)
                return;
        rc = est_send_probe_status(&estimator.ch, estimator_rpc_id,
                                   status, results);
        BUG_ON(rc);
}

static inline void
rpc_status(int status)
{
        rpc_status_results(status, NULL);
}

static inline void
abort_error(int status)
{
        fprintf(stderr, "++++++++ stopping pathload, status %d, rpcID %d\n", status, estimator_rpc_id);
        rpc_status(status);
        pathload_estimator_done();
        fprintf(stderr, "++++++++ stopped pathload, status %d, rpcID %d\n", status, estimator_rpc_id);
        exit(-1);
}

static inline void
abort_noports()
{
        abort_error(EST_PROBE_STATUS_NOPORTS);
}


static inline void
abort_failed()
{
        abort_error(EST_PROBE_STATUS_FAILED);
}

static inline void
abort_alarm()
{
        abort_error(EST_PROBE_STATUS_ALARM);
}

static inline void
abort_terminated()
{
        abort_error(EST_PROBE_STATUS_TERMINATED);
}
#else

static inline void
abort_noports()
{
        exit(-1);
}


static inline void
abort_failed()
{
        exit(-1);
}

static inline void
abort_alarm()
{
        exit(0);
}

static inline void
abort_terminated()
{
        exit(0);
}
#endif

static inline void
log_reset_probe_stats(long long usec)
{
        double mbps = 0;
        
        if (usec > 0)
                mbps = (8.0 * pathload_probe_bytes) / usec;
        printf(">>> bytes %d pkts %d mbps %.2f\n",
               pathload_probe_bytes,
               pathload_probe_pkts,
               mbps);

        pathload_probe_bytes = 0;
        pathload_probe_pkts = 0;
}

// XXX hack so that we can send/recv the hostid as a ctr_code
union hostid_hack {
        char hostid_str[4];
        l_int32 hostid_code;
};

#endif
