/*
 This file is part of pathload.

 pathload is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 pathload is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with pathload; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*-------------------------------------------------
   pathload : an end-to-end available bandwidth 
              estimation tool
   Author   : Manish Jain (jain@cc.gatech.edu)
              Constantinos Dovrolis (dovrolis@cc.gatech.edu)
   Release  : Ver 1.3.2
   Support  : This work was supported by the SciDAC
              program of the US department 
--------------------------------------------------*/
#define LOCAL
#include "pathload_gbls.h"
#include "pathload_snd.h"

int pathload_probe_bytes = 0;
int pathload_probe_pkts = 0;

#ifdef MGRP
int adjust = DEFAULT_PATHLOAD_ADJUST_TYPE;
int tcp_port = TCPSND_PORT;
int udp_port = UDPRCV_PORT;
int use_kernel_probes = 0;
// only matters when using mgrp
int allow_piggyback = 0;
int send_estimator_info = 0;
int use_estimator_rpc = 0;
int estimator_rpc_id = 0;
struct est_estimator estimator = { 0 };

char local_hostid[EST_INFO_IDLEN] = { 0 };
char remote_hostid[EST_INFO_IDLEN] = { 0 };

int faster_pathload = 0;
int debug_on = 0;
#endif

// XXX hack so that we can send/recv the hostid as a ctr_code
void
send_hostid(char *hostid, int len)
{
    char ctr_buff[8];
    union hostid_hack hack;

    // truncates hostid to 4 bytes
    strncpy(hack.hostid_str, hostid, sizeof(hack.hostid_str));
    hack.hostid_str[sizeof(hack.hostid_str)-1] = 0;
    send_ctr_mesg(ctr_buff, hack.hostid_code);

    LOG(">>> sent hostid %s (original %s)", hack.hostid_str, hostid);
}

void
recv_hostid(char *hostid, int len)
{
    char ctr_buff[8];
    union hostid_hack hack;

    while ((hack.hostid_code = recv_ctr_mesg(ctr_buff)) == -1);

    strncpy(hostid, hack.hostid_str, len);
    hostid[len-1] = 0;

    LOG(">>> received hostid %s (stored %s)", hack.hostid_str, hostid);
}

int main(int argc, char* argv[])
{
  struct hostent *host_rcv;
  struct timeval tv1,tv2;
  l_uint32 snd_time ;
  l_int32 ctr_code ;
  time_t localtm;
#ifdef MGRP
  int found;
  int mss;
  unsigned int opt_len;
#else
  int opt_len,mss;
#endif
  int ret_val ;
  int iterate=0;
  int done=0;
  int latency[30],ord_latency[30];
  int i;
  int c ;
  int errflg=0;
  char pkt_buf[256];
  char ctr_buff[8];

  quiet=0;
#ifdef MGRP
  while ((c = getopt(argc, argv, "a:ihHqxpT:U:P:I:fd")) != EOF)
#else
  while ((c = getopt(argc, argv, "ihHq")) != EOF)
#endif
    switch (c) 
    {
      case 'H':
      case 'h':
        help() ;
        break ;
      case 'i':
        iterate=1;
        break;
      case 'q':
        quiet=1;
        break;
#ifdef MGRP
      case 'T':
	found = sscanf(optarg, "%u", &tcp_port);
	if (found != 1) {
		fprintf(stderr, "invalid TCP send port '%s'", optarg);
		abort_failed();
	}
        break ;
      case 'U':
	found = sscanf(optarg, "%u", &udp_port);
	if (found != 1) {
		fprintf(stderr, "invalid UDP recv port '%s'", optarg);
		abort_failed();
	}
        break ;
      case 'a':
        adjust = pathload_get_adjust_type(optarg);
        break;
      case 'x':
        use_kernel_probes=1;
        break;
      case 'p':
        allow_piggyback = 1;
        break ;

      // Estimator MODES
      //
      // no -P, -I, -F               default, no interaction with estimators
      //
      // -P<estimator> -I<rpc-ID>    Use estimator RPC to send results.
      //                             This mode is used when the estimator
      //                             initiates the pathload session.
      //
      //                             For example, when avbw2.c connects to
      //                             the estimator to request probing.
      // 
    
      // read an estimator address
      case 'P':
        est_parse_estimator(optarg, &estimator);
        break;
      // estimator_rpc
      case 'I':
        if (!estimator.valid) {
                fprintf(stderr, "error: option -P must precede -I\n");
                abort_failed();
        }
        found = sscanf(optarg, "%u", &estimator_rpc_id);
        use_estimator_rpc = 1;
        break;
      case 'f':
        faster_pathload = 1;
        break;

      case 'd':
        debug_on = 1;
        break;
#endif

      case '?':
        errflg++;
    }
  if (errflg)
  {
    fprintf(stderr, "usage: pathload_snd [-q] [-H|-h]\n");
    abort_failed();
  }

#ifdef MGRP
  pathload_estimator_init();
#endif

  num_stream = NUM_STREAM ;
  min_sleeptime();

  /* gettimeofday latency */
#if defined(MGRP) && defined(DEBUG)
  printf("DEBUG :: calculating gettimeofday latency\n");
#endif
  for(i=0;i<30;i++)
  {
    gettimeofday(&tv1,NULL);
    gettimeofday(&tv2,NULL);
    latency[i]=tv2.tv_sec*1000000+tv2.tv_usec-tv1.tv_sec*1000000-tv1.tv_usec;
  }
  order_int(latency,ord_latency,30);
  gettimeofday_latency = ord_latency[15];  
#ifdef DEBUG
  printf("DEBUG :: gettimeofday_latency = %d\n",gettimeofday_latency);
#endif
  /* Control stream: TCP connection */
  if ((sock_tcp=socket(AF_INET, SOCK_STREAM, 0)) < 0)
  {
    perror("socket(AF_INET,SOCK_STREAM,0):");
    abort_failed();
  }
#if defined(MGRP) && defined(DEBUG)
  printf("DEBUG :: open TCP socket(%d), listen on %d\n",
  	sock_tcp, get_tcp_port());
#endif
  opt_len=1;
  if (setsockopt(sock_tcp, SOL_SOCKET, SO_REUSEADDR, (char*)&opt_len, 
        sizeof(opt_len)) < 0)
  {
    perror("setsockopt(SOL_SOCKET,SO_REUSEADDR):");
    abort_failed();
  }
  bzero((char*)&snd_tcp_addr, sizeof(snd_tcp_addr));
  snd_tcp_addr.sin_family         = AF_INET;
  snd_tcp_addr.sin_addr.s_addr    = htonl(INADDR_ANY);
#ifdef MGRP
  snd_tcp_addr.sin_port           = htons(get_tcp_port());
#else
  snd_tcp_addr.sin_port           = htons(TCPSND_PORT);
#endif
  if (bind(sock_tcp, (struct sockaddr*)&snd_tcp_addr,sizeof(snd_tcp_addr)) < 0)
  {
    perror("bind(sock_tcp):");
    abort_noports();
  }
  if (listen(sock_tcp,1) < 0)
  {
    perror("listen(sock_tcp,1):");
    abort_failed();
  }
  /* Data stream: UDP socket */
#ifdef MGRP
  if (use_kernel_probes) {
	  printf("sender: opening a MGRP socket explicitly\n");
	  if ((sock_udp = socket(AF_INET, SOCK_DGRAM, IPPROTO_MGRP)) < 0)
	  {
	    perror("socket(AF_INET,SOCK_DGRAM,IPPROTO_MGRP):");
	    abort_failed();
	  }
  } else
#endif
  if ((sock_udp=socket(AF_INET, SOCK_DGRAM, 0)) < 0)
  {
    perror("socket(AF_INET,SOCK_DGRAM,0):");
    abort_failed();
  }
#if defined(MGRP) && defined(DEBUG)
  printf("DEBUG :: open UDP socket(%d)\n", sock_udp);
#endif
#ifdef MGRP
  /* ignore the return value; just us it to print it on stdout */
  get_udp_ip_overhead(sock_udp);
#endif
  bzero((char*)&snd_udp_addr, sizeof(snd_udp_addr));
  snd_udp_addr.sin_family         = AF_INET;
  snd_udp_addr.sin_addr.s_addr    = htonl(INADDR_ANY);
  snd_udp_addr.sin_port           = htons(0);
  if (bind(sock_udp, (struct sockaddr*)&snd_udp_addr, 
        sizeof(snd_udp_addr)) < 0)
  {
    perror("bind(sock_udp):");
    abort_failed();
  }
  send_buff_sz = UDP_BUFFER_SZ;
  if (setsockopt(sock_udp, SOL_SOCKET, SO_SNDBUF, (char*)&send_buff_sz, 
        sizeof(send_buff_sz)) < 0)
  {
    send_buff_sz/=2;
    if (setsockopt(sock_udp, SOL_SOCKET, SO_SNDBUF, (char*)&send_buff_sz, 
        sizeof(send_buff_sz)) < 0)
    {
      perror("setsockopt(SOL_SOCKET,SO_SNDBUF):");
      abort_failed();
    }
  }
#ifdef MGRP
  rpc_status(EST_PROBE_STATUS_SUCCESS);
#endif
  do
  {
    struct timeval start_time;
    struct timeval end_time;

    gettimeofday(&start_time, NULL);
    if ( !quiet)
      printf("\n\nWaiting for receiver to establish control stream => ");
    fflush(stdout);
    /* 
      Wait until receiver attempts to connect, 
      starting new measurement cycle
    */
    rcv_tcp_adrlen = sizeof(rcv_tcp_addr);
    ctr_strm = accept(sock_tcp, (struct sockaddr*)&rcv_tcp_addr, &rcv_tcp_adrlen);
    if (ctr_strm < 0)
    {
      perror("accept(sock_tcp):");
      abort_failed();
    }  
    if ( !quiet)
      printf("OK\n");
    localtm = time(NULL); 
    gethostname(pkt_buf, 256);
    host_rcv=gethostbyaddr((char*)&(rcv_tcp_addr.sin_addr), sizeof(rcv_tcp_addr.sin_addr), AF_INET);
    if (host_rcv!=NULL)
    {
      if ( !quiet)
        printf("Receiver %s starts measurements on %s", host_rcv->h_name, ctime(&localtm));
    }
    else
      if ( !quiet)
        printf("Unknown receiver starts measurements at %s", ctime(&localtm));

    /* Form receiving UDP address */
    bzero((char*)&rcv_udp_addr, sizeof(rcv_udp_addr));
    rcv_udp_addr.sin_family         = AF_INET;
    rcv_udp_addr.sin_addr.s_addr    = rcv_tcp_addr.sin_addr.s_addr;
#ifdef MGRP
    rcv_udp_addr.sin_port           = htons(get_udp_port());
#else
    rcv_udp_addr.sin_port           = htons(UDPRCV_PORT);
#endif
    /* Connect UDP socket */
    connect(sock_udp , (struct sockaddr *)&rcv_udp_addr , sizeof(rcv_udp_addr)); 
    /* Make TCP socket non-blocking */
    if (fcntl(ctr_strm, F_SETFL, O_NONBLOCK)<0)
    {
      perror("fcntl(ctr_strm, F_SETFL, O_NONBLOCK):");
      abort_failed();
    }
    opt_len = sizeof(mss) ;
    if (getsockopt(ctr_strm, IPPROTO_TCP, TCP_MAXSEG, (char*)&mss, &opt_len)<0)
    {
      perror("getsockopt(sock_tcp,IPPROTO_TCP,TCP_MAXSEG):");
      abort_failed();
    }
    snd_max_pkt_sz = mss;
    if (snd_max_pkt_sz == 0 || snd_max_pkt_sz== 1448) 
      snd_max_pkt_sz = 1472;   /* Make it Ethernet sized MTU */
    else
      snd_max_pkt_sz = mss+12;

    /* exchange hostids */
    send_hostid(local_hostid, sizeof(local_hostid));
    recv_hostid(remote_hostid, sizeof(remote_hostid));

    /* tell receiver if we allow piggybacking */
    if (allow_piggyback)
            send_ctr_mesg(ctr_buff, CTR_CODE | PIGGYBACK_ALLOWED);
    else
            send_ctr_mesg(ctr_buff, CTR_CODE | PIGGYBACK_NOTUSED);
    
    /* tell receiver our max packet sz */
    send_ctr_mesg(ctr_buff, snd_max_pkt_sz) ;

    /* receiver's maxp packet size */
    while ((rcv_max_pkt_sz = recv_ctr_mesg( ctr_buff)) == -1);
    max_pkt_sz = (rcv_max_pkt_sz < snd_max_pkt_sz) ? rcv_max_pkt_sz:snd_max_pkt_sz ;
    if ( !quiet )
      printf("Maximum packet size          :: %ld bytes\n",max_pkt_sz);
    /* tell receiver our send latency */
    snd_time = (l_int32) send_latency();
    send_ctr_mesg(ctr_buff, snd_time) ;

#if defined(MGRP) && defined(DEBUG)
  printf("DEBUG :: wait for receiver to start ADR measurement\n");
#endif
    /* wait for receiver to start ADR measurement */
    if((ret_val=recv_ctr_mesg(ctr_buff)) == -1 )break;
    if ( (((ret_val & CTR_CODE) >> 31) == 1) && ((ret_val & 0x7fffffff) == SEND_TRAIN ) ) 
    {
      if ( !quiet)
        printf("Estimating ADR to initialize rate adjustment algorithm => ");
      fflush(stdout);
#ifdef MGRP
      // send a marker to indicate the begining of a group of fleet_stats
      pathload_send_fleet_stats_marker("FLEET_PREAMBLE");
#endif
      if ( send_train() == -1 ) continue ;
      if ( !quiet)
        printf("Done\n");
    }
    fleet_id=0;
    done=0;
#if defined(MGRP) && defined(DEBUG)
  printf("DEBUG ::  Start avail-bw measurement\n");
#endif
#ifdef MGRP
    // send a marker to indicate the begining of a group of fleet_stats
    pathload_send_fleet_stats_marker("FLEET_BEGIN");
#endif
    /* Start avail-bw measurement */
    while(!done)
    {
      if (( ret_val  = recv_ctr_mesg ( ctr_buff ) ) == -1 ) break ;
      if((((ret_val & CTR_CODE) >> 31) == 1) &&((ret_val&0x7fffffff) == TERMINATE)) 
      {
        if ( !quiet)
          printf("Terminating current run.\n");
        done=1;
      }
      else
      {
        transmission_rate = ret_val ;
        if ((cur_pkt_sz = recv_ctr_mesg( ctr_buff)) <= 0 )break;
        if ((stream_len = recv_ctr_mesg( ctr_buff))  <= 0 )break;
        if ((time_interval = recv_ctr_mesg( ctr_buff)) <= 0 )break;
        if ((ret_val = recv_ctr_mesg ( ctr_buff )) == -1 )break;
        /* ret_val = SENd_FLEET */
        ctr_code = RECV_FLEET | CTR_CODE ;
        if ( send_ctr_mesg(ctr_buff,  ctr_code  ) == -1 ) break;
        if(send_fleet()==-1) break ;
        if ( !quiet) printf("\n");
        fleet_id++ ;
      }
    }
#ifdef MGRP
    // send a marker to indicate the end of a group of fleet_stats
    // the purpose is to store in the stats-pktrain datafile (for plotting)
    pathload_send_fleet_stats_marker("FLEET_END");
#endif
    close(ctr_strm);
    gettimeofday(&end_time, NULL);
    log_reset_probe_stats(time_to_us_delta(start_time, end_time));
  }while (iterate);
  
#ifdef MGRP
  pathload_estimator_done();
#endif
  return 0;
}

