#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <tcpdump-stdinc.h>

#include <stdio.h>
#include <string.h>

#include "interface.h"
#include "addrtoname.h"
#include "extract.h"			/* must come after interface.h */
#include "ip.h"
#include "ipproto.h"

#include "mgrp/bhtp.h"
#include "mgrp/udpex.h"
#include "mgrp/frag.h"
#include "pktrain.h"

#include "tcp.h"


struct bhtp_state bhtp = {0};


/* HACK: copied from linux/time.h */
/* 	 is this in a user-space include file? */

#define NSEC_PER_SEC	1000000000L
#define NSEC_PER_USEC	1000L

static inline __s64 timeval_to_ns(const struct timeval *tv)
{                                    
	return ((__s64) tv->tv_sec * NSEC_PER_SEC) +       
		tv->tv_usec * NSEC_PER_USEC;                            
}                                                 

static inline void
mgrp_frag_print(const u_char *data, int len, const u_char *data2, int fragmented)
{
        const struct mgrp_fraghdr *hdr = (struct mgrp_fraghdr *)data;

        int next_proto = hdr->proto;
        const u_char *next_proto_data = data + sizeof(struct mgrp_fraghdr);
        int next_proto_len = len - sizeof(struct mgrp_fraghdr);

        if (len < sizeof(struct mgrp_fraghdr)) {
                printf("frag:error");
                return;
        }

        printf("frag len:%-4d proto:%d flags:%s%s%s id:%d off:%-4d",
               len,
               next_proto,
               (hdr->duplicate       ? "D" : " "),
               (hdr->fragmented      ? "F" : " "),
               (hdr->more_fragments  ? "M" : " "),
               hdr->fragid,
               hdr->fragoff);

        // for the first TCP fragment of a TCP packet
        if (next_proto == BHTP_PROTO_TCP && hdr->fragoff == 0) {
                printf(" TCP ");
                tcp_print(next_proto_data, next_proto_len, data2, fragmented);
        }
}

static inline int
find_source_index(__u32 src)
{
	int i = 0;
	for (i = 0; i < bhtp.sources; i++)
		if (bhtp.stats[i].addr == src)
			return i;

	return -1;
}

static inline int
new_source_index(__u32 src)
{
	int i = bhtp.sources;
	if (bhtp.sources == BHTP_MAX_SOURCES)
		return -1;

	memset(&bhtp.stats[i], 0, sizeof(struct bhtp_source_stats));
	bhtp.stats[i].addr = src;
	bhtp.sources++;
	return i;
}

static inline unsigned long long
tstamp_diff(const struct bhtphdr *bh, const struct ip *iph,
		unsigned long long *tspkt_diff)
{
	__u32 src = iph->ip_src.s_addr;
	int i = find_source_index(src);
	struct bhtp_source_stats *stats = NULL;
        unsigned long long sndstamp = bhtp_get_sndstamp_ns(bh);

	/* diff in nanosec of source timestamp inside bhtp header */
	unsigned long long diff = 0;

	/* diff in nanosec of tcpdump timestamp of current packet */
	unsigned long long tsdiff = 0;

	if (i < 0)
		i = new_source_index(src);

	if (i < 0) {
		printf("[tstamp error] ");
		return 0;
	}
		
	stats = &bhtp.stats[i];

	/* tcpdump timestamp in global @tspkt */
	if (stats->sndstamp_last > 0) {
		diff = sndstamp - stats->sndstamp_last;
		tsdiff = timeval_to_ns(&bhtp.tspkt) -
			 timeval_to_ns(&stats->tspkt_last);
	}

	stats->sndstamp_last = sndstamp;
	stats->tspkt_last = bhtp.tspkt;

	if (tspkt_diff != NULL)
		*tspkt_diff = tsdiff;

	return diff;
}

void
pktrain_print(const u_char *pd)
{
	printf("pktrain ");
	pktrain_print_pd((struct pktrain_payload *)pd);
}

static inline void
bhtp_print_tunproto(char *proto, int len)
{
	if (bhtp.op.stats)
		return;
	if (bhtp.op.compact)
		printf("%s:%d ", proto, len);
	else
		printf("[%s]", proto);
}

/* dead code: counts tcp packets with certain flags; does not work */
static inline void
bhtp_detect_riders(const u_char *hdr)
{
	struct tcphdr *tcp = (struct tcphdr *)hdr;

	if (!tcp)
		return;
	
	if (tcp->th_flags & TH_FIN)
		bhtp.tcp.fin++;	

	if (tcp->th_flags & (TH_SYN|TH_ACK))
		bhtp.tcp.synack++;	
	else if (tcp->th_flags & TH_SYN)
		bhtp.tcp.syn++;	

	if (tcp->th_flags & TH_RST)
		bhtp.tcp.rst++;	

}

/**
 * bhtp_print - show bhtp packet
 * @bp - beginning of bhtp packet
 * @data2 - beginning of enclosing 
 * @len - lenght of ip packet
 */
void
bhtp_print(const u_char *bp, u_int pktlen,
	   const u_char *data2, int fragmented)
{
	register const struct bhtphdr *bh = (const struct bhtphdr *)bp;
	int t = 0;
	unsigned long long bhdiff = 0;
	unsigned long long tsdiff = 0;
        unsigned long long seqno = 0;

	int tuns = 0;

	char tmp[100];
	char *str = tmp;
	int strlen = 0;

	char *stun_tcp = "";
	char *stun_udp = "";
	char *stun_pad = "";
	char *stun_unknown = "";

        int probed = 0;

        int pad_orig = 0;
        int pad_reused = 0;
        int pad_unused = 0;
        char error[200];

        int pkt_is_probe = 0;

        // This BHTP packet looks awfully like a TCP packet
        // (due to an ongoing, unresolved MGRP bug)
        if (bhtp.op.tsniff && (pktlen == 20 || pktlen == 32)) {
                printf("[bhtperror: looks like tcp] ");
                tcp_print(bp, pktlen, data2, fragmented);
                return;
        }

        // sanity check, did we get enough of the header?
	if (pktlen < sizeof(struct bhtphdr)) {
		printf("[|bhtperror got %d needs %d]",
                       pktlen, sizeof(struct bhtphdr));
		return;
        }

        // deconstruct the padding into @pad_orig (== @pad_reused + @pad_unused)
        pad_orig = bhtp_orig_pad_bytes(bh, pktlen,
                                       &pad_reused, &pad_unused,
                                       error, sizeof(error));
        if (pad_orig == -1) {
		printf("[|bhtperror %s]", error);
		return;
        }
        pkt_is_probe = (pad_orig > 0);

	bhdiff = tstamp_diff(bh, (struct ip *)data2, &tsdiff);

	tuns = bhtp_header_tuns(bh);
        seqno = bh->seqno;

	printf("bhtp len:%-4d seq:%-4lld tuns:%d ",
		pktlen,
		seqno,
		tuns);

	if (bhtp.op.tstamp) {
		printf("ts:%llu ", bhtp_get_sndstamp_ns(bh));
	}
	printf("sd:%-10llu ", bhdiff);
	printf("td:%-10llu", tsdiff);

	for (t = 0; t < tuns; t++) {

		const struct bhtphdr_tun *tun = &bh->tun[t];
		const unsigned char *tdata = bhtp_get_tun_chunk(bh, t, bp);

		int tunlen = tun->len;
		int proto = tun->proto;

		if (tun->needs_padding)
			probed += tunlen;

		if (!bhtp.op.compact && !bhtp.op.stats)
			printf(" T%d> prot:%d len:%-4d pad:%d dup:%d ",
				t, proto, tunlen,
                                (tun->needs_padding ? pad_orig : 0),
                                (tun->duplicate ? 1 : 0));

		switch (proto) {
		
		case BHTP_PROTO_TCP:
			/* shouldn't we pass length as well? */
			/* bhtp_detect_riders(tdata); */
			stun_tcp = "tcp:";
			if (bhtp.op.tun || bhtp.op.tcp)
				tcp_print(tdata, tunlen, data2, fragmented);
			else
				bhtp_print_tunproto("tcp", tunlen);
			break;

		case BHTP_PROTO_UDP:
			stun_udp = "udp:";
			if (bhtp.op.tun)
				udp_print(tdata, tunlen, data2, fragmented);
			else
				bhtp_print_tunproto("udp", tunlen);
			break;

                case MGRP_PROTO_FRAG:
                        stun_udp = "frag";
                        if (bhtp.op.tun)
                                mgrp_frag_print(tdata, tunlen, data2, fragmented);
                        else
                                bhtp_print_tunproto("frag", tunlen);
                        break;

		default:
			stun_unknown = "UNKNOWN:";
			printf("[unknown]");
			break;

		}
	}

	if (pkt_is_probe)
		printf(" probed:%-4d reused:%-4d padded:%-4d",
			probed, pad_reused, pad_unused);

	if (bhtp.op.stun)
		printf(" mgrp:%s%s%s%s",
			stun_udp,
			stun_tcp,
			stun_pad,
			stun_unknown);


}

